/* Generated by Cython 3.0.7 */

#ifndef __PYX_HAVE__asynctnt__iproto__protocol
#define __PYX_HAVE__asynctnt__iproto__protocol

#include "Python.h"
struct C_Metadata;

/* "asynctnt/iproto/schema.pxd":14
 * 
 * 
 * cdef public class Metadata [object C_Metadata, type C_Metadata_Type]:             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly list fields
 */
struct C_Metadata {
  PyObject_HEAD
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Metadata *__pyx_vtab;
  PyObject *fields;
  PyObject *name_id_map;
  PyObject *names;
};

#ifndef __PYX_HAVE_API__asynctnt__iproto__protocol

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C DL_IMPORT(PyTypeObject) C_Metadata_Type;

#endif /* !__PYX_HAVE_API__asynctnt__iproto__protocol */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initprotocol(void);
#else
/* WARNING: Use PyImport_AppendInittab("protocol", PyInit_protocol) instead of calling PyInit_protocol directly from Python 3.5 */
PyMODINIT_FUNC PyInit_protocol(void);

#if PY_VERSION_HEX >= 0x03050000 && (defined(__GNUC__) || defined(__clang__) || defined(_MSC_VER) || (defined(__cplusplus) && __cplusplus >= 201402L))
#if defined(__cplusplus) && __cplusplus >= 201402L
[[deprecated("Use PyImport_AppendInittab(\"protocol\", PyInit_protocol) instead of calling PyInit_protocol directly.")]] inline
#elif defined(__GNUC__) || defined(__clang__)
__attribute__ ((__deprecated__("Use PyImport_AppendInittab(\"protocol\", PyInit_protocol) instead of calling PyInit_protocol directly."), __unused__)) __inline__
#elif defined(_MSC_VER)
__declspec(deprecated("Use PyImport_AppendInittab(\"protocol\", PyInit_protocol) instead of calling PyInit_protocol directly.")) __inline
#endif
static PyObject* __PYX_WARN_IF_PyInit_protocol_INIT_CALLED(PyObject* res) {
  return res;
}
#define PyInit_protocol() __PYX_WARN_IF_PyInit_protocol_INIT_CALLED(PyInit_protocol())
#endif
#endif

#endif /* !__PYX_HAVE__asynctnt__iproto__protocol */
