/* Generated by Cython 3.0.7 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "asynctnt/iproto/tupleobj/tupleobj.h",
            "third_party/msgpuck/msgpuck.h",
            "third_party/xd.h"
        ],
        "include_dirs": [
            "asynctnt/iproto",
            "third_party"
        ],
        "name": "asynctnt.iproto.protocol",
        "sources": [
            "asynctnt/iproto/protocol.pyx",
            "third_party/msgpuck/msgpuck.c",
            "third_party/msgpuck/hints.c",
            "asynctnt/iproto/tupleobj/tupleobj.c"
        ]
    },
    "module_name": "asynctnt.iproto.protocol"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif
    

    #if (PY_VERSION_HEX < 0x030700b1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600)) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_7" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x030007F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info; // borrowed
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11; // we don't yet need to distinguish between versions > 11
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);  // we don't have access to __pyx_empty_bytes here
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__asynctnt__iproto__protocol
#define __PYX_HAVE_API__asynctnt__iproto__protocol
/* Early includes */
#include <string.h>
#include <stdio.h>
#include <stddef.h>
#include "pythread.h"
#include "tupleobj/tupleobj.h"
#include <stdint.h>
#include "../../third_party/msgpuck/msgpuck.h"
#include "../../third_party/xd.h"
#include "datetime.h"

    /* Backport for Python 2.x */
    #if PY_MAJOR_VERSION < 3
        #ifndef PyDateTime_DELTA_GET_DAYS
            #define PyDateTime_DELTA_GET_DAYS(o) (((PyDateTime_Delta*)o)->days)
        #endif
        #ifndef PyDateTime_DELTA_GET_SECONDS
            #define PyDateTime_DELTA_GET_SECONDS(o) (((PyDateTime_Delta*)o)->seconds)
        #endif
        #ifndef PyDateTime_DELTA_GET_MICROSECONDS
            #define PyDateTime_DELTA_GET_MICROSECONDS(o) (((PyDateTime_Delta*)o)->microseconds)
        #endif
    #endif
    /* Backport for Python < 3.6 */
    #if PY_VERSION_HEX < 0x030600a4
        #ifndef PyDateTime_TIME_GET_FOLD
            #define PyDateTime_TIME_GET_FOLD(o) ((void)(o), 0)
        #endif
        #ifndef PyDateTime_DATE_GET_FOLD
            #define PyDateTime_DATE_GET_FOLD(o) ((void)(o), 0)
        #endif
    #endif
    /* Backport for Python < 3.6 */
    #if PY_VERSION_HEX < 0x030600a4
        #define __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             ((void)(fold), PyDateTimeAPI->DateTime_FromDateAndTime(year, month, day, hour, minute, second,                 microsecond, tz, PyDateTimeAPI->DateTimeType))
        #define __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             ((void)(fold), PyDateTimeAPI->Time_FromTime(hour, minute, second, microsecond, tz, PyDateTimeAPI->TimeType))
    #else /* For Python 3.6+ so that we can pass tz */
        #define __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             PyDateTimeAPI->DateTime_FromDateAndTimeAndFold(year, month, day, hour, minute, second,                 microsecond, tz, fold, PyDateTimeAPI->DateTimeType)
        #define __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             PyDateTimeAPI->Time_FromTimeAndFold(hour, minute, second, microsecond, tz, fold, PyDateTimeAPI->TimeType)
    #endif
    /* Backport for Python < 3.7 */
    #if PY_VERSION_HEX < 0x030700b1
        #define __Pyx_TimeZone_UTC NULL
        #define __Pyx_TimeZone_FromOffset(offset) ((void)(offset), (PyObject*)NULL)
        #define __Pyx_TimeZone_FromOffsetAndName(offset, name) ((void)(offset), (void)(name), (PyObject*)NULL)
    #else
        #define __Pyx_TimeZone_UTC PyDateTime_TimeZone_UTC
        #define __Pyx_TimeZone_FromOffset(offset) PyTimeZone_FromOffset(offset)
        #define __Pyx_TimeZone_FromOffsetAndName(offset, name) PyTimeZone_FromOffsetAndName(offset, name)
    #endif
    /* Backport for Python < 3.10 */
    #if PY_VERSION_HEX < 0x030a00a1
        #ifndef PyDateTime_TIME_GET_TZINFO
            #define PyDateTime_TIME_GET_TZINFO(o)                 ((((PyDateTime_Time*)o)->hastzinfo) ? ((PyDateTime_Time*)o)->tzinfo : Py_None)
        #endif
        #ifndef PyDateTime_DATE_GET_TZINFO
            #define PyDateTime_DATE_GET_TZINFO(o)                 ((((PyDateTime_DateTime*)o)->hastzinfo) ? ((PyDateTime_DateTime*)o)->tzinfo : Py_None)
        #endif
    #endif
    

    /* Backport for Python 2.x */
    #if PY_MAJOR_VERSION < 3
        #ifndef PyDateTime_DELTA_GET_DAYS
            #define PyDateTime_DELTA_GET_DAYS(o) (((PyDateTime_Delta*)o)->days)
        #endif
        #ifndef PyDateTime_DELTA_GET_SECONDS
            #define PyDateTime_DELTA_GET_SECONDS(o) (((PyDateTime_Delta*)o)->seconds)
        #endif
        #ifndef PyDateTime_DELTA_GET_MICROSECONDS
            #define PyDateTime_DELTA_GET_MICROSECONDS(o) (((PyDateTime_Delta*)o)->microseconds)
        #endif
    #endif

    /* Backport for Python < 3.6 */
    #if PY_VERSION_HEX < 0x030600a4
        #ifndef PyDateTime_TIME_GET_FOLD
            #define PyDateTime_TIME_GET_FOLD(o) ((void)(o), 0)
        #endif
        #ifndef PyDateTime_DATE_GET_FOLD
            #define PyDateTime_DATE_GET_FOLD(o) ((void)(o), 0)
        #endif
    #endif

    /* Backport for Python < 3.6 */
    #if PY_VERSION_HEX < 0x030600a4
        #define __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             ((void)(fold), PyDateTimeAPI->DateTime_FromDateAndTime(year, month, day, hour, minute, second,                 microsecond, tz, PyDateTimeAPI->DateTimeType))
        #define __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             ((void)(fold), PyDateTimeAPI->Time_FromTime(hour, minute, second, microsecond, tz, PyDateTimeAPI->TimeType))
    #else /* For Python 3.6+ so that we can pass tz */
        #define __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             PyDateTimeAPI->DateTime_FromDateAndTimeAndFold(year, month, day, hour, minute, second,                 microsecond, tz, fold, PyDateTimeAPI->DateTimeType)
        #define __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             PyDateTimeAPI->Time_FromTimeAndFold(hour, minute, second, microsecond, tz, fold, PyDateTimeAPI->TimeType)
    #endif

    /* Backport for Python < 3.7 */
    #if PY_VERSION_HEX < 0x030700b1
        #define __Pyx_TimeZone_UTC NULL
        #define __Pyx_TimeZone_FromOffsetAndName(offset, name) ((void)(offset), (void)(name), (PyObject*)NULL)
    #else
        #define __Pyx_TimeZone_UTC PyDateTime_TimeZone_UTC
        #define __Pyx_TimeZone_FromOffsetAndName(offset, name) PyTimeZone_FromOffsetAndName(offset, name)
    #endif

    /* Backport for Python < 3.10 */
    #if PY_VERSION_HEX < 0x030a00a1
        #ifndef PyDateTime_TIME_GET_TZINFO
            #define PyDateTime_TIME_GET_TZINFO(o)                 ((((PyDateTime_Time*)o)->hastzinfo) ? ((PyDateTime_Time*)o)->tzinfo : Py_None)
        #endif
        #ifndef PyDateTime_DATE_GET_TZINFO
            #define PyDateTime_DATE_GET_TZINFO(o)                 ((((PyDateTime_DateTime*)o)->hastzinfo) ? ((PyDateTime_DateTime*)o)->tzinfo : Py_None)
        #endif
    #endif
    
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const wchar_t *u)
{
    const wchar_t *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#endif
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else  // Py < 3.12
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "<stringsource>",
  "asynctnt/iproto/schema.pyx",
  "asynctnt/iproto/ext.pyx",
  "asynctnt/iproto/buffer.pyx",
  "asynctnt/iproto/requests/update.pyx",
  "asynctnt/iproto/response.pyx",
  "asynctnt/iproto/protocol.pyx",
  "contextvars.pxd",
  "datetime.pxd",
  "asynctnt/iproto/python.pxd",
  "asynctnt/iproto/unicodeutil.pyx",
  "asynctnt/iproto/schema.pxd",
  "asynctnt/iproto/rbuffer.pyx",
  "asynctnt/iproto/requests/base.pyx",
  "asynctnt/iproto/requests/call.pyx",
  "asynctnt/iproto/requests/eval.pyx",
  "asynctnt/iproto/requests/select.pyx",
  "asynctnt/iproto/requests/insert.pyx",
  "asynctnt/iproto/requests/delete.pyx",
  "asynctnt/iproto/requests/upsert.pyx",
  "asynctnt/iproto/requests/prepare.pyx",
  "asynctnt/iproto/requests/execute.pyx",
  "asynctnt/iproto/requests/id.pyx",
  "asynctnt/iproto/requests/auth.pyx",
  "asynctnt/iproto/requests/streams.pyx",
  "asynctnt/iproto/response.pxd",
  "asynctnt/iproto/db.pyx",
  "asynctnt/iproto/push.pyx",
  "asynctnt/iproto/coreproto.pyx",
  "asynctnt/iproto/requests/ping.pyx",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "asynctnt/iproto/ttuple.pyx",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_8asynctnt_6iproto_8protocol_Field;
struct C_Metadata;
struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex;
struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaDummyIndex;
struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace;
struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaDummySpace;
struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema;
struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer;
struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer;
struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest;
struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame;
struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError;
struct __pyx_obj_8asynctnt_6iproto_8protocol_Response;
struct __pyx_obj_8asynctnt_6iproto_8protocol_Db;
struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator;
struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol;
struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol;
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex;
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr;
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__;
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id;
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth;
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema;

/* "asynctnt/iproto/tarantool.pxd":1
 * cdef enum iproto_header_key:             # <<<<<<<<<<<<<<
 *     IPROTO_REQUEST_TYPE = 0x00
 *     IPROTO_SYNC = 0x01
 */
enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_header_key {
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_REQUEST_TYPE = 0x00,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SYNC = 0x01,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_REPLICA_ID = 0x02,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_LSN = 0x03,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TIMESTAMP = 0x04,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SCHEMA_VERSION = 0x05,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SERVER_VERSION = 0x06,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_GROUP_ID = 0x07,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_STREAM_ID = 0x0a
};

/* "asynctnt/iproto/tarantool.pxd":13
 * 
 * 
 * cdef enum iproto_key:             # <<<<<<<<<<<<<<
 *     IPROTO_SPACE_ID = 0x10
 *     IPROTO_INDEX_ID = 0x11
 */
enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_key {
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SPACE_ID = 0x10,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_INDEX_ID = 0x11,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_LIMIT = 0x12,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OFFSET = 0x13,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_ITERATOR = 0x14,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_INDEX_BASE = 0x15,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_KEY = 0x20,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TUPLE = 0x21,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FUNCTION_NAME = 0x22,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_USER_NAME = 0x23,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_INSTANCE_UUID = 0x24,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_CLUSTER_UUID = 0x25,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_VCLOCK = 0x26,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_EXPR = 0x27,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OPS = 0x28,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_BALLOT = 0x29,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TUPLE_META = 0x2a,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OPTIONS = 0x2b,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_DATA = 0x30,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_ERROR_24 = 0x31,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_METADATA = 0x32,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_BIND_METADATA = 0x33,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_BIND_COUNT = 0x34,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SQL_TEXT = 0x40,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SQL_BIND = 0x41,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SQL_INFO = 0x42,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_STMT_ID = 0x43,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_ERROR = 0x52,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_VERSION = 0x54,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FEATURES = 0x55,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_AUTH_TYPE = 0x5b,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TIMEOUT = 0x56,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TXN_ISOLATION = 0x59,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_CHUNK = 0x80
};

/* "asynctnt/iproto/tarantool.pxd":55
 * 
 * 
 * cdef enum iproto_metadata_key:             # <<<<<<<<<<<<<<
 *     IPROTO_FIELD_NAME = 0x00
 *     IPROTO_FIELD_TYPE = 0x01
 */
enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_metadata_key {
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_NAME = 0x00,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_TYPE = 0x01,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_COLL = 0x02,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_IS_NULLABLE = 0x03,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_IS_AUTOINCREMENT = 0x04,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_SPAN = 0x05
};

/* "asynctnt/iproto/tarantool.pxd":64
 * 
 * 
 * cdef enum iproto_sql_info_key:             # <<<<<<<<<<<<<<
 *     SQL_INFO_ROW_COUNT = 0x00
 *     SQL_INFO_AUTOINCREMENT_IDS = 0x01
 */
enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_sql_info_key {
  __pyx_e_8asynctnt_6iproto_9tarantool_SQL_INFO_ROW_COUNT = 0x00,
  __pyx_e_8asynctnt_6iproto_9tarantool_SQL_INFO_AUTOINCREMENT_IDS = 0x01
};

/* "asynctnt/iproto/tarantool.pxd":69
 * 
 * 
 * cdef enum iproto_type:             # <<<<<<<<<<<<<<
 *     IPROTO_SELECT = 0x01
 *     IPROTO_INSERT = 0x02
 */
enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type {
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SELECT = 0x01,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_INSERT = 0x02,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_REPLACE = 0x03,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_UPDATE = 0x04,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_DELETE = 0x05,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_CALL_16 = 0x06,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_AUTH = 0x07,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_EVAL = 0x08,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_UPSERT = 0x09,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_CALL = 0x0a,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_EXECUTE = 0x0b,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_PREPARE = 0x0d,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_BEGIN = 0x0e,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_COMMIT = 0x0f,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_ROLLBACK = 0x10,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_PING = 0x40,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_ID = 0x49
};

/* "asynctnt/iproto/tarantool.pxd":89
 * 
 * 
 * cdef enum iproto_update_operation:             # <<<<<<<<<<<<<<
 *     IPROTO_OP_ADD = 43  # ord('+')
 *     IPROTO_OP_SUB = 45  # ord('-')
 */
enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_update_operation {
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_ADD = 43,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_SUB = 45,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_AND = 38,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_XOR = 94,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_OR = 0x7C,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_DELETE = 35,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_INSERT = 33,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_ASSIGN = 61,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_SPLICE = 58
};

/* "asynctnt/iproto/tarantool.pxd":101
 * 
 * 
 * cdef enum mp_extension_type:             # <<<<<<<<<<<<<<
 *     MP_UNKNOWN_EXTENSION = 0
 *     MP_DECIMAL = 1
 */
enum __pyx_t_8asynctnt_6iproto_9tarantool_mp_extension_type {
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_UNKNOWN_EXTENSION = 0,
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_DECIMAL = 1,
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_UUID = 2,
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR = 3,
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_DATETIME = 4
};

/* "asynctnt/iproto/tarantool.pxd":108
 *     MP_DATETIME = 4
 * 
 * cdef enum iproto_features:             # <<<<<<<<<<<<<<
 *     IPROTO_FEATURE_STREAMS = 0
 *     IPROTO_FEATURE_TRANSACTIONS = 1
 */
enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_features {
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FEATURE_STREAMS = 0,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FEATURE_TRANSACTIONS = 1,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FEATURE_ERROR_EXTENSION = 2,
  __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FEATURE_WATCHERS = 3
};

/* "asynctnt/iproto/tarantool.pxd":114
 *     IPROTO_FEATURE_WATCHERS = 3
 * 
 * cdef enum iproto_error_fields:             # <<<<<<<<<<<<<<
 *     MP_ERROR_STACK = 0x00
 * 
 */
enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_error_fields {
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_STACK = 0x00
};

/* "asynctnt/iproto/tarantool.pxd":117
 *     MP_ERROR_STACK = 0x00
 * 
 * cdef enum iproto_error_stack_fields:             # <<<<<<<<<<<<<<
 *     MP_ERROR_TYPE = 0x00
 *     MP_ERROR_FILE = 0x01
 */
enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_error_stack_fields {
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_TYPE = 0x00,
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_FILE = 0x01,
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_LINE = 0x02,
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_MESSAGE = 0x03,
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_ERRNO = 0x04,
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_ERRCODE = 0x05,
  __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_FIELDS = 0x06
};
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value {
  int __pyx_n;
  PyObject *default_value;
};

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {
  int __pyx_n;
  PyObject *default_value;
};
struct __pyx_opt_args_7cpython_8datetime_time_new;
struct __pyx_opt_args_7cpython_8datetime_datetime_new;
struct __pyx_opt_args_7cpython_8datetime_timezone_new;
struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp;

/* "cpython/datetime.pxd":292
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
 */
struct __pyx_opt_args_7cpython_8datetime_time_new {
  int __pyx_n;
  int fold;
};

/* "cpython/datetime.pxd":297
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
 */
struct __pyx_opt_args_7cpython_8datetime_datetime_new {
  int __pyx_n;
  int fold;
};

/* "cpython/datetime.pxd":306
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */
struct __pyx_opt_args_7cpython_8datetime_timezone_new {
  int __pyx_n;
  PyObject *name;
};

/* "cpython/datetime.pxd":312
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 */
struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp {
  int __pyx_n;
  PyObject *tz;
};
struct __pyx_opt_args_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index;
struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime;
struct __pyx_opt_args_8asynctnt_6iproto_8protocol_10ReadBuffer_create;
struct __pyx_t_8asynctnt_6iproto_8protocol_Header;

/* "asynctnt/iproto/coreproto.pxd":1
 * cdef enum ProtocolState:             # <<<<<<<<<<<<<<
 *     PROTOCOL_IDLE = 0
 *     PROTOCOL_GREETING = 1
 */
enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState {
  __pyx_e_8asynctnt_6iproto_8protocol_PROTOCOL_IDLE = 0,
  __pyx_e_8asynctnt_6iproto_8protocol_PROTOCOL_GREETING = 1,
  __pyx_e_8asynctnt_6iproto_8protocol_PROTOCOL_NORMAL = 2
};

/* "asynctnt/iproto/coreproto.pxd":7
 * 
 * 
 * cdef enum ConnectionState:             # <<<<<<<<<<<<<<
 *     CONNECTION_BAD = 0
 *     CONNECTION_CONNECTED = 1
 */
enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState {
  __pyx_e_8asynctnt_6iproto_8protocol_CONNECTION_BAD = 0,
  __pyx_e_8asynctnt_6iproto_8protocol_CONNECTION_CONNECTED = 1,
  __pyx_e_8asynctnt_6iproto_8protocol_CONNECTION_FULL = 2
};

/* "asynctnt/iproto/protocol.pxd":38
 * include "coreproto.pxd"
 * 
 * cdef enum PostConnectionState:             # <<<<<<<<<<<<<<
 *     POST_CONNECTION_NONE = 0
 *     POST_CONNECTION_ID = 10
 */
enum __pyx_t_8asynctnt_6iproto_8protocol_PostConnectionState {
  __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_NONE = 0,
  __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_ID = 10,
  __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_AUTH = 20,
  __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_SCHEMA = 30,
  __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_DONE = 0x64
};

/* "asynctnt/iproto/schema.pxd":54
 * 
 *     cdef void add_index(self, SchemaIndex idx)
 *     cdef SchemaIndex get_index(self, index, create_dummy=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index {
  int __pyx_n;
  PyObject *create_dummy;
};

/* "asynctnt/iproto/ext.pxd":19
 * cdef object uuid_decode(const char ** p, uint32_t length)
 * 
 * cdef struct IProtoDateTime:             # <<<<<<<<<<<<<<
 *     int64_t seconds
 *     int32_t nsec
 */
struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime {
  int64_t seconds;
  int32_t nsec;
  int16_t tzoffset;
  int16_t tzindex;
};

/* "asynctnt/iproto/rbuffer.pxd":31
 * 
 *     @staticmethod
 *     cdef ReadBuffer create(str encoding, size_t initial_buffer_size= *)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _reallocate(self, size_t new_size) except *
 */
struct __pyx_opt_args_8asynctnt_6iproto_8protocol_10ReadBuffer_create {
  int __pyx_n;
  size_t initial_buffer_size;
};

/* "asynctnt/iproto/response.pxd":5
 * 
 * 
 * cdef struct Header:             # <<<<<<<<<<<<<<
 *     int32_t code
 *     int32_t return_code
 */
struct __pyx_t_8asynctnt_6iproto_8protocol_Header {
  int32_t code;
  int32_t return_code;
  uint64_t sync;
  int64_t schema_id;
};

/* "asynctnt/iproto/protocol.pxd":46
 * 
 * 
 * ctypedef object (*req_execute_func)(BaseProtocol, BaseRequest, float)             # <<<<<<<<<<<<<<
 * 
 * cdef class BaseProtocol(CoreProtocol):
 */
typedef PyObject *(*__pyx_t_8asynctnt_6iproto_8protocol_req_execute_func)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, float);

/* "asynctnt/iproto/schema.pxd":4
 * 
 * 
 * cdef class Field:             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly str name
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_Field {
  PyObject_HEAD
  PyObject *name;
  PyObject *type;
  PyObject *collation;
  PyObject *is_nullable;
  PyObject *is_autoincrement;
  PyObject *span;
};


/* "asynctnt/iproto/schema.pxd":14
 * 
 * 
 * cdef public class Metadata [object C_Metadata, type C_Metadata_Type]:             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly list fields
 */
struct C_Metadata {
  PyObject_HEAD
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Metadata *__pyx_vtab;
  PyObject *fields;
  PyObject *name_id_map;
  PyObject *names;
};

__PYX_EXTERN_C DL_EXPORT(PyTypeObject) C_Metadata_Type;

/* "asynctnt/iproto/schema.pxd":27
 * 
 * 
 * cdef class SchemaIndex:             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly int sid
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex {
  PyObject_HEAD
  int sid;
  int iid;
  PyObject *name;
  PyObject *index_type;
  PyObject *unique;
  struct C_Metadata *metadata;
};


/* "asynctnt/iproto/schema.pxd":37
 * 
 * 
 * cdef class SchemaDummyIndex(SchemaIndex):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaDummyIndex {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex __pyx_base;
};


/* "asynctnt/iproto/schema.pxd":41
 * 
 * 
 * cdef class SchemaSpace:             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly int sid
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace {
  PyObject_HEAD
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_vtab;
  int sid;
  int owner;
  PyObject *name;
  PyObject *engine;
  int field_count;
  PyObject *flags;
  struct C_Metadata *metadata;
  PyObject *indexes;
};


/* "asynctnt/iproto/schema.pxd":57
 * 
 * 
 * cdef class SchemaDummySpace(SchemaSpace):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaDummySpace {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace __pyx_base;
};


/* "asynctnt/iproto/schema.pxd":61
 * 
 * 
 * cdef class Schema:             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly dict spaces
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema {
  PyObject_HEAD
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Schema *__pyx_vtab;
  PyObject *spaces;
  int id;
};


/* "asynctnt/iproto/buffer.pxd":6
 * 
 * @cython.final
 * cdef class WriteBuffer:             # <<<<<<<<<<<<<<
 *     cdef:
 *         # Preallocated small buffer
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer {
  PyObject_HEAD
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_vtab;
  int _smallbuf_inuse;
  char _smallbuf[0x400];
  char *_buf;
  Py_ssize_t _size;
  Py_ssize_t _length;
  int _view_count;
  PyObject *_encoding;
};


/* "asynctnt/iproto/rbuffer.pxd":21
 * 
 * @cython.final
 * cdef class ReadBuffer:             # <<<<<<<<<<<<<<
 *     cdef:
 *         char *buf
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer {
  PyObject_HEAD
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_vtab;
  char *buf;
  size_t initial_buffer_size;
  size_t len;
  size_t use;
  PyObject *encoding;
};


/* "asynctnt/iproto/requests/base.pxd":4
 * 
 * 
 * cdef class BaseRequest:             # <<<<<<<<<<<<<<
 *     cdef:
 *         tarantool.iproto_type op
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest {
  PyObject_HEAD
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_vtab;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type op;
  uint64_t sync;
  int64_t schema_id;
  uint64_t stream_id;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *space;
  PyObject *waiter;
  PyObject *timeout_handle;
  int parse_metadata;
  int parse_as_tuples;
  int push_subscribe;
  int check_schema_change;
};


/* "asynctnt/iproto/requests/ping.pxd":1
 * cdef class PingRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};


/* "asynctnt/iproto/requests/call.pxd":1
 * cdef class CallRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str func_name
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
  PyObject *func_name;
  PyObject *args;
};


/* "asynctnt/iproto/requests/eval.pxd":1
 * cdef class EvalRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str expression
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
  PyObject *expression;
  PyObject *args;
};


/* "asynctnt/iproto/requests/select.pxd":1
 * cdef class SelectRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         SchemaIndex index
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *index;
  PyObject *key;
  uint64_t offset;
  uint64_t limit;
  uint32_t iterator;
};


/* "asynctnt/iproto/requests/insert.pxd":1
 * cdef class InsertRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object t
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
  PyObject *t;
};


/* "asynctnt/iproto/requests/delete.pxd":1
 * cdef class DeleteRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         SchemaIndex index
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *index;
  PyObject *key;
};


/* "asynctnt/iproto/requests/update.pxd":1
 * cdef class UpdateRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         SchemaIndex index
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *index;
  PyObject *key;
  PyObject *operations;
};


/* "asynctnt/iproto/requests/upsert.pxd":1
 * cdef class UpsertRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object t
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
  PyObject *t;
  PyObject *operations;
};


/* "asynctnt/iproto/requests/prepare.pxd":1
 * cdef class PrepareRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str query
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
  PyObject *query;
  uint64_t statement_id;
};


/* "asynctnt/iproto/requests/execute.pxd":1
 * cdef class ExecuteRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str query
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
  PyObject *query;
  uint64_t statement_id;
  PyObject *args;
};


/* "asynctnt/iproto/requests/id.pxd":1
 * cdef class IDRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};


/* "asynctnt/iproto/requests/auth.pxd":1
 * cdef class AuthRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes salt
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
  PyObject *salt;
  PyObject *username;
  PyObject *password;
};


/* "asynctnt/iproto/requests/streams.pxd":1
 * cdef class BeginRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t isolation
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
  uint32_t isolation;
  double tx_timeout;
};


/* "asynctnt/iproto/requests/streams.pxd":6
 *         double tx_timeout
 * 
 * cdef class CommitRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};


/* "asynctnt/iproto/requests/streams.pxd":9
 *     pass
 * 
 * cdef class RollbackRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};


/* "asynctnt/iproto/response.pxd":11
 *     int64_t schema_id
 * 
 * cdef class IProtoErrorStackFrame:             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly str error_type
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame {
  PyObject_HEAD
  PyObject *error_type;
  PyObject *file;
  int line;
  PyObject *message;
  int err_no;
  int code;
  PyObject *fields;
};


/* "asynctnt/iproto/response.pxd":21
 *         readonly dict fields
 * 
 * cdef class IProtoError:             # <<<<<<<<<<<<<<
 *     cdef:
 *         readonly list trace
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError {
  PyObject_HEAD
  PyObject *trace;
};


/* "asynctnt/iproto/response.pxd":25
 *         readonly list trace
 * 
 * cdef class Response:             # <<<<<<<<<<<<<<
 *     cdef:
 *         int32_t code_
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_Response {
  PyObject_HEAD
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Response *__pyx_vtab;
  int32_t code_;
  int32_t return_code_;
  uint64_t sync_;
  int64_t schema_id_;
  PyObject *errmsg;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *error;
  int _rowcount;
  PyObject *body;
  PyObject *encoding;
  struct C_Metadata *metadata;
  struct C_Metadata *params;
  int params_count;
  PyObject *autoincrement_ids;
  uint64_t stmt_id_;
  int _push_subscribe;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *request_;
  PyObject *_exception;
  PyObject *_q;
  PyObject *_push_event;
  PyObject *_q_append;
  PyObject *_q_popleft;
  PyObject *_push_event_set;
  PyObject *_push_event_clear;
};


/* "asynctnt/iproto/db.pxd":6
 * 
 * @cython.final
 * cdef class Db:             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint64_t _stream_id
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_Db {
  PyObject_HEAD
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Db *__pyx_vtab;
  uint64_t _stream_id;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *_protocol;
  PyObject *_encoding;
};


/* "asynctnt/iproto/push.pxd":5
 * 
 * @cython.final
 * cdef class PushIterator:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _fut
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator {
  PyObject_HEAD
  PyObject *_fut;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *_request;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *_response;
};


/* "asynctnt/iproto/coreproto.pxd":13
 * 
 * 
 * cdef class CoreProtocol:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object host
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol {
  PyObject_HEAD
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_vtab;
  PyObject *host;
  PyObject *port;
  PyObject *encoding;
  enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState state;
  enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState con_state;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *rbuf;
  PyObject *version;
  PyObject *salt;
};


/* "asynctnt/iproto/protocol.pxd":48
 * ctypedef object (*req_execute_func)(BaseProtocol, BaseRequest, float)
 * 
 * cdef class BaseProtocol(CoreProtocol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object loop
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol __pyx_base;
  PyObject *loop;
  PyObject *username;
  PyObject *password;
  int fetch_schema;
  int auto_refetch_schema;
  float request_timeout;
  int post_con_state;
  PyObject *connected_fut;
  PyObject *on_connection_made_cb;
  PyObject *on_connection_lost_cb;
  PyObject *_on_request_completed_cb;
  PyObject *_on_request_timeout_cb;
  PyObject *_reqs;
  uint64_t _sync;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *_schema;
  int64_t _schema_id;
  int _schema_fetch_in_progress;
  PyObject *_refetch_schema_future;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *_db;
  __pyx_t_8asynctnt_6iproto_8protocol_req_execute_func execute;
  PyObject *create_future;
};


/* "asynctnt/iproto/buffer.pyx":54
 *         self._view_count -= 1
 * 
 *     def hex(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))
 *                         for i in range(self._length))
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex {
  PyObject_HEAD
  struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self;
};


/* "asynctnt/iproto/buffer.pyx":55
 * 
 *     def hex(self):  # pragma: nocover
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))             # <<<<<<<<<<<<<<
 *                         for i in range(self._length))
 * 
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *__pyx_outer_scope;
  Py_ssize_t __pyx_genexpr_arg_0;
  Py_ssize_t __pyx_v_i;
};


/* "asynctnt/iproto/push.pyx":62
 *         return self
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __anext__(self):
 *         cdef Response response
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ {
  PyObject_HEAD
  PyObject *__pyx_v_ev;
  PyObject *__pyx_v_exc;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_response;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self;
};


/* "asynctnt/iproto/protocol.pyx":247
 *         waiter.set_result(response)
 * 
 *     cdef void _do_id(self):             # <<<<<<<<<<<<<<
 *         fut = self._db._id(0.0)
 * 
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id {
  PyObject_HEAD
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self;
};


/* "asynctnt/iproto/protocol.pyx":268
 *         fut.add_done_callback(on_id)
 * 
 *     cdef void _do_auth(self, str username, str password):             # <<<<<<<<<<<<<<
 *         # No extra error handling from Db.execute
 *         fut = self._db._auth(self.salt, username, password, 0)
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth {
  PyObject_HEAD
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self;
};


/* "asynctnt/iproto/protocol.pyx":290
 *         fut.add_done_callback(on_authorized)
 * 
 *     cdef void _do_fetch_schema(self, object fut):             # <<<<<<<<<<<<<<
 *         if self._schema_fetch_in_progress:
 *             return
 */
struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema {
  PyObject_HEAD
  PyObject *__pyx_v_fut;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self;
};


__PYX_EXTERN_C DL_EXPORT(PyTypeObject) C_Metadata_Type;

/* "asynctnt/iproto/schema.pyx":29
 * @cython.final
 * @cython.freelist(METADATA_FREELIST_SIZE)
 * cdef class Metadata:             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self.fields = []
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Metadata {
  int (*len)(struct C_Metadata *);
  void (*add)(struct C_Metadata *, int, struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *);
  PyObject *(*name_by_id)(struct C_Metadata *, int);
  int (*id_by_name)(struct C_Metadata *, PyObject *);
  int (*id_by_name_safe)(struct C_Metadata *, PyObject *);
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Metadata *__pyx_vtabptr_8asynctnt_6iproto_8protocol_Metadata;
static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_len(struct C_Metadata *);
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_add(struct C_Metadata *, int, struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *);
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_8Metadata_name_by_id(struct C_Metadata *, int);
static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_id_by_name(struct C_Metadata *, PyObject *);
static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_id_by_name_safe(struct C_Metadata *, PyObject *);


/* "asynctnt/iproto/schema.pyx":88
 * 
 * @cython.final
 * cdef class SchemaSpace:             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self.sid = -1
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SchemaSpace {
  void (*add_index)(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *);
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *(*get_index)(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *, PyObject *, struct __pyx_opt_args_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index *__pyx_optional_args);
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_vtabptr_8asynctnt_6iproto_8protocol_SchemaSpace;
static void __pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_add_index(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *);
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *, PyObject *, struct __pyx_opt_args_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index *__pyx_optional_args);


/* "asynctnt/iproto/schema.pxd":57
 * 
 * 
 * cdef class SchemaDummySpace(SchemaSpace):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SchemaDummySpace {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SchemaSpace __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SchemaDummySpace *__pyx_vtabptr_8asynctnt_6iproto_8protocol_SchemaDummySpace;


/* "asynctnt/iproto/schema.pyx":145
 * 
 * @cython.final
 * cdef class Schema:             # <<<<<<<<<<<<<<
 *     def __cinit__(self, int schema_id):
 *         self.id = schema_id
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Schema {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *(*get_space)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *);
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *(*create_dummy_space)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, int);
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *(*get_or_create_space)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *);
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *(*parse_space)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *);
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *(*parse_index)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *);
  PyObject *(*clear)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *);
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *(*parse)(int64_t, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Schema *__pyx_vtabptr_8asynctnt_6iproto_8protocol_Schema;
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_space(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *);
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_create_dummy_space(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, int);
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_or_create_space(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *);
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse_space(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *);
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse_index(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_clear(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *);
static struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse(int64_t, PyObject *, PyObject *);


/* "asynctnt/iproto/buffer.pyx":24
 * @cython.final
 * @cython.freelist(_BUFFER_FREELIST_SIZE)
 * cdef class WriteBuffer:             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self._smallbuf_inuse = True
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_WriteBuffer {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *(*create)(PyObject *);
  PyObject *(*_check_readonly)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *);
  PyObject *(*len)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *);
  int (*ensure_allocated)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, Py_ssize_t);
  char *(*_ensure_allocated)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, Py_ssize_t);
  int (*_reallocate)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, Py_ssize_t);
  int (*write_buffer)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *);
  int (*write_header)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, uint64_t, enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, int64_t, uint64_t);
  void (*write_length)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *);
  char *(*mp_encode_nil)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *);
  char *(*mp_encode_bool)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, int);
  char *(*mp_encode_double)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, double);
  char *(*mp_encode_uint)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, uint64_t);
  char *(*mp_encode_int)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, int64_t);
  char *(*mp_encode_str)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, char const *, uint32_t);
  char *(*mp_encode_bin)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, char const *, uint32_t);
  char *(*mp_encode_decimal)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
  char *(*mp_encode_uuid)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
  char *(*mp_encode_datetime)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
  char *(*mp_encode_array)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, uint32_t);
  char *(*mp_encode_map)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, uint32_t);
  char *(*mp_encode_list)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
  char *(*mp_encode_tuple)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
  char *(*mp_encode_dict)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
  char *(*mp_encode_obj)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_vtabptr_8asynctnt_6iproto_8protocol_WriteBuffer;
static struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_create(PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__check_readonly(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *);
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_len(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *);
static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, Py_ssize_t);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, Py_ssize_t);
static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__reallocate(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, Py_ssize_t);
static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_buffer(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *);
static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_header(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, uint64_t, enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, int64_t, uint64_t);
static void __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_length(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_nil(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_bool(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, int);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_double(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, double);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uint(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, uint64_t);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_int(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, int64_t);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_str(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, char const *, uint32_t);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_bin(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, char const *, uint32_t);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_decimal(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uuid(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_datetime(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_array(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, uint32_t);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_map(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, uint32_t);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_list(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_tuple(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_dict(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_obj(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *);


/* "asynctnt/iproto/rbuffer.pyx":8
 * @cython.no_gc_clear
 * @cython.final
 * cdef class ReadBuffer:             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self.buf = NULL
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_ReadBuffer {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *(*create)(PyObject *, struct __pyx_opt_args_8asynctnt_6iproto_8protocol_10ReadBuffer_create *__pyx_optional_args);
  void (*_reallocate)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t);
  int (*extend)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, char const *, size_t);
  void (*move)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t);
  void (*move_offset)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, Py_ssize_t, size_t);
  PyObject *(*get_slice)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t, size_t);
  PyObject *(*get_slice_begin)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t);
  PyObject *(*get_slice_end)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t);
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_vtabptr_8asynctnt_6iproto_8protocol_ReadBuffer;
static struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_create(PyObject *, struct __pyx_opt_args_8asynctnt_6iproto_8protocol_10ReadBuffer_create *__pyx_optional_args);
static void __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer__reallocate(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t);
static int __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_extend(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, char const *, size_t);
static void __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_move(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t);
static void __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_move_offset(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, Py_ssize_t, size_t);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t, size_t);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice_begin(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice_end(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t);


/* "asynctnt/iproto/requests/base.pyx":6
 * 
 * @cython.freelist(REQUEST_FREELIST)
 * cdef class BaseRequest:             # <<<<<<<<<<<<<<
 * 
 *     # def __cinit__(self):
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest {
  struct C_Metadata *(*metadata)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *);
  struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *(*encode)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, PyObject *);
  int (*encode_body)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *);
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
static CYTHON_INLINE struct C_Metadata *__pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_metadata(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *);
static CYTHON_INLINE struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_encode(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, PyObject *);


/* "asynctnt/iproto/requests/ping.pyx":5
 * 
 * @cython.final
 * cdef class PingRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_PingRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_PingRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_PingRequest;


/* "asynctnt/iproto/requests/call.pyx":7
 * 
 * @cython.final
 * cdef class CallRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         cdef:
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CallRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CallRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_CallRequest;


/* "asynctnt/iproto/requests/eval.pyx":5
 * 
 * @cython.final
 * cdef class EvalRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         cdef:
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_EvalRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_EvalRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_EvalRequest;


/* "asynctnt/iproto/requests/select.pyx":5
 * 
 * @cython.final
 * cdef class SelectRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         cdef:
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SelectRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SelectRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_SelectRequest;


/* "asynctnt/iproto/requests/insert.pyx":5
 * 
 * @cython.final
 * cdef class InsertRequest(BaseRequest):             # <<<<<<<<<<<<<<
 * 
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_InsertRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_InsertRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_InsertRequest;


/* "asynctnt/iproto/requests/delete.pyx":5
 * 
 * @cython.final
 * cdef class DeleteRequest(BaseRequest):             # <<<<<<<<<<<<<<
 * 
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_DeleteRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_DeleteRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_DeleteRequest;


/* "asynctnt/iproto/requests/update.pyx":5
 * 
 * @cython.final
 * cdef class UpdateRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         return encode_request_update(buffer, self.space, self.index,
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_UpdateRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_UpdateRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_UpdateRequest;


/* "asynctnt/iproto/requests/upsert.pyx":5
 * 
 * @cython.final
 * cdef class UpsertRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         return encode_request_update(buffer, self.space, self.space.get_index(0),
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_UpsertRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_UpsertRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_UpsertRequest;


/* "asynctnt/iproto/requests/prepare.pyx":5
 * 
 * @cython.final
 * cdef class PrepareRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         cdef:
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_PrepareRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_PrepareRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_PrepareRequest;


/* "asynctnt/iproto/requests/execute.pyx":5
 * 
 * @cython.final
 * cdef class ExecuteRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         cdef:
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_ExecuteRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_ExecuteRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_ExecuteRequest;


/* "asynctnt/iproto/requests/id.pyx":6
 * 
 * @cython.final
 * cdef class IDRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         cdef:
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_IDRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_IDRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_IDRequest;


/* "asynctnt/iproto/requests/auth.pyx":31
 * 
 * @cython.final
 * cdef class AuthRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         cdef:
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_AuthRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_AuthRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_AuthRequest;


/* "asynctnt/iproto/requests/streams.pyx":5
 * 
 * @cython.final
 * cdef class BeginRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         cdef:
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BeginRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BeginRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BeginRequest;


/* "asynctnt/iproto/requests/streams.pyx":33
 * 
 * @cython.final
 * cdef class CommitRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CommitRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CommitRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_CommitRequest;


/* "asynctnt/iproto/requests/streams.pyx":37
 * 
 * @cython.final
 * cdef class RollbackRequest(BaseRequest):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_RollbackRequest {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_base;
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_RollbackRequest *__pyx_vtabptr_8asynctnt_6iproto_8protocol_RollbackRequest;


/* "asynctnt/iproto/response.pyx":30
 * @cython.final
 * @cython.freelist(REQUEST_FREELIST)
 * cdef class Response:             # <<<<<<<<<<<<<<
 *     """
 *         Response object for all the requests to Tarantool
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Response {
  int (*is_error)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
  uint32_t (*_len)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
  void (*init_push)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
  void (*add_push)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, PyObject *);
  PyObject *(*pop_push)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
  int (*push_len)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
  void (*set_data)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, PyObject *);
  void (*set_exception)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, PyObject *);
  PyObject *(*get_exception)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
  void (*notify)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Response *__pyx_vtabptr_8asynctnt_6iproto_8protocol_Response;
static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Response_is_error(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
static CYTHON_INLINE uint32_t __pyx_f_8asynctnt_6iproto_8protocol_8Response__len(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_init_push(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_add_push(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_8Response_pop_push(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Response_push_len(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_set_data(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, PyObject *);
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_set_exception(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_8Response_get_exception(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_notify(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *);


/* "asynctnt/iproto/db.pyx":5
 * 
 * @cython.final
 * cdef class Db:             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self._stream_id = 0
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Db {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *(*create)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, uint64_t);
  uint64_t (*next_sync)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *);
  PyObject *(*_ping)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float);
  PyObject *(*_id)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float);
  PyObject *(*_auth)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, float);
  PyObject *(*_call)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, PyObject *, PyObject *, float, int);
  PyObject *(*_eval)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, float, int);
  PyObject *(*_select)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, uint64_t, uint64_t, PyObject *, float, int);
  PyObject *(*_insert)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, int, float);
  PyObject *(*_delete)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, float);
  PyObject *(*_update)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, PyObject *, float);
  PyObject *(*_upsert)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, float);
  PyObject *(*_execute)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, int, float);
  PyObject *(*_prepare)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, int, float);
  PyObject *(*_begin)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, uint32_t, double, float);
  PyObject *(*_commit)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float);
  PyObject *(*_rollback)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float);
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Db *__pyx_vtabptr_8asynctnt_6iproto_8protocol_Db;
static CYTHON_INLINE struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_f_8asynctnt_6iproto_8protocol_2Db_create(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, uint64_t);
static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__ping(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__id(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__auth(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, float);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__call(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, PyObject *, PyObject *, float, int);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__eval(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, float, int);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__select(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, uint64_t, uint64_t, PyObject *, float, int);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__insert(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, int, float);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__delete(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, float);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__update(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, PyObject *, float);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__upsert(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, float);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__execute(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, int, float);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__prepare(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, int, float);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__begin(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, uint32_t, double, float);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__commit(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float);
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__rollback(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float);


/* "asynctnt/iproto/coreproto.pyx":18
 * 
 * 
 * cdef class CoreProtocol:             # <<<<<<<<<<<<<<
 *     def __init__(self,
 *                  host, port,
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol {
  int (*_is_connected)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *);
  int (*_is_fully_connected)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *);
  void (*_write)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *, PyObject *);
  void (*_on_data_received)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *, PyObject *);
  void (*_process__greeting)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *);
  void (*_on_greeting_received)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *);
  void (*_on_response_received)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *, char const *, uint32_t);
  void (*_on_connection_made)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *);
  void (*_on_connection_lost)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *, PyObject *);
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_vtabptr_8asynctnt_6iproto_8protocol_CoreProtocol;


/* "asynctnt/iproto/protocol.pyx":62
 * 
 * 
 * cdef class BaseProtocol(CoreProtocol):             # <<<<<<<<<<<<<<
 *     def __init__(self, host, port,
 *                  username, password,
 */

struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol {
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol __pyx_base;
  void (*_set_connection_ready)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *);
  void (*_set_connection_error)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, PyObject *);
  void (*_post_con_state_machine)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *);
  void (*_do_id)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *);
  void (*_do_auth)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, PyObject *, PyObject *);
  void (*_do_fetch_schema)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, PyObject *);
  PyObject *(*_refetch_schema)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *);
  uint64_t (*next_sync)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *);
  uint64_t (*next_stream_id)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *);
  uint32_t (*transform_iterator)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, PyObject *);
  PyObject *(*_new_waiter_for_request)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, float);
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *(*_create_db)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, int);
  PyObject *(*_execute_bad)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, float);
  PyObject *(*_execute_normal)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, float);
};
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseProtocol;
static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_next_sync(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *);
static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_next_stream_id(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *);
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* IncludeStringH.proto */
#include <string.h>

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg // no-op
    #define __Pyx_Arg_XDECREF_VARARGS(arg) // no-op - arg is borrowed
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg // no-op, __Pyx_Arg_FASTCALL is direct and this needs
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)  // no-op - arg was returned from array
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObject_Str.proto */
#define __Pyx_PyObject_Str(obj)\
    (likely(PyString_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* py_dict_clear.proto */
#define __Pyx_PyDict_Clear(d) (PyDict_Clear(d), 0)

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (1)
#elif CYTHON_COMPILING_IN_LIMITED_API  ||  (CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030C0000)
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* UnicodeAsUCS4.proto */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject*);

/* object_ord.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Ord(c)\
    (likely(PyUnicode_Check(c)) ? (long)__Pyx_PyUnicode_AsPy_UCS4(c) : __Pyx__PyObject_Ord(c))
#else
#define __Pyx_PyObject_Ord(c) __Pyx__PyObject_Ord(c)
#endif
static long __Pyx__PyObject_Ord(PyObject* c);

/* RaiseClosureNameError.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right)
    #endif
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_ConcatInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_ConcatInPlaceSafe(left, right) ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_ConcatInPlace(left, right))

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_enum__mp_type(enum mp_type value, Py_ssize_t width, char padding_char, char format_char);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* StopAsyncIteration.proto */
#define __Pyx_StopAsyncIteration_USED
static PyObject *__Pyx_PyExc_StopAsyncIteration;
static int __pyx_StopAsyncIteration_init(PyObject *module);

/* ReturnWithStopIteration.proto */
#define __Pyx_ReturnWithStopIteration(value)\
    if (value == Py_None) PyErr_SetNone(PyExc_StopIteration); else __Pyx__ReturnWithStopIteration(value)
static void __Pyx__ReturnWithStopIteration(PyObject* value);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Coroutine.proto */
#define __Pyx_Coroutine_USED
#define __Pyx_Coroutine_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_CoroutineType)
#define __Pyx_Coroutine_Check(obj) __Pyx_Coroutine_CheckExact(obj)
#define __Pyx_CoroutineAwait_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_CoroutineAwaitType)
#define __Pyx_Coroutine_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_CoroutineType, body, code, closure, name, qualname, module_name)
static int __pyx_Coroutine_init(PyObject *module);
static PyObject *__Pyx__Coroutine_await(PyObject *coroutine);
typedef struct {
    PyObject_HEAD
    PyObject *coroutine;
} __pyx_CoroutineAwaitObject;
static PyObject *__Pyx_CoroutineAwait_Close(__pyx_CoroutineAwaitObject *self, PyObject *arg);
static PyObject *__Pyx_CoroutineAwait_Throw(__pyx_CoroutineAwaitObject *self, PyObject *args);

/* GetAwaitIter.proto */
static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAwaitableIter(PyObject *o);
static PyObject *__Pyx__Coroutine_GetAwaitableIter(PyObject *o);

/* CoroutineYieldFrom.proto */
static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_7
#define __PYX_HAVE_RT_ImportType_proto_3_0_7
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_7(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_7(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_7 {
   __Pyx_ImportType_CheckSize_Error_3_0_7 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_7 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_7 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_7(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_7 check_size);
#endif

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* PatchInspect.proto */
static PyObject* __Pyx_patch_inspect(PyObject* module);

/* PatchAsyncIO.proto */
static PyObject* __Pyx_patch_asyncio(PyObject* module);

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type value);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_8protocol_ConnectionState(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState value);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState __Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_8protocol_ConnectionState(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_8protocol_ProtocolState(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState value);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState __Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_8protocol_ProtocolState(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__mp_type(enum mp_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int8_t(int8_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);

/* IterableCoroutine.proto */
#define __Pyx_IterableCoroutine_USED
#undef __Pyx_Coroutine_Check
#define __Pyx_Coroutine_Check(obj) (__Pyx_Coroutine_CheckExact(obj) || __Pyx_IS_TYPE(obj, __pyx_IterableCoroutineType))
#define __Pyx_IterableCoroutine_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_IterableCoroutineType, body, code, closure, name, qualname, module_name)
static int __pyx_IterableCoroutine_init(PyObject *module);

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_4year_year(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_5month_month(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_3day_day(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4hour_hour(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6minute_minute(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6second_second(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_11microsecond_microsecond(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_4time_6tzinfo_tzinfo(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4fold_fold(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4year_year(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_5month_month(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_3day_day(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4hour_hour(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6minute_minute(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6second_second(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_11microsecond_microsecond(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_8datetime_6tzinfo_tzinfo(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4fold_fold(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_3day_day(PyDateTime_Delta *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_6second_second(PyDateTime_Delta *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_11microsecond_microsecond(PyDateTime_Delta *__pyx_v_self); /* proto*/
static CYTHON_INLINE struct C_Metadata *__pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_metadata(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self); /* proto*/
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_add(struct C_Metadata *__pyx_v_self, int __pyx_v_id, struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_field); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_8Metadata_name_by_id(struct C_Metadata *__pyx_v_self, int __pyx_v_id); /* proto*/
static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_id_by_name(struct C_Metadata *__pyx_v_self, PyObject *__pyx_v_name); /* proto*/
static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_id_by_name_safe(struct C_Metadata *__pyx_v_self, PyObject *__pyx_v_name); /* proto*/
static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_len(struct C_Metadata *__pyx_v_self); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_add_index(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_idx); /* proto*/
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self, PyObject *__pyx_v_index, struct __pyx_opt_args_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index *__pyx_optional_args); /* proto*/
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_space(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, PyObject *__pyx_v_space); /* proto*/
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_or_create_space(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, PyObject *__pyx_v_space); /* proto*/
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_create_dummy_space(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, int __pyx_v_space_id); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_clear(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self); /* proto*/
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse_space(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, PyObject *__pyx_v_space_row); /* proto*/
static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse_index(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, PyObject *__pyx_v_index_row); /* proto*/
static struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse(int64_t __pyx_v_schema_id, PyObject *__pyx_v_spaces, PyObject *__pyx_v_indexes); /* proto*/
static struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_create(PyObject *__pyx_v_encoding); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__check_readonly(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_len(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, Py_ssize_t __pyx_v_extra_length); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, Py_ssize_t __pyx_v_extra_length); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__reallocate(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, Py_ssize_t __pyx_v_new_size); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_buffer(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buf); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_header(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, uint64_t __pyx_v_sync, enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_v_op, int64_t __pyx_v_schema_id, uint64_t __pyx_v_stream_id); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_length(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_nil(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_bool(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, int __pyx_v_value); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_double(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, double __pyx_v_value); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uint(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, uint64_t __pyx_v_value); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_int(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, int64_t __pyx_v_value); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_str(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, char const *__pyx_v_str, uint32_t __pyx_v_len); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_bin(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, char const *__pyx_v_data, uint32_t __pyx_v_len); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_decimal(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_value); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uuid(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_value); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_datetime(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_value); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_array(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, uint32_t __pyx_v_len); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_map(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, uint32_t __pyx_v_len); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_list(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_arr); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_tuple(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_t); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_dict(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_d); /* proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_obj(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_o); /* proto*/
static struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_create(PyObject *__pyx_v_encoding, struct __pyx_opt_args_8asynctnt_6iproto_8protocol_10ReadBuffer_create *__pyx_optional_args); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer__reallocate(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, size_t __pyx_v_new_size); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_extend(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, char const *__pyx_v_data, size_t __pyx_v_len); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_move(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, size_t __pyx_v_pos); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_move_offset(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, Py_ssize_t __pyx_v_offset, size_t __pyx_v_size); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, size_t __pyx_v_begin, size_t __pyx_v_end); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice_begin(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, size_t __pyx_v_begin); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice_end(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, size_t __pyx_v_end); /* proto*/
static CYTHON_INLINE struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_encode(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self, PyObject *__pyx_v_encoding); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_encode_body(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_11CallRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_11EvalRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_13SelectRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_13InsertRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_13DeleteRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_13UpdateRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_13UpsertRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_14PrepareRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_14ExecuteRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_9IDRequest_encode_body(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_11AuthRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_12BeginRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer); /* proto*/
static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Response_is_error(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto*/
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_init_push(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto*/
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_add_push(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self, PyObject *__pyx_v_push); /* proto*/
static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Response_push_len(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_8Response_pop_push(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto*/
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_set_data(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self, PyObject *__pyx_v_data); /* proto*/
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_set_exception(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self, PyObject *__pyx_v_exc); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_8Response_get_exception(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto*/
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_notify(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto*/
static CYTHON_INLINE uint32_t __pyx_f_8asynctnt_6iproto_8protocol_8Response__len(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto*/
static CYTHON_INLINE struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_f_8asynctnt_6iproto_8protocol_2Db_create(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_protocol, uint64_t __pyx_v_stream_id); /* proto*/
static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__ping(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__id(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__auth(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_salt, PyObject *__pyx_v_username, PyObject *__pyx_v_password, float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__call(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_v_op, PyObject *__pyx_v_func_name, PyObject *__pyx_v_args, float __pyx_v_timeout, int __pyx_v_push_subscribe); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__eval(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_expression, PyObject *__pyx_v_args, float __pyx_v_timeout, int __pyx_v_push_subscribe); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__select(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_index, PyObject *__pyx_v_key, uint64_t __pyx_v_offset, uint64_t __pyx_v_limit, PyObject *__pyx_v_iterator, float __pyx_v_timeout, int __pyx_v_check_schema_change); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__insert(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_t, int __pyx_v_replace, float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__delete(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_index, PyObject *__pyx_v_key, float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__update(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_index, PyObject *__pyx_v_key, PyObject *__pyx_v_operations, float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__upsert(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_t, PyObject *__pyx_v_operations, float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__execute(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_query, PyObject *__pyx_v_args, int __pyx_v_parse_metadata, float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__prepare(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_query, int __pyx_v_parse_metadata, float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__begin(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, uint32_t __pyx_v_isolation, double __pyx_v_tx_timeout, float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__commit(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__rollback(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__is_connected(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self); /* proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__is_fully_connected(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__write(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_buf); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_data_received(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_data); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__process__greeting(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_greeting_received(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_response_received(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, CYTHON_UNUSED char const *__pyx_v_buf, CYTHON_UNUSED uint32_t __pyx_v_buf_len); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_connection_made(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_connection_lost(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_exc); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__set_connection_ready(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__set_connection_error(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_e); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_greeting_received(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__post_con_state_machine(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_response_received(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, char const *__pyx_v_buf, uint32_t __pyx_v_buf_len); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__do_id(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__do_auth(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_username, PyObject *__pyx_v_password); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__do_fetch_schema(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_fut); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_connection_made(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_connection_lost(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_exc); /* proto*/
static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_next_sync(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto*/
static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_next_stream_id(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__new_waiter_for_request(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_response, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req, float __pyx_v_timeout); /* proto*/
static struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__create_db(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, int __pyx_v_gen_stream_id); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__execute_bad(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req, CYTHON_UNUSED float __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__execute_normal(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req, float __pyx_v_timeout); /* proto*/
static uint32_t __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_transform_iterator(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_iterator); /* proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__refetch_schema(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto*/

/* Module declarations from "asynctnt.iproto.tarantool" */

/* Module declarations from "cpython.version" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "cpython.int" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "cpython.string" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.cobject" */

/* Module declarations from "cpython.oldbuffer" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "datetime" */

/* Module declarations from "cpython.datetime" */
static CYTHON_INLINE void __pyx_f_7cpython_8datetime_import_datetime(void); /*proto*/
static CYTHON_INLINE PyDateTime_Delta *__pyx_f_7cpython_8datetime_timedelta_new(int, int, int); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_datetime_tzinfo(PyObject *); /*proto*/

/* Module declarations from "cpython.bytearray" */

/* Module declarations from "cpython" */

/* Module declarations from "asynctnt.iproto.tupleobj" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "libc" */

/* Module declarations from "libc.math" */

/* Module declarations from "cython" */

/* Module declarations from "asynctnt.iproto.protocol" */
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_timezone_new(PyObject *); /*proto*/
static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_load_u64(void const *); /*proto*/
static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_load_u32(void const *); /*proto*/
static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_load_u16(void const *); /*proto*/
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_store_u64(void *, uint64_t); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string(PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_decode_string(PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_dict_to_list_fields(PyObject *, struct C_Metadata *, int); /*proto*/
static CYTHON_INLINE uint32_t __pyx_f_8asynctnt_6iproto_8protocol_bcd_len(uint32_t); /*proto*/
static uint32_t __pyx_f_8asynctnt_6iproto_8protocol_decimal_len(int, uint32_t); /*proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_decimal_encode(char *, uint32_t, uint8_t, PyObject *, int); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_decimal_decode(char const **, uint32_t); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_uuid_decode(char const **, uint32_t); /*proto*/
static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_datetime_zero(struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *); /*proto*/
static CYTHON_INLINE uint32_t __pyx_f_8asynctnt_6iproto_8protocol_datetime_len(struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *); /*proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_datetime_encode(char *, struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *); /*proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_datetime_decode(char const **, uint32_t, struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *); /*proto*/
static void __pyx_f_8asynctnt_6iproto_8protocol_datetime_from_py(PyDateTime_DateTime *, struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_datetime_to_py(struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *); /*proto*/
static CYTHON_INLINE size_t __pyx_f_8asynctnt_6iproto_8protocol_size_t_max(size_t, size_t); /*proto*/
static CYTHON_INLINE uint32_t __pyx_f_8asynctnt_6iproto_8protocol_nearest_power_of_2(uint32_t); /*proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_encode_key_sequence(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *, struct C_Metadata *, int); /*proto*/
static char *__pyx_f_8asynctnt_6iproto_8protocol_encode_update_ops(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *); /*proto*/
static int __pyx_f_8asynctnt_6iproto_8protocol_encode_request_update(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *, PyObject *, PyObject *, int); /*proto*/
static Py_ssize_t __pyx_f_8asynctnt_6iproto_8protocol_response_parse_header(char const *, uint32_t, struct __pyx_t_8asynctnt_6iproto_8protocol_Header *); /*proto*/
static Py_ssize_t __pyx_f_8asynctnt_6iproto_8protocol_response_parse_body(char const *, uint32_t, struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, int); /*proto*/
static CYTHON_INLINE struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *__pyx_f_8asynctnt_6iproto_8protocol_parse_iproto_error(char const **, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol__sha1(PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol__strxor(PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol__decode_obj(char const **, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol__response_parse_body_data(char const **, struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *); /*proto*/
static struct C_Metadata *__pyx_f_8asynctnt_6iproto_8protocol_response_parse_metadata(char const **, PyObject *); /*proto*/
static CYTHON_INLINE struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_f_8asynctnt_6iproto_8protocol_parse_iproto_error_stack_frame(char const **, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_BaseRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PingRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CallRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_EvalRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_SelectRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_InsertRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_DeleteRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_UpdateRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_UpsertRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PrepareRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_ExecuteRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IDRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_AuthRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_BeginRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CommitRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_RollbackRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IProtoErrorStackFrame__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IProtoError__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PushIterator__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *, PyObject *); /*proto*/
static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CoreProtocol__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "asynctnt.iproto.protocol"
extern int __pyx_module_is_main_asynctnt__iproto__protocol;
int __pyx_module_is_main_asynctnt__iproto__protocol = 0;

/* Implementation of "asynctnt.iproto.protocol" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_BufferError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_UnicodeDecodeError;
static PyObject *__pyx_builtin_hex;
static PyObject *__pyx_builtin_AttributeError;
/* #### Code section: string_decls ### */
static const char __pyx_k_e[] = "e";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_Db[] = "Db";
static const char __pyx_k_EQ[] = "EQ";
static const char __pyx_k_GE[] = "GE";
static const char __pyx_k_GT[] = "GT";
static const char __pyx_k_LE[] = "LE";
static const char __pyx_k_LT[] = "LT";
static const char __pyx_k__2[] = ".";
static const char __pyx_k__3[] = "";
static const char __pyx_k__5[] = ":";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_re[] = "re";
static const char __pyx_k_02x[] = "{:02x}";
static const char __pyx_k_ALL[] = "ALL";
static const char __pyx_k_REQ[] = "REQ";
static const char __pyx_k__14[] = ", ";
static const char __pyx_k__17[] = " ... ";
static const char __pyx_k__18[] = "[";
static const char __pyx_k__19[] = "]";
static const char __pyx_k__65[] = "*";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_exc[] = "exc";
static const char __pyx_k_fut[] = "fut";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_got[] = "{}, got: {}";
static const char __pyx_k_hex[] = "hex";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_req[] = "req";
static const char __pyx_k_set[] = "set";
static const char __pyx_k_ver[] = "ver";
static const char __pyx_k_UUID[] = "UUID";
static const char __pyx_k__211[] = "?";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_call[] = "call";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_done[] = "done";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_eval[] = "eval";
static const char __pyx_k_host[] = "host";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_loop[] = "loop";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_ping[] = "ping";
static const char __pyx_k_port[] = "port";
static const char __pyx_k_repr[] = "repr";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sha1[] = "sha1";
static const char __pyx_k_sign[] = "sign";
static const char __pyx_k_sock[] = "sock";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_sync[] = "sync";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_uuid[] = "uuid";
static const char __pyx_k_wait[] = "wait";
static const char __pyx_k_Event[] = "Event";
static const char __pyx_k_Field[] = "Field";
static const char __pyx_k_anext[] = "__anext__";
static const char __pyx_k_ascii[] = "ascii";
static const char __pyx_k_await[] = "__await__";
static const char __pyx_k_begin[] = "begin";
static const char __pyx_k_bytes[] = "bytes";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_deque[] = "deque";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_field[] = "field";
static const char __pyx_k_group[] = "group";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_limit[] = "limit";
static const char __pyx_k_match[] = "match";
static const char __pyx_k_on_id[] = "on_id";
static const char __pyx_k_query[] = "query";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_space[] = "space";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_Future[] = "Future";
static const char __pyx_k_Schema[] = "Schema";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_base64[] = "base64";
static const char __pyx_k_call16[] = "call16";
static const char __pyx_k_cancel[] = "cancel";
static const char __pyx_k_commit[] = "commit";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_delete[] = "delete";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_digest[] = "digest";
static const char __pyx_k_digits[] = "digits";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_family[] = "family";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_gather[] = "gather";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_insert[] = "insert";
static const char __pyx_k_logger[] = "logger";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_offset[] = "offset";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_select[] = "select";
static const char __pyx_k_socket[] = "socket";
static const char __pyx_k_spaces[] = "spaces";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_upsert[] = "upsert";
static const char __pyx_k_waiter[] = "waiter";
static const char __pyx_k_AF_UNIX[] = "AF_UNIX";
static const char __pyx_k_Db_call[] = "Db.call";
static const char __pyx_k_Db_eval[] = "Db.eval";
static const char __pyx_k_Db_ping[] = "Db.ping";
static const char __pyx_k_Decimal[] = "Decimal";
static const char __pyx_k_IntEnum[] = "IntEnum";
static const char __pyx_k_asyncio[] = "asyncio";
static const char __pyx_k_closing[] = "_closing";
static const char __pyx_k_compile[] = "compile";
static const char __pyx_k_decimal[] = "decimal";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_execute[] = "execute";
static const char __pyx_k_futures[] = "futures";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_hashlib[] = "hashlib";
static const char __pyx_k_indexes[] = "indexes";
static const char __pyx_k_inspect[] = "inspect";
static const char __pyx_k_popleft[] = "popleft";
static const char __pyx_k_prepare[] = "prepare";
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_timeout[] = "timeout";
static const char __pyx_k_version[] = "version";
static const char __pyx_k_warning[] = "warning";
static const char __pyx_k_Db_begin[] = "Db.begin";
static const char __pyx_k_Iterator[] = "Iterator";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_Metadata[] = "Metadata";
static const char __pyx_k_NEIGHBOR[] = "NEIGHBOR";
static const char __pyx_k_OVERLAPS[] = "OVERLAPS";
static const char __pyx_k_Optional[] = "Optional";
static const char __pyx_k_Protocol[] = "Protocol";
static const char __pyx_k_Response[] = "Response";
static const char __pyx_k_as_tuple[] = "as_tuple";
static const char __pyx_k_asynctnt[] = "asynctnt";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_encoding[] = "encoding";
static const char __pyx_k_exponent[] = "exponent";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_iterator[] = "iterator";
static const char __pyx_k_on_fetch[] = "on_fetch";
static const char __pyx_k_password[] = "password";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_response[] = "_response";
static const char __pyx_k_rollback[] = "rollback";
static const char __pyx_k_rowcount[] = "rowcount";
static const char __pyx_k_set_name[] = "__set_name__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_timezone[] = "timezone";
static const char __pyx_k_username[] = "username";
static const char __pyx_k_Db_call16[] = "Db.call16";
static const char __pyx_k_Db_commit[] = "Db.commit";
static const char __pyx_k_Db_delete[] = "Db.delete";
static const char __pyx_k_Db_insert[] = "Db.insert";
static const char __pyx_k_Db_select[] = "Db.select";
static const char __pyx_k_Db_update[] = "Db.update";
static const char __pyx_k_Db_upsert[] = "Db.upsert";
static const char __pyx_k_IDRequest[] = "IDRequest";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_b64decode[] = "b64decode";
static const char __pyx_k_cancelled[] = "cancelled";
static const char __pyx_k_create_db[] = "create_db";
static const char __pyx_k_exception[] = "exception";
static const char __pyx_k_func_name[] = "func_name";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_isolation[] = "isolation";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_prepare_2[] = "__prepare__";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_schema_id[] = "schema_id";
static const char __pyx_k_stream_id[] = "stream_id";
static const char __pyx_k_timestamp[] = "timestamp";
static const char __pyx_k_transport[] = "transport";
static const char __pyx_k_utcoffset[] = "utcoffset";
static const char __pyx_k_Db_execute[] = "Db.execute";
static const char __pyx_k_Db_prepare[] = "Db.prepare";
static const char __pyx_k_Db_replace[] = "Db.replace";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ReadBuffer[] = "ReadBuffer";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_call_later[] = "call_later";
static const char __pyx_k_expression[] = "expression";
static const char __pyx_k_operations[] = "operations";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_response_2[] = "response";
static const char __pyx_k_set_result[] = "set_result";
static const char __pyx_k_setsockopt[] = "setsockopt";
static const char __pyx_k_tx_timeout[] = "tx_timeout";
static const char __pyx_k_AuthRequest[] = "AuthRequest";
static const char __pyx_k_BaseRequest[] = "BaseRequest";
static const char __pyx_k_BufferError[] = "BufferError";
static const char __pyx_k_CallRequest[] = "CallRequest";
static const char __pyx_k_Db_rollback[] = "Db.rollback";
static const char __pyx_k_EvalRequest[] = "EvalRequest";
static const char __pyx_k_IPPROTO_TCP[] = "IPPROTO_TCP";
static const char __pyx_k_IProtoError[] = "IProtoError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_PingRequest[] = "PingRequest";
static const char __pyx_k_SchemaIndex[] = "SchemaIndex";
static const char __pyx_k_SchemaSpace[] = "SchemaSpace";
static const char __pyx_k_TCP_NODELAY[] = "TCP_NODELAY";
static const char __pyx_k_WriteBuffer[] = "WriteBuffer";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_get_version[] = "get_version";
static const char __pyx_k_is_nullable[] = "is_nullable";
static const char __pyx_k_mro_entries[] = "__mro_entries__";
static const char __pyx_k_BITS_ALL_SET[] = "BITS_ALL_SET";
static const char __pyx_k_BITS_ANY_SET[] = "BITS_ANY_SET";
static const char __pyx_k_BaseProtocol[] = "BaseProtocol";
static const char __pyx_k_BeginRequest[] = "BeginRequest";
static const char __pyx_k_CoreProtocol[] = "CoreProtocol";
static const char __pyx_k_PushIterator[] = "PushIterator";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_TimeoutError[] = "TimeoutError";
static const char __pyx_k_asynctnt_log[] = "asynctnt.log";
static const char __pyx_k_fetch_schema[] = "fetch_schema";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_connected[] = "is_connected";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_CommitRequest[] = "CommitRequest";
static const char __pyx_k_DeleteRequest[] = "DeleteRequest";
static const char __pyx_k_InsertRequest[] = "InsertRequest";
static const char __pyx_k_Response_done[] = "Response.done";
static const char __pyx_k_Schema_spaces[] = "<Schema spaces={}>";
static const char __pyx_k_SelectRequest[] = "SelectRequest";
static const char __pyx_k_UpdateRequest[] = "UpdateRequest";
static const char __pyx_k_UpsertRequest[] = "UpsertRequest";
static const char __pyx_k_asyncio_tasks[] = "asyncio.tasks";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_connected_fut[] = "connected_fut";
static const char __pyx_k_create_future[] = "create_future";
static const char __pyx_k_data_received[] = "data_received";
static const char __pyx_k_gen_stream_id[] = "gen_stream_id";
static const char __pyx_k_get_common_db[] = "get_common_db";
static const char __pyx_k_init_subclass[] = "__init_subclass__";
static const char __pyx_k_on_authorized[] = "on_authorized";
static const char __pyx_k_parse_version[] = "_parse_version";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_set_exception[] = "set_exception";
static const char __pyx_k_set_stream_id[] = "set_stream_id";
static const char __pyx_k_total_seconds[] = "total_seconds";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_CancelledError[] = "CancelledError";
static const char __pyx_k_ExecuteRequest[] = "ExecuteRequest";
static const char __pyx_k_PrepareRequest[] = "PrepareRequest";
static const char __pyx_k_TarantoolTuple[] = "TarantoolTuple";
static const char __pyx_k_get_extra_info[] = "get_extra_info";
static const char __pyx_k_id_name_engine[] = "<{} id={} name={} engine={}>";
static const char __pyx_k_last_stream_id[] = "_last_stream_id";
static const char __pyx_k_parse_metadata[] = "parse_metadata";
static const char __pyx_k_push_subscribe[] = "push_subscribe";
static const char __pyx_k_refetch_schema[] = "refetch_schema";
static const char __pyx_k_Field_not_found[] = "Field '{}' not found";
static const char __pyx_k_RollbackRequest[] = "RollbackRequest";
static const char __pyx_k_Space_not_found[] = "Space {} not found";
static const char __pyx_k_WriteBuffer_hex[] = "WriteBuffer.hex";
static const char __pyx_k_connection_lost[] = "connection_lost";
static const char __pyx_k_connection_made[] = "connection_made";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_request_timeout[] = "request_timeout";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_BITS_ALL_NOT_SET[] = "BITS_ALL_NOT_SET";
static const char __pyx_k_Db_set_stream_id[] = "Db.set_stream_id";
static const char __pyx_k_IProto_version_s[] = "IProto version: %s";
static const char __pyx_k_SchemaDummyIndex[] = "SchemaDummyIndex";
static const char __pyx_k_SchemaDummySpace[] = "SchemaDummySpace";
static const char __pyx_k_exceeded_timeout[] = "{} exceeded timeout";
static const char __pyx_k_sync_d_not_found[] = "sync %d not found";
static const char __pyx_k_add_done_callback[] = "add_done_callback";
static const char __pyx_k_return_exceptions[] = "return_exceptions";
static const char __pyx_k_s_Tarantool_s_d_s[] = "\\s*Tarantool\\s+([\\d.]+)\\s+.*";
static const char __pyx_k_Db___reduce_cython[] = "Db.__reduce_cython__";
static const char __pyx_k_IProto_auth_type_s[] = "IProto auth type: %s";
static const char __pyx_k_Part_len_must_be_2[] = "Part len must be 2";
static const char __pyx_k_UnicodeDecodeError[] = "UnicodeDecodeError";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_hex_locals_genexpr[] = "hex.<locals>.genexpr";
static const char __pyx_k_is_fully_connected[] = "is_fully_connected";
static const char __pyx_k_on_connection_lost[] = "on_connection_lost";
static const char __pyx_k_on_connection_made[] = "on_connection_made";
static const char __pyx_k_on_request_timeout[] = "_on_request_timeout";
static const char __pyx_k_sync_rowcount_data[] = "<{} sync={} rowcount={} data={}>";
static const char __pyx_k_asynctnt_exceptions[] = "asynctnt.exceptions";
static const char __pyx_k_auto_refetch_schema[] = "auto_refetch_schema";
static const char __pyx_k_check_schema_change[] = "check_schema_change";
static const char __pyx_k_ConnectionResetError[] = "ConnectionResetError";
static const char __pyx_k_Db___setstate_cython[] = "Db.__setstate_cython__";
static const char __pyx_k_PushIterator___anext[] = "PushIterator.__anext__";
static const char __pyx_k_Schema_fetch_timeout[] = "Schema fetch timeout";
static const char __pyx_k_TarantoolSchemaError[] = "TarantoolSchemaError";
static const char __pyx_k_VERSION_STRING_REGEX[] = "VERSION_STRING_REGEX";
static const char __pyx_k_on_request_completed[] = "_on_request_completed";
static const char __pyx_k_repr___locals_lambda[] = "__repr__.<locals>.<lambda>";
static const char __pyx_k_Field___reduce_cython[] = "Field.__reduce_cython__";
static const char __pyx_k_IProtoErrorStackFrame[] = "IProtoErrorStackFrame";
static const char __pyx_k_Metadata_fields_count[] = "<Metadata [fields_count={}]>";
static const char __pyx_k_Unexpected_ext_type_d[] = "Unexpected ext type: %d";
static const char __pyx_k_Unexpected_obj_type_s[] = "Unexpected obj type: %s";
static const char __pyx_k_BaseProtocol_create_db[] = "BaseProtocol.create_db";
static const char __pyx_k_ConnectionRefusedError[] = "ConnectionRefusedError";
static const char __pyx_k_Schema___reduce_cython[] = "Schema.__reduce_cython__";
static const char __pyx_k_TarantoolDatabaseError[] = "TarantoolDatabaseError";
static const char __pyx_k_asynctnt_iproto_db_pyx[] = "asynctnt/iproto/db.pyx";
static const char __pyx_k_create_future_fallback[] = "_create_future_fallback";
static const char __pyx_k_pyx_unpickle_IDRequest[] = "__pyx_unpickle_IDRequest";
static const char __pyx_k_Field___setstate_cython[] = "Field.__setstate_cython__";
static const char __pyx_k_Frame_type_code_message[] = "<Frame type={} code={} message={}>";
static const char __pyx_k_Space_with_id_not_found[] = "Space with id {} not found";
static const char __pyx_k_sid_id_name_type_unique[] = "<{} sid={}, id={}, name={}, type={}, unique={}>";
static const char __pyx_k_Available_Iterator_types[] = "\n        Available Iterator types\n    ";
static const char __pyx_k_CoreProtocol_get_version[] = "CoreProtocol.get_version";
static const char __pyx_k_Index_not_found_in_space[] = "Index {} not found in space {}/{}";
static const char __pyx_k_Metadata___reduce_cython[] = "Metadata.__reduce_cython__";
static const char __pyx_k_Response___reduce_cython[] = "Response.__reduce_cython__";
static const char __pyx_k_Schema___setstate_cython[] = "Schema.__setstate_cython__";
static const char __pyx_k_asynctnt_iproto_protocol[] = "asynctnt.iproto.protocol";
static const char __pyx_k_initial_read_buffer_size[] = "initial_read_buffer_size";
static const char __pyx_k_pyx_unpickle_AuthRequest[] = "__pyx_unpickle_AuthRequest";
static const char __pyx_k_pyx_unpickle_BaseRequest[] = "__pyx_unpickle_BaseRequest";
static const char __pyx_k_pyx_unpickle_CallRequest[] = "__pyx_unpickle_CallRequest";
static const char __pyx_k_pyx_unpickle_EvalRequest[] = "__pyx_unpickle_EvalRequest";
static const char __pyx_k_pyx_unpickle_IProtoError[] = "__pyx_unpickle_IProtoError";
static const char __pyx_k_pyx_unpickle_PingRequest[] = "__pyx_unpickle_PingRequest";
static const char __pyx_k_unexpected_exponent_type[] = "unexpected exponent type: {}";
static const char __pyx_k_CoreProtocol_is_connected[] = "CoreProtocol.is_connected";
static const char __pyx_k_IDRequest___reduce_cython[] = "IDRequest.__reduce_cython__";
static const char __pyx_k_Splice_offset_must_be_int[] = "Splice offset must be int";
static const char __pyx_k_pyx_unpickle_BeginRequest[] = "__pyx_unpickle_BeginRequest";
static const char __pyx_k_pyx_unpickle_CoreProtocol[] = "__pyx_unpickle_CoreProtocol";
static const char __pyx_k_pyx_unpickle_PushIterator[] = "__pyx_unpickle_PushIterator";
static const char __pyx_k_unknown_key_in_body_map_s[] = "unknown key in body map: %s";
static const char __pyx_k_BaseProtocol_get_common_db[] = "BaseProtocol.get_common_db";
static const char __pyx_k_CoreProtocol_data_received[] = "CoreProtocol.data_received";
static const char __pyx_k_Metadata___setstate_cython[] = "Metadata.__setstate_cython__";
static const char __pyx_k_ReadBuffer___reduce_cython[] = "ReadBuffer.__reduce_cython__";
static const char __pyx_k_Response___setstate_cython[] = "Response.__setstate_cython__";
static const char __pyx_k_TarantoolNotConnectedError[] = "TarantoolNotConnectedError";
static const char __pyx_k_Tarantool_is_not_connected[] = "Tarantool is not connected";
static const char __pyx_k_asynctnt_iproto_buffer_pyx[] = "asynctnt/iproto/buffer.pyx";
static const char __pyx_k_pyx_unpickle_CommitRequest[] = "__pyx_unpickle_CommitRequest";
static const char __pyx_k_pyx_unpickle_DeleteRequest[] = "__pyx_unpickle_DeleteRequest";
static const char __pyx_k_pyx_unpickle_InsertRequest[] = "__pyx_unpickle_InsertRequest";
static const char __pyx_k_pyx_unpickle_SelectRequest[] = "__pyx_unpickle_SelectRequest";
static const char __pyx_k_pyx_unpickle_UpdateRequest[] = "__pyx_unpickle_UpdateRequest";
static const char __pyx_k_pyx_unpickle_UpsertRequest[] = "__pyx_unpickle_UpsertRequest";
static const char __pyx_k_AuthRequest___reduce_cython[] = "AuthRequest.__reduce_cython__";
static const char __pyx_k_BaseProtocol_refetch_schema[] = "BaseProtocol.refetch_schema";
static const char __pyx_k_BaseRequest___reduce_cython[] = "BaseRequest.__reduce_cython__";
static const char __pyx_k_CallRequest___reduce_cython[] = "CallRequest.__reduce_cython__";
static const char __pyx_k_CoreProtocol__parse_version[] = "CoreProtocol._parse_version";
static const char __pyx_k_EvalRequest___reduce_cython[] = "EvalRequest.__reduce_cython__";
static const char __pyx_k_Field_name_type_is_nullable[] = "<Field name={} type={} is_nullable={}>";
static const char __pyx_k_IDRequest___setstate_cython[] = "IDRequest.__setstate_cython__";
static const char __pyx_k_IProtoError___reduce_cython[] = "IProtoError.__reduce_cython__";
static const char __pyx_k_IProto_features_available_s[] = "IProto features available: %s";
static const char __pyx_k_PingRequest___reduce_cython[] = "PingRequest.__reduce_cython__";
static const char __pyx_k_SchemaIndex___reduce_cython[] = "SchemaIndex.__reduce_cython__";
static const char __pyx_k_SchemaSpace___reduce_cython[] = "SchemaSpace.__reduce_cython__";
static const char __pyx_k_Splice_position_must_be_int[] = "Splice position must be int";
static const char __pyx_k_WriteBuffer___reduce_cython[] = "WriteBuffer.__reduce_cython__";
static const char __pyx_k_code_type_must_be_a_MP_UINT[] = "code type must be a MP_UINT";
static const char __pyx_k_field_name_must_not_be_None[] = "field.name must not be None";
static const char __pyx_k_pyx_unpickle_ExecuteRequest[] = "__pyx_unpickle_ExecuteRequest";
static const char __pyx_k_pyx_unpickle_PrepareRequest[] = "__pyx_unpickle_PrepareRequest";
static const char __pyx_k_sync_type_must_be_a_MP_UINT[] = "sync type must be a MP_UINT";
static const char __pyx_k_BaseProtocol___reduce_cython[] = "BaseProtocol.__reduce_cython__";
static const char __pyx_k_BeginRequest___reduce_cython[] = "BeginRequest.__reduce_cython__";
static const char __pyx_k_CoreProtocol___reduce_cython[] = "CoreProtocol.__reduce_cython__";
static const char __pyx_k_CoreProtocol_connection_lost[] = "CoreProtocol.connection_lost";
static const char __pyx_k_CoreProtocol_connection_made[] = "CoreProtocol.connection_made";
static const char __pyx_k_Header_key_must_be_a_MP_UINT[] = "Header key must be a MP_UINT";
static const char __pyx_k_Lost_connection_to_Tarantool[] = "Lost connection to Tarantool";
static const char __pyx_k_Offset_incorrect_Got_use_len[] = "Offset incorrect. Got: {}. use:{}, len:{}";
static const char __pyx_k_PushIterator___init___line_9[] = "PushIterator.__init__ (line 9)";
static const char __pyx_k_PushIterator___reduce_cython[] = "PushIterator.__reduce_cython__";
static const char __pyx_k_ReadBuffer___setstate_cython[] = "ReadBuffer.__setstate_cython__";
static const char __pyx_k_Unexpected_key_type_in_map_s[] = "Unexpected key type in map: %s";
static const char __pyx_k_asynctnt_iproto_protocol_pyx[] = "asynctnt/iproto/protocol.pyx";
static const char __pyx_k_asynctnt_iproto_response_pyx[] = "asynctnt/iproto/response.pyx";
static const char __pyx_k_errstr_type_must_be_a_MP_STR[] = "errstr type must be a MP_STR";
static const char __pyx_k_pyx_unpickle_RollbackRequest[] = "__pyx_unpickle_RollbackRequest";
static const char __pyx_k_AuthRequest___setstate_cython[] = "AuthRequest.__setstate_cython__";
static const char __pyx_k_BaseRequest___setstate_cython[] = "BaseRequest.__setstate_cython__";
static const char __pyx_k_CallRequest___setstate_cython[] = "CallRequest.__setstate_cython__";
static const char __pyx_k_CommitRequest___reduce_cython[] = "CommitRequest.__reduce_cython__";
static const char __pyx_k_DeleteRequest___reduce_cython[] = "DeleteRequest.__reduce_cython__";
static const char __pyx_k_EvalRequest___setstate_cython[] = "EvalRequest.__setstate_cython__";
static const char __pyx_k_IProtoError___setstate_cython[] = "IProtoError.__setstate_cython__";
static const char __pyx_k_InsertRequest___reduce_cython[] = "InsertRequest.__reduce_cython__";
static const char __pyx_k_PingRequest___setstate_cython[] = "PingRequest.__setstate_cython__";
static const char __pyx_k_SchemaIndex___setstate_cython[] = "SchemaIndex.__setstate_cython__";
static const char __pyx_k_SchemaSpace___setstate_cython[] = "SchemaSpace.__setstate_cython__";
static const char __pyx_k_SelectRequest___reduce_cython[] = "SelectRequest.__reduce_cython__";
static const char __pyx_k_Unknown_update_operation_type[] = "Unknown update operation type `{}`";
static const char __pyx_k_UpdateRequest___reduce_cython[] = "UpdateRequest.__reduce_cython__";
static const char __pyx_k_UpsertRequest___reduce_cython[] = "UpsertRequest.__reduce_cython__";
static const char __pyx_k_WriteBuffer___setstate_cython[] = "WriteBuffer.__setstate_cython__";
static const char __pyx_k_asynctnt_iproto_coreproto_pyx[] = "asynctnt/iproto/coreproto.pyx";
static const char __pyx_k_BaseProtocol___setstate_cython[] = "BaseProtocol.__setstate_cython__";
static const char __pyx_k_BeginRequest___setstate_cython[] = "BeginRequest.__setstate_cython__";
static const char __pyx_k_CoreProtocol___setstate_cython[] = "CoreProtocol.__setstate_cython__";
static const char __pyx_k_Creates_PushIterator_object_In[] = "\n            Creates PushIterator object. In order to receive push notifications\n            this iterator must be created.\n\n            Example:\n\n            .. code-block:: pycon\n\n                # tarantool function:\n                # function f()\n                #     box.session.push('hello')\n                #     return 'finished'\n                # end\n\n                >>> fut = conn.call('async_function', push_subscribe=True)\n                >>> async for item in PushIterator(fut):\n                ...     print(item)\n\n\n            :param fut: Future object returned from call_async, eval_sync\n                        functions\n            :type fut: asyncio.Future\n        ";
static const char __pyx_k_ExecuteRequest___reduce_cython[] = "ExecuteRequest.__reduce_cython__";
static const char __pyx_k_Lost_connection_to_Tarantool_2[] = "Lost connection to Tarantool: {}: {}";
static const char __pyx_k_Lost_connection_to_Tarantool_3[] = "Lost connection to Tarantool: {}";
static const char __pyx_k_PrepareRequest___reduce_cython[] = "PrepareRequest.__reduce_cython__";
static const char __pyx_k_PushIterator___setstate_cython[] = "PushIterator.__setstate_cython__";
static const char __pyx_k_Request_op_sync_schema_id_push[] = "<Request op={} sync={} schema_id={} push_subscribe={}>";
static const char __pyx_k_Response_body_must_be_a_MP_MAP[] = "Response body must be a MP_MAP";
static const char __pyx_k_pyx_unpickle_IProtoErrorStackF[] = "__pyx_unpickle_IProtoErrorStackFrame";
static const char __pyx_k_BaseProtocol__do_auth_locals_on[] = "BaseProtocol._do_auth.<locals>.on_authorized";
static const char __pyx_k_CommitRequest___setstate_cython[] = "CommitRequest.__setstate_cython__";
static const char __pyx_k_CoreProtocol_is_fully_connected[] = "CoreProtocol.is_fully_connected";
static const char __pyx_k_DeleteRequest___setstate_cython[] = "DeleteRequest.__setstate_cython__";
static const char __pyx_k_Field_map_must_contain_at_least[] = "Field map must contain at least 1 element - field_name";
static const char __pyx_k_Index_must_be_either_str_or_int[] = "Index must be either str or int, got: {}";
static const char __pyx_k_InsertRequest___setstate_cython[] = "InsertRequest.__setstate_cython__";
static const char __pyx_k_Iterator_is_of_unsupported_type[] = "Iterator is of unsupported type (asynctnt.Iterator, int, str)";
static const char __pyx_k_Operation_type_must_of_a_str_or[] = "Operation type must of a str or bytes type";
static const char __pyx_k_RollbackRequest___reduce_cython[] = "RollbackRequest.__reduce_cython__";
static const char __pyx_k_Schema_fetch_stopped_connection[] = "Schema fetch stopped: connection is closed";
static const char __pyx_k_SelectRequest___setstate_cython[] = "SelectRequest.__setstate_cython__";
static const char __pyx_k_Space_must_be_either_str_or_int[] = "Space must be either str or int, got: {}";
static const char __pyx_k_UpdateRequest___setstate_cython[] = "UpdateRequest.__setstate_cython__";
static const char __pyx_k_UpsertRequest___setstate_cython[] = "UpsertRequest.__setstate_cython__";
static const char __pyx_k_invalid_datetime_size_got_extra[] = "invalid datetime size. got {} extra bytes";
static const char __pyx_k_iproto_error_stack_frame_fields[] = "iproto_error stack frame fields must be a map, but got ";
static const char __pyx_k_on_data_received_expected_bytes[] = "_on_data_received: expected bytes or bytearray object, got {}";
static const char __pyx_k_query_must_be_either_str_or_int[] = "query must be either str or int";
static const char __pyx_k_sequence_must_be_either_list_or[] = "sequence must be either list or tuple";
static const char __pyx_k_the_buffer_is_in_read_only_mode[] = "the buffer is in read-only mode";
static const char __pyx_k_unexpected_type_of_part_must_be[] = "unexpected type of part: {}. must be either dict or list/tuple";
static const char __pyx_k_BaseProtocol__create_future_fall[] = "BaseProtocol._create_future_fallback";
static const char __pyx_k_BaseProtocol__do_fetch_schema_lo[] = "BaseProtocol._do_fetch_schema.<locals>.on_fetch";
static const char __pyx_k_BaseProtocol__do_id_locals_on_id[] = "BaseProtocol._do_id.<locals>.on_id";
static const char __pyx_k_BaseProtocol__on_request_complet[] = "BaseProtocol._on_request_completed";
static const char __pyx_k_BaseProtocol__on_request_timeout[] = "BaseProtocol._on_request_timeout";
static const char __pyx_k_Cannot_pop_push_from_a_non_async[] = "Cannot pop push from a non-async response";
static const char __pyx_k_Cannot_use_iter_with_PushIterato[] = "Cannot use iter with PushIterator - use aiter";
static const char __pyx_k_Cannot_use_next_with_PushIterato[] = "Cannot use next with PushIterator - use anext";
static const char __pyx_k_Deallocating_buffer_with_attache[] = "Deallocating buffer with attached memoryviews";
static const char __pyx_k_Error_happened_while_parsing_sch[] = "Error happened while parsing schema. Space, fields and index names currently not working. Please file an issue at https://github.com/igorcoding/asynctnt";
static const char __pyx_k_ExecuteRequest___setstate_cython[] = "ExecuteRequest.__setstate_cython__";
static const char __pyx_k_Field_d_of_space_s_is_not_in_spa[] = "Field #%d of space %s is not in space format definition";
static const char __pyx_k_Field_s_in_supplied_dict_is_unkn[] = "Field '%s' in supplied dict is unknown as a tuple field for selected index. Skipping.";
static const char __pyx_k_Future_is_invalid_Make_sure_to_c[] = "Future is invalid. Make sure to call with a future returned from a method with push_subscribe=True flag";
static const char __pyx_k_IPROTO_ERROR_type_must_be_a_MP_M[] = "IPROTO_ERROR type must be a MP_MAP";
static const char __pyx_k_IPROTO_FIELD_SPAN_must_be_either[] = "IPROTO_FIELD_SPAN must be either STR or NIL";
static const char __pyx_k_IPROTO_STMT_ID_type_must_be_a_MP[] = "IPROTO_STMT_ID type must be a MP_UINT, but got ";
static const char __pyx_k_IProtoErrorStackFrame___reduce_c[] = "IProtoErrorStackFrame.__reduce_cython__";
static const char __pyx_k_IProtoErrorStackFrame___setstate[] = "IProtoErrorStackFrame.__setstate_cython__";
static const char __pyx_k_Impossible_condition_happened_Pl[] = "Impossible condition happened. Please file a bug to https://github.com/igorcoding/asynctnt";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))";
static const char __pyx_k_Index_s_not_found_in_space_s_s_C[] = "Index %s not found in space %s/%s. Creating dummy.";
static const char __pyx_k_Operation_field_no_must_be_of_ei[] = "Operation field_no must be of either int or str type";
static const char __pyx_k_Operation_length_must_be_at_leas[] = "Operation length must be at least 3";
static const char __pyx_k_PrepareRequest___setstate_cython[] = "PrepareRequest.__setstate_cython__";
static const char __pyx_k_Response_header_must_be_a_MP_MAP[] = "Response header must be a MP_MAP";
static const char __pyx_k_RollbackRequest___setstate_cytho[] = "RollbackRequest.__setstate_cython__";
static const char __pyx_k_Single_operation_must_be_a_tuple[] = "Single operation must be a tuple or list";
static const char __pyx_k_Space_s_not_found_Creating_dummy[] = "Space %s not found. Creating dummy.";
static const char __pyx_k_Splice_operation_must_have_lengt[] = "Splice operation must have length of 5, but got: {}";
static const char __pyx_k_Tarantool_s_s_Authorization_fail[] = "Tarantool[%s:%s] Authorization failed: %s";
static const char __pyx_k_Tarantool_s_s_Authorized_success[] = "Tarantool[%s:%s] Authorized successfully";
static const char __pyx_k_Tarantool_s_s_Schema_fetch_faile[] = "Tarantool[%s:%s] Schema fetch failed: %s";
static const char __pyx_k_Tarantool_s_s_Schema_fetch_succe[] = "Tarantool[%s:%s] Schema fetch succeeded. Version: %d, Spaces: %d, Indexes: %d.";
static const char __pyx_k_Tarantool_s_s_identification_fai[] = "Tarantool[%s:%s] identification failed: %s";
static const char __pyx_k_Tarantool_s_s_identified_success[] = "Tarantool[%s:%s] identified successfully";
static const char __pyx_k_Time_zones_are_not_available_fro[] = "Time zones are not available from the C-API.";
static const char __pyx_k_Tuple_must_be_an_array_when_deco[] = "Tuple must be an array when decoding as TarantoolTuple";
static const char __pyx_k_Type_is_not_supported_for_encodi[] = "Type `{}` is not supported for encoding";
static const char __pyx_k_Unknown_key_with_code_d_in_heade[] = "Unknown key with code '%d' in header. Skipping.";
static const char __pyx_k_body_data_type_must_be_a_MP_ARRA[] = "body data type must be a MP_ARRAY";
static const char __pyx_k_encoding_must_be_either_str_or_b[] = "encoding must be either str or bytes";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_schema_id_type_must_be_a_MP_UINT[] = "schema_id type must be a MP_UINT";
static const char __pyx_k_self_execute_cannot_be_converted[] = "self.execute cannot be converted to a Python object for pickling";
static const char __pyx_k_sequence_must_be_either_list_tup[] = "sequence must be either list, tuple or dict";
static const char __pyx_k_unknown_iproto_error_map_field_w[] = "unknown iproto_error map field with key ";
static const char __pyx_k_unknown_iproto_error_stack_eleme[] = "unknown iproto_error stack element with key ";
static const char __pyx_k_unknown_key_in_metadata_decoding[] = "unknown key in metadata decoding: %d";
static const char __pyx_k_unknown_key_in_sql_info_decoding[] = "unknown key in sql info decoding: %d";
static const char __pyx_k_Field_map_must_contain_at_least_2[] = "Field map must contain at least 1 element - rowcount";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0x2d5a81d, 0xac90343, 0xcd9f35b) = (args, check_schema_change, func_name, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0xafafa67, 0x3c16c97, 0x904102b) = (args, check_schema_change, expression, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_4[] = "Incompatible checksums (0x%x vs (0x8e2dc2b, 0x62f76f3, 0x57fe0b4) = (check_schema_change, index, iterator, key, limit, offset, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_5[] = "Incompatible checksums (0x%x vs (0xb91d1f2, 0xff182e4, 0x15f8d7d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_6[] = "Incompatible checksums (0x%x vs (0xf383ff7, 0x8ac4f18, 0xe9836c0) = (check_schema_change, index, key, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_7[] = "Incompatible checksums (0x%x vs (0x815e48a, 0x8a9c73b, 0x180001a) = (check_schema_change, index, key, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_8[] = "Incompatible checksums (0x%x vs (0xc823323, 0xea6e875, 0x0493034) = (check_schema_change, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_9[] = "Incompatible checksums (0x%x vs (0xd0934a3, 0xcc899a0, 0x7b3178d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_10[] = "Incompatible checksums (0x%x vs (0x314ca88, 0x64ec1ad, 0xd228fa0) = (args, check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_11[] = "Incompatible checksums (0x%x vs (0x335e105, 0xf5a1799, 0xa2d5d5d) = (check_schema_change, op, parse_as_tuples, parse_metadata, password, push_subscribe, salt, schema_id, space, stream_id, sync, timeout_handle, username, waiter))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_12[] = "Incompatible checksums (0x%x vs (0x30f95df, 0xa4f125c, 0xfc952a6) = (check_schema_change, isolation, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, tx_timeout, waiter))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_13[] = "Incompatible checksums (0x%x vs (0xab276bf, 0xb7ed9c7, 0x6438658) = (code, err_no, error_type, fields, file, line, message))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_14[] = "Incompatible checksums (0x%x vs (0xeafe895, 0xd0fa554, 0x04a7503) = (trace))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_15[] = "Incompatible checksums (0x%x vs (0x98137af, 0xbb11453, 0xd6d3964) = (_fut, _request, _response))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_16[] = "Incompatible checksums (0x%x vs (0x89d360e, 0xe0a6eac, 0xf188771) = (con_state, encoding, host, port, rbuf, salt, state, version))";
/* #### Code section: decls ### */
static int __pyx_pf_8asynctnt_6iproto_8protocol_5Field___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_2__repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_4name___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_4type___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_9collation___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_11is_nullable___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_16is_autoincrement___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_4span___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_8Metadata___cinit__(struct C_Metadata *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_2__repr__(struct C_Metadata *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_6fields___get__(struct C_Metadata *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_11name_id_map___get__(struct C_Metadata *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_4__reduce_cython__(CYTHON_UNUSED struct C_Metadata *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_6__setstate_cython__(CYTHON_UNUSED struct C_Metadata *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_2__repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_3sid___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_3iid___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_4name___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_10index_type___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_6unique___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_8metadata___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_2__repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_3sid___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_5owner___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_4name___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_6engine___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_11field_count___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_5flags___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_8metadata___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_7indexes___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_6Schema___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, int __pyx_v_schema_id); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6Schema_2__repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6Schema_6spaces___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6Schema_2id___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6Schema_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6Schema_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self); /* proto */
static void __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_2__dealloc__(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_4__getbuffer__(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /* proto */
static void __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_6__releasebuffer__(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_3hex_genexpr(PyObject *__pyx_self, Py_ssize_t __pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_8hex(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self); /* proto */
static void __pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer_2__dealloc__(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11BaseRequest___repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11BaseRequest_2__reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11BaseRequest_4__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11PingRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11PingRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11CallRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11CallRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11EvalRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11EvalRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13SelectRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13SelectRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13InsertRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13InsertRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13DeleteRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13DeleteRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13UpdateRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13UpdateRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13UpsertRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13UpsertRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_14PrepareRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_14PrepareRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_14ExecuteRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_14ExecuteRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_9IDRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_9IDRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11AuthRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11AuthRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BeginRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BeginRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13CommitRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13CommitRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_15RollbackRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_15RollbackRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame___repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_10error_type___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4file___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4line___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_7message___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6err_no___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4code___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6fields___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_2__reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11IProtoError_5trace___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11IProtoError___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11IProtoError_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_8Response___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_2__repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_4sync___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_4code___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_11return_code___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_9schema_id___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_8rowcount___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_7stmt_id___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_4done(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_8asynctnt_6iproto_8protocol_8Response_6__len__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_8__getitem__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_10__iter__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_6errmsg___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_5error___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_4body___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_8encoding___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_8metadata___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_6params___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_12params_count___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_17autoincrement_ids___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_2_q___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_11_push_event___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_2Db___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_9stream_id___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_2set_stream_id(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, int __pyx_v_stream_id); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_4ping(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_6call16(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_func_name, PyObject *__pyx_v_args, float __pyx_v_timeout, int __pyx_v_push_subscribe); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_8call(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_func_name, PyObject *__pyx_v_args, float __pyx_v_timeout, int __pyx_v_push_subscribe); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_10eval(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_expression, PyObject *__pyx_v_args, float __pyx_v_timeout, int __pyx_v_push_subscribe); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_12select(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_key, int __pyx_v_offset, int __pyx_v_limit, PyObject *__pyx_v_index, PyObject *__pyx_v_iterator, float __pyx_v_timeout, int __pyx_v_check_schema_change); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_14insert(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_t, int __pyx_v_replace, float __pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_16replace(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_t, float __pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_18delete(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_key, PyObject *__pyx_v_index, float __pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_20update(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_key, PyObject *__pyx_v_operations, PyObject *__pyx_v_index, float __pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_22upsert(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_t, PyObject *__pyx_v_operations, float __pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_24execute(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_query, PyObject *__pyx_v_args, int __pyx_v_parse_metadata, float __pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_26prepare(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_query, int __pyx_v_parse_metadata, float __pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_28begin(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, uint32_t __pyx_v_isolation, float __pyx_v_tx_timeout, float __pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_30commit(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_32rollback(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_34__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_36__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator___init__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self, PyObject *__pyx_v_fut); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_2__iter__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_4__next__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_6__aiter__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_8__anext__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_8response___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_11__reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_13__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol___init__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_host, PyObject *__pyx_v_port, PyObject *__pyx_v_encoding, PyObject *__pyx_v_initial_read_buffer_size); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_2is_connected(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_4is_fully_connected(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_6get_version(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_8_parse_version(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_version); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_10data_received(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_12connection_made(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_transport); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_14connection_lost(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_exc); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_16__reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_18__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol___init__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_host, PyObject *__pyx_v_port, PyObject *__pyx_v_username, PyObject *__pyx_v_password, PyObject *__pyx_v_fetch_schema, PyObject *__pyx_v_auto_refetch_schema, PyObject *__pyx_v_connected_fut, PyObject *__pyx_v_on_connection_made, PyObject *__pyx_v_on_connection_lost, PyObject *__pyx_v_loop, PyObject *__pyx_v_request_timeout, PyObject *__pyx_v_encoding, PyObject *__pyx_v_initial_read_buffer_size); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_2_create_future_fallback(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_6schema___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_9schema_id___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_6_do_id_on_id(PyObject *__pyx_self, PyObject *__pyx_v_f); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_8_do_auth_on_authorized(PyObject *__pyx_self, PyObject *__pyx_v_f); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_16_do_fetch_schema_on_fetch(PyObject *__pyx_self, PyObject *__pyx_v_f); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_4_on_request_timeout(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_waiter); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_6_on_request_completed(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_fut); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_8create_db(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, int __pyx_v_gen_stream_id); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_10get_common_db(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_12refetch_schema(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol___pyx_unpickle_BaseRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2__pyx_unpickle_PingRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_4__pyx_unpickle_CallRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6__pyx_unpickle_EvalRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8__pyx_unpickle_SelectRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_10__pyx_unpickle_InsertRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12__pyx_unpickle_DeleteRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_14__pyx_unpickle_UpdateRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_16__pyx_unpickle_UpsertRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_18__pyx_unpickle_PrepareRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_20__pyx_unpickle_ExecuteRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_22__pyx_unpickle_IDRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_24__pyx_unpickle_AuthRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_26__pyx_unpickle_BeginRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_28__pyx_unpickle_CommitRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_30__pyx_unpickle_RollbackRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_32__pyx_unpickle_IProtoErrorStackFrame(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_34__pyx_unpickle_IProtoError(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_36__pyx_unpickle_PushIterator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_38__pyx_unpickle_CoreProtocol(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_Field(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_Metadata(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaIndex(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaSpace(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaDummySpace(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_Schema(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_WriteBuffer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_ReadBuffer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_PingRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_CallRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_EvalRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_SelectRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_InsertRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_DeleteRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_UpdateRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_UpsertRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_PrepareRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_ExecuteRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_IDRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_AuthRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_BeginRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_CommitRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_RollbackRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_IProtoError(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_Response(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_Db(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_PushIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_CoreProtocol(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_BaseProtocol(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_8datetime_date;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_time;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_datetime;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_timedelta;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_tzinfo;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_Field;
  PyObject *C_Metadata_Type;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_Schema;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_BaseRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_PingRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_CallRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_EvalRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_SelectRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_InsertRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_IDRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_AuthRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_BeginRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_CommitRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_IProtoError;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_Response;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_Db;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_PushIterator;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth;
  PyObject *__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema;
  #endif
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_Field;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_Schema;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_Response;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_Db;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth;
  PyTypeObject *__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema;
  PyObject *__pyx_kp_u_02x;
  PyObject *__pyx_n_s_AF_UNIX;
  PyObject *__pyx_n_u_AF_UNIX;
  PyObject *__pyx_n_s_ALL;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_s_AttributeError;
  PyObject *__pyx_n_s_AuthRequest;
  PyObject *__pyx_n_s_AuthRequest___reduce_cython;
  PyObject *__pyx_n_s_AuthRequest___setstate_cython;
  PyObject *__pyx_kp_s_Available_Iterator_types;
  PyObject *__pyx_n_s_BITS_ALL_NOT_SET;
  PyObject *__pyx_n_s_BITS_ALL_SET;
  PyObject *__pyx_n_s_BITS_ANY_SET;
  PyObject *__pyx_n_s_BaseProtocol;
  PyObject *__pyx_n_s_BaseProtocol___reduce_cython;
  PyObject *__pyx_n_s_BaseProtocol___setstate_cython;
  PyObject *__pyx_n_s_BaseProtocol__create_future_fall;
  PyObject *__pyx_n_s_BaseProtocol__do_auth_locals_on;
  PyObject *__pyx_n_s_BaseProtocol__do_fetch_schema_lo;
  PyObject *__pyx_n_s_BaseProtocol__do_id_locals_on_id;
  PyObject *__pyx_n_s_BaseProtocol__on_request_complet;
  PyObject *__pyx_n_s_BaseProtocol__on_request_timeout;
  PyObject *__pyx_n_s_BaseProtocol_create_db;
  PyObject *__pyx_n_s_BaseProtocol_get_common_db;
  PyObject *__pyx_n_s_BaseProtocol_refetch_schema;
  PyObject *__pyx_n_s_BaseRequest;
  PyObject *__pyx_n_s_BaseRequest___reduce_cython;
  PyObject *__pyx_n_s_BaseRequest___setstate_cython;
  PyObject *__pyx_n_s_BeginRequest;
  PyObject *__pyx_n_s_BeginRequest___reduce_cython;
  PyObject *__pyx_n_s_BeginRequest___setstate_cython;
  PyObject *__pyx_n_s_BufferError;
  PyObject *__pyx_n_s_CallRequest;
  PyObject *__pyx_n_s_CallRequest___reduce_cython;
  PyObject *__pyx_n_s_CallRequest___setstate_cython;
  PyObject *__pyx_n_s_CancelledError;
  PyObject *__pyx_kp_u_Cannot_pop_push_from_a_non_async;
  PyObject *__pyx_kp_u_Cannot_use_iter_with_PushIterato;
  PyObject *__pyx_kp_u_Cannot_use_next_with_PushIterato;
  PyObject *__pyx_n_s_CommitRequest;
  PyObject *__pyx_n_s_CommitRequest___reduce_cython;
  PyObject *__pyx_n_s_CommitRequest___setstate_cython;
  PyObject *__pyx_n_s_ConnectionRefusedError;
  PyObject *__pyx_n_s_ConnectionResetError;
  PyObject *__pyx_n_s_CoreProtocol;
  PyObject *__pyx_n_s_CoreProtocol___reduce_cython;
  PyObject *__pyx_n_s_CoreProtocol___setstate_cython;
  PyObject *__pyx_n_s_CoreProtocol__parse_version;
  PyObject *__pyx_n_s_CoreProtocol_connection_lost;
  PyObject *__pyx_n_s_CoreProtocol_connection_made;
  PyObject *__pyx_n_s_CoreProtocol_data_received;
  PyObject *__pyx_n_s_CoreProtocol_get_version;
  PyObject *__pyx_n_s_CoreProtocol_is_connected;
  PyObject *__pyx_n_s_CoreProtocol_is_fully_connected;
  PyObject *__pyx_kp_u_Creates_PushIterator_object_In;
  PyObject *__pyx_n_s_Db;
  PyObject *__pyx_n_s_Db___reduce_cython;
  PyObject *__pyx_n_s_Db___setstate_cython;
  PyObject *__pyx_n_s_Db_begin;
  PyObject *__pyx_n_s_Db_call;
  PyObject *__pyx_n_s_Db_call16;
  PyObject *__pyx_n_s_Db_commit;
  PyObject *__pyx_n_s_Db_delete;
  PyObject *__pyx_n_s_Db_eval;
  PyObject *__pyx_n_s_Db_execute;
  PyObject *__pyx_n_s_Db_insert;
  PyObject *__pyx_n_s_Db_ping;
  PyObject *__pyx_n_s_Db_prepare;
  PyObject *__pyx_n_s_Db_replace;
  PyObject *__pyx_n_s_Db_rollback;
  PyObject *__pyx_n_s_Db_select;
  PyObject *__pyx_n_s_Db_set_stream_id;
  PyObject *__pyx_n_s_Db_update;
  PyObject *__pyx_n_s_Db_upsert;
  PyObject *__pyx_kp_u_Deallocating_buffer_with_attache;
  PyObject *__pyx_n_s_Decimal;
  PyObject *__pyx_n_s_DeleteRequest;
  PyObject *__pyx_n_s_DeleteRequest___reduce_cython;
  PyObject *__pyx_n_s_DeleteRequest___setstate_cython;
  PyObject *__pyx_n_s_EQ;
  PyObject *__pyx_kp_u_Error_happened_while_parsing_sch;
  PyObject *__pyx_n_s_EvalRequest;
  PyObject *__pyx_n_s_EvalRequest___reduce_cython;
  PyObject *__pyx_n_s_EvalRequest___setstate_cython;
  PyObject *__pyx_n_s_Event;
  PyObject *__pyx_n_s_ExecuteRequest;
  PyObject *__pyx_n_s_ExecuteRequest___reduce_cython;
  PyObject *__pyx_n_s_ExecuteRequest___setstate_cython;
  PyObject *__pyx_n_s_Field;
  PyObject *__pyx_n_s_Field___reduce_cython;
  PyObject *__pyx_n_s_Field___setstate_cython;
  PyObject *__pyx_kp_u_Field_d_of_space_s_is_not_in_spa;
  PyObject *__pyx_kp_u_Field_map_must_contain_at_least;
  PyObject *__pyx_kp_u_Field_map_must_contain_at_least_2;
  PyObject *__pyx_kp_u_Field_name_type_is_nullable;
  PyObject *__pyx_kp_u_Field_not_found;
  PyObject *__pyx_kp_u_Field_s_in_supplied_dict_is_unkn;
  PyObject *__pyx_kp_u_Frame_type_code_message;
  PyObject *__pyx_n_s_Future;
  PyObject *__pyx_kp_u_Future_is_invalid_Make_sure_to_c;
  PyObject *__pyx_n_s_GE;
  PyObject *__pyx_n_s_GT;
  PyObject *__pyx_kp_u_Header_key_must_be_a_MP_UINT;
  PyObject *__pyx_n_s_IDRequest;
  PyObject *__pyx_n_s_IDRequest___reduce_cython;
  PyObject *__pyx_n_s_IDRequest___setstate_cython;
  PyObject *__pyx_n_s_IPPROTO_TCP;
  PyObject *__pyx_kp_u_IPROTO_ERROR_type_must_be_a_MP_M;
  PyObject *__pyx_kp_u_IPROTO_FIELD_SPAN_must_be_either;
  PyObject *__pyx_kp_u_IPROTO_STMT_ID_type_must_be_a_MP;
  PyObject *__pyx_n_s_IProtoError;
  PyObject *__pyx_n_s_IProtoErrorStackFrame;
  PyObject *__pyx_n_s_IProtoErrorStackFrame___reduce_c;
  PyObject *__pyx_n_s_IProtoErrorStackFrame___setstate;
  PyObject *__pyx_n_s_IProtoError___reduce_cython;
  PyObject *__pyx_n_s_IProtoError___setstate_cython;
  PyObject *__pyx_kp_u_IProto_auth_type_s;
  PyObject *__pyx_kp_u_IProto_features_available_s;
  PyObject *__pyx_kp_u_IProto_version_s;
  PyObject *__pyx_kp_u_Impossible_condition_happened_Pl;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9;
  PyObject *__pyx_n_s_IndexError;
  PyObject *__pyx_kp_u_Index_must_be_either_str_or_int;
  PyObject *__pyx_kp_u_Index_not_found_in_space;
  PyObject *__pyx_kp_u_Index_s_not_found_in_space_s_s_C;
  PyObject *__pyx_n_s_InsertRequest;
  PyObject *__pyx_n_s_InsertRequest___reduce_cython;
  PyObject *__pyx_n_s_InsertRequest___setstate_cython;
  PyObject *__pyx_n_s_IntEnum;
  PyObject *__pyx_n_s_Iterator;
  PyObject *__pyx_kp_u_Iterator_is_of_unsupported_type;
  PyObject *__pyx_n_s_KeyError;
  PyObject *__pyx_n_s_LE;
  PyObject *__pyx_n_s_LT;
  PyObject *__pyx_kp_u_Lost_connection_to_Tarantool;
  PyObject *__pyx_kp_u_Lost_connection_to_Tarantool_2;
  PyObject *__pyx_kp_u_Lost_connection_to_Tarantool_3;
  PyObject *__pyx_n_s_MemoryError;
  PyObject *__pyx_n_s_Metadata;
  PyObject *__pyx_n_s_Metadata___reduce_cython;
  PyObject *__pyx_n_s_Metadata___setstate_cython;
  PyObject *__pyx_kp_u_Metadata_fields_count;
  PyObject *__pyx_n_s_NEIGHBOR;
  PyObject *__pyx_n_s_OVERLAPS;
  PyObject *__pyx_kp_u_Offset_incorrect_Got_use_len;
  PyObject *__pyx_kp_u_Operation_field_no_must_be_of_ei;
  PyObject *__pyx_kp_u_Operation_length_must_be_at_leas;
  PyObject *__pyx_kp_u_Operation_type_must_of_a_str_or;
  PyObject *__pyx_n_s_Optional;
  PyObject *__pyx_kp_u_Part_len_must_be_2;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_n_s_PingRequest;
  PyObject *__pyx_n_s_PingRequest___reduce_cython;
  PyObject *__pyx_n_s_PingRequest___setstate_cython;
  PyObject *__pyx_n_s_PrepareRequest;
  PyObject *__pyx_n_s_PrepareRequest___reduce_cython;
  PyObject *__pyx_n_s_PrepareRequest___setstate_cython;
  PyObject *__pyx_n_s_Protocol;
  PyObject *__pyx_n_s_PushIterator;
  PyObject *__pyx_n_s_PushIterator___anext;
  PyObject *__pyx_kp_u_PushIterator___init___line_9;
  PyObject *__pyx_n_s_PushIterator___reduce_cython;
  PyObject *__pyx_n_s_PushIterator___setstate_cython;
  PyObject *__pyx_n_s_REQ;
  PyObject *__pyx_n_s_ReadBuffer;
  PyObject *__pyx_n_s_ReadBuffer___reduce_cython;
  PyObject *__pyx_n_s_ReadBuffer___setstate_cython;
  PyObject *__pyx_kp_u_Request_op_sync_schema_id_push;
  PyObject *__pyx_n_s_Response;
  PyObject *__pyx_n_s_Response___reduce_cython;
  PyObject *__pyx_n_s_Response___setstate_cython;
  PyObject *__pyx_kp_u_Response_body_must_be_a_MP_MAP;
  PyObject *__pyx_n_s_Response_done;
  PyObject *__pyx_kp_u_Response_header_must_be_a_MP_MAP;
  PyObject *__pyx_n_s_RollbackRequest;
  PyObject *__pyx_n_s_RollbackRequest___reduce_cython;
  PyObject *__pyx_n_s_RollbackRequest___setstate_cytho;
  PyObject *__pyx_n_s_RuntimeError;
  PyObject *__pyx_n_s_Schema;
  PyObject *__pyx_n_s_SchemaDummyIndex;
  PyObject *__pyx_n_s_SchemaDummySpace;
  PyObject *__pyx_n_s_SchemaIndex;
  PyObject *__pyx_n_s_SchemaIndex___reduce_cython;
  PyObject *__pyx_n_s_SchemaIndex___setstate_cython;
  PyObject *__pyx_n_s_SchemaSpace;
  PyObject *__pyx_n_s_SchemaSpace___reduce_cython;
  PyObject *__pyx_n_s_SchemaSpace___setstate_cython;
  PyObject *__pyx_n_s_Schema___reduce_cython;
  PyObject *__pyx_n_s_Schema___setstate_cython;
  PyObject *__pyx_kp_u_Schema_fetch_stopped_connection;
  PyObject *__pyx_kp_u_Schema_fetch_timeout;
  PyObject *__pyx_kp_u_Schema_spaces;
  PyObject *__pyx_n_s_SelectRequest;
  PyObject *__pyx_n_s_SelectRequest___reduce_cython;
  PyObject *__pyx_n_s_SelectRequest___setstate_cython;
  PyObject *__pyx_kp_u_Single_operation_must_be_a_tuple;
  PyObject *__pyx_kp_u_Space_must_be_either_str_or_int;
  PyObject *__pyx_kp_u_Space_not_found;
  PyObject *__pyx_kp_u_Space_s_not_found_Creating_dummy;
  PyObject *__pyx_kp_u_Space_with_id_not_found;
  PyObject *__pyx_kp_u_Splice_offset_must_be_int;
  PyObject *__pyx_kp_u_Splice_operation_must_have_lengt;
  PyObject *__pyx_kp_u_Splice_position_must_be_int;
  PyObject *__pyx_n_s_TCP_NODELAY;
  PyObject *__pyx_n_s_TarantoolDatabaseError;
  PyObject *__pyx_n_s_TarantoolNotConnectedError;
  PyObject *__pyx_n_s_TarantoolSchemaError;
  PyObject *__pyx_n_s_TarantoolTuple;
  PyObject *__pyx_kp_u_Tarantool_is_not_connected;
  PyObject *__pyx_kp_u_Tarantool_s_s_Authorization_fail;
  PyObject *__pyx_kp_u_Tarantool_s_s_Authorized_success;
  PyObject *__pyx_kp_u_Tarantool_s_s_Schema_fetch_faile;
  PyObject *__pyx_kp_u_Tarantool_s_s_Schema_fetch_succe;
  PyObject *__pyx_kp_u_Tarantool_s_s_identification_fai;
  PyObject *__pyx_kp_u_Tarantool_s_s_identified_success;
  PyObject *__pyx_kp_u_Time_zones_are_not_available_fro;
  PyObject *__pyx_n_s_TimeoutError;
  PyObject *__pyx_kp_u_Tuple_must_be_an_array_when_deco;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_kp_u_Type_is_not_supported_for_encodi;
  PyObject *__pyx_n_s_UUID;
  PyObject *__pyx_kp_u_Unexpected_ext_type_d;
  PyObject *__pyx_kp_u_Unexpected_key_type_in_map_s;
  PyObject *__pyx_kp_u_Unexpected_obj_type_s;
  PyObject *__pyx_n_s_UnicodeDecodeError;
  PyObject *__pyx_kp_u_Unknown_key_with_code_d_in_heade;
  PyObject *__pyx_kp_u_Unknown_update_operation_type;
  PyObject *__pyx_n_s_UpdateRequest;
  PyObject *__pyx_n_s_UpdateRequest___reduce_cython;
  PyObject *__pyx_n_s_UpdateRequest___setstate_cython;
  PyObject *__pyx_n_s_UpsertRequest;
  PyObject *__pyx_n_s_UpsertRequest___reduce_cython;
  PyObject *__pyx_n_s_UpsertRequest___setstate_cython;
  PyObject *__pyx_n_s_VERSION_STRING_REGEX;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_n_s_WriteBuffer;
  PyObject *__pyx_n_s_WriteBuffer___reduce_cython;
  PyObject *__pyx_n_s_WriteBuffer___setstate_cython;
  PyObject *__pyx_n_s_WriteBuffer_hex;
  PyObject *__pyx_kp_u__14;
  PyObject *__pyx_kp_u__17;
  PyObject *__pyx_kp_u__18;
  PyObject *__pyx_kp_u__19;
  PyObject *__pyx_kp_u__2;
  PyObject *__pyx_n_s__211;
  PyObject *__pyx_kp_u__3;
  PyObject *__pyx_kp_u__5;
  PyObject *__pyx_n_s__65;
  PyObject *__pyx_n_s_add_done_callback;
  PyObject *__pyx_n_s_anext;
  PyObject *__pyx_n_s_append;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_as_tuple;
  PyObject *__pyx_n_u_ascii;
  PyObject *__pyx_n_s_asyncio;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_asyncio_tasks;
  PyObject *__pyx_n_s_asynctnt;
  PyObject *__pyx_n_s_asynctnt_exceptions;
  PyObject *__pyx_kp_s_asynctnt_iproto_buffer_pyx;
  PyObject *__pyx_kp_s_asynctnt_iproto_coreproto_pyx;
  PyObject *__pyx_kp_s_asynctnt_iproto_db_pyx;
  PyObject *__pyx_n_s_asynctnt_iproto_protocol;
  PyObject *__pyx_kp_s_asynctnt_iproto_protocol_pyx;
  PyObject *__pyx_kp_s_asynctnt_iproto_response_pyx;
  PyObject *__pyx_n_s_asynctnt_log;
  PyObject *__pyx_n_s_auto_refetch_schema;
  PyObject *__pyx_n_s_await;
  PyObject *__pyx_n_s_b64decode;
  PyObject *__pyx_n_s_base64;
  PyObject *__pyx_n_s_begin;
  PyObject *__pyx_kp_u_body_data_type_must_be_a_MP_ARRA;
  PyObject *__pyx_n_s_bytes;
  PyObject *__pyx_n_s_call;
  PyObject *__pyx_n_s_call16;
  PyObject *__pyx_n_s_call_later;
  PyObject *__pyx_n_s_cancel;
  PyObject *__pyx_n_s_cancelled;
  PyObject *__pyx_n_s_check_schema_change;
  PyObject *__pyx_n_s_class;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_clear;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_s_closing;
  PyObject *__pyx_kp_u_code_type_must_be_a_MP_UINT;
  PyObject *__pyx_n_s_collections;
  PyObject *__pyx_n_s_commit;
  PyObject *__pyx_n_s_compile;
  PyObject *__pyx_n_s_connected_fut;
  PyObject *__pyx_n_s_connection_lost;
  PyObject *__pyx_n_s_connection_made;
  PyObject *__pyx_n_s_create_db;
  PyObject *__pyx_n_s_create_future;
  PyObject *__pyx_n_s_create_future_fallback;
  PyObject *__pyx_n_s_data;
  PyObject *__pyx_n_s_data_received;
  PyObject *__pyx_n_s_datetime;
  PyObject *__pyx_n_s_debug;
  PyObject *__pyx_n_s_decimal;
  PyObject *__pyx_n_s_decode;
  PyObject *__pyx_n_s_delete;
  PyObject *__pyx_n_s_deque;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_n_s_digest;
  PyObject *__pyx_n_s_digits;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_doc;
  PyObject *__pyx_n_s_done;
  PyObject *__pyx_n_s_e;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_encode;
  PyObject *__pyx_n_s_encoding;
  PyObject *__pyx_kp_u_encoding_must_be_either_str_or_b;
  PyObject *__pyx_n_s_enum;
  PyObject *__pyx_n_s_error;
  PyObject *__pyx_kp_u_errstr_type_must_be_a_MP_STR;
  PyObject *__pyx_n_s_eval;
  PyObject *__pyx_n_s_exc;
  PyObject *__pyx_kp_u_exceeded_timeout;
  PyObject *__pyx_n_s_exception;
  PyObject *__pyx_n_s_execute;
  PyObject *__pyx_n_s_exponent;
  PyObject *__pyx_n_s_expression;
  PyObject *__pyx_n_s_f;
  PyObject *__pyx_n_s_family;
  PyObject *__pyx_n_s_fetch_schema;
  PyObject *__pyx_n_u_field;
  PyObject *__pyx_kp_u_field_name_must_not_be_None;
  PyObject *__pyx_n_s_format;
  PyObject *__pyx_n_s_func_name;
  PyObject *__pyx_n_s_fut;
  PyObject *__pyx_n_s_futures;
  PyObject *__pyx_n_s_gather;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_gen_stream_id;
  PyObject *__pyx_n_s_genexpr;
  PyObject *__pyx_n_s_get;
  PyObject *__pyx_n_s_get_common_db;
  PyObject *__pyx_n_s_get_extra_info;
  PyObject *__pyx_n_s_get_version;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_kp_u_got;
  PyObject *__pyx_n_s_group;
  PyObject *__pyx_n_s_hashlib;
  PyObject *__pyx_n_s_hex;
  PyObject *__pyx_n_s_hex_locals_genexpr;
  PyObject *__pyx_n_s_host;
  PyObject *__pyx_kp_u_id_name_engine;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_index;
  PyObject *__pyx_n_s_indexes;
  PyObject *__pyx_n_s_init;
  PyObject *__pyx_n_s_init_subclass;
  PyObject *__pyx_n_s_initial_read_buffer_size;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_insert;
  PyObject *__pyx_n_s_inspect;
  PyObject *__pyx_kp_u_invalid_datetime_size_got_extra;
  PyObject *__pyx_kp_u_iproto_error_stack_frame_fields;
  PyObject *__pyx_n_s_is_connected;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_n_s_is_fully_connected;
  PyObject *__pyx_n_u_is_nullable;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_isolation;
  PyObject *__pyx_n_s_iterator;
  PyObject *__pyx_n_s_key;
  PyObject *__pyx_n_s_last_stream_id;
  PyObject *__pyx_n_s_limit;
  PyObject *__pyx_n_s_logger;
  PyObject *__pyx_n_s_loop;
  PyObject *__pyx_n_s_m;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_map;
  PyObject *__pyx_n_s_match;
  PyObject *__pyx_n_s_metaclass;
  PyObject *__pyx_n_s_module;
  PyObject *__pyx_n_s_mro_entries;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_u_name_2;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
  PyObject *__pyx_n_s_offset;
  PyObject *__pyx_n_s_on_authorized;
  PyObject *__pyx_n_s_on_connection_lost;
  PyObject *__pyx_n_s_on_connection_made;
  PyObject *__pyx_kp_u_on_data_received_expected_bytes;
  PyObject *__pyx_n_s_on_fetch;
  PyObject *__pyx_n_s_on_id;
  PyObject *__pyx_n_s_on_request_completed;
  PyObject *__pyx_n_s_on_request_timeout;
  PyObject *__pyx_n_s_operations;
  PyObject *__pyx_n_s_parse_metadata;
  PyObject *__pyx_n_s_parse_version;
  PyObject *__pyx_n_s_password;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_s_ping;
  PyObject *__pyx_n_s_popleft;
  PyObject *__pyx_n_s_port;
  PyObject *__pyx_n_s_prepare;
  PyObject *__pyx_n_s_prepare_2;
  PyObject *__pyx_n_s_push_subscribe;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_AuthRequest;
  PyObject *__pyx_n_s_pyx_unpickle_BaseRequest;
  PyObject *__pyx_n_s_pyx_unpickle_BeginRequest;
  PyObject *__pyx_n_s_pyx_unpickle_CallRequest;
  PyObject *__pyx_n_s_pyx_unpickle_CommitRequest;
  PyObject *__pyx_n_s_pyx_unpickle_CoreProtocol;
  PyObject *__pyx_n_s_pyx_unpickle_DeleteRequest;
  PyObject *__pyx_n_s_pyx_unpickle_EvalRequest;
  PyObject *__pyx_n_s_pyx_unpickle_ExecuteRequest;
  PyObject *__pyx_n_s_pyx_unpickle_IDRequest;
  PyObject *__pyx_n_s_pyx_unpickle_IProtoError;
  PyObject *__pyx_n_s_pyx_unpickle_IProtoErrorStackF;
  PyObject *__pyx_n_s_pyx_unpickle_InsertRequest;
  PyObject *__pyx_n_s_pyx_unpickle_PingRequest;
  PyObject *__pyx_n_s_pyx_unpickle_PrepareRequest;
  PyObject *__pyx_n_s_pyx_unpickle_PushIterator;
  PyObject *__pyx_n_s_pyx_unpickle_RollbackRequest;
  PyObject *__pyx_n_s_pyx_unpickle_SelectRequest;
  PyObject *__pyx_n_s_pyx_unpickle_UpdateRequest;
  PyObject *__pyx_n_s_pyx_unpickle_UpsertRequest;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_qualname;
  PyObject *__pyx_n_s_query;
  PyObject *__pyx_kp_u_query_must_be_either_str_or_int;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_re;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_refetch_schema;
  PyObject *__pyx_n_s_replace;
  PyObject *__pyx_n_s_repr;
  PyObject *__pyx_n_s_repr___locals_lambda;
  PyObject *__pyx_n_s_req;
  PyObject *__pyx_n_s_request_timeout;
  PyObject *__pyx_n_s_response;
  PyObject *__pyx_n_u_response;
  PyObject *__pyx_n_s_response_2;
  PyObject *__pyx_n_s_result;
  PyObject *__pyx_n_s_return_exceptions;
  PyObject *__pyx_n_s_rollback;
  PyObject *__pyx_n_s_rowcount;
  PyObject *__pyx_kp_u_s_Tarantool_s_d_s;
  PyObject *__pyx_n_s_schema_id;
  PyObject *__pyx_kp_u_schema_id_type_must_be_a_MP_UINT;
  PyObject *__pyx_n_s_select;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_kp_s_self_execute_cannot_be_converted;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_kp_u_sequence_must_be_either_list_or;
  PyObject *__pyx_kp_u_sequence_must_be_either_list_tup;
  PyObject *__pyx_n_s_set;
  PyObject *__pyx_n_s_set_exception;
  PyObject *__pyx_n_s_set_name;
  PyObject *__pyx_n_s_set_result;
  PyObject *__pyx_n_s_set_stream_id;
  PyObject *__pyx_n_s_setsockopt;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_sha1;
  PyObject *__pyx_kp_u_sid_id_name_type_unique;
  PyObject *__pyx_n_s_sign;
  PyObject *__pyx_n_s_sock;
  PyObject *__pyx_n_s_socket;
  PyObject *__pyx_n_u_socket;
  PyObject *__pyx_n_s_space;
  PyObject *__pyx_n_s_spaces;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_split;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_stream_id;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_super;
  PyObject *__pyx_n_s_sync;
  PyObject *__pyx_kp_u_sync_d_not_found;
  PyObject *__pyx_kp_u_sync_rowcount_data;
  PyObject *__pyx_kp_u_sync_type_must_be_a_MP_UINT;
  PyObject *__pyx_n_s_t;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_kp_u_the_buffer_is_in_read_only_mode;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_timeout;
  PyObject *__pyx_n_s_timestamp;
  PyObject *__pyx_n_s_timezone;
  PyObject *__pyx_n_s_total_seconds;
  PyObject *__pyx_n_s_transport;
  PyObject *__pyx_n_s_tx_timeout;
  PyObject *__pyx_n_u_type;
  PyObject *__pyx_n_s_typing;
  PyObject *__pyx_kp_u_unexpected_exponent_type;
  PyObject *__pyx_kp_u_unexpected_type_of_part_must_be;
  PyObject *__pyx_kp_u_unknown_iproto_error_map_field_w;
  PyObject *__pyx_kp_u_unknown_iproto_error_stack_eleme;
  PyObject *__pyx_kp_u_unknown_key_in_body_map_s;
  PyObject *__pyx_kp_u_unknown_key_in_metadata_decoding;
  PyObject *__pyx_kp_u_unknown_key_in_sql_info_decoding;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_upsert;
  PyObject *__pyx_n_s_use_setstate;
  PyObject *__pyx_n_s_username;
  PyObject *__pyx_n_s_utcoffset;
  PyObject *__pyx_kp_b_utf_8;
  PyObject *__pyx_n_s_uuid;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_s_ver;
  PyObject *__pyx_n_s_version;
  PyObject *__pyx_n_s_wait;
  PyObject *__pyx_n_s_waiter;
  PyObject *__pyx_n_s_warning;
  PyObject *__pyx_n_s_write;
  PyObject *__pyx_n_s_x;
  PyObject *__pyx_float_0_0;
  PyObject *__pyx_float_neg_1_0;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_2;
  PyObject *__pyx_int_3;
  PyObject *__pyx_int_4;
  PyObject *__pyx_int_5;
  PyObject *__pyx_int_6;
  PyObject *__pyx_int_7;
  PyObject *__pyx_int_8;
  PyObject *__pyx_int_9;
  PyObject *__pyx_int_10;
  PyObject *__pyx_int_11;
  PyObject *__pyx_int_20;
  PyObject *__pyx_int_281;
  PyObject *__pyx_int_289;
  PyObject *__pyx_int_4796468;
  PyObject *__pyx_int_4879619;
  PyObject *__pyx_int_23039357;
  PyObject *__pyx_int_25165850;
  PyObject *__pyx_int_47556637;
  PyObject *__pyx_int_51353055;
  PyObject *__pyx_int_51694216;
  PyObject *__pyx_int_53862661;
  PyObject *__pyx_int_63007895;
  PyObject *__pyx_int_92266676;
  PyObject *__pyx_int_103773939;
  PyObject *__pyx_int_105088600;
  PyObject *__pyx_int_105824685;
  PyObject *__pyx_int_129177485;
  PyObject *__pyx_int_135652490;
  PyObject *__pyx_int_144520718;
  PyObject *__pyx_int_145344315;
  PyObject *__pyx_int_145510168;
  PyObject *__pyx_int_149085227;
  PyObject *__pyx_int_150999062;
  PyObject *__pyx_int_151261227;
  PyObject *__pyx_int_159463343;
  PyObject *__pyx_int_170745181;
  PyObject *__pyx_int_172954204;
  PyObject *__pyx_int_179467967;
  PyObject *__pyx_int_180945731;
  PyObject *__pyx_int_184220263;
  PyObject *__pyx_int_192862663;
  PyObject *__pyx_int_194105842;
  PyObject *__pyx_int_196154451;
  PyObject *__pyx_int_201714318;
  PyObject *__pyx_int_209859363;
  PyObject *__pyx_int_214473120;
  PyObject *__pyx_int_215610203;
  PyObject *__pyx_int_218707107;
  PyObject *__pyx_int_219129172;
  PyObject *__pyx_int_220368800;
  PyObject *__pyx_int_225261924;
  PyObject *__pyx_int_235564716;
  PyObject *__pyx_int_244856512;
  PyObject *__pyx_int_245819509;
  PyObject *__pyx_int_246409365;
  PyObject *__pyx_int_253265777;
  PyObject *__pyx_int_255344631;
  PyObject *__pyx_int_257562521;
  PyObject *__pyx_int_264852134;
  PyObject *__pyx_int_266727818;
  PyObject *__pyx_int_267485924;
  PyObject *__pyx_int_4294967295;
  PyObject *__pyx_int_neg_2;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_tuple__4;
  PyObject *__pyx_tuple__6;
  PyObject *__pyx_tuple__7;
  PyObject *__pyx_tuple__8;
  PyObject *__pyx_tuple__9;
  PyObject *__pyx_slice__15;
  PyObject *__pyx_slice__16;
  PyObject *__pyx_slice__39;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_tuple__13;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__21;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__23;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__25;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__27;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__29;
  PyObject *__pyx_tuple__30;
  PyObject *__pyx_tuple__31;
  PyObject *__pyx_tuple__32;
  PyObject *__pyx_tuple__33;
  PyObject *__pyx_tuple__34;
  PyObject *__pyx_tuple__35;
  PyObject *__pyx_tuple__36;
  PyObject *__pyx_tuple__37;
  PyObject *__pyx_tuple__38;
  PyObject *__pyx_tuple__40;
  PyObject *__pyx_tuple__41;
  PyObject *__pyx_tuple__44;
  PyObject *__pyx_tuple__46;
  PyObject *__pyx_tuple__47;
  PyObject *__pyx_tuple__48;
  PyObject *__pyx_tuple__49;
  PyObject *__pyx_tuple__50;
  PyObject *__pyx_tuple__51;
  PyObject *__pyx_tuple__52;
  PyObject *__pyx_tuple__53;
  PyObject *__pyx_tuple__54;
  PyObject *__pyx_tuple__55;
  PyObject *__pyx_tuple__56;
  PyObject *__pyx_tuple__57;
  PyObject *__pyx_tuple__58;
  PyObject *__pyx_tuple__59;
  PyObject *__pyx_tuple__60;
  PyObject *__pyx_tuple__61;
  PyObject *__pyx_tuple__62;
  PyObject *__pyx_tuple__63;
  PyObject *__pyx_tuple__64;
  PyObject *__pyx_tuple__66;
  PyObject *__pyx_tuple__68;
  PyObject *__pyx_tuple__78;
  PyObject *__pyx_tuple__84;
  PyObject *__pyx_tuple__124;
  PyObject *__pyx_tuple__126;
  PyObject *__pyx_tuple__128;
  PyObject *__pyx_tuple__129;
  PyObject *__pyx_tuple__131;
  PyObject *__pyx_tuple__133;
  PyObject *__pyx_tuple__135;
  PyObject *__pyx_tuple__137;
  PyObject *__pyx_tuple__138;
  PyObject *__pyx_tuple__140;
  PyObject *__pyx_tuple__141;
  PyObject *__pyx_tuple__143;
  PyObject *__pyx_tuple__145;
  PyObject *__pyx_tuple__146;
  PyObject *__pyx_tuple__148;
  PyObject *__pyx_tuple__150;
  PyObject *__pyx_tuple__152;
  PyObject *__pyx_tuple__153;
  PyObject *__pyx_tuple__155;
  PyObject *__pyx_tuple__157;
  PyObject *__pyx_tuple__164;
  PyObject *__pyx_tuple__168;
  PyObject *__pyx_tuple__170;
  PyObject *__pyx_tuple__172;
  PyObject *__pyx_tuple__174;
  PyObject *__pyx_tuple__179;
  PyObject *__pyx_tuple__181;
  PyObject *__pyx_tuple__183;
  PyObject *__pyx_tuple__185;
  PyObject *__pyx_tuple__190;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__43;
  PyObject *__pyx_codeobj__45;
  PyObject *__pyx_codeobj__67;
  PyObject *__pyx_codeobj__69;
  PyObject *__pyx_codeobj__70;
  PyObject *__pyx_codeobj__71;
  PyObject *__pyx_codeobj__72;
  PyObject *__pyx_codeobj__73;
  PyObject *__pyx_codeobj__74;
  PyObject *__pyx_codeobj__75;
  PyObject *__pyx_codeobj__76;
  PyObject *__pyx_codeobj__77;
  PyObject *__pyx_codeobj__79;
  PyObject *__pyx_codeobj__80;
  PyObject *__pyx_codeobj__81;
  PyObject *__pyx_codeobj__82;
  PyObject *__pyx_codeobj__83;
  PyObject *__pyx_codeobj__85;
  PyObject *__pyx_codeobj__86;
  PyObject *__pyx_codeobj__87;
  PyObject *__pyx_codeobj__88;
  PyObject *__pyx_codeobj__89;
  PyObject *__pyx_codeobj__90;
  PyObject *__pyx_codeobj__91;
  PyObject *__pyx_codeobj__92;
  PyObject *__pyx_codeobj__93;
  PyObject *__pyx_codeobj__94;
  PyObject *__pyx_codeobj__95;
  PyObject *__pyx_codeobj__96;
  PyObject *__pyx_codeobj__97;
  PyObject *__pyx_codeobj__98;
  PyObject *__pyx_codeobj__99;
  PyObject *__pyx_codeobj__100;
  PyObject *__pyx_codeobj__101;
  PyObject *__pyx_codeobj__102;
  PyObject *__pyx_codeobj__103;
  PyObject *__pyx_codeobj__104;
  PyObject *__pyx_codeobj__105;
  PyObject *__pyx_codeobj__106;
  PyObject *__pyx_codeobj__107;
  PyObject *__pyx_codeobj__108;
  PyObject *__pyx_codeobj__109;
  PyObject *__pyx_codeobj__110;
  PyObject *__pyx_codeobj__111;
  PyObject *__pyx_codeobj__112;
  PyObject *__pyx_codeobj__113;
  PyObject *__pyx_codeobj__114;
  PyObject *__pyx_codeobj__115;
  PyObject *__pyx_codeobj__116;
  PyObject *__pyx_codeobj__117;
  PyObject *__pyx_codeobj__118;
  PyObject *__pyx_codeobj__119;
  PyObject *__pyx_codeobj__120;
  PyObject *__pyx_codeobj__121;
  PyObject *__pyx_codeobj__122;
  PyObject *__pyx_codeobj__123;
  PyObject *__pyx_codeobj__125;
  PyObject *__pyx_codeobj__127;
  PyObject *__pyx_codeobj__130;
  PyObject *__pyx_codeobj__132;
  PyObject *__pyx_codeobj__134;
  PyObject *__pyx_codeobj__136;
  PyObject *__pyx_codeobj__139;
  PyObject *__pyx_codeobj__142;
  PyObject *__pyx_codeobj__144;
  PyObject *__pyx_codeobj__147;
  PyObject *__pyx_codeobj__149;
  PyObject *__pyx_codeobj__151;
  PyObject *__pyx_codeobj__154;
  PyObject *__pyx_codeobj__156;
  PyObject *__pyx_codeobj__158;
  PyObject *__pyx_codeobj__159;
  PyObject *__pyx_codeobj__160;
  PyObject *__pyx_codeobj__161;
  PyObject *__pyx_codeobj__162;
  PyObject *__pyx_codeobj__163;
  PyObject *__pyx_codeobj__165;
  PyObject *__pyx_codeobj__166;
  PyObject *__pyx_codeobj__167;
  PyObject *__pyx_codeobj__169;
  PyObject *__pyx_codeobj__171;
  PyObject *__pyx_codeobj__173;
  PyObject *__pyx_codeobj__175;
  PyObject *__pyx_codeobj__176;
  PyObject *__pyx_codeobj__177;
  PyObject *__pyx_codeobj__178;
  PyObject *__pyx_codeobj__180;
  PyObject *__pyx_codeobj__182;
  PyObject *__pyx_codeobj__184;
  PyObject *__pyx_codeobj__186;
  PyObject *__pyx_codeobj__187;
  PyObject *__pyx_codeobj__188;
  PyObject *__pyx_codeobj__189;
  PyObject *__pyx_codeobj__191;
  PyObject *__pyx_codeobj__192;
  PyObject *__pyx_codeobj__193;
  PyObject *__pyx_codeobj__194;
  PyObject *__pyx_codeobj__195;
  PyObject *__pyx_codeobj__196;
  PyObject *__pyx_codeobj__197;
  PyObject *__pyx_codeobj__198;
  PyObject *__pyx_codeobj__199;
  PyObject *__pyx_codeobj__200;
  PyObject *__pyx_codeobj__201;
  PyObject *__pyx_codeobj__202;
  PyObject *__pyx_codeobj__203;
  PyObject *__pyx_codeobj__204;
  PyObject *__pyx_codeobj__205;
  PyObject *__pyx_codeobj__206;
  PyObject *__pyx_codeobj__207;
  PyObject *__pyx_codeobj__208;
  PyObject *__pyx_codeobj__209;
  PyObject *__pyx_codeobj__210;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_date);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_time);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_datetime);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_timedelta);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_tzinfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_Field);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_Field);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata);
  Py_CLEAR(clear_module_state->C_Metadata_Type);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_Schema);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_Schema);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_BaseRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_PingRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_CallRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_EvalRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_SelectRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_InsertRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_IDRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_AuthRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_BeginRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_CommitRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_IProtoError);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_Response);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_Response);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_Db);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_Db);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_PushIterator);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth);
  Py_CLEAR(clear_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema);
  Py_CLEAR(clear_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema);
  Py_CLEAR(clear_module_state->__pyx_kp_u_02x);
  Py_CLEAR(clear_module_state->__pyx_n_s_AF_UNIX);
  Py_CLEAR(clear_module_state->__pyx_n_u_AF_UNIX);
  Py_CLEAR(clear_module_state->__pyx_n_s_ALL);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_AttributeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_AuthRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_AuthRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_AuthRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Available_Iterator_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_BITS_ALL_NOT_SET);
  Py_CLEAR(clear_module_state->__pyx_n_s_BITS_ALL_SET);
  Py_CLEAR(clear_module_state->__pyx_n_s_BITS_ANY_SET);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol__create_future_fall);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol__do_auth_locals_on);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol__do_fetch_schema_lo);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol__do_id_locals_on_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol__on_request_complet);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol__on_request_timeout);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol_create_db);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol_get_common_db);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol_refetch_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BeginRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_BeginRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BeginRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BufferError);
  Py_CLEAR(clear_module_state->__pyx_n_s_CallRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_CallRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_CallRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_CancelledError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_pop_push_from_a_non_async);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_use_iter_with_PushIterato);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_use_next_with_PushIterato);
  Py_CLEAR(clear_module_state->__pyx_n_s_CommitRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_CommitRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_CommitRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConnectionRefusedError);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConnectionResetError);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoreProtocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoreProtocol___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoreProtocol___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoreProtocol__parse_version);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoreProtocol_connection_lost);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoreProtocol_connection_made);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoreProtocol_data_received);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoreProtocol_get_version);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoreProtocol_is_connected);
  Py_CLEAR(clear_module_state->__pyx_n_s_CoreProtocol_is_fully_connected);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Creates_PushIterator_object_In);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_begin);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_call);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_call16);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_commit);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_delete);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_eval);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_execute);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_insert);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_ping);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_prepare);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_replace);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_rollback);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_select);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_set_stream_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_Db_upsert);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Deallocating_buffer_with_attache);
  Py_CLEAR(clear_module_state->__pyx_n_s_Decimal);
  Py_CLEAR(clear_module_state->__pyx_n_s_DeleteRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_DeleteRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_DeleteRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_EQ);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Error_happened_while_parsing_sch);
  Py_CLEAR(clear_module_state->__pyx_n_s_EvalRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_EvalRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_EvalRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Event);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExecuteRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExecuteRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExecuteRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Field);
  Py_CLEAR(clear_module_state->__pyx_n_s_Field___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Field___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Field_d_of_space_s_is_not_in_spa);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Field_map_must_contain_at_least);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Field_map_must_contain_at_least_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Field_name_type_is_nullable);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Field_not_found);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Field_s_in_supplied_dict_is_unkn);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Frame_type_code_message);
  Py_CLEAR(clear_module_state->__pyx_n_s_Future);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Future_is_invalid_Make_sure_to_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_GE);
  Py_CLEAR(clear_module_state->__pyx_n_s_GT);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Header_key_must_be_a_MP_UINT);
  Py_CLEAR(clear_module_state->__pyx_n_s_IDRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_IDRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_IDRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_IPPROTO_TCP);
  Py_CLEAR(clear_module_state->__pyx_kp_u_IPROTO_ERROR_type_must_be_a_MP_M);
  Py_CLEAR(clear_module_state->__pyx_kp_u_IPROTO_FIELD_SPAN_must_be_either);
  Py_CLEAR(clear_module_state->__pyx_kp_u_IPROTO_STMT_ID_type_must_be_a_MP);
  Py_CLEAR(clear_module_state->__pyx_n_s_IProtoError);
  Py_CLEAR(clear_module_state->__pyx_n_s_IProtoErrorStackFrame);
  Py_CLEAR(clear_module_state->__pyx_n_s_IProtoErrorStackFrame___reduce_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_IProtoErrorStackFrame___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_IProtoError___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_IProtoError___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_IProto_auth_type_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_IProto_features_available_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_IProto_version_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Impossible_condition_happened_Pl);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Index_must_be_either_str_or_int);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Index_not_found_in_space);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Index_s_not_found_in_space_s_s_C);
  Py_CLEAR(clear_module_state->__pyx_n_s_InsertRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_InsertRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_InsertRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntEnum);
  Py_CLEAR(clear_module_state->__pyx_n_s_Iterator);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Iterator_is_of_unsupported_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_KeyError);
  Py_CLEAR(clear_module_state->__pyx_n_s_LE);
  Py_CLEAR(clear_module_state->__pyx_n_s_LT);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Lost_connection_to_Tarantool);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Lost_connection_to_Tarantool_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Lost_connection_to_Tarantool_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Metadata);
  Py_CLEAR(clear_module_state->__pyx_n_s_Metadata___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Metadata___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Metadata_fields_count);
  Py_CLEAR(clear_module_state->__pyx_n_s_NEIGHBOR);
  Py_CLEAR(clear_module_state->__pyx_n_s_OVERLAPS);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Offset_incorrect_Got_use_len);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Operation_field_no_must_be_of_ei);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Operation_length_must_be_at_leas);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Operation_type_must_of_a_str_or);
  Py_CLEAR(clear_module_state->__pyx_n_s_Optional);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Part_len_must_be_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_PingRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_PingRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PingRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrepareRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrepareRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrepareRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_PushIterator);
  Py_CLEAR(clear_module_state->__pyx_n_s_PushIterator___anext);
  Py_CLEAR(clear_module_state->__pyx_kp_u_PushIterator___init___line_9);
  Py_CLEAR(clear_module_state->__pyx_n_s_PushIterator___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PushIterator___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_REQ);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadBuffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadBuffer___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadBuffer___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Request_op_sync_schema_id_push);
  Py_CLEAR(clear_module_state->__pyx_n_s_Response);
  Py_CLEAR(clear_module_state->__pyx_n_s_Response___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Response___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Response_body_must_be_a_MP_MAP);
  Py_CLEAR(clear_module_state->__pyx_n_s_Response_done);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Response_header_must_be_a_MP_MAP);
  Py_CLEAR(clear_module_state->__pyx_n_s_RollbackRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_RollbackRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_RollbackRequest___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_RuntimeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_SchemaDummyIndex);
  Py_CLEAR(clear_module_state->__pyx_n_s_SchemaDummySpace);
  Py_CLEAR(clear_module_state->__pyx_n_s_SchemaIndex);
  Py_CLEAR(clear_module_state->__pyx_n_s_SchemaIndex___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SchemaIndex___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SchemaSpace);
  Py_CLEAR(clear_module_state->__pyx_n_s_SchemaSpace___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SchemaSpace___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Schema_fetch_stopped_connection);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Schema_fetch_timeout);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Schema_spaces);
  Py_CLEAR(clear_module_state->__pyx_n_s_SelectRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_SelectRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SelectRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Single_operation_must_be_a_tuple);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Space_must_be_either_str_or_int);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Space_not_found);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Space_s_not_found_Creating_dummy);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Space_with_id_not_found);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Splice_offset_must_be_int);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Splice_operation_must_have_lengt);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Splice_position_must_be_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_TCP_NODELAY);
  Py_CLEAR(clear_module_state->__pyx_n_s_TarantoolDatabaseError);
  Py_CLEAR(clear_module_state->__pyx_n_s_TarantoolNotConnectedError);
  Py_CLEAR(clear_module_state->__pyx_n_s_TarantoolSchemaError);
  Py_CLEAR(clear_module_state->__pyx_n_s_TarantoolTuple);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Tarantool_is_not_connected);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Tarantool_s_s_Authorization_fail);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Tarantool_s_s_Authorized_success);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Tarantool_s_s_Schema_fetch_faile);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Tarantool_s_s_Schema_fetch_succe);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Tarantool_s_s_identification_fai);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Tarantool_s_s_identified_success);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Time_zones_are_not_available_fro);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimeoutError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Tuple_must_be_an_array_when_deco);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Type_is_not_supported_for_encodi);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUID);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unexpected_ext_type_d);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unexpected_key_type_in_map_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unexpected_obj_type_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnicodeDecodeError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unknown_key_with_code_d_in_heade);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unknown_update_operation_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_UpdateRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_UpdateRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UpdateRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UpsertRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_UpsertRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UpsertRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_VERSION_STRING_REGEX);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteBuffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteBuffer___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteBuffer___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteBuffer_hex);
  Py_CLEAR(clear_module_state->__pyx_kp_u__14);
  Py_CLEAR(clear_module_state->__pyx_kp_u__17);
  Py_CLEAR(clear_module_state->__pyx_kp_u__18);
  Py_CLEAR(clear_module_state->__pyx_kp_u__19);
  Py_CLEAR(clear_module_state->__pyx_kp_u__2);
  Py_CLEAR(clear_module_state->__pyx_n_s__211);
  Py_CLEAR(clear_module_state->__pyx_kp_u__3);
  Py_CLEAR(clear_module_state->__pyx_kp_u__5);
  Py_CLEAR(clear_module_state->__pyx_n_s__65);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_done_callback);
  Py_CLEAR(clear_module_state->__pyx_n_s_anext);
  Py_CLEAR(clear_module_state->__pyx_n_s_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_as_tuple);
  Py_CLEAR(clear_module_state->__pyx_n_u_ascii);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_tasks);
  Py_CLEAR(clear_module_state->__pyx_n_s_asynctnt);
  Py_CLEAR(clear_module_state->__pyx_n_s_asynctnt_exceptions);
  Py_CLEAR(clear_module_state->__pyx_kp_s_asynctnt_iproto_buffer_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_asynctnt_iproto_coreproto_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_asynctnt_iproto_db_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_asynctnt_iproto_protocol);
  Py_CLEAR(clear_module_state->__pyx_kp_s_asynctnt_iproto_protocol_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_asynctnt_iproto_response_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_asynctnt_log);
  Py_CLEAR(clear_module_state->__pyx_n_s_auto_refetch_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_await);
  Py_CLEAR(clear_module_state->__pyx_n_s_b64decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_base64);
  Py_CLEAR(clear_module_state->__pyx_n_s_begin);
  Py_CLEAR(clear_module_state->__pyx_kp_u_body_data_type_must_be_a_MP_ARRA);
  Py_CLEAR(clear_module_state->__pyx_n_s_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_call);
  Py_CLEAR(clear_module_state->__pyx_n_s_call16);
  Py_CLEAR(clear_module_state->__pyx_n_s_call_later);
  Py_CLEAR(clear_module_state->__pyx_n_s_cancel);
  Py_CLEAR(clear_module_state->__pyx_n_s_cancelled);
  Py_CLEAR(clear_module_state->__pyx_n_s_check_schema_change);
  Py_CLEAR(clear_module_state->__pyx_n_s_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_clear);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_closing);
  Py_CLEAR(clear_module_state->__pyx_kp_u_code_type_must_be_a_MP_UINT);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections);
  Py_CLEAR(clear_module_state->__pyx_n_s_commit);
  Py_CLEAR(clear_module_state->__pyx_n_s_compile);
  Py_CLEAR(clear_module_state->__pyx_n_s_connected_fut);
  Py_CLEAR(clear_module_state->__pyx_n_s_connection_lost);
  Py_CLEAR(clear_module_state->__pyx_n_s_connection_made);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_db);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_future);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_future_fallback);
  Py_CLEAR(clear_module_state->__pyx_n_s_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_data_received);
  Py_CLEAR(clear_module_state->__pyx_n_s_datetime);
  Py_CLEAR(clear_module_state->__pyx_n_s_debug);
  Py_CLEAR(clear_module_state->__pyx_n_s_decimal);
  Py_CLEAR(clear_module_state->__pyx_n_s_decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_delete);
  Py_CLEAR(clear_module_state->__pyx_n_s_deque);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_digest);
  Py_CLEAR(clear_module_state->__pyx_n_s_digits);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_doc);
  Py_CLEAR(clear_module_state->__pyx_n_s_done);
  Py_CLEAR(clear_module_state->__pyx_n_s_e);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_encoding);
  Py_CLEAR(clear_module_state->__pyx_kp_u_encoding_must_be_either_str_or_b);
  Py_CLEAR(clear_module_state->__pyx_n_s_enum);
  Py_CLEAR(clear_module_state->__pyx_n_s_error);
  Py_CLEAR(clear_module_state->__pyx_kp_u_errstr_type_must_be_a_MP_STR);
  Py_CLEAR(clear_module_state->__pyx_n_s_eval);
  Py_CLEAR(clear_module_state->__pyx_n_s_exc);
  Py_CLEAR(clear_module_state->__pyx_kp_u_exceeded_timeout);
  Py_CLEAR(clear_module_state->__pyx_n_s_exception);
  Py_CLEAR(clear_module_state->__pyx_n_s_execute);
  Py_CLEAR(clear_module_state->__pyx_n_s_exponent);
  Py_CLEAR(clear_module_state->__pyx_n_s_expression);
  Py_CLEAR(clear_module_state->__pyx_n_s_f);
  Py_CLEAR(clear_module_state->__pyx_n_s_family);
  Py_CLEAR(clear_module_state->__pyx_n_s_fetch_schema);
  Py_CLEAR(clear_module_state->__pyx_n_u_field);
  Py_CLEAR(clear_module_state->__pyx_kp_u_field_name_must_not_be_None);
  Py_CLEAR(clear_module_state->__pyx_n_s_format);
  Py_CLEAR(clear_module_state->__pyx_n_s_func_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_fut);
  Py_CLEAR(clear_module_state->__pyx_n_s_futures);
  Py_CLEAR(clear_module_state->__pyx_n_s_gather);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_gen_stream_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_common_db);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_extra_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_version);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_kp_u_got);
  Py_CLEAR(clear_module_state->__pyx_n_s_group);
  Py_CLEAR(clear_module_state->__pyx_n_s_hashlib);
  Py_CLEAR(clear_module_state->__pyx_n_s_hex);
  Py_CLEAR(clear_module_state->__pyx_n_s_hex_locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_host);
  Py_CLEAR(clear_module_state->__pyx_kp_u_id_name_engine);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_indexes);
  Py_CLEAR(clear_module_state->__pyx_n_s_init);
  Py_CLEAR(clear_module_state->__pyx_n_s_init_subclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_initial_read_buffer_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_insert);
  Py_CLEAR(clear_module_state->__pyx_n_s_inspect);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_datetime_size_got_extra);
  Py_CLEAR(clear_module_state->__pyx_kp_u_iproto_error_stack_frame_fields);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_connected);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_fully_connected);
  Py_CLEAR(clear_module_state->__pyx_n_u_is_nullable);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_isolation);
  Py_CLEAR(clear_module_state->__pyx_n_s_iterator);
  Py_CLEAR(clear_module_state->__pyx_n_s_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_last_stream_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_limit);
  Py_CLEAR(clear_module_state->__pyx_n_s_logger);
  Py_CLEAR(clear_module_state->__pyx_n_s_loop);
  Py_CLEAR(clear_module_state->__pyx_n_s_m);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_match);
  Py_CLEAR(clear_module_state->__pyx_n_s_metaclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_module);
  Py_CLEAR(clear_module_state->__pyx_n_s_mro_entries);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_CLEAR(clear_module_state->__pyx_n_s_offset);
  Py_CLEAR(clear_module_state->__pyx_n_s_on_authorized);
  Py_CLEAR(clear_module_state->__pyx_n_s_on_connection_lost);
  Py_CLEAR(clear_module_state->__pyx_n_s_on_connection_made);
  Py_CLEAR(clear_module_state->__pyx_kp_u_on_data_received_expected_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_on_fetch);
  Py_CLEAR(clear_module_state->__pyx_n_s_on_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_on_request_completed);
  Py_CLEAR(clear_module_state->__pyx_n_s_on_request_timeout);
  Py_CLEAR(clear_module_state->__pyx_n_s_operations);
  Py_CLEAR(clear_module_state->__pyx_n_s_parse_metadata);
  Py_CLEAR(clear_module_state->__pyx_n_s_parse_version);
  Py_CLEAR(clear_module_state->__pyx_n_s_password);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_s_ping);
  Py_CLEAR(clear_module_state->__pyx_n_s_popleft);
  Py_CLEAR(clear_module_state->__pyx_n_s_port);
  Py_CLEAR(clear_module_state->__pyx_n_s_prepare);
  Py_CLEAR(clear_module_state->__pyx_n_s_prepare_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_push_subscribe);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_AuthRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_BaseRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_BeginRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_CallRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_CommitRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_CoreProtocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_DeleteRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_EvalRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_ExecuteRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_IDRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_IProtoError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_IProtoErrorStackF);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_InsertRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PingRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PrepareRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PushIterator);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_RollbackRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_SelectRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_UpdateRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_UpsertRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_qualname);
  Py_CLEAR(clear_module_state->__pyx_n_s_query);
  Py_CLEAR(clear_module_state->__pyx_kp_u_query_must_be_either_str_or_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_re);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_refetch_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_replace);
  Py_CLEAR(clear_module_state->__pyx_n_s_repr);
  Py_CLEAR(clear_module_state->__pyx_n_s_repr___locals_lambda);
  Py_CLEAR(clear_module_state->__pyx_n_s_req);
  Py_CLEAR(clear_module_state->__pyx_n_s_request_timeout);
  Py_CLEAR(clear_module_state->__pyx_n_s_response);
  Py_CLEAR(clear_module_state->__pyx_n_u_response);
  Py_CLEAR(clear_module_state->__pyx_n_s_response_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_return_exceptions);
  Py_CLEAR(clear_module_state->__pyx_n_s_rollback);
  Py_CLEAR(clear_module_state->__pyx_n_s_rowcount);
  Py_CLEAR(clear_module_state->__pyx_kp_u_s_Tarantool_s_d_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_schema_id);
  Py_CLEAR(clear_module_state->__pyx_kp_u_schema_id_type_must_be_a_MP_UINT);
  Py_CLEAR(clear_module_state->__pyx_n_s_select);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_kp_s_self_execute_cannot_be_converted);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sequence_must_be_either_list_or);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sequence_must_be_either_list_tup);
  Py_CLEAR(clear_module_state->__pyx_n_s_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_exception);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_stream_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_setsockopt);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_sha1);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sid_id_name_type_unique);
  Py_CLEAR(clear_module_state->__pyx_n_s_sign);
  Py_CLEAR(clear_module_state->__pyx_n_s_sock);
  Py_CLEAR(clear_module_state->__pyx_n_s_socket);
  Py_CLEAR(clear_module_state->__pyx_n_u_socket);
  Py_CLEAR(clear_module_state->__pyx_n_s_space);
  Py_CLEAR(clear_module_state->__pyx_n_s_spaces);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_split);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_stream_id);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_super);
  Py_CLEAR(clear_module_state->__pyx_n_s_sync);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sync_d_not_found);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sync_rowcount_data);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sync_type_must_be_a_MP_UINT);
  Py_CLEAR(clear_module_state->__pyx_n_s_t);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_kp_u_the_buffer_is_in_read_only_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_timeout);
  Py_CLEAR(clear_module_state->__pyx_n_s_timestamp);
  Py_CLEAR(clear_module_state->__pyx_n_s_timezone);
  Py_CLEAR(clear_module_state->__pyx_n_s_total_seconds);
  Py_CLEAR(clear_module_state->__pyx_n_s_transport);
  Py_CLEAR(clear_module_state->__pyx_n_s_tx_timeout);
  Py_CLEAR(clear_module_state->__pyx_n_u_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_typing);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unexpected_exponent_type);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unexpected_type_of_part_must_be);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unknown_iproto_error_map_field_w);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unknown_iproto_error_stack_eleme);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unknown_key_in_body_map_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unknown_key_in_metadata_decoding);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unknown_key_in_sql_info_decoding);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_upsert);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_username);
  Py_CLEAR(clear_module_state->__pyx_n_s_utcoffset);
  Py_CLEAR(clear_module_state->__pyx_kp_b_utf_8);
  Py_CLEAR(clear_module_state->__pyx_n_s_uuid);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_ver);
  Py_CLEAR(clear_module_state->__pyx_n_s_version);
  Py_CLEAR(clear_module_state->__pyx_n_s_wait);
  Py_CLEAR(clear_module_state->__pyx_n_s_waiter);
  Py_CLEAR(clear_module_state->__pyx_n_s_warning);
  Py_CLEAR(clear_module_state->__pyx_n_s_write);
  Py_CLEAR(clear_module_state->__pyx_n_s_x);
  Py_CLEAR(clear_module_state->__pyx_float_0_0);
  Py_CLEAR(clear_module_state->__pyx_float_neg_1_0);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_2);
  Py_CLEAR(clear_module_state->__pyx_int_3);
  Py_CLEAR(clear_module_state->__pyx_int_4);
  Py_CLEAR(clear_module_state->__pyx_int_5);
  Py_CLEAR(clear_module_state->__pyx_int_6);
  Py_CLEAR(clear_module_state->__pyx_int_7);
  Py_CLEAR(clear_module_state->__pyx_int_8);
  Py_CLEAR(clear_module_state->__pyx_int_9);
  Py_CLEAR(clear_module_state->__pyx_int_10);
  Py_CLEAR(clear_module_state->__pyx_int_11);
  Py_CLEAR(clear_module_state->__pyx_int_20);
  Py_CLEAR(clear_module_state->__pyx_int_281);
  Py_CLEAR(clear_module_state->__pyx_int_289);
  Py_CLEAR(clear_module_state->__pyx_int_4796468);
  Py_CLEAR(clear_module_state->__pyx_int_4879619);
  Py_CLEAR(clear_module_state->__pyx_int_23039357);
  Py_CLEAR(clear_module_state->__pyx_int_25165850);
  Py_CLEAR(clear_module_state->__pyx_int_47556637);
  Py_CLEAR(clear_module_state->__pyx_int_51353055);
  Py_CLEAR(clear_module_state->__pyx_int_51694216);
  Py_CLEAR(clear_module_state->__pyx_int_53862661);
  Py_CLEAR(clear_module_state->__pyx_int_63007895);
  Py_CLEAR(clear_module_state->__pyx_int_92266676);
  Py_CLEAR(clear_module_state->__pyx_int_103773939);
  Py_CLEAR(clear_module_state->__pyx_int_105088600);
  Py_CLEAR(clear_module_state->__pyx_int_105824685);
  Py_CLEAR(clear_module_state->__pyx_int_129177485);
  Py_CLEAR(clear_module_state->__pyx_int_135652490);
  Py_CLEAR(clear_module_state->__pyx_int_144520718);
  Py_CLEAR(clear_module_state->__pyx_int_145344315);
  Py_CLEAR(clear_module_state->__pyx_int_145510168);
  Py_CLEAR(clear_module_state->__pyx_int_149085227);
  Py_CLEAR(clear_module_state->__pyx_int_150999062);
  Py_CLEAR(clear_module_state->__pyx_int_151261227);
  Py_CLEAR(clear_module_state->__pyx_int_159463343);
  Py_CLEAR(clear_module_state->__pyx_int_170745181);
  Py_CLEAR(clear_module_state->__pyx_int_172954204);
  Py_CLEAR(clear_module_state->__pyx_int_179467967);
  Py_CLEAR(clear_module_state->__pyx_int_180945731);
  Py_CLEAR(clear_module_state->__pyx_int_184220263);
  Py_CLEAR(clear_module_state->__pyx_int_192862663);
  Py_CLEAR(clear_module_state->__pyx_int_194105842);
  Py_CLEAR(clear_module_state->__pyx_int_196154451);
  Py_CLEAR(clear_module_state->__pyx_int_201714318);
  Py_CLEAR(clear_module_state->__pyx_int_209859363);
  Py_CLEAR(clear_module_state->__pyx_int_214473120);
  Py_CLEAR(clear_module_state->__pyx_int_215610203);
  Py_CLEAR(clear_module_state->__pyx_int_218707107);
  Py_CLEAR(clear_module_state->__pyx_int_219129172);
  Py_CLEAR(clear_module_state->__pyx_int_220368800);
  Py_CLEAR(clear_module_state->__pyx_int_225261924);
  Py_CLEAR(clear_module_state->__pyx_int_235564716);
  Py_CLEAR(clear_module_state->__pyx_int_244856512);
  Py_CLEAR(clear_module_state->__pyx_int_245819509);
  Py_CLEAR(clear_module_state->__pyx_int_246409365);
  Py_CLEAR(clear_module_state->__pyx_int_253265777);
  Py_CLEAR(clear_module_state->__pyx_int_255344631);
  Py_CLEAR(clear_module_state->__pyx_int_257562521);
  Py_CLEAR(clear_module_state->__pyx_int_264852134);
  Py_CLEAR(clear_module_state->__pyx_int_266727818);
  Py_CLEAR(clear_module_state->__pyx_int_267485924);
  Py_CLEAR(clear_module_state->__pyx_int_4294967295);
  Py_CLEAR(clear_module_state->__pyx_int_neg_2);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_tuple__4);
  Py_CLEAR(clear_module_state->__pyx_tuple__6);
  Py_CLEAR(clear_module_state->__pyx_tuple__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__8);
  Py_CLEAR(clear_module_state->__pyx_tuple__9);
  Py_CLEAR(clear_module_state->__pyx_slice__15);
  Py_CLEAR(clear_module_state->__pyx_slice__16);
  Py_CLEAR(clear_module_state->__pyx_slice__39);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_tuple__13);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__21);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__23);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__25);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__27);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__29);
  Py_CLEAR(clear_module_state->__pyx_tuple__30);
  Py_CLEAR(clear_module_state->__pyx_tuple__31);
  Py_CLEAR(clear_module_state->__pyx_tuple__32);
  Py_CLEAR(clear_module_state->__pyx_tuple__33);
  Py_CLEAR(clear_module_state->__pyx_tuple__34);
  Py_CLEAR(clear_module_state->__pyx_tuple__35);
  Py_CLEAR(clear_module_state->__pyx_tuple__36);
  Py_CLEAR(clear_module_state->__pyx_tuple__37);
  Py_CLEAR(clear_module_state->__pyx_tuple__38);
  Py_CLEAR(clear_module_state->__pyx_tuple__40);
  Py_CLEAR(clear_module_state->__pyx_tuple__41);
  Py_CLEAR(clear_module_state->__pyx_tuple__44);
  Py_CLEAR(clear_module_state->__pyx_tuple__46);
  Py_CLEAR(clear_module_state->__pyx_tuple__47);
  Py_CLEAR(clear_module_state->__pyx_tuple__48);
  Py_CLEAR(clear_module_state->__pyx_tuple__49);
  Py_CLEAR(clear_module_state->__pyx_tuple__50);
  Py_CLEAR(clear_module_state->__pyx_tuple__51);
  Py_CLEAR(clear_module_state->__pyx_tuple__52);
  Py_CLEAR(clear_module_state->__pyx_tuple__53);
  Py_CLEAR(clear_module_state->__pyx_tuple__54);
  Py_CLEAR(clear_module_state->__pyx_tuple__55);
  Py_CLEAR(clear_module_state->__pyx_tuple__56);
  Py_CLEAR(clear_module_state->__pyx_tuple__57);
  Py_CLEAR(clear_module_state->__pyx_tuple__58);
  Py_CLEAR(clear_module_state->__pyx_tuple__59);
  Py_CLEAR(clear_module_state->__pyx_tuple__60);
  Py_CLEAR(clear_module_state->__pyx_tuple__61);
  Py_CLEAR(clear_module_state->__pyx_tuple__62);
  Py_CLEAR(clear_module_state->__pyx_tuple__63);
  Py_CLEAR(clear_module_state->__pyx_tuple__64);
  Py_CLEAR(clear_module_state->__pyx_tuple__66);
  Py_CLEAR(clear_module_state->__pyx_tuple__68);
  Py_CLEAR(clear_module_state->__pyx_tuple__78);
  Py_CLEAR(clear_module_state->__pyx_tuple__84);
  Py_CLEAR(clear_module_state->__pyx_tuple__124);
  Py_CLEAR(clear_module_state->__pyx_tuple__126);
  Py_CLEAR(clear_module_state->__pyx_tuple__128);
  Py_CLEAR(clear_module_state->__pyx_tuple__129);
  Py_CLEAR(clear_module_state->__pyx_tuple__131);
  Py_CLEAR(clear_module_state->__pyx_tuple__133);
  Py_CLEAR(clear_module_state->__pyx_tuple__135);
  Py_CLEAR(clear_module_state->__pyx_tuple__137);
  Py_CLEAR(clear_module_state->__pyx_tuple__138);
  Py_CLEAR(clear_module_state->__pyx_tuple__140);
  Py_CLEAR(clear_module_state->__pyx_tuple__141);
  Py_CLEAR(clear_module_state->__pyx_tuple__143);
  Py_CLEAR(clear_module_state->__pyx_tuple__145);
  Py_CLEAR(clear_module_state->__pyx_tuple__146);
  Py_CLEAR(clear_module_state->__pyx_tuple__148);
  Py_CLEAR(clear_module_state->__pyx_tuple__150);
  Py_CLEAR(clear_module_state->__pyx_tuple__152);
  Py_CLEAR(clear_module_state->__pyx_tuple__153);
  Py_CLEAR(clear_module_state->__pyx_tuple__155);
  Py_CLEAR(clear_module_state->__pyx_tuple__157);
  Py_CLEAR(clear_module_state->__pyx_tuple__164);
  Py_CLEAR(clear_module_state->__pyx_tuple__168);
  Py_CLEAR(clear_module_state->__pyx_tuple__170);
  Py_CLEAR(clear_module_state->__pyx_tuple__172);
  Py_CLEAR(clear_module_state->__pyx_tuple__174);
  Py_CLEAR(clear_module_state->__pyx_tuple__179);
  Py_CLEAR(clear_module_state->__pyx_tuple__181);
  Py_CLEAR(clear_module_state->__pyx_tuple__183);
  Py_CLEAR(clear_module_state->__pyx_tuple__185);
  Py_CLEAR(clear_module_state->__pyx_tuple__190);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__43);
  Py_CLEAR(clear_module_state->__pyx_codeobj__45);
  Py_CLEAR(clear_module_state->__pyx_codeobj__67);
  Py_CLEAR(clear_module_state->__pyx_codeobj__69);
  Py_CLEAR(clear_module_state->__pyx_codeobj__70);
  Py_CLEAR(clear_module_state->__pyx_codeobj__71);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  Py_CLEAR(clear_module_state->__pyx_codeobj__73);
  Py_CLEAR(clear_module_state->__pyx_codeobj__74);
  Py_CLEAR(clear_module_state->__pyx_codeobj__75);
  Py_CLEAR(clear_module_state->__pyx_codeobj__76);
  Py_CLEAR(clear_module_state->__pyx_codeobj__77);
  Py_CLEAR(clear_module_state->__pyx_codeobj__79);
  Py_CLEAR(clear_module_state->__pyx_codeobj__80);
  Py_CLEAR(clear_module_state->__pyx_codeobj__81);
  Py_CLEAR(clear_module_state->__pyx_codeobj__82);
  Py_CLEAR(clear_module_state->__pyx_codeobj__83);
  Py_CLEAR(clear_module_state->__pyx_codeobj__85);
  Py_CLEAR(clear_module_state->__pyx_codeobj__86);
  Py_CLEAR(clear_module_state->__pyx_codeobj__87);
  Py_CLEAR(clear_module_state->__pyx_codeobj__88);
  Py_CLEAR(clear_module_state->__pyx_codeobj__89);
  Py_CLEAR(clear_module_state->__pyx_codeobj__90);
  Py_CLEAR(clear_module_state->__pyx_codeobj__91);
  Py_CLEAR(clear_module_state->__pyx_codeobj__92);
  Py_CLEAR(clear_module_state->__pyx_codeobj__93);
  Py_CLEAR(clear_module_state->__pyx_codeobj__94);
  Py_CLEAR(clear_module_state->__pyx_codeobj__95);
  Py_CLEAR(clear_module_state->__pyx_codeobj__96);
  Py_CLEAR(clear_module_state->__pyx_codeobj__97);
  Py_CLEAR(clear_module_state->__pyx_codeobj__98);
  Py_CLEAR(clear_module_state->__pyx_codeobj__99);
  Py_CLEAR(clear_module_state->__pyx_codeobj__100);
  Py_CLEAR(clear_module_state->__pyx_codeobj__101);
  Py_CLEAR(clear_module_state->__pyx_codeobj__102);
  Py_CLEAR(clear_module_state->__pyx_codeobj__103);
  Py_CLEAR(clear_module_state->__pyx_codeobj__104);
  Py_CLEAR(clear_module_state->__pyx_codeobj__105);
  Py_CLEAR(clear_module_state->__pyx_codeobj__106);
  Py_CLEAR(clear_module_state->__pyx_codeobj__107);
  Py_CLEAR(clear_module_state->__pyx_codeobj__108);
  Py_CLEAR(clear_module_state->__pyx_codeobj__109);
  Py_CLEAR(clear_module_state->__pyx_codeobj__110);
  Py_CLEAR(clear_module_state->__pyx_codeobj__111);
  Py_CLEAR(clear_module_state->__pyx_codeobj__112);
  Py_CLEAR(clear_module_state->__pyx_codeobj__113);
  Py_CLEAR(clear_module_state->__pyx_codeobj__114);
  Py_CLEAR(clear_module_state->__pyx_codeobj__115);
  Py_CLEAR(clear_module_state->__pyx_codeobj__116);
  Py_CLEAR(clear_module_state->__pyx_codeobj__117);
  Py_CLEAR(clear_module_state->__pyx_codeobj__118);
  Py_CLEAR(clear_module_state->__pyx_codeobj__119);
  Py_CLEAR(clear_module_state->__pyx_codeobj__120);
  Py_CLEAR(clear_module_state->__pyx_codeobj__121);
  Py_CLEAR(clear_module_state->__pyx_codeobj__122);
  Py_CLEAR(clear_module_state->__pyx_codeobj__123);
  Py_CLEAR(clear_module_state->__pyx_codeobj__125);
  Py_CLEAR(clear_module_state->__pyx_codeobj__127);
  Py_CLEAR(clear_module_state->__pyx_codeobj__130);
  Py_CLEAR(clear_module_state->__pyx_codeobj__132);
  Py_CLEAR(clear_module_state->__pyx_codeobj__134);
  Py_CLEAR(clear_module_state->__pyx_codeobj__136);
  Py_CLEAR(clear_module_state->__pyx_codeobj__139);
  Py_CLEAR(clear_module_state->__pyx_codeobj__142);
  Py_CLEAR(clear_module_state->__pyx_codeobj__144);
  Py_CLEAR(clear_module_state->__pyx_codeobj__147);
  Py_CLEAR(clear_module_state->__pyx_codeobj__149);
  Py_CLEAR(clear_module_state->__pyx_codeobj__151);
  Py_CLEAR(clear_module_state->__pyx_codeobj__154);
  Py_CLEAR(clear_module_state->__pyx_codeobj__156);
  Py_CLEAR(clear_module_state->__pyx_codeobj__158);
  Py_CLEAR(clear_module_state->__pyx_codeobj__159);
  Py_CLEAR(clear_module_state->__pyx_codeobj__160);
  Py_CLEAR(clear_module_state->__pyx_codeobj__161);
  Py_CLEAR(clear_module_state->__pyx_codeobj__162);
  Py_CLEAR(clear_module_state->__pyx_codeobj__163);
  Py_CLEAR(clear_module_state->__pyx_codeobj__165);
  Py_CLEAR(clear_module_state->__pyx_codeobj__166);
  Py_CLEAR(clear_module_state->__pyx_codeobj__167);
  Py_CLEAR(clear_module_state->__pyx_codeobj__169);
  Py_CLEAR(clear_module_state->__pyx_codeobj__171);
  Py_CLEAR(clear_module_state->__pyx_codeobj__173);
  Py_CLEAR(clear_module_state->__pyx_codeobj__175);
  Py_CLEAR(clear_module_state->__pyx_codeobj__176);
  Py_CLEAR(clear_module_state->__pyx_codeobj__177);
  Py_CLEAR(clear_module_state->__pyx_codeobj__178);
  Py_CLEAR(clear_module_state->__pyx_codeobj__180);
  Py_CLEAR(clear_module_state->__pyx_codeobj__182);
  Py_CLEAR(clear_module_state->__pyx_codeobj__184);
  Py_CLEAR(clear_module_state->__pyx_codeobj__186);
  Py_CLEAR(clear_module_state->__pyx_codeobj__187);
  Py_CLEAR(clear_module_state->__pyx_codeobj__188);
  Py_CLEAR(clear_module_state->__pyx_codeobj__189);
  Py_CLEAR(clear_module_state->__pyx_codeobj__191);
  Py_CLEAR(clear_module_state->__pyx_codeobj__192);
  Py_CLEAR(clear_module_state->__pyx_codeobj__193);
  Py_CLEAR(clear_module_state->__pyx_codeobj__194);
  Py_CLEAR(clear_module_state->__pyx_codeobj__195);
  Py_CLEAR(clear_module_state->__pyx_codeobj__196);
  Py_CLEAR(clear_module_state->__pyx_codeobj__197);
  Py_CLEAR(clear_module_state->__pyx_codeobj__198);
  Py_CLEAR(clear_module_state->__pyx_codeobj__199);
  Py_CLEAR(clear_module_state->__pyx_codeobj__200);
  Py_CLEAR(clear_module_state->__pyx_codeobj__201);
  Py_CLEAR(clear_module_state->__pyx_codeobj__202);
  Py_CLEAR(clear_module_state->__pyx_codeobj__203);
  Py_CLEAR(clear_module_state->__pyx_codeobj__204);
  Py_CLEAR(clear_module_state->__pyx_codeobj__205);
  Py_CLEAR(clear_module_state->__pyx_codeobj__206);
  Py_CLEAR(clear_module_state->__pyx_codeobj__207);
  Py_CLEAR(clear_module_state->__pyx_codeobj__208);
  Py_CLEAR(clear_module_state->__pyx_codeobj__209);
  Py_CLEAR(clear_module_state->__pyx_codeobj__210);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_date);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_time);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_datetime);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_timedelta);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_tzinfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_Field);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_Field);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata);
  Py_VISIT(traverse_module_state->C_Metadata_Type);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_Schema);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_Schema);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_BaseRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_PingRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_CallRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_EvalRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_SelectRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_InsertRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_IDRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_AuthRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_BeginRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_CommitRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_IProtoError);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_Response);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_Response);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_Db);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_Db);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_PushIterator);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth);
  Py_VISIT(traverse_module_state->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema);
  Py_VISIT(traverse_module_state->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema);
  Py_VISIT(traverse_module_state->__pyx_kp_u_02x);
  Py_VISIT(traverse_module_state->__pyx_n_s_AF_UNIX);
  Py_VISIT(traverse_module_state->__pyx_n_u_AF_UNIX);
  Py_VISIT(traverse_module_state->__pyx_n_s_ALL);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_AttributeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_AuthRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_AuthRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_AuthRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Available_Iterator_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_BITS_ALL_NOT_SET);
  Py_VISIT(traverse_module_state->__pyx_n_s_BITS_ALL_SET);
  Py_VISIT(traverse_module_state->__pyx_n_s_BITS_ANY_SET);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol__create_future_fall);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol__do_auth_locals_on);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol__do_fetch_schema_lo);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol__do_id_locals_on_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol__on_request_complet);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol__on_request_timeout);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol_create_db);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol_get_common_db);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol_refetch_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BeginRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_BeginRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BeginRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BufferError);
  Py_VISIT(traverse_module_state->__pyx_n_s_CallRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_CallRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_CallRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_CancelledError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_pop_push_from_a_non_async);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_use_iter_with_PushIterato);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_use_next_with_PushIterato);
  Py_VISIT(traverse_module_state->__pyx_n_s_CommitRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_CommitRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_CommitRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConnectionRefusedError);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConnectionResetError);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoreProtocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoreProtocol___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoreProtocol___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoreProtocol__parse_version);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoreProtocol_connection_lost);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoreProtocol_connection_made);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoreProtocol_data_received);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoreProtocol_get_version);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoreProtocol_is_connected);
  Py_VISIT(traverse_module_state->__pyx_n_s_CoreProtocol_is_fully_connected);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Creates_PushIterator_object_In);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_begin);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_call);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_call16);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_commit);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_delete);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_eval);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_execute);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_insert);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_ping);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_prepare);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_replace);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_rollback);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_select);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_set_stream_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_Db_upsert);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Deallocating_buffer_with_attache);
  Py_VISIT(traverse_module_state->__pyx_n_s_Decimal);
  Py_VISIT(traverse_module_state->__pyx_n_s_DeleteRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_DeleteRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_DeleteRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_EQ);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Error_happened_while_parsing_sch);
  Py_VISIT(traverse_module_state->__pyx_n_s_EvalRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_EvalRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_EvalRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Event);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExecuteRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExecuteRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExecuteRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Field);
  Py_VISIT(traverse_module_state->__pyx_n_s_Field___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Field___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Field_d_of_space_s_is_not_in_spa);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Field_map_must_contain_at_least);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Field_map_must_contain_at_least_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Field_name_type_is_nullable);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Field_not_found);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Field_s_in_supplied_dict_is_unkn);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Frame_type_code_message);
  Py_VISIT(traverse_module_state->__pyx_n_s_Future);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Future_is_invalid_Make_sure_to_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_GE);
  Py_VISIT(traverse_module_state->__pyx_n_s_GT);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Header_key_must_be_a_MP_UINT);
  Py_VISIT(traverse_module_state->__pyx_n_s_IDRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_IDRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_IDRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_IPPROTO_TCP);
  Py_VISIT(traverse_module_state->__pyx_kp_u_IPROTO_ERROR_type_must_be_a_MP_M);
  Py_VISIT(traverse_module_state->__pyx_kp_u_IPROTO_FIELD_SPAN_must_be_either);
  Py_VISIT(traverse_module_state->__pyx_kp_u_IPROTO_STMT_ID_type_must_be_a_MP);
  Py_VISIT(traverse_module_state->__pyx_n_s_IProtoError);
  Py_VISIT(traverse_module_state->__pyx_n_s_IProtoErrorStackFrame);
  Py_VISIT(traverse_module_state->__pyx_n_s_IProtoErrorStackFrame___reduce_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_IProtoErrorStackFrame___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_IProtoError___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_IProtoError___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_IProto_auth_type_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_IProto_features_available_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_IProto_version_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Impossible_condition_happened_Pl);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Index_must_be_either_str_or_int);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Index_not_found_in_space);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Index_s_not_found_in_space_s_s_C);
  Py_VISIT(traverse_module_state->__pyx_n_s_InsertRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_InsertRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_InsertRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntEnum);
  Py_VISIT(traverse_module_state->__pyx_n_s_Iterator);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Iterator_is_of_unsupported_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_KeyError);
  Py_VISIT(traverse_module_state->__pyx_n_s_LE);
  Py_VISIT(traverse_module_state->__pyx_n_s_LT);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Lost_connection_to_Tarantool);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Lost_connection_to_Tarantool_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Lost_connection_to_Tarantool_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Metadata);
  Py_VISIT(traverse_module_state->__pyx_n_s_Metadata___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Metadata___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Metadata_fields_count);
  Py_VISIT(traverse_module_state->__pyx_n_s_NEIGHBOR);
  Py_VISIT(traverse_module_state->__pyx_n_s_OVERLAPS);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Offset_incorrect_Got_use_len);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Operation_field_no_must_be_of_ei);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Operation_length_must_be_at_leas);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Operation_type_must_of_a_str_or);
  Py_VISIT(traverse_module_state->__pyx_n_s_Optional);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Part_len_must_be_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_PingRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_PingRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PingRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrepareRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrepareRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrepareRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_PushIterator);
  Py_VISIT(traverse_module_state->__pyx_n_s_PushIterator___anext);
  Py_VISIT(traverse_module_state->__pyx_kp_u_PushIterator___init___line_9);
  Py_VISIT(traverse_module_state->__pyx_n_s_PushIterator___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PushIterator___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_REQ);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadBuffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadBuffer___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadBuffer___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Request_op_sync_schema_id_push);
  Py_VISIT(traverse_module_state->__pyx_n_s_Response);
  Py_VISIT(traverse_module_state->__pyx_n_s_Response___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Response___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Response_body_must_be_a_MP_MAP);
  Py_VISIT(traverse_module_state->__pyx_n_s_Response_done);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Response_header_must_be_a_MP_MAP);
  Py_VISIT(traverse_module_state->__pyx_n_s_RollbackRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_RollbackRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_RollbackRequest___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_RuntimeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_SchemaDummyIndex);
  Py_VISIT(traverse_module_state->__pyx_n_s_SchemaDummySpace);
  Py_VISIT(traverse_module_state->__pyx_n_s_SchemaIndex);
  Py_VISIT(traverse_module_state->__pyx_n_s_SchemaIndex___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SchemaIndex___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SchemaSpace);
  Py_VISIT(traverse_module_state->__pyx_n_s_SchemaSpace___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SchemaSpace___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Schema_fetch_stopped_connection);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Schema_fetch_timeout);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Schema_spaces);
  Py_VISIT(traverse_module_state->__pyx_n_s_SelectRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_SelectRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SelectRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Single_operation_must_be_a_tuple);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Space_must_be_either_str_or_int);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Space_not_found);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Space_s_not_found_Creating_dummy);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Space_with_id_not_found);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Splice_offset_must_be_int);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Splice_operation_must_have_lengt);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Splice_position_must_be_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_TCP_NODELAY);
  Py_VISIT(traverse_module_state->__pyx_n_s_TarantoolDatabaseError);
  Py_VISIT(traverse_module_state->__pyx_n_s_TarantoolNotConnectedError);
  Py_VISIT(traverse_module_state->__pyx_n_s_TarantoolSchemaError);
  Py_VISIT(traverse_module_state->__pyx_n_s_TarantoolTuple);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Tarantool_is_not_connected);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Tarantool_s_s_Authorization_fail);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Tarantool_s_s_Authorized_success);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Tarantool_s_s_Schema_fetch_faile);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Tarantool_s_s_Schema_fetch_succe);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Tarantool_s_s_identification_fai);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Tarantool_s_s_identified_success);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Time_zones_are_not_available_fro);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimeoutError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Tuple_must_be_an_array_when_deco);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Type_is_not_supported_for_encodi);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUID);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unexpected_ext_type_d);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unexpected_key_type_in_map_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unexpected_obj_type_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnicodeDecodeError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unknown_key_with_code_d_in_heade);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unknown_update_operation_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_UpdateRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_UpdateRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UpdateRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UpsertRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_UpsertRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UpsertRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_VERSION_STRING_REGEX);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteBuffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteBuffer___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteBuffer___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteBuffer_hex);
  Py_VISIT(traverse_module_state->__pyx_kp_u__14);
  Py_VISIT(traverse_module_state->__pyx_kp_u__17);
  Py_VISIT(traverse_module_state->__pyx_kp_u__18);
  Py_VISIT(traverse_module_state->__pyx_kp_u__19);
  Py_VISIT(traverse_module_state->__pyx_kp_u__2);
  Py_VISIT(traverse_module_state->__pyx_n_s__211);
  Py_VISIT(traverse_module_state->__pyx_kp_u__3);
  Py_VISIT(traverse_module_state->__pyx_kp_u__5);
  Py_VISIT(traverse_module_state->__pyx_n_s__65);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_done_callback);
  Py_VISIT(traverse_module_state->__pyx_n_s_anext);
  Py_VISIT(traverse_module_state->__pyx_n_s_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_as_tuple);
  Py_VISIT(traverse_module_state->__pyx_n_u_ascii);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_tasks);
  Py_VISIT(traverse_module_state->__pyx_n_s_asynctnt);
  Py_VISIT(traverse_module_state->__pyx_n_s_asynctnt_exceptions);
  Py_VISIT(traverse_module_state->__pyx_kp_s_asynctnt_iproto_buffer_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_asynctnt_iproto_coreproto_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_asynctnt_iproto_db_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_asynctnt_iproto_protocol);
  Py_VISIT(traverse_module_state->__pyx_kp_s_asynctnt_iproto_protocol_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_asynctnt_iproto_response_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_asynctnt_log);
  Py_VISIT(traverse_module_state->__pyx_n_s_auto_refetch_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_await);
  Py_VISIT(traverse_module_state->__pyx_n_s_b64decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_base64);
  Py_VISIT(traverse_module_state->__pyx_n_s_begin);
  Py_VISIT(traverse_module_state->__pyx_kp_u_body_data_type_must_be_a_MP_ARRA);
  Py_VISIT(traverse_module_state->__pyx_n_s_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_call);
  Py_VISIT(traverse_module_state->__pyx_n_s_call16);
  Py_VISIT(traverse_module_state->__pyx_n_s_call_later);
  Py_VISIT(traverse_module_state->__pyx_n_s_cancel);
  Py_VISIT(traverse_module_state->__pyx_n_s_cancelled);
  Py_VISIT(traverse_module_state->__pyx_n_s_check_schema_change);
  Py_VISIT(traverse_module_state->__pyx_n_s_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_clear);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_closing);
  Py_VISIT(traverse_module_state->__pyx_kp_u_code_type_must_be_a_MP_UINT);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections);
  Py_VISIT(traverse_module_state->__pyx_n_s_commit);
  Py_VISIT(traverse_module_state->__pyx_n_s_compile);
  Py_VISIT(traverse_module_state->__pyx_n_s_connected_fut);
  Py_VISIT(traverse_module_state->__pyx_n_s_connection_lost);
  Py_VISIT(traverse_module_state->__pyx_n_s_connection_made);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_db);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_future);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_future_fallback);
  Py_VISIT(traverse_module_state->__pyx_n_s_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_data_received);
  Py_VISIT(traverse_module_state->__pyx_n_s_datetime);
  Py_VISIT(traverse_module_state->__pyx_n_s_debug);
  Py_VISIT(traverse_module_state->__pyx_n_s_decimal);
  Py_VISIT(traverse_module_state->__pyx_n_s_decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_delete);
  Py_VISIT(traverse_module_state->__pyx_n_s_deque);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_digest);
  Py_VISIT(traverse_module_state->__pyx_n_s_digits);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_doc);
  Py_VISIT(traverse_module_state->__pyx_n_s_done);
  Py_VISIT(traverse_module_state->__pyx_n_s_e);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_encoding);
  Py_VISIT(traverse_module_state->__pyx_kp_u_encoding_must_be_either_str_or_b);
  Py_VISIT(traverse_module_state->__pyx_n_s_enum);
  Py_VISIT(traverse_module_state->__pyx_n_s_error);
  Py_VISIT(traverse_module_state->__pyx_kp_u_errstr_type_must_be_a_MP_STR);
  Py_VISIT(traverse_module_state->__pyx_n_s_eval);
  Py_VISIT(traverse_module_state->__pyx_n_s_exc);
  Py_VISIT(traverse_module_state->__pyx_kp_u_exceeded_timeout);
  Py_VISIT(traverse_module_state->__pyx_n_s_exception);
  Py_VISIT(traverse_module_state->__pyx_n_s_execute);
  Py_VISIT(traverse_module_state->__pyx_n_s_exponent);
  Py_VISIT(traverse_module_state->__pyx_n_s_expression);
  Py_VISIT(traverse_module_state->__pyx_n_s_f);
  Py_VISIT(traverse_module_state->__pyx_n_s_family);
  Py_VISIT(traverse_module_state->__pyx_n_s_fetch_schema);
  Py_VISIT(traverse_module_state->__pyx_n_u_field);
  Py_VISIT(traverse_module_state->__pyx_kp_u_field_name_must_not_be_None);
  Py_VISIT(traverse_module_state->__pyx_n_s_format);
  Py_VISIT(traverse_module_state->__pyx_n_s_func_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_fut);
  Py_VISIT(traverse_module_state->__pyx_n_s_futures);
  Py_VISIT(traverse_module_state->__pyx_n_s_gather);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_gen_stream_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_common_db);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_extra_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_version);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_kp_u_got);
  Py_VISIT(traverse_module_state->__pyx_n_s_group);
  Py_VISIT(traverse_module_state->__pyx_n_s_hashlib);
  Py_VISIT(traverse_module_state->__pyx_n_s_hex);
  Py_VISIT(traverse_module_state->__pyx_n_s_hex_locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_host);
  Py_VISIT(traverse_module_state->__pyx_kp_u_id_name_engine);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_indexes);
  Py_VISIT(traverse_module_state->__pyx_n_s_init);
  Py_VISIT(traverse_module_state->__pyx_n_s_init_subclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_initial_read_buffer_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_insert);
  Py_VISIT(traverse_module_state->__pyx_n_s_inspect);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_datetime_size_got_extra);
  Py_VISIT(traverse_module_state->__pyx_kp_u_iproto_error_stack_frame_fields);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_connected);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_fully_connected);
  Py_VISIT(traverse_module_state->__pyx_n_u_is_nullable);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_isolation);
  Py_VISIT(traverse_module_state->__pyx_n_s_iterator);
  Py_VISIT(traverse_module_state->__pyx_n_s_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_last_stream_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_limit);
  Py_VISIT(traverse_module_state->__pyx_n_s_logger);
  Py_VISIT(traverse_module_state->__pyx_n_s_loop);
  Py_VISIT(traverse_module_state->__pyx_n_s_m);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_match);
  Py_VISIT(traverse_module_state->__pyx_n_s_metaclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_module);
  Py_VISIT(traverse_module_state->__pyx_n_s_mro_entries);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_VISIT(traverse_module_state->__pyx_n_s_offset);
  Py_VISIT(traverse_module_state->__pyx_n_s_on_authorized);
  Py_VISIT(traverse_module_state->__pyx_n_s_on_connection_lost);
  Py_VISIT(traverse_module_state->__pyx_n_s_on_connection_made);
  Py_VISIT(traverse_module_state->__pyx_kp_u_on_data_received_expected_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_on_fetch);
  Py_VISIT(traverse_module_state->__pyx_n_s_on_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_on_request_completed);
  Py_VISIT(traverse_module_state->__pyx_n_s_on_request_timeout);
  Py_VISIT(traverse_module_state->__pyx_n_s_operations);
  Py_VISIT(traverse_module_state->__pyx_n_s_parse_metadata);
  Py_VISIT(traverse_module_state->__pyx_n_s_parse_version);
  Py_VISIT(traverse_module_state->__pyx_n_s_password);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_s_ping);
  Py_VISIT(traverse_module_state->__pyx_n_s_popleft);
  Py_VISIT(traverse_module_state->__pyx_n_s_port);
  Py_VISIT(traverse_module_state->__pyx_n_s_prepare);
  Py_VISIT(traverse_module_state->__pyx_n_s_prepare_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_push_subscribe);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_AuthRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_BaseRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_BeginRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_CallRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_CommitRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_CoreProtocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_DeleteRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_EvalRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_ExecuteRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_IDRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_IProtoError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_IProtoErrorStackF);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_InsertRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PingRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PrepareRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PushIterator);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_RollbackRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_SelectRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_UpdateRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_UpsertRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_qualname);
  Py_VISIT(traverse_module_state->__pyx_n_s_query);
  Py_VISIT(traverse_module_state->__pyx_kp_u_query_must_be_either_str_or_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_re);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_refetch_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_replace);
  Py_VISIT(traverse_module_state->__pyx_n_s_repr);
  Py_VISIT(traverse_module_state->__pyx_n_s_repr___locals_lambda);
  Py_VISIT(traverse_module_state->__pyx_n_s_req);
  Py_VISIT(traverse_module_state->__pyx_n_s_request_timeout);
  Py_VISIT(traverse_module_state->__pyx_n_s_response);
  Py_VISIT(traverse_module_state->__pyx_n_u_response);
  Py_VISIT(traverse_module_state->__pyx_n_s_response_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_return_exceptions);
  Py_VISIT(traverse_module_state->__pyx_n_s_rollback);
  Py_VISIT(traverse_module_state->__pyx_n_s_rowcount);
  Py_VISIT(traverse_module_state->__pyx_kp_u_s_Tarantool_s_d_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_schema_id);
  Py_VISIT(traverse_module_state->__pyx_kp_u_schema_id_type_must_be_a_MP_UINT);
  Py_VISIT(traverse_module_state->__pyx_n_s_select);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_kp_s_self_execute_cannot_be_converted);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sequence_must_be_either_list_or);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sequence_must_be_either_list_tup);
  Py_VISIT(traverse_module_state->__pyx_n_s_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_exception);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_stream_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_setsockopt);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_sha1);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sid_id_name_type_unique);
  Py_VISIT(traverse_module_state->__pyx_n_s_sign);
  Py_VISIT(traverse_module_state->__pyx_n_s_sock);
  Py_VISIT(traverse_module_state->__pyx_n_s_socket);
  Py_VISIT(traverse_module_state->__pyx_n_u_socket);
  Py_VISIT(traverse_module_state->__pyx_n_s_space);
  Py_VISIT(traverse_module_state->__pyx_n_s_spaces);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_split);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_stream_id);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_super);
  Py_VISIT(traverse_module_state->__pyx_n_s_sync);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sync_d_not_found);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sync_rowcount_data);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sync_type_must_be_a_MP_UINT);
  Py_VISIT(traverse_module_state->__pyx_n_s_t);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_kp_u_the_buffer_is_in_read_only_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_timeout);
  Py_VISIT(traverse_module_state->__pyx_n_s_timestamp);
  Py_VISIT(traverse_module_state->__pyx_n_s_timezone);
  Py_VISIT(traverse_module_state->__pyx_n_s_total_seconds);
  Py_VISIT(traverse_module_state->__pyx_n_s_transport);
  Py_VISIT(traverse_module_state->__pyx_n_s_tx_timeout);
  Py_VISIT(traverse_module_state->__pyx_n_u_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_typing);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unexpected_exponent_type);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unexpected_type_of_part_must_be);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unknown_iproto_error_map_field_w);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unknown_iproto_error_stack_eleme);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unknown_key_in_body_map_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unknown_key_in_metadata_decoding);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unknown_key_in_sql_info_decoding);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_upsert);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_username);
  Py_VISIT(traverse_module_state->__pyx_n_s_utcoffset);
  Py_VISIT(traverse_module_state->__pyx_kp_b_utf_8);
  Py_VISIT(traverse_module_state->__pyx_n_s_uuid);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_ver);
  Py_VISIT(traverse_module_state->__pyx_n_s_version);
  Py_VISIT(traverse_module_state->__pyx_n_s_wait);
  Py_VISIT(traverse_module_state->__pyx_n_s_waiter);
  Py_VISIT(traverse_module_state->__pyx_n_s_warning);
  Py_VISIT(traverse_module_state->__pyx_n_s_write);
  Py_VISIT(traverse_module_state->__pyx_n_s_x);
  Py_VISIT(traverse_module_state->__pyx_float_0_0);
  Py_VISIT(traverse_module_state->__pyx_float_neg_1_0);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_2);
  Py_VISIT(traverse_module_state->__pyx_int_3);
  Py_VISIT(traverse_module_state->__pyx_int_4);
  Py_VISIT(traverse_module_state->__pyx_int_5);
  Py_VISIT(traverse_module_state->__pyx_int_6);
  Py_VISIT(traverse_module_state->__pyx_int_7);
  Py_VISIT(traverse_module_state->__pyx_int_8);
  Py_VISIT(traverse_module_state->__pyx_int_9);
  Py_VISIT(traverse_module_state->__pyx_int_10);
  Py_VISIT(traverse_module_state->__pyx_int_11);
  Py_VISIT(traverse_module_state->__pyx_int_20);
  Py_VISIT(traverse_module_state->__pyx_int_281);
  Py_VISIT(traverse_module_state->__pyx_int_289);
  Py_VISIT(traverse_module_state->__pyx_int_4796468);
  Py_VISIT(traverse_module_state->__pyx_int_4879619);
  Py_VISIT(traverse_module_state->__pyx_int_23039357);
  Py_VISIT(traverse_module_state->__pyx_int_25165850);
  Py_VISIT(traverse_module_state->__pyx_int_47556637);
  Py_VISIT(traverse_module_state->__pyx_int_51353055);
  Py_VISIT(traverse_module_state->__pyx_int_51694216);
  Py_VISIT(traverse_module_state->__pyx_int_53862661);
  Py_VISIT(traverse_module_state->__pyx_int_63007895);
  Py_VISIT(traverse_module_state->__pyx_int_92266676);
  Py_VISIT(traverse_module_state->__pyx_int_103773939);
  Py_VISIT(traverse_module_state->__pyx_int_105088600);
  Py_VISIT(traverse_module_state->__pyx_int_105824685);
  Py_VISIT(traverse_module_state->__pyx_int_129177485);
  Py_VISIT(traverse_module_state->__pyx_int_135652490);
  Py_VISIT(traverse_module_state->__pyx_int_144520718);
  Py_VISIT(traverse_module_state->__pyx_int_145344315);
  Py_VISIT(traverse_module_state->__pyx_int_145510168);
  Py_VISIT(traverse_module_state->__pyx_int_149085227);
  Py_VISIT(traverse_module_state->__pyx_int_150999062);
  Py_VISIT(traverse_module_state->__pyx_int_151261227);
  Py_VISIT(traverse_module_state->__pyx_int_159463343);
  Py_VISIT(traverse_module_state->__pyx_int_170745181);
  Py_VISIT(traverse_module_state->__pyx_int_172954204);
  Py_VISIT(traverse_module_state->__pyx_int_179467967);
  Py_VISIT(traverse_module_state->__pyx_int_180945731);
  Py_VISIT(traverse_module_state->__pyx_int_184220263);
  Py_VISIT(traverse_module_state->__pyx_int_192862663);
  Py_VISIT(traverse_module_state->__pyx_int_194105842);
  Py_VISIT(traverse_module_state->__pyx_int_196154451);
  Py_VISIT(traverse_module_state->__pyx_int_201714318);
  Py_VISIT(traverse_module_state->__pyx_int_209859363);
  Py_VISIT(traverse_module_state->__pyx_int_214473120);
  Py_VISIT(traverse_module_state->__pyx_int_215610203);
  Py_VISIT(traverse_module_state->__pyx_int_218707107);
  Py_VISIT(traverse_module_state->__pyx_int_219129172);
  Py_VISIT(traverse_module_state->__pyx_int_220368800);
  Py_VISIT(traverse_module_state->__pyx_int_225261924);
  Py_VISIT(traverse_module_state->__pyx_int_235564716);
  Py_VISIT(traverse_module_state->__pyx_int_244856512);
  Py_VISIT(traverse_module_state->__pyx_int_245819509);
  Py_VISIT(traverse_module_state->__pyx_int_246409365);
  Py_VISIT(traverse_module_state->__pyx_int_253265777);
  Py_VISIT(traverse_module_state->__pyx_int_255344631);
  Py_VISIT(traverse_module_state->__pyx_int_257562521);
  Py_VISIT(traverse_module_state->__pyx_int_264852134);
  Py_VISIT(traverse_module_state->__pyx_int_266727818);
  Py_VISIT(traverse_module_state->__pyx_int_267485924);
  Py_VISIT(traverse_module_state->__pyx_int_4294967295);
  Py_VISIT(traverse_module_state->__pyx_int_neg_2);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_tuple__4);
  Py_VISIT(traverse_module_state->__pyx_tuple__6);
  Py_VISIT(traverse_module_state->__pyx_tuple__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__8);
  Py_VISIT(traverse_module_state->__pyx_tuple__9);
  Py_VISIT(traverse_module_state->__pyx_slice__15);
  Py_VISIT(traverse_module_state->__pyx_slice__16);
  Py_VISIT(traverse_module_state->__pyx_slice__39);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_tuple__13);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__21);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__23);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__25);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__27);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__29);
  Py_VISIT(traverse_module_state->__pyx_tuple__30);
  Py_VISIT(traverse_module_state->__pyx_tuple__31);
  Py_VISIT(traverse_module_state->__pyx_tuple__32);
  Py_VISIT(traverse_module_state->__pyx_tuple__33);
  Py_VISIT(traverse_module_state->__pyx_tuple__34);
  Py_VISIT(traverse_module_state->__pyx_tuple__35);
  Py_VISIT(traverse_module_state->__pyx_tuple__36);
  Py_VISIT(traverse_module_state->__pyx_tuple__37);
  Py_VISIT(traverse_module_state->__pyx_tuple__38);
  Py_VISIT(traverse_module_state->__pyx_tuple__40);
  Py_VISIT(traverse_module_state->__pyx_tuple__41);
  Py_VISIT(traverse_module_state->__pyx_tuple__44);
  Py_VISIT(traverse_module_state->__pyx_tuple__46);
  Py_VISIT(traverse_module_state->__pyx_tuple__47);
  Py_VISIT(traverse_module_state->__pyx_tuple__48);
  Py_VISIT(traverse_module_state->__pyx_tuple__49);
  Py_VISIT(traverse_module_state->__pyx_tuple__50);
  Py_VISIT(traverse_module_state->__pyx_tuple__51);
  Py_VISIT(traverse_module_state->__pyx_tuple__52);
  Py_VISIT(traverse_module_state->__pyx_tuple__53);
  Py_VISIT(traverse_module_state->__pyx_tuple__54);
  Py_VISIT(traverse_module_state->__pyx_tuple__55);
  Py_VISIT(traverse_module_state->__pyx_tuple__56);
  Py_VISIT(traverse_module_state->__pyx_tuple__57);
  Py_VISIT(traverse_module_state->__pyx_tuple__58);
  Py_VISIT(traverse_module_state->__pyx_tuple__59);
  Py_VISIT(traverse_module_state->__pyx_tuple__60);
  Py_VISIT(traverse_module_state->__pyx_tuple__61);
  Py_VISIT(traverse_module_state->__pyx_tuple__62);
  Py_VISIT(traverse_module_state->__pyx_tuple__63);
  Py_VISIT(traverse_module_state->__pyx_tuple__64);
  Py_VISIT(traverse_module_state->__pyx_tuple__66);
  Py_VISIT(traverse_module_state->__pyx_tuple__68);
  Py_VISIT(traverse_module_state->__pyx_tuple__78);
  Py_VISIT(traverse_module_state->__pyx_tuple__84);
  Py_VISIT(traverse_module_state->__pyx_tuple__124);
  Py_VISIT(traverse_module_state->__pyx_tuple__126);
  Py_VISIT(traverse_module_state->__pyx_tuple__128);
  Py_VISIT(traverse_module_state->__pyx_tuple__129);
  Py_VISIT(traverse_module_state->__pyx_tuple__131);
  Py_VISIT(traverse_module_state->__pyx_tuple__133);
  Py_VISIT(traverse_module_state->__pyx_tuple__135);
  Py_VISIT(traverse_module_state->__pyx_tuple__137);
  Py_VISIT(traverse_module_state->__pyx_tuple__138);
  Py_VISIT(traverse_module_state->__pyx_tuple__140);
  Py_VISIT(traverse_module_state->__pyx_tuple__141);
  Py_VISIT(traverse_module_state->__pyx_tuple__143);
  Py_VISIT(traverse_module_state->__pyx_tuple__145);
  Py_VISIT(traverse_module_state->__pyx_tuple__146);
  Py_VISIT(traverse_module_state->__pyx_tuple__148);
  Py_VISIT(traverse_module_state->__pyx_tuple__150);
  Py_VISIT(traverse_module_state->__pyx_tuple__152);
  Py_VISIT(traverse_module_state->__pyx_tuple__153);
  Py_VISIT(traverse_module_state->__pyx_tuple__155);
  Py_VISIT(traverse_module_state->__pyx_tuple__157);
  Py_VISIT(traverse_module_state->__pyx_tuple__164);
  Py_VISIT(traverse_module_state->__pyx_tuple__168);
  Py_VISIT(traverse_module_state->__pyx_tuple__170);
  Py_VISIT(traverse_module_state->__pyx_tuple__172);
  Py_VISIT(traverse_module_state->__pyx_tuple__174);
  Py_VISIT(traverse_module_state->__pyx_tuple__179);
  Py_VISIT(traverse_module_state->__pyx_tuple__181);
  Py_VISIT(traverse_module_state->__pyx_tuple__183);
  Py_VISIT(traverse_module_state->__pyx_tuple__185);
  Py_VISIT(traverse_module_state->__pyx_tuple__190);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__43);
  Py_VISIT(traverse_module_state->__pyx_codeobj__45);
  Py_VISIT(traverse_module_state->__pyx_codeobj__67);
  Py_VISIT(traverse_module_state->__pyx_codeobj__69);
  Py_VISIT(traverse_module_state->__pyx_codeobj__70);
  Py_VISIT(traverse_module_state->__pyx_codeobj__71);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  Py_VISIT(traverse_module_state->__pyx_codeobj__73);
  Py_VISIT(traverse_module_state->__pyx_codeobj__74);
  Py_VISIT(traverse_module_state->__pyx_codeobj__75);
  Py_VISIT(traverse_module_state->__pyx_codeobj__76);
  Py_VISIT(traverse_module_state->__pyx_codeobj__77);
  Py_VISIT(traverse_module_state->__pyx_codeobj__79);
  Py_VISIT(traverse_module_state->__pyx_codeobj__80);
  Py_VISIT(traverse_module_state->__pyx_codeobj__81);
  Py_VISIT(traverse_module_state->__pyx_codeobj__82);
  Py_VISIT(traverse_module_state->__pyx_codeobj__83);
  Py_VISIT(traverse_module_state->__pyx_codeobj__85);
  Py_VISIT(traverse_module_state->__pyx_codeobj__86);
  Py_VISIT(traverse_module_state->__pyx_codeobj__87);
  Py_VISIT(traverse_module_state->__pyx_codeobj__88);
  Py_VISIT(traverse_module_state->__pyx_codeobj__89);
  Py_VISIT(traverse_module_state->__pyx_codeobj__90);
  Py_VISIT(traverse_module_state->__pyx_codeobj__91);
  Py_VISIT(traverse_module_state->__pyx_codeobj__92);
  Py_VISIT(traverse_module_state->__pyx_codeobj__93);
  Py_VISIT(traverse_module_state->__pyx_codeobj__94);
  Py_VISIT(traverse_module_state->__pyx_codeobj__95);
  Py_VISIT(traverse_module_state->__pyx_codeobj__96);
  Py_VISIT(traverse_module_state->__pyx_codeobj__97);
  Py_VISIT(traverse_module_state->__pyx_codeobj__98);
  Py_VISIT(traverse_module_state->__pyx_codeobj__99);
  Py_VISIT(traverse_module_state->__pyx_codeobj__100);
  Py_VISIT(traverse_module_state->__pyx_codeobj__101);
  Py_VISIT(traverse_module_state->__pyx_codeobj__102);
  Py_VISIT(traverse_module_state->__pyx_codeobj__103);
  Py_VISIT(traverse_module_state->__pyx_codeobj__104);
  Py_VISIT(traverse_module_state->__pyx_codeobj__105);
  Py_VISIT(traverse_module_state->__pyx_codeobj__106);
  Py_VISIT(traverse_module_state->__pyx_codeobj__107);
  Py_VISIT(traverse_module_state->__pyx_codeobj__108);
  Py_VISIT(traverse_module_state->__pyx_codeobj__109);
  Py_VISIT(traverse_module_state->__pyx_codeobj__110);
  Py_VISIT(traverse_module_state->__pyx_codeobj__111);
  Py_VISIT(traverse_module_state->__pyx_codeobj__112);
  Py_VISIT(traverse_module_state->__pyx_codeobj__113);
  Py_VISIT(traverse_module_state->__pyx_codeobj__114);
  Py_VISIT(traverse_module_state->__pyx_codeobj__115);
  Py_VISIT(traverse_module_state->__pyx_codeobj__116);
  Py_VISIT(traverse_module_state->__pyx_codeobj__117);
  Py_VISIT(traverse_module_state->__pyx_codeobj__118);
  Py_VISIT(traverse_module_state->__pyx_codeobj__119);
  Py_VISIT(traverse_module_state->__pyx_codeobj__120);
  Py_VISIT(traverse_module_state->__pyx_codeobj__121);
  Py_VISIT(traverse_module_state->__pyx_codeobj__122);
  Py_VISIT(traverse_module_state->__pyx_codeobj__123);
  Py_VISIT(traverse_module_state->__pyx_codeobj__125);
  Py_VISIT(traverse_module_state->__pyx_codeobj__127);
  Py_VISIT(traverse_module_state->__pyx_codeobj__130);
  Py_VISIT(traverse_module_state->__pyx_codeobj__132);
  Py_VISIT(traverse_module_state->__pyx_codeobj__134);
  Py_VISIT(traverse_module_state->__pyx_codeobj__136);
  Py_VISIT(traverse_module_state->__pyx_codeobj__139);
  Py_VISIT(traverse_module_state->__pyx_codeobj__142);
  Py_VISIT(traverse_module_state->__pyx_codeobj__144);
  Py_VISIT(traverse_module_state->__pyx_codeobj__147);
  Py_VISIT(traverse_module_state->__pyx_codeobj__149);
  Py_VISIT(traverse_module_state->__pyx_codeobj__151);
  Py_VISIT(traverse_module_state->__pyx_codeobj__154);
  Py_VISIT(traverse_module_state->__pyx_codeobj__156);
  Py_VISIT(traverse_module_state->__pyx_codeobj__158);
  Py_VISIT(traverse_module_state->__pyx_codeobj__159);
  Py_VISIT(traverse_module_state->__pyx_codeobj__160);
  Py_VISIT(traverse_module_state->__pyx_codeobj__161);
  Py_VISIT(traverse_module_state->__pyx_codeobj__162);
  Py_VISIT(traverse_module_state->__pyx_codeobj__163);
  Py_VISIT(traverse_module_state->__pyx_codeobj__165);
  Py_VISIT(traverse_module_state->__pyx_codeobj__166);
  Py_VISIT(traverse_module_state->__pyx_codeobj__167);
  Py_VISIT(traverse_module_state->__pyx_codeobj__169);
  Py_VISIT(traverse_module_state->__pyx_codeobj__171);
  Py_VISIT(traverse_module_state->__pyx_codeobj__173);
  Py_VISIT(traverse_module_state->__pyx_codeobj__175);
  Py_VISIT(traverse_module_state->__pyx_codeobj__176);
  Py_VISIT(traverse_module_state->__pyx_codeobj__177);
  Py_VISIT(traverse_module_state->__pyx_codeobj__178);
  Py_VISIT(traverse_module_state->__pyx_codeobj__180);
  Py_VISIT(traverse_module_state->__pyx_codeobj__182);
  Py_VISIT(traverse_module_state->__pyx_codeobj__184);
  Py_VISIT(traverse_module_state->__pyx_codeobj__186);
  Py_VISIT(traverse_module_state->__pyx_codeobj__187);
  Py_VISIT(traverse_module_state->__pyx_codeobj__188);
  Py_VISIT(traverse_module_state->__pyx_codeobj__189);
  Py_VISIT(traverse_module_state->__pyx_codeobj__191);
  Py_VISIT(traverse_module_state->__pyx_codeobj__192);
  Py_VISIT(traverse_module_state->__pyx_codeobj__193);
  Py_VISIT(traverse_module_state->__pyx_codeobj__194);
  Py_VISIT(traverse_module_state->__pyx_codeobj__195);
  Py_VISIT(traverse_module_state->__pyx_codeobj__196);
  Py_VISIT(traverse_module_state->__pyx_codeobj__197);
  Py_VISIT(traverse_module_state->__pyx_codeobj__198);
  Py_VISIT(traverse_module_state->__pyx_codeobj__199);
  Py_VISIT(traverse_module_state->__pyx_codeobj__200);
  Py_VISIT(traverse_module_state->__pyx_codeobj__201);
  Py_VISIT(traverse_module_state->__pyx_codeobj__202);
  Py_VISIT(traverse_module_state->__pyx_codeobj__203);
  Py_VISIT(traverse_module_state->__pyx_codeobj__204);
  Py_VISIT(traverse_module_state->__pyx_codeobj__205);
  Py_VISIT(traverse_module_state->__pyx_codeobj__206);
  Py_VISIT(traverse_module_state->__pyx_codeobj__207);
  Py_VISIT(traverse_module_state->__pyx_codeobj__208);
  Py_VISIT(traverse_module_state->__pyx_codeobj__209);
  Py_VISIT(traverse_module_state->__pyx_codeobj__210);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4bool_bool __pyx_mstate_global->__pyx_ptype_7cpython_4bool_bool
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_7complex_complex __pyx_mstate_global->__pyx_ptype_7cpython_7complex_complex
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_8datetime_date __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_date
#define __pyx_ptype_7cpython_8datetime_time __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_time
#define __pyx_ptype_7cpython_8datetime_datetime __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_datetime
#define __pyx_ptype_7cpython_8datetime_timedelta __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_timedelta
#define __pyx_ptype_7cpython_8datetime_tzinfo __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_tzinfo
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_8asynctnt_6iproto_8protocol_Field __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_Field
#define C_Metadata_Type __pyx_mstate_global->C_Metadata_Type
#define __pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex
#define __pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex
#define __pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace
#define __pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace
#define __pyx_type_8asynctnt_6iproto_8protocol_Schema __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_Schema
#define __pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer
#define __pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer
#define __pyx_type_8asynctnt_6iproto_8protocol_BaseRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_BaseRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_PingRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_PingRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_CallRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_CallRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_EvalRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_EvalRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_SelectRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_SelectRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_InsertRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_InsertRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_IDRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_IDRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_AuthRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_AuthRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_BeginRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_BeginRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_CommitRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_CommitRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest
#define __pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame
#define __pyx_type_8asynctnt_6iproto_8protocol_IProtoError __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_IProtoError
#define __pyx_type_8asynctnt_6iproto_8protocol_Response __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_Response
#define __pyx_type_8asynctnt_6iproto_8protocol_Db __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_Db
#define __pyx_type_8asynctnt_6iproto_8protocol_PushIterator __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_PushIterator
#define __pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol
#define __pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol
#define __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex
#define __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr
#define __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__
#define __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id
#define __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth
#define __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema __pyx_mstate_global->__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema
#endif
#define __pyx_ptype_8asynctnt_6iproto_8protocol_Field __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_Field
#define __pyx_ptype_8asynctnt_6iproto_8protocol_Metadata __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata
#define __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex
#define __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex
#define __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace
#define __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace
#define __pyx_ptype_8asynctnt_6iproto_8protocol_Schema __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_Schema
#define __pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer
#define __pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer
#define __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest
#define __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame
#define __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError
#define __pyx_ptype_8asynctnt_6iproto_8protocol_Response __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_Response
#define __pyx_ptype_8asynctnt_6iproto_8protocol_Db __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_Db
#define __pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator
#define __pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol
#define __pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol
#define __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex
#define __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr
#define __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__
#define __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id
#define __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth
#define __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema __pyx_mstate_global->__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema
#define __pyx_kp_u_02x __pyx_mstate_global->__pyx_kp_u_02x
#define __pyx_n_s_AF_UNIX __pyx_mstate_global->__pyx_n_s_AF_UNIX
#define __pyx_n_u_AF_UNIX __pyx_mstate_global->__pyx_n_u_AF_UNIX
#define __pyx_n_s_ALL __pyx_mstate_global->__pyx_n_s_ALL
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_s_AttributeError __pyx_mstate_global->__pyx_n_s_AttributeError
#define __pyx_n_s_AuthRequest __pyx_mstate_global->__pyx_n_s_AuthRequest
#define __pyx_n_s_AuthRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_AuthRequest___reduce_cython
#define __pyx_n_s_AuthRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_AuthRequest___setstate_cython
#define __pyx_kp_s_Available_Iterator_types __pyx_mstate_global->__pyx_kp_s_Available_Iterator_types
#define __pyx_n_s_BITS_ALL_NOT_SET __pyx_mstate_global->__pyx_n_s_BITS_ALL_NOT_SET
#define __pyx_n_s_BITS_ALL_SET __pyx_mstate_global->__pyx_n_s_BITS_ALL_SET
#define __pyx_n_s_BITS_ANY_SET __pyx_mstate_global->__pyx_n_s_BITS_ANY_SET
#define __pyx_n_s_BaseProtocol __pyx_mstate_global->__pyx_n_s_BaseProtocol
#define __pyx_n_s_BaseProtocol___reduce_cython __pyx_mstate_global->__pyx_n_s_BaseProtocol___reduce_cython
#define __pyx_n_s_BaseProtocol___setstate_cython __pyx_mstate_global->__pyx_n_s_BaseProtocol___setstate_cython
#define __pyx_n_s_BaseProtocol__create_future_fall __pyx_mstate_global->__pyx_n_s_BaseProtocol__create_future_fall
#define __pyx_n_s_BaseProtocol__do_auth_locals_on __pyx_mstate_global->__pyx_n_s_BaseProtocol__do_auth_locals_on
#define __pyx_n_s_BaseProtocol__do_fetch_schema_lo __pyx_mstate_global->__pyx_n_s_BaseProtocol__do_fetch_schema_lo
#define __pyx_n_s_BaseProtocol__do_id_locals_on_id __pyx_mstate_global->__pyx_n_s_BaseProtocol__do_id_locals_on_id
#define __pyx_n_s_BaseProtocol__on_request_complet __pyx_mstate_global->__pyx_n_s_BaseProtocol__on_request_complet
#define __pyx_n_s_BaseProtocol__on_request_timeout __pyx_mstate_global->__pyx_n_s_BaseProtocol__on_request_timeout
#define __pyx_n_s_BaseProtocol_create_db __pyx_mstate_global->__pyx_n_s_BaseProtocol_create_db
#define __pyx_n_s_BaseProtocol_get_common_db __pyx_mstate_global->__pyx_n_s_BaseProtocol_get_common_db
#define __pyx_n_s_BaseProtocol_refetch_schema __pyx_mstate_global->__pyx_n_s_BaseProtocol_refetch_schema
#define __pyx_n_s_BaseRequest __pyx_mstate_global->__pyx_n_s_BaseRequest
#define __pyx_n_s_BaseRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_BaseRequest___reduce_cython
#define __pyx_n_s_BaseRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_BaseRequest___setstate_cython
#define __pyx_n_s_BeginRequest __pyx_mstate_global->__pyx_n_s_BeginRequest
#define __pyx_n_s_BeginRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_BeginRequest___reduce_cython
#define __pyx_n_s_BeginRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_BeginRequest___setstate_cython
#define __pyx_n_s_BufferError __pyx_mstate_global->__pyx_n_s_BufferError
#define __pyx_n_s_CallRequest __pyx_mstate_global->__pyx_n_s_CallRequest
#define __pyx_n_s_CallRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_CallRequest___reduce_cython
#define __pyx_n_s_CallRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_CallRequest___setstate_cython
#define __pyx_n_s_CancelledError __pyx_mstate_global->__pyx_n_s_CancelledError
#define __pyx_kp_u_Cannot_pop_push_from_a_non_async __pyx_mstate_global->__pyx_kp_u_Cannot_pop_push_from_a_non_async
#define __pyx_kp_u_Cannot_use_iter_with_PushIterato __pyx_mstate_global->__pyx_kp_u_Cannot_use_iter_with_PushIterato
#define __pyx_kp_u_Cannot_use_next_with_PushIterato __pyx_mstate_global->__pyx_kp_u_Cannot_use_next_with_PushIterato
#define __pyx_n_s_CommitRequest __pyx_mstate_global->__pyx_n_s_CommitRequest
#define __pyx_n_s_CommitRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_CommitRequest___reduce_cython
#define __pyx_n_s_CommitRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_CommitRequest___setstate_cython
#define __pyx_n_s_ConnectionRefusedError __pyx_mstate_global->__pyx_n_s_ConnectionRefusedError
#define __pyx_n_s_ConnectionResetError __pyx_mstate_global->__pyx_n_s_ConnectionResetError
#define __pyx_n_s_CoreProtocol __pyx_mstate_global->__pyx_n_s_CoreProtocol
#define __pyx_n_s_CoreProtocol___reduce_cython __pyx_mstate_global->__pyx_n_s_CoreProtocol___reduce_cython
#define __pyx_n_s_CoreProtocol___setstate_cython __pyx_mstate_global->__pyx_n_s_CoreProtocol___setstate_cython
#define __pyx_n_s_CoreProtocol__parse_version __pyx_mstate_global->__pyx_n_s_CoreProtocol__parse_version
#define __pyx_n_s_CoreProtocol_connection_lost __pyx_mstate_global->__pyx_n_s_CoreProtocol_connection_lost
#define __pyx_n_s_CoreProtocol_connection_made __pyx_mstate_global->__pyx_n_s_CoreProtocol_connection_made
#define __pyx_n_s_CoreProtocol_data_received __pyx_mstate_global->__pyx_n_s_CoreProtocol_data_received
#define __pyx_n_s_CoreProtocol_get_version __pyx_mstate_global->__pyx_n_s_CoreProtocol_get_version
#define __pyx_n_s_CoreProtocol_is_connected __pyx_mstate_global->__pyx_n_s_CoreProtocol_is_connected
#define __pyx_n_s_CoreProtocol_is_fully_connected __pyx_mstate_global->__pyx_n_s_CoreProtocol_is_fully_connected
#define __pyx_kp_u_Creates_PushIterator_object_In __pyx_mstate_global->__pyx_kp_u_Creates_PushIterator_object_In
#define __pyx_n_s_Db __pyx_mstate_global->__pyx_n_s_Db
#define __pyx_n_s_Db___reduce_cython __pyx_mstate_global->__pyx_n_s_Db___reduce_cython
#define __pyx_n_s_Db___setstate_cython __pyx_mstate_global->__pyx_n_s_Db___setstate_cython
#define __pyx_n_s_Db_begin __pyx_mstate_global->__pyx_n_s_Db_begin
#define __pyx_n_s_Db_call __pyx_mstate_global->__pyx_n_s_Db_call
#define __pyx_n_s_Db_call16 __pyx_mstate_global->__pyx_n_s_Db_call16
#define __pyx_n_s_Db_commit __pyx_mstate_global->__pyx_n_s_Db_commit
#define __pyx_n_s_Db_delete __pyx_mstate_global->__pyx_n_s_Db_delete
#define __pyx_n_s_Db_eval __pyx_mstate_global->__pyx_n_s_Db_eval
#define __pyx_n_s_Db_execute __pyx_mstate_global->__pyx_n_s_Db_execute
#define __pyx_n_s_Db_insert __pyx_mstate_global->__pyx_n_s_Db_insert
#define __pyx_n_s_Db_ping __pyx_mstate_global->__pyx_n_s_Db_ping
#define __pyx_n_s_Db_prepare __pyx_mstate_global->__pyx_n_s_Db_prepare
#define __pyx_n_s_Db_replace __pyx_mstate_global->__pyx_n_s_Db_replace
#define __pyx_n_s_Db_rollback __pyx_mstate_global->__pyx_n_s_Db_rollback
#define __pyx_n_s_Db_select __pyx_mstate_global->__pyx_n_s_Db_select
#define __pyx_n_s_Db_set_stream_id __pyx_mstate_global->__pyx_n_s_Db_set_stream_id
#define __pyx_n_s_Db_update __pyx_mstate_global->__pyx_n_s_Db_update
#define __pyx_n_s_Db_upsert __pyx_mstate_global->__pyx_n_s_Db_upsert
#define __pyx_kp_u_Deallocating_buffer_with_attache __pyx_mstate_global->__pyx_kp_u_Deallocating_buffer_with_attache
#define __pyx_n_s_Decimal __pyx_mstate_global->__pyx_n_s_Decimal
#define __pyx_n_s_DeleteRequest __pyx_mstate_global->__pyx_n_s_DeleteRequest
#define __pyx_n_s_DeleteRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_DeleteRequest___reduce_cython
#define __pyx_n_s_DeleteRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_DeleteRequest___setstate_cython
#define __pyx_n_s_EQ __pyx_mstate_global->__pyx_n_s_EQ
#define __pyx_kp_u_Error_happened_while_parsing_sch __pyx_mstate_global->__pyx_kp_u_Error_happened_while_parsing_sch
#define __pyx_n_s_EvalRequest __pyx_mstate_global->__pyx_n_s_EvalRequest
#define __pyx_n_s_EvalRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_EvalRequest___reduce_cython
#define __pyx_n_s_EvalRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_EvalRequest___setstate_cython
#define __pyx_n_s_Event __pyx_mstate_global->__pyx_n_s_Event
#define __pyx_n_s_ExecuteRequest __pyx_mstate_global->__pyx_n_s_ExecuteRequest
#define __pyx_n_s_ExecuteRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_ExecuteRequest___reduce_cython
#define __pyx_n_s_ExecuteRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_ExecuteRequest___setstate_cython
#define __pyx_n_s_Field __pyx_mstate_global->__pyx_n_s_Field
#define __pyx_n_s_Field___reduce_cython __pyx_mstate_global->__pyx_n_s_Field___reduce_cython
#define __pyx_n_s_Field___setstate_cython __pyx_mstate_global->__pyx_n_s_Field___setstate_cython
#define __pyx_kp_u_Field_d_of_space_s_is_not_in_spa __pyx_mstate_global->__pyx_kp_u_Field_d_of_space_s_is_not_in_spa
#define __pyx_kp_u_Field_map_must_contain_at_least __pyx_mstate_global->__pyx_kp_u_Field_map_must_contain_at_least
#define __pyx_kp_u_Field_map_must_contain_at_least_2 __pyx_mstate_global->__pyx_kp_u_Field_map_must_contain_at_least_2
#define __pyx_kp_u_Field_name_type_is_nullable __pyx_mstate_global->__pyx_kp_u_Field_name_type_is_nullable
#define __pyx_kp_u_Field_not_found __pyx_mstate_global->__pyx_kp_u_Field_not_found
#define __pyx_kp_u_Field_s_in_supplied_dict_is_unkn __pyx_mstate_global->__pyx_kp_u_Field_s_in_supplied_dict_is_unkn
#define __pyx_kp_u_Frame_type_code_message __pyx_mstate_global->__pyx_kp_u_Frame_type_code_message
#define __pyx_n_s_Future __pyx_mstate_global->__pyx_n_s_Future
#define __pyx_kp_u_Future_is_invalid_Make_sure_to_c __pyx_mstate_global->__pyx_kp_u_Future_is_invalid_Make_sure_to_c
#define __pyx_n_s_GE __pyx_mstate_global->__pyx_n_s_GE
#define __pyx_n_s_GT __pyx_mstate_global->__pyx_n_s_GT
#define __pyx_kp_u_Header_key_must_be_a_MP_UINT __pyx_mstate_global->__pyx_kp_u_Header_key_must_be_a_MP_UINT
#define __pyx_n_s_IDRequest __pyx_mstate_global->__pyx_n_s_IDRequest
#define __pyx_n_s_IDRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_IDRequest___reduce_cython
#define __pyx_n_s_IDRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_IDRequest___setstate_cython
#define __pyx_n_s_IPPROTO_TCP __pyx_mstate_global->__pyx_n_s_IPPROTO_TCP
#define __pyx_kp_u_IPROTO_ERROR_type_must_be_a_MP_M __pyx_mstate_global->__pyx_kp_u_IPROTO_ERROR_type_must_be_a_MP_M
#define __pyx_kp_u_IPROTO_FIELD_SPAN_must_be_either __pyx_mstate_global->__pyx_kp_u_IPROTO_FIELD_SPAN_must_be_either
#define __pyx_kp_u_IPROTO_STMT_ID_type_must_be_a_MP __pyx_mstate_global->__pyx_kp_u_IPROTO_STMT_ID_type_must_be_a_MP
#define __pyx_n_s_IProtoError __pyx_mstate_global->__pyx_n_s_IProtoError
#define __pyx_n_s_IProtoErrorStackFrame __pyx_mstate_global->__pyx_n_s_IProtoErrorStackFrame
#define __pyx_n_s_IProtoErrorStackFrame___reduce_c __pyx_mstate_global->__pyx_n_s_IProtoErrorStackFrame___reduce_c
#define __pyx_n_s_IProtoErrorStackFrame___setstate __pyx_mstate_global->__pyx_n_s_IProtoErrorStackFrame___setstate
#define __pyx_n_s_IProtoError___reduce_cython __pyx_mstate_global->__pyx_n_s_IProtoError___reduce_cython
#define __pyx_n_s_IProtoError___setstate_cython __pyx_mstate_global->__pyx_n_s_IProtoError___setstate_cython
#define __pyx_kp_u_IProto_auth_type_s __pyx_mstate_global->__pyx_kp_u_IProto_auth_type_s
#define __pyx_kp_u_IProto_features_available_s __pyx_mstate_global->__pyx_kp_u_IProto_features_available_s
#define __pyx_kp_u_IProto_version_s __pyx_mstate_global->__pyx_kp_u_IProto_version_s
#define __pyx_kp_u_Impossible_condition_happened_Pl __pyx_mstate_global->__pyx_kp_u_Impossible_condition_happened_Pl
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9
#define __pyx_n_s_IndexError __pyx_mstate_global->__pyx_n_s_IndexError
#define __pyx_kp_u_Index_must_be_either_str_or_int __pyx_mstate_global->__pyx_kp_u_Index_must_be_either_str_or_int
#define __pyx_kp_u_Index_not_found_in_space __pyx_mstate_global->__pyx_kp_u_Index_not_found_in_space
#define __pyx_kp_u_Index_s_not_found_in_space_s_s_C __pyx_mstate_global->__pyx_kp_u_Index_s_not_found_in_space_s_s_C
#define __pyx_n_s_InsertRequest __pyx_mstate_global->__pyx_n_s_InsertRequest
#define __pyx_n_s_InsertRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_InsertRequest___reduce_cython
#define __pyx_n_s_InsertRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_InsertRequest___setstate_cython
#define __pyx_n_s_IntEnum __pyx_mstate_global->__pyx_n_s_IntEnum
#define __pyx_n_s_Iterator __pyx_mstate_global->__pyx_n_s_Iterator
#define __pyx_kp_u_Iterator_is_of_unsupported_type __pyx_mstate_global->__pyx_kp_u_Iterator_is_of_unsupported_type
#define __pyx_n_s_KeyError __pyx_mstate_global->__pyx_n_s_KeyError
#define __pyx_n_s_LE __pyx_mstate_global->__pyx_n_s_LE
#define __pyx_n_s_LT __pyx_mstate_global->__pyx_n_s_LT
#define __pyx_kp_u_Lost_connection_to_Tarantool __pyx_mstate_global->__pyx_kp_u_Lost_connection_to_Tarantool
#define __pyx_kp_u_Lost_connection_to_Tarantool_2 __pyx_mstate_global->__pyx_kp_u_Lost_connection_to_Tarantool_2
#define __pyx_kp_u_Lost_connection_to_Tarantool_3 __pyx_mstate_global->__pyx_kp_u_Lost_connection_to_Tarantool_3
#define __pyx_n_s_MemoryError __pyx_mstate_global->__pyx_n_s_MemoryError
#define __pyx_n_s_Metadata __pyx_mstate_global->__pyx_n_s_Metadata
#define __pyx_n_s_Metadata___reduce_cython __pyx_mstate_global->__pyx_n_s_Metadata___reduce_cython
#define __pyx_n_s_Metadata___setstate_cython __pyx_mstate_global->__pyx_n_s_Metadata___setstate_cython
#define __pyx_kp_u_Metadata_fields_count __pyx_mstate_global->__pyx_kp_u_Metadata_fields_count
#define __pyx_n_s_NEIGHBOR __pyx_mstate_global->__pyx_n_s_NEIGHBOR
#define __pyx_n_s_OVERLAPS __pyx_mstate_global->__pyx_n_s_OVERLAPS
#define __pyx_kp_u_Offset_incorrect_Got_use_len __pyx_mstate_global->__pyx_kp_u_Offset_incorrect_Got_use_len
#define __pyx_kp_u_Operation_field_no_must_be_of_ei __pyx_mstate_global->__pyx_kp_u_Operation_field_no_must_be_of_ei
#define __pyx_kp_u_Operation_length_must_be_at_leas __pyx_mstate_global->__pyx_kp_u_Operation_length_must_be_at_leas
#define __pyx_kp_u_Operation_type_must_of_a_str_or __pyx_mstate_global->__pyx_kp_u_Operation_type_must_of_a_str_or
#define __pyx_n_s_Optional __pyx_mstate_global->__pyx_n_s_Optional
#define __pyx_kp_u_Part_len_must_be_2 __pyx_mstate_global->__pyx_kp_u_Part_len_must_be_2
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_n_s_PingRequest __pyx_mstate_global->__pyx_n_s_PingRequest
#define __pyx_n_s_PingRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_PingRequest___reduce_cython
#define __pyx_n_s_PingRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_PingRequest___setstate_cython
#define __pyx_n_s_PrepareRequest __pyx_mstate_global->__pyx_n_s_PrepareRequest
#define __pyx_n_s_PrepareRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_PrepareRequest___reduce_cython
#define __pyx_n_s_PrepareRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_PrepareRequest___setstate_cython
#define __pyx_n_s_Protocol __pyx_mstate_global->__pyx_n_s_Protocol
#define __pyx_n_s_PushIterator __pyx_mstate_global->__pyx_n_s_PushIterator
#define __pyx_n_s_PushIterator___anext __pyx_mstate_global->__pyx_n_s_PushIterator___anext
#define __pyx_kp_u_PushIterator___init___line_9 __pyx_mstate_global->__pyx_kp_u_PushIterator___init___line_9
#define __pyx_n_s_PushIterator___reduce_cython __pyx_mstate_global->__pyx_n_s_PushIterator___reduce_cython
#define __pyx_n_s_PushIterator___setstate_cython __pyx_mstate_global->__pyx_n_s_PushIterator___setstate_cython
#define __pyx_n_s_REQ __pyx_mstate_global->__pyx_n_s_REQ
#define __pyx_n_s_ReadBuffer __pyx_mstate_global->__pyx_n_s_ReadBuffer
#define __pyx_n_s_ReadBuffer___reduce_cython __pyx_mstate_global->__pyx_n_s_ReadBuffer___reduce_cython
#define __pyx_n_s_ReadBuffer___setstate_cython __pyx_mstate_global->__pyx_n_s_ReadBuffer___setstate_cython
#define __pyx_kp_u_Request_op_sync_schema_id_push __pyx_mstate_global->__pyx_kp_u_Request_op_sync_schema_id_push
#define __pyx_n_s_Response __pyx_mstate_global->__pyx_n_s_Response
#define __pyx_n_s_Response___reduce_cython __pyx_mstate_global->__pyx_n_s_Response___reduce_cython
#define __pyx_n_s_Response___setstate_cython __pyx_mstate_global->__pyx_n_s_Response___setstate_cython
#define __pyx_kp_u_Response_body_must_be_a_MP_MAP __pyx_mstate_global->__pyx_kp_u_Response_body_must_be_a_MP_MAP
#define __pyx_n_s_Response_done __pyx_mstate_global->__pyx_n_s_Response_done
#define __pyx_kp_u_Response_header_must_be_a_MP_MAP __pyx_mstate_global->__pyx_kp_u_Response_header_must_be_a_MP_MAP
#define __pyx_n_s_RollbackRequest __pyx_mstate_global->__pyx_n_s_RollbackRequest
#define __pyx_n_s_RollbackRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_RollbackRequest___reduce_cython
#define __pyx_n_s_RollbackRequest___setstate_cytho __pyx_mstate_global->__pyx_n_s_RollbackRequest___setstate_cytho
#define __pyx_n_s_RuntimeError __pyx_mstate_global->__pyx_n_s_RuntimeError
#define __pyx_n_s_Schema __pyx_mstate_global->__pyx_n_s_Schema
#define __pyx_n_s_SchemaDummyIndex __pyx_mstate_global->__pyx_n_s_SchemaDummyIndex
#define __pyx_n_s_SchemaDummySpace __pyx_mstate_global->__pyx_n_s_SchemaDummySpace
#define __pyx_n_s_SchemaIndex __pyx_mstate_global->__pyx_n_s_SchemaIndex
#define __pyx_n_s_SchemaIndex___reduce_cython __pyx_mstate_global->__pyx_n_s_SchemaIndex___reduce_cython
#define __pyx_n_s_SchemaIndex___setstate_cython __pyx_mstate_global->__pyx_n_s_SchemaIndex___setstate_cython
#define __pyx_n_s_SchemaSpace __pyx_mstate_global->__pyx_n_s_SchemaSpace
#define __pyx_n_s_SchemaSpace___reduce_cython __pyx_mstate_global->__pyx_n_s_SchemaSpace___reduce_cython
#define __pyx_n_s_SchemaSpace___setstate_cython __pyx_mstate_global->__pyx_n_s_SchemaSpace___setstate_cython
#define __pyx_n_s_Schema___reduce_cython __pyx_mstate_global->__pyx_n_s_Schema___reduce_cython
#define __pyx_n_s_Schema___setstate_cython __pyx_mstate_global->__pyx_n_s_Schema___setstate_cython
#define __pyx_kp_u_Schema_fetch_stopped_connection __pyx_mstate_global->__pyx_kp_u_Schema_fetch_stopped_connection
#define __pyx_kp_u_Schema_fetch_timeout __pyx_mstate_global->__pyx_kp_u_Schema_fetch_timeout
#define __pyx_kp_u_Schema_spaces __pyx_mstate_global->__pyx_kp_u_Schema_spaces
#define __pyx_n_s_SelectRequest __pyx_mstate_global->__pyx_n_s_SelectRequest
#define __pyx_n_s_SelectRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_SelectRequest___reduce_cython
#define __pyx_n_s_SelectRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_SelectRequest___setstate_cython
#define __pyx_kp_u_Single_operation_must_be_a_tuple __pyx_mstate_global->__pyx_kp_u_Single_operation_must_be_a_tuple
#define __pyx_kp_u_Space_must_be_either_str_or_int __pyx_mstate_global->__pyx_kp_u_Space_must_be_either_str_or_int
#define __pyx_kp_u_Space_not_found __pyx_mstate_global->__pyx_kp_u_Space_not_found
#define __pyx_kp_u_Space_s_not_found_Creating_dummy __pyx_mstate_global->__pyx_kp_u_Space_s_not_found_Creating_dummy
#define __pyx_kp_u_Space_with_id_not_found __pyx_mstate_global->__pyx_kp_u_Space_with_id_not_found
#define __pyx_kp_u_Splice_offset_must_be_int __pyx_mstate_global->__pyx_kp_u_Splice_offset_must_be_int
#define __pyx_kp_u_Splice_operation_must_have_lengt __pyx_mstate_global->__pyx_kp_u_Splice_operation_must_have_lengt
#define __pyx_kp_u_Splice_position_must_be_int __pyx_mstate_global->__pyx_kp_u_Splice_position_must_be_int
#define __pyx_n_s_TCP_NODELAY __pyx_mstate_global->__pyx_n_s_TCP_NODELAY
#define __pyx_n_s_TarantoolDatabaseError __pyx_mstate_global->__pyx_n_s_TarantoolDatabaseError
#define __pyx_n_s_TarantoolNotConnectedError __pyx_mstate_global->__pyx_n_s_TarantoolNotConnectedError
#define __pyx_n_s_TarantoolSchemaError __pyx_mstate_global->__pyx_n_s_TarantoolSchemaError
#define __pyx_n_s_TarantoolTuple __pyx_mstate_global->__pyx_n_s_TarantoolTuple
#define __pyx_kp_u_Tarantool_is_not_connected __pyx_mstate_global->__pyx_kp_u_Tarantool_is_not_connected
#define __pyx_kp_u_Tarantool_s_s_Authorization_fail __pyx_mstate_global->__pyx_kp_u_Tarantool_s_s_Authorization_fail
#define __pyx_kp_u_Tarantool_s_s_Authorized_success __pyx_mstate_global->__pyx_kp_u_Tarantool_s_s_Authorized_success
#define __pyx_kp_u_Tarantool_s_s_Schema_fetch_faile __pyx_mstate_global->__pyx_kp_u_Tarantool_s_s_Schema_fetch_faile
#define __pyx_kp_u_Tarantool_s_s_Schema_fetch_succe __pyx_mstate_global->__pyx_kp_u_Tarantool_s_s_Schema_fetch_succe
#define __pyx_kp_u_Tarantool_s_s_identification_fai __pyx_mstate_global->__pyx_kp_u_Tarantool_s_s_identification_fai
#define __pyx_kp_u_Tarantool_s_s_identified_success __pyx_mstate_global->__pyx_kp_u_Tarantool_s_s_identified_success
#define __pyx_kp_u_Time_zones_are_not_available_fro __pyx_mstate_global->__pyx_kp_u_Time_zones_are_not_available_fro
#define __pyx_n_s_TimeoutError __pyx_mstate_global->__pyx_n_s_TimeoutError
#define __pyx_kp_u_Tuple_must_be_an_array_when_deco __pyx_mstate_global->__pyx_kp_u_Tuple_must_be_an_array_when_deco
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_kp_u_Type_is_not_supported_for_encodi __pyx_mstate_global->__pyx_kp_u_Type_is_not_supported_for_encodi
#define __pyx_n_s_UUID __pyx_mstate_global->__pyx_n_s_UUID
#define __pyx_kp_u_Unexpected_ext_type_d __pyx_mstate_global->__pyx_kp_u_Unexpected_ext_type_d
#define __pyx_kp_u_Unexpected_key_type_in_map_s __pyx_mstate_global->__pyx_kp_u_Unexpected_key_type_in_map_s
#define __pyx_kp_u_Unexpected_obj_type_s __pyx_mstate_global->__pyx_kp_u_Unexpected_obj_type_s
#define __pyx_n_s_UnicodeDecodeError __pyx_mstate_global->__pyx_n_s_UnicodeDecodeError
#define __pyx_kp_u_Unknown_key_with_code_d_in_heade __pyx_mstate_global->__pyx_kp_u_Unknown_key_with_code_d_in_heade
#define __pyx_kp_u_Unknown_update_operation_type __pyx_mstate_global->__pyx_kp_u_Unknown_update_operation_type
#define __pyx_n_s_UpdateRequest __pyx_mstate_global->__pyx_n_s_UpdateRequest
#define __pyx_n_s_UpdateRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_UpdateRequest___reduce_cython
#define __pyx_n_s_UpdateRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_UpdateRequest___setstate_cython
#define __pyx_n_s_UpsertRequest __pyx_mstate_global->__pyx_n_s_UpsertRequest
#define __pyx_n_s_UpsertRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_UpsertRequest___reduce_cython
#define __pyx_n_s_UpsertRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_UpsertRequest___setstate_cython
#define __pyx_n_s_VERSION_STRING_REGEX __pyx_mstate_global->__pyx_n_s_VERSION_STRING_REGEX
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_n_s_WriteBuffer __pyx_mstate_global->__pyx_n_s_WriteBuffer
#define __pyx_n_s_WriteBuffer___reduce_cython __pyx_mstate_global->__pyx_n_s_WriteBuffer___reduce_cython
#define __pyx_n_s_WriteBuffer___setstate_cython __pyx_mstate_global->__pyx_n_s_WriteBuffer___setstate_cython
#define __pyx_n_s_WriteBuffer_hex __pyx_mstate_global->__pyx_n_s_WriteBuffer_hex
#define __pyx_kp_u__14 __pyx_mstate_global->__pyx_kp_u__14
#define __pyx_kp_u__17 __pyx_mstate_global->__pyx_kp_u__17
#define __pyx_kp_u__18 __pyx_mstate_global->__pyx_kp_u__18
#define __pyx_kp_u__19 __pyx_mstate_global->__pyx_kp_u__19
#define __pyx_kp_u__2 __pyx_mstate_global->__pyx_kp_u__2
#define __pyx_n_s__211 __pyx_mstate_global->__pyx_n_s__211
#define __pyx_kp_u__3 __pyx_mstate_global->__pyx_kp_u__3
#define __pyx_kp_u__5 __pyx_mstate_global->__pyx_kp_u__5
#define __pyx_n_s__65 __pyx_mstate_global->__pyx_n_s__65
#define __pyx_n_s_add_done_callback __pyx_mstate_global->__pyx_n_s_add_done_callback
#define __pyx_n_s_anext __pyx_mstate_global->__pyx_n_s_anext
#define __pyx_n_s_append __pyx_mstate_global->__pyx_n_s_append
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_as_tuple __pyx_mstate_global->__pyx_n_s_as_tuple
#define __pyx_n_u_ascii __pyx_mstate_global->__pyx_n_u_ascii
#define __pyx_n_s_asyncio __pyx_mstate_global->__pyx_n_s_asyncio
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_asyncio_tasks __pyx_mstate_global->__pyx_n_s_asyncio_tasks
#define __pyx_n_s_asynctnt __pyx_mstate_global->__pyx_n_s_asynctnt
#define __pyx_n_s_asynctnt_exceptions __pyx_mstate_global->__pyx_n_s_asynctnt_exceptions
#define __pyx_kp_s_asynctnt_iproto_buffer_pyx __pyx_mstate_global->__pyx_kp_s_asynctnt_iproto_buffer_pyx
#define __pyx_kp_s_asynctnt_iproto_coreproto_pyx __pyx_mstate_global->__pyx_kp_s_asynctnt_iproto_coreproto_pyx
#define __pyx_kp_s_asynctnt_iproto_db_pyx __pyx_mstate_global->__pyx_kp_s_asynctnt_iproto_db_pyx
#define __pyx_n_s_asynctnt_iproto_protocol __pyx_mstate_global->__pyx_n_s_asynctnt_iproto_protocol
#define __pyx_kp_s_asynctnt_iproto_protocol_pyx __pyx_mstate_global->__pyx_kp_s_asynctnt_iproto_protocol_pyx
#define __pyx_kp_s_asynctnt_iproto_response_pyx __pyx_mstate_global->__pyx_kp_s_asynctnt_iproto_response_pyx
#define __pyx_n_s_asynctnt_log __pyx_mstate_global->__pyx_n_s_asynctnt_log
#define __pyx_n_s_auto_refetch_schema __pyx_mstate_global->__pyx_n_s_auto_refetch_schema
#define __pyx_n_s_await __pyx_mstate_global->__pyx_n_s_await
#define __pyx_n_s_b64decode __pyx_mstate_global->__pyx_n_s_b64decode
#define __pyx_n_s_base64 __pyx_mstate_global->__pyx_n_s_base64
#define __pyx_n_s_begin __pyx_mstate_global->__pyx_n_s_begin
#define __pyx_kp_u_body_data_type_must_be_a_MP_ARRA __pyx_mstate_global->__pyx_kp_u_body_data_type_must_be_a_MP_ARRA
#define __pyx_n_s_bytes __pyx_mstate_global->__pyx_n_s_bytes
#define __pyx_n_s_call __pyx_mstate_global->__pyx_n_s_call
#define __pyx_n_s_call16 __pyx_mstate_global->__pyx_n_s_call16
#define __pyx_n_s_call_later __pyx_mstate_global->__pyx_n_s_call_later
#define __pyx_n_s_cancel __pyx_mstate_global->__pyx_n_s_cancel
#define __pyx_n_s_cancelled __pyx_mstate_global->__pyx_n_s_cancelled
#define __pyx_n_s_check_schema_change __pyx_mstate_global->__pyx_n_s_check_schema_change
#define __pyx_n_s_class __pyx_mstate_global->__pyx_n_s_class
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_clear __pyx_mstate_global->__pyx_n_s_clear
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_s_closing __pyx_mstate_global->__pyx_n_s_closing
#define __pyx_kp_u_code_type_must_be_a_MP_UINT __pyx_mstate_global->__pyx_kp_u_code_type_must_be_a_MP_UINT
#define __pyx_n_s_collections __pyx_mstate_global->__pyx_n_s_collections
#define __pyx_n_s_commit __pyx_mstate_global->__pyx_n_s_commit
#define __pyx_n_s_compile __pyx_mstate_global->__pyx_n_s_compile
#define __pyx_n_s_connected_fut __pyx_mstate_global->__pyx_n_s_connected_fut
#define __pyx_n_s_connection_lost __pyx_mstate_global->__pyx_n_s_connection_lost
#define __pyx_n_s_connection_made __pyx_mstate_global->__pyx_n_s_connection_made
#define __pyx_n_s_create_db __pyx_mstate_global->__pyx_n_s_create_db
#define __pyx_n_s_create_future __pyx_mstate_global->__pyx_n_s_create_future
#define __pyx_n_s_create_future_fallback __pyx_mstate_global->__pyx_n_s_create_future_fallback
#define __pyx_n_s_data __pyx_mstate_global->__pyx_n_s_data
#define __pyx_n_s_data_received __pyx_mstate_global->__pyx_n_s_data_received
#define __pyx_n_s_datetime __pyx_mstate_global->__pyx_n_s_datetime
#define __pyx_n_s_debug __pyx_mstate_global->__pyx_n_s_debug
#define __pyx_n_s_decimal __pyx_mstate_global->__pyx_n_s_decimal
#define __pyx_n_s_decode __pyx_mstate_global->__pyx_n_s_decode
#define __pyx_n_s_delete __pyx_mstate_global->__pyx_n_s_delete
#define __pyx_n_s_deque __pyx_mstate_global->__pyx_n_s_deque
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_n_s_digest __pyx_mstate_global->__pyx_n_s_digest
#define __pyx_n_s_digits __pyx_mstate_global->__pyx_n_s_digits
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_doc __pyx_mstate_global->__pyx_n_s_doc
#define __pyx_n_s_done __pyx_mstate_global->__pyx_n_s_done
#define __pyx_n_s_e __pyx_mstate_global->__pyx_n_s_e
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_encode __pyx_mstate_global->__pyx_n_s_encode
#define __pyx_n_s_encoding __pyx_mstate_global->__pyx_n_s_encoding
#define __pyx_kp_u_encoding_must_be_either_str_or_b __pyx_mstate_global->__pyx_kp_u_encoding_must_be_either_str_or_b
#define __pyx_n_s_enum __pyx_mstate_global->__pyx_n_s_enum
#define __pyx_n_s_error __pyx_mstate_global->__pyx_n_s_error
#define __pyx_kp_u_errstr_type_must_be_a_MP_STR __pyx_mstate_global->__pyx_kp_u_errstr_type_must_be_a_MP_STR
#define __pyx_n_s_eval __pyx_mstate_global->__pyx_n_s_eval
#define __pyx_n_s_exc __pyx_mstate_global->__pyx_n_s_exc
#define __pyx_kp_u_exceeded_timeout __pyx_mstate_global->__pyx_kp_u_exceeded_timeout
#define __pyx_n_s_exception __pyx_mstate_global->__pyx_n_s_exception
#define __pyx_n_s_execute __pyx_mstate_global->__pyx_n_s_execute
#define __pyx_n_s_exponent __pyx_mstate_global->__pyx_n_s_exponent
#define __pyx_n_s_expression __pyx_mstate_global->__pyx_n_s_expression
#define __pyx_n_s_f __pyx_mstate_global->__pyx_n_s_f
#define __pyx_n_s_family __pyx_mstate_global->__pyx_n_s_family
#define __pyx_n_s_fetch_schema __pyx_mstate_global->__pyx_n_s_fetch_schema
#define __pyx_n_u_field __pyx_mstate_global->__pyx_n_u_field
#define __pyx_kp_u_field_name_must_not_be_None __pyx_mstate_global->__pyx_kp_u_field_name_must_not_be_None
#define __pyx_n_s_format __pyx_mstate_global->__pyx_n_s_format
#define __pyx_n_s_func_name __pyx_mstate_global->__pyx_n_s_func_name
#define __pyx_n_s_fut __pyx_mstate_global->__pyx_n_s_fut
#define __pyx_n_s_futures __pyx_mstate_global->__pyx_n_s_futures
#define __pyx_n_s_gather __pyx_mstate_global->__pyx_n_s_gather
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_gen_stream_id __pyx_mstate_global->__pyx_n_s_gen_stream_id
#define __pyx_n_s_genexpr __pyx_mstate_global->__pyx_n_s_genexpr
#define __pyx_n_s_get __pyx_mstate_global->__pyx_n_s_get
#define __pyx_n_s_get_common_db __pyx_mstate_global->__pyx_n_s_get_common_db
#define __pyx_n_s_get_extra_info __pyx_mstate_global->__pyx_n_s_get_extra_info
#define __pyx_n_s_get_version __pyx_mstate_global->__pyx_n_s_get_version
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_kp_u_got __pyx_mstate_global->__pyx_kp_u_got
#define __pyx_n_s_group __pyx_mstate_global->__pyx_n_s_group
#define __pyx_n_s_hashlib __pyx_mstate_global->__pyx_n_s_hashlib
#define __pyx_n_s_hex __pyx_mstate_global->__pyx_n_s_hex
#define __pyx_n_s_hex_locals_genexpr __pyx_mstate_global->__pyx_n_s_hex_locals_genexpr
#define __pyx_n_s_host __pyx_mstate_global->__pyx_n_s_host
#define __pyx_kp_u_id_name_engine __pyx_mstate_global->__pyx_kp_u_id_name_engine
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_index __pyx_mstate_global->__pyx_n_s_index
#define __pyx_n_s_indexes __pyx_mstate_global->__pyx_n_s_indexes
#define __pyx_n_s_init __pyx_mstate_global->__pyx_n_s_init
#define __pyx_n_s_init_subclass __pyx_mstate_global->__pyx_n_s_init_subclass
#define __pyx_n_s_initial_read_buffer_size __pyx_mstate_global->__pyx_n_s_initial_read_buffer_size
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_insert __pyx_mstate_global->__pyx_n_s_insert
#define __pyx_n_s_inspect __pyx_mstate_global->__pyx_n_s_inspect
#define __pyx_kp_u_invalid_datetime_size_got_extra __pyx_mstate_global->__pyx_kp_u_invalid_datetime_size_got_extra
#define __pyx_kp_u_iproto_error_stack_frame_fields __pyx_mstate_global->__pyx_kp_u_iproto_error_stack_frame_fields
#define __pyx_n_s_is_connected __pyx_mstate_global->__pyx_n_s_is_connected
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_n_s_is_fully_connected __pyx_mstate_global->__pyx_n_s_is_fully_connected
#define __pyx_n_u_is_nullable __pyx_mstate_global->__pyx_n_u_is_nullable
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_isolation __pyx_mstate_global->__pyx_n_s_isolation
#define __pyx_n_s_iterator __pyx_mstate_global->__pyx_n_s_iterator
#define __pyx_n_s_key __pyx_mstate_global->__pyx_n_s_key
#define __pyx_n_s_last_stream_id __pyx_mstate_global->__pyx_n_s_last_stream_id
#define __pyx_n_s_limit __pyx_mstate_global->__pyx_n_s_limit
#define __pyx_n_s_logger __pyx_mstate_global->__pyx_n_s_logger
#define __pyx_n_s_loop __pyx_mstate_global->__pyx_n_s_loop
#define __pyx_n_s_m __pyx_mstate_global->__pyx_n_s_m
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_map __pyx_mstate_global->__pyx_n_s_map
#define __pyx_n_s_match __pyx_mstate_global->__pyx_n_s_match
#define __pyx_n_s_metaclass __pyx_mstate_global->__pyx_n_s_metaclass
#define __pyx_n_s_module __pyx_mstate_global->__pyx_n_s_module
#define __pyx_n_s_mro_entries __pyx_mstate_global->__pyx_n_s_mro_entries
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_u_name_2 __pyx_mstate_global->__pyx_n_u_name_2
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_kp_s_no_default___reduce___due_to_non __pyx_mstate_global->__pyx_kp_s_no_default___reduce___due_to_non
#define __pyx_n_s_offset __pyx_mstate_global->__pyx_n_s_offset
#define __pyx_n_s_on_authorized __pyx_mstate_global->__pyx_n_s_on_authorized
#define __pyx_n_s_on_connection_lost __pyx_mstate_global->__pyx_n_s_on_connection_lost
#define __pyx_n_s_on_connection_made __pyx_mstate_global->__pyx_n_s_on_connection_made
#define __pyx_kp_u_on_data_received_expected_bytes __pyx_mstate_global->__pyx_kp_u_on_data_received_expected_bytes
#define __pyx_n_s_on_fetch __pyx_mstate_global->__pyx_n_s_on_fetch
#define __pyx_n_s_on_id __pyx_mstate_global->__pyx_n_s_on_id
#define __pyx_n_s_on_request_completed __pyx_mstate_global->__pyx_n_s_on_request_completed
#define __pyx_n_s_on_request_timeout __pyx_mstate_global->__pyx_n_s_on_request_timeout
#define __pyx_n_s_operations __pyx_mstate_global->__pyx_n_s_operations
#define __pyx_n_s_parse_metadata __pyx_mstate_global->__pyx_n_s_parse_metadata
#define __pyx_n_s_parse_version __pyx_mstate_global->__pyx_n_s_parse_version
#define __pyx_n_s_password __pyx_mstate_global->__pyx_n_s_password
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_s_ping __pyx_mstate_global->__pyx_n_s_ping
#define __pyx_n_s_popleft __pyx_mstate_global->__pyx_n_s_popleft
#define __pyx_n_s_port __pyx_mstate_global->__pyx_n_s_port
#define __pyx_n_s_prepare __pyx_mstate_global->__pyx_n_s_prepare
#define __pyx_n_s_prepare_2 __pyx_mstate_global->__pyx_n_s_prepare_2
#define __pyx_n_s_push_subscribe __pyx_mstate_global->__pyx_n_s_push_subscribe
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_AuthRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_AuthRequest
#define __pyx_n_s_pyx_unpickle_BaseRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_BaseRequest
#define __pyx_n_s_pyx_unpickle_BeginRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_BeginRequest
#define __pyx_n_s_pyx_unpickle_CallRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_CallRequest
#define __pyx_n_s_pyx_unpickle_CommitRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_CommitRequest
#define __pyx_n_s_pyx_unpickle_CoreProtocol __pyx_mstate_global->__pyx_n_s_pyx_unpickle_CoreProtocol
#define __pyx_n_s_pyx_unpickle_DeleteRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_DeleteRequest
#define __pyx_n_s_pyx_unpickle_EvalRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_EvalRequest
#define __pyx_n_s_pyx_unpickle_ExecuteRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_ExecuteRequest
#define __pyx_n_s_pyx_unpickle_IDRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_IDRequest
#define __pyx_n_s_pyx_unpickle_IProtoError __pyx_mstate_global->__pyx_n_s_pyx_unpickle_IProtoError
#define __pyx_n_s_pyx_unpickle_IProtoErrorStackF __pyx_mstate_global->__pyx_n_s_pyx_unpickle_IProtoErrorStackF
#define __pyx_n_s_pyx_unpickle_InsertRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_InsertRequest
#define __pyx_n_s_pyx_unpickle_PingRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PingRequest
#define __pyx_n_s_pyx_unpickle_PrepareRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PrepareRequest
#define __pyx_n_s_pyx_unpickle_PushIterator __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PushIterator
#define __pyx_n_s_pyx_unpickle_RollbackRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_RollbackRequest
#define __pyx_n_s_pyx_unpickle_SelectRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_SelectRequest
#define __pyx_n_s_pyx_unpickle_UpdateRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_UpdateRequest
#define __pyx_n_s_pyx_unpickle_UpsertRequest __pyx_mstate_global->__pyx_n_s_pyx_unpickle_UpsertRequest
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_qualname __pyx_mstate_global->__pyx_n_s_qualname
#define __pyx_n_s_query __pyx_mstate_global->__pyx_n_s_query
#define __pyx_kp_u_query_must_be_either_str_or_int __pyx_mstate_global->__pyx_kp_u_query_must_be_either_str_or_int
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_re __pyx_mstate_global->__pyx_n_s_re
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_refetch_schema __pyx_mstate_global->__pyx_n_s_refetch_schema
#define __pyx_n_s_replace __pyx_mstate_global->__pyx_n_s_replace
#define __pyx_n_s_repr __pyx_mstate_global->__pyx_n_s_repr
#define __pyx_n_s_repr___locals_lambda __pyx_mstate_global->__pyx_n_s_repr___locals_lambda
#define __pyx_n_s_req __pyx_mstate_global->__pyx_n_s_req
#define __pyx_n_s_request_timeout __pyx_mstate_global->__pyx_n_s_request_timeout
#define __pyx_n_s_response __pyx_mstate_global->__pyx_n_s_response
#define __pyx_n_u_response __pyx_mstate_global->__pyx_n_u_response
#define __pyx_n_s_response_2 __pyx_mstate_global->__pyx_n_s_response_2
#define __pyx_n_s_result __pyx_mstate_global->__pyx_n_s_result
#define __pyx_n_s_return_exceptions __pyx_mstate_global->__pyx_n_s_return_exceptions
#define __pyx_n_s_rollback __pyx_mstate_global->__pyx_n_s_rollback
#define __pyx_n_s_rowcount __pyx_mstate_global->__pyx_n_s_rowcount
#define __pyx_kp_u_s_Tarantool_s_d_s __pyx_mstate_global->__pyx_kp_u_s_Tarantool_s_d_s
#define __pyx_n_s_schema_id __pyx_mstate_global->__pyx_n_s_schema_id
#define __pyx_kp_u_schema_id_type_must_be_a_MP_UINT __pyx_mstate_global->__pyx_kp_u_schema_id_type_must_be_a_MP_UINT
#define __pyx_n_s_select __pyx_mstate_global->__pyx_n_s_select
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_kp_s_self_execute_cannot_be_converted __pyx_mstate_global->__pyx_kp_s_self_execute_cannot_be_converted
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_kp_u_sequence_must_be_either_list_or __pyx_mstate_global->__pyx_kp_u_sequence_must_be_either_list_or
#define __pyx_kp_u_sequence_must_be_either_list_tup __pyx_mstate_global->__pyx_kp_u_sequence_must_be_either_list_tup
#define __pyx_n_s_set __pyx_mstate_global->__pyx_n_s_set
#define __pyx_n_s_set_exception __pyx_mstate_global->__pyx_n_s_set_exception
#define __pyx_n_s_set_name __pyx_mstate_global->__pyx_n_s_set_name
#define __pyx_n_s_set_result __pyx_mstate_global->__pyx_n_s_set_result
#define __pyx_n_s_set_stream_id __pyx_mstate_global->__pyx_n_s_set_stream_id
#define __pyx_n_s_setsockopt __pyx_mstate_global->__pyx_n_s_setsockopt
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_sha1 __pyx_mstate_global->__pyx_n_s_sha1
#define __pyx_kp_u_sid_id_name_type_unique __pyx_mstate_global->__pyx_kp_u_sid_id_name_type_unique
#define __pyx_n_s_sign __pyx_mstate_global->__pyx_n_s_sign
#define __pyx_n_s_sock __pyx_mstate_global->__pyx_n_s_sock
#define __pyx_n_s_socket __pyx_mstate_global->__pyx_n_s_socket
#define __pyx_n_u_socket __pyx_mstate_global->__pyx_n_u_socket
#define __pyx_n_s_space __pyx_mstate_global->__pyx_n_s_space
#define __pyx_n_s_spaces __pyx_mstate_global->__pyx_n_s_spaces
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_split __pyx_mstate_global->__pyx_n_s_split
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_stream_id __pyx_mstate_global->__pyx_n_s_stream_id
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_super __pyx_mstate_global->__pyx_n_s_super
#define __pyx_n_s_sync __pyx_mstate_global->__pyx_n_s_sync
#define __pyx_kp_u_sync_d_not_found __pyx_mstate_global->__pyx_kp_u_sync_d_not_found
#define __pyx_kp_u_sync_rowcount_data __pyx_mstate_global->__pyx_kp_u_sync_rowcount_data
#define __pyx_kp_u_sync_type_must_be_a_MP_UINT __pyx_mstate_global->__pyx_kp_u_sync_type_must_be_a_MP_UINT
#define __pyx_n_s_t __pyx_mstate_global->__pyx_n_s_t
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_kp_u_the_buffer_is_in_read_only_mode __pyx_mstate_global->__pyx_kp_u_the_buffer_is_in_read_only_mode
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_timeout __pyx_mstate_global->__pyx_n_s_timeout
#define __pyx_n_s_timestamp __pyx_mstate_global->__pyx_n_s_timestamp
#define __pyx_n_s_timezone __pyx_mstate_global->__pyx_n_s_timezone
#define __pyx_n_s_total_seconds __pyx_mstate_global->__pyx_n_s_total_seconds
#define __pyx_n_s_transport __pyx_mstate_global->__pyx_n_s_transport
#define __pyx_n_s_tx_timeout __pyx_mstate_global->__pyx_n_s_tx_timeout
#define __pyx_n_u_type __pyx_mstate_global->__pyx_n_u_type
#define __pyx_n_s_typing __pyx_mstate_global->__pyx_n_s_typing
#define __pyx_kp_u_unexpected_exponent_type __pyx_mstate_global->__pyx_kp_u_unexpected_exponent_type
#define __pyx_kp_u_unexpected_type_of_part_must_be __pyx_mstate_global->__pyx_kp_u_unexpected_type_of_part_must_be
#define __pyx_kp_u_unknown_iproto_error_map_field_w __pyx_mstate_global->__pyx_kp_u_unknown_iproto_error_map_field_w
#define __pyx_kp_u_unknown_iproto_error_stack_eleme __pyx_mstate_global->__pyx_kp_u_unknown_iproto_error_stack_eleme
#define __pyx_kp_u_unknown_key_in_body_map_s __pyx_mstate_global->__pyx_kp_u_unknown_key_in_body_map_s
#define __pyx_kp_u_unknown_key_in_metadata_decoding __pyx_mstate_global->__pyx_kp_u_unknown_key_in_metadata_decoding
#define __pyx_kp_u_unknown_key_in_sql_info_decoding __pyx_mstate_global->__pyx_kp_u_unknown_key_in_sql_info_decoding
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_upsert __pyx_mstate_global->__pyx_n_s_upsert
#define __pyx_n_s_use_setstate __pyx_mstate_global->__pyx_n_s_use_setstate
#define __pyx_n_s_username __pyx_mstate_global->__pyx_n_s_username
#define __pyx_n_s_utcoffset __pyx_mstate_global->__pyx_n_s_utcoffset
#define __pyx_kp_b_utf_8 __pyx_mstate_global->__pyx_kp_b_utf_8
#define __pyx_n_s_uuid __pyx_mstate_global->__pyx_n_s_uuid
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_s_ver __pyx_mstate_global->__pyx_n_s_ver
#define __pyx_n_s_version __pyx_mstate_global->__pyx_n_s_version
#define __pyx_n_s_wait __pyx_mstate_global->__pyx_n_s_wait
#define __pyx_n_s_waiter __pyx_mstate_global->__pyx_n_s_waiter
#define __pyx_n_s_warning __pyx_mstate_global->__pyx_n_s_warning
#define __pyx_n_s_write __pyx_mstate_global->__pyx_n_s_write
#define __pyx_n_s_x __pyx_mstate_global->__pyx_n_s_x
#define __pyx_float_0_0 __pyx_mstate_global->__pyx_float_0_0
#define __pyx_float_neg_1_0 __pyx_mstate_global->__pyx_float_neg_1_0
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_2 __pyx_mstate_global->__pyx_int_2
#define __pyx_int_3 __pyx_mstate_global->__pyx_int_3
#define __pyx_int_4 __pyx_mstate_global->__pyx_int_4
#define __pyx_int_5 __pyx_mstate_global->__pyx_int_5
#define __pyx_int_6 __pyx_mstate_global->__pyx_int_6
#define __pyx_int_7 __pyx_mstate_global->__pyx_int_7
#define __pyx_int_8 __pyx_mstate_global->__pyx_int_8
#define __pyx_int_9 __pyx_mstate_global->__pyx_int_9
#define __pyx_int_10 __pyx_mstate_global->__pyx_int_10
#define __pyx_int_11 __pyx_mstate_global->__pyx_int_11
#define __pyx_int_20 __pyx_mstate_global->__pyx_int_20
#define __pyx_int_281 __pyx_mstate_global->__pyx_int_281
#define __pyx_int_289 __pyx_mstate_global->__pyx_int_289
#define __pyx_int_4796468 __pyx_mstate_global->__pyx_int_4796468
#define __pyx_int_4879619 __pyx_mstate_global->__pyx_int_4879619
#define __pyx_int_23039357 __pyx_mstate_global->__pyx_int_23039357
#define __pyx_int_25165850 __pyx_mstate_global->__pyx_int_25165850
#define __pyx_int_47556637 __pyx_mstate_global->__pyx_int_47556637
#define __pyx_int_51353055 __pyx_mstate_global->__pyx_int_51353055
#define __pyx_int_51694216 __pyx_mstate_global->__pyx_int_51694216
#define __pyx_int_53862661 __pyx_mstate_global->__pyx_int_53862661
#define __pyx_int_63007895 __pyx_mstate_global->__pyx_int_63007895
#define __pyx_int_92266676 __pyx_mstate_global->__pyx_int_92266676
#define __pyx_int_103773939 __pyx_mstate_global->__pyx_int_103773939
#define __pyx_int_105088600 __pyx_mstate_global->__pyx_int_105088600
#define __pyx_int_105824685 __pyx_mstate_global->__pyx_int_105824685
#define __pyx_int_129177485 __pyx_mstate_global->__pyx_int_129177485
#define __pyx_int_135652490 __pyx_mstate_global->__pyx_int_135652490
#define __pyx_int_144520718 __pyx_mstate_global->__pyx_int_144520718
#define __pyx_int_145344315 __pyx_mstate_global->__pyx_int_145344315
#define __pyx_int_145510168 __pyx_mstate_global->__pyx_int_145510168
#define __pyx_int_149085227 __pyx_mstate_global->__pyx_int_149085227
#define __pyx_int_150999062 __pyx_mstate_global->__pyx_int_150999062
#define __pyx_int_151261227 __pyx_mstate_global->__pyx_int_151261227
#define __pyx_int_159463343 __pyx_mstate_global->__pyx_int_159463343
#define __pyx_int_170745181 __pyx_mstate_global->__pyx_int_170745181
#define __pyx_int_172954204 __pyx_mstate_global->__pyx_int_172954204
#define __pyx_int_179467967 __pyx_mstate_global->__pyx_int_179467967
#define __pyx_int_180945731 __pyx_mstate_global->__pyx_int_180945731
#define __pyx_int_184220263 __pyx_mstate_global->__pyx_int_184220263
#define __pyx_int_192862663 __pyx_mstate_global->__pyx_int_192862663
#define __pyx_int_194105842 __pyx_mstate_global->__pyx_int_194105842
#define __pyx_int_196154451 __pyx_mstate_global->__pyx_int_196154451
#define __pyx_int_201714318 __pyx_mstate_global->__pyx_int_201714318
#define __pyx_int_209859363 __pyx_mstate_global->__pyx_int_209859363
#define __pyx_int_214473120 __pyx_mstate_global->__pyx_int_214473120
#define __pyx_int_215610203 __pyx_mstate_global->__pyx_int_215610203
#define __pyx_int_218707107 __pyx_mstate_global->__pyx_int_218707107
#define __pyx_int_219129172 __pyx_mstate_global->__pyx_int_219129172
#define __pyx_int_220368800 __pyx_mstate_global->__pyx_int_220368800
#define __pyx_int_225261924 __pyx_mstate_global->__pyx_int_225261924
#define __pyx_int_235564716 __pyx_mstate_global->__pyx_int_235564716
#define __pyx_int_244856512 __pyx_mstate_global->__pyx_int_244856512
#define __pyx_int_245819509 __pyx_mstate_global->__pyx_int_245819509
#define __pyx_int_246409365 __pyx_mstate_global->__pyx_int_246409365
#define __pyx_int_253265777 __pyx_mstate_global->__pyx_int_253265777
#define __pyx_int_255344631 __pyx_mstate_global->__pyx_int_255344631
#define __pyx_int_257562521 __pyx_mstate_global->__pyx_int_257562521
#define __pyx_int_264852134 __pyx_mstate_global->__pyx_int_264852134
#define __pyx_int_266727818 __pyx_mstate_global->__pyx_int_266727818
#define __pyx_int_267485924 __pyx_mstate_global->__pyx_int_267485924
#define __pyx_int_4294967295 __pyx_mstate_global->__pyx_int_4294967295
#define __pyx_int_neg_2 __pyx_mstate_global->__pyx_int_neg_2
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_tuple__4 __pyx_mstate_global->__pyx_tuple__4
#define __pyx_tuple__6 __pyx_mstate_global->__pyx_tuple__6
#define __pyx_tuple__7 __pyx_mstate_global->__pyx_tuple__7
#define __pyx_tuple__8 __pyx_mstate_global->__pyx_tuple__8
#define __pyx_tuple__9 __pyx_mstate_global->__pyx_tuple__9
#define __pyx_slice__15 __pyx_mstate_global->__pyx_slice__15
#define __pyx_slice__16 __pyx_mstate_global->__pyx_slice__16
#define __pyx_slice__39 __pyx_mstate_global->__pyx_slice__39
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_tuple__13 __pyx_mstate_global->__pyx_tuple__13
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__21 __pyx_mstate_global->__pyx_tuple__21
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__23 __pyx_mstate_global->__pyx_tuple__23
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__25 __pyx_mstate_global->__pyx_tuple__25
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__27 __pyx_mstate_global->__pyx_tuple__27
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__29 __pyx_mstate_global->__pyx_tuple__29
#define __pyx_tuple__30 __pyx_mstate_global->__pyx_tuple__30
#define __pyx_tuple__31 __pyx_mstate_global->__pyx_tuple__31
#define __pyx_tuple__32 __pyx_mstate_global->__pyx_tuple__32
#define __pyx_tuple__33 __pyx_mstate_global->__pyx_tuple__33
#define __pyx_tuple__34 __pyx_mstate_global->__pyx_tuple__34
#define __pyx_tuple__35 __pyx_mstate_global->__pyx_tuple__35
#define __pyx_tuple__36 __pyx_mstate_global->__pyx_tuple__36
#define __pyx_tuple__37 __pyx_mstate_global->__pyx_tuple__37
#define __pyx_tuple__38 __pyx_mstate_global->__pyx_tuple__38
#define __pyx_tuple__40 __pyx_mstate_global->__pyx_tuple__40
#define __pyx_tuple__41 __pyx_mstate_global->__pyx_tuple__41
#define __pyx_tuple__44 __pyx_mstate_global->__pyx_tuple__44
#define __pyx_tuple__46 __pyx_mstate_global->__pyx_tuple__46
#define __pyx_tuple__47 __pyx_mstate_global->__pyx_tuple__47
#define __pyx_tuple__48 __pyx_mstate_global->__pyx_tuple__48
#define __pyx_tuple__49 __pyx_mstate_global->__pyx_tuple__49
#define __pyx_tuple__50 __pyx_mstate_global->__pyx_tuple__50
#define __pyx_tuple__51 __pyx_mstate_global->__pyx_tuple__51
#define __pyx_tuple__52 __pyx_mstate_global->__pyx_tuple__52
#define __pyx_tuple__53 __pyx_mstate_global->__pyx_tuple__53
#define __pyx_tuple__54 __pyx_mstate_global->__pyx_tuple__54
#define __pyx_tuple__55 __pyx_mstate_global->__pyx_tuple__55
#define __pyx_tuple__56 __pyx_mstate_global->__pyx_tuple__56
#define __pyx_tuple__57 __pyx_mstate_global->__pyx_tuple__57
#define __pyx_tuple__58 __pyx_mstate_global->__pyx_tuple__58
#define __pyx_tuple__59 __pyx_mstate_global->__pyx_tuple__59
#define __pyx_tuple__60 __pyx_mstate_global->__pyx_tuple__60
#define __pyx_tuple__61 __pyx_mstate_global->__pyx_tuple__61
#define __pyx_tuple__62 __pyx_mstate_global->__pyx_tuple__62
#define __pyx_tuple__63 __pyx_mstate_global->__pyx_tuple__63
#define __pyx_tuple__64 __pyx_mstate_global->__pyx_tuple__64
#define __pyx_tuple__66 __pyx_mstate_global->__pyx_tuple__66
#define __pyx_tuple__68 __pyx_mstate_global->__pyx_tuple__68
#define __pyx_tuple__78 __pyx_mstate_global->__pyx_tuple__78
#define __pyx_tuple__84 __pyx_mstate_global->__pyx_tuple__84
#define __pyx_tuple__124 __pyx_mstate_global->__pyx_tuple__124
#define __pyx_tuple__126 __pyx_mstate_global->__pyx_tuple__126
#define __pyx_tuple__128 __pyx_mstate_global->__pyx_tuple__128
#define __pyx_tuple__129 __pyx_mstate_global->__pyx_tuple__129
#define __pyx_tuple__131 __pyx_mstate_global->__pyx_tuple__131
#define __pyx_tuple__133 __pyx_mstate_global->__pyx_tuple__133
#define __pyx_tuple__135 __pyx_mstate_global->__pyx_tuple__135
#define __pyx_tuple__137 __pyx_mstate_global->__pyx_tuple__137
#define __pyx_tuple__138 __pyx_mstate_global->__pyx_tuple__138
#define __pyx_tuple__140 __pyx_mstate_global->__pyx_tuple__140
#define __pyx_tuple__141 __pyx_mstate_global->__pyx_tuple__141
#define __pyx_tuple__143 __pyx_mstate_global->__pyx_tuple__143
#define __pyx_tuple__145 __pyx_mstate_global->__pyx_tuple__145
#define __pyx_tuple__146 __pyx_mstate_global->__pyx_tuple__146
#define __pyx_tuple__148 __pyx_mstate_global->__pyx_tuple__148
#define __pyx_tuple__150 __pyx_mstate_global->__pyx_tuple__150
#define __pyx_tuple__152 __pyx_mstate_global->__pyx_tuple__152
#define __pyx_tuple__153 __pyx_mstate_global->__pyx_tuple__153
#define __pyx_tuple__155 __pyx_mstate_global->__pyx_tuple__155
#define __pyx_tuple__157 __pyx_mstate_global->__pyx_tuple__157
#define __pyx_tuple__164 __pyx_mstate_global->__pyx_tuple__164
#define __pyx_tuple__168 __pyx_mstate_global->__pyx_tuple__168
#define __pyx_tuple__170 __pyx_mstate_global->__pyx_tuple__170
#define __pyx_tuple__172 __pyx_mstate_global->__pyx_tuple__172
#define __pyx_tuple__174 __pyx_mstate_global->__pyx_tuple__174
#define __pyx_tuple__179 __pyx_mstate_global->__pyx_tuple__179
#define __pyx_tuple__181 __pyx_mstate_global->__pyx_tuple__181
#define __pyx_tuple__183 __pyx_mstate_global->__pyx_tuple__183
#define __pyx_tuple__185 __pyx_mstate_global->__pyx_tuple__185
#define __pyx_tuple__190 __pyx_mstate_global->__pyx_tuple__190
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__43 __pyx_mstate_global->__pyx_codeobj__43
#define __pyx_codeobj__45 __pyx_mstate_global->__pyx_codeobj__45
#define __pyx_codeobj__67 __pyx_mstate_global->__pyx_codeobj__67
#define __pyx_codeobj__69 __pyx_mstate_global->__pyx_codeobj__69
#define __pyx_codeobj__70 __pyx_mstate_global->__pyx_codeobj__70
#define __pyx_codeobj__71 __pyx_mstate_global->__pyx_codeobj__71
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
#define __pyx_codeobj__73 __pyx_mstate_global->__pyx_codeobj__73
#define __pyx_codeobj__74 __pyx_mstate_global->__pyx_codeobj__74
#define __pyx_codeobj__75 __pyx_mstate_global->__pyx_codeobj__75
#define __pyx_codeobj__76 __pyx_mstate_global->__pyx_codeobj__76
#define __pyx_codeobj__77 __pyx_mstate_global->__pyx_codeobj__77
#define __pyx_codeobj__79 __pyx_mstate_global->__pyx_codeobj__79
#define __pyx_codeobj__80 __pyx_mstate_global->__pyx_codeobj__80
#define __pyx_codeobj__81 __pyx_mstate_global->__pyx_codeobj__81
#define __pyx_codeobj__82 __pyx_mstate_global->__pyx_codeobj__82
#define __pyx_codeobj__83 __pyx_mstate_global->__pyx_codeobj__83
#define __pyx_codeobj__85 __pyx_mstate_global->__pyx_codeobj__85
#define __pyx_codeobj__86 __pyx_mstate_global->__pyx_codeobj__86
#define __pyx_codeobj__87 __pyx_mstate_global->__pyx_codeobj__87
#define __pyx_codeobj__88 __pyx_mstate_global->__pyx_codeobj__88
#define __pyx_codeobj__89 __pyx_mstate_global->__pyx_codeobj__89
#define __pyx_codeobj__90 __pyx_mstate_global->__pyx_codeobj__90
#define __pyx_codeobj__91 __pyx_mstate_global->__pyx_codeobj__91
#define __pyx_codeobj__92 __pyx_mstate_global->__pyx_codeobj__92
#define __pyx_codeobj__93 __pyx_mstate_global->__pyx_codeobj__93
#define __pyx_codeobj__94 __pyx_mstate_global->__pyx_codeobj__94
#define __pyx_codeobj__95 __pyx_mstate_global->__pyx_codeobj__95
#define __pyx_codeobj__96 __pyx_mstate_global->__pyx_codeobj__96
#define __pyx_codeobj__97 __pyx_mstate_global->__pyx_codeobj__97
#define __pyx_codeobj__98 __pyx_mstate_global->__pyx_codeobj__98
#define __pyx_codeobj__99 __pyx_mstate_global->__pyx_codeobj__99
#define __pyx_codeobj__100 __pyx_mstate_global->__pyx_codeobj__100
#define __pyx_codeobj__101 __pyx_mstate_global->__pyx_codeobj__101
#define __pyx_codeobj__102 __pyx_mstate_global->__pyx_codeobj__102
#define __pyx_codeobj__103 __pyx_mstate_global->__pyx_codeobj__103
#define __pyx_codeobj__104 __pyx_mstate_global->__pyx_codeobj__104
#define __pyx_codeobj__105 __pyx_mstate_global->__pyx_codeobj__105
#define __pyx_codeobj__106 __pyx_mstate_global->__pyx_codeobj__106
#define __pyx_codeobj__107 __pyx_mstate_global->__pyx_codeobj__107
#define __pyx_codeobj__108 __pyx_mstate_global->__pyx_codeobj__108
#define __pyx_codeobj__109 __pyx_mstate_global->__pyx_codeobj__109
#define __pyx_codeobj__110 __pyx_mstate_global->__pyx_codeobj__110
#define __pyx_codeobj__111 __pyx_mstate_global->__pyx_codeobj__111
#define __pyx_codeobj__112 __pyx_mstate_global->__pyx_codeobj__112
#define __pyx_codeobj__113 __pyx_mstate_global->__pyx_codeobj__113
#define __pyx_codeobj__114 __pyx_mstate_global->__pyx_codeobj__114
#define __pyx_codeobj__115 __pyx_mstate_global->__pyx_codeobj__115
#define __pyx_codeobj__116 __pyx_mstate_global->__pyx_codeobj__116
#define __pyx_codeobj__117 __pyx_mstate_global->__pyx_codeobj__117
#define __pyx_codeobj__118 __pyx_mstate_global->__pyx_codeobj__118
#define __pyx_codeobj__119 __pyx_mstate_global->__pyx_codeobj__119
#define __pyx_codeobj__120 __pyx_mstate_global->__pyx_codeobj__120
#define __pyx_codeobj__121 __pyx_mstate_global->__pyx_codeobj__121
#define __pyx_codeobj__122 __pyx_mstate_global->__pyx_codeobj__122
#define __pyx_codeobj__123 __pyx_mstate_global->__pyx_codeobj__123
#define __pyx_codeobj__125 __pyx_mstate_global->__pyx_codeobj__125
#define __pyx_codeobj__127 __pyx_mstate_global->__pyx_codeobj__127
#define __pyx_codeobj__130 __pyx_mstate_global->__pyx_codeobj__130
#define __pyx_codeobj__132 __pyx_mstate_global->__pyx_codeobj__132
#define __pyx_codeobj__134 __pyx_mstate_global->__pyx_codeobj__134
#define __pyx_codeobj__136 __pyx_mstate_global->__pyx_codeobj__136
#define __pyx_codeobj__139 __pyx_mstate_global->__pyx_codeobj__139
#define __pyx_codeobj__142 __pyx_mstate_global->__pyx_codeobj__142
#define __pyx_codeobj__144 __pyx_mstate_global->__pyx_codeobj__144
#define __pyx_codeobj__147 __pyx_mstate_global->__pyx_codeobj__147
#define __pyx_codeobj__149 __pyx_mstate_global->__pyx_codeobj__149
#define __pyx_codeobj__151 __pyx_mstate_global->__pyx_codeobj__151
#define __pyx_codeobj__154 __pyx_mstate_global->__pyx_codeobj__154
#define __pyx_codeobj__156 __pyx_mstate_global->__pyx_codeobj__156
#define __pyx_codeobj__158 __pyx_mstate_global->__pyx_codeobj__158
#define __pyx_codeobj__159 __pyx_mstate_global->__pyx_codeobj__159
#define __pyx_codeobj__160 __pyx_mstate_global->__pyx_codeobj__160
#define __pyx_codeobj__161 __pyx_mstate_global->__pyx_codeobj__161
#define __pyx_codeobj__162 __pyx_mstate_global->__pyx_codeobj__162
#define __pyx_codeobj__163 __pyx_mstate_global->__pyx_codeobj__163
#define __pyx_codeobj__165 __pyx_mstate_global->__pyx_codeobj__165
#define __pyx_codeobj__166 __pyx_mstate_global->__pyx_codeobj__166
#define __pyx_codeobj__167 __pyx_mstate_global->__pyx_codeobj__167
#define __pyx_codeobj__169 __pyx_mstate_global->__pyx_codeobj__169
#define __pyx_codeobj__171 __pyx_mstate_global->__pyx_codeobj__171
#define __pyx_codeobj__173 __pyx_mstate_global->__pyx_codeobj__173
#define __pyx_codeobj__175 __pyx_mstate_global->__pyx_codeobj__175
#define __pyx_codeobj__176 __pyx_mstate_global->__pyx_codeobj__176
#define __pyx_codeobj__177 __pyx_mstate_global->__pyx_codeobj__177
#define __pyx_codeobj__178 __pyx_mstate_global->__pyx_codeobj__178
#define __pyx_codeobj__180 __pyx_mstate_global->__pyx_codeobj__180
#define __pyx_codeobj__182 __pyx_mstate_global->__pyx_codeobj__182
#define __pyx_codeobj__184 __pyx_mstate_global->__pyx_codeobj__184
#define __pyx_codeobj__186 __pyx_mstate_global->__pyx_codeobj__186
#define __pyx_codeobj__187 __pyx_mstate_global->__pyx_codeobj__187
#define __pyx_codeobj__188 __pyx_mstate_global->__pyx_codeobj__188
#define __pyx_codeobj__189 __pyx_mstate_global->__pyx_codeobj__189
#define __pyx_codeobj__191 __pyx_mstate_global->__pyx_codeobj__191
#define __pyx_codeobj__192 __pyx_mstate_global->__pyx_codeobj__192
#define __pyx_codeobj__193 __pyx_mstate_global->__pyx_codeobj__193
#define __pyx_codeobj__194 __pyx_mstate_global->__pyx_codeobj__194
#define __pyx_codeobj__195 __pyx_mstate_global->__pyx_codeobj__195
#define __pyx_codeobj__196 __pyx_mstate_global->__pyx_codeobj__196
#define __pyx_codeobj__197 __pyx_mstate_global->__pyx_codeobj__197
#define __pyx_codeobj__198 __pyx_mstate_global->__pyx_codeobj__198
#define __pyx_codeobj__199 __pyx_mstate_global->__pyx_codeobj__199
#define __pyx_codeobj__200 __pyx_mstate_global->__pyx_codeobj__200
#define __pyx_codeobj__201 __pyx_mstate_global->__pyx_codeobj__201
#define __pyx_codeobj__202 __pyx_mstate_global->__pyx_codeobj__202
#define __pyx_codeobj__203 __pyx_mstate_global->__pyx_codeobj__203
#define __pyx_codeobj__204 __pyx_mstate_global->__pyx_codeobj__204
#define __pyx_codeobj__205 __pyx_mstate_global->__pyx_codeobj__205
#define __pyx_codeobj__206 __pyx_mstate_global->__pyx_codeobj__206
#define __pyx_codeobj__207 __pyx_mstate_global->__pyx_codeobj__207
#define __pyx_codeobj__208 __pyx_mstate_global->__pyx_codeobj__208
#define __pyx_codeobj__209 __pyx_mstate_global->__pyx_codeobj__209
#define __pyx_codeobj__210 __pyx_mstate_global->__pyx_codeobj__210
/* #### Code section: module_code ### */

/* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self):             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":20
 *         @property
 *         cdef inline double real(self):
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = __pyx_v_self->cval.real;
  goto __pyx_L0;

  /* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self):             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self):             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":24
 *         @property
 *         cdef inline double imag(self):
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
 */
  __pyx_r = __pyx_v_self->cval.imag;
  goto __pyx_L0;

  /* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self):             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":117
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":118
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(7, 118, __pyx_L1_error)

  /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
  __pyx_t_2 = (__pyx_v_value == NULL);
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":121
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
 */
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
    goto __pyx_L3;
  }

  /* "cpython/contextvars.pxd":124
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_pyvalue = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":125
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
 */
    Py_XDECREF(__pyx_v_value);
  }
  __pyx_L3:;

  /* "cpython/contextvars.pxd":126
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_no_default", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":135
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":136
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(7, 136, __pyx_L1_error)

  /* "cpython/contextvars.pxd":138
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_value);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pyvalue = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":139
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 */
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":140
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":72
 *     ctypedef extern class datetime.date[object PyDateTime_Date]:
 *         @property
 *         cdef inline int year(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_4year_year(PyDateTime_Date *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":73
 *         @property
 *         cdef inline int year(self):
 *             return PyDateTime_GET_YEAR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_YEAR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":72
 *     ctypedef extern class datetime.date[object PyDateTime_Date]:
 *         @property
 *         cdef inline int year(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":76
 * 
 *         @property
 *         cdef inline int month(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_5month_month(PyDateTime_Date *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":77
 *         @property
 *         cdef inline int month(self):
 *             return PyDateTime_GET_MONTH(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_MONTH(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":76
 * 
 *         @property
 *         cdef inline int month(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":80
 * 
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_3day_day(PyDateTime_Date *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":81
 *         @property
 *         cdef inline int day(self):
 *             return PyDateTime_GET_DAY(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
 */
  __pyx_r = PyDateTime_GET_DAY(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":80
 * 
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":85
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
 *         @property
 *         cdef inline int hour(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_HOUR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4hour_hour(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":86
 *         @property
 *         cdef inline int hour(self):
 *             return PyDateTime_TIME_GET_HOUR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_HOUR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":85
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
 *         @property
 *         cdef inline int hour(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_HOUR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":89
 * 
 *         @property
 *         cdef inline int minute(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MINUTE(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6minute_minute(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":90
 *         @property
 *         cdef inline int minute(self):
 *             return PyDateTime_TIME_GET_MINUTE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_MINUTE(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":89
 * 
 *         @property
 *         cdef inline int minute(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MINUTE(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":93
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_SECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6second_second(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":94
 *         @property
 *         cdef inline int second(self):
 *             return PyDateTime_TIME_GET_SECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_SECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":93
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_SECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":97
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MICROSECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_11microsecond_microsecond(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":98
 *         @property
 *         cdef inline int microsecond(self):
 *             return PyDateTime_TIME_GET_MICROSECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_MICROSECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":97
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MICROSECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":101
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_4time_6tzinfo_tzinfo(PyDateTime_Time *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("tzinfo", 1);

  /* "cpython/datetime.pxd":102
 *         @property
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_TIME_GET_TZINFO(((PyObject *)__pyx_v_self));
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":101
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":105
 * 
 *         @property
 *         cdef inline int fold(self):             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_TIME_GET_FOLD(self)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4fold_fold(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":107
 *         cdef inline int fold(self):
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_TIME_GET_FOLD(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
 */
  __pyx_r = PyDateTime_TIME_GET_FOLD(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":105
 * 
 *         @property
 *         cdef inline int fold(self):             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_TIME_GET_FOLD(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":111
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
 *         @property
 *         cdef inline int year(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4year_year(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":112
 *         @property
 *         cdef inline int year(self):
 *             return PyDateTime_GET_YEAR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_YEAR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":111
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
 *         @property
 *         cdef inline int year(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":115
 * 
 *         @property
 *         cdef inline int month(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_5month_month(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":116
 *         @property
 *         cdef inline int month(self):
 *             return PyDateTime_GET_MONTH(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_MONTH(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":115
 * 
 *         @property
 *         cdef inline int month(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":119
 * 
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_3day_day(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":120
 *         @property
 *         cdef inline int day(self):
 *             return PyDateTime_GET_DAY(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_DAY(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":119
 * 
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":123
 * 
 *         @property
 *         cdef inline int hour(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_HOUR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4hour_hour(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":124
 *         @property
 *         cdef inline int hour(self):
 *             return PyDateTime_DATE_GET_HOUR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_HOUR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":123
 * 
 *         @property
 *         cdef inline int hour(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_HOUR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":127
 * 
 *         @property
 *         cdef inline int minute(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MINUTE(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6minute_minute(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":128
 *         @property
 *         cdef inline int minute(self):
 *             return PyDateTime_DATE_GET_MINUTE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_MINUTE(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":127
 * 
 *         @property
 *         cdef inline int minute(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MINUTE(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":131
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_SECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6second_second(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":132
 *         @property
 *         cdef inline int second(self):
 *             return PyDateTime_DATE_GET_SECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_SECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":131
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_SECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":135
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MICROSECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_11microsecond_microsecond(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":136
 *         @property
 *         cdef inline int microsecond(self):
 *             return PyDateTime_DATE_GET_MICROSECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_MICROSECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":135
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MICROSECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":139
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_8datetime_6tzinfo_tzinfo(PyDateTime_DateTime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("tzinfo", 1);

  /* "cpython/datetime.pxd":140
 *         @property
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_DATE_GET_TZINFO(((PyObject *)__pyx_v_self));
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":139
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":143
 * 
 *         @property
 *         cdef inline int fold(self):             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_DATE_GET_FOLD(self)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4fold_fold(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":145
 *         cdef inline int fold(self):
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_DATE_GET_FOLD(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
 */
  __pyx_r = PyDateTime_DATE_GET_FOLD(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":143
 * 
 *         @property
 *         cdef inline int fold(self):             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_DATE_GET_FOLD(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":149
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_DAYS(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_3day_day(PyDateTime_Delta *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":150
 *         @property
 *         cdef inline int day(self):
 *             return PyDateTime_DELTA_GET_DAYS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DELTA_GET_DAYS(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":149
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_DAYS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":153
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_SECONDS(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_6second_second(PyDateTime_Delta *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":154
 *         @property
 *         cdef inline int second(self):
 *             return PyDateTime_DELTA_GET_SECONDS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DELTA_GET_SECONDS(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":153
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_SECONDS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":157
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_11microsecond_microsecond(PyDateTime_Delta *__pyx_v_self) {
  int __pyx_r;

  /* "cpython/datetime.pxd":158
 *         @property
 *         cdef inline int microsecond(self):
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.tzinfo[object PyDateTime_TZInfo]:
 */
  __pyx_r = PyDateTime_DELTA_GET_MICROSECONDS(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":157
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":282
 * # Datetime C API initialization function.
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime():             # <<<<<<<<<<<<<<
 *     PyDateTime_IMPORT
 * 
 */

static CYTHON_INLINE void __pyx_f_7cpython_8datetime_import_datetime(void) {

  /* "cpython/datetime.pxd":283
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime():
 *     PyDateTime_IMPORT             # <<<<<<<<<<<<<<
 * 
 * # Create date object using DateTime CAPI factory function.
 */
  (void)(PyDateTime_IMPORT);

  /* "cpython/datetime.pxd":282
 * # Datetime C API initialization function.
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime():             # <<<<<<<<<<<<<<
 *     PyDateTime_IMPORT
 * 
 */

  /* function exit code */
}

/* "cpython/datetime.pxd":287
 * # Create date object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)
 * 
 */

static CYTHON_INLINE PyDateTime_Date *__pyx_f_7cpython_8datetime_date_new(int __pyx_v_year, int __pyx_v_month, int __pyx_v_day) {
  PyDateTime_Date *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("date_new", 1);

  /* "cpython/datetime.pxd":288
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)             # <<<<<<<<<<<<<<
 * 
 * # Create time object using DateTime CAPI factory function
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)PyDateTimeAPI->Date_FromDate(__pyx_v_year, __pyx_v_month, __pyx_v_day, PyDateTimeAPI->DateType)); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyDateTime_Date *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":287
 * # Create date object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.date_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":292
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
 */

static CYTHON_INLINE PyDateTime_Time *__pyx_f_7cpython_8datetime_time_new(int __pyx_v_hour, int __pyx_v_minute, int __pyx_v_second, int __pyx_v_microsecond, PyObject *__pyx_v_tz, struct __pyx_opt_args_7cpython_8datetime_time_new *__pyx_optional_args) {
  int __pyx_v_fold = ((int)0);
  PyDateTime_Time *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("time_new", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_fold = __pyx_optional_args->fold;
    }
  }

  /* "cpython/datetime.pxd":293
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             # <<<<<<<<<<<<<<
 * 
 * # Create datetime object using DateTime CAPI factory function.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__Pyx_DateTime_TimeWithFold(__pyx_v_hour, __pyx_v_minute, __pyx_v_second, __pyx_v_microsecond, __pyx_v_tz, __pyx_v_fold)); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7cpython_8datetime_time))))) __PYX_ERR(8, 293, __pyx_L1_error)
  __pyx_r = ((PyDateTime_Time *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":292
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.time_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":297
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
 */

static CYTHON_INLINE PyDateTime_DateTime *__pyx_f_7cpython_8datetime_datetime_new(int __pyx_v_year, int __pyx_v_month, int __pyx_v_day, int __pyx_v_hour, int __pyx_v_minute, int __pyx_v_second, int __pyx_v_microsecond, PyObject *__pyx_v_tz, struct __pyx_opt_args_7cpython_8datetime_datetime_new *__pyx_optional_args) {
  int __pyx_v_fold = ((int)0);
  PyDateTime_DateTime *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("datetime_new", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_fold = __pyx_optional_args->fold;
    }
  }

  /* "cpython/datetime.pxd":298
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             # <<<<<<<<<<<<<<
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__Pyx_DateTime_DateTimeWithFold(__pyx_v_year, __pyx_v_month, __pyx_v_day, __pyx_v_hour, __pyx_v_minute, __pyx_v_second, __pyx_v_microsecond, __pyx_v_tz, __pyx_v_fold)); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyDateTime_DateTime *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":297
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.datetime_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":302
 * # Create timedelta object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)
 * 
 */

static CYTHON_INLINE PyDateTime_Delta *__pyx_f_7cpython_8datetime_timedelta_new(int __pyx_v_days, int __pyx_v_seconds, int __pyx_v_useconds) {
  PyDateTime_Delta *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("timedelta_new", 1);

  /* "cpython/datetime.pxd":303
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)             # <<<<<<<<<<<<<<
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)PyDateTimeAPI->Delta_FromDelta(__pyx_v_days, __pyx_v_seconds, __pyx_v_useconds, 1, PyDateTimeAPI->DeltaType)); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyDateTime_Delta *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":302
 * # Create timedelta object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.timedelta_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":306
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_timezone_new(PyObject *__pyx_v_offset, struct __pyx_opt_args_7cpython_8datetime_timezone_new *__pyx_optional_args) {
  PyObject *__pyx_v_name = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("timezone_new", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_name = __pyx_optional_args->name;
    }
  }

  /* "cpython/datetime.pxd":307
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 */
  __pyx_t_1 = (PY_VERSION_HEX < 0x030700b1);
  if (unlikely(__pyx_t_1)) {

    /* "cpython/datetime.pxd":308
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')             # <<<<<<<<<<<<<<
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(8, 308, __pyx_L1_error)

    /* "cpython/datetime.pxd":307
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 */
  }

  /* "cpython/datetime.pxd":309
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)             # <<<<<<<<<<<<<<
 * 
 * # Create datetime object using DB API constructor.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = (__pyx_v_name != Py_None);
  if (__pyx_t_1) {
    __pyx_t_3 = ((PyObject *)__pyx_v_name);
  } else {
    __pyx_t_3 = NULL;
  }
  __pyx_t_2 = __Pyx_TimeZone_FromOffsetAndName(__pyx_v_offset, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":306
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.datetime.timezone_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":312
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 */

static CYTHON_INLINE PyDateTime_DateTime *__pyx_f_7cpython_8datetime_datetime_from_timestamp(PyObject *__pyx_v_timestamp, struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp *__pyx_optional_args) {
  PyObject *__pyx_v_tz = ((PyObject *)Py_None);
  PyDateTime_DateTime *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("datetime_from_timestamp", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_tz = __pyx_optional_args->tz;
    }
  }

  /* "cpython/datetime.pxd":313
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(             # <<<<<<<<<<<<<<
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);

  /* "cpython/datetime.pxd":314
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)             # <<<<<<<<<<<<<<
 * 
 * # Create date object using DB API constructor.
 */
  __pyx_t_2 = (__pyx_v_tz != Py_None);
  if (__pyx_t_2) {
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_timestamp);
    __Pyx_GIVEREF(__pyx_v_timestamp);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_timestamp)) __PYX_ERR(8, 314, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_tz);
    __Pyx_GIVEREF(__pyx_v_tz);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_tz)) __PYX_ERR(8, 314, __pyx_L1_error);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_timestamp);
    __Pyx_GIVEREF(__pyx_v_timestamp);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_timestamp)) __PYX_ERR(8, 314, __pyx_L1_error);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }

  /* "cpython/datetime.pxd":313
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(             # <<<<<<<<<<<<<<
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 * 
 */
  __pyx_t_3 = ((PyObject *)PyDateTimeAPI->DateTime_FromTimestamp(((PyObject *)PyDateTimeAPI->DateTimeType), __pyx_t_1, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyDateTime_DateTime *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":312
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.datetime.datetime_from_timestamp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":317
 * 
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))
 * 
 */

static CYTHON_INLINE PyDateTime_Date *__pyx_f_7cpython_8datetime_date_from_timestamp(PyObject *__pyx_v_timestamp) {
  PyDateTime_Date *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("date_from_timestamp", 1);

  /* "cpython/datetime.pxd":318
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))             # <<<<<<<<<<<<<<
 * 
 * # More recognizable getters for date/time/datetime/timedelta.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_timestamp);
  __Pyx_GIVEREF(__pyx_v_timestamp);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_timestamp)) __PYX_ERR(8, 318, __pyx_L1_error);
  __pyx_t_2 = ((PyObject *)PyDateTimeAPI->Date_FromTimestamp(((PyObject *)PyDateTimeAPI->DateType), __pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyDateTime_Date *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":317
 * 
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.datetime.date_from_timestamp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":326
 * 
 * # Get UTC singleton
 * cdef inline object get_utc():             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_get_utc(void) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_utc", 1);

  /* "cpython/datetime.pxd":327
 * # Get UTC singleton
 * cdef inline object get_utc():
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return <object>__Pyx_TimeZone_UTC
 */
  __pyx_t_1 = (PY_VERSION_HEX < 0x030700b1);
  if (unlikely(__pyx_t_1)) {

    /* "cpython/datetime.pxd":328
 * cdef inline object get_utc():
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')             # <<<<<<<<<<<<<<
 *     return <object>__Pyx_TimeZone_UTC
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(8, 328, __pyx_L1_error)

    /* "cpython/datetime.pxd":327
 * # Get UTC singleton
 * cdef inline object get_utc():
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return <object>__Pyx_TimeZone_UTC
 */
  }

  /* "cpython/datetime.pxd":329
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return <object>__Pyx_TimeZone_UTC             # <<<<<<<<<<<<<<
 * 
 * # Get tzinfo of time
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__Pyx_TimeZone_UTC));
  __pyx_r = ((PyObject *)__Pyx_TimeZone_UTC);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":326
 * 
 * # Get UTC singleton
 * cdef inline object get_utc():             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.datetime.get_utc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":332
 * 
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_time_tzinfo(PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("time_tzinfo", 1);

  /* "cpython/datetime.pxd":333
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)             # <<<<<<<<<<<<<<
 * 
 * # Get tzinfo of datetime
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_TIME_GET_TZINFO(__pyx_v_o);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":332
 * 
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":336
 * 
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_datetime_tzinfo(PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("datetime_tzinfo", 1);

  /* "cpython/datetime.pxd":337
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)             # <<<<<<<<<<<<<<
 * 
 * # Get year of date
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_DATE_GET_TZINFO(__pyx_v_o);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":336
 * 
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":340
 * 
 * # Get year of date
 * cdef inline int date_year(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_year(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":341
 * # Get year of date
 * cdef inline int date_year(object o):
 *     return PyDateTime_GET_YEAR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get month of date
 */
  __pyx_r = PyDateTime_GET_YEAR(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":340
 * 
 * # Get year of date
 * cdef inline int date_year(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":344
 * 
 * # Get month of date
 * cdef inline int date_month(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_month(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":345
 * # Get month of date
 * cdef inline int date_month(object o):
 *     return PyDateTime_GET_MONTH(o)             # <<<<<<<<<<<<<<
 * 
 * # Get day of date
 */
  __pyx_r = PyDateTime_GET_MONTH(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":344
 * 
 * # Get month of date
 * cdef inline int date_month(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":348
 * 
 * # Get day of date
 * cdef inline int date_day(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_day(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":349
 * # Get day of date
 * cdef inline int date_day(object o):
 *     return PyDateTime_GET_DAY(o)             # <<<<<<<<<<<<<<
 * 
 * # Get year of datetime
 */
  __pyx_r = PyDateTime_GET_DAY(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":348
 * 
 * # Get day of date
 * cdef inline int date_day(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":352
 * 
 * # Get year of datetime
 * cdef inline int datetime_year(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_year(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":353
 * # Get year of datetime
 * cdef inline int datetime_year(object o):
 *     return PyDateTime_GET_YEAR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get month of datetime
 */
  __pyx_r = PyDateTime_GET_YEAR(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":352
 * 
 * # Get year of datetime
 * cdef inline int datetime_year(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":356
 * 
 * # Get month of datetime
 * cdef inline int datetime_month(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_month(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":357
 * # Get month of datetime
 * cdef inline int datetime_month(object o):
 *     return PyDateTime_GET_MONTH(o)             # <<<<<<<<<<<<<<
 * 
 * # Get day of datetime
 */
  __pyx_r = PyDateTime_GET_MONTH(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":356
 * 
 * # Get month of datetime
 * cdef inline int datetime_month(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":360
 * 
 * # Get day of datetime
 * cdef inline int datetime_day(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_day(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":361
 * # Get day of datetime
 * cdef inline int datetime_day(object o):
 *     return PyDateTime_GET_DAY(o)             # <<<<<<<<<<<<<<
 * 
 * # Get hour of time
 */
  __pyx_r = PyDateTime_GET_DAY(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":360
 * 
 * # Get day of datetime
 * cdef inline int datetime_day(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":364
 * 
 * # Get hour of time
 * cdef inline int time_hour(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_HOUR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_hour(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":365
 * # Get hour of time
 * cdef inline int time_hour(object o):
 *     return PyDateTime_TIME_GET_HOUR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get minute of time
 */
  __pyx_r = PyDateTime_TIME_GET_HOUR(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":364
 * 
 * # Get hour of time
 * cdef inline int time_hour(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_HOUR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":368
 * 
 * # Get minute of time
 * cdef inline int time_minute(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MINUTE(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_minute(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":369
 * # Get minute of time
 * cdef inline int time_minute(object o):
 *     return PyDateTime_TIME_GET_MINUTE(o)             # <<<<<<<<<<<<<<
 * 
 * # Get second of time
 */
  __pyx_r = PyDateTime_TIME_GET_MINUTE(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":368
 * 
 * # Get minute of time
 * cdef inline int time_minute(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MINUTE(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":372
 * 
 * # Get second of time
 * cdef inline int time_second(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_SECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_second(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":373
 * # Get second of time
 * cdef inline int time_second(object o):
 *     return PyDateTime_TIME_GET_SECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get microsecond of time
 */
  __pyx_r = PyDateTime_TIME_GET_SECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":372
 * 
 * # Get second of time
 * cdef inline int time_second(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_SECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":376
 * 
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MICROSECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_microsecond(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":377
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o):
 *     return PyDateTime_TIME_GET_MICROSECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get fold of time
 */
  __pyx_r = PyDateTime_TIME_GET_MICROSECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":376
 * 
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MICROSECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":380
 * 
 * # Get fold of time
 * cdef inline int time_fold(object o):             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_fold(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":382
 * cdef inline int time_fold(object o):
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)             # <<<<<<<<<<<<<<
 * 
 * # Get hour of datetime
 */
  __pyx_r = PyDateTime_TIME_GET_FOLD(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":380
 * 
 * # Get fold of time
 * cdef inline int time_fold(object o):             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":385
 * 
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_HOUR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_hour(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":386
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o):
 *     return PyDateTime_DATE_GET_HOUR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get minute of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_HOUR(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":385
 * 
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_HOUR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":389
 * 
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MINUTE(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_minute(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":390
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o):
 *     return PyDateTime_DATE_GET_MINUTE(o)             # <<<<<<<<<<<<<<
 * 
 * # Get second of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_MINUTE(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":389
 * 
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MINUTE(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":393
 * 
 * # Get second of datetime
 * cdef inline int datetime_second(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_SECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_second(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":394
 * # Get second of datetime
 * cdef inline int datetime_second(object o):
 *     return PyDateTime_DATE_GET_SECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get microsecond of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_SECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":393
 * 
 * # Get second of datetime
 * cdef inline int datetime_second(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_SECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":397
 * 
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MICROSECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_microsecond(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":398
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o):
 *     return PyDateTime_DATE_GET_MICROSECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get fold of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_MICROSECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":397
 * 
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MICROSECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":401
 * 
 * # Get fold of datetime
 * cdef inline int datetime_fold(object o):             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_fold(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":403
 * cdef inline int datetime_fold(object o):
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)             # <<<<<<<<<<<<<<
 * 
 * # Get days of timedelta
 */
  __pyx_r = PyDateTime_DATE_GET_FOLD(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":401
 * 
 * # Get fold of datetime
 * cdef inline int datetime_fold(object o):             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":406
 * 
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).days
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_days(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":407
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o):
 *     return (<PyDateTime_Delta*>o).days             # <<<<<<<<<<<<<<
 * 
 * # Get seconds of timedelta
 */
  __pyx_r = ((PyDateTime_Delta *)__pyx_v_o)->days;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":406
 * 
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).days
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":410
 * 
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).seconds
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_seconds(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":411
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o):
 *     return (<PyDateTime_Delta*>o).seconds             # <<<<<<<<<<<<<<
 * 
 * # Get microseconds of timedelta
 */
  __pyx_r = ((PyDateTime_Delta *)__pyx_v_o)->seconds;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":410
 * 
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).seconds
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":414
 * 
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_microseconds(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "cpython/datetime.pxd":415
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o):
 *     return (<PyDateTime_Delta*>o).microseconds             # <<<<<<<<<<<<<<
 * 
 * cdef inline double total_seconds(timedelta obj):
 */
  __pyx_r = ((PyDateTime_Delta *)__pyx_v_o)->microseconds;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":414
 * 
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/datetime.pxd":417
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 * cdef inline double total_seconds(timedelta obj):             # <<<<<<<<<<<<<<
 *     # Mirrors the "timedelta.total_seconds()" method.
 *     # Note that this implementation is not guaranteed to give *exactly* the same
 */

static CYTHON_INLINE double __pyx_f_7cpython_8datetime_total_seconds(PyDateTime_Delta *__pyx_v_obj) {
  double __pyx_v_days;
  double __pyx_v_seconds;
  double __pyx_v_micros;
  double __pyx_r;

  /* "cpython/datetime.pxd":423
 *     cdef:
 *         double days, seconds, micros
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)             # <<<<<<<<<<<<<<
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
 */
  __pyx_v_days = ((double)PyDateTime_DELTA_GET_DAYS(((PyObject *)__pyx_v_obj)));

  /* "cpython/datetime.pxd":424
 *         double days, seconds, micros
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)             # <<<<<<<<<<<<<<
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
 *     return days * 24 * 3600 + seconds + micros / 1_000_000
 */
  __pyx_v_seconds = ((double)PyDateTime_DELTA_GET_SECONDS(((PyObject *)__pyx_v_obj)));

  /* "cpython/datetime.pxd":425
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)             # <<<<<<<<<<<<<<
 *     return days * 24 * 3600 + seconds + micros / 1_000_000
 */
  __pyx_v_micros = ((double)PyDateTime_DELTA_GET_MICROSECONDS(((PyObject *)__pyx_v_obj)));

  /* "cpython/datetime.pxd":426
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
 *     return days * 24 * 3600 + seconds + micros / 1_000_000             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((((__pyx_v_days * 24.0) * 3600.0) + __pyx_v_seconds) + (__pyx_v_micros / 1000000.0));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":417
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 * cdef inline double total_seconds(timedelta obj):             # <<<<<<<<<<<<<<
 *     # Mirrors the "timedelta.total_seconds()" method.
 *     # Note that this implementation is not guaranteed to give *exactly* the same
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/python.pxd":73
 * 
 * 
 * cdef inline object timezone_new(object offset):             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         from datetime import timezone
 */

static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_timezone_new(PyObject *__pyx_v_offset) {
  PyObject *__pyx_v_timezone = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("timezone_new", 1);

  /* "asynctnt/iproto/python.pxd":74
 * 
 * cdef inline object timezone_new(object offset):
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         from datetime import timezone
 *         return timezone(offset)
 */
  __pyx_t_1 = (PY_VERSION_HEX < 0x030700b1);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/python.pxd":75
 * cdef inline object timezone_new(object offset):
 *     if PY_VERSION_HEX < 0x030700b1:
 *         from datetime import timezone             # <<<<<<<<<<<<<<
 *         return timezone(offset)
 *     return __Pyx_TimeZone_FromOffset(offset)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_timezone);
    __Pyx_GIVEREF(__pyx_n_s_timezone);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_timezone)) __PYX_ERR(9, 75, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_datetime, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_timezone); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_timezone = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "asynctnt/iproto/python.pxd":76
 *     if PY_VERSION_HEX < 0x030700b1:
 *         from datetime import timezone
 *         return timezone(offset)             # <<<<<<<<<<<<<<
 *     return __Pyx_TimeZone_FromOffset(offset)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_timezone);
    __pyx_t_2 = __pyx_v_timezone; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_offset};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/python.pxd":74
 * 
 * cdef inline object timezone_new(object offset):
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         from datetime import timezone
 *         return timezone(offset)
 */
  }

  /* "asynctnt/iproto/python.pxd":77
 *         from datetime import timezone
 *         return timezone(offset)
 *     return __Pyx_TimeZone_FromOffset(offset)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_TimeZone_FromOffset(__pyx_v_offset); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/python.pxd":73
 * 
 * 
 * cdef inline object timezone_new(object offset):             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         from datetime import timezone
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.timezone_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_timezone);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/bit.pxd":5
 * 
 * 
 * cdef inline uint64_t load_u64(const void * p):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint64_t res
 */

static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_load_u64(void const *__pyx_v_p) {
  uint64_t __pyx_v_res;
  uint64_t __pyx_r;

  /* "asynctnt/iproto/bit.pxd":9
 *         uint64_t res
 * 
 *     res = 0             # <<<<<<<<<<<<<<
 *     memcpy(&res, p, sizeof(res))
 *     return res
 */
  __pyx_v_res = 0;

  /* "asynctnt/iproto/bit.pxd":10
 * 
 *     res = 0
 *     memcpy(&res, p, sizeof(res))             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  (void)(memcpy((&__pyx_v_res), __pyx_v_p, (sizeof(__pyx_v_res))));

  /* "asynctnt/iproto/bit.pxd":11
 *     res = 0
 *     memcpy(&res, p, sizeof(res))
 *     return res             # <<<<<<<<<<<<<<
 * 
 * cdef inline uint64_t load_u32(const void * p):
 */
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "asynctnt/iproto/bit.pxd":5
 * 
 * 
 * cdef inline uint64_t load_u64(const void * p):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint64_t res
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/bit.pxd":13
 *     return res
 * 
 * cdef inline uint64_t load_u32(const void * p):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t res
 */

static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_load_u32(void const *__pyx_v_p) {
  uint32_t __pyx_v_res;
  uint64_t __pyx_r;

  /* "asynctnt/iproto/bit.pxd":17
 *         uint32_t res
 * 
 *     res = 0             # <<<<<<<<<<<<<<
 *     memcpy(&res, p, sizeof(res))
 *     return res
 */
  __pyx_v_res = 0;

  /* "asynctnt/iproto/bit.pxd":18
 * 
 *     res = 0
 *     memcpy(&res, p, sizeof(res))             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  (void)(memcpy((&__pyx_v_res), __pyx_v_p, (sizeof(__pyx_v_res))));

  /* "asynctnt/iproto/bit.pxd":19
 *     res = 0
 *     memcpy(&res, p, sizeof(res))
 *     return res             # <<<<<<<<<<<<<<
 * 
 * cdef inline uint64_t load_u16(const void * p):
 */
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "asynctnt/iproto/bit.pxd":13
 *     return res
 * 
 * cdef inline uint64_t load_u32(const void * p):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t res
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/bit.pxd":21
 *     return res
 * 
 * cdef inline uint64_t load_u16(const void * p):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint16_t res
 */

static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_load_u16(void const *__pyx_v_p) {
  uint16_t __pyx_v_res;
  uint64_t __pyx_r;

  /* "asynctnt/iproto/bit.pxd":25
 *         uint16_t res
 * 
 *     res = 0             # <<<<<<<<<<<<<<
 *     memcpy(&res, p, sizeof(res))
 *     return res
 */
  __pyx_v_res = 0;

  /* "asynctnt/iproto/bit.pxd":26
 * 
 *     res = 0
 *     memcpy(&res, p, sizeof(res))             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  (void)(memcpy((&__pyx_v_res), __pyx_v_p, (sizeof(__pyx_v_res))));

  /* "asynctnt/iproto/bit.pxd":27
 *     res = 0
 *     memcpy(&res, p, sizeof(res))
 *     return res             # <<<<<<<<<<<<<<
 * 
 * cdef inline void store_u64(void * p, uint64_t v):
 */
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "asynctnt/iproto/bit.pxd":21
 *     return res
 * 
 * cdef inline uint64_t load_u16(const void * p):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint16_t res
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/bit.pxd":29
 *     return res
 * 
 * cdef inline void store_u64(void * p, uint64_t v):             # <<<<<<<<<<<<<<
 *     memcpy(p, &v, sizeof(v))
 * 
 */

static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_store_u64(void *__pyx_v_p, uint64_t __pyx_v_v) {

  /* "asynctnt/iproto/bit.pxd":30
 * 
 * cdef inline void store_u64(void * p, uint64_t v):
 *     memcpy(p, &v, sizeof(v))             # <<<<<<<<<<<<<<
 * 
 * cdef inline void store_u32(void * p, uint32_t v):
 */
  (void)(memcpy(__pyx_v_p, (&__pyx_v_v), (sizeof(__pyx_v_v))));

  /* "asynctnt/iproto/bit.pxd":29
 *     return res
 * 
 * cdef inline void store_u64(void * p, uint64_t v):             # <<<<<<<<<<<<<<
 *     memcpy(p, &v, sizeof(v))
 * 
 */

  /* function exit code */
}

/* "asynctnt/iproto/bit.pxd":32
 *     memcpy(p, &v, sizeof(v))
 * 
 * cdef inline void store_u32(void * p, uint32_t v):             # <<<<<<<<<<<<<<
 *     memcpy(p, &v, sizeof(v))
 * 
 */

static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_store_u32(void *__pyx_v_p, uint32_t __pyx_v_v) {

  /* "asynctnt/iproto/bit.pxd":33
 * 
 * cdef inline void store_u32(void * p, uint32_t v):
 *     memcpy(p, &v, sizeof(v))             # <<<<<<<<<<<<<<
 * 
 * cdef inline void store_u16(void * p, uint16_t v):
 */
  (void)(memcpy(__pyx_v_p, (&__pyx_v_v), (sizeof(__pyx_v_v))));

  /* "asynctnt/iproto/bit.pxd":32
 *     memcpy(p, &v, sizeof(v))
 * 
 * cdef inline void store_u32(void * p, uint32_t v):             # <<<<<<<<<<<<<<
 *     memcpy(p, &v, sizeof(v))
 * 
 */

  /* function exit code */
}

/* "asynctnt/iproto/bit.pxd":35
 *     memcpy(p, &v, sizeof(v))
 * 
 * cdef inline void store_u16(void * p, uint16_t v):             # <<<<<<<<<<<<<<
 *     memcpy(p, &v, sizeof(v))
 */

static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_store_u16(void *__pyx_v_p, uint16_t __pyx_v_v) {

  /* "asynctnt/iproto/bit.pxd":36
 * 
 * cdef inline void store_u16(void * p, uint16_t v):
 *     memcpy(p, &v, sizeof(v))             # <<<<<<<<<<<<<<
 */
  (void)(memcpy(__pyx_v_p, (&__pyx_v_v), (sizeof(__pyx_v_v))));

  /* "asynctnt/iproto/bit.pxd":35
 *     memcpy(p, &v, sizeof(v))
 * 
 * cdef inline void store_u16(void * p, uint16_t v):             # <<<<<<<<<<<<<<
 *     memcpy(p, &v, sizeof(v))
 */

  /* function exit code */
}

/* "asynctnt/iproto/ext.pxd":6
 * 
 * 
 * cdef inline uint32_t bcd_len(uint32_t digits_len):             # <<<<<<<<<<<<<<
 *     return <uint32_t> math.floor(digits_len / 2) + 1
 * 
 */

static CYTHON_INLINE uint32_t __pyx_f_8asynctnt_6iproto_8protocol_bcd_len(uint32_t __pyx_v_digits_len) {
  uint32_t __pyx_r;

  /* "asynctnt/iproto/ext.pxd":7
 * 
 * cdef inline uint32_t bcd_len(uint32_t digits_len):
 *     return <uint32_t> math.floor(digits_len / 2) + 1             # <<<<<<<<<<<<<<
 * 
 * cdef uint32_t decimal_len(int exponent, uint32_t digits_count)
 */
  __pyx_r = (((uint32_t)floor((((double)__pyx_v_digits_len) / 2.0))) + 1);
  goto __pyx_L0;

  /* "asynctnt/iproto/ext.pxd":6
 * 
 * 
 * cdef inline uint32_t bcd_len(uint32_t digits_len):             # <<<<<<<<<<<<<<
 *     return <uint32_t> math.floor(digits_len / 2) + 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/rbuffer.pxd":5
 * 
 * 
 * cdef inline size_t size_t_max(size_t a, size_t b):             # <<<<<<<<<<<<<<
 *     if a > b:
 *         return a
 */

static CYTHON_INLINE size_t __pyx_f_8asynctnt_6iproto_8protocol_size_t_max(size_t __pyx_v_a, size_t __pyx_v_b) {
  size_t __pyx_r;
  int __pyx_t_1;

  /* "asynctnt/iproto/rbuffer.pxd":6
 * 
 * cdef inline size_t size_t_max(size_t a, size_t b):
 *     if a > b:             # <<<<<<<<<<<<<<
 *         return a
 *     return b
 */
  __pyx_t_1 = (__pyx_v_a > __pyx_v_b);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/rbuffer.pxd":7
 * cdef inline size_t size_t_max(size_t a, size_t b):
 *     if a > b:
 *         return a             # <<<<<<<<<<<<<<
 *     return b
 * 
 */
    __pyx_r = __pyx_v_a;
    goto __pyx_L0;

    /* "asynctnt/iproto/rbuffer.pxd":6
 * 
 * cdef inline size_t size_t_max(size_t a, size_t b):
 *     if a > b:             # <<<<<<<<<<<<<<
 *         return a
 *     return b
 */
  }

  /* "asynctnt/iproto/rbuffer.pxd":8
 *     if a > b:
 *         return a
 *     return b             # <<<<<<<<<<<<<<
 * 
 * cdef inline uint32_t nearest_power_of_2(uint32_t v):
 */
  __pyx_r = __pyx_v_b;
  goto __pyx_L0;

  /* "asynctnt/iproto/rbuffer.pxd":5
 * 
 * 
 * cdef inline size_t size_t_max(size_t a, size_t b):             # <<<<<<<<<<<<<<
 *     if a > b:
 *         return a
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/rbuffer.pxd":10
 *     return b
 * 
 * cdef inline uint32_t nearest_power_of_2(uint32_t v):             # <<<<<<<<<<<<<<
 *     v -= 1
 *     v |= v >> 1
 */

static CYTHON_INLINE uint32_t __pyx_f_8asynctnt_6iproto_8protocol_nearest_power_of_2(uint32_t __pyx_v_v) {
  uint32_t __pyx_r;

  /* "asynctnt/iproto/rbuffer.pxd":11
 * 
 * cdef inline uint32_t nearest_power_of_2(uint32_t v):
 *     v -= 1             # <<<<<<<<<<<<<<
 *     v |= v >> 1
 *     v |= v >> 2
 */
  __pyx_v_v = (__pyx_v_v - 1);

  /* "asynctnt/iproto/rbuffer.pxd":12
 * cdef inline uint32_t nearest_power_of_2(uint32_t v):
 *     v -= 1
 *     v |= v >> 1             # <<<<<<<<<<<<<<
 *     v |= v >> 2
 *     v |= v >> 4
 */
  __pyx_v_v = (__pyx_v_v | (__pyx_v_v >> 1));

  /* "asynctnt/iproto/rbuffer.pxd":13
 *     v -= 1
 *     v |= v >> 1
 *     v |= v >> 2             # <<<<<<<<<<<<<<
 *     v |= v >> 4
 *     v |= v >> 8
 */
  __pyx_v_v = (__pyx_v_v | (__pyx_v_v >> 2));

  /* "asynctnt/iproto/rbuffer.pxd":14
 *     v |= v >> 1
 *     v |= v >> 2
 *     v |= v >> 4             # <<<<<<<<<<<<<<
 *     v |= v >> 8
 *     v |= v >> 16
 */
  __pyx_v_v = (__pyx_v_v | (__pyx_v_v >> 4));

  /* "asynctnt/iproto/rbuffer.pxd":15
 *     v |= v >> 2
 *     v |= v >> 4
 *     v |= v >> 8             # <<<<<<<<<<<<<<
 *     v |= v >> 16
 *     v += 1
 */
  __pyx_v_v = (__pyx_v_v | (__pyx_v_v >> 8));

  /* "asynctnt/iproto/rbuffer.pxd":16
 *     v |= v >> 4
 *     v |= v >> 8
 *     v |= v >> 16             # <<<<<<<<<<<<<<
 *     v += 1
 *     return v
 */
  __pyx_v_v = (__pyx_v_v | (__pyx_v_v >> 16));

  /* "asynctnt/iproto/rbuffer.pxd":17
 *     v |= v >> 8
 *     v |= v >> 16
 *     v += 1             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
  __pyx_v_v = (__pyx_v_v + 1);

  /* "asynctnt/iproto/rbuffer.pxd":18
 *     v |= v >> 16
 *     v += 1
 *     return v             # <<<<<<<<<<<<<<
 * 
 * @cython.final
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "asynctnt/iproto/rbuffer.pxd":10
 *     return b
 * 
 * cdef inline uint32_t nearest_power_of_2(uint32_t v):             # <<<<<<<<<<<<<<
 *     v -= 1
 *     v |= v >> 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/requests/base.pxd":18
 *         bint check_schema_change
 * 
 *     cdef inline Metadata metadata(self):             # <<<<<<<<<<<<<<
 *         if self.space is None:
 *             return None
 */

static CYTHON_INLINE struct C_Metadata *__pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_metadata(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self) {
  struct C_Metadata *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("metadata", 1);

  /* "asynctnt/iproto/requests/base.pxd":19
 * 
 *     cdef inline Metadata metadata(self):
 *         if self.space is None:             # <<<<<<<<<<<<<<
 *             return None
 *         return self.space.metadata
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->space) == Py_None);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/requests/base.pxd":20
 *     cdef inline Metadata metadata(self):
 *         if self.space is None:
 *             return None             # <<<<<<<<<<<<<<
 *         return self.space.metadata
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct C_Metadata *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynctnt/iproto/requests/base.pxd":19
 * 
 *     cdef inline Metadata metadata(self):
 *         if self.space is None:             # <<<<<<<<<<<<<<
 *             return None
 *         return self.space.metadata
 */
  }

  /* "asynctnt/iproto/requests/base.pxd":21
 *         if self.space is None:
 *             return None
 *         return self.space.metadata             # <<<<<<<<<<<<<<
 * 
 *     cdef inline WriteBuffer encode(self, bytes encoding)
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->space->metadata);
  __pyx_r = __pyx_v_self->space->metadata;
  goto __pyx_L0;

  /* "asynctnt/iproto/requests/base.pxd":18
 *         bint check_schema_change
 * 
 *     cdef inline Metadata metadata(self):             # <<<<<<<<<<<<<<
 *         if self.space is None:
 *             return None
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/unicodeutil.pyx":5
 * 
 * 
 * cdef bytes encode_unicode_string(object s, bytes encoding):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes b
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string(PyObject *__pyx_v_s, PyObject *__pyx_v_encoding) {
  PyObject *__pyx_v_b = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_unicode_string", 1);

  /* "asynctnt/iproto/unicodeutil.pyx":11
 * 
 *     b = <bytes><object>cpython.unicode.PyUnicode_AsEncodedString(
 *         s, encoding, b'strict'             # <<<<<<<<<<<<<<
 *     )
 *     return b
 */
  if (unlikely(__pyx_v_encoding == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(10, 11, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoding); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(10, 11, __pyx_L1_error)

  /* "asynctnt/iproto/unicodeutil.pyx":10
 *         PyObject *p
 * 
 *     b = <bytes><object>cpython.unicode.PyUnicode_AsEncodedString(             # <<<<<<<<<<<<<<
 *         s, encoding, b'strict'
 *     )
 */
  __pyx_t_2 = PyUnicode_AsEncodedString(__pyx_v_s, __pyx_t_1, ((char *)"strict")); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_b = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/unicodeutil.pyx":13
 *         s, encoding, b'strict'
 *     )
 *     return b             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_b);
  __pyx_r = __pyx_v_b;
  goto __pyx_L0;

  /* "asynctnt/iproto/unicodeutil.pyx":5
 * 
 * 
 * cdef bytes encode_unicode_string(object s, bytes encoding):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes b
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.encode_unicode_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/unicodeutil.pyx":16
 * 
 * 
 * cdef str decode_string(bytes b, bytes encoding):             # <<<<<<<<<<<<<<
 *     return <str><object>cpython.unicode.PyUnicode_FromEncodedObject(
 *         b, encoding, b'strict'
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_decode_string(PyObject *__pyx_v_b, PyObject *__pyx_v_encoding) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_string", 1);

  /* "asynctnt/iproto/unicodeutil.pyx":17
 * 
 * cdef str decode_string(bytes b, bytes encoding):
 *     return <str><object>cpython.unicode.PyUnicode_FromEncodedObject(             # <<<<<<<<<<<<<<
 *         b, encoding, b'strict'
 *     )
 */
  __Pyx_XDECREF(__pyx_r);

  /* "asynctnt/iproto/unicodeutil.pyx":18
 * cdef str decode_string(bytes b, bytes encoding):
 *     return <str><object>cpython.unicode.PyUnicode_FromEncodedObject(
 *         b, encoding, b'strict'             # <<<<<<<<<<<<<<
 *     )
 */
  if (unlikely(__pyx_v_encoding == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(10, 18, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoding); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(10, 18, __pyx_L1_error)

  /* "asynctnt/iproto/unicodeutil.pyx":17
 * 
 * cdef str decode_string(bytes b, bytes encoding):
 *     return <str><object>cpython.unicode.PyUnicode_FromEncodedObject(             # <<<<<<<<<<<<<<
 *         b, encoding, b'strict'
 *     )
 */
  __pyx_t_2 = PyUnicode_FromEncodedObject(__pyx_v_b, __pyx_t_1, ((char *)"strict")); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject*)((PyObject *)__pyx_t_2)));
  __pyx_r = ((PyObject*)__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/unicodeutil.pyx":16
 * 
 * 
 * cdef str decode_string(bytes b, bytes encoding):             # <<<<<<<<<<<<<<
 *     return <str><object>cpython.unicode.PyUnicode_FromEncodedObject(
 *         b, encoding, b'strict'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.decode_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":14
 * @cython.final
 * cdef class Field:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.name = None
 *         self.type = None
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_5Field_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8asynctnt_6iproto_8protocol_5Field_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_5Field___cinit__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_5Field___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "asynctnt/iproto/schema.pyx":15
 * cdef class Field:
 *     def __cinit__(self):
 *         self.name = None             # <<<<<<<<<<<<<<
 *         self.type = None
 *         self.collation = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)Py_None);

  /* "asynctnt/iproto/schema.pyx":16
 *     def __cinit__(self):
 *         self.name = None
 *         self.type = None             # <<<<<<<<<<<<<<
 *         self.collation = None
 *         self.is_nullable = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->type);
  __Pyx_DECREF(__pyx_v_self->type);
  __pyx_v_self->type = ((PyObject*)Py_None);

  /* "asynctnt/iproto/schema.pyx":17
 *         self.name = None
 *         self.type = None
 *         self.collation = None             # <<<<<<<<<<<<<<
 *         self.is_nullable = None
 *         self.is_autoincrement = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->collation);
  __Pyx_DECREF(__pyx_v_self->collation);
  __pyx_v_self->collation = ((PyObject*)Py_None);

  /* "asynctnt/iproto/schema.pyx":18
 *         self.type = None
 *         self.collation = None
 *         self.is_nullable = None             # <<<<<<<<<<<<<<
 *         self.is_autoincrement = None
 *         self.span = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->is_nullable);
  __Pyx_DECREF(__pyx_v_self->is_nullable);
  __pyx_v_self->is_nullable = Py_None;

  /* "asynctnt/iproto/schema.pyx":19
 *         self.collation = None
 *         self.is_nullable = None
 *         self.is_autoincrement = None             # <<<<<<<<<<<<<<
 *         self.span = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->is_autoincrement);
  __Pyx_DECREF(__pyx_v_self->is_autoincrement);
  __pyx_v_self->is_autoincrement = Py_None;

  /* "asynctnt/iproto/schema.pyx":20
 *         self.is_nullable = None
 *         self.is_autoincrement = None
 *         self.span = None             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):  # pragma: nocover
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->span);
  __Pyx_DECREF(__pyx_v_self->span);
  __pyx_v_self->span = ((PyObject*)Py_None);

  /* "asynctnt/iproto/schema.pyx":14
 * @cython.final
 * cdef class Field:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.name = None
 *         self.type = None
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":22
 *         self.span = None
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return "<Field name={} type={} is_nullable={}>".format(
 *             self.name, self.type, self.is_nullable
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_5Field_2__repr__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_2__repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "asynctnt/iproto/schema.pyx":23
 * 
 *     def __repr__(self):  # pragma: nocover
 *         return "<Field name={} type={} is_nullable={}>".format(             # <<<<<<<<<<<<<<
 *             self.name, self.type, self.is_nullable
 *         )
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_name_type_is_nullable, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "asynctnt/iproto/schema.pyx":24
 *     def __repr__(self):  # pragma: nocover
 *         return "<Field name={} type={} is_nullable={}>".format(
 *             self.name, self.type, self.is_nullable             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_self->name, __pyx_v_self->type, __pyx_v_self->is_nullable};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":22
 *         self.span = None
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return "<Field name={} type={} is_nullable={}>".format(
 *             self.name, self.type, self.is_nullable
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Field.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":6
 * cdef class Field:
 *     cdef:
 *         readonly str name             # <<<<<<<<<<<<<<
 *         readonly str type
 *         readonly str collation
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_5Field_4name___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_4name___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":7
 *     cdef:
 *         readonly str name
 *         readonly str type             # <<<<<<<<<<<<<<
 *         readonly str collation
 *         readonly object is_nullable
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_4type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_4type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_5Field_4type___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_4type___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->type);
  __pyx_r = __pyx_v_self->type;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":8
 *         readonly str name
 *         readonly str type
 *         readonly str collation             # <<<<<<<<<<<<<<
 *         readonly object is_nullable
 *         readonly object is_autoincrement
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_9collation_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_9collation_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_5Field_9collation___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_9collation___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->collation);
  __pyx_r = __pyx_v_self->collation;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":9
 *         readonly str type
 *         readonly str collation
 *         readonly object is_nullable             # <<<<<<<<<<<<<<
 *         readonly object is_autoincrement
 *         readonly str span
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_11is_nullable_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_11is_nullable_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_5Field_11is_nullable___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_11is_nullable___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->is_nullable);
  __pyx_r = __pyx_v_self->is_nullable;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":10
 *         readonly str collation
 *         readonly object is_nullable
 *         readonly object is_autoincrement             # <<<<<<<<<<<<<<
 *         readonly str span
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_16is_autoincrement_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_16is_autoincrement_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_5Field_16is_autoincrement___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_16is_autoincrement___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->is_autoincrement);
  __pyx_r = __pyx_v_self->is_autoincrement;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":11
 *         readonly object is_nullable
 *         readonly object is_autoincrement
 *         readonly str span             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_4span_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_4span_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_5Field_4span___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_4span___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->span);
  __pyx_r = __pyx_v_self->span;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_5Field_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_5Field_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_5Field_4__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Field.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_5Field_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_5Field_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5Field_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Field.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_5Field_6__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_5Field_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Field.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":30
 * @cython.freelist(METADATA_FREELIST_SIZE)
 * cdef class Metadata:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.fields = []
 *         self.names = []
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Metadata___cinit__(((struct C_Metadata *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_8Metadata___cinit__(struct C_Metadata *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "asynctnt/iproto/schema.pyx":31
 * cdef class Metadata:
 *     def __cinit__(self):
 *         self.fields = []             # <<<<<<<<<<<<<<
 *         self.names = []
 *         self.name_id_map = {}
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->fields);
  __Pyx_DECREF(__pyx_v_self->fields);
  __pyx_v_self->fields = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":32
 *     def __cinit__(self):
 *         self.fields = []
 *         self.names = []             # <<<<<<<<<<<<<<
 *         self.name_id_map = {}
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->names);
  __Pyx_DECREF(__pyx_v_self->names);
  __pyx_v_self->names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":33
 *         self.fields = []
 *         self.names = []
 *         self.name_id_map = {}             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void add(self, int id, Field field):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name_id_map);
  __Pyx_DECREF(__pyx_v_self->name_id_map);
  __pyx_v_self->name_id_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":30
 * @cython.freelist(METADATA_FREELIST_SIZE)
 * cdef class Metadata:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.fields = []
 *         self.names = []
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Metadata.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":35
 *         self.name_id_map = {}
 * 
 *     cdef inline void add(self, int id, Field field):             # <<<<<<<<<<<<<<
 *         cpython.list.PyList_Append(self.fields, field)
 *         cpython.list.PyList_Append(self.names, field.name)
 */

static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_add(struct C_Metadata *__pyx_v_self, int __pyx_v_id, struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_field) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add", 1);

  /* "asynctnt/iproto/schema.pyx":36
 * 
 *     cdef inline void add(self, int id, Field field):
 *         cpython.list.PyList_Append(self.fields, field)             # <<<<<<<<<<<<<<
 *         cpython.list.PyList_Append(self.names, field.name)
 *         cpython.dict.PyDict_SetItem(self.name_id_map, field.name, id)
 */
  __pyx_t_1 = __pyx_v_self->fields;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyList_Append(__pyx_t_1, ((PyObject *)__pyx_v_field)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":37
 *     cdef inline void add(self, int id, Field field):
 *         cpython.list.PyList_Append(self.fields, field)
 *         cpython.list.PyList_Append(self.names, field.name)             # <<<<<<<<<<<<<<
 *         cpython.dict.PyDict_SetItem(self.name_id_map, field.name, id)
 * 
 */
  __pyx_t_1 = __pyx_v_self->names;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_field->name;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_2 = PyList_Append(__pyx_t_1, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/schema.pyx":38
 *         cpython.list.PyList_Append(self.fields, field)
 *         cpython.list.PyList_Append(self.names, field.name)
 *         cpython.dict.PyDict_SetItem(self.name_id_map, field.name, id)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline str name_by_id(self, int id):
 */
  __pyx_t_3 = __pyx_v_self->name_id_map;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_v_field->name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyDict_SetItem(__pyx_t_3, __pyx_t_1, __pyx_t_4); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "asynctnt/iproto/schema.pyx":35
 *         self.name_id_map = {}
 * 
 *     cdef inline void add(self, int id, Field field):             # <<<<<<<<<<<<<<
 *         cpython.list.PyList_Append(self.fields, field)
 *         cpython.list.PyList_Append(self.names, field.name)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Metadata.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/schema.pyx":40
 *         cpython.dict.PyDict_SetItem(self.name_id_map, field.name, id)
 * 
 *     cdef inline str name_by_id(self, int id):             # <<<<<<<<<<<<<<
 *         cdef Field field
 *         field = <Field> cpython.list.PyList_GetItem(self.fields, id)
 */

static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_8Metadata_name_by_id(struct C_Metadata *__pyx_v_self, int __pyx_v_id) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_field = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("name_by_id", 1);

  /* "asynctnt/iproto/schema.pyx":42
 *     cdef inline str name_by_id(self, int id):
 *         cdef Field field
 *         field = <Field> cpython.list.PyList_GetItem(self.fields, id)             # <<<<<<<<<<<<<<
 *         return field.name
 * 
 */
  __pyx_t_1 = __pyx_v_self->fields;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyList_GetItem(__pyx_t_1, __pyx_v_id); if (unlikely(__pyx_t_2 == ((PyObject *)NULL))) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = ((PyObject *)__pyx_t_2);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_field = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":43
 *         cdef Field field
 *         field = <Field> cpython.list.PyList_GetItem(self.fields, id)
 *         return field.name             # <<<<<<<<<<<<<<
 * 
 *     cdef inline int id_by_name(self, str name) except *:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_field->name);
  __pyx_r = __pyx_v_field->name;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":40
 *         cpython.dict.PyDict_SetItem(self.name_id_map, field.name, id)
 * 
 *     cdef inline str name_by_id(self, int id):             # <<<<<<<<<<<<<<
 *         cdef Field field
 *         field = <Field> cpython.list.PyList_GetItem(self.fields, id)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Metadata.name_by_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":45
 *         return field.name
 * 
 *     cdef inline int id_by_name(self, str name) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *fld
 */

static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_id_by_name(struct C_Metadata *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_v_fld;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("id_by_name", 1);

  /* "asynctnt/iproto/schema.pyx":49
 *             PyObject *fld
 * 
 *         fld = cpython.dict.PyDict_GetItem(self.name_id_map, name)             # <<<<<<<<<<<<<<
 *         if fld == NULL:
 *             raise KeyError('Field \'{}\' not found'.format(name))
 */
  __pyx_t_1 = __pyx_v_self->name_id_map;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_fld = PyDict_GetItem(__pyx_t_1, __pyx_v_name);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":50
 * 
 *         fld = cpython.dict.PyDict_GetItem(self.name_id_map, name)
 *         if fld == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError('Field \'{}\' not found'.format(name))
 *         return <int> <object> fld
 */
  __pyx_t_2 = (__pyx_v_fld == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "asynctnt/iproto/schema.pyx":51
 *         fld = cpython.dict.PyDict_GetItem(self.name_id_map, name)
 *         if fld == NULL:
 *             raise KeyError('Field \'{}\' not found'.format(name))             # <<<<<<<<<<<<<<
 *         return <int> <object> fld
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Field_not_found, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_name};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 51, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 51, __pyx_L1_error)

    /* "asynctnt/iproto/schema.pyx":50
 * 
 *         fld = cpython.dict.PyDict_GetItem(self.name_id_map, name)
 *         if fld == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError('Field \'{}\' not found'.format(name))
 *         return <int> <object> fld
 */
  }

  /* "asynctnt/iproto/schema.pyx":52
 *         if fld == NULL:
 *             raise KeyError('Field \'{}\' not found'.format(name))
 *         return <int> <object> fld             # <<<<<<<<<<<<<<
 * 
 *     cdef inline int id_by_name_safe(self, str name) except *:
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(((PyObject *)__pyx_v_fld)); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 52, __pyx_L1_error)
  __pyx_r = ((int)__pyx_t_5);
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":45
 *         return field.name
 * 
 *     cdef inline int id_by_name(self, str name) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *fld
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Metadata.id_by_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":54
 *         return <int> <object> fld
 * 
 *     cdef inline int id_by_name_safe(self, str name) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *fld
 */

static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_id_by_name_safe(struct C_Metadata *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_v_fld;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("id_by_name_safe", 1);

  /* "asynctnt/iproto/schema.pyx":58
 *             PyObject *fld
 * 
 *         fld = cpython.dict.PyDict_GetItem(self.name_id_map, name)             # <<<<<<<<<<<<<<
 *         if fld == NULL:
 *             return -1
 */
  __pyx_t_1 = __pyx_v_self->name_id_map;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_fld = PyDict_GetItem(__pyx_t_1, __pyx_v_name);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":59
 * 
 *         fld = cpython.dict.PyDict_GetItem(self.name_id_map, name)
 *         if fld == NULL:             # <<<<<<<<<<<<<<
 *             return -1
 *         return <int> <object> fld
 */
  __pyx_t_2 = (__pyx_v_fld == NULL);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/schema.pyx":60
 *         fld = cpython.dict.PyDict_GetItem(self.name_id_map, name)
 *         if fld == NULL:
 *             return -1             # <<<<<<<<<<<<<<
 *         return <int> <object> fld
 * 
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "asynctnt/iproto/schema.pyx":59
 * 
 *         fld = cpython.dict.PyDict_GetItem(self.name_id_map, name)
 *         if fld == NULL:             # <<<<<<<<<<<<<<
 *             return -1
 *         return <int> <object> fld
 */
  }

  /* "asynctnt/iproto/schema.pyx":61
 *         if fld == NULL:
 *             return -1
 *         return <int> <object> fld             # <<<<<<<<<<<<<<
 * 
 *     cdef inline int len(self):
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(((PyObject *)__pyx_v_fld)); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 61, __pyx_L1_error)
  __pyx_r = ((int)__pyx_t_3);
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":54
 *         return <int> <object> fld
 * 
 *     cdef inline int id_by_name_safe(self, str name) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *fld
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Metadata.id_by_name_safe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":63
 *         return <int> <object> fld
 * 
 *     cdef inline int len(self):             # <<<<<<<<<<<<<<
 *         return <int> cpython.list.PyList_GET_SIZE(self.fields)
 * 
 */

static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_len(struct C_Metadata *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("len", 1);

  /* "asynctnt/iproto/schema.pyx":64
 * 
 *     cdef inline int len(self):
 *         return <int> cpython.list.PyList_GET_SIZE(self.fields)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):  # pragma: nocover
 */
  __pyx_t_1 = __pyx_v_self->fields;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_r = ((int)PyList_GET_SIZE(__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":63
 *         return <int> <object> fld
 * 
 *     cdef inline int len(self):             # <<<<<<<<<<<<<<
 *         return <int> cpython.list.PyList_GET_SIZE(self.fields)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":66
 *         return <int> cpython.list.PyList_GET_SIZE(self.fields)
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return '<Metadata [fields_count={}]>'.format(self.len())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_2__repr__(((struct C_Metadata *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_2__repr__(struct C_Metadata *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "asynctnt/iproto/schema.pyx":67
 * 
 *     def __repr__(self):  # pragma: nocover
 *         return '<Metadata [fields_count={}]>'.format(self.len())             # <<<<<<<<<<<<<<
 * 
 * @cython.final
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Metadata_fields_count, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_len(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 67, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_3 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_3 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_3, 1+__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":66
 *         return <int> cpython.list.PyList_GET_SIZE(self.fields)
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return '<Metadata [fields_count={}]>'.format(self.len())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Metadata.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":16
 * cdef public class Metadata [object C_Metadata, type C_Metadata_Type]:
 *     cdef:
 *         readonly list fields             # <<<<<<<<<<<<<<
 *         readonly dict name_id_map
 *         list names
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_6fields_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_6fields_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_6fields___get__(((struct C_Metadata *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_6fields___get__(struct C_Metadata *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->fields);
  __pyx_r = __pyx_v_self->fields;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":17
 *     cdef:
 *         readonly list fields
 *         readonly dict name_id_map             # <<<<<<<<<<<<<<
 *         list names
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_11name_id_map_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_11name_id_map_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_11name_id_map___get__(((struct C_Metadata *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_11name_id_map___get__(struct C_Metadata *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name_id_map);
  __pyx_r = __pyx_v_self->name_id_map;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_8Metadata_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_4__reduce_cython__(((struct C_Metadata *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_4__reduce_cython__(CYTHON_UNUSED struct C_Metadata *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Metadata.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_8Metadata_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Metadata.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_6__setstate_cython__(((struct C_Metadata *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Metadata_6__setstate_cython__(CYTHON_UNUSED struct C_Metadata *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Metadata.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":71
 * @cython.final
 * cdef class SchemaIndex:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.iid = -1
 *         self.sid = -1
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex___cinit__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "asynctnt/iproto/schema.pyx":72
 * cdef class SchemaIndex:
 *     def __cinit__(self):
 *         self.iid = -1             # <<<<<<<<<<<<<<
 *         self.sid = -1
 *         self.name = None
 */
  __pyx_v_self->iid = -1;

  /* "asynctnt/iproto/schema.pyx":73
 *     def __cinit__(self):
 *         self.iid = -1
 *         self.sid = -1             # <<<<<<<<<<<<<<
 *         self.name = None
 *         self.index_type = None
 */
  __pyx_v_self->sid = -1;

  /* "asynctnt/iproto/schema.pyx":74
 *         self.iid = -1
 *         self.sid = -1
 *         self.name = None             # <<<<<<<<<<<<<<
 *         self.index_type = None
 *         self.unique = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)Py_None);

  /* "asynctnt/iproto/schema.pyx":75
 *         self.sid = -1
 *         self.name = None
 *         self.index_type = None             # <<<<<<<<<<<<<<
 *         self.unique = None
 *         self.metadata = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->index_type);
  __Pyx_DECREF(__pyx_v_self->index_type);
  __pyx_v_self->index_type = ((PyObject*)Py_None);

  /* "asynctnt/iproto/schema.pyx":76
 *         self.name = None
 *         self.index_type = None
 *         self.unique = None             # <<<<<<<<<<<<<<
 *         self.metadata = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->unique);
  __Pyx_DECREF(__pyx_v_self->unique);
  __pyx_v_self->unique = Py_None;

  /* "asynctnt/iproto/schema.pyx":77
 *         self.index_type = None
 *         self.unique = None
 *         self.metadata = None             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):  # pragma: nocover
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->metadata);
  __Pyx_DECREF((PyObject *)__pyx_v_self->metadata);
  __pyx_v_self->metadata = ((struct C_Metadata *)Py_None);

  /* "asynctnt/iproto/schema.pyx":71
 * @cython.final
 * cdef class SchemaIndex:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.iid = -1
 *         self.sid = -1
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":79
 *         self.metadata = None
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return \
 *             '<{} sid={}, id={}, name={}, ' \
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_2__repr__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_2__repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "asynctnt/iproto/schema.pyx":80
 * 
 *     def __repr__(self):  # pragma: nocover
 *         return \             # <<<<<<<<<<<<<<
 *             '<{} sid={}, id={}, name={}, ' \
 *             'type={}, unique={}>'.format(
 */
  __Pyx_XDECREF(__pyx_r);

  /* "asynctnt/iproto/schema.pyx":82
 *         return \
 *             '<{} sid={}, id={}, name={}, ' \
 *             'type={}, unique={}>'.format(             # <<<<<<<<<<<<<<
 *                 self.__class__.__name__,
 *                 self.sid, self.iid, self.name, self.index_type, self.unique
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_sid_id_name_type_unique, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "asynctnt/iproto/schema.pyx":83
 *             '<{} sid={}, id={}, name={}, ' \
 *             'type={}, unique={}>'.format(
 *                 self.__class__.__name__,             # <<<<<<<<<<<<<<
 *                 self.sid, self.iid, self.name, self.index_type, self.unique
 *             )
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/schema.pyx":84
 *             'type={}, unique={}>'.format(
 *                 self.__class__.__name__,
 *                 self.sid, self.iid, self.name, self.index_type, self.unique             # <<<<<<<<<<<<<<
 *             )
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->sid); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->iid); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[7] = {__pyx_t_6, __pyx_t_4, __pyx_t_3, __pyx_t_5, __pyx_v_self->name, __pyx_v_self->index_type, __pyx_v_self->unique};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_7, 6+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":79
 *         self.metadata = None
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return \
 *             '<{} sid={}, id={}, name={}, ' \
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaIndex.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":29
 * cdef class SchemaIndex:
 *     cdef:
 *         readonly int sid             # <<<<<<<<<<<<<<
 *         readonly int iid
 *         readonly str name
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3sid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3sid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_3sid___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_3sid___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->sid); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaIndex.sid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":30
 *     cdef:
 *         readonly int sid
 *         readonly int iid             # <<<<<<<<<<<<<<
 *         readonly str name
 *         readonly str index_type
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3iid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3iid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_3iid___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_3iid___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->iid); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaIndex.iid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":31
 *         readonly int sid
 *         readonly int iid
 *         readonly str name             # <<<<<<<<<<<<<<
 *         readonly str index_type
 *         readonly object unique
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_4name___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_4name___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":32
 *         readonly int iid
 *         readonly str name
 *         readonly str index_type             # <<<<<<<<<<<<<<
 *         readonly object unique
 *         readonly Metadata metadata
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_10index_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_10index_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_10index_type___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_10index_type___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->index_type);
  __pyx_r = __pyx_v_self->index_type;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":33
 *         readonly str name
 *         readonly str index_type
 *         readonly object unique             # <<<<<<<<<<<<<<
 *         readonly Metadata metadata
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_6unique_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_6unique_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_6unique___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_6unique___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->unique);
  __pyx_r = __pyx_v_self->unique;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":34
 *         readonly str index_type
 *         readonly object unique
 *         readonly Metadata metadata             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_8metadata_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_8metadata_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_8metadata___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_8metadata___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->metadata);
  __pyx_r = ((PyObject *)__pyx_v_self->metadata);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11SchemaIndex_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_4__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaIndex.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11SchemaIndex_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaIndex.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_6__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaIndex_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaIndex.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":89
 * @cython.final
 * cdef class SchemaSpace:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.sid = -1
 *         self.owner = -1
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace___cinit__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "asynctnt/iproto/schema.pyx":90
 * cdef class SchemaSpace:
 *     def __cinit__(self):
 *         self.sid = -1             # <<<<<<<<<<<<<<
 *         self.owner = -1
 *         self.name = None
 */
  __pyx_v_self->sid = -1;

  /* "asynctnt/iproto/schema.pyx":91
 *     def __cinit__(self):
 *         self.sid = -1
 *         self.owner = -1             # <<<<<<<<<<<<<<
 *         self.name = None
 *         self.engine = None
 */
  __pyx_v_self->owner = -1;

  /* "asynctnt/iproto/schema.pyx":92
 *         self.sid = -1
 *         self.owner = -1
 *         self.name = None             # <<<<<<<<<<<<<<
 *         self.engine = None
 *         self.field_count = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)Py_None);

  /* "asynctnt/iproto/schema.pyx":93
 *         self.owner = -1
 *         self.name = None
 *         self.engine = None             # <<<<<<<<<<<<<<
 *         self.field_count = 0
 *         self.flags = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->engine);
  __Pyx_DECREF(__pyx_v_self->engine);
  __pyx_v_self->engine = ((PyObject*)Py_None);

  /* "asynctnt/iproto/schema.pyx":94
 *         self.name = None
 *         self.engine = None
 *         self.field_count = 0             # <<<<<<<<<<<<<<
 *         self.flags = None
 * 
 */
  __pyx_v_self->field_count = 0;

  /* "asynctnt/iproto/schema.pyx":95
 *         self.engine = None
 *         self.field_count = 0
 *         self.flags = None             # <<<<<<<<<<<<<<
 * 
 *         self.metadata = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->flags);
  __Pyx_DECREF(__pyx_v_self->flags);
  __pyx_v_self->flags = Py_None;

  /* "asynctnt/iproto/schema.pyx":97
 *         self.flags = None
 * 
 *         self.metadata = None             # <<<<<<<<<<<<<<
 *         self.indexes = {}
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->metadata);
  __Pyx_DECREF((PyObject *)__pyx_v_self->metadata);
  __pyx_v_self->metadata = ((struct C_Metadata *)Py_None);

  /* "asynctnt/iproto/schema.pyx":98
 * 
 *         self.metadata = None
 *         self.indexes = {}             # <<<<<<<<<<<<<<
 * 
 *     cdef void add_index(self, SchemaIndex idx):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->indexes);
  __Pyx_DECREF(__pyx_v_self->indexes);
  __pyx_v_self->indexes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":89
 * @cython.final
 * cdef class SchemaSpace:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.sid = -1
 *         self.owner = -1
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaSpace.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":100
 *         self.indexes = {}
 * 
 *     cdef void add_index(self, SchemaIndex idx):             # <<<<<<<<<<<<<<
 *         cpython.dict.PyDict_SetItem(self.indexes, idx.iid, idx)
 *         if idx.name:
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_add_index(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_idx) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_index", 1);

  /* "asynctnt/iproto/schema.pyx":101
 * 
 *     cdef void add_index(self, SchemaIndex idx):
 *         cpython.dict.PyDict_SetItem(self.indexes, idx.iid, idx)             # <<<<<<<<<<<<<<
 *         if idx.name:
 *             cpython.dict.PyDict_SetItem(self.indexes, idx.name, idx)
 */
  __pyx_t_1 = __pyx_v_self->indexes;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_idx->iid); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyDict_SetItem(__pyx_t_1, __pyx_t_2, ((PyObject *)__pyx_v_idx)); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/schema.pyx":102
 *     cdef void add_index(self, SchemaIndex idx):
 *         cpython.dict.PyDict_SetItem(self.indexes, idx.iid, idx)
 *         if idx.name:             # <<<<<<<<<<<<<<
 *             cpython.dict.PyDict_SetItem(self.indexes, idx.name, idx)
 * 
 */
  __pyx_t_4 = (__pyx_v_idx->name != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_v_idx->name) != 0);
  if (__pyx_t_4) {

    /* "asynctnt/iproto/schema.pyx":103
 *         cpython.dict.PyDict_SetItem(self.indexes, idx.iid, idx)
 *         if idx.name:
 *             cpython.dict.PyDict_SetItem(self.indexes, idx.name, idx)             # <<<<<<<<<<<<<<
 * 
 *     cdef SchemaIndex get_index(self, index, create_dummy=True):
 */
    __pyx_t_2 = __pyx_v_self->indexes;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_v_idx->name;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = PyDict_SetItem(__pyx_t_2, __pyx_t_1, ((PyObject *)__pyx_v_idx)); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 103, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/schema.pyx":102
 *     cdef void add_index(self, SchemaIndex idx):
 *         cpython.dict.PyDict_SetItem(self.indexes, idx.iid, idx)
 *         if idx.name:             # <<<<<<<<<<<<<<
 *             cpython.dict.PyDict_SetItem(self.indexes, idx.name, idx)
 * 
 */
  }

  /* "asynctnt/iproto/schema.pyx":100
 *         self.indexes = {}
 * 
 *     cdef void add_index(self, SchemaIndex idx):             # <<<<<<<<<<<<<<
 *         cpython.dict.PyDict_SetItem(self.indexes, idx.iid, idx)
 *         if idx.name:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaSpace.add_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/schema.pyx":105
 *             cpython.dict.PyDict_SetItem(self.indexes, idx.name, idx)
 * 
 *     cdef SchemaIndex get_index(self, index, create_dummy=True):             # <<<<<<<<<<<<<<
 *         cdef:
 *             SchemaIndex idx
 */

static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self, PyObject *__pyx_v_index, struct __pyx_opt_args_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index *__pyx_optional_args) {
  PyObject *__pyx_v_create_dummy = ((PyObject *)Py_True);
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_idx = 0;
  int __pyx_v_is_str;
  int __pyx_v_is_int;
  PyObject *__pyx_v_obj_p;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_index", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_create_dummy = __pyx_optional_args->create_dummy;
    }
  }

  /* "asynctnt/iproto/schema.pyx":110
 *             bint is_str, is_int
 *             PyObject *obj_p
 *         is_str = isinstance(index, str)             # <<<<<<<<<<<<<<
 *         is_int = isinstance(index, int)
 *         if not is_str and not is_int:
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_index); 
  __pyx_v_is_str = __pyx_t_1;

  /* "asynctnt/iproto/schema.pyx":111
 *             PyObject *obj_p
 *         is_str = isinstance(index, str)
 *         is_int = isinstance(index, int)             # <<<<<<<<<<<<<<
 *         if not is_str and not is_int:
 *             raise TypeError(
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_index); 
  __pyx_v_is_int = __pyx_t_1;

  /* "asynctnt/iproto/schema.pyx":112
 *         is_str = isinstance(index, str)
 *         is_int = isinstance(index, int)
 *         if not is_str and not is_int:             # <<<<<<<<<<<<<<
 *             raise TypeError(
 *                 'Index must be either str or int, got: {}'.format(type(index)))
 */
  __pyx_t_2 = (!__pyx_v_is_str);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (!__pyx_v_is_int);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "asynctnt/iproto/schema.pyx":114
 *         if not is_str and not is_int:
 *             raise TypeError(
 *                 'Index must be either str or int, got: {}'.format(type(index)))             # <<<<<<<<<<<<<<
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(self.indexes, index)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Index_must_be_either_str_or_int, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)Py_TYPE(__pyx_v_index))};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 114, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }

    /* "asynctnt/iproto/schema.pyx":113
 *         is_int = isinstance(index, int)
 *         if not is_str and not is_int:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 'Index must be either str or int, got: {}'.format(type(index)))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 113, __pyx_L1_error)

    /* "asynctnt/iproto/schema.pyx":112
 *         is_str = isinstance(index, str)
 *         is_int = isinstance(index, int)
 *         if not is_str and not is_int:             # <<<<<<<<<<<<<<
 *             raise TypeError(
 *                 'Index must be either str or int, got: {}'.format(type(index)))
 */
  }

  /* "asynctnt/iproto/schema.pyx":116
 *                 'Index must be either str or int, got: {}'.format(type(index)))
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(self.indexes, index)             # <<<<<<<<<<<<<<
 *         if obj_p is not NULL:
 *             return <SchemaIndex> obj_p
 */
  __pyx_t_4 = __pyx_v_self->indexes;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_obj_p = PyDict_GetItem(__pyx_t_4, __pyx_v_index);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "asynctnt/iproto/schema.pyx":117
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(self.indexes, index)
 *         if obj_p is not NULL:             # <<<<<<<<<<<<<<
 *             return <SchemaIndex> obj_p
 *         else:
 */
  __pyx_t_1 = (__pyx_v_obj_p != NULL);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/schema.pyx":118
 *         obj_p = cpython.dict.PyDict_GetItem(self.indexes, index)
 *         if obj_p is not NULL:
 *             return <SchemaIndex> obj_p             # <<<<<<<<<<<<<<
 *         else:
 *             if is_int and create_dummy:
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_obj_p));
    __pyx_r = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_v_obj_p);
    goto __pyx_L0;

    /* "asynctnt/iproto/schema.pyx":117
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(self.indexes, index)
 *         if obj_p is not NULL:             # <<<<<<<<<<<<<<
 *             return <SchemaIndex> obj_p
 *         else:
 */
  }

  /* "asynctnt/iproto/schema.pyx":120
 *             return <SchemaIndex> obj_p
 *         else:
 *             if is_int and create_dummy:             # <<<<<<<<<<<<<<
 *                 logger.debug(
 *                     'Index %s not found in space %s/%s. Creating dummy.',
 */
  /*else*/ {
    if (__pyx_v_is_int) {
    } else {
      __pyx_t_1 = __pyx_v_is_int;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_create_dummy); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 120, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_2;
    __pyx_L8_bool_binop_done:;
    if (likely(__pyx_t_1)) {

      /* "asynctnt/iproto/schema.pyx":121
 *         else:
 *             if is_int and create_dummy:
 *                 logger.debug(             # <<<<<<<<<<<<<<
 *                     'Index %s not found in space %s/%s. Creating dummy.',
 *                     index, self.sid, self.name
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "asynctnt/iproto/schema.pyx":123
 *                 logger.debug(
 *                     'Index %s not found in space %s/%s. Creating dummy.',
 *                     index, self.sid, self.name             # <<<<<<<<<<<<<<
 *                 )
 *                 idx = <SchemaIndex> SchemaIndex.__new__(SchemaIndex)
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->sid); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_kp_u_Index_s_not_found_in_space_s_s_C, __pyx_v_index, __pyx_t_3, __pyx_v_self->name};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 4+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 121, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "asynctnt/iproto/schema.pyx":125
 *                     index, self.sid, self.name
 *                 )
 *                 idx = <SchemaIndex> SchemaIndex.__new__(SchemaIndex)             # <<<<<<<<<<<<<<
 *                 idx.iid = index
 *                 idx.sid = self.sid
 */
      __pyx_t_4 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaIndex(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 125, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_4);
      __pyx_t_5 = ((PyObject *)__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_DECREF((PyObject *)__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_idx = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "asynctnt/iproto/schema.pyx":126
 *                 )
 *                 idx = <SchemaIndex> SchemaIndex.__new__(SchemaIndex)
 *                 idx.iid = index             # <<<<<<<<<<<<<<
 *                 idx.sid = self.sid
 *                 idx.name = str(index)
 */
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_index); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 126, __pyx_L1_error)
      __pyx_v_idx->iid = __pyx_t_6;

      /* "asynctnt/iproto/schema.pyx":127
 *                 idx = <SchemaIndex> SchemaIndex.__new__(SchemaIndex)
 *                 idx.iid = index
 *                 idx.sid = self.sid             # <<<<<<<<<<<<<<
 *                 idx.name = str(index)
 *                 cpython.dict.PyDict_SetItem(self.indexes, index, idx)
 */
      __pyx_t_6 = __pyx_v_self->sid;
      __pyx_v_idx->sid = __pyx_t_6;

      /* "asynctnt/iproto/schema.pyx":128
 *                 idx.iid = index
 *                 idx.sid = self.sid
 *                 idx.name = str(index)             # <<<<<<<<<<<<<<
 *                 cpython.dict.PyDict_SetItem(self.indexes, index, idx)
 *                 return idx
 */
      __pyx_t_5 = __Pyx_PyObject_Str(__pyx_v_index); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_v_idx->name);
      __Pyx_DECREF(__pyx_v_idx->name);
      __pyx_v_idx->name = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "asynctnt/iproto/schema.pyx":129
 *                 idx.sid = self.sid
 *                 idx.name = str(index)
 *                 cpython.dict.PyDict_SetItem(self.indexes, index, idx)             # <<<<<<<<<<<<<<
 *                 return idx
 *             else:
 */
      __pyx_t_5 = __pyx_v_self->indexes;
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_6 = PyDict_SetItem(__pyx_t_5, __pyx_v_index, ((PyObject *)__pyx_v_idx)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 129, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "asynctnt/iproto/schema.pyx":130
 *                 idx.name = str(index)
 *                 cpython.dict.PyDict_SetItem(self.indexes, index, idx)
 *                 return idx             # <<<<<<<<<<<<<<
 *             else:
 *                 raise TarantoolSchemaError(
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __Pyx_INCREF((PyObject *)__pyx_v_idx);
      __pyx_r = __pyx_v_idx;
      goto __pyx_L0;

      /* "asynctnt/iproto/schema.pyx":120
 *             return <SchemaIndex> obj_p
 *         else:
 *             if is_int and create_dummy:             # <<<<<<<<<<<<<<
 *                 logger.debug(
 *                     'Index %s not found in space %s/%s. Creating dummy.',
 */
    }

    /* "asynctnt/iproto/schema.pyx":132
 *                 return idx
 *             else:
 *                 raise TarantoolSchemaError(             # <<<<<<<<<<<<<<
 *                     'Index {} not found in space {}/{}'.format(
 *                         index, self.sid, self.name
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TarantoolSchemaError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "asynctnt/iproto/schema.pyx":133
 *             else:
 *                 raise TarantoolSchemaError(
 *                     'Index {} not found in space {}/{}'.format(             # <<<<<<<<<<<<<<
 *                         index, self.sid, self.name
 *                     )
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Index_not_found_in_space, __pyx_n_s_format); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "asynctnt/iproto/schema.pyx":134
 *                 raise TarantoolSchemaError(
 *                     'Index {} not found in space {}/{}'.format(
 *                         index, self.sid, self.name             # <<<<<<<<<<<<<<
 *                     )
 *                 )
 */
      __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_self->sid); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 134, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_9, __pyx_v_index, __pyx_t_8, __pyx_v_self->name};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 133, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __pyx_t_7 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_3};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 132, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(1, 132, __pyx_L1_error)
    }
  }

  /* "asynctnt/iproto/schema.pyx":105
 *             cpython.dict.PyDict_SetItem(self.indexes, idx.name, idx)
 * 
 *     cdef SchemaIndex get_index(self, index, create_dummy=True):             # <<<<<<<<<<<<<<
 *         cdef:
 *             SchemaIndex idx
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaSpace.get_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_idx);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":138
 *                 )
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return '<{} id={} name={} engine={}>'.format(
 *             self.__class__.__name__,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_2__repr__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_2__repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "asynctnt/iproto/schema.pyx":139
 * 
 *     def __repr__(self):  # pragma: nocover
 *         return '<{} id={} name={} engine={}>'.format(             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.sid, self.name, self.engine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_id_name_engine, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "asynctnt/iproto/schema.pyx":140
 *     def __repr__(self):  # pragma: nocover
 *         return '<{} id={} name={} engine={}>'.format(
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.sid, self.name, self.engine
 *         )
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/schema.pyx":141
 *         return '<{} id={} name={} engine={}>'.format(
 *             self.__class__.__name__,
 *             self.sid, self.name, self.engine             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->sid); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_5, __pyx_t_4, __pyx_t_3, __pyx_v_self->name, __pyx_v_self->engine};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 4+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":138
 *                 )
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return '<{} id={} name={} engine={}>'.format(
 *             self.__class__.__name__,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaSpace.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":43
 * cdef class SchemaSpace:
 *     cdef:
 *         readonly int sid             # <<<<<<<<<<<<<<
 *         readonly int owner
 *         readonly str name
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_3sid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_3sid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_3sid___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_3sid___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->sid); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaSpace.sid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":44
 *     cdef:
 *         readonly int sid
 *         readonly int owner             # <<<<<<<<<<<<<<
 *         readonly str name
 *         readonly str engine
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_5owner_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_5owner_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_5owner___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_5owner___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->owner); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaSpace.owner.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":45
 *         readonly int sid
 *         readonly int owner
 *         readonly str name             # <<<<<<<<<<<<<<
 *         readonly str engine
 *         readonly int field_count
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_4name___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_4name___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":46
 *         readonly int owner
 *         readonly str name
 *         readonly str engine             # <<<<<<<<<<<<<<
 *         readonly int field_count
 *         readonly object flags
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_6engine_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_6engine_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_6engine___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_6engine___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->engine);
  __pyx_r = __pyx_v_self->engine;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":47
 *         readonly str name
 *         readonly str engine
 *         readonly int field_count             # <<<<<<<<<<<<<<
 *         readonly object flags
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_11field_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_11field_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_11field_count___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_11field_count___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->field_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaSpace.field_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":48
 *         readonly str engine
 *         readonly int field_count
 *         readonly object flags             # <<<<<<<<<<<<<<
 * 
 *         readonly Metadata metadata
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_5flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_5flags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_5flags___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_5flags___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->flags);
  __pyx_r = __pyx_v_self->flags;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":50
 *         readonly object flags
 * 
 *         readonly Metadata metadata             # <<<<<<<<<<<<<<
 *         readonly dict indexes
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_8metadata_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_8metadata_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_8metadata___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_8metadata___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->metadata);
  __pyx_r = ((PyObject *)__pyx_v_self->metadata);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":51
 * 
 *         readonly Metadata metadata
 *         readonly dict indexes             # <<<<<<<<<<<<<<
 * 
 *     cdef void add_index(self, SchemaIndex idx)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_7indexes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_7indexes_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_7indexes___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_7indexes___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->indexes);
  __pyx_r = __pyx_v_self->indexes;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11SchemaSpace_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_4__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaSpace.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11SchemaSpace_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaSpace.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_6__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11SchemaSpace_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SchemaSpace.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":146
 * @cython.final
 * cdef class Schema:
 *     def __cinit__(self, int schema_id):             # <<<<<<<<<<<<<<
 *         self.id = schema_id
 *         self.spaces = {}
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_6Schema_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8asynctnt_6iproto_8protocol_6Schema_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_schema_id;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema_id,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema_id)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 146, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 146, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_schema_id = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_schema_id == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 146, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 146, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_6Schema___cinit__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)__pyx_v_self), __pyx_v_schema_id);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_6Schema___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, int __pyx_v_schema_id) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "asynctnt/iproto/schema.pyx":147
 * cdef class Schema:
 *     def __cinit__(self, int schema_id):
 *         self.id = schema_id             # <<<<<<<<<<<<<<
 *         self.spaces = {}
 * 
 */
  __pyx_v_self->id = __pyx_v_schema_id;

  /* "asynctnt/iproto/schema.pyx":148
 *     def __cinit__(self, int schema_id):
 *         self.id = schema_id
 *         self.spaces = {}             # <<<<<<<<<<<<<<
 * 
 *     cdef SchemaSpace get_space(self, space):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->spaces);
  __Pyx_DECREF(__pyx_v_self->spaces);
  __pyx_v_self->spaces = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":146
 * @cython.final
 * cdef class Schema:
 *     def __cinit__(self, int schema_id):             # <<<<<<<<<<<<<<
 *         self.id = schema_id
 *         self.spaces = {}
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":150
 *         self.spaces = {}
 * 
 *     cdef SchemaSpace get_space(self, space):             # <<<<<<<<<<<<<<
 *         cdef PyObject *obj_p = \
 *             cpython.dict.PyDict_GetItem(self.spaces, space)
 */

static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_space(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, PyObject *__pyx_v_space) {
  PyObject *__pyx_v_obj_p;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("get_space", 1);

  /* "asynctnt/iproto/schema.pyx":152
 *     cdef SchemaSpace get_space(self, space):
 *         cdef PyObject *obj_p = \
 *             cpython.dict.PyDict_GetItem(self.spaces, space)             # <<<<<<<<<<<<<<
 *         if obj_p is NULL:
 *             return None
 */
  __pyx_t_1 = __pyx_v_self->spaces;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_obj_p = PyDict_GetItem(__pyx_t_1, __pyx_v_space);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":153
 *         cdef PyObject *obj_p = \
 *             cpython.dict.PyDict_GetItem(self.spaces, space)
 *         if obj_p is NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         return <SchemaSpace> obj_p
 */
  __pyx_t_2 = (__pyx_v_obj_p == NULL);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/schema.pyx":154
 *             cpython.dict.PyDict_GetItem(self.spaces, space)
 *         if obj_p is NULL:
 *             return None             # <<<<<<<<<<<<<<
 *         return <SchemaSpace> obj_p
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynctnt/iproto/schema.pyx":153
 *         cdef PyObject *obj_p = \
 *             cpython.dict.PyDict_GetItem(self.spaces, space)
 *         if obj_p is NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         return <SchemaSpace> obj_p
 */
  }

  /* "asynctnt/iproto/schema.pyx":155
 *         if obj_p is NULL:
 *             return None
 *         return <SchemaSpace> obj_p             # <<<<<<<<<<<<<<
 * 
 *     cdef SchemaSpace get_or_create_space(self, space):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_obj_p));
  __pyx_r = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_obj_p);
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":150
 *         self.spaces = {}
 * 
 *     cdef SchemaSpace get_space(self, space):             # <<<<<<<<<<<<<<
 *         cdef PyObject *obj_p = \
 *             cpython.dict.PyDict_GetItem(self.spaces, space)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":157
 *         return <SchemaSpace> obj_p
 * 
 *     cdef SchemaSpace get_or_create_space(self, space):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bint is_str, is_int
 */

static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_or_create_space(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, PyObject *__pyx_v_space) {
  int __pyx_v_is_str;
  int __pyx_v_is_int;
  PyObject *__pyx_v_obj_p;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_or_create_space", 1);

  /* "asynctnt/iproto/schema.pyx":161
 *             bint is_str, is_int
 *             PyObject *obj_p
 *         is_str = isinstance(space, str)             # <<<<<<<<<<<<<<
 *         is_int = isinstance(space, int)
 *         if not is_str and not is_int:
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_space); 
  __pyx_v_is_str = __pyx_t_1;

  /* "asynctnt/iproto/schema.pyx":162
 *             PyObject *obj_p
 *         is_str = isinstance(space, str)
 *         is_int = isinstance(space, int)             # <<<<<<<<<<<<<<
 *         if not is_str and not is_int:
 *             raise TypeError(
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_space); 
  __pyx_v_is_int = __pyx_t_1;

  /* "asynctnt/iproto/schema.pyx":163
 *         is_str = isinstance(space, str)
 *         is_int = isinstance(space, int)
 *         if not is_str and not is_int:             # <<<<<<<<<<<<<<
 *             raise TypeError(
 *                 'Space must be either str or int, got: {}'.format(type(space)))
 */
  __pyx_t_2 = (!__pyx_v_is_str);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (!__pyx_v_is_int);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "asynctnt/iproto/schema.pyx":165
 *         if not is_str and not is_int:
 *             raise TypeError(
 *                 'Space must be either str or int, got: {}'.format(type(space)))             # <<<<<<<<<<<<<<
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(self.spaces, space)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Space_must_be_either_str_or_int, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)Py_TYPE(__pyx_v_space))};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }

    /* "asynctnt/iproto/schema.pyx":164
 *         is_int = isinstance(space, int)
 *         if not is_str and not is_int:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 'Space must be either str or int, got: {}'.format(type(space)))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 164, __pyx_L1_error)

    /* "asynctnt/iproto/schema.pyx":163
 *         is_str = isinstance(space, str)
 *         is_int = isinstance(space, int)
 *         if not is_str and not is_int:             # <<<<<<<<<<<<<<
 *             raise TypeError(
 *                 'Space must be either str or int, got: {}'.format(type(space)))
 */
  }

  /* "asynctnt/iproto/schema.pyx":167
 *                 'Space must be either str or int, got: {}'.format(type(space)))
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(self.spaces, space)             # <<<<<<<<<<<<<<
 *         if obj_p is NULL:
 *             if is_str:
 */
  __pyx_t_4 = __pyx_v_self->spaces;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_obj_p = PyDict_GetItem(__pyx_t_4, __pyx_v_space);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "asynctnt/iproto/schema.pyx":168
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(self.spaces, space)
 *         if obj_p is NULL:             # <<<<<<<<<<<<<<
 *             if is_str:
 *                 raise TarantoolSchemaError(
 */
  __pyx_t_1 = (__pyx_v_obj_p == NULL);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/schema.pyx":169
 *         obj_p = cpython.dict.PyDict_GetItem(self.spaces, space)
 *         if obj_p is NULL:
 *             if is_str:             # <<<<<<<<<<<<<<
 *                 raise TarantoolSchemaError(
 *                     'Space {} not found'.format(space)
 */
    if (unlikely(__pyx_v_is_str)) {

      /* "asynctnt/iproto/schema.pyx":170
 *         if obj_p is NULL:
 *             if is_str:
 *                 raise TarantoolSchemaError(             # <<<<<<<<<<<<<<
 *                     'Space {} not found'.format(space)
 *                 )
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_TarantoolSchemaError); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "asynctnt/iproto/schema.pyx":171
 *             if is_str:
 *                 raise TarantoolSchemaError(
 *                     'Space {} not found'.format(space)             # <<<<<<<<<<<<<<
 *                 )
 *             else:
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Space_not_found, __pyx_n_s_format); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_space};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 171, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __pyx_t_7 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 170, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(1, 170, __pyx_L1_error)

      /* "asynctnt/iproto/schema.pyx":169
 *         obj_p = cpython.dict.PyDict_GetItem(self.spaces, space)
 *         if obj_p is NULL:
 *             if is_str:             # <<<<<<<<<<<<<<
 *                 raise TarantoolSchemaError(
 *                     'Space {} not found'.format(space)
 */
    }

    /* "asynctnt/iproto/schema.pyx":174
 *                 )
 *             else:
 *                 return self.create_dummy_space(space)             # <<<<<<<<<<<<<<
 *         return <SchemaSpace> obj_p
 * 
 */
    /*else*/ {
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_space); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 174, __pyx_L1_error)
      __pyx_t_4 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_6Schema_create_dummy_space(__pyx_v_self, __pyx_t_6)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_4);
      __pyx_t_4 = 0;
      goto __pyx_L0;
    }

    /* "asynctnt/iproto/schema.pyx":168
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(self.spaces, space)
 *         if obj_p is NULL:             # <<<<<<<<<<<<<<
 *             if is_str:
 *                 raise TarantoolSchemaError(
 */
  }

  /* "asynctnt/iproto/schema.pyx":175
 *             else:
 *                 return self.create_dummy_space(space)
 *         return <SchemaSpace> obj_p             # <<<<<<<<<<<<<<
 * 
 *     cdef SchemaSpace create_dummy_space(self, int space_id):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_obj_p));
  __pyx_r = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_v_obj_p);
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":157
 *         return <SchemaSpace> obj_p
 * 
 *     cdef SchemaSpace get_or_create_space(self, space):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bint is_str, is_int
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.get_or_create_space", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":177
 *         return <SchemaSpace> obj_p
 * 
 *     cdef SchemaSpace create_dummy_space(self, int space_id):             # <<<<<<<<<<<<<<
 *         cdef SchemaSpace s
 *         logger.debug('Space %s not found. Creating dummy.', space_id)
 */

static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_create_dummy_space(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, int __pyx_v_space_id) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_s = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_dummy_space", 1);

  /* "asynctnt/iproto/schema.pyx":179
 *     cdef SchemaSpace create_dummy_space(self, int space_id):
 *         cdef SchemaSpace s
 *         logger.debug('Space %s not found. Creating dummy.', space_id)             # <<<<<<<<<<<<<<
 *         s = <SchemaSpace> SchemaSpace.__new__(SchemaSpace)
 *         s.sid = space_id
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_space_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_kp_u_Space_s_not_found_Creating_dummy, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":180
 *         cdef SchemaSpace s
 *         logger.debug('Space %s not found. Creating dummy.', space_id)
 *         s = <SchemaSpace> SchemaSpace.__new__(SchemaSpace)             # <<<<<<<<<<<<<<
 *         s.sid = space_id
 *         s.name = str(space_id)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaSpace(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 180, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_t_3 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_s = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/schema.pyx":181
 *         logger.debug('Space %s not found. Creating dummy.', space_id)
 *         s = <SchemaSpace> SchemaSpace.__new__(SchemaSpace)
 *         s.sid = space_id             # <<<<<<<<<<<<<<
 *         s.name = str(space_id)
 *         cpython.dict.PyDict_SetItem(self.spaces, space_id, s)
 */
  __pyx_v_s->sid = __pyx_v_space_id;

  /* "asynctnt/iproto/schema.pyx":182
 *         s = <SchemaSpace> SchemaSpace.__new__(SchemaSpace)
 *         s.sid = space_id
 *         s.name = str(space_id)             # <<<<<<<<<<<<<<
 *         cpython.dict.PyDict_SetItem(self.spaces, space_id, s)
 *         return s
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_space_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_Str(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_s->name);
  __Pyx_DECREF(__pyx_v_s->name);
  __pyx_v_s->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/schema.pyx":183
 *         s.sid = space_id
 *         s.name = str(space_id)
 *         cpython.dict.PyDict_SetItem(self.spaces, space_id, s)             # <<<<<<<<<<<<<<
 *         return s
 * 
 */
  __pyx_t_1 = __pyx_v_self->spaces;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_space_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyDict_SetItem(__pyx_t_1, __pyx_t_3, ((PyObject *)__pyx_v_s)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/schema.pyx":184
 *         s.name = str(space_id)
 *         cpython.dict.PyDict_SetItem(self.spaces, space_id, s)
 *         return s             # <<<<<<<<<<<<<<
 * 
 *     cdef inline clear(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_s);
  __pyx_r = __pyx_v_s;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":177
 *         return <SchemaSpace> obj_p
 * 
 *     cdef SchemaSpace create_dummy_space(self, int space_id):             # <<<<<<<<<<<<<<
 *         cdef SchemaSpace s
 *         logger.debug('Space %s not found. Creating dummy.', space_id)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.create_dummy_space", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_s);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":186
 *         return s
 * 
 *     cdef inline clear(self):             # <<<<<<<<<<<<<<
 *         self.spaces.clear()
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_clear(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 1);

  /* "asynctnt/iproto/schema.pyx":187
 * 
 *     cdef inline clear(self):
 *         self.spaces.clear()             # <<<<<<<<<<<<<<
 * 
 *     cdef SchemaSpace parse_space(self, space_row):
 */
  if (unlikely(__pyx_v_self->spaces == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
    __PYX_ERR(1, 187, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Clear(__pyx_v_self->spaces); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 187, __pyx_L1_error)

  /* "asynctnt/iproto/schema.pyx":186
 *         return s
 * 
 *     cdef inline clear(self):             # <<<<<<<<<<<<<<
 *         self.spaces.clear()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":189
 *         self.spaces.clear()
 * 
 *     cdef SchemaSpace parse_space(self, space_row):             # <<<<<<<<<<<<<<
 *         cdef:
 *             SchemaSpace sp
 */

static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse_space(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, PyObject *__pyx_v_space_row) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_sp = 0;
  size_t __pyx_v_k;
  size_t __pyx_v_row_len;
  PyObject *__pyx_v_format_list = 0;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_field_id;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_field = NULL;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse_space", 1);

  /* "asynctnt/iproto/schema.pyx":196
 *             list format_list
 * 
 *         assert space_row is not None             # <<<<<<<<<<<<<<
 * 
 *         sp = <SchemaSpace> SchemaSpace.__new__(SchemaSpace)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__pyx_v_space_row != Py_None);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(1, 196, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(1, 196, __pyx_L1_error)
  #endif

  /* "asynctnt/iproto/schema.pyx":198
 *         assert space_row is not None
 * 
 *         sp = <SchemaSpace> SchemaSpace.__new__(SchemaSpace)             # <<<<<<<<<<<<<<
 *         row_len = <size_t> len(space_row)
 * 
 */
  __pyx_t_2 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaSpace(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 198, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_2);
  __pyx_t_3 = ((PyObject *)__pyx_t_2);
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_DECREF((PyObject *)__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sp = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/schema.pyx":199
 * 
 *         sp = <SchemaSpace> SchemaSpace.__new__(SchemaSpace)
 *         row_len = <size_t> len(space_row)             # <<<<<<<<<<<<<<
 * 
 *         k = 0
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_space_row); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 199, __pyx_L1_error)
  __pyx_v_row_len = ((size_t)__pyx_t_4);

  /* "asynctnt/iproto/schema.pyx":201
 *         row_len = <size_t> len(space_row)
 * 
 *         k = 0             # <<<<<<<<<<<<<<
 *         sp.sid = <int> space_row[k]
 *         k += 1
 */
  __pyx_v_k = 0;

  /* "asynctnt/iproto/schema.pyx":202
 * 
 *         k = 0
 *         sp.sid = <int> space_row[k]             # <<<<<<<<<<<<<<
 *         k += 1
 *         if k < row_len:
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_space_row, __pyx_v_k, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sp->sid = ((int)__pyx_t_5);

  /* "asynctnt/iproto/schema.pyx":203
 *         k = 0
 *         sp.sid = <int> space_row[k]
 *         k += 1             # <<<<<<<<<<<<<<
 *         if k < row_len:
 *             sp.owner = <int> space_row[k]
 */
  __pyx_v_k = (__pyx_v_k + 1);

  /* "asynctnt/iproto/schema.pyx":204
 *         sp.sid = <int> space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             sp.owner = <int> space_row[k]
 *         k += 1
 */
  __pyx_t_1 = (__pyx_v_k < __pyx_v_row_len);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/schema.pyx":205
 *         k += 1
 *         if k < row_len:
 *             sp.owner = <int> space_row[k]             # <<<<<<<<<<<<<<
 *         k += 1
 *         if k < row_len:
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_space_row, __pyx_v_k, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 205, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_sp->owner = ((int)__pyx_t_5);

    /* "asynctnt/iproto/schema.pyx":204
 *         sp.sid = <int> space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             sp.owner = <int> space_row[k]
 *         k += 1
 */
  }

  /* "asynctnt/iproto/schema.pyx":206
 *         if k < row_len:
 *             sp.owner = <int> space_row[k]
 *         k += 1             # <<<<<<<<<<<<<<
 *         if k < row_len:
 *             sp.name = <str> space_row[k]
 */
  __pyx_v_k = (__pyx_v_k + 1);

  /* "asynctnt/iproto/schema.pyx":207
 *             sp.owner = <int> space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             sp.name = <str> space_row[k]
 *         k += 1
 */
  __pyx_t_1 = (__pyx_v_k < __pyx_v_row_len);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/schema.pyx":208
 *         k += 1
 *         if k < row_len:
 *             sp.name = <str> space_row[k]             # <<<<<<<<<<<<<<
 *         k += 1
 *         if k < row_len:
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_space_row, __pyx_v_k, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_sp->name);
    __Pyx_DECREF(__pyx_v_sp->name);
    __pyx_v_sp->name = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynctnt/iproto/schema.pyx":207
 *             sp.owner = <int> space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             sp.name = <str> space_row[k]
 *         k += 1
 */
  }

  /* "asynctnt/iproto/schema.pyx":209
 *         if k < row_len:
 *             sp.name = <str> space_row[k]
 *         k += 1             # <<<<<<<<<<<<<<
 *         if k < row_len:
 *             sp.engine = <str> space_row[k]
 */
  __pyx_v_k = (__pyx_v_k + 1);

  /* "asynctnt/iproto/schema.pyx":210
 *             sp.name = <str> space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             sp.engine = <str> space_row[k]
 *         k += 1
 */
  __pyx_t_1 = (__pyx_v_k < __pyx_v_row_len);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/schema.pyx":211
 *         k += 1
 *         if k < row_len:
 *             sp.engine = <str> space_row[k]             # <<<<<<<<<<<<<<
 *         k += 1
 *         if k < row_len:
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_space_row, __pyx_v_k, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_sp->engine);
    __Pyx_DECREF(__pyx_v_sp->engine);
    __pyx_v_sp->engine = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "asynctnt/iproto/schema.pyx":210
 *             sp.name = <str> space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             sp.engine = <str> space_row[k]
 *         k += 1
 */
  }

  /* "asynctnt/iproto/schema.pyx":212
 *         if k < row_len:
 *             sp.engine = <str> space_row[k]
 *         k += 1             # <<<<<<<<<<<<<<
 *         if k < row_len:
 *             sp.field_count = <int> space_row[k]
 */
  __pyx_v_k = (__pyx_v_k + 1);

  /* "asynctnt/iproto/schema.pyx":213
 *             sp.engine = <str> space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             sp.field_count = <int> space_row[k]
 *         k += 1
 */
  __pyx_t_1 = (__pyx_v_k < __pyx_v_row_len);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/schema.pyx":214
 *         k += 1
 *         if k < row_len:
 *             sp.field_count = <int> space_row[k]             # <<<<<<<<<<<<<<
 *         k += 1
 *         if k < row_len:
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_space_row, __pyx_v_k, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 214, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_sp->field_count = ((int)__pyx_t_5);

    /* "asynctnt/iproto/schema.pyx":213
 *             sp.engine = <str> space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             sp.field_count = <int> space_row[k]
 *         k += 1
 */
  }

  /* "asynctnt/iproto/schema.pyx":215
 *         if k < row_len:
 *             sp.field_count = <int> space_row[k]
 *         k += 1             # <<<<<<<<<<<<<<
 *         if k < row_len:
 *             sp.flags = space_row[k]
 */
  __pyx_v_k = (__pyx_v_k + 1);

  /* "asynctnt/iproto/schema.pyx":216
 *             sp.field_count = <int> space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             sp.flags = space_row[k]
 *         k += 1
 */
  __pyx_t_1 = (__pyx_v_k < __pyx_v_row_len);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/schema.pyx":217
 *         k += 1
 *         if k < row_len:
 *             sp.flags = space_row[k]             # <<<<<<<<<<<<<<
 *         k += 1
 *         if k < row_len:
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_space_row, __pyx_v_k, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_sp->flags);
    __Pyx_DECREF(__pyx_v_sp->flags);
    __pyx_v_sp->flags = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "asynctnt/iproto/schema.pyx":216
 *             sp.field_count = <int> space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             sp.flags = space_row[k]
 *         k += 1
 */
  }

  /* "asynctnt/iproto/schema.pyx":218
 *         if k < row_len:
 *             sp.flags = space_row[k]
 *         k += 1             # <<<<<<<<<<<<<<
 *         if k < row_len:
 *             # format
 */
  __pyx_v_k = (__pyx_v_k + 1);

  /* "asynctnt/iproto/schema.pyx":219
 *             sp.flags = space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             # format
 *             format_list = <list> space_row[k]
 */
  __pyx_t_1 = (__pyx_v_k < __pyx_v_row_len);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/schema.pyx":221
 *         if k < row_len:
 *             # format
 *             format_list = <list> space_row[k]             # <<<<<<<<<<<<<<
 *             if len(format_list) > 0:
 *                 sp.metadata = <Metadata> Metadata.__new__(Metadata)
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_space_row, __pyx_v_k, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_format_list = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynctnt/iproto/schema.pyx":222
 *             # format
 *             format_list = <list> space_row[k]
 *             if len(format_list) > 0:             # <<<<<<<<<<<<<<
 *                 sp.metadata = <Metadata> Metadata.__new__(Metadata)
 *                 for i in range(len(format_list)):
 */
    if (unlikely(__pyx_v_format_list == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 222, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyList_GET_SIZE(__pyx_v_format_list); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 222, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_4 > 0);
    if (__pyx_t_1) {

      /* "asynctnt/iproto/schema.pyx":223
 *             format_list = <list> space_row[k]
 *             if len(format_list) > 0:
 *                 sp.metadata = <Metadata> Metadata.__new__(Metadata)             # <<<<<<<<<<<<<<
 *                 for i in range(len(format_list)):
 *                     field_id = i
 */
      __pyx_t_2 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Metadata(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 223, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
      __pyx_t_3 = ((PyObject *)__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_DECREF((PyObject *)__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_GOTREF((PyObject *)__pyx_v_sp->metadata);
      __Pyx_DECREF((PyObject *)__pyx_v_sp->metadata);
      __pyx_v_sp->metadata = ((struct C_Metadata *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "asynctnt/iproto/schema.pyx":224
 *             if len(format_list) > 0:
 *                 sp.metadata = <Metadata> Metadata.__new__(Metadata)
 *                 for i in range(len(format_list)):             # <<<<<<<<<<<<<<
 *                     field_id = i
 *                     field = <Field> Field.__new__(Field)
 */
      if (unlikely(__pyx_v_format_list == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(1, 224, __pyx_L1_error)
      }
      __pyx_t_4 = __Pyx_PyList_GET_SIZE(__pyx_v_format_list); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 224, __pyx_L1_error)
      __pyx_t_6 = __pyx_t_4;
      for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
        __pyx_v_i = __pyx_t_7;

        /* "asynctnt/iproto/schema.pyx":225
 *                 sp.metadata = <Metadata> Metadata.__new__(Metadata)
 *                 for i in range(len(format_list)):
 *                     field_id = i             # <<<<<<<<<<<<<<
 *                     field = <Field> Field.__new__(Field)
 * 
 */
        __pyx_v_field_id = __pyx_v_i;

        /* "asynctnt/iproto/schema.pyx":226
 *                 for i in range(len(format_list)):
 *                     field_id = i
 *                     field = <Field> Field.__new__(Field)             # <<<<<<<<<<<<<<
 * 
 *                     field.name = format_list[i]['name']
 */
        __pyx_t_3 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Field(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Field), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 226, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_3);
        __pyx_t_2 = ((PyObject *)__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "asynctnt/iproto/schema.pyx":228
 *                     field = <Field> Field.__new__(Field)
 * 
 *                     field.name = format_list[i]['name']             # <<<<<<<<<<<<<<
 *                     field.type = format_list[i]['type']
 *                     field.is_nullable = format_list[i].get('is_nullable')
 */
        if (unlikely(__pyx_v_format_list == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(1, 228, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_format_list, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 228, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_name_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 228, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(1, 228, __pyx_L1_error)
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_v_field->name);
        __Pyx_DECREF(__pyx_v_field->name);
        __pyx_v_field->name = ((PyObject*)__pyx_t_3);
        __pyx_t_3 = 0;

        /* "asynctnt/iproto/schema.pyx":229
 * 
 *                     field.name = format_list[i]['name']
 *                     field.type = format_list[i]['type']             # <<<<<<<<<<<<<<
 *                     field.is_nullable = format_list[i].get('is_nullable')
 * 
 */
        if (unlikely(__pyx_v_format_list == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(1, 229, __pyx_L1_error)
        }
        __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_format_list, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 229, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 229, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(1, 229, __pyx_L1_error)
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_v_field->type);
        __Pyx_DECREF(__pyx_v_field->type);
        __pyx_v_field->type = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;

        /* "asynctnt/iproto/schema.pyx":230
 *                     field.name = format_list[i]['name']
 *                     field.type = format_list[i]['type']
 *                     field.is_nullable = format_list[i].get('is_nullable')             # <<<<<<<<<<<<<<
 * 
 *                     sp.metadata.add(field_id, field)
 */
        if (unlikely(__pyx_v_format_list == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(1, 230, __pyx_L1_error)
        }
        __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_format_list, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 230, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 230, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_u_is_nullable};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 230, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_v_field->is_nullable);
        __Pyx_DECREF(__pyx_v_field->is_nullable);
        __pyx_v_field->is_nullable = __pyx_t_2;
        __pyx_t_2 = 0;

        /* "asynctnt/iproto/schema.pyx":232
 *                     field.is_nullable = format_list[i].get('is_nullable')
 * 
 *                     sp.metadata.add(field_id, field)             # <<<<<<<<<<<<<<
 * 
 *         return sp
 */
        __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_add(__pyx_v_sp->metadata, __pyx_v_field_id, __pyx_v_field); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 232, __pyx_L1_error)
      }

      /* "asynctnt/iproto/schema.pyx":222
 *             # format
 *             format_list = <list> space_row[k]
 *             if len(format_list) > 0:             # <<<<<<<<<<<<<<
 *                 sp.metadata = <Metadata> Metadata.__new__(Metadata)
 *                 for i in range(len(format_list)):
 */
    }

    /* "asynctnt/iproto/schema.pyx":219
 *             sp.flags = space_row[k]
 *         k += 1
 *         if k < row_len:             # <<<<<<<<<<<<<<
 *             # format
 *             format_list = <list> space_row[k]
 */
  }

  /* "asynctnt/iproto/schema.pyx":234
 *                     sp.metadata.add(field_id, field)
 * 
 *         return sp             # <<<<<<<<<<<<<<
 * 
 *     cdef SchemaIndex parse_index(self, index_row):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_sp);
  __pyx_r = __pyx_v_sp;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":189
 *         self.spaces.clear()
 * 
 *     cdef SchemaSpace parse_space(self, space_row):             # <<<<<<<<<<<<<<
 *         cdef:
 *             SchemaSpace sp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.parse_space", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sp);
  __Pyx_XDECREF(__pyx_v_format_list);
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":236
 *         return sp
 * 
 *     cdef SchemaIndex parse_index(self, index_row):             # <<<<<<<<<<<<<<
 *         cdef:
 *             SchemaIndex idx
 */

static struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse_index(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, PyObject *__pyx_v_index_row) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_idx = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_sp = 0;
  uint32_t __pyx_v_i;
  int __pyx_v_field_id;
  PyObject *__pyx_v_parts = NULL;
  PyObject *__pyx_v_part = NULL;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_field = NULL;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  Py_ssize_t __pyx_t_12;
  long __pyx_t_13;
  long __pyx_t_14;
  uint32_t __pyx_t_15;
  long __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse_index", 1);

  /* "asynctnt/iproto/schema.pyx":241
 *             SchemaSpace sp
 *             uint32_t i
 *             int field_id = -1             # <<<<<<<<<<<<<<
 *             str field_type
 * 
 */
  __pyx_v_field_id = -1;

  /* "asynctnt/iproto/schema.pyx":244
 *             str field_type
 * 
 *         assert index_row is not None             # <<<<<<<<<<<<<<
 * 
 *         idx = <SchemaIndex> SchemaIndex.__new__(SchemaIndex)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__pyx_v_index_row != Py_None);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(1, 244, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(1, 244, __pyx_L1_error)
  #endif

  /* "asynctnt/iproto/schema.pyx":246
 *         assert index_row is not None
 * 
 *         idx = <SchemaIndex> SchemaIndex.__new__(SchemaIndex)             # <<<<<<<<<<<<<<
 *         idx.sid = <int> index_row[0]
 *         idx.iid = <int> index_row[1]
 */
  __pyx_t_2 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaIndex(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 246, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_2);
  __pyx_t_3 = ((PyObject *)__pyx_t_2);
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_DECREF((PyObject *)__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_idx = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/schema.pyx":247
 * 
 *         idx = <SchemaIndex> SchemaIndex.__new__(SchemaIndex)
 *         idx.sid = <int> index_row[0]             # <<<<<<<<<<<<<<
 *         idx.iid = <int> index_row[1]
 *         idx.name = <str> index_row[2]
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_index_row, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 247, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_idx->sid = ((int)__pyx_t_4);

  /* "asynctnt/iproto/schema.pyx":248
 *         idx = <SchemaIndex> SchemaIndex.__new__(SchemaIndex)
 *         idx.sid = <int> index_row[0]
 *         idx.iid = <int> index_row[1]             # <<<<<<<<<<<<<<
 *         idx.name = <str> index_row[2]
 *         idx.index_type = <str> index_row[3]
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_index_row, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 248, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_idx->iid = ((int)__pyx_t_4);

  /* "asynctnt/iproto/schema.pyx":249
 *         idx.sid = <int> index_row[0]
 *         idx.iid = <int> index_row[1]
 *         idx.name = <str> index_row[2]             # <<<<<<<<<<<<<<
 *         idx.index_type = <str> index_row[3]
 *         idx.unique = <object> index_row[4]
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_index_row, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_idx->name);
  __Pyx_DECREF(__pyx_v_idx->name);
  __pyx_v_idx->name = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/schema.pyx":250
 *         idx.iid = <int> index_row[1]
 *         idx.name = <str> index_row[2]
 *         idx.index_type = <str> index_row[3]             # <<<<<<<<<<<<<<
 *         idx.unique = <object> index_row[4]
 *         idx.metadata = None
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_index_row, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_idx->index_type);
  __Pyx_DECREF(__pyx_v_idx->index_type);
  __pyx_v_idx->index_type = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/schema.pyx":251
 *         idx.name = <str> index_row[2]
 *         idx.index_type = <str> index_row[3]
 *         idx.unique = <object> index_row[4]             # <<<<<<<<<<<<<<
 *         idx.metadata = None
 * 
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_index_row, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_idx->unique);
  __Pyx_DECREF(__pyx_v_idx->unique);
  __pyx_v_idx->unique = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/schema.pyx":252
 *         idx.index_type = <str> index_row[3]
 *         idx.unique = <object> index_row[4]
 *         idx.metadata = None             # <<<<<<<<<<<<<<
 * 
 *         sp = self.get_space(idx.sid)
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_idx->metadata);
  __Pyx_DECREF((PyObject *)__pyx_v_idx->metadata);
  __pyx_v_idx->metadata = ((struct C_Metadata *)Py_None);

  /* "asynctnt/iproto/schema.pyx":254
 *         idx.metadata = None
 * 
 *         sp = self.get_space(idx.sid)             # <<<<<<<<<<<<<<
 *         if sp is None:
 *             raise TarantoolSchemaError(
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_idx->sid); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_space(__pyx_v_self, __pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sp = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/schema.pyx":255
 * 
 *         sp = self.get_space(idx.sid)
 *         if sp is None:             # <<<<<<<<<<<<<<
 *             raise TarantoolSchemaError(
 *                 'Space with id {} not found'.format(idx.sid))
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_sp) == Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "asynctnt/iproto/schema.pyx":256
 *         sp = self.get_space(idx.sid)
 *         if sp is None:
 *             raise TarantoolSchemaError(             # <<<<<<<<<<<<<<
 *                 'Space with id {} not found'.format(idx.sid))
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_TarantoolSchemaError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "asynctnt/iproto/schema.pyx":257
 *         if sp is None:
 *             raise TarantoolSchemaError(
 *                 'Space with id {} not found'.format(idx.sid))             # <<<<<<<<<<<<<<
 * 
 *         if sp.metadata is not None:
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Space_with_id_not_found, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_idx->sid); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_7};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_t_6 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 256, __pyx_L1_error)

    /* "asynctnt/iproto/schema.pyx":255
 * 
 *         sp = self.get_space(idx.sid)
 *         if sp is None:             # <<<<<<<<<<<<<<
 *             raise TarantoolSchemaError(
 *                 'Space with id {} not found'.format(idx.sid))
 */
  }

  /* "asynctnt/iproto/schema.pyx":259
 *                 'Space with id {} not found'.format(idx.sid))
 * 
 *         if sp.metadata is not None:             # <<<<<<<<<<<<<<
 *             parts = index_row[5]
 *             idx.metadata = <Metadata> Metadata.__new__(Metadata)
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_sp->metadata) != Py_None);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/schema.pyx":260
 * 
 *         if sp.metadata is not None:
 *             parts = index_row[5]             # <<<<<<<<<<<<<<
 *             idx.metadata = <Metadata> Metadata.__new__(Metadata)
 * 
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_index_row, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_parts = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "asynctnt/iproto/schema.pyx":261
 *         if sp.metadata is not None:
 *             parts = index_row[5]
 *             idx.metadata = <Metadata> Metadata.__new__(Metadata)             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(parts, (list, tuple)):
 */
    __pyx_t_3 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Metadata(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 261, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
    __pyx_t_2 = ((PyObject *)__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF((PyObject *)__pyx_v_idx->metadata);
    __Pyx_DECREF((PyObject *)__pyx_v_idx->metadata);
    __pyx_v_idx->metadata = ((struct C_Metadata *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynctnt/iproto/schema.pyx":263
 *             idx.metadata = <Metadata> Metadata.__new__(Metadata)
 * 
 *             if isinstance(parts, (list, tuple)):             # <<<<<<<<<<<<<<
 *                 for part in parts:
 *                     field = <Field> Field.__new__(Field)
 */
    __pyx_t_9 = PyList_Check(__pyx_v_parts); 
    if (!__pyx_t_9) {
    } else {
      __pyx_t_1 = __pyx_t_9;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_9 = PyTuple_Check(__pyx_v_parts); 
    __pyx_t_1 = __pyx_t_9;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_1) {

      /* "asynctnt/iproto/schema.pyx":264
 * 
 *             if isinstance(parts, (list, tuple)):
 *                 for part in parts:             # <<<<<<<<<<<<<<
 *                     field = <Field> Field.__new__(Field)
 *                     field.name = ''
 */
      if (likely(PyList_CheckExact(__pyx_v_parts)) || PyTuple_CheckExact(__pyx_v_parts)) {
        __pyx_t_2 = __pyx_v_parts; __Pyx_INCREF(__pyx_t_2);
        __pyx_t_10 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_10 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_parts); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 264, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 264, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 264, __pyx_L1_error)
              #endif
              if (__pyx_t_10 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_10); __Pyx_INCREF(__pyx_t_3); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(1, 264, __pyx_L1_error)
            #else
            __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 264, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 264, __pyx_L1_error)
              #endif
              if (__pyx_t_10 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_10); __Pyx_INCREF(__pyx_t_3); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(1, 264, __pyx_L1_error)
            #else
            __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 264, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_11(__pyx_t_2);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(1, 264, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_v_part, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "asynctnt/iproto/schema.pyx":265
 *             if isinstance(parts, (list, tuple)):
 *                 for part in parts:
 *                     field = <Field> Field.__new__(Field)             # <<<<<<<<<<<<<<
 *                     field.name = ''
 * 
 */
        __pyx_t_3 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Field(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Field), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 265, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_3);
        __pyx_t_5 = ((PyObject *)__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_t_5));
        __pyx_t_5 = 0;

        /* "asynctnt/iproto/schema.pyx":266
 *                 for part in parts:
 *                     field = <Field> Field.__new__(Field)
 *                     field.name = ''             # <<<<<<<<<<<<<<
 * 
 *                     if isinstance(part, (list, tuple)):
 */
        __Pyx_INCREF(__pyx_kp_u__3);
        __Pyx_GIVEREF(__pyx_kp_u__3);
        __Pyx_GOTREF(__pyx_v_field->name);
        __Pyx_DECREF(__pyx_v_field->name);
        __pyx_v_field->name = __pyx_kp_u__3;

        /* "asynctnt/iproto/schema.pyx":268
 *                     field.name = ''
 * 
 *                     if isinstance(part, (list, tuple)):             # <<<<<<<<<<<<<<
 *                         assert len(part) == 2, 'Part len must be 2'
 *                         field_id = part[0]
 */
        __pyx_t_9 = PyList_Check(__pyx_v_part); 
        if (!__pyx_t_9) {
        } else {
          __pyx_t_1 = __pyx_t_9;
          goto __pyx_L11_bool_binop_done;
        }
        __pyx_t_9 = PyTuple_Check(__pyx_v_part); 
        __pyx_t_1 = __pyx_t_9;
        __pyx_L11_bool_binop_done:;
        if (__pyx_t_1) {

          /* "asynctnt/iproto/schema.pyx":269
 * 
 *                     if isinstance(part, (list, tuple)):
 *                         assert len(part) == 2, 'Part len must be 2'             # <<<<<<<<<<<<<<
 *                         field_id = part[0]
 *                         field.type = part[1]
 */
          #ifndef CYTHON_WITHOUT_ASSERTIONS
          if (unlikely(__pyx_assertions_enabled())) {
            __pyx_t_12 = PyObject_Length(__pyx_v_part); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(1, 269, __pyx_L1_error)
            __pyx_t_1 = (__pyx_t_12 == 2);
            if (unlikely(!__pyx_t_1)) {
              __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Part_len_must_be_2, 0, 0);
              __PYX_ERR(1, 269, __pyx_L1_error)
            }
          }
          #else
          if ((1)); else __PYX_ERR(1, 269, __pyx_L1_error)
          #endif

          /* "asynctnt/iproto/schema.pyx":270
 *                     if isinstance(part, (list, tuple)):
 *                         assert len(part) == 2, 'Part len must be 2'
 *                         field_id = part[0]             # <<<<<<<<<<<<<<
 *                         field.type = part[1]
 *                     elif isinstance(part, dict):
 */
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_part, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 270, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 270, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_field_id = __pyx_t_4;

          /* "asynctnt/iproto/schema.pyx":271
 *                         assert len(part) == 2, 'Part len must be 2'
 *                         field_id = part[0]
 *                         field.type = part[1]             # <<<<<<<<<<<<<<
 *                     elif isinstance(part, dict):
 *                         field_id = part['field']
 */
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_part, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 271, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(1, 271, __pyx_L1_error)
          __Pyx_GIVEREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_v_field->type);
          __Pyx_DECREF(__pyx_v_field->type);
          __pyx_v_field->type = ((PyObject*)__pyx_t_5);
          __pyx_t_5 = 0;

          /* "asynctnt/iproto/schema.pyx":268
 *                     field.name = ''
 * 
 *                     if isinstance(part, (list, tuple)):             # <<<<<<<<<<<<<<
 *                         assert len(part) == 2, 'Part len must be 2'
 *                         field_id = part[0]
 */
          goto __pyx_L10;
        }

        /* "asynctnt/iproto/schema.pyx":272
 *                         field_id = part[0]
 *                         field.type = part[1]
 *                     elif isinstance(part, dict):             # <<<<<<<<<<<<<<
 *                         field_id = part['field']
 *                         field.type = part['type']
 */
        __pyx_t_1 = PyDict_Check(__pyx_v_part); 
        if (likely(__pyx_t_1)) {

          /* "asynctnt/iproto/schema.pyx":273
 *                         field.type = part[1]
 *                     elif isinstance(part, dict):
 *                         field_id = part['field']             # <<<<<<<<<<<<<<
 *                         field.type = part['type']
 *                         # TODO: add is_nullable and collation if we really need
 */
          __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_part, __pyx_n_u_field); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 273, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 273, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_field_id = __pyx_t_4;

          /* "asynctnt/iproto/schema.pyx":274
 *                     elif isinstance(part, dict):
 *                         field_id = part['field']
 *                         field.type = part['type']             # <<<<<<<<<<<<<<
 *                         # TODO: add is_nullable and collation if we really need
 *                         # TODO: it in a python driver
 */
          __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_part, __pyx_n_u_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 274, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(1, 274, __pyx_L1_error)
          __Pyx_GIVEREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_v_field->type);
          __Pyx_DECREF(__pyx_v_field->type);
          __pyx_v_field->type = ((PyObject*)__pyx_t_5);
          __pyx_t_5 = 0;

          /* "asynctnt/iproto/schema.pyx":272
 *                         field_id = part[0]
 *                         field.type = part[1]
 *                     elif isinstance(part, dict):             # <<<<<<<<<<<<<<
 *                         field_id = part['field']
 *                         field.type = part['type']
 */
          goto __pyx_L10;
        }

        /* "asynctnt/iproto/schema.pyx":278
 *                         # TODO: it in a python driver
 *                     else:
 *                         raise TypeError(             # <<<<<<<<<<<<<<
 *                             "unexpected type of part: {}. "
 *                             "must be either dict or list/tuple".format(
 */
        /*else*/ {

          /* "asynctnt/iproto/schema.pyx":280
 *                         raise TypeError(
 *                             "unexpected type of part: {}. "
 *                             "must be either dict or list/tuple".format(             # <<<<<<<<<<<<<<
 *                                 type(part)
 *                             )
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_unexpected_type_of_part_must_be, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 280, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);

          /* "asynctnt/iproto/schema.pyx":281
 *                             "unexpected type of part: {}. "
 *                             "must be either dict or list/tuple".format(
 *                                 type(part)             # <<<<<<<<<<<<<<
 *                             )
 *                         )
 */
          __pyx_t_6 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_6, ((PyObject *)Py_TYPE(__pyx_v_part))};
            __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 280, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }

          /* "asynctnt/iproto/schema.pyx":278
 *                         # TODO: it in a python driver
 *                     else:
 *                         raise TypeError(             # <<<<<<<<<<<<<<
 *                             "unexpected type of part: {}. "
 *                             "must be either dict or list/tuple".format(
 */
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 278, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(1, 278, __pyx_L1_error)
        }
        __pyx_L10:;

        /* "asynctnt/iproto/schema.pyx":285
 *                         )
 * 
 *                     if field_id < sp.metadata.len():             # <<<<<<<<<<<<<<
 *                         field.name = sp.metadata.name_by_id(field_id)
 *                     else:
 */
        __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_len(__pyx_v_sp->metadata); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 285, __pyx_L1_error)
        __pyx_t_1 = (__pyx_v_field_id < __pyx_t_4);
        if (__pyx_t_1) {

          /* "asynctnt/iproto/schema.pyx":286
 * 
 *                     if field_id < sp.metadata.len():
 *                         field.name = sp.metadata.name_by_id(field_id)             # <<<<<<<<<<<<<<
 *                     else:
 *                         logger.debug(
 */
          __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_name_by_id(__pyx_v_sp->metadata, __pyx_v_field_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 286, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_v_field->name);
          __Pyx_DECREF(__pyx_v_field->name);
          __pyx_v_field->name = ((PyObject*)__pyx_t_3);
          __pyx_t_3 = 0;

          /* "asynctnt/iproto/schema.pyx":285
 *                         )
 * 
 *                     if field_id < sp.metadata.len():             # <<<<<<<<<<<<<<
 *                         field.name = sp.metadata.name_by_id(field_id)
 *                     else:
 */
          goto __pyx_L13;
        }

        /* "asynctnt/iproto/schema.pyx":288
 *                         field.name = sp.metadata.name_by_id(field_id)
 *                     else:
 *                         logger.debug(             # <<<<<<<<<<<<<<
 *                             'Field #%d of space %s is not '
 *                             'in space format definition', field_id, sp.name)
 */
        /*else*/ {
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_logger); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "asynctnt/iproto/schema.pyx":290
 *                         logger.debug(
 *                             'Field #%d of space %s is not '
 *                             'in space format definition', field_id, sp.name)             # <<<<<<<<<<<<<<
 * 
 *                     idx.metadata.add(field_id, field)
 */
          __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_field_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 290, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[4] = {__pyx_t_7, __pyx_kp_u_Field_d_of_space_s_is_not_in_spa, __pyx_t_5, __pyx_v_sp->name};
            __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 288, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_L13:;

        /* "asynctnt/iproto/schema.pyx":292
 *                             'in space format definition', field_id, sp.name)
 * 
 *                     idx.metadata.add(field_id, field)             # <<<<<<<<<<<<<<
 *             else:
 *                 for i in range(parts):
 */
        __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_add(__pyx_v_idx->metadata, __pyx_v_field_id, __pyx_v_field); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 292, __pyx_L1_error)

        /* "asynctnt/iproto/schema.pyx":264
 * 
 *             if isinstance(parts, (list, tuple)):
 *                 for part in parts:             # <<<<<<<<<<<<<<
 *                     field = <Field> Field.__new__(Field)
 *                     field.name = ''
 */
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "asynctnt/iproto/schema.pyx":263
 *             idx.metadata = <Metadata> Metadata.__new__(Metadata)
 * 
 *             if isinstance(parts, (list, tuple)):             # <<<<<<<<<<<<<<
 *                 for part in parts:
 *                     field = <Field> Field.__new__(Field)
 */
      goto __pyx_L5;
    }

    /* "asynctnt/iproto/schema.pyx":294
 *                     idx.metadata.add(field_id, field)
 *             else:
 *                 for i in range(parts):             # <<<<<<<<<<<<<<
 *                     field = <Field> Field.__new__(Field)
 * 
 */
    /*else*/ {
      __pyx_t_13 = __Pyx_PyInt_As_long(__pyx_v_parts); if (unlikely((__pyx_t_13 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 294, __pyx_L1_error)
      __pyx_t_14 = __pyx_t_13;
      for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
        __pyx_v_i = __pyx_t_15;

        /* "asynctnt/iproto/schema.pyx":295
 *             else:
 *                 for i in range(parts):
 *                     field = <Field> Field.__new__(Field)             # <<<<<<<<<<<<<<
 * 
 *                     field_id = index_row[5 + 1 + i * 2]
 */
        __pyx_t_2 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Field(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Field), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 295, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_2);
        __pyx_t_3 = ((PyObject *)__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_DECREF((PyObject *)__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_t_3));
        __pyx_t_3 = 0;

        /* "asynctnt/iproto/schema.pyx":297
 *                     field = <Field> Field.__new__(Field)
 * 
 *                     field_id = index_row[5 + 1 + i * 2]             # <<<<<<<<<<<<<<
 *                     field.type = index_row[5 + 2 + i * 2]
 *                     field.name = ''
 */
        __pyx_t_16 = (0x6 + (__pyx_v_i * 2));
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_index_row, __pyx_t_16, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 297, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 297, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_field_id = __pyx_t_4;

        /* "asynctnt/iproto/schema.pyx":298
 * 
 *                     field_id = index_row[5 + 1 + i * 2]
 *                     field.type = index_row[5 + 2 + i * 2]             # <<<<<<<<<<<<<<
 *                     field.name = ''
 * 
 */
        __pyx_t_16 = (0x7 + (__pyx_v_i * 2));
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_index_row, __pyx_t_16, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 298, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(1, 298, __pyx_L1_error)
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_v_field->type);
        __Pyx_DECREF(__pyx_v_field->type);
        __pyx_v_field->type = ((PyObject*)__pyx_t_3);
        __pyx_t_3 = 0;

        /* "asynctnt/iproto/schema.pyx":299
 *                     field_id = index_row[5 + 1 + i * 2]
 *                     field.type = index_row[5 + 2 + i * 2]
 *                     field.name = ''             # <<<<<<<<<<<<<<
 * 
 *                     if field_id < sp.metadata.len():
 */
        __Pyx_INCREF(__pyx_kp_u__3);
        __Pyx_GIVEREF(__pyx_kp_u__3);
        __Pyx_GOTREF(__pyx_v_field->name);
        __Pyx_DECREF(__pyx_v_field->name);
        __pyx_v_field->name = __pyx_kp_u__3;

        /* "asynctnt/iproto/schema.pyx":301
 *                     field.name = ''
 * 
 *                     if field_id < sp.metadata.len():             # <<<<<<<<<<<<<<
 *                         field.name = sp.metadata.name_by_id(field_id)
 *                     else:
 */
        __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_len(__pyx_v_sp->metadata); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 301, __pyx_L1_error)
        __pyx_t_1 = (__pyx_v_field_id < __pyx_t_4);
        if (__pyx_t_1) {

          /* "asynctnt/iproto/schema.pyx":302
 * 
 *                     if field_id < sp.metadata.len():
 *                         field.name = sp.metadata.name_by_id(field_id)             # <<<<<<<<<<<<<<
 *                     else:
 *                         logger.debug(
 */
          __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_name_by_id(__pyx_v_sp->metadata, __pyx_v_field_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 302, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_v_field->name);
          __Pyx_DECREF(__pyx_v_field->name);
          __pyx_v_field->name = ((PyObject*)__pyx_t_3);
          __pyx_t_3 = 0;

          /* "asynctnt/iproto/schema.pyx":301
 *                     field.name = ''
 * 
 *                     if field_id < sp.metadata.len():             # <<<<<<<<<<<<<<
 *                         field.name = sp.metadata.name_by_id(field_id)
 *                     else:
 */
          goto __pyx_L17;
        }

        /* "asynctnt/iproto/schema.pyx":304
 *                         field.name = sp.metadata.name_by_id(field_id)
 *                     else:
 *                         logger.debug(             # <<<<<<<<<<<<<<
 *                             'Field #%d of space %s is not '
 *                             'in space format definition', field_id, sp.name)
 */
        /*else*/ {
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 304, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 304, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "asynctnt/iproto/schema.pyx":306
 *                         logger.debug(
 *                             'Field #%d of space %s is not '
 *                             'in space format definition', field_id, sp.name)             # <<<<<<<<<<<<<<
 * 
 *                     idx.metadata.add(field_id, field)
 */
          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_field_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 306, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_kp_u_Field_d_of_space_s_is_not_in_spa, __pyx_t_2, __pyx_v_sp->name};
            __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 304, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_L17:;

        /* "asynctnt/iproto/schema.pyx":308
 *                             'in space format definition', field_id, sp.name)
 * 
 *                     idx.metadata.add(field_id, field)             # <<<<<<<<<<<<<<
 * 
 *         return idx
 */
        __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_add(__pyx_v_idx->metadata, __pyx_v_field_id, __pyx_v_field); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 308, __pyx_L1_error)
      }
    }
    __pyx_L5:;

    /* "asynctnt/iproto/schema.pyx":259
 *                 'Space with id {} not found'.format(idx.sid))
 * 
 *         if sp.metadata is not None:             # <<<<<<<<<<<<<<
 *             parts = index_row[5]
 *             idx.metadata = <Metadata> Metadata.__new__(Metadata)
 */
  }

  /* "asynctnt/iproto/schema.pyx":310
 *                     idx.metadata.add(field_id, field)
 * 
 *         return idx             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_idx);
  __pyx_r = __pyx_v_idx;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":236
 *         return sp
 * 
 *     cdef SchemaIndex parse_index(self, index_row):             # <<<<<<<<<<<<<<
 *         cdef:
 *             SchemaIndex idx
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.parse_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_idx);
  __Pyx_XDECREF((PyObject *)__pyx_v_sp);
  __Pyx_XDECREF(__pyx_v_parts);
  __Pyx_XDECREF(__pyx_v_part);
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":313
 * 
 *     @staticmethod
 *     cdef Schema parse(int64_t schema_id, spaces, indexes):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Schema s
 */

static struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse(int64_t __pyx_v_schema_id, PyObject *__pyx_v_spaces, PyObject *__pyx_v_indexes) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_s = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_sp = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_idx = 0;
  PyObject *__pyx_v_space_row = NULL;
  PyObject *__pyx_v_index_row = NULL;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse", 1);

  /* "asynctnt/iproto/schema.pyx":319
 *             SchemaIndex idx
 * 
 *         s = <Schema> Schema.__new__(Schema, <int> schema_id)             # <<<<<<<<<<<<<<
 *         for space_row in spaces:
 *             sp = s.parse_space(space_row)
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(((int)__pyx_v_schema_id)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 319, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Schema(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Schema), __pyx_t_2, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 319, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_s = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/schema.pyx":320
 * 
 *         s = <Schema> Schema.__new__(Schema, <int> schema_id)
 *         for space_row in spaces:             # <<<<<<<<<<<<<<
 *             sp = s.parse_space(space_row)
 *             cpython.dict.PyDict_SetItem(s.spaces, sp.sid, sp)
 */
  if (likely(PyList_CheckExact(__pyx_v_spaces)) || PyTuple_CheckExact(__pyx_v_spaces)) {
    __pyx_t_2 = __pyx_v_spaces; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_spaces); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 320, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 320, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(1, 320, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 320, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(1, 320, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 320, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_space_row, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/schema.pyx":321
 *         s = <Schema> Schema.__new__(Schema, <int> schema_id)
 *         for space_row in spaces:
 *             sp = s.parse_space(space_row)             # <<<<<<<<<<<<<<
 *             cpython.dict.PyDict_SetItem(s.spaces, sp.sid, sp)
 *             if sp.name:
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse_space(__pyx_v_s, __pyx_v_space_row)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sp, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/schema.pyx":322
 *         for space_row in spaces:
 *             sp = s.parse_space(space_row)
 *             cpython.dict.PyDict_SetItem(s.spaces, sp.sid, sp)             # <<<<<<<<<<<<<<
 *             if sp.name:
 *                 cpython.dict.PyDict_SetItem(s.spaces, sp.name, sp)
 */
    __pyx_t_1 = __pyx_v_s->spaces;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_sp->sid); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyDict_SetItem(__pyx_t_1, __pyx_t_5, ((PyObject *)__pyx_v_sp)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 322, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "asynctnt/iproto/schema.pyx":323
 *             sp = s.parse_space(space_row)
 *             cpython.dict.PyDict_SetItem(s.spaces, sp.sid, sp)
 *             if sp.name:             # <<<<<<<<<<<<<<
 *                 cpython.dict.PyDict_SetItem(s.spaces, sp.name, sp)
 * 
 */
    __pyx_t_7 = (__pyx_v_sp->name != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_v_sp->name) != 0);
    if (__pyx_t_7) {

      /* "asynctnt/iproto/schema.pyx":324
 *             cpython.dict.PyDict_SetItem(s.spaces, sp.sid, sp)
 *             if sp.name:
 *                 cpython.dict.PyDict_SetItem(s.spaces, sp.name, sp)             # <<<<<<<<<<<<<<
 * 
 *         for index_row in indexes:
 */
      __pyx_t_5 = __pyx_v_s->spaces;
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_1 = __pyx_v_sp->name;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_6 = PyDict_SetItem(__pyx_t_5, __pyx_t_1, ((PyObject *)__pyx_v_sp)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 324, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynctnt/iproto/schema.pyx":323
 *             sp = s.parse_space(space_row)
 *             cpython.dict.PyDict_SetItem(s.spaces, sp.sid, sp)
 *             if sp.name:             # <<<<<<<<<<<<<<
 *                 cpython.dict.PyDict_SetItem(s.spaces, sp.name, sp)
 * 
 */
    }

    /* "asynctnt/iproto/schema.pyx":320
 * 
 *         s = <Schema> Schema.__new__(Schema, <int> schema_id)
 *         for space_row in spaces:             # <<<<<<<<<<<<<<
 *             sp = s.parse_space(space_row)
 *             cpython.dict.PyDict_SetItem(s.spaces, sp.sid, sp)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/schema.pyx":326
 *                 cpython.dict.PyDict_SetItem(s.spaces, sp.name, sp)
 * 
 *         for index_row in indexes:             # <<<<<<<<<<<<<<
 *             idx = s.parse_index(index_row)
 *             sp = s.spaces[idx.sid]
 */
  if (likely(PyList_CheckExact(__pyx_v_indexes)) || PyTuple_CheckExact(__pyx_v_indexes)) {
    __pyx_t_2 = __pyx_v_indexes; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_indexes); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 326, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 326, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(1, 326, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 326, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 326, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(1, 326, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 326, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 326, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_index_row, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/schema.pyx":327
 * 
 *         for index_row in indexes:
 *             idx = s.parse_index(index_row)             # <<<<<<<<<<<<<<
 *             sp = s.spaces[idx.sid]
 *             sp.add_index(idx)
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse_index(__pyx_v_s, __pyx_v_index_row)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_idx, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/schema.pyx":328
 *         for index_row in indexes:
 *             idx = s.parse_index(index_row)
 *             sp = s.spaces[idx.sid]             # <<<<<<<<<<<<<<
 *             sp.add_index(idx)
 * 
 */
    if (unlikely(__pyx_v_s->spaces == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 328, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_idx->sid); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_s->spaces, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(1, 328, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_sp, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "asynctnt/iproto/schema.pyx":329
 *             idx = s.parse_index(index_row)
 *             sp = s.spaces[idx.sid]
 *             sp.add_index(idx)             # <<<<<<<<<<<<<<
 * 
 *         return s
 */
    __pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_add_index(__pyx_v_sp, __pyx_v_idx); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 329, __pyx_L1_error)

    /* "asynctnt/iproto/schema.pyx":326
 *                 cpython.dict.PyDict_SetItem(s.spaces, sp.name, sp)
 * 
 *         for index_row in indexes:             # <<<<<<<<<<<<<<
 *             idx = s.parse_index(index_row)
 *             sp = s.spaces[idx.sid]
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/schema.pyx":331
 *             sp.add_index(idx)
 * 
 *         return s             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):  # pragma: nocover
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_s);
  __pyx_r = __pyx_v_s;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":313
 * 
 *     @staticmethod
 *     cdef Schema parse(int64_t schema_id, spaces, indexes):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Schema s
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.parse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_s);
  __Pyx_XDECREF((PyObject *)__pyx_v_sp);
  __Pyx_XDECREF((PyObject *)__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_space_row);
  __Pyx_XDECREF(__pyx_v_index_row);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":333
 *         return s
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return '<Schema spaces={}>'.format(len(self.spaces))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_6Schema_2__repr__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6Schema_2__repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "asynctnt/iproto/schema.pyx":334
 * 
 *     def __repr__(self):  # pragma: nocover
 *         return '<Schema spaces={}>'.format(len(self.spaces))             # <<<<<<<<<<<<<<
 * 
 * cdef list dict_to_list_fields(dict d, Metadata metadata, bint default_none):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Schema_spaces, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_v_self->spaces;
  __Pyx_INCREF(__pyx_t_3);
  if (unlikely(__pyx_t_3 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 334, __pyx_L1_error)
  }
  __pyx_t_4 = PyDict_Size(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":333
 *         return s
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return '<Schema spaces={}>'.format(len(self.spaces))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":63
 * cdef class Schema:
 *     cdef:
 *         readonly dict spaces             # <<<<<<<<<<<<<<
 *         readonly int id
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_6spaces_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_6spaces_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_6Schema_6spaces___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6Schema_6spaces___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->spaces);
  __pyx_r = __pyx_v_self->spaces;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pxd":64
 *     cdef:
 *         readonly dict spaces
 *         readonly int id             # <<<<<<<<<<<<<<
 * 
 *     cdef SchemaSpace get_space(self, space)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_2id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_2id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_6Schema_2id___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6Schema_2id___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->id); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_6Schema_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_6Schema_4__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6Schema_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_6Schema_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_6Schema_6__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6Schema_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Schema.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/schema.pyx":336
 *         return '<Schema spaces={}>'.format(len(self.spaces))
 * 
 * cdef list dict_to_list_fields(dict d, Metadata metadata, bint default_none):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list tpl
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_dict_to_list_fields(PyObject *__pyx_v_d, struct C_Metadata *__pyx_v_metadata, int __pyx_v_default_none) {
  PyObject *__pyx_v_tpl = 0;
  PyObject *__pyx_v_used = 0;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_field_name = 0;
  int32_t __pyx_v_used_count;
  int __pyx_v_field_id;
  PyObject *__pyx_v_obj_p;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dict_to_list_fields", 1);

  /* "asynctnt/iproto/schema.pyx":346
 *         PyObject *obj_p
 * 
 *     assert metadata is not None             # <<<<<<<<<<<<<<
 *     assert d is not None
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (((PyObject *)__pyx_v_metadata) != Py_None);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(1, 346, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(1, 346, __pyx_L1_error)
  #endif

  /* "asynctnt/iproto/schema.pyx":347
 * 
 *     assert metadata is not None
 *     assert d is not None             # <<<<<<<<<<<<<<
 * 
 *     tpl = []
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__pyx_v_d != ((PyObject*)Py_None));
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(1, 347, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(1, 347, __pyx_L1_error)
  #endif

  /* "asynctnt/iproto/schema.pyx":349
 *     assert d is not None
 * 
 *     tpl = []             # <<<<<<<<<<<<<<
 *     used_count = <int32_t> cpython.dict.PyDict_Size(d)
 * 
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_tpl = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/schema.pyx":350
 * 
 *     tpl = []
 *     used_count = <int32_t> cpython.dict.PyDict_Size(d)             # <<<<<<<<<<<<<<
 * 
 *     for field_id in range(metadata.len()):
 */
  __pyx_t_3 = PyDict_Size(__pyx_v_d); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(1, 350, __pyx_L1_error)
  __pyx_v_used_count = ((int32_t)__pyx_t_3);

  /* "asynctnt/iproto/schema.pyx":352
 *     used_count = <int32_t> cpython.dict.PyDict_Size(d)
 * 
 *     for field_id in range(metadata.len()):             # <<<<<<<<<<<<<<
 *         field_name = metadata.name_by_id(field_id)
 * 
 */
  __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_len(__pyx_v_metadata); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 352, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_field_id = __pyx_t_6;

    /* "asynctnt/iproto/schema.pyx":353
 * 
 *     for field_id in range(metadata.len()):
 *         field_name = metadata.name_by_id(field_id)             # <<<<<<<<<<<<<<
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(d, field_name)
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_name_by_id(__pyx_v_metadata, __pyx_v_field_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_field_name, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "asynctnt/iproto/schema.pyx":355
 *         field_name = metadata.name_by_id(field_id)
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(d, field_name)             # <<<<<<<<<<<<<<
 *         if obj_p is not NULL:
 *             value = <object> obj_p
 */
    __pyx_v_obj_p = PyDict_GetItem(__pyx_v_d, __pyx_v_field_name);

    /* "asynctnt/iproto/schema.pyx":356
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(d, field_name)
 *         if obj_p is not NULL:             # <<<<<<<<<<<<<<
 *             value = <object> obj_p
 *             used_count -= 1
 */
    __pyx_t_1 = (__pyx_v_obj_p != NULL);
    if (__pyx_t_1) {

      /* "asynctnt/iproto/schema.pyx":357
 *         obj_p = cpython.dict.PyDict_GetItem(d, field_name)
 *         if obj_p is not NULL:
 *             value = <object> obj_p             # <<<<<<<<<<<<<<
 *             used_count -= 1
 *             tpl.append(value)
 */
      __pyx_t_2 = ((PyObject *)__pyx_v_obj_p);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "asynctnt/iproto/schema.pyx":358
 *         if obj_p is not NULL:
 *             value = <object> obj_p
 *             used_count -= 1             # <<<<<<<<<<<<<<
 *             tpl.append(value)
 *         else:
 */
      __pyx_v_used_count = (__pyx_v_used_count - 1);

      /* "asynctnt/iproto/schema.pyx":359
 *             value = <object> obj_p
 *             used_count -= 1
 *             tpl.append(value)             # <<<<<<<<<<<<<<
 *         else:
 *             if default_none:
 */
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_tpl, __pyx_v_value); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 359, __pyx_L1_error)

      /* "asynctnt/iproto/schema.pyx":356
 * 
 *         obj_p = cpython.dict.PyDict_GetItem(d, field_name)
 *         if obj_p is not NULL:             # <<<<<<<<<<<<<<
 *             value = <object> obj_p
 *             used_count -= 1
 */
      goto __pyx_L5;
    }

    /* "asynctnt/iproto/schema.pyx":361
 *             tpl.append(value)
 *         else:
 *             if default_none:             # <<<<<<<<<<<<<<
 *                 tpl.append(None)
 * 
 */
    /*else*/ {
      if (__pyx_v_default_none) {

        /* "asynctnt/iproto/schema.pyx":362
 *         else:
 *             if default_none:
 *                 tpl.append(None)             # <<<<<<<<<<<<<<
 * 
 *     # Warn user if he used any of unknown fields
 */
        __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_tpl, Py_None); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 362, __pyx_L1_error)

        /* "asynctnt/iproto/schema.pyx":361
 *             tpl.append(value)
 *         else:
 *             if default_none:             # <<<<<<<<<<<<<<
 *                 tpl.append(None)
 * 
 */
      }
    }
    __pyx_L5:;
  }

  /* "asynctnt/iproto/schema.pyx":365
 * 
 *     # Warn user if he used any of unknown fields
 *     if used_count != 0:             # <<<<<<<<<<<<<<
 *         used = {}
 *         for field_id in range(metadata.len()):
 */
  __pyx_t_1 = (__pyx_v_used_count != 0);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/schema.pyx":366
 *     # Warn user if he used any of unknown fields
 *     if used_count != 0:
 *         used = {}             # <<<<<<<<<<<<<<
 *         for field_id in range(metadata.len()):
 *             field_name = metadata.name_by_id(field_id)
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_used = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynctnt/iproto/schema.pyx":367
 *     if used_count != 0:
 *         used = {}
 *         for field_id in range(metadata.len()):             # <<<<<<<<<<<<<<
 *             field_name = metadata.name_by_id(field_id)
 * 
 */
    __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_len(__pyx_v_metadata); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 367, __pyx_L1_error)
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_field_id = __pyx_t_6;

      /* "asynctnt/iproto/schema.pyx":368
 *         used = {}
 *         for field_id in range(metadata.len()):
 *             field_name = metadata.name_by_id(field_id)             # <<<<<<<<<<<<<<
 * 
 *             if <bint> cpython.dict.PyDict_Contains(d, field_name):
 */
      __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_name_by_id(__pyx_v_metadata, __pyx_v_field_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_field_name, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "asynctnt/iproto/schema.pyx":370
 *             field_name = metadata.name_by_id(field_id)
 * 
 *             if <bint> cpython.dict.PyDict_Contains(d, field_name):             # <<<<<<<<<<<<<<
 *                 used[field_name] = None
 *         if <bint> cpython.dict.PyDict_Contains(d, ''):
 */
      __pyx_t_8 = PyDict_Contains(__pyx_v_d, __pyx_v_field_name); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(1, 370, __pyx_L1_error)
      __pyx_t_1 = (__pyx_t_8 != 0);
      if (__pyx_t_1) {

        /* "asynctnt/iproto/schema.pyx":371
 * 
 *             if <bint> cpython.dict.PyDict_Contains(d, field_name):
 *                 used[field_name] = None             # <<<<<<<<<<<<<<
 *         if <bint> cpython.dict.PyDict_Contains(d, ''):
 *             used[''] = None
 */
        if (unlikely((PyDict_SetItem(__pyx_v_used, __pyx_v_field_name, Py_None) < 0))) __PYX_ERR(1, 371, __pyx_L1_error)

        /* "asynctnt/iproto/schema.pyx":370
 *             field_name = metadata.name_by_id(field_id)
 * 
 *             if <bint> cpython.dict.PyDict_Contains(d, field_name):             # <<<<<<<<<<<<<<
 *                 used[field_name] = None
 *         if <bint> cpython.dict.PyDict_Contains(d, ''):
 */
      }
    }

    /* "asynctnt/iproto/schema.pyx":372
 *             if <bint> cpython.dict.PyDict_Contains(d, field_name):
 *                 used[field_name] = None
 *         if <bint> cpython.dict.PyDict_Contains(d, ''):             # <<<<<<<<<<<<<<
 *             used[''] = None
 * 
 */
    __pyx_t_4 = PyDict_Contains(__pyx_v_d, __pyx_kp_u__3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 372, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_4 != 0);
    if (__pyx_t_1) {

      /* "asynctnt/iproto/schema.pyx":373
 *                 used[field_name] = None
 *         if <bint> cpython.dict.PyDict_Contains(d, ''):
 *             used[''] = None             # <<<<<<<<<<<<<<
 * 
 *         for f in d:
 */
      if (unlikely((PyDict_SetItem(__pyx_v_used, __pyx_kp_u__3, Py_None) < 0))) __PYX_ERR(1, 373, __pyx_L1_error)

      /* "asynctnt/iproto/schema.pyx":372
 *             if <bint> cpython.dict.PyDict_Contains(d, field_name):
 *                 used[field_name] = None
 *         if <bint> cpython.dict.PyDict_Contains(d, ''):             # <<<<<<<<<<<<<<
 *             used[''] = None
 * 
 */
    }

    /* "asynctnt/iproto/schema.pyx":375
 *             used[''] = None
 * 
 *         for f in d:             # <<<<<<<<<<<<<<
 *             if f not in used:
 *                 logger.warning(
 */
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_d == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(1, 375, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_dict_iterator(__pyx_v_d, 1, ((PyObject *)NULL), (&__pyx_t_9), (&__pyx_t_4)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_10;
    __pyx_t_10 = 0;
    while (1) {
      __pyx_t_5 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_9, &__pyx_t_3, &__pyx_t_10, NULL, NULL, __pyx_t_4);
      if (unlikely(__pyx_t_5 == 0)) break;
      if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(1, 375, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_XDECREF_SET(__pyx_v_f, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "asynctnt/iproto/schema.pyx":376
 * 
 *         for f in d:
 *             if f not in used:             # <<<<<<<<<<<<<<
 *                 logger.warning(
 *                     'Field \'%s\' in supplied dict is unknown as '
 */
      __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_f, __pyx_v_used, Py_NE)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 376, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "asynctnt/iproto/schema.pyx":377
 *         for f in d:
 *             if f not in used:
 *                 logger.warning(             # <<<<<<<<<<<<<<
 *                     'Field \'%s\' in supplied dict is unknown as '
 *                     'a tuple field for selected index. Skipping.',
 */
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_logger); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 377, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_warning); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 377, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "asynctnt/iproto/schema.pyx":380
 *                     'Field \'%s\' in supplied dict is unknown as '
 *                     'a tuple field for selected index. Skipping.',
 *                     f             # <<<<<<<<<<<<<<
 *                 )
 *     return tpl
 */
        __pyx_t_11 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_11, __pyx_kp_u_Field_s_in_supplied_dict_is_unkn, __pyx_v_f};
          __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 377, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "asynctnt/iproto/schema.pyx":376
 * 
 *         for f in d:
 *             if f not in used:             # <<<<<<<<<<<<<<
 *                 logger.warning(
 *                     'Field \'%s\' in supplied dict is unknown as '
 */
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynctnt/iproto/schema.pyx":365
 * 
 *     # Warn user if he used any of unknown fields
 *     if used_count != 0:             # <<<<<<<<<<<<<<
 *         used = {}
 *         for field_id in range(metadata.len()):
 */
  }

  /* "asynctnt/iproto/schema.pyx":382
 *                     f
 *                 )
 *     return tpl             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tpl);
  __pyx_r = __pyx_v_tpl;
  goto __pyx_L0;

  /* "asynctnt/iproto/schema.pyx":336
 *         return '<Schema spaces={}>'.format(len(self.spaces))
 * 
 * cdef list dict_to_list_fields(dict d, Metadata metadata, bint default_none):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list tpl
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.dict_to_list_fields", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tpl);
  __Pyx_XDECREF(__pyx_v_used);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_field_name);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/ext.pyx":10
 * 
 * 
 * cdef uint32_t decimal_len(int exponent, uint32_t digits_count):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t length
 */

static uint32_t __pyx_f_8asynctnt_6iproto_8protocol_decimal_len(int __pyx_v_exponent, uint32_t __pyx_v_digits_count) {
  uint32_t __pyx_v_length;
  uint32_t __pyx_r;
  uint32_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/ext.pyx":14
 *         uint32_t length
 * 
 *     length = bcd_len(digits_count)             # <<<<<<<<<<<<<<
 *     if exponent > 0:
 *         length += mp_sizeof_int(-exponent)
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_bcd_len(__pyx_v_digits_count); if (unlikely(__pyx_t_1 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 14, __pyx_L1_error)
  __pyx_v_length = __pyx_t_1;

  /* "asynctnt/iproto/ext.pyx":15
 * 
 *     length = bcd_len(digits_count)
 *     if exponent > 0:             # <<<<<<<<<<<<<<
 *         length += mp_sizeof_int(-exponent)
 *     else:
 */
  __pyx_t_2 = (__pyx_v_exponent > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/ext.pyx":16
 *     length = bcd_len(digits_count)
 *     if exponent > 0:
 *         length += mp_sizeof_int(-exponent)             # <<<<<<<<<<<<<<
 *     else:
 *         length += mp_sizeof_uint(-exponent)
 */
    __pyx_v_length = (__pyx_v_length + mp_sizeof_int((-__pyx_v_exponent)));

    /* "asynctnt/iproto/ext.pyx":15
 * 
 *     length = bcd_len(digits_count)
 *     if exponent > 0:             # <<<<<<<<<<<<<<
 *         length += mp_sizeof_int(-exponent)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/ext.pyx":18
 *         length += mp_sizeof_int(-exponent)
 *     else:
 *         length += mp_sizeof_uint(-exponent)             # <<<<<<<<<<<<<<
 * 
 *     return length
 */
  /*else*/ {
    __pyx_v_length = (__pyx_v_length + mp_sizeof_uint((-__pyx_v_exponent)));
  }
  __pyx_L3:;

  /* "asynctnt/iproto/ext.pyx":20
 *         length += mp_sizeof_uint(-exponent)
 * 
 *     return length             # <<<<<<<<<<<<<<
 * 
 * cdef char *decimal_encode(char *p,
 */
  __pyx_r = __pyx_v_length;
  goto __pyx_L0;

  /* "asynctnt/iproto/ext.pyx":10
 * 
 * 
 * cdef uint32_t decimal_len(int exponent, uint32_t digits_count):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t length
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.decimal_len", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/ext.pyx":22
 *     return length
 * 
 * cdef char *decimal_encode(char *p,             # <<<<<<<<<<<<<<
 *                           uint32_t digits_count,
 *                           uint8_t sign,
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_decimal_encode(char *__pyx_v_p, uint32_t __pyx_v_digits_count, uint8_t __pyx_v_sign, PyObject *__pyx_v_digits, int __pyx_v_exponent) {
  int __pyx_v_i;
  uint8_t __pyx_v_byte;
  char *__pyx_v_out;
  uint32_t __pyx_v_length;
  char *__pyx_r;
  int __pyx_t_1;
  uint32_t __pyx_t_2;
  PyObject *__pyx_t_3;
  uint8_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/ext.pyx":34
 * 
 *     # encode exponent
 *     if exponent > 0:             # <<<<<<<<<<<<<<
 *         p = mp_encode_int(p, -exponent)
 *     else:
 */
  __pyx_t_1 = (__pyx_v_exponent > 0);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/ext.pyx":35
 *     # encode exponent
 *     if exponent > 0:
 *         p = mp_encode_int(p, -exponent)             # <<<<<<<<<<<<<<
 *     else:
 *         p = mp_encode_uint(p, -exponent)
 */
    __pyx_v_p = mp_encode_int(__pyx_v_p, (-__pyx_v_exponent));

    /* "asynctnt/iproto/ext.pyx":34
 * 
 *     # encode exponent
 *     if exponent > 0:             # <<<<<<<<<<<<<<
 *         p = mp_encode_int(p, -exponent)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/ext.pyx":37
 *         p = mp_encode_int(p, -exponent)
 *     else:
 *         p = mp_encode_uint(p, -exponent)             # <<<<<<<<<<<<<<
 * 
 *     length = bcd_len(digits_count)
 */
  /*else*/ {
    __pyx_v_p = mp_encode_uint(__pyx_v_p, (-__pyx_v_exponent));
  }
  __pyx_L3:;

  /* "asynctnt/iproto/ext.pyx":39
 *         p = mp_encode_uint(p, -exponent)
 * 
 *     length = bcd_len(digits_count)             # <<<<<<<<<<<<<<
 * 
 *     out = &p[length - 1]
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_bcd_len(__pyx_v_digits_count); if (unlikely(__pyx_t_2 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 39, __pyx_L1_error)
  __pyx_v_length = __pyx_t_2;

  /* "asynctnt/iproto/ext.pyx":41
 *     length = bcd_len(digits_count)
 * 
 *     out = &p[length - 1]             # <<<<<<<<<<<<<<
 *     if sign == 1:
 *         byte = 0x0d
 */
  __pyx_v_out = (&(__pyx_v_p[(__pyx_v_length - 1)]));

  /* "asynctnt/iproto/ext.pyx":42
 * 
 *     out = &p[length - 1]
 *     if sign == 1:             # <<<<<<<<<<<<<<
 *         byte = 0x0d
 *     else:
 */
  __pyx_t_1 = (__pyx_v_sign == 1);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/ext.pyx":43
 *     out = &p[length - 1]
 *     if sign == 1:
 *         byte = 0x0d             # <<<<<<<<<<<<<<
 *     else:
 *         byte = 0x0c
 */
    __pyx_v_byte = 0x0d;

    /* "asynctnt/iproto/ext.pyx":42
 * 
 *     out = &p[length - 1]
 *     if sign == 1:             # <<<<<<<<<<<<<<
 *         byte = 0x0d
 *     else:
 */
    goto __pyx_L4;
  }

  /* "asynctnt/iproto/ext.pyx":45
 *         byte = 0x0d
 *     else:
 *         byte = 0x0c             # <<<<<<<<<<<<<<
 * 
 *     i = digits_count - 1
 */
  /*else*/ {
    __pyx_v_byte = 0x0c;
  }
  __pyx_L4:;

  /* "asynctnt/iproto/ext.pyx":47
 *         byte = 0x0c
 * 
 *     i = digits_count - 1             # <<<<<<<<<<<<<<
 *     while out >= p:
 *         if i >= 0:
 */
  __pyx_v_i = (__pyx_v_digits_count - 1);

  /* "asynctnt/iproto/ext.pyx":48
 * 
 *     i = digits_count - 1
 *     while out >= p:             # <<<<<<<<<<<<<<
 *         if i >= 0:
 *             byte |= (<uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i)) << 4
 */
  while (1) {
    __pyx_t_1 = (__pyx_v_out >= __pyx_v_p);
    if (!__pyx_t_1) break;

    /* "asynctnt/iproto/ext.pyx":49
 *     i = digits_count - 1
 *     while out >= p:
 *         if i >= 0:             # <<<<<<<<<<<<<<
 *             byte |= (<uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i)) << 4
 * 
 */
    __pyx_t_1 = (__pyx_v_i >= 0);
    if (__pyx_t_1) {

      /* "asynctnt/iproto/ext.pyx":50
 *     while out >= p:
 *         if i >= 0:
 *             byte |= (<uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i)) << 4             # <<<<<<<<<<<<<<
 * 
 *         out[0] = byte
 */
      __pyx_t_3 = PyTuple_GET_ITEM(__pyx_v_digits, __pyx_v_i);
      __pyx_t_4 = __Pyx_PyInt_As_uint8_t(((PyObject *)__pyx_t_3)); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(2, 50, __pyx_L1_error)
      __pyx_v_byte = (__pyx_v_byte | (((uint8_t)__pyx_t_4) << 4));

      /* "asynctnt/iproto/ext.pyx":49
 *     i = digits_count - 1
 *     while out >= p:
 *         if i >= 0:             # <<<<<<<<<<<<<<
 *             byte |= (<uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i)) << 4
 * 
 */
    }

    /* "asynctnt/iproto/ext.pyx":52
 *             byte |= (<uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i)) << 4
 * 
 *         out[0] = byte             # <<<<<<<<<<<<<<
 *         byte = 0
 * 
 */
    (__pyx_v_out[0]) = __pyx_v_byte;

    /* "asynctnt/iproto/ext.pyx":53
 * 
 *         out[0] = byte
 *         byte = 0             # <<<<<<<<<<<<<<
 * 
 *         if i > 0:
 */
    __pyx_v_byte = 0;

    /* "asynctnt/iproto/ext.pyx":55
 *         byte = 0
 * 
 *         if i > 0:             # <<<<<<<<<<<<<<
 *             byte = <uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i - 1) & 0xf
 * 
 */
    __pyx_t_1 = (__pyx_v_i > 0);
    if (__pyx_t_1) {

      /* "asynctnt/iproto/ext.pyx":56
 * 
 *         if i > 0:
 *             byte = <uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i - 1) & 0xf             # <<<<<<<<<<<<<<
 * 
 *         out -= 1
 */
      __pyx_t_3 = PyTuple_GET_ITEM(__pyx_v_digits, (__pyx_v_i - 1));
      __pyx_t_4 = __Pyx_PyInt_As_uint8_t(((PyObject *)__pyx_t_3)); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(2, 56, __pyx_L1_error)
      __pyx_v_byte = (((uint8_t)__pyx_t_4) & 0xf);

      /* "asynctnt/iproto/ext.pyx":55
 *         byte = 0
 * 
 *         if i > 0:             # <<<<<<<<<<<<<<
 *             byte = <uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i - 1) & 0xf
 * 
 */
    }

    /* "asynctnt/iproto/ext.pyx":58
 *             byte = <uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i - 1) & 0xf
 * 
 *         out -= 1             # <<<<<<<<<<<<<<
 *         i -= 2
 * 
 */
    __pyx_v_out = (__pyx_v_out - 1);

    /* "asynctnt/iproto/ext.pyx":59
 * 
 *         out -= 1
 *         i -= 2             # <<<<<<<<<<<<<<
 * 
 *     p = &p[length]
 */
    __pyx_v_i = (__pyx_v_i - 2);
  }

  /* "asynctnt/iproto/ext.pyx":61
 *         i -= 2
 * 
 *     p = &p[length]             # <<<<<<<<<<<<<<
 *     return p
 * 
 */
  __pyx_v_p = (&(__pyx_v_p[__pyx_v_length]));

  /* "asynctnt/iproto/ext.pyx":62
 * 
 *     p = &p[length]
 *     return p             # <<<<<<<<<<<<<<
 * 
 * cdef object decimal_decode(const char ** p, uint32_t length):
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/ext.pyx":22
 *     return length
 * 
 * cdef char *decimal_encode(char *p,             # <<<<<<<<<<<<<<
 *                           uint32_t digits_count,
 *                           uint8_t sign,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.decimal_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/ext.pyx":64
 *     return p
 * 
 * cdef object decimal_decode(const char ** p, uint32_t length):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int exponent
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_decimal_decode(char const **__pyx_v_p, uint32_t __pyx_v_length) {
  int __pyx_v_exponent;
  uint8_t __pyx_v_sign;
  enum mp_type __pyx_v_obj_type;
  char const *__pyx_v_first;
  char const *__pyx_v_last;
  uint32_t __pyx_v_digits_count;
  uint8_t __pyx_v_nibble;
  PyObject *__pyx_v_digits = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  long __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decimal_decode", 1);

  /* "asynctnt/iproto/ext.pyx":74
 *         uint8_t nibble
 * 
 *     sign = 0             # <<<<<<<<<<<<<<
 *     first = &p[0][0]
 *     last = first + length - 1
 */
  __pyx_v_sign = 0;

  /* "asynctnt/iproto/ext.pyx":75
 * 
 *     sign = 0
 *     first = &p[0][0]             # <<<<<<<<<<<<<<
 *     last = first + length - 1
 * 
 */
  __pyx_v_first = (&((__pyx_v_p[0])[0]));

  /* "asynctnt/iproto/ext.pyx":76
 *     sign = 0
 *     first = &p[0][0]
 *     last = first + length - 1             # <<<<<<<<<<<<<<
 * 
 *     # decode exponent
 */
  __pyx_v_last = ((__pyx_v_first + __pyx_v_length) - 1);

  /* "asynctnt/iproto/ext.pyx":79
 * 
 *     # decode exponent
 *     obj_type = mp_typeof(p[0][0])             # <<<<<<<<<<<<<<
 *     if obj_type == MP_UINT:
 *         exponent = -<int> mp_decode_uint(p)
 */
  __pyx_v_obj_type = mp_typeof(((__pyx_v_p[0])[0]));

  /* "asynctnt/iproto/ext.pyx":80
 *     # decode exponent
 *     obj_type = mp_typeof(p[0][0])
 *     if obj_type == MP_UINT:             # <<<<<<<<<<<<<<
 *         exponent = -<int> mp_decode_uint(p)
 *     elif obj_type == MP_INT:
 */
  switch (__pyx_v_obj_type) {
    case MP_UINT:

    /* "asynctnt/iproto/ext.pyx":81
 *     obj_type = mp_typeof(p[0][0])
 *     if obj_type == MP_UINT:
 *         exponent = -<int> mp_decode_uint(p)             # <<<<<<<<<<<<<<
 *     elif obj_type == MP_INT:
 *         exponent = -<int> mp_decode_int(p)
 */
    __pyx_v_exponent = (-((int)mp_decode_uint(__pyx_v_p)));

    /* "asynctnt/iproto/ext.pyx":80
 *     # decode exponent
 *     obj_type = mp_typeof(p[0][0])
 *     if obj_type == MP_UINT:             # <<<<<<<<<<<<<<
 *         exponent = -<int> mp_decode_uint(p)
 *     elif obj_type == MP_INT:
 */
    break;
    case MP_INT:

    /* "asynctnt/iproto/ext.pyx":83
 *         exponent = -<int> mp_decode_uint(p)
 *     elif obj_type == MP_INT:
 *         exponent = -<int> mp_decode_int(p)             # <<<<<<<<<<<<<<
 *     else:
 *         raise TypeError('unexpected exponent type: {}'.format(obj_type))
 */
    __pyx_v_exponent = (-((int)mp_decode_int(__pyx_v_p)));

    /* "asynctnt/iproto/ext.pyx":82
 *     if obj_type == MP_UINT:
 *         exponent = -<int> mp_decode_uint(p)
 *     elif obj_type == MP_INT:             # <<<<<<<<<<<<<<
 *         exponent = -<int> mp_decode_int(p)
 *     else:
 */
    break;
    default:

    /* "asynctnt/iproto/ext.pyx":85
 *         exponent = -<int> mp_decode_int(p)
 *     else:
 *         raise TypeError('unexpected exponent type: {}'.format(obj_type))             # <<<<<<<<<<<<<<
 * 
 *     length -= (&p[0][0] - first)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_unexpected_exponent_type, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_enum__mp_type(__pyx_v_obj_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 85, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 85, __pyx_L1_error)
    break;
  }

  /* "asynctnt/iproto/ext.pyx":87
 *         raise TypeError('unexpected exponent type: {}'.format(obj_type))
 * 
 *     length -= (&p[0][0] - first)             # <<<<<<<<<<<<<<
 *     first = &p[0][0]
 * 
 */
  __pyx_v_length = (__pyx_v_length - ((&((__pyx_v_p[0])[0])) - __pyx_v_first));

  /* "asynctnt/iproto/ext.pyx":88
 * 
 *     length -= (&p[0][0] - first)
 *     first = &p[0][0]             # <<<<<<<<<<<<<<
 * 
 *     while first[0] == 0:
 */
  __pyx_v_first = (&((__pyx_v_p[0])[0]));

  /* "asynctnt/iproto/ext.pyx":90
 *     first = &p[0][0]
 * 
 *     while first[0] == 0:             # <<<<<<<<<<<<<<
 *         first += 1  # skipping leading zeros
 * 
 */
  while (1) {
    __pyx_t_6 = ((__pyx_v_first[0]) == 0);
    if (!__pyx_t_6) break;

    /* "asynctnt/iproto/ext.pyx":91
 * 
 *     while first[0] == 0:
 *         first += 1  # skipping leading zeros             # <<<<<<<<<<<<<<
 * 
 *     sign = last[0] & 0xf  # extract sign
 */
    __pyx_v_first = (__pyx_v_first + 1);
  }

  /* "asynctnt/iproto/ext.pyx":93
 *         first += 1  # skipping leading zeros
 * 
 *     sign = last[0] & 0xf  # extract sign             # <<<<<<<<<<<<<<
 *     if sign == 0x0a or sign == 0x0c or sign == 0x0e or sign == 0x0f:
 *         sign = 0
 */
  __pyx_v_sign = ((__pyx_v_last[0]) & 0xf);

  /* "asynctnt/iproto/ext.pyx":94
 * 
 *     sign = last[0] & 0xf  # extract sign
 *     if sign == 0x0a or sign == 0x0c or sign == 0x0e or sign == 0x0f:             # <<<<<<<<<<<<<<
 *         sign = 0
 *     else:
 */
  switch (__pyx_v_sign) {
    case 0x0a:
    case 0x0c:
    case 0x0e:
    case 0x0f:

    /* "asynctnt/iproto/ext.pyx":95
 *     sign = last[0] & 0xf  # extract sign
 *     if sign == 0x0a or sign == 0x0c or sign == 0x0e or sign == 0x0f:
 *         sign = 0             # <<<<<<<<<<<<<<
 *     else:
 *         sign = 1
 */
    __pyx_v_sign = 0;

    /* "asynctnt/iproto/ext.pyx":94
 * 
 *     sign = last[0] & 0xf  # extract sign
 *     if sign == 0x0a or sign == 0x0c or sign == 0x0e or sign == 0x0f:             # <<<<<<<<<<<<<<
 *         sign = 0
 *     else:
 */
    break;
    default:

    /* "asynctnt/iproto/ext.pyx":97
 *         sign = 0
 *     else:
 *         sign = 1             # <<<<<<<<<<<<<<
 * 
 *     # decode digits
 */
    __pyx_v_sign = 1;
    break;
  }

  /* "asynctnt/iproto/ext.pyx":100
 * 
 *     # decode digits
 *     digits_count = (last - first) * 2 + 1             # <<<<<<<<<<<<<<
 *     if first[0] & 0xf0 == 0:
 *         digits_count -= 1  # adjust for leading zero nibble
 */
  __pyx_v_digits_count = (((__pyx_v_last - __pyx_v_first) * 2) + 1);

  /* "asynctnt/iproto/ext.pyx":101
 *     # decode digits
 *     digits_count = (last - first) * 2 + 1
 *     if first[0] & 0xf0 == 0:             # <<<<<<<<<<<<<<
 *         digits_count -= 1  # adjust for leading zero nibble
 * 
 */
  __pyx_t_6 = (((__pyx_v_first[0]) & 0xf0) == 0);
  if (__pyx_t_6) {

    /* "asynctnt/iproto/ext.pyx":102
 *     digits_count = (last - first) * 2 + 1
 *     if first[0] & 0xf0 == 0:
 *         digits_count -= 1  # adjust for leading zero nibble             # <<<<<<<<<<<<<<
 * 
 *     digits = cpython.tuple.PyTuple_New(digits_count)
 */
    __pyx_v_digits_count = (__pyx_v_digits_count - 1);

    /* "asynctnt/iproto/ext.pyx":101
 *     # decode digits
 *     digits_count = (last - first) * 2 + 1
 *     if first[0] & 0xf0 == 0:             # <<<<<<<<<<<<<<
 *         digits_count -= 1  # adjust for leading zero nibble
 * 
 */
  }

  /* "asynctnt/iproto/ext.pyx":104
 *         digits_count -= 1  # adjust for leading zero nibble
 * 
 *     digits = cpython.tuple.PyTuple_New(digits_count)             # <<<<<<<<<<<<<<
 * 
 *     if digits_count > 0:
 */
  __pyx_t_2 = PyTuple_New(__pyx_v_digits_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_digits = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/ext.pyx":106
 *     digits = cpython.tuple.PyTuple_New(digits_count)
 * 
 *     if digits_count > 0:             # <<<<<<<<<<<<<<
 *         while True:
 *             nibble = (last[0] & 0xf0) >> 4  # left nibble first
 */
  __pyx_t_6 = (__pyx_v_digits_count > 0);
  if (__pyx_t_6) {

    /* "asynctnt/iproto/ext.pyx":107
 * 
 *     if digits_count > 0:
 *         while True:             # <<<<<<<<<<<<<<
 *             nibble = (last[0] & 0xf0) >> 4  # left nibble first
 *             item = <object> <int> nibble
 */
    while (1) {

      /* "asynctnt/iproto/ext.pyx":108
 *     if digits_count > 0:
 *         while True:
 *             nibble = (last[0] & 0xf0) >> 4  # left nibble first             # <<<<<<<<<<<<<<
 *             item = <object> <int> nibble
 *             cpython.Py_INCREF(item)
 */
      __pyx_v_nibble = (((__pyx_v_last[0]) & 0xf0) >> 4);

      /* "asynctnt/iproto/ext.pyx":109
 *         while True:
 *             nibble = (last[0] & 0xf0) >> 4  # left nibble first
 *             item = <object> <int> nibble             # <<<<<<<<<<<<<<
 *             cpython.Py_INCREF(item)
 *             cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)
 */
      __pyx_t_2 = __Pyx_PyInt_From_int(((int)__pyx_v_nibble)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __pyx_t_2;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "asynctnt/iproto/ext.pyx":110
 *             nibble = (last[0] & 0xf0) >> 4  # left nibble first
 *             item = <object> <int> nibble
 *             cpython.Py_INCREF(item)             # <<<<<<<<<<<<<<
 *             cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)
 * 
 */
      Py_INCREF(__pyx_v_item);

      /* "asynctnt/iproto/ext.pyx":111
 *             item = <object> <int> nibble
 *             cpython.Py_INCREF(item)
 *             cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)             # <<<<<<<<<<<<<<
 * 
 *             digits_count -= 1
 */
      PyTuple_SET_ITEM(__pyx_v_digits, (__pyx_v_digits_count - 1), __pyx_v_item);

      /* "asynctnt/iproto/ext.pyx":113
 *             cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)
 * 
 *             digits_count -= 1             # <<<<<<<<<<<<<<
 *             if digits_count == 0:
 *                 break
 */
      __pyx_v_digits_count = (__pyx_v_digits_count - 1);

      /* "asynctnt/iproto/ext.pyx":114
 * 
 *             digits_count -= 1
 *             if digits_count == 0:             # <<<<<<<<<<<<<<
 *                 break
 *             last -= 1
 */
      __pyx_t_6 = (__pyx_v_digits_count == 0);
      if (__pyx_t_6) {

        /* "asynctnt/iproto/ext.pyx":115
 *             digits_count -= 1
 *             if digits_count == 0:
 *                 break             # <<<<<<<<<<<<<<
 *             last -= 1
 * 
 */
        goto __pyx_L8_break;

        /* "asynctnt/iproto/ext.pyx":114
 * 
 *             digits_count -= 1
 *             if digits_count == 0:             # <<<<<<<<<<<<<<
 *                 break
 *             last -= 1
 */
      }

      /* "asynctnt/iproto/ext.pyx":116
 *             if digits_count == 0:
 *                 break
 *             last -= 1             # <<<<<<<<<<<<<<
 * 
 *             nibble = last[0] & 0x0f  # right nibble
 */
      __pyx_v_last = (__pyx_v_last - 1);

      /* "asynctnt/iproto/ext.pyx":118
 *             last -= 1
 * 
 *             nibble = last[0] & 0x0f  # right nibble             # <<<<<<<<<<<<<<
 *             item = <object> <int> nibble
 *             cpython.Py_INCREF(item)
 */
      __pyx_v_nibble = ((__pyx_v_last[0]) & 0x0f);

      /* "asynctnt/iproto/ext.pyx":119
 * 
 *             nibble = last[0] & 0x0f  # right nibble
 *             item = <object> <int> nibble             # <<<<<<<<<<<<<<
 *             cpython.Py_INCREF(item)
 *             cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(((int)__pyx_v_nibble)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_item, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "asynctnt/iproto/ext.pyx":120
 *             nibble = last[0] & 0x0f  # right nibble
 *             item = <object> <int> nibble
 *             cpython.Py_INCREF(item)             # <<<<<<<<<<<<<<
 *             cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)
 * 
 */
      Py_INCREF(__pyx_v_item);

      /* "asynctnt/iproto/ext.pyx":121
 *             item = <object> <int> nibble
 *             cpython.Py_INCREF(item)
 *             cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)             # <<<<<<<<<<<<<<
 * 
 *             digits_count -= 1
 */
      PyTuple_SET_ITEM(__pyx_v_digits, (__pyx_v_digits_count - 1), __pyx_v_item);

      /* "asynctnt/iproto/ext.pyx":123
 *             cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)
 * 
 *             digits_count -= 1             # <<<<<<<<<<<<<<
 *             if digits_count == 0:
 *                 break
 */
      __pyx_v_digits_count = (__pyx_v_digits_count - 1);

      /* "asynctnt/iproto/ext.pyx":124
 * 
 *             digits_count -= 1
 *             if digits_count == 0:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      __pyx_t_6 = (__pyx_v_digits_count == 0);
      if (__pyx_t_6) {

        /* "asynctnt/iproto/ext.pyx":125
 *             digits_count -= 1
 *             if digits_count == 0:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *     p[0] += length
 */
        goto __pyx_L8_break;

        /* "asynctnt/iproto/ext.pyx":124
 * 
 *             digits_count -= 1
 *             if digits_count == 0:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      }
    }
    __pyx_L8_break:;

    /* "asynctnt/iproto/ext.pyx":106
 *     digits = cpython.tuple.PyTuple_New(digits_count)
 * 
 *     if digits_count > 0:             # <<<<<<<<<<<<<<
 *         while True:
 *             nibble = (last[0] & 0xf0) >> 4  # left nibble first
 */
  }

  /* "asynctnt/iproto/ext.pyx":127
 *                 break
 * 
 *     p[0] += length             # <<<<<<<<<<<<<<
 * 
 *     return Decimal((<object> <int> sign, digits, <object> exponent))
 */
  __pyx_t_7 = 0;
  (__pyx_v_p[__pyx_t_7]) = ((__pyx_v_p[__pyx_t_7]) + __pyx_v_length);

  /* "asynctnt/iproto/ext.pyx":129
 *     p[0] += length
 * 
 *     return Decimal((<object> <int> sign, digits, <object> exponent))             # <<<<<<<<<<<<<<
 * 
 * cdef object uuid_decode(const char ** p, uint32_t length):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_int(((int)__pyx_v_sign)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_exponent); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_t_3));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_3));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3)) __PYX_ERR(2, 129, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_digits);
  __Pyx_GIVEREF(__pyx_v_digits);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_digits)) __PYX_ERR(2, 129, __pyx_L1_error);
  __Pyx_INCREF(((PyObject *)__pyx_t_4));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_4));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_4)) __PYX_ERR(2, 129, __pyx_L1_error);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_8};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/ext.pyx":64
 *     return p
 * 
 * cdef object decimal_decode(const char ** p, uint32_t length):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int exponent
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.decimal_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_digits);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/ext.pyx":131
 *     return Decimal((<object> <int> sign, digits, <object> exponent))
 * 
 * cdef object uuid_decode(const char ** p, uint32_t length):             # <<<<<<<<<<<<<<
 *     data = cpython.bytes.PyBytes_FromStringAndSize(p[0], length)
 *     p[0] += length
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_uuid_decode(char const **__pyx_v_p, uint32_t __pyx_v_length) {
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("uuid_decode", 1);

  /* "asynctnt/iproto/ext.pyx":132
 * 
 * cdef object uuid_decode(const char ** p, uint32_t length):
 *     data = cpython.bytes.PyBytes_FromStringAndSize(p[0], length)             # <<<<<<<<<<<<<<
 *     p[0] += length
 *     return UUID(bytes=data)
 */
  __pyx_t_1 = PyBytes_FromStringAndSize((__pyx_v_p[0]), __pyx_v_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/ext.pyx":133
 * cdef object uuid_decode(const char ** p, uint32_t length):
 *     data = cpython.bytes.PyBytes_FromStringAndSize(p[0], length)
 *     p[0] += length             # <<<<<<<<<<<<<<
 *     return UUID(bytes=data)
 * 
 */
  __pyx_t_2 = 0;
  (__pyx_v_p[__pyx_t_2]) = ((__pyx_v_p[__pyx_t_2]) + __pyx_v_length);

  /* "asynctnt/iproto/ext.pyx":134
 *     data = cpython.bytes.PyBytes_FromStringAndSize(p[0], length)
 *     p[0] += length
 *     return UUID(bytes=data)             # <<<<<<<<<<<<<<
 * 
 * cdef inline void datetime_zero(IProtoDateTime *dt):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_UUID); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_bytes, __pyx_v_data) < 0) __PYX_ERR(2, 134, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/ext.pyx":131
 *     return Decimal((<object> <int> sign, digits, <object> exponent))
 * 
 * cdef object uuid_decode(const char ** p, uint32_t length):             # <<<<<<<<<<<<<<
 *     data = cpython.bytes.PyBytes_FromStringAndSize(p[0], length)
 *     p[0] += length
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.uuid_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/ext.pyx":136
 *     return UUID(bytes=data)
 * 
 * cdef inline void datetime_zero(IProtoDateTime *dt):             # <<<<<<<<<<<<<<
 *     dt.seconds = 0
 *     dt.nsec = 0
 */

static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_datetime_zero(struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *__pyx_v_dt) {

  /* "asynctnt/iproto/ext.pyx":137
 * 
 * cdef inline void datetime_zero(IProtoDateTime *dt):
 *     dt.seconds = 0             # <<<<<<<<<<<<<<
 *     dt.nsec = 0
 *     dt.tzoffset = 0
 */
  __pyx_v_dt->seconds = 0;

  /* "asynctnt/iproto/ext.pyx":138
 * cdef inline void datetime_zero(IProtoDateTime *dt):
 *     dt.seconds = 0
 *     dt.nsec = 0             # <<<<<<<<<<<<<<
 *     dt.tzoffset = 0
 *     dt.tzindex = 0
 */
  __pyx_v_dt->nsec = 0;

  /* "asynctnt/iproto/ext.pyx":139
 *     dt.seconds = 0
 *     dt.nsec = 0
 *     dt.tzoffset = 0             # <<<<<<<<<<<<<<
 *     dt.tzindex = 0
 * 
 */
  __pyx_v_dt->tzoffset = 0;

  /* "asynctnt/iproto/ext.pyx":140
 *     dt.nsec = 0
 *     dt.tzoffset = 0
 *     dt.tzindex = 0             # <<<<<<<<<<<<<<
 * 
 * cdef inline uint32_t datetime_len(IProtoDateTime *dt):
 */
  __pyx_v_dt->tzindex = 0;

  /* "asynctnt/iproto/ext.pyx":136
 *     return UUID(bytes=data)
 * 
 * cdef inline void datetime_zero(IProtoDateTime *dt):             # <<<<<<<<<<<<<<
 *     dt.seconds = 0
 *     dt.nsec = 0
 */

  /* function exit code */
}

/* "asynctnt/iproto/ext.pyx":142
 *     dt.tzindex = 0
 * 
 * cdef inline uint32_t datetime_len(IProtoDateTime *dt):             # <<<<<<<<<<<<<<
 *     cdef uint32_t sz
 *     sz = sizeof(int64_t)
 */

static CYTHON_INLINE uint32_t __pyx_f_8asynctnt_6iproto_8protocol_datetime_len(struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *__pyx_v_dt) {
  uint32_t __pyx_v_sz;
  uint32_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "asynctnt/iproto/ext.pyx":144
 * cdef inline uint32_t datetime_len(IProtoDateTime *dt):
 *     cdef uint32_t sz
 *     sz = sizeof(int64_t)             # <<<<<<<<<<<<<<
 *     if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:
 *         return sz + DATETIME_TAIL_SZ
 */
  __pyx_v_sz = (sizeof(int64_t));

  /* "asynctnt/iproto/ext.pyx":145
 *     cdef uint32_t sz
 *     sz = sizeof(int64_t)
 *     if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:             # <<<<<<<<<<<<<<
 *         return sz + DATETIME_TAIL_SZ
 *     return sz
 */
  __pyx_t_2 = (__pyx_v_dt->nsec != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_dt->tzoffset != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_dt->tzindex != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "asynctnt/iproto/ext.pyx":146
 *     sz = sizeof(int64_t)
 *     if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:
 *         return sz + DATETIME_TAIL_SZ             # <<<<<<<<<<<<<<
 *     return sz
 * 
 */
    __pyx_r = (__pyx_v_sz + 8);
    goto __pyx_L0;

    /* "asynctnt/iproto/ext.pyx":145
 *     cdef uint32_t sz
 *     sz = sizeof(int64_t)
 *     if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:             # <<<<<<<<<<<<<<
 *         return sz + DATETIME_TAIL_SZ
 *     return sz
 */
  }

  /* "asynctnt/iproto/ext.pyx":147
 *     if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:
 *         return sz + DATETIME_TAIL_SZ
 *     return sz             # <<<<<<<<<<<<<<
 * 
 * cdef char *datetime_encode(char *p, IProtoDateTime *dt) except NULL:
 */
  __pyx_r = __pyx_v_sz;
  goto __pyx_L0;

  /* "asynctnt/iproto/ext.pyx":142
 *     dt.tzindex = 0
 * 
 * cdef inline uint32_t datetime_len(IProtoDateTime *dt):             # <<<<<<<<<<<<<<
 *     cdef uint32_t sz
 *     sz = sizeof(int64_t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/ext.pyx":149
 *     return sz
 * 
 * cdef char *datetime_encode(char *p, IProtoDateTime *dt) except NULL:             # <<<<<<<<<<<<<<
 *     store_u64(p, dt.seconds)
 *     p += sizeof(dt.seconds)
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_datetime_encode(char *__pyx_v_p, struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *__pyx_v_dt) {
  char *__pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/ext.pyx":150
 * 
 * cdef char *datetime_encode(char *p, IProtoDateTime *dt) except NULL:
 *     store_u64(p, dt.seconds)             # <<<<<<<<<<<<<<
 *     p += sizeof(dt.seconds)
 *     if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:
 */
  __pyx_f_8asynctnt_6iproto_8protocol_store_u64(__pyx_v_p, __pyx_v_dt->seconds); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 150, __pyx_L1_error)

  /* "asynctnt/iproto/ext.pyx":151
 * cdef char *datetime_encode(char *p, IProtoDateTime *dt) except NULL:
 *     store_u64(p, dt.seconds)
 *     p += sizeof(dt.seconds)             # <<<<<<<<<<<<<<
 *     if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:
 *         memcpy(p, &dt.nsec, DATETIME_TAIL_SZ)
 */
  __pyx_v_p = (__pyx_v_p + (sizeof(__pyx_v_dt->seconds)));

  /* "asynctnt/iproto/ext.pyx":152
 *     store_u64(p, dt.seconds)
 *     p += sizeof(dt.seconds)
 *     if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:             # <<<<<<<<<<<<<<
 *         memcpy(p, &dt.nsec, DATETIME_TAIL_SZ)
 *         p += DATETIME_TAIL_SZ
 */
  __pyx_t_2 = (__pyx_v_dt->nsec != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_dt->tzoffset != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_dt->tzindex != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "asynctnt/iproto/ext.pyx":153
 *     p += sizeof(dt.seconds)
 *     if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:
 *         memcpy(p, &dt.nsec, DATETIME_TAIL_SZ)             # <<<<<<<<<<<<<<
 *         p += DATETIME_TAIL_SZ
 *     return p
 */
    (void)(memcpy(__pyx_v_p, (&__pyx_v_dt->nsec), 8));

    /* "asynctnt/iproto/ext.pyx":154
 *     if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:
 *         memcpy(p, &dt.nsec, DATETIME_TAIL_SZ)
 *         p += DATETIME_TAIL_SZ             # <<<<<<<<<<<<<<
 *     return p
 * 
 */
    __pyx_v_p = (__pyx_v_p + 8);

    /* "asynctnt/iproto/ext.pyx":152
 *     store_u64(p, dt.seconds)
 *     p += sizeof(dt.seconds)
 *     if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:             # <<<<<<<<<<<<<<
 *         memcpy(p, &dt.nsec, DATETIME_TAIL_SZ)
 *         p += DATETIME_TAIL_SZ
 */
  }

  /* "asynctnt/iproto/ext.pyx":155
 *         memcpy(p, &dt.nsec, DATETIME_TAIL_SZ)
 *         p += DATETIME_TAIL_SZ
 *     return p             # <<<<<<<<<<<<<<
 * 
 * cdef int datetime_decode(
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/ext.pyx":149
 *     return sz
 * 
 * cdef char *datetime_encode(char *p, IProtoDateTime *dt) except NULL:             # <<<<<<<<<<<<<<
 *     store_u64(p, dt.seconds)
 *     p += sizeof(dt.seconds)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.datetime_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/ext.pyx":157
 *     return p
 * 
 * cdef int datetime_decode(             # <<<<<<<<<<<<<<
 *         const char ** p,
 *         uint32_t length,
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_datetime_decode(char const **__pyx_v_p, uint32_t __pyx_v_length, struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *__pyx_v_dt) {
  CYTHON_UNUSED PyObject *__pyx_v_delta = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_tz = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  uint64_t __pyx_t_1;
  long __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("datetime_decode", 1);

  /* "asynctnt/iproto/ext.pyx":162
 *         IProtoDateTime *dt
 * ) except -1:
 *     delta = None             # <<<<<<<<<<<<<<
 *     tz = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_delta = Py_None;

  /* "asynctnt/iproto/ext.pyx":163
 * ) except -1:
 *     delta = None
 *     tz = None             # <<<<<<<<<<<<<<
 * 
 *     dt.seconds = load_u64(p[0])
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_tz = Py_None;

  /* "asynctnt/iproto/ext.pyx":165
 *     tz = None
 * 
 *     dt.seconds = load_u64(p[0])             # <<<<<<<<<<<<<<
 *     p[0] += sizeof(dt.seconds)
 *     length -= sizeof(dt.seconds)
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_load_u64((__pyx_v_p[0])); if (unlikely(__pyx_t_1 == ((uint64_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 165, __pyx_L1_error)
  __pyx_v_dt->seconds = __pyx_t_1;

  /* "asynctnt/iproto/ext.pyx":166
 * 
 *     dt.seconds = load_u64(p[0])
 *     p[0] += sizeof(dt.seconds)             # <<<<<<<<<<<<<<
 *     length -= sizeof(dt.seconds)
 * 
 */
  __pyx_t_2 = 0;
  (__pyx_v_p[__pyx_t_2]) = ((__pyx_v_p[__pyx_t_2]) + (sizeof(__pyx_v_dt->seconds)));

  /* "asynctnt/iproto/ext.pyx":167
 *     dt.seconds = load_u64(p[0])
 *     p[0] += sizeof(dt.seconds)
 *     length -= sizeof(dt.seconds)             # <<<<<<<<<<<<<<
 * 
 *     if length == 0:
 */
  __pyx_v_length = (__pyx_v_length - (sizeof(__pyx_v_dt->seconds)));

  /* "asynctnt/iproto/ext.pyx":169
 *     length -= sizeof(dt.seconds)
 * 
 *     if length == 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_3 = (__pyx_v_length == 0);
  if (__pyx_t_3) {

    /* "asynctnt/iproto/ext.pyx":170
 * 
 *     if length == 0:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     if length != DATETIME_TAIL_SZ:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/ext.pyx":169
 *     length -= sizeof(dt.seconds)
 * 
 *     if length == 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  }

  /* "asynctnt/iproto/ext.pyx":172
 *         return 0
 * 
 *     if length != DATETIME_TAIL_SZ:             # <<<<<<<<<<<<<<
 *         raise ValueError("invalid datetime size. got {} extra bytes".format(
 *             length
 */
  __pyx_t_3 = (__pyx_v_length != 8);
  if (unlikely(__pyx_t_3)) {

    /* "asynctnt/iproto/ext.pyx":173
 * 
 *     if length != DATETIME_TAIL_SZ:
 *         raise ValueError("invalid datetime size. got {} extra bytes".format(             # <<<<<<<<<<<<<<
 *             length
 *         ))
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_datetime_size_got_extra, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "asynctnt/iproto/ext.pyx":174
 *     if length != DATETIME_TAIL_SZ:
 *         raise ValueError("invalid datetime size. got {} extra bytes".format(
 *             length             # <<<<<<<<<<<<<<
 *         ))
 * 
 */
    __pyx_t_6 = __Pyx_PyInt_From_uint32_t(__pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }

    /* "asynctnt/iproto/ext.pyx":173
 * 
 *     if length != DATETIME_TAIL_SZ:
 *         raise ValueError("invalid datetime size. got {} extra bytes".format(             # <<<<<<<<<<<<<<
 *             length
 *         ))
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(2, 173, __pyx_L1_error)

    /* "asynctnt/iproto/ext.pyx":172
 *         return 0
 * 
 *     if length != DATETIME_TAIL_SZ:             # <<<<<<<<<<<<<<
 *         raise ValueError("invalid datetime size. got {} extra bytes".format(
 *             length
 */
  }

  /* "asynctnt/iproto/ext.pyx":177
 *         ))
 * 
 *     dt.nsec = load_u32(p[0])             # <<<<<<<<<<<<<<
 *     p[0] += 4
 *     dt.tzoffset = load_u16(p[0])
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_load_u32((__pyx_v_p[0])); if (unlikely(__pyx_t_1 == ((uint64_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 177, __pyx_L1_error)
  __pyx_v_dt->nsec = __pyx_t_1;

  /* "asynctnt/iproto/ext.pyx":178
 * 
 *     dt.nsec = load_u32(p[0])
 *     p[0] += 4             # <<<<<<<<<<<<<<
 *     dt.tzoffset = load_u16(p[0])
 *     p[0] += 2
 */
  __pyx_t_2 = 0;
  (__pyx_v_p[__pyx_t_2]) = ((__pyx_v_p[__pyx_t_2]) + 4);

  /* "asynctnt/iproto/ext.pyx":179
 *     dt.nsec = load_u32(p[0])
 *     p[0] += 4
 *     dt.tzoffset = load_u16(p[0])             # <<<<<<<<<<<<<<
 *     p[0] += 2
 *     dt.tzindex = load_u16(p[0])
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_load_u16((__pyx_v_p[0])); if (unlikely(__pyx_t_1 == ((uint64_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 179, __pyx_L1_error)
  __pyx_v_dt->tzoffset = __pyx_t_1;

  /* "asynctnt/iproto/ext.pyx":180
 *     p[0] += 4
 *     dt.tzoffset = load_u16(p[0])
 *     p[0] += 2             # <<<<<<<<<<<<<<
 *     dt.tzindex = load_u16(p[0])
 *     p[0] += 2
 */
  __pyx_t_2 = 0;
  (__pyx_v_p[__pyx_t_2]) = ((__pyx_v_p[__pyx_t_2]) + 2);

  /* "asynctnt/iproto/ext.pyx":181
 *     dt.tzoffset = load_u16(p[0])
 *     p[0] += 2
 *     dt.tzindex = load_u16(p[0])             # <<<<<<<<<<<<<<
 *     p[0] += 2
 * 
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_load_u16((__pyx_v_p[0])); if (unlikely(__pyx_t_1 == ((uint64_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 181, __pyx_L1_error)
  __pyx_v_dt->tzindex = __pyx_t_1;

  /* "asynctnt/iproto/ext.pyx":182
 *     p[0] += 2
 *     dt.tzindex = load_u16(p[0])
 *     p[0] += 2             # <<<<<<<<<<<<<<
 * 
 * cdef void datetime_from_py(datetime ob, IProtoDateTime *dt):
 */
  __pyx_t_2 = 0;
  (__pyx_v_p[__pyx_t_2]) = ((__pyx_v_p[__pyx_t_2]) + 2);

  /* "asynctnt/iproto/ext.pyx":157
 *     return p
 * 
 * cdef int datetime_decode(             # <<<<<<<<<<<<<<
 *         const char ** p,
 *         uint32_t length,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.datetime_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_delta);
  __Pyx_XDECREF(__pyx_v_tz);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/ext.pyx":184
 *     p[0] += 2
 * 
 * cdef void datetime_from_py(datetime ob, IProtoDateTime *dt):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double ts
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_datetime_from_py(PyDateTime_DateTime *__pyx_v_ob, struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *__pyx_v_dt) {
  double __pyx_v_ts;
  int __pyx_v_offset;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  double __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("datetime_from_py", 1);

  /* "asynctnt/iproto/ext.pyx":188
 *         double ts
 *         int offset
 *     ts = <double> ob.timestamp()             # <<<<<<<<<<<<<<
 *     dt.seconds = <int64_t> ts
 *     dt.nsec = <int32_t> ((ts - <double> dt.seconds) * 1000000) * 1000
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_ob), __pyx_n_s_timestamp); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ts = ((double)__pyx_t_5);

  /* "asynctnt/iproto/ext.pyx":189
 *         int offset
 *     ts = <double> ob.timestamp()
 *     dt.seconds = <int64_t> ts             # <<<<<<<<<<<<<<
 *     dt.nsec = <int32_t> ((ts - <double> dt.seconds) * 1000000) * 1000
 *     if dt.nsec < 0:
 */
  __pyx_v_dt->seconds = ((int64_t)__pyx_v_ts);

  /* "asynctnt/iproto/ext.pyx":190
 *     ts = <double> ob.timestamp()
 *     dt.seconds = <int64_t> ts
 *     dt.nsec = <int32_t> ((ts - <double> dt.seconds) * 1000000) * 1000             # <<<<<<<<<<<<<<
 *     if dt.nsec < 0:
 *         # correction for negative dates
 */
  __pyx_v_dt->nsec = (((int32_t)((__pyx_v_ts - ((double)__pyx_v_dt->seconds)) * 1000000.0)) * 0x3E8);

  /* "asynctnt/iproto/ext.pyx":191
 *     dt.seconds = <int64_t> ts
 *     dt.nsec = <int32_t> ((ts - <double> dt.seconds) * 1000000) * 1000
 *     if dt.nsec < 0:             # <<<<<<<<<<<<<<
 *         # correction for negative dates
 *         dt.seconds -= 1
 */
  __pyx_t_6 = (__pyx_v_dt->nsec < 0);
  if (__pyx_t_6) {

    /* "asynctnt/iproto/ext.pyx":193
 *     if dt.nsec < 0:
 *         # correction for negative dates
 *         dt.seconds -= 1             # <<<<<<<<<<<<<<
 *         dt.nsec += 1000000000
 * 
 */
    __pyx_v_dt->seconds = (__pyx_v_dt->seconds - 1);

    /* "asynctnt/iproto/ext.pyx":194
 *         # correction for negative dates
 *         dt.seconds -= 1
 *         dt.nsec += 1000000000             # <<<<<<<<<<<<<<
 * 
 *     if datetime_tzinfo(ob) is not None:
 */
    __pyx_v_dt->nsec = (__pyx_v_dt->nsec + 0x3B9ACA00);

    /* "asynctnt/iproto/ext.pyx":191
 *     dt.seconds = <int64_t> ts
 *     dt.nsec = <int32_t> ((ts - <double> dt.seconds) * 1000000) * 1000
 *     if dt.nsec < 0:             # <<<<<<<<<<<<<<
 *         # correction for negative dates
 *         dt.seconds -= 1
 */
  }

  /* "asynctnt/iproto/ext.pyx":196
 *         dt.nsec += 1000000000
 * 
 *     if datetime_tzinfo(ob) is not None:             # <<<<<<<<<<<<<<
 *         offset = ob.utcoffset().total_seconds()
 *         dt.tzoffset = <int16_t> (offset / 60)
 */
  __pyx_t_1 = __pyx_f_7cpython_8datetime_datetime_tzinfo(((PyObject *)__pyx_v_ob)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "asynctnt/iproto/ext.pyx":197
 * 
 *     if datetime_tzinfo(ob) is not None:
 *         offset = ob.utcoffset().total_seconds()             # <<<<<<<<<<<<<<
 *         dt.tzoffset = <int16_t> (offset / 60)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_ob), __pyx_n_s_utcoffset); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_total_seconds); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 197, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_offset = __pyx_t_4;

    /* "asynctnt/iproto/ext.pyx":198
 *     if datetime_tzinfo(ob) is not None:
 *         offset = ob.utcoffset().total_seconds()
 *         dt.tzoffset = <int16_t> (offset / 60)             # <<<<<<<<<<<<<<
 * 
 * cdef object datetime_to_py(IProtoDateTime *dt):
 */
    __pyx_v_dt->tzoffset = ((int16_t)(((double)__pyx_v_offset) / 60.0));

    /* "asynctnt/iproto/ext.pyx":196
 *         dt.nsec += 1000000000
 * 
 *     if datetime_tzinfo(ob) is not None:             # <<<<<<<<<<<<<<
 *         offset = ob.utcoffset().total_seconds()
 *         dt.tzoffset = <int16_t> (offset / 60)
 */
  }

  /* "asynctnt/iproto/ext.pyx":184
 *     p[0] += 2
 * 
 * cdef void datetime_from_py(datetime ob, IProtoDateTime *dt):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double ts
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.datetime_from_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/ext.pyx":200
 *         dt.tzoffset = <int16_t> (offset / 60)
 * 
 * cdef object datetime_to_py(IProtoDateTime *dt):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double timestamp
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_datetime_to_py(struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *__pyx_v_dt) {
  double __pyx_v_timestamp;
  PyObject *__pyx_v_tz = 0;
  PyDateTime_Delta *__pyx_v_delta = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("datetime_to_py", 1);

  /* "asynctnt/iproto/ext.pyx":205
 *         object tz
 * 
 *     tz = None             # <<<<<<<<<<<<<<
 * 
 *     if dt.tzoffset != 0:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_tz = Py_None;

  /* "asynctnt/iproto/ext.pyx":207
 *     tz = None
 * 
 *     if dt.tzoffset != 0:             # <<<<<<<<<<<<<<
 *         delta = timedelta_new(0, <int> dt.tzoffset * 60, 0)
 *         tz = timezone_new(delta)
 */
  __pyx_t_1 = (__pyx_v_dt->tzoffset != 0);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/ext.pyx":208
 * 
 *     if dt.tzoffset != 0:
 *         delta = timedelta_new(0, <int> dt.tzoffset * 60, 0)             # <<<<<<<<<<<<<<
 *         tz = timezone_new(delta)
 * 
 */
    __pyx_t_2 = ((PyObject *)__pyx_f_7cpython_8datetime_timedelta_new(0, (((int)__pyx_v_dt->tzoffset) * 60), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_delta = ((PyDateTime_Delta *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynctnt/iproto/ext.pyx":209
 *     if dt.tzoffset != 0:
 *         delta = timedelta_new(0, <int> dt.tzoffset * 60, 0)
 *         tz = timezone_new(delta)             # <<<<<<<<<<<<<<
 * 
 *     timestamp = dt.seconds + (<double> dt.nsec) / 1e9
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_timezone_new(((PyObject *)__pyx_v_delta)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_tz, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynctnt/iproto/ext.pyx":207
 *     tz = None
 * 
 *     if dt.tzoffset != 0:             # <<<<<<<<<<<<<<
 *         delta = timedelta_new(0, <int> dt.tzoffset * 60, 0)
 *         tz = timezone_new(delta)
 */
  }

  /* "asynctnt/iproto/ext.pyx":211
 *         tz = timezone_new(delta)
 * 
 *     timestamp = dt.seconds + (<double> dt.nsec) / 1e9             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject *>PyDateTimeAPI.DateTimeType,
 */
  __pyx_v_timestamp = (__pyx_v_dt->seconds + (((double)__pyx_v_dt->nsec) / 1e9));

  /* "asynctnt/iproto/ext.pyx":212
 * 
 *     timestamp = dt.seconds + (<double> dt.nsec) / 1e9
 *     return PyDateTimeAPI.DateTime_FromTimestamp(             # <<<<<<<<<<<<<<
 *         <PyObject *>PyDateTimeAPI.DateTimeType,
 *         (timestamp,) if tz is None else (timestamp, tz),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "asynctnt/iproto/ext.pyx":214
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject *>PyDateTimeAPI.DateTimeType,
 *         (timestamp,) if tz is None else (timestamp, tz),             # <<<<<<<<<<<<<<
 *         NULL,
 *     )
 */
  __pyx_t_1 = (__pyx_v_tz == Py_None);
  if (__pyx_t_1) {
    __pyx_t_3 = PyFloat_FromDouble(__pyx_v_timestamp); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(2, 214, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_2 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __pyx_t_4 = PyFloat_FromDouble(__pyx_v_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(2, 214, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_tz);
    __Pyx_GIVEREF(__pyx_v_tz);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_tz)) __PYX_ERR(2, 214, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;
  }

  /* "asynctnt/iproto/ext.pyx":212
 * 
 *     timestamp = dt.seconds + (<double> dt.nsec) / 1e9
 *     return PyDateTimeAPI.DateTime_FromTimestamp(             # <<<<<<<<<<<<<<
 *         <PyObject *>PyDateTimeAPI.DateTimeType,
 *         (timestamp,) if tz is None else (timestamp, tz),
 */
  __pyx_t_3 = ((PyObject *)PyDateTimeAPI->DateTime_FromTimestamp(((PyObject *)PyDateTimeAPI->DateTimeType), __pyx_t_2, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/ext.pyx":200
 *         dt.tzoffset = <int16_t> (offset / 60)
 * 
 * cdef object datetime_to_py(IProtoDateTime *dt):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double timestamp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.datetime_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tz);
  __Pyx_XDECREF((PyObject *)__pyx_v_delta);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":25
 * @cython.freelist(_BUFFER_FREELIST_SIZE)
 * cdef class WriteBuffer:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._smallbuf_inuse = True
 *         self._buf = self._smallbuf
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer___cinit__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "asynctnt/iproto/buffer.pyx":26
 * cdef class WriteBuffer:
 *     def __cinit__(self):
 *         self._smallbuf_inuse = True             # <<<<<<<<<<<<<<
 *         self._buf = self._smallbuf
 *         self._size = _BUFFER_INITIAL_SIZE
 */
  __pyx_v_self->_smallbuf_inuse = 1;

  /* "asynctnt/iproto/buffer.pyx":27
 *     def __cinit__(self):
 *         self._smallbuf_inuse = True
 *         self._buf = self._smallbuf             # <<<<<<<<<<<<<<
 *         self._size = _BUFFER_INITIAL_SIZE
 *         self._length = 0
 */
  __pyx_t_1 = __pyx_v_self->_smallbuf;
  __pyx_v_self->_buf = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":28
 *         self._smallbuf_inuse = True
 *         self._buf = self._smallbuf
 *         self._size = _BUFFER_INITIAL_SIZE             # <<<<<<<<<<<<<<
 *         self._length = 0
 *         self._encoding = None
 */
  __pyx_v_self->_size = 0x400;

  /* "asynctnt/iproto/buffer.pyx":29
 *         self._buf = self._smallbuf
 *         self._size = _BUFFER_INITIAL_SIZE
 *         self._length = 0             # <<<<<<<<<<<<<<
 *         self._encoding = None
 * 
 */
  __pyx_v_self->_length = 0;

  /* "asynctnt/iproto/buffer.pyx":30
 *         self._size = _BUFFER_INITIAL_SIZE
 *         self._length = 0
 *         self._encoding = None             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_encoding);
  __Pyx_DECREF(__pyx_v_self->_encoding);
  __pyx_v_self->_encoding = ((PyObject*)Py_None);

  /* "asynctnt/iproto/buffer.pyx":25
 * @cython.freelist(_BUFFER_FREELIST_SIZE)
 * cdef class WriteBuffer:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._smallbuf_inuse = True
 *         self._buf = self._smallbuf
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":32
 *         self._encoding = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._buf is not NULL and not self._smallbuf_inuse:
 *             PyMem_Free(self._buf)
 */

/* Python wrapper */
static void __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_2__dealloc__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_2__dealloc__(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "asynctnt/iproto/buffer.pyx":33
 * 
 *     def __dealloc__(self):
 *         if self._buf is not NULL and not self._smallbuf_inuse:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self._buf)
 *             self._buf = NULL
 */
  __pyx_t_2 = (__pyx_v_self->_buf != NULL);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (!__pyx_v_self->_smallbuf_inuse);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":34
 *     def __dealloc__(self):
 *         if self._buf is not NULL and not self._smallbuf_inuse:
 *             PyMem_Free(self._buf)             # <<<<<<<<<<<<<<
 *             self._buf = NULL
 *             self._size = 0
 */
    PyMem_Free(__pyx_v_self->_buf);

    /* "asynctnt/iproto/buffer.pyx":35
 *         if self._buf is not NULL and not self._smallbuf_inuse:
 *             PyMem_Free(self._buf)
 *             self._buf = NULL             # <<<<<<<<<<<<<<
 *             self._size = 0
 * 
 */
    __pyx_v_self->_buf = NULL;

    /* "asynctnt/iproto/buffer.pyx":36
 *             PyMem_Free(self._buf)
 *             self._buf = NULL
 *             self._size = 0             # <<<<<<<<<<<<<<
 * 
 *         if self._view_count:  # pragma: nocover
 */
    __pyx_v_self->_size = 0;

    /* "asynctnt/iproto/buffer.pyx":33
 * 
 *     def __dealloc__(self):
 *         if self._buf is not NULL and not self._smallbuf_inuse:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self._buf)
 *             self._buf = NULL
 */
  }

  /* "asynctnt/iproto/buffer.pyx":38
 *             self._size = 0
 * 
 *         if self._view_count:  # pragma: nocover             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'Deallocating buffer with attached memoryviews')
 */
  __pyx_t_1 = (__pyx_v_self->_view_count != 0);
  if (unlikely(__pyx_t_1)) {

    /* "asynctnt/iproto/buffer.pyx":39
 * 
 *         if self._view_count:  # pragma: nocover
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'Deallocating buffer with attached memoryviews')
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 39, __pyx_L1_error)

    /* "asynctnt/iproto/buffer.pyx":38
 *             self._size = 0
 * 
 *         if self._view_count:  # pragma: nocover             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'Deallocating buffer with attached memoryviews')
 */
  }

  /* "asynctnt/iproto/buffer.pyx":32
 *         self._encoding = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._buf is not NULL and not self._smallbuf_inuse:
 *             PyMem_Free(self._buf)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("asynctnt.iproto.protocol.WriteBuffer.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/buffer.pyx":42
 *                 'Deallocating buffer with attached memoryviews')
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 *         self._view_count += 1
 * 
 */

/* Python wrapper */
CYTHON_UNUSED static int __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_5__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
CYTHON_UNUSED static int __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_5__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_4__getbuffer__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_4__getbuffer__(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (unlikely(__pyx_v_buffer == NULL)) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "asynctnt/iproto/buffer.pyx":43
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 *         self._view_count += 1             # <<<<<<<<<<<<<<
 * 
 *         cpython.PyBuffer_FillInfo(
 */
  __pyx_v_self->_view_count = (__pyx_v_self->_view_count + 1);

  /* "asynctnt/iproto/buffer.pyx":45
 *         self._view_count += 1
 * 
 *         cpython.PyBuffer_FillInfo(             # <<<<<<<<<<<<<<
 *             buffer, self, self._buf, self._length,
 *             1,  # read-only
 */
  __pyx_t_1 = PyBuffer_FillInfo(__pyx_v_buffer, ((PyObject *)__pyx_v_self), __pyx_v_self->_buf, __pyx_v_self->_length, 1, __pyx_v_flags); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 45, __pyx_L1_error)

  /* "asynctnt/iproto/buffer.pyx":42
 *                 'Deallocating buffer with attached memoryviews')
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 *         self._view_count += 1
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_buffer->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":51
 *         )
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         self._view_count -= 1
 * 
 */

/* Python wrapper */
CYTHON_UNUSED static void __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_7__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
CYTHON_UNUSED static void __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_7__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_6__releasebuffer__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_6__releasebuffer__(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {

  /* "asynctnt/iproto/buffer.pyx":52
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 *         self._view_count -= 1             # <<<<<<<<<<<<<<
 * 
 *     def hex(self):  # pragma: nocover
 */
  __pyx_v_self->_view_count = (__pyx_v_self->_view_count - 1);

  /* "asynctnt/iproto/buffer.pyx":51
 *         )
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         self._view_count -= 1
 * 
 */

  /* function exit code */
}

/* "asynctnt/iproto/buffer.pyx":54
 *         self._view_count -= 1
 * 
 *     def hex(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))
 *                         for i in range(self._length))
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_9hex(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11WriteBuffer_9hex = {"hex", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_9hex, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_9hex(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hex (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("hex", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "hex", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_8hex(((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8asynctnt_6iproto_8protocol_11WriteBuffer_3hex_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "asynctnt/iproto/buffer.pyx":55
 * 
 *     def hex(self):  # pragma: nocover
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))             # <<<<<<<<<<<<<<
 *                         for i in range(self._length))
 * 
 */

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_3hex_genexpr(PyObject *__pyx_self, Py_ssize_t __pyx_genexpr_arg_0) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(3, 55, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8asynctnt_6iproto_8protocol_11WriteBuffer_3hex_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_hex_locals_genexpr, __pyx_n_s_asynctnt_iproto_protocol); if (unlikely(!gen)) __PYX_ERR(3, 55, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.hex.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8asynctnt_6iproto_8protocol_11WriteBuffer_3hex_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(3, 55, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(3, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);

  /* "asynctnt/iproto/buffer.pyx":56
 *     def hex(self):  # pragma: nocover
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))
 *                         for i in range(self._length))             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_cur_scope->__pyx_v_i = __pyx_t_3;

    /* "asynctnt/iproto/buffer.pyx":55
 * 
 *     def hex(self):  # pragma: nocover
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))             # <<<<<<<<<<<<<<
 *                         for i in range(self._length))
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_02x, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(3, 55, __pyx_L1_error) }
    __pyx_t_6 = PyBytes_FromStringAndSize(&(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->_buf[__pyx_cur_scope->__pyx_v_i]), 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_t_6); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(3, 55, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyInt_From_long(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 55, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_4))) __PYX_ERR(3, 55, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":54
 *         self._view_count -= 1
 * 
 *     def hex(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))
 *                         for i in range(self._length))
 */

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_8hex(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *__pyx_cur_scope;
  PyObject *__pyx_gb_8asynctnt_6iproto_8protocol_11WriteBuffer_3hex_2generator1 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hex", 0);
  __pyx_cur_scope = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(3, 54, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "asynctnt/iproto/buffer.pyx":55
 * 
 *     def hex(self):  # pragma: nocover
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))             # <<<<<<<<<<<<<<
 *                         for i in range(self._length))
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "asynctnt/iproto/buffer.pyx":56
 *     def hex(self):  # pragma: nocover
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))
 *                         for i in range(self._length))             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_t_1 = __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_3hex_genexpr(((PyObject*)__pyx_cur_scope), __pyx_cur_scope->__pyx_v_self->_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "asynctnt/iproto/buffer.pyx":55
 * 
 *     def hex(self):  # pragma: nocover
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))             # <<<<<<<<<<<<<<
 *                         for i in range(self._length))
 * 
 */
  __pyx_t_2 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":54
 *         self._view_count -= 1
 * 
 *     def hex(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))
 *                         for i in range(self._length))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.hex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_8asynctnt_6iproto_8protocol_11WriteBuffer_3hex_2generator1);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":59
 * 
 *     @staticmethod
 *     cdef WriteBuffer create(bytes encoding):             # <<<<<<<<<<<<<<
 *         cdef WriteBuffer buf
 *         buf = WriteBuffer.__new__(WriteBuffer)
 */

static struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_create(PyObject *__pyx_v_encoding) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buf = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create", 1);

  /* "asynctnt/iproto/buffer.pyx":61
 *     cdef WriteBuffer create(bytes encoding):
 *         cdef WriteBuffer buf
 *         buf = WriteBuffer.__new__(WriteBuffer)             # <<<<<<<<<<<<<<
 *         buf._encoding = encoding
 *         return buf
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_WriteBuffer(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 61, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_buf = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/buffer.pyx":62
 *         cdef WriteBuffer buf
 *         buf = WriteBuffer.__new__(WriteBuffer)
 *         buf._encoding = encoding             # <<<<<<<<<<<<<<
 *         return buf
 * 
 */
  __Pyx_INCREF(__pyx_v_encoding);
  __Pyx_GIVEREF(__pyx_v_encoding);
  __Pyx_GOTREF(__pyx_v_buf->_encoding);
  __Pyx_DECREF(__pyx_v_buf->_encoding);
  __pyx_v_buf->_encoding = __pyx_v_encoding;

  /* "asynctnt/iproto/buffer.pyx":63
 *         buf = WriteBuffer.__new__(WriteBuffer)
 *         buf._encoding = encoding
 *         return buf             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _check_readonly(self):  # pragma: nocover
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_buf);
  __pyx_r = __pyx_v_buf;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":59
 * 
 *     @staticmethod
 *     cdef WriteBuffer create(bytes encoding):             # <<<<<<<<<<<<<<
 *         cdef WriteBuffer buf
 *         buf = WriteBuffer.__new__(WriteBuffer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_buf);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":65
 *         return buf
 * 
 *     cdef inline _check_readonly(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         if self._view_count:
 *             raise BufferError('the buffer is in read-only mode')
 */

static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__check_readonly(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_readonly", 1);

  /* "asynctnt/iproto/buffer.pyx":66
 * 
 *     cdef inline _check_readonly(self):  # pragma: nocover
 *         if self._view_count:             # <<<<<<<<<<<<<<
 *             raise BufferError('the buffer is in read-only mode')
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_view_count != 0);
  if (unlikely(__pyx_t_1)) {

    /* "asynctnt/iproto/buffer.pyx":67
 *     cdef inline _check_readonly(self):  # pragma: nocover
 *         if self._view_count:
 *             raise BufferError('the buffer is in read-only mode')             # <<<<<<<<<<<<<<
 * 
 *     cdef inline len(self):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_BufferError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(3, 67, __pyx_L1_error)

    /* "asynctnt/iproto/buffer.pyx":66
 * 
 *     cdef inline _check_readonly(self):  # pragma: nocover
 *         if self._view_count:             # <<<<<<<<<<<<<<
 *             raise BufferError('the buffer is in read-only mode')
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":65
 *         return buf
 * 
 *     cdef inline _check_readonly(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         if self._view_count:
 *             raise BufferError('the buffer is in read-only mode')
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer._check_readonly", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":69
 *             raise BufferError('the buffer is in read-only mode')
 * 
 *     cdef inline len(self):             # <<<<<<<<<<<<<<
 *         return self._length
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_len(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("len", 1);

  /* "asynctnt/iproto/buffer.pyx":70
 * 
 *     cdef inline len(self):
 *         return self._length             # <<<<<<<<<<<<<<
 * 
 *     cdef int ensure_allocated(self, ssize_t extra_length) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":69
 *             raise BufferError('the buffer is in read-only mode')
 * 
 *     cdef inline len(self):             # <<<<<<<<<<<<<<
 *         return self._length
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.len", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":72
 *         return self._length
 * 
 *     cdef int ensure_allocated(self, ssize_t extra_length) except -1:             # <<<<<<<<<<<<<<
 *         cdef ssize_t new_size = extra_length + self._length
 * 
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, Py_ssize_t __pyx_v_extra_length) {
  Py_ssize_t __pyx_v_new_size;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":73
 * 
 *     cdef int ensure_allocated(self, ssize_t extra_length) except -1:
 *         cdef ssize_t new_size = extra_length + self._length             # <<<<<<<<<<<<<<
 * 
 *         if new_size > self._size:
 */
  __pyx_v_new_size = (__pyx_v_extra_length + __pyx_v_self->_length);

  /* "asynctnt/iproto/buffer.pyx":75
 *         cdef ssize_t new_size = extra_length + self._length
 * 
 *         if new_size > self._size:             # <<<<<<<<<<<<<<
 *             self._reallocate(new_size)
 * 
 */
  __pyx_t_1 = (__pyx_v_new_size > __pyx_v_self->_size);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":76
 * 
 *         if new_size > self._size:
 *             self._reallocate(new_size)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *_ensure_allocated(self, char *p,
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__reallocate(__pyx_v_self, __pyx_v_new_size); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 76, __pyx_L1_error)

    /* "asynctnt/iproto/buffer.pyx":75
 *         cdef ssize_t new_size = extra_length + self._length
 * 
 *         if new_size > self._size:             # <<<<<<<<<<<<<<
 *             self._reallocate(new_size)
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":72
 *         return self._length
 * 
 *     cdef int ensure_allocated(self, ssize_t extra_length) except -1:             # <<<<<<<<<<<<<<
 *         cdef ssize_t new_size = extra_length + self._length
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.ensure_allocated", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":78
 *             self._reallocate(new_size)
 * 
 *     cdef char *_ensure_allocated(self, char *p,             # <<<<<<<<<<<<<<
 *                                  ssize_t extra_length) except NULL:
 *         cdef:
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, Py_ssize_t __pyx_v_extra_length) {
  Py_ssize_t __pyx_v_new_size;
  ptrdiff_t __pyx_v_used;
  char *__pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":83
 *             ssize_t new_size
 * 
 *         new_size = extra_length + self._length             # <<<<<<<<<<<<<<
 * 
 *         if new_size > self._size:
 */
  __pyx_v_new_size = (__pyx_v_extra_length + __pyx_v_self->_length);

  /* "asynctnt/iproto/buffer.pyx":85
 *         new_size = extra_length + self._length
 * 
 *         if new_size > self._size:             # <<<<<<<<<<<<<<
 *             used = p - self._buf
 *             self._reallocate(new_size)
 */
  __pyx_t_1 = (__pyx_v_new_size > __pyx_v_self->_size);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":86
 * 
 *         if new_size > self._size:
 *             used = p - self._buf             # <<<<<<<<<<<<<<
 *             self._reallocate(new_size)
 *             p = &self._buf[used]
 */
    __pyx_v_used = (__pyx_v_p - __pyx_v_self->_buf);

    /* "asynctnt/iproto/buffer.pyx":87
 *         if new_size > self._size:
 *             used = p - self._buf
 *             self._reallocate(new_size)             # <<<<<<<<<<<<<<
 *             p = &self._buf[used]
 *         return p
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__reallocate(__pyx_v_self, __pyx_v_new_size); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 87, __pyx_L1_error)

    /* "asynctnt/iproto/buffer.pyx":88
 *             used = p - self._buf
 *             self._reallocate(new_size)
 *             p = &self._buf[used]             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
    __pyx_v_p = (&(__pyx_v_self->_buf[__pyx_v_used]));

    /* "asynctnt/iproto/buffer.pyx":85
 *         new_size = extra_length + self._length
 * 
 *         if new_size > self._size:             # <<<<<<<<<<<<<<
 *             used = p - self._buf
 *             self._reallocate(new_size)
 */
  }

  /* "asynctnt/iproto/buffer.pyx":89
 *             self._reallocate(new_size)
 *             p = &self._buf[used]
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef int _reallocate(self, ssize_t new_size) except -1:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":78
 *             self._reallocate(new_size)
 * 
 *     cdef char *_ensure_allocated(self, char *p,             # <<<<<<<<<<<<<<
 *                                  ssize_t extra_length) except NULL:
 *         cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer._ensure_allocated", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":91
 *         return p
 * 
 *     cdef int _reallocate(self, ssize_t new_size) except -1:             # <<<<<<<<<<<<<<
 *         cdef char *new_buf
 * 
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__reallocate(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, Py_ssize_t __pyx_v_new_size) {
  char *__pyx_v_new_buf;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":94
 *         cdef char *new_buf
 * 
 *         if new_size < _BUFFER_MAX_GROW:             # <<<<<<<<<<<<<<
 *             new_size = _BUFFER_MAX_GROW
 *         else:
 */
  __pyx_t_1 = (__pyx_v_new_size < 0x10000);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":95
 * 
 *         if new_size < _BUFFER_MAX_GROW:
 *             new_size = _BUFFER_MAX_GROW             # <<<<<<<<<<<<<<
 *         else:
 *             # Add a little extra
 */
    __pyx_v_new_size = 0x10000;

    /* "asynctnt/iproto/buffer.pyx":94
 *         cdef char *new_buf
 * 
 *         if new_size < _BUFFER_MAX_GROW:             # <<<<<<<<<<<<<<
 *             new_size = _BUFFER_MAX_GROW
 *         else:
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/buffer.pyx":98
 *         else:
 *             # Add a little extra
 *             new_size += _BUFFER_INITIAL_SIZE             # <<<<<<<<<<<<<<
 * 
 *         if self._smallbuf_inuse:
 */
  /*else*/ {
    __pyx_v_new_size = (__pyx_v_new_size + 0x400);
  }
  __pyx_L3:;

  /* "asynctnt/iproto/buffer.pyx":100
 *             new_size += _BUFFER_INITIAL_SIZE
 * 
 *         if self._smallbuf_inuse:             # <<<<<<<<<<<<<<
 *             new_buf = <char*> PyMem_Malloc(sizeof(char) * <size_t> new_size)
 *             if new_buf is NULL:  # pragma: nocover
 */
  if (__pyx_v_self->_smallbuf_inuse) {

    /* "asynctnt/iproto/buffer.pyx":101
 * 
 *         if self._smallbuf_inuse:
 *             new_buf = <char*> PyMem_Malloc(sizeof(char) * <size_t> new_size)             # <<<<<<<<<<<<<<
 *             if new_buf is NULL:  # pragma: nocover
 *                 self._buf = NULL
 */
    __pyx_v_new_buf = ((char *)PyMem_Malloc(((sizeof(char)) * ((size_t)__pyx_v_new_size))));

    /* "asynctnt/iproto/buffer.pyx":102
 *         if self._smallbuf_inuse:
 *             new_buf = <char*> PyMem_Malloc(sizeof(char) * <size_t> new_size)
 *             if new_buf is NULL:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 self._buf = NULL
 *                 self._size = 0
 */
    __pyx_t_1 = (__pyx_v_new_buf == NULL);
    if (unlikely(__pyx_t_1)) {

      /* "asynctnt/iproto/buffer.pyx":103
 *             new_buf = <char*> PyMem_Malloc(sizeof(char) * <size_t> new_size)
 *             if new_buf is NULL:  # pragma: nocover
 *                 self._buf = NULL             # <<<<<<<<<<<<<<
 *                 self._size = 0
 *                 self._length = 0
 */
      __pyx_v_self->_buf = NULL;

      /* "asynctnt/iproto/buffer.pyx":104
 *             if new_buf is NULL:  # pragma: nocover
 *                 self._buf = NULL
 *                 self._size = 0             # <<<<<<<<<<<<<<
 *                 self._length = 0
 *                 raise MemoryError
 */
      __pyx_v_self->_size = 0;

      /* "asynctnt/iproto/buffer.pyx":105
 *                 self._buf = NULL
 *                 self._size = 0
 *                 self._length = 0             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             memcpy(new_buf, self._buf, <size_t> self._size)
 */
      __pyx_v_self->_length = 0;

      /* "asynctnt/iproto/buffer.pyx":106
 *                 self._size = 0
 *                 self._length = 0
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             memcpy(new_buf, self._buf, <size_t> self._size)
 *             self._size = new_size
 */
      PyErr_NoMemory(); __PYX_ERR(3, 106, __pyx_L1_error)

      /* "asynctnt/iproto/buffer.pyx":102
 *         if self._smallbuf_inuse:
 *             new_buf = <char*> PyMem_Malloc(sizeof(char) * <size_t> new_size)
 *             if new_buf is NULL:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 self._buf = NULL
 *                 self._size = 0
 */
    }

    /* "asynctnt/iproto/buffer.pyx":107
 *                 self._length = 0
 *                 raise MemoryError
 *             memcpy(new_buf, self._buf, <size_t> self._size)             # <<<<<<<<<<<<<<
 *             self._size = new_size
 *             self._buf = new_buf
 */
    (void)(memcpy(__pyx_v_new_buf, __pyx_v_self->_buf, ((size_t)__pyx_v_self->_size)));

    /* "asynctnt/iproto/buffer.pyx":108
 *                 raise MemoryError
 *             memcpy(new_buf, self._buf, <size_t> self._size)
 *             self._size = new_size             # <<<<<<<<<<<<<<
 *             self._buf = new_buf
 *             self._smallbuf_inuse = False
 */
    __pyx_v_self->_size = __pyx_v_new_size;

    /* "asynctnt/iproto/buffer.pyx":109
 *             memcpy(new_buf, self._buf, <size_t> self._size)
 *             self._size = new_size
 *             self._buf = new_buf             # <<<<<<<<<<<<<<
 *             self._smallbuf_inuse = False
 *         else:
 */
    __pyx_v_self->_buf = __pyx_v_new_buf;

    /* "asynctnt/iproto/buffer.pyx":110
 *             self._size = new_size
 *             self._buf = new_buf
 *             self._smallbuf_inuse = False             # <<<<<<<<<<<<<<
 *         else:
 *             new_buf = <char*> PyMem_Realloc(<void*> self._buf, <size_t> new_size)
 */
    __pyx_v_self->_smallbuf_inuse = 0;

    /* "asynctnt/iproto/buffer.pyx":100
 *             new_size += _BUFFER_INITIAL_SIZE
 * 
 *         if self._smallbuf_inuse:             # <<<<<<<<<<<<<<
 *             new_buf = <char*> PyMem_Malloc(sizeof(char) * <size_t> new_size)
 *             if new_buf is NULL:  # pragma: nocover
 */
    goto __pyx_L4;
  }

  /* "asynctnt/iproto/buffer.pyx":112
 *             self._smallbuf_inuse = False
 *         else:
 *             new_buf = <char*> PyMem_Realloc(<void*> self._buf, <size_t> new_size)             # <<<<<<<<<<<<<<
 *             if new_buf is NULL:
 *                 PyMem_Free(self._buf)
 */
  /*else*/ {
    __pyx_v_new_buf = ((char *)PyMem_Realloc(((void *)__pyx_v_self->_buf), ((size_t)__pyx_v_new_size)));

    /* "asynctnt/iproto/buffer.pyx":113
 *         else:
 *             new_buf = <char*> PyMem_Realloc(<void*> self._buf, <size_t> new_size)
 *             if new_buf is NULL:             # <<<<<<<<<<<<<<
 *                 PyMem_Free(self._buf)
 *                 self._buf = NULL
 */
    __pyx_t_1 = (__pyx_v_new_buf == NULL);
    if (unlikely(__pyx_t_1)) {

      /* "asynctnt/iproto/buffer.pyx":114
 *             new_buf = <char*> PyMem_Realloc(<void*> self._buf, <size_t> new_size)
 *             if new_buf is NULL:
 *                 PyMem_Free(self._buf)             # <<<<<<<<<<<<<<
 *                 self._buf = NULL
 *                 self._size = 0
 */
      PyMem_Free(__pyx_v_self->_buf);

      /* "asynctnt/iproto/buffer.pyx":115
 *             if new_buf is NULL:
 *                 PyMem_Free(self._buf)
 *                 self._buf = NULL             # <<<<<<<<<<<<<<
 *                 self._size = 0
 *                 self._length = 0
 */
      __pyx_v_self->_buf = NULL;

      /* "asynctnt/iproto/buffer.pyx":116
 *                 PyMem_Free(self._buf)
 *                 self._buf = NULL
 *                 self._size = 0             # <<<<<<<<<<<<<<
 *                 self._length = 0
 *                 raise MemoryError
 */
      __pyx_v_self->_size = 0;

      /* "asynctnt/iproto/buffer.pyx":117
 *                 self._buf = NULL
 *                 self._size = 0
 *                 self._length = 0             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             self._buf = new_buf
 */
      __pyx_v_self->_length = 0;

      /* "asynctnt/iproto/buffer.pyx":118
 *                 self._size = 0
 *                 self._length = 0
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             self._buf = new_buf
 *             self._size = new_size
 */
      PyErr_NoMemory(); __PYX_ERR(3, 118, __pyx_L1_error)

      /* "asynctnt/iproto/buffer.pyx":113
 *         else:
 *             new_buf = <char*> PyMem_Realloc(<void*> self._buf, <size_t> new_size)
 *             if new_buf is NULL:             # <<<<<<<<<<<<<<
 *                 PyMem_Free(self._buf)
 *                 self._buf = NULL
 */
    }

    /* "asynctnt/iproto/buffer.pyx":119
 *                 self._length = 0
 *                 raise MemoryError
 *             self._buf = new_buf             # <<<<<<<<<<<<<<
 *             self._size = new_size
 * 
 */
    __pyx_v_self->_buf = __pyx_v_new_buf;

    /* "asynctnt/iproto/buffer.pyx":120
 *                 raise MemoryError
 *             self._buf = new_buf
 *             self._size = new_size             # <<<<<<<<<<<<<<
 * 
 *     cdef int write_buffer(self, WriteBuffer buf) except -1:
 */
    __pyx_v_self->_size = __pyx_v_new_size;
  }
  __pyx_L4:;

  /* "asynctnt/iproto/buffer.pyx":91
 *         return p
 * 
 *     cdef int _reallocate(self, ssize_t new_size) except -1:             # <<<<<<<<<<<<<<
 *         cdef char *new_buf
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer._reallocate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":122
 *             self._size = new_size
 * 
 *     cdef int write_buffer(self, WriteBuffer buf) except -1:             # <<<<<<<<<<<<<<
 *         if not buf._length:
 *             return 0
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_buffer(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buf) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":123
 * 
 *     cdef int write_buffer(self, WriteBuffer buf) except -1:
 *         if not buf._length:             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
  __pyx_t_1 = (!(__pyx_v_buf->_length != 0));
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":124
 *     cdef int write_buffer(self, WriteBuffer buf) except -1:
 *         if not buf._length:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         self.ensure_allocated(buf._length)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":123
 * 
 *     cdef int write_buffer(self, WriteBuffer buf) except -1:
 *         if not buf._length:             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":126
 *             return 0
 * 
 *         self.ensure_allocated(buf._length)             # <<<<<<<<<<<<<<
 *         memcpy(self._buf + self._length,
 *                <void*> buf._buf,
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_self, __pyx_v_buf->_length); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 126, __pyx_L1_error)

  /* "asynctnt/iproto/buffer.pyx":127
 * 
 *         self.ensure_allocated(buf._length)
 *         memcpy(self._buf + self._length,             # <<<<<<<<<<<<<<
 *                <void*> buf._buf,
 *                <size_t> buf._length)
 */
  (void)(memcpy((__pyx_v_self->_buf + __pyx_v_self->_length), ((void *)__pyx_v_buf->_buf), ((size_t)__pyx_v_buf->_length)));

  /* "asynctnt/iproto/buffer.pyx":130
 *                <void*> buf._buf,
 *                <size_t> buf._length)
 *         self._length += buf._length             # <<<<<<<<<<<<<<
 * 
 *     cdef int write_header(self, uint64_t sync,
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + __pyx_v_buf->_length);

  /* "asynctnt/iproto/buffer.pyx":122
 *             self._size = new_size
 * 
 *     cdef int write_buffer(self, WriteBuffer buf) except -1:             # <<<<<<<<<<<<<<
 *         if not buf._length:
 *             return 0
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.write_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":132
 *         self._length += buf._length
 * 
 *     cdef int write_header(self, uint64_t sync,             # <<<<<<<<<<<<<<
 *                           tarantool.iproto_type op,
 *                           int64_t schema_id,
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_header(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, uint64_t __pyx_v_sync, enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_v_op, int64_t __pyx_v_schema_id, uint64_t __pyx_v_stream_id) {
  char *__pyx_v_begin;
  char *__pyx_v_p;
  uint32_t __pyx_v_map_size;
  int __pyx_r;
  int __pyx_t_1;
  char *__pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":137
 *                           uint64_t stream_id) except -1:
 *         cdef:
 *             char *begin = NULL             # <<<<<<<<<<<<<<
 *             char *p = NULL
 *             uint32_t map_size
 */
  __pyx_v_begin = NULL;

  /* "asynctnt/iproto/buffer.pyx":138
 *         cdef:
 *             char *begin = NULL
 *             char *p = NULL             # <<<<<<<<<<<<<<
 *             uint32_t map_size
 *         self.ensure_allocated(HEADER_CONST_LEN)
 */
  __pyx_v_p = NULL;

  /* "asynctnt/iproto/buffer.pyx":140
 *             char *p = NULL
 *             uint32_t map_size
 *         self.ensure_allocated(HEADER_CONST_LEN)             # <<<<<<<<<<<<<<
 * 
 *         map_size = 2 \
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_self, 33); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 140, __pyx_L1_error)

  /* "asynctnt/iproto/buffer.pyx":144
 *         map_size = 2 \
 *                     + (<uint32_t> (schema_id > 0)) \
 *                     + (<uint32_t> (stream_id > 0))             # <<<<<<<<<<<<<<
 * 
 *         p = begin = &self._buf[self._length]
 */
  __pyx_v_map_size = ((2 + ((uint32_t)(__pyx_v_schema_id > 0))) + ((uint32_t)(__pyx_v_stream_id > 0)));

  /* "asynctnt/iproto/buffer.pyx":146
 *                     + (<uint32_t> (stream_id > 0))
 * 
 *         p = begin = &self._buf[self._length]             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(&p[5], map_size)
 *         p = mp_encode_uint(p, tarantool.IPROTO_REQUEST_TYPE)
 */
  __pyx_t_2 = (&(__pyx_v_self->_buf[__pyx_v_self->_length]));
  __pyx_v_p = __pyx_t_2;
  __pyx_v_begin = __pyx_t_2;

  /* "asynctnt/iproto/buffer.pyx":147
 * 
 *         p = begin = &self._buf[self._length]
 *         p = mp_encode_map(&p[5], map_size)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_REQUEST_TYPE)
 *         p = mp_encode_uint(p, <uint32_t> op)
 */
  __pyx_v_p = mp_encode_map((&(__pyx_v_p[5])), __pyx_v_map_size);

  /* "asynctnt/iproto/buffer.pyx":148
 *         p = begin = &self._buf[self._length]
 *         p = mp_encode_map(&p[5], map_size)
 *         p = mp_encode_uint(p, tarantool.IPROTO_REQUEST_TYPE)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, <uint32_t> op)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SYNC)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_REQUEST_TYPE);

  /* "asynctnt/iproto/buffer.pyx":149
 *         p = mp_encode_map(&p[5], map_size)
 *         p = mp_encode_uint(p, tarantool.IPROTO_REQUEST_TYPE)
 *         p = mp_encode_uint(p, <uint32_t> op)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_SYNC)
 *         p = mp_encode_uint(p, sync)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, ((uint32_t)__pyx_v_op));

  /* "asynctnt/iproto/buffer.pyx":150
 *         p = mp_encode_uint(p, tarantool.IPROTO_REQUEST_TYPE)
 *         p = mp_encode_uint(p, <uint32_t> op)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SYNC)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, sync)
 * 
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SYNC);

  /* "asynctnt/iproto/buffer.pyx":151
 *         p = mp_encode_uint(p, <uint32_t> op)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SYNC)
 *         p = mp_encode_uint(p, sync)             # <<<<<<<<<<<<<<
 * 
 *         if schema_id > 0:  # pragma: nocover  # asynctnt does not send schema_id
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_sync);

  /* "asynctnt/iproto/buffer.pyx":153
 *         p = mp_encode_uint(p, sync)
 * 
 *         if schema_id > 0:  # pragma: nocover  # asynctnt does not send schema_id             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_SCHEMA_VERSION)
 *             p = mp_store_u8(p, 0xce)
 */
  __pyx_t_3 = (__pyx_v_schema_id > 0);
  if (__pyx_t_3) {

    /* "asynctnt/iproto/buffer.pyx":154
 * 
 *         if schema_id > 0:  # pragma: nocover  # asynctnt does not send schema_id
 *             p = mp_encode_uint(p, tarantool.IPROTO_SCHEMA_VERSION)             # <<<<<<<<<<<<<<
 *             p = mp_store_u8(p, 0xce)
 *             p = mp_store_u32(p, schema_id)
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SCHEMA_VERSION);

    /* "asynctnt/iproto/buffer.pyx":155
 *         if schema_id > 0:  # pragma: nocover  # asynctnt does not send schema_id
 *             p = mp_encode_uint(p, tarantool.IPROTO_SCHEMA_VERSION)
 *             p = mp_store_u8(p, 0xce)             # <<<<<<<<<<<<<<
 *             p = mp_store_u32(p, schema_id)
 * 
 */
    __pyx_v_p = mp_store_u8(__pyx_v_p, 0xce);

    /* "asynctnt/iproto/buffer.pyx":156
 *             p = mp_encode_uint(p, tarantool.IPROTO_SCHEMA_VERSION)
 *             p = mp_store_u8(p, 0xce)
 *             p = mp_store_u32(p, schema_id)             # <<<<<<<<<<<<<<
 * 
 *         if stream_id > 0:
 */
    __pyx_v_p = mp_store_u32(__pyx_v_p, __pyx_v_schema_id);

    /* "asynctnt/iproto/buffer.pyx":153
 *         p = mp_encode_uint(p, sync)
 * 
 *         if schema_id > 0:  # pragma: nocover  # asynctnt does not send schema_id             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_SCHEMA_VERSION)
 *             p = mp_store_u8(p, 0xce)
 */
  }

  /* "asynctnt/iproto/buffer.pyx":158
 *             p = mp_store_u32(p, schema_id)
 * 
 *         if stream_id > 0:             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_STREAM_ID)
 *             p = mp_encode_uint(p, stream_id)
 */
  __pyx_t_3 = (__pyx_v_stream_id > 0);
  if (__pyx_t_3) {

    /* "asynctnt/iproto/buffer.pyx":159
 * 
 *         if stream_id > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_STREAM_ID)             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, stream_id)
 * 
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_STREAM_ID);

    /* "asynctnt/iproto/buffer.pyx":160
 *         if stream_id > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_STREAM_ID)
 *             p = mp_encode_uint(p, stream_id)             # <<<<<<<<<<<<<<
 * 
 *         self._length += (p - begin)
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_stream_id);

    /* "asynctnt/iproto/buffer.pyx":158
 *             p = mp_store_u32(p, schema_id)
 * 
 *         if stream_id > 0:             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_STREAM_ID)
 *             p = mp_encode_uint(p, stream_id)
 */
  }

  /* "asynctnt/iproto/buffer.pyx":162
 *             p = mp_encode_uint(p, stream_id)
 * 
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_length(self):
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":132
 *         self._length += buf._length
 * 
 *     cdef int write_header(self, uint64_t sync,             # <<<<<<<<<<<<<<
 *                           tarantool.iproto_type op,
 *                           int64_t schema_id,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.write_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":164
 *         self._length += (p - begin)
 * 
 *     cdef void write_length(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *p
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_length(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self) {
  char *__pyx_v_p;
  char *__pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":167
 *         cdef:
 *             char *p
 *         p = self._buf             # <<<<<<<<<<<<<<
 *         p = mp_store_u8(p, 0xce)
 *         p = mp_store_u32(p, self._length - 5)
 */
  __pyx_t_1 = __pyx_v_self->_buf;
  __pyx_v_p = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":168
 *             char *p
 *         p = self._buf
 *         p = mp_store_u8(p, 0xce)             # <<<<<<<<<<<<<<
 *         p = mp_store_u32(p, self._length - 5)
 * 
 */
  __pyx_v_p = mp_store_u8(__pyx_v_p, 0xce);

  /* "asynctnt/iproto/buffer.pyx":169
 *         p = self._buf
 *         p = mp_store_u8(p, 0xce)
 *         p = mp_store_u32(p, self._length - 5)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_nil(self, char *p) except NULL:
 */
  __pyx_v_p = mp_store_u32(__pyx_v_p, (__pyx_v_self->_length - 5));

  /* "asynctnt/iproto/buffer.pyx":164
 *         self._length += (p - begin)
 * 
 *     cdef void write_length(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *p
 */

  /* function exit code */
}

/* "asynctnt/iproto/buffer.pyx":171
 *         p = mp_store_u32(p, self._length - 5)
 * 
 *     cdef char *mp_encode_nil(self, char *p) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, 1)
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_nil(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p) {
  char *__pyx_v_begin;
  char *__pyx_r;
  char *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":173
 *     cdef char *mp_encode_nil(self, char *p) except NULL:
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, 1)             # <<<<<<<<<<<<<<
 *         p = mp_encode_nil(p)
 *         self._length += (p - begin)
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, 1); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(3, 173, __pyx_L1_error)
  __pyx_v_p = __pyx_t_1;
  __pyx_v_begin = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":174
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, 1)
 *         p = mp_encode_nil(p)             # <<<<<<<<<<<<<<
 *         self._length += (p - begin)
 *         return p
 */
  __pyx_v_p = mp_encode_nil(__pyx_v_p);

  /* "asynctnt/iproto/buffer.pyx":175
 *         p = begin = self._ensure_allocated(p, 1)
 *         p = mp_encode_nil(p)
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":176
 *         p = mp_encode_nil(p)
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_bool(self, char *p, bint value) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":171
 *         p = mp_store_u32(p, self._length - 5)
 * 
 *     cdef char *mp_encode_nil(self, char *p) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, 1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_nil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":178
 *         return p
 * 
 *     cdef char *mp_encode_bool(self, char *p, bint value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, 1)
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_bool(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, int __pyx_v_value) {
  char *__pyx_v_begin;
  char *__pyx_r;
  char *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":180
 *     cdef char *mp_encode_bool(self, char *p, bint value) except NULL:
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, 1)             # <<<<<<<<<<<<<<
 *         p = mp_encode_bool(p, value)
 *         self._length += (p - begin)
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, 1); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(3, 180, __pyx_L1_error)
  __pyx_v_p = __pyx_t_1;
  __pyx_v_begin = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":181
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, 1)
 *         p = mp_encode_bool(p, value)             # <<<<<<<<<<<<<<
 *         self._length += (p - begin)
 *         return p
 */
  __pyx_v_p = mp_encode_bool(__pyx_v_p, __pyx_v_value);

  /* "asynctnt/iproto/buffer.pyx":182
 *         p = begin = self._ensure_allocated(p, 1)
 *         p = mp_encode_bool(p, value)
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":183
 *         p = mp_encode_bool(p, value)
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_double(self, char *p, double value) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":178
 *         return p
 * 
 *     cdef char *mp_encode_bool(self, char *p, bint value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, 1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_bool", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":185
 *         return p
 * 
 *     cdef char *mp_encode_double(self, char *p, double value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_double(value))
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_double(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, double __pyx_v_value) {
  char *__pyx_v_begin;
  char *__pyx_r;
  char *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":187
 *     cdef char *mp_encode_double(self, char *p, double value) except NULL:
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_double(value))             # <<<<<<<<<<<<<<
 *         p = mp_encode_double(p, value)
 *         self._length += (p - begin)
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, mp_sizeof_double(__pyx_v_value)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(3, 187, __pyx_L1_error)
  __pyx_v_p = __pyx_t_1;
  __pyx_v_begin = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":188
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_double(value))
 *         p = mp_encode_double(p, value)             # <<<<<<<<<<<<<<
 *         self._length += (p - begin)
 *         return p
 */
  __pyx_v_p = mp_encode_double(__pyx_v_p, __pyx_v_value);

  /* "asynctnt/iproto/buffer.pyx":189
 *         p = begin = self._ensure_allocated(p, mp_sizeof_double(value))
 *         p = mp_encode_double(p, value)
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":190
 *         p = mp_encode_double(p, value)
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_uint(self, char *p, uint64_t value) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":185
 *         return p
 * 
 *     cdef char *mp_encode_double(self, char *p, double value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_double(value))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":192
 *         return p
 * 
 *     cdef char *mp_encode_uint(self, char *p, uint64_t value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_uint(value))
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uint(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, uint64_t __pyx_v_value) {
  char *__pyx_v_begin;
  char *__pyx_r;
  char *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":194
 *     cdef char *mp_encode_uint(self, char *p, uint64_t value) except NULL:
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_uint(value))             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, value)
 *         self._length += (p - begin)
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, mp_sizeof_uint(__pyx_v_value)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(3, 194, __pyx_L1_error)
  __pyx_v_p = __pyx_t_1;
  __pyx_v_begin = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":195
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_uint(value))
 *         p = mp_encode_uint(p, value)             # <<<<<<<<<<<<<<
 *         self._length += (p - begin)
 *         return p
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_value);

  /* "asynctnt/iproto/buffer.pyx":196
 *         p = begin = self._ensure_allocated(p, mp_sizeof_uint(value))
 *         p = mp_encode_uint(p, value)
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":197
 *         p = mp_encode_uint(p, value)
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_int(self, char *p, int64_t value) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":192
 *         return p
 * 
 *     cdef char *mp_encode_uint(self, char *p, uint64_t value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_uint(value))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_uint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":199
 *         return p
 * 
 *     cdef char *mp_encode_int(self, char *p, int64_t value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_int(value))
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_int(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, int64_t __pyx_v_value) {
  char *__pyx_v_begin;
  char *__pyx_r;
  char *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":201
 *     cdef char *mp_encode_int(self, char *p, int64_t value) except NULL:
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_int(value))             # <<<<<<<<<<<<<<
 *         p = mp_encode_int(p, value)
 *         self._length += (p - begin)
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, mp_sizeof_int(__pyx_v_value)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(3, 201, __pyx_L1_error)
  __pyx_v_p = __pyx_t_1;
  __pyx_v_begin = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":202
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_int(value))
 *         p = mp_encode_int(p, value)             # <<<<<<<<<<<<<<
 *         self._length += (p - begin)
 *         return p
 */
  __pyx_v_p = mp_encode_int(__pyx_v_p, __pyx_v_value);

  /* "asynctnt/iproto/buffer.pyx":203
 *         p = begin = self._ensure_allocated(p, mp_sizeof_int(value))
 *         p = mp_encode_int(p, value)
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":204
 *         p = mp_encode_int(p, value)
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_str(self, char *p,
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":199
 *         return p
 * 
 *     cdef char *mp_encode_int(self, char *p, int64_t value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_int(value))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":206
 *         return p
 * 
 *     cdef char *mp_encode_str(self, char *p,             # <<<<<<<<<<<<<<
 *                              const char *str, uint32_t len) except NULL:
 *         cdef char *begin
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_str(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, char const *__pyx_v_str, uint32_t __pyx_v_len) {
  char *__pyx_v_begin;
  char *__pyx_r;
  char *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":209
 *                              const char *str, uint32_t len) except NULL:
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_str(len))             # <<<<<<<<<<<<<<
 *         p = mp_encode_str(p, str, len)
 *         self._length += (p - begin)
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, mp_sizeof_str(__pyx_v_len)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(3, 209, __pyx_L1_error)
  __pyx_v_p = __pyx_t_1;
  __pyx_v_begin = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":210
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_str(len))
 *         p = mp_encode_str(p, str, len)             # <<<<<<<<<<<<<<
 *         self._length += (p - begin)
 *         return p
 */
  __pyx_v_p = mp_encode_str(__pyx_v_p, __pyx_v_str, __pyx_v_len);

  /* "asynctnt/iproto/buffer.pyx":211
 *         p = begin = self._ensure_allocated(p, mp_sizeof_str(len))
 *         p = mp_encode_str(p, str, len)
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":212
 *         p = mp_encode_str(p, str, len)
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_bin(self, char *p,
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":206
 *         return p
 * 
 *     cdef char *mp_encode_str(self, char *p,             # <<<<<<<<<<<<<<
 *                              const char *str, uint32_t len) except NULL:
 *         cdef char *begin
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":214
 *         return p
 * 
 *     cdef char *mp_encode_bin(self, char *p,             # <<<<<<<<<<<<<<
 *                              const char *data, uint32_t len) except NULL:
 *         cdef char *begin
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_bin(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, char const *__pyx_v_data, uint32_t __pyx_v_len) {
  char *__pyx_v_begin;
  char *__pyx_r;
  char *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":217
 *                              const char *data, uint32_t len) except NULL:
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_bin(len))             # <<<<<<<<<<<<<<
 *         p = mp_encode_bin(p, data, len)
 *         self._length += (p - begin)
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, mp_sizeof_bin(__pyx_v_len)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(3, 217, __pyx_L1_error)
  __pyx_v_p = __pyx_t_1;
  __pyx_v_begin = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":218
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_bin(len))
 *         p = mp_encode_bin(p, data, len)             # <<<<<<<<<<<<<<
 *         self._length += (p - begin)
 *         return p
 */
  __pyx_v_p = mp_encode_bin(__pyx_v_p, __pyx_v_data, __pyx_v_len);

  /* "asynctnt/iproto/buffer.pyx":219
 *         p = begin = self._ensure_allocated(p, mp_sizeof_bin(len))
 *         p = mp_encode_bin(p, data, len)
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":220
 *         p = mp_encode_bin(p, data, len)
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_decimal(self, char *p, object value) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":214
 *         return p
 * 
 *     cdef char *mp_encode_bin(self, char *p,             # <<<<<<<<<<<<<<
 *                              const char *data, uint32_t len) except NULL:
 *         cdef char *begin
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_bin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":222
 *         return p
 * 
 *     cdef char *mp_encode_decimal(self, char *p, object value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_decimal(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_value) {
  char *__pyx_v_begin;
  uint8_t __pyx_v_sign;
  PyObject *__pyx_v_digits = 0;
  uint32_t __pyx_v_digits_count;
  int __pyx_v_exponent;
  uint32_t __pyx_v_length;
  PyObject *__pyx_v_decimal_tuple = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  uint8_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  uint32_t __pyx_t_7;
  char *__pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mp_encode_decimal", 1);

  /* "asynctnt/iproto/buffer.pyx":231
 *             uint32_t length
 * 
 *         decimal_tuple = value.as_tuple()             # <<<<<<<<<<<<<<
 *         sign = <uint8_t> decimal_tuple.sign
 *         digits = <tuple> decimal_tuple.digits
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_as_tuple); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_decimal_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/buffer.pyx":232
 * 
 *         decimal_tuple = value.as_tuple()
 *         sign = <uint8_t> decimal_tuple.sign             # <<<<<<<<<<<<<<
 *         digits = <tuple> decimal_tuple.digits
 *         exponent = <int> decimal_tuple.exponent
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_decimal_tuple, __pyx_n_s_sign); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(3, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sign = ((uint8_t)__pyx_t_5);

  /* "asynctnt/iproto/buffer.pyx":233
 *         decimal_tuple = value.as_tuple()
 *         sign = <uint8_t> decimal_tuple.sign
 *         digits = <tuple> decimal_tuple.digits             # <<<<<<<<<<<<<<
 *         exponent = <int> decimal_tuple.exponent
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_decimal_tuple, __pyx_n_s_digits); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_digits = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/buffer.pyx":234
 *         sign = <uint8_t> decimal_tuple.sign
 *         digits = <tuple> decimal_tuple.digits
 *         exponent = <int> decimal_tuple.exponent             # <<<<<<<<<<<<<<
 * 
 *         digits_count = <uint32_t> len(digits)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_decimal_tuple, __pyx_n_s_exponent); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_exponent = ((int)__pyx_t_4);

  /* "asynctnt/iproto/buffer.pyx":236
 *         exponent = <int> decimal_tuple.exponent
 * 
 *         digits_count = <uint32_t> len(digits)             # <<<<<<<<<<<<<<
 *         length = decimal_len(exponent, digits_count)
 * 
 */
  if (unlikely(__pyx_v_digits == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 236, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v_digits); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(3, 236, __pyx_L1_error)
  __pyx_v_digits_count = ((uint32_t)__pyx_t_6);

  /* "asynctnt/iproto/buffer.pyx":237
 * 
 *         digits_count = <uint32_t> len(digits)
 *         length = decimal_len(exponent, digits_count)             # <<<<<<<<<<<<<<
 * 
 *         p = begin = self._ensure_allocated(p, mp_sizeof_ext(length))
 */
  __pyx_t_7 = __pyx_f_8asynctnt_6iproto_8protocol_decimal_len(__pyx_v_exponent, __pyx_v_digits_count); if (unlikely(__pyx_t_7 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 237, __pyx_L1_error)
  __pyx_v_length = __pyx_t_7;

  /* "asynctnt/iproto/buffer.pyx":239
 *         length = decimal_len(exponent, digits_count)
 * 
 *         p = begin = self._ensure_allocated(p, mp_sizeof_ext(length))             # <<<<<<<<<<<<<<
 * 
 *         # encode header
 */
  __pyx_t_8 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, mp_sizeof_ext(__pyx_v_length)); if (unlikely(__pyx_t_8 == ((char *)NULL))) __PYX_ERR(3, 239, __pyx_L1_error)
  __pyx_v_p = __pyx_t_8;
  __pyx_v_begin = __pyx_t_8;

  /* "asynctnt/iproto/buffer.pyx":242
 * 
 *         # encode header
 *         p = mp_encode_extl(p, tarantool.MP_DECIMAL, length)             # <<<<<<<<<<<<<<
 * 
 *         # encode decimal
 */
  __pyx_v_p = mp_encode_extl(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_MP_DECIMAL, __pyx_v_length);

  /* "asynctnt/iproto/buffer.pyx":245
 * 
 *         # encode decimal
 *         p = decimal_encode(p, digits_count, sign, digits, exponent)             # <<<<<<<<<<<<<<
 * 
 *         self._length += (p - begin)
 */
  __pyx_t_8 = __pyx_f_8asynctnt_6iproto_8protocol_decimal_encode(__pyx_v_p, __pyx_v_digits_count, __pyx_v_sign, __pyx_v_digits, __pyx_v_exponent); if (unlikely(__pyx_t_8 == ((char *)NULL))) __PYX_ERR(3, 245, __pyx_L1_error)
  __pyx_v_p = __pyx_t_8;

  /* "asynctnt/iproto/buffer.pyx":247
 *         p = decimal_encode(p, digits_count, sign, digits, exponent)
 * 
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":248
 * 
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_uuid(self, char *p, object value) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":222
 *         return p
 * 
 *     cdef char *mp_encode_decimal(self, char *p, object value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_decimal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_digits);
  __Pyx_XDECREF(__pyx_v_decimal_tuple);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":250
 *         return p
 * 
 *     cdef char *mp_encode_uuid(self, char *p, object value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uuid(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_value) {
  char *__pyx_v_begin;
  char *__pyx_v_data_p;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mp_encode_uuid", 1);

  /* "asynctnt/iproto/buffer.pyx":255
 *             char *data_p
 * 
 *         p = begin = self._ensure_allocated(p, mp_sizeof_ext(16))             # <<<<<<<<<<<<<<
 *         data_p = cpython.bytes.PyBytes_AS_STRING(<bytes> value.bytes)
 * 
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, mp_sizeof_ext(16)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(3, 255, __pyx_L1_error)
  __pyx_v_p = __pyx_t_1;
  __pyx_v_begin = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":256
 * 
 *         p = begin = self._ensure_allocated(p, mp_sizeof_ext(16))
 *         data_p = cpython.bytes.PyBytes_AS_STRING(<bytes> value.bytes)             # <<<<<<<<<<<<<<
 * 
 *         # encode header
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_bytes); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_data_p = PyBytes_AS_STRING(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/buffer.pyx":259
 * 
 *         # encode header
 *         p = mp_encode_ext(p, tarantool.MP_UUID, data_p, 16)  # uuid is exactly 16 bytes             # <<<<<<<<<<<<<<
 * 
 *         self._length += (p - begin)
 */
  __pyx_v_p = mp_encode_ext(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_MP_UUID, __pyx_v_data_p, 16);

  /* "asynctnt/iproto/buffer.pyx":261
 *         p = mp_encode_ext(p, tarantool.MP_UUID, data_p, 16)  # uuid is exactly 16 bytes
 * 
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":262
 * 
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_datetime(self, char *p, object value) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":250
 *         return p
 * 
 *     cdef char *mp_encode_uuid(self, char *p, object value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_uuid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":264
 *         return p
 * 
 *     cdef char *mp_encode_datetime(self, char *p, object value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_datetime(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_value) {
  char *__pyx_v_begin;
  uint32_t __pyx_v_length;
  PyDateTime_DateTime *__pyx_v_pydt = 0;
  struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime __pyx_v_dt;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint32_t __pyx_t_2;
  char *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mp_encode_datetime", 1);

  /* "asynctnt/iproto/buffer.pyx":271
 *             IProtoDateTime dt
 * 
 *         pydt = <datetime> value             # <<<<<<<<<<<<<<
 *         datetime_zero(&dt)
 *         datetime_from_py(pydt, &dt)
 */
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_pydt = ((PyDateTime_DateTime *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/buffer.pyx":272
 * 
 *         pydt = <datetime> value
 *         datetime_zero(&dt)             # <<<<<<<<<<<<<<
 *         datetime_from_py(pydt, &dt)
 * 
 */
  __pyx_f_8asynctnt_6iproto_8protocol_datetime_zero((&__pyx_v_dt)); if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 272, __pyx_L1_error)

  /* "asynctnt/iproto/buffer.pyx":273
 *         pydt = <datetime> value
 *         datetime_zero(&dt)
 *         datetime_from_py(pydt, &dt)             # <<<<<<<<<<<<<<
 * 
 *         length = datetime_len(&dt)
 */
  __pyx_f_8asynctnt_6iproto_8protocol_datetime_from_py(__pyx_v_pydt, (&__pyx_v_dt)); if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 273, __pyx_L1_error)

  /* "asynctnt/iproto/buffer.pyx":275
 *         datetime_from_py(pydt, &dt)
 * 
 *         length = datetime_len(&dt)             # <<<<<<<<<<<<<<
 *         p = begin = self._ensure_allocated(p, mp_sizeof_ext(length))
 *         p = mp_encode_extl(p, tarantool.MP_DATETIME, length)
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_datetime_len((&__pyx_v_dt)); if (unlikely(__pyx_t_2 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 275, __pyx_L1_error)
  __pyx_v_length = __pyx_t_2;

  /* "asynctnt/iproto/buffer.pyx":276
 * 
 *         length = datetime_len(&dt)
 *         p = begin = self._ensure_allocated(p, mp_sizeof_ext(length))             # <<<<<<<<<<<<<<
 *         p = mp_encode_extl(p, tarantool.MP_DATETIME, length)
 *         p = datetime_encode(p, &dt)
 */
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, mp_sizeof_ext(__pyx_v_length)); if (unlikely(__pyx_t_3 == ((char *)NULL))) __PYX_ERR(3, 276, __pyx_L1_error)
  __pyx_v_p = __pyx_t_3;
  __pyx_v_begin = __pyx_t_3;

  /* "asynctnt/iproto/buffer.pyx":277
 *         length = datetime_len(&dt)
 *         p = begin = self._ensure_allocated(p, mp_sizeof_ext(length))
 *         p = mp_encode_extl(p, tarantool.MP_DATETIME, length)             # <<<<<<<<<<<<<<
 *         p = datetime_encode(p, &dt)
 *         self._length += (p - begin)
 */
  __pyx_v_p = mp_encode_extl(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_MP_DATETIME, __pyx_v_length);

  /* "asynctnt/iproto/buffer.pyx":278
 *         p = begin = self._ensure_allocated(p, mp_sizeof_ext(length))
 *         p = mp_encode_extl(p, tarantool.MP_DATETIME, length)
 *         p = datetime_encode(p, &dt)             # <<<<<<<<<<<<<<
 *         self._length += (p - begin)
 *         return p
 */
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_datetime_encode(__pyx_v_p, (&__pyx_v_dt)); if (unlikely(__pyx_t_3 == ((char *)NULL))) __PYX_ERR(3, 278, __pyx_L1_error)
  __pyx_v_p = __pyx_t_3;

  /* "asynctnt/iproto/buffer.pyx":279
 *         p = mp_encode_extl(p, tarantool.MP_DATETIME, length)
 *         p = datetime_encode(p, &dt)
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":280
 *         p = datetime_encode(p, &dt)
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_array(self, char *p, uint32_t len) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":264
 *         return p
 * 
 *     cdef char *mp_encode_datetime(self, char *p, object value) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_datetime", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_pydt);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":282
 *         return p
 * 
 *     cdef char *mp_encode_array(self, char *p, uint32_t len) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_array(len))
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_array(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, uint32_t __pyx_v_len) {
  char *__pyx_v_begin;
  char *__pyx_r;
  char *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":284
 *     cdef char *mp_encode_array(self, char *p, uint32_t len) except NULL:
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_array(len))             # <<<<<<<<<<<<<<
 *         p = mp_encode_array(p, len)
 *         self._length += (p - begin)
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, mp_sizeof_array(__pyx_v_len)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(3, 284, __pyx_L1_error)
  __pyx_v_p = __pyx_t_1;
  __pyx_v_begin = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":285
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_array(len))
 *         p = mp_encode_array(p, len)             # <<<<<<<<<<<<<<
 *         self._length += (p - begin)
 *         return p
 */
  __pyx_v_p = mp_encode_array(__pyx_v_p, __pyx_v_len);

  /* "asynctnt/iproto/buffer.pyx":286
 *         p = begin = self._ensure_allocated(p, mp_sizeof_array(len))
 *         p = mp_encode_array(p, len)
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":287
 *         p = mp_encode_array(p, len)
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_map(self, char *p, uint32_t len) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":282
 *         return p
 * 
 *     cdef char *mp_encode_array(self, char *p, uint32_t len) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_array(len))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":289
 *         return p
 * 
 *     cdef char *mp_encode_map(self, char *p, uint32_t len) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_map(len))
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_map(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, uint32_t __pyx_v_len) {
  char *__pyx_v_begin;
  char *__pyx_r;
  char *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/buffer.pyx":291
 *     cdef char *mp_encode_map(self, char *p, uint32_t len) except NULL:
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_map(len))             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(p, len)
 *         self._length += (p - begin)
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_self, __pyx_v_p, mp_sizeof_map(__pyx_v_len)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(3, 291, __pyx_L1_error)
  __pyx_v_p = __pyx_t_1;
  __pyx_v_begin = __pyx_t_1;

  /* "asynctnt/iproto/buffer.pyx":292
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_map(len))
 *         p = mp_encode_map(p, len)             # <<<<<<<<<<<<<<
 *         self._length += (p - begin)
 *         return p
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_len);

  /* "asynctnt/iproto/buffer.pyx":293
 *         p = begin = self._ensure_allocated(p, mp_sizeof_map(len))
 *         p = mp_encode_map(p, len)
 *         self._length += (p - begin)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_v_self->_length = (__pyx_v_self->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/buffer.pyx":294
 *         p = mp_encode_map(p, len)
 *         self._length += (p - begin)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_list(self, char *p, list arr) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":289
 *         return p
 * 
 *     cdef char *mp_encode_map(self, char *p, uint32_t len) except NULL:             # <<<<<<<<<<<<<<
 *         cdef char *begin
 *         p = begin = self._ensure_allocated(p, mp_sizeof_map(len))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":296
 *         return p
 * 
 *     cdef char *mp_encode_list(self, char *p, list arr) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint32_t arr_len
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_list(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_arr) {
  uint32_t __pyx_v_arr_len;
  PyObject *__pyx_v_item_ptr;
  PyObject *__pyx_v_item = 0;
  uint32_t __pyx_v_i;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  char *__pyx_t_2;
  uint32_t __pyx_t_3;
  uint32_t __pyx_t_4;
  uint32_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mp_encode_list", 1);

  /* "asynctnt/iproto/buffer.pyx":302
 *             object item
 * 
 *         if arr is not None:             # <<<<<<<<<<<<<<
 *             arr_len = <uint32_t> cpython.list.PyList_GET_SIZE(arr)
 *         else:  # pragma: nocover
 */
  __pyx_t_1 = (__pyx_v_arr != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":303
 * 
 *         if arr is not None:
 *             arr_len = <uint32_t> cpython.list.PyList_GET_SIZE(arr)             # <<<<<<<<<<<<<<
 *         else:  # pragma: nocover
 *             arr_len = 0
 */
    __pyx_v_arr_len = ((uint32_t)PyList_GET_SIZE(__pyx_v_arr));

    /* "asynctnt/iproto/buffer.pyx":302
 *             object item
 * 
 *         if arr is not None:             # <<<<<<<<<<<<<<
 *             arr_len = <uint32_t> cpython.list.PyList_GET_SIZE(arr)
 *         else:  # pragma: nocover
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/buffer.pyx":305
 *             arr_len = <uint32_t> cpython.list.PyList_GET_SIZE(arr)
 *         else:  # pragma: nocover
 *             arr_len = 0             # <<<<<<<<<<<<<<
 *         p = self.mp_encode_array(p, arr_len)
 * 
 */
  /*else*/ {
    __pyx_v_arr_len = 0;
  }
  __pyx_L3:;

  /* "asynctnt/iproto/buffer.pyx":306
 *         else:  # pragma: nocover
 *             arr_len = 0
 *         p = self.mp_encode_array(p, arr_len)             # <<<<<<<<<<<<<<
 * 
 *         if arr_len > 0:
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_array(__pyx_v_self, __pyx_v_p, __pyx_v_arr_len); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 306, __pyx_L1_error)
  __pyx_v_p = __pyx_t_2;

  /* "asynctnt/iproto/buffer.pyx":308
 *         p = self.mp_encode_array(p, arr_len)
 * 
 *         if arr_len > 0:             # <<<<<<<<<<<<<<
 *             for i in range(arr_len):
 *                 item_ptr = cpython.list.PyList_GET_ITEM(arr, i)
 */
  __pyx_t_1 = (__pyx_v_arr_len > 0);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":309
 * 
 *         if arr_len > 0:
 *             for i in range(arr_len):             # <<<<<<<<<<<<<<
 *                 item_ptr = cpython.list.PyList_GET_ITEM(arr, i)
 *                 item = <object> item_ptr
 */
    __pyx_t_3 = __pyx_v_arr_len;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "asynctnt/iproto/buffer.pyx":310
 *         if arr_len > 0:
 *             for i in range(arr_len):
 *                 item_ptr = cpython.list.PyList_GET_ITEM(arr, i)             # <<<<<<<<<<<<<<
 *                 item = <object> item_ptr
 *                 p = self.mp_encode_obj(p, item)
 */
      __pyx_v_item_ptr = PyList_GET_ITEM(__pyx_v_arr, __pyx_v_i);

      /* "asynctnt/iproto/buffer.pyx":311
 *             for i in range(arr_len):
 *                 item_ptr = cpython.list.PyList_GET_ITEM(arr, i)
 *                 item = <object> item_ptr             # <<<<<<<<<<<<<<
 *                 p = self.mp_encode_obj(p, item)
 *         return p
 */
      __pyx_t_6 = ((PyObject *)__pyx_v_item_ptr);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "asynctnt/iproto/buffer.pyx":312
 *                 item_ptr = cpython.list.PyList_GET_ITEM(arr, i)
 *                 item = <object> item_ptr
 *                 p = self.mp_encode_obj(p, item)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
      __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_obj(__pyx_v_self, __pyx_v_p, __pyx_v_item); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 312, __pyx_L1_error)
      __pyx_v_p = __pyx_t_2;
    }

    /* "asynctnt/iproto/buffer.pyx":308
 *         p = self.mp_encode_array(p, arr_len)
 * 
 *         if arr_len > 0:             # <<<<<<<<<<<<<<
 *             for i in range(arr_len):
 *                 item_ptr = cpython.list.PyList_GET_ITEM(arr, i)
 */
  }

  /* "asynctnt/iproto/buffer.pyx":313
 *                 item = <object> item_ptr
 *                 p = self.mp_encode_obj(p, item)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_tuple(self, char *p, tuple t) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":296
 *         return p
 * 
 *     cdef char *mp_encode_list(self, char *p, list arr) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint32_t arr_len
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":315
 *         return p
 * 
 *     cdef char *mp_encode_tuple(self, char *p, tuple t) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint32_t t_len
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_tuple(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_t) {
  uint32_t __pyx_v_t_len;
  PyObject *__pyx_v_item_ptr;
  PyObject *__pyx_v_item = 0;
  uint32_t __pyx_v_i;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  char *__pyx_t_2;
  uint32_t __pyx_t_3;
  uint32_t __pyx_t_4;
  uint32_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mp_encode_tuple", 1);

  /* "asynctnt/iproto/buffer.pyx":321
 *             object item
 * 
 *         if t is not None:             # <<<<<<<<<<<<<<
 *             t_len = <uint32_t> cpython.tuple.PyTuple_GET_SIZE(t)
 *         else:  # pragma: nocover
 */
  __pyx_t_1 = (__pyx_v_t != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":322
 * 
 *         if t is not None:
 *             t_len = <uint32_t> cpython.tuple.PyTuple_GET_SIZE(t)             # <<<<<<<<<<<<<<
 *         else:  # pragma: nocover
 *             t_len = 0
 */
    __pyx_v_t_len = ((uint32_t)PyTuple_GET_SIZE(__pyx_v_t));

    /* "asynctnt/iproto/buffer.pyx":321
 *             object item
 * 
 *         if t is not None:             # <<<<<<<<<<<<<<
 *             t_len = <uint32_t> cpython.tuple.PyTuple_GET_SIZE(t)
 *         else:  # pragma: nocover
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/buffer.pyx":324
 *             t_len = <uint32_t> cpython.tuple.PyTuple_GET_SIZE(t)
 *         else:  # pragma: nocover
 *             t_len = 0             # <<<<<<<<<<<<<<
 *         p = self.mp_encode_array(p, t_len)
 * 
 */
  /*else*/ {
    __pyx_v_t_len = 0;
  }
  __pyx_L3:;

  /* "asynctnt/iproto/buffer.pyx":325
 *         else:  # pragma: nocover
 *             t_len = 0
 *         p = self.mp_encode_array(p, t_len)             # <<<<<<<<<<<<<<
 * 
 *         if t_len > 0:
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_array(__pyx_v_self, __pyx_v_p, __pyx_v_t_len); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 325, __pyx_L1_error)
  __pyx_v_p = __pyx_t_2;

  /* "asynctnt/iproto/buffer.pyx":327
 *         p = self.mp_encode_array(p, t_len)
 * 
 *         if t_len > 0:             # <<<<<<<<<<<<<<
 *             for i in range(t_len):
 *                 item_ptr = cpython.tuple.PyTuple_GET_ITEM(t, i)
 */
  __pyx_t_1 = (__pyx_v_t_len > 0);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":328
 * 
 *         if t_len > 0:
 *             for i in range(t_len):             # <<<<<<<<<<<<<<
 *                 item_ptr = cpython.tuple.PyTuple_GET_ITEM(t, i)
 *                 item = <object> item_ptr
 */
    __pyx_t_3 = __pyx_v_t_len;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "asynctnt/iproto/buffer.pyx":329
 *         if t_len > 0:
 *             for i in range(t_len):
 *                 item_ptr = cpython.tuple.PyTuple_GET_ITEM(t, i)             # <<<<<<<<<<<<<<
 *                 item = <object> item_ptr
 *                 p = self.mp_encode_obj(p, item)
 */
      __pyx_v_item_ptr = PyTuple_GET_ITEM(__pyx_v_t, __pyx_v_i);

      /* "asynctnt/iproto/buffer.pyx":330
 *             for i in range(t_len):
 *                 item_ptr = cpython.tuple.PyTuple_GET_ITEM(t, i)
 *                 item = <object> item_ptr             # <<<<<<<<<<<<<<
 *                 p = self.mp_encode_obj(p, item)
 *         return p
 */
      __pyx_t_6 = ((PyObject *)__pyx_v_item_ptr);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "asynctnt/iproto/buffer.pyx":331
 *                 item_ptr = cpython.tuple.PyTuple_GET_ITEM(t, i)
 *                 item = <object> item_ptr
 *                 p = self.mp_encode_obj(p, item)             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
      __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_obj(__pyx_v_self, __pyx_v_p, __pyx_v_item); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 331, __pyx_L1_error)
      __pyx_v_p = __pyx_t_2;
    }

    /* "asynctnt/iproto/buffer.pyx":327
 *         p = self.mp_encode_array(p, t_len)
 * 
 *         if t_len > 0:             # <<<<<<<<<<<<<<
 *             for i in range(t_len):
 *                 item_ptr = cpython.tuple.PyTuple_GET_ITEM(t, i)
 */
  }

  /* "asynctnt/iproto/buffer.pyx":332
 *                 item = <object> item_ptr
 *                 p = self.mp_encode_obj(p, item)
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_dict(self, char *p, dict d) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":315
 *         return p
 * 
 *     cdef char *mp_encode_tuple(self, char *p, tuple t) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint32_t t_len
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_tuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":334
 *         return p
 * 
 *     cdef char *mp_encode_dict(self, char *p, dict d) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint32_t d_len
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_dict(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_d) {
  uint32_t __pyx_v_d_len;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pvalue;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  Py_ssize_t __pyx_v_pos;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  char *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mp_encode_dict", 1);

  /* "asynctnt/iproto/buffer.pyx":342
 *             Py_ssize_t pos
 * 
 *         if d is not None:             # <<<<<<<<<<<<<<
 *             d_len = <uint32_t> cpython.dict.PyDict_Size(d)
 *         else:  # pragma: nocover
 */
  __pyx_t_1 = (__pyx_v_d != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":343
 * 
 *         if d is not None:
 *             d_len = <uint32_t> cpython.dict.PyDict_Size(d)             # <<<<<<<<<<<<<<
 *         else:  # pragma: nocover
 *             d_len = 0
 */
    __pyx_t_2 = PyDict_Size(__pyx_v_d); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(3, 343, __pyx_L1_error)
    __pyx_v_d_len = ((uint32_t)__pyx_t_2);

    /* "asynctnt/iproto/buffer.pyx":342
 *             Py_ssize_t pos
 * 
 *         if d is not None:             # <<<<<<<<<<<<<<
 *             d_len = <uint32_t> cpython.dict.PyDict_Size(d)
 *         else:  # pragma: nocover
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/buffer.pyx":345
 *             d_len = <uint32_t> cpython.dict.PyDict_Size(d)
 *         else:  # pragma: nocover
 *             d_len = 0             # <<<<<<<<<<<<<<
 *         p = self.mp_encode_map(p, d_len)
 * 
 */
  /*else*/ {
    __pyx_v_d_len = 0;
  }
  __pyx_L3:;

  /* "asynctnt/iproto/buffer.pyx":346
 *         else:  # pragma: nocover
 *             d_len = 0
 *         p = self.mp_encode_map(p, d_len)             # <<<<<<<<<<<<<<
 * 
 *         pos = 0
 */
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_map(__pyx_v_self, __pyx_v_p, __pyx_v_d_len); if (unlikely(__pyx_t_3 == ((char *)NULL))) __PYX_ERR(3, 346, __pyx_L1_error)
  __pyx_v_p = __pyx_t_3;

  /* "asynctnt/iproto/buffer.pyx":348
 *         p = self.mp_encode_map(p, d_len)
 * 
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while cpython.dict.PyDict_Next(d, &pos, &pkey, &pvalue):
 *             key = <object> pkey
 */
  __pyx_v_pos = 0;

  /* "asynctnt/iproto/buffer.pyx":349
 * 
 *         pos = 0
 *         while cpython.dict.PyDict_Next(d, &pos, &pkey, &pvalue):             # <<<<<<<<<<<<<<
 *             key = <object> pkey
 *             value = <object> pvalue
 */
  while (1) {
    __pyx_t_1 = (PyDict_Next(__pyx_v_d, (&__pyx_v_pos), (&__pyx_v_pkey), (&__pyx_v_pvalue)) != 0);
    if (!__pyx_t_1) break;

    /* "asynctnt/iproto/buffer.pyx":350
 *         pos = 0
 *         while cpython.dict.PyDict_Next(d, &pos, &pkey, &pvalue):
 *             key = <object> pkey             # <<<<<<<<<<<<<<
 *             value = <object> pvalue
 *             p = self.mp_encode_obj(p, key)
 */
    __pyx_t_4 = ((PyObject *)__pyx_v_pkey);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "asynctnt/iproto/buffer.pyx":351
 *         while cpython.dict.PyDict_Next(d, &pos, &pkey, &pvalue):
 *             key = <object> pkey
 *             value = <object> pvalue             # <<<<<<<<<<<<<<
 *             p = self.mp_encode_obj(p, key)
 *             p = self.mp_encode_obj(p, value)
 */
    __pyx_t_4 = ((PyObject *)__pyx_v_pvalue);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "asynctnt/iproto/buffer.pyx":352
 *             key = <object> pkey
 *             value = <object> pvalue
 *             p = self.mp_encode_obj(p, key)             # <<<<<<<<<<<<<<
 *             p = self.mp_encode_obj(p, value)
 * 
 */
    __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_obj(__pyx_v_self, __pyx_v_p, __pyx_v_key); if (unlikely(__pyx_t_3 == ((char *)NULL))) __PYX_ERR(3, 352, __pyx_L1_error)
    __pyx_v_p = __pyx_t_3;

    /* "asynctnt/iproto/buffer.pyx":353
 *             value = <object> pvalue
 *             p = self.mp_encode_obj(p, key)
 *             p = self.mp_encode_obj(p, value)             # <<<<<<<<<<<<<<
 * 
 *         return p
 */
    __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_obj(__pyx_v_self, __pyx_v_p, __pyx_v_value); if (unlikely(__pyx_t_3 == ((char *)NULL))) __PYX_ERR(3, 353, __pyx_L1_error)
    __pyx_v_p = __pyx_t_3;
  }

  /* "asynctnt/iproto/buffer.pyx":355
 *             p = self.mp_encode_obj(p, value)
 * 
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef char *mp_encode_obj(self, char *p, object o) except NULL:
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/buffer.pyx":334
 *         return p
 * 
 *     cdef char *mp_encode_dict(self, char *p, dict d) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint32_t d_len
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/buffer.pyx":357
 *         return p
 * 
 *     cdef char *mp_encode_obj(self, char *p, object o) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             bytes o_string_temp
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_obj(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, char *__pyx_v_p, PyObject *__pyx_v_o) {
  PyObject *__pyx_v_o_string_temp = 0;
  char *__pyx_v_o_string_str;
  Py_ssize_t __pyx_v_o_string_len;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  char *__pyx_t_2;
  double __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  uint64_t __pyx_t_5;
  int64_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mp_encode_obj", 1);

  /* "asynctnt/iproto/buffer.pyx":363
 *             ssize_t o_string_len
 * 
 *         if o is None:             # <<<<<<<<<<<<<<
 *             return self.mp_encode_nil(p)
 * 
 */
  __pyx_t_1 = (__pyx_v_o == Py_None);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":364
 * 
 *         if o is None:
 *             return self.mp_encode_nil(p)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(o, float):
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_nil(__pyx_v_self, __pyx_v_p); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 364, __pyx_L1_error)
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":363
 *             ssize_t o_string_len
 * 
 *         if o is None:             # <<<<<<<<<<<<<<
 *             return self.mp_encode_nil(p)
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":366
 *             return self.mp_encode_nil(p)
 * 
 *         elif isinstance(o, float):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_double(p, <double> o)
 * 
 */
  __pyx_t_1 = PyFloat_Check(__pyx_v_o); 
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":367
 * 
 *         elif isinstance(o, float):
 *             return self.mp_encode_double(p, <double> o)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(o, bool):
 */
    __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_v_o); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 367, __pyx_L1_error)
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_double(__pyx_v_self, __pyx_v_p, ((double)__pyx_t_3)); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 367, __pyx_L1_error)
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":366
 *             return self.mp_encode_nil(p)
 * 
 *         elif isinstance(o, float):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_double(p, <double> o)
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":369
 *             return self.mp_encode_double(p, <double> o)
 * 
 *         elif isinstance(o, bool):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_bool(p, <bint> o)
 * 
 */
  __pyx_t_4 = ((PyObject*)&PyBool_Type);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_o, __pyx_t_4); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 369, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":370
 * 
 *         elif isinstance(o, bool):
 *             return self.mp_encode_bool(p, <bint> o)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(o, int):
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_o); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 370, __pyx_L1_error)
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_bool(__pyx_v_self, __pyx_v_p, __pyx_t_1); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 370, __pyx_L1_error)
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":369
 *             return self.mp_encode_double(p, <double> o)
 * 
 *         elif isinstance(o, bool):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_bool(p, <bint> o)
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":372
 *             return self.mp_encode_bool(p, <bint> o)
 * 
 *         elif isinstance(o, int):             # <<<<<<<<<<<<<<
 *             if o >= 0:
 *                 return self.mp_encode_uint(p, <uint64_t> o)
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_o); 
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":373
 * 
 *         elif isinstance(o, int):
 *             if o >= 0:             # <<<<<<<<<<<<<<
 *                 return self.mp_encode_uint(p, <uint64_t> o)
 *             else:
 */
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_o, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 373, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 373, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_1) {

      /* "asynctnt/iproto/buffer.pyx":374
 *         elif isinstance(o, int):
 *             if o >= 0:
 *                 return self.mp_encode_uint(p, <uint64_t> o)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self.mp_encode_int(p, <int64_t> o)
 */
      __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_v_o); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(3, 374, __pyx_L1_error)
      __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uint(__pyx_v_self, __pyx_v_p, ((uint64_t)__pyx_t_5)); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 374, __pyx_L1_error)
      __pyx_r = __pyx_t_2;
      goto __pyx_L0;

      /* "asynctnt/iproto/buffer.pyx":373
 * 
 *         elif isinstance(o, int):
 *             if o >= 0:             # <<<<<<<<<<<<<<
 *                 return self.mp_encode_uint(p, <uint64_t> o)
 *             else:
 */
    }

    /* "asynctnt/iproto/buffer.pyx":376
 *                 return self.mp_encode_uint(p, <uint64_t> o)
 *             else:
 *                 return self.mp_encode_int(p, <int64_t> o)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(o, bytes):
 */
    /*else*/ {
      __pyx_t_6 = __Pyx_PyInt_As_int64_t(__pyx_v_o); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(3, 376, __pyx_L1_error)
      __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_int(__pyx_v_self, __pyx_v_p, ((int64_t)__pyx_t_6)); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 376, __pyx_L1_error)
      __pyx_r = __pyx_t_2;
      goto __pyx_L0;
    }

    /* "asynctnt/iproto/buffer.pyx":372
 *             return self.mp_encode_bool(p, <bint> o)
 * 
 *         elif isinstance(o, int):             # <<<<<<<<<<<<<<
 *             if o >= 0:
 *                 return self.mp_encode_uint(p, <uint64_t> o)
 */
  }

  /* "asynctnt/iproto/buffer.pyx":378
 *                 return self.mp_encode_int(p, <int64_t> o)
 * 
 *         elif isinstance(o, bytes):             # <<<<<<<<<<<<<<
 *             o_string_str = NULL
 *             o_string_len = 0
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_o); 
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":379
 * 
 *         elif isinstance(o, bytes):
 *             o_string_str = NULL             # <<<<<<<<<<<<<<
 *             o_string_len = 0
 *             cpython.bytes.PyBytes_AsStringAndSize(o,
 */
    __pyx_v_o_string_str = NULL;

    /* "asynctnt/iproto/buffer.pyx":380
 *         elif isinstance(o, bytes):
 *             o_string_str = NULL
 *             o_string_len = 0             # <<<<<<<<<<<<<<
 *             cpython.bytes.PyBytes_AsStringAndSize(o,
 *                                                   &o_string_str,
 */
    __pyx_v_o_string_len = 0;

    /* "asynctnt/iproto/buffer.pyx":381
 *             o_string_str = NULL
 *             o_string_len = 0
 *             cpython.bytes.PyBytes_AsStringAndSize(o,             # <<<<<<<<<<<<<<
 *                                                   &o_string_str,
 *                                                   &o_string_len)
 */
    __pyx_t_7 = PyBytes_AsStringAndSize(__pyx_v_o, (&__pyx_v_o_string_str), ((Py_ssize_t *)(&__pyx_v_o_string_len))); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(3, 381, __pyx_L1_error)

    /* "asynctnt/iproto/buffer.pyx":384
 *                                                   &o_string_str,
 *                                                   &o_string_len)
 *             return self.mp_encode_bin(p, o_string_str, <uint32_t> o_string_len)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(o, str):
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_bin(__pyx_v_self, __pyx_v_p, __pyx_v_o_string_str, ((uint32_t)__pyx_v_o_string_len)); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 384, __pyx_L1_error)
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":378
 *                 return self.mp_encode_int(p, <int64_t> o)
 * 
 *         elif isinstance(o, bytes):             # <<<<<<<<<<<<<<
 *             o_string_str = NULL
 *             o_string_len = 0
 */
  }

  /* "asynctnt/iproto/buffer.pyx":386
 *             return self.mp_encode_bin(p, o_string_str, <uint32_t> o_string_len)
 * 
 *         elif isinstance(o, str):             # <<<<<<<<<<<<<<
 *             o_string_temp = encode_unicode_string(o, self._encoding)
 *             o_string_str = NULL
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_o); 
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":387
 * 
 *         elif isinstance(o, str):
 *             o_string_temp = encode_unicode_string(o, self._encoding)             # <<<<<<<<<<<<<<
 *             o_string_str = NULL
 *             o_string_len = 0
 */
    __pyx_t_4 = __pyx_v_self->_encoding;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_8 = __pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string(__pyx_v_o, ((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_o_string_temp = ((PyObject*)__pyx_t_8);
    __pyx_t_8 = 0;

    /* "asynctnt/iproto/buffer.pyx":388
 *         elif isinstance(o, str):
 *             o_string_temp = encode_unicode_string(o, self._encoding)
 *             o_string_str = NULL             # <<<<<<<<<<<<<<
 *             o_string_len = 0
 *             cpython.bytes.PyBytes_AsStringAndSize(o_string_temp,
 */
    __pyx_v_o_string_str = NULL;

    /* "asynctnt/iproto/buffer.pyx":389
 *             o_string_temp = encode_unicode_string(o, self._encoding)
 *             o_string_str = NULL
 *             o_string_len = 0             # <<<<<<<<<<<<<<
 *             cpython.bytes.PyBytes_AsStringAndSize(o_string_temp,
 *                                                   &o_string_str,
 */
    __pyx_v_o_string_len = 0;

    /* "asynctnt/iproto/buffer.pyx":390
 *             o_string_str = NULL
 *             o_string_len = 0
 *             cpython.bytes.PyBytes_AsStringAndSize(o_string_temp,             # <<<<<<<<<<<<<<
 *                                                   &o_string_str,
 *                                                   &o_string_len)
 */
    __pyx_t_7 = PyBytes_AsStringAndSize(__pyx_v_o_string_temp, (&__pyx_v_o_string_str), ((Py_ssize_t *)(&__pyx_v_o_string_len))); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(3, 390, __pyx_L1_error)

    /* "asynctnt/iproto/buffer.pyx":394
 *                                                   &o_string_len)
 * 
 *             p = self.mp_encode_str(p, o_string_str, <uint32_t> o_string_len)             # <<<<<<<<<<<<<<
 *             return p
 * 
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_str(__pyx_v_self, __pyx_v_p, __pyx_v_o_string_str, ((uint32_t)__pyx_v_o_string_len)); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 394, __pyx_L1_error)
    __pyx_v_p = __pyx_t_2;

    /* "asynctnt/iproto/buffer.pyx":395
 * 
 *             p = self.mp_encode_str(p, o_string_str, <uint32_t> o_string_len)
 *             return p             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(o, list):
 */
    __pyx_r = __pyx_v_p;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":386
 *             return self.mp_encode_bin(p, o_string_str, <uint32_t> o_string_len)
 * 
 *         elif isinstance(o, str):             # <<<<<<<<<<<<<<
 *             o_string_temp = encode_unicode_string(o, self._encoding)
 *             o_string_str = NULL
 */
  }

  /* "asynctnt/iproto/buffer.pyx":397
 *             return p
 * 
 *         elif isinstance(o, list):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_list(p, <list> o)
 * 
 */
  __pyx_t_1 = PyList_Check(__pyx_v_o); 
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":398
 * 
 *         elif isinstance(o, list):
 *             return self.mp_encode_list(p, <list> o)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(o, tuple):
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_list(__pyx_v_self, __pyx_v_p, ((PyObject*)__pyx_v_o)); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 398, __pyx_L1_error)
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":397
 *             return p
 * 
 *         elif isinstance(o, list):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_list(p, <list> o)
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":400
 *             return self.mp_encode_list(p, <list> o)
 * 
 *         elif isinstance(o, tuple):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_tuple(p, <tuple> o)
 * 
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_o); 
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":401
 * 
 *         elif isinstance(o, tuple):
 *             return self.mp_encode_tuple(p, <tuple> o)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(o, dict):
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_tuple(__pyx_v_self, __pyx_v_p, ((PyObject*)__pyx_v_o)); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 401, __pyx_L1_error)
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":400
 *             return self.mp_encode_list(p, <list> o)
 * 
 *         elif isinstance(o, tuple):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_tuple(p, <tuple> o)
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":403
 *             return self.mp_encode_tuple(p, <tuple> o)
 * 
 *         elif isinstance(o, dict):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_dict(p, <dict> o)
 * 
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_o); 
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":404
 * 
 *         elif isinstance(o, dict):
 *             return self.mp_encode_dict(p, <dict> o)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(o, datetime):
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_dict(__pyx_v_self, __pyx_v_p, ((PyObject*)__pyx_v_o)); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 404, __pyx_L1_error)
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":403
 *             return self.mp_encode_tuple(p, <tuple> o)
 * 
 *         elif isinstance(o, dict):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_dict(p, <dict> o)
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":406
 *             return self.mp_encode_dict(p, <dict> o)
 * 
 *         elif isinstance(o, datetime):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_datetime(p, o)
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_7cpython_8datetime_datetime); 
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":407
 * 
 *         elif isinstance(o, datetime):
 *             return self.mp_encode_datetime(p, o)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(o, Decimal):
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_datetime(__pyx_v_self, __pyx_v_p, __pyx_v_o); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 407, __pyx_L1_error)
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":406
 *             return self.mp_encode_dict(p, <dict> o)
 * 
 *         elif isinstance(o, datetime):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_datetime(p, o)
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":409
 *             return self.mp_encode_datetime(p, o)
 * 
 *         elif isinstance(o, Decimal):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_decimal(p, o)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_o, __pyx_t_8); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 409, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__pyx_t_1) {

    /* "asynctnt/iproto/buffer.pyx":410
 * 
 *         elif isinstance(o, Decimal):
 *             return self.mp_encode_decimal(p, o)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(o, UUID):
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_decimal(__pyx_v_self, __pyx_v_p, __pyx_v_o); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 410, __pyx_L1_error)
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":409
 *             return self.mp_encode_datetime(p, o)
 * 
 *         elif isinstance(o, Decimal):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_decimal(p, o)
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":412
 *             return self.mp_encode_decimal(p, o)
 * 
 *         elif isinstance(o, UUID):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_uuid(p, o)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_UUID); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_o, __pyx_t_8); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (likely(__pyx_t_1)) {

    /* "asynctnt/iproto/buffer.pyx":413
 * 
 *         elif isinstance(o, UUID):
 *             return self.mp_encode_uuid(p, o)             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uuid(__pyx_v_self, __pyx_v_p, __pyx_v_o); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(3, 413, __pyx_L1_error)
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/buffer.pyx":412
 *             return self.mp_encode_decimal(p, o)
 * 
 *         elif isinstance(o, UUID):             # <<<<<<<<<<<<<<
 *             return self.mp_encode_uuid(p, o)
 * 
 */
  }

  /* "asynctnt/iproto/buffer.pyx":416
 * 
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 'Type `{}` is not supported for encoding'.format(type(o)))
 */
  /*else*/ {

    /* "asynctnt/iproto/buffer.pyx":417
 *         else:
 *             raise TypeError(
 *                 'Type `{}` is not supported for encoding'.format(type(o)))             # <<<<<<<<<<<<<<
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Type_is_not_supported_for_encodi, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, ((PyObject *)Py_TYPE(__pyx_v_o))};
      __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }

    /* "asynctnt/iproto/buffer.pyx":416
 * 
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 'Type `{}` is not supported for encoding'.format(type(o)))
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(3, 416, __pyx_L1_error)
  }

  /* "asynctnt/iproto/buffer.pyx":357
 *         return p
 * 
 *     cdef char *mp_encode_obj(self, char *p, object o) except NULL:             # <<<<<<<<<<<<<<
 *         cdef:
 *             bytes o_string_temp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.mp_encode_obj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_o_string_temp);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11WriteBuffer_11__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_10__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11WriteBuffer_13__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_12__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11WriteBuffer_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.WriteBuffer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/rbuffer.pyx":9
 * @cython.final
 * cdef class ReadBuffer:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.buf = NULL
 *         self.initial_buffer_size = 0
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer___cinit__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "asynctnt/iproto/rbuffer.pyx":10
 * cdef class ReadBuffer:
 *     def __cinit__(self):
 *         self.buf = NULL             # <<<<<<<<<<<<<<
 *         self.initial_buffer_size = 0
 *         self.len = 0
 */
  __pyx_v_self->buf = NULL;

  /* "asynctnt/iproto/rbuffer.pyx":11
 *     def __cinit__(self):
 *         self.buf = NULL
 *         self.initial_buffer_size = 0             # <<<<<<<<<<<<<<
 *         self.len = 0
 *         self.use = 0
 */
  __pyx_v_self->initial_buffer_size = 0;

  /* "asynctnt/iproto/rbuffer.pyx":12
 *         self.buf = NULL
 *         self.initial_buffer_size = 0
 *         self.len = 0             # <<<<<<<<<<<<<<
 *         self.use = 0
 *         self.encoding = None
 */
  __pyx_v_self->len = 0;

  /* "asynctnt/iproto/rbuffer.pyx":13
 *         self.initial_buffer_size = 0
 *         self.len = 0
 *         self.use = 0             # <<<<<<<<<<<<<<
 *         self.encoding = None
 * 
 */
  __pyx_v_self->use = 0;

  /* "asynctnt/iproto/rbuffer.pyx":14
 *         self.len = 0
 *         self.use = 0
 *         self.encoding = None             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->encoding);
  __Pyx_DECREF(__pyx_v_self->encoding);
  __pyx_v_self->encoding = ((PyObject*)Py_None);

  /* "asynctnt/iproto/rbuffer.pyx":9
 * @cython.final
 * cdef class ReadBuffer:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.buf = NULL
 *         self.initial_buffer_size = 0
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/rbuffer.pyx":17
 * 
 *     @staticmethod
 *     cdef ReadBuffer create(str encoding, size_t initial_buffer_size=0x80000):             # <<<<<<<<<<<<<<
 *         cdef ReadBuffer b
 *         b = ReadBuffer.__new__(ReadBuffer)
 */

static struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_create(PyObject *__pyx_v_encoding, struct __pyx_opt_args_8asynctnt_6iproto_8protocol_10ReadBuffer_create *__pyx_optional_args) {
  size_t __pyx_v_initial_buffer_size = ((size_t)0x80000);
  struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_b = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_initial_buffer_size = __pyx_optional_args->initial_buffer_size;
    }
  }

  /* "asynctnt/iproto/rbuffer.pyx":19
 *     cdef ReadBuffer create(str encoding, size_t initial_buffer_size=0x80000):
 *         cdef ReadBuffer b
 *         b = ReadBuffer.__new__(ReadBuffer)             # <<<<<<<<<<<<<<
 * 
 *         b.buf = <char*>PyMem_Malloc(sizeof(char) * <size_t>initial_buffer_size)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_ReadBuffer(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 19, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_b = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/rbuffer.pyx":21
 *         b = ReadBuffer.__new__(ReadBuffer)
 * 
 *         b.buf = <char*>PyMem_Malloc(sizeof(char) * <size_t>initial_buffer_size)             # <<<<<<<<<<<<<<
 *         if b.buf is NULL:
 *             raise MemoryError
 */
  __pyx_v_b->buf = ((char *)PyMem_Malloc(((sizeof(char)) * ((size_t)__pyx_v_initial_buffer_size))));

  /* "asynctnt/iproto/rbuffer.pyx":22
 * 
 *         b.buf = <char*>PyMem_Malloc(sizeof(char) * <size_t>initial_buffer_size)
 *         if b.buf is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 * 
 */
  __pyx_t_2 = (__pyx_v_b->buf == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "asynctnt/iproto/rbuffer.pyx":23
 *         b.buf = <char*>PyMem_Malloc(sizeof(char) * <size_t>initial_buffer_size)
 *         if b.buf is NULL:
 *             raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         b.initial_buffer_size = initial_buffer_size
 */
    PyErr_NoMemory(); __PYX_ERR(12, 23, __pyx_L1_error)

    /* "asynctnt/iproto/rbuffer.pyx":22
 * 
 *         b.buf = <char*>PyMem_Malloc(sizeof(char) * <size_t>initial_buffer_size)
 *         if b.buf is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 * 
 */
  }

  /* "asynctnt/iproto/rbuffer.pyx":25
 *             raise MemoryError
 * 
 *         b.initial_buffer_size = initial_buffer_size             # <<<<<<<<<<<<<<
 *         b.len = initial_buffer_size
 *         b.use = 0
 */
  __pyx_v_b->initial_buffer_size = __pyx_v_initial_buffer_size;

  /* "asynctnt/iproto/rbuffer.pyx":26
 * 
 *         b.initial_buffer_size = initial_buffer_size
 *         b.len = initial_buffer_size             # <<<<<<<<<<<<<<
 *         b.use = 0
 *         b.encoding = encoding
 */
  __pyx_v_b->len = __pyx_v_initial_buffer_size;

  /* "asynctnt/iproto/rbuffer.pyx":27
 *         b.initial_buffer_size = initial_buffer_size
 *         b.len = initial_buffer_size
 *         b.use = 0             # <<<<<<<<<<<<<<
 *         b.encoding = encoding
 *         return b
 */
  __pyx_v_b->use = 0;

  /* "asynctnt/iproto/rbuffer.pyx":28
 *         b.len = initial_buffer_size
 *         b.use = 0
 *         b.encoding = encoding             # <<<<<<<<<<<<<<
 *         return b
 * 
 */
  __Pyx_INCREF(__pyx_v_encoding);
  __Pyx_GIVEREF(__pyx_v_encoding);
  __Pyx_GOTREF(__pyx_v_b->encoding);
  __Pyx_DECREF(__pyx_v_b->encoding);
  __pyx_v_b->encoding = __pyx_v_encoding;

  /* "asynctnt/iproto/rbuffer.pyx":29
 *         b.use = 0
 *         b.encoding = encoding
 *         return b             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_b);
  __pyx_r = __pyx_v_b;
  goto __pyx_L0;

  /* "asynctnt/iproto/rbuffer.pyx":17
 * 
 *     @staticmethod
 *     cdef ReadBuffer create(str encoding, size_t initial_buffer_size=0x80000):             # <<<<<<<<<<<<<<
 *         cdef ReadBuffer b
 *         b = ReadBuffer.__new__(ReadBuffer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ReadBuffer.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_b);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/rbuffer.pyx":31
 *         return b
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.buf is not NULL:
 *             PyMem_Free(self.buf)
 */

/* Python wrapper */
static void __pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer_2__dealloc__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer_2__dealloc__(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self) {
  int __pyx_t_1;

  /* "asynctnt/iproto/rbuffer.pyx":32
 * 
 *     def __dealloc__(self):
 *         if self.buf is not NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.buf)
 *             self.buf = NULL
 */
  __pyx_t_1 = (__pyx_v_self->buf != NULL);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/rbuffer.pyx":33
 *     def __dealloc__(self):
 *         if self.buf is not NULL:
 *             PyMem_Free(self.buf)             # <<<<<<<<<<<<<<
 *             self.buf = NULL
 *         self.initial_buffer_size = 0
 */
    PyMem_Free(__pyx_v_self->buf);

    /* "asynctnt/iproto/rbuffer.pyx":34
 *         if self.buf is not NULL:
 *             PyMem_Free(self.buf)
 *             self.buf = NULL             # <<<<<<<<<<<<<<
 *         self.initial_buffer_size = 0
 *         self.len = 0
 */
    __pyx_v_self->buf = NULL;

    /* "asynctnt/iproto/rbuffer.pyx":32
 * 
 *     def __dealloc__(self):
 *         if self.buf is not NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.buf)
 *             self.buf = NULL
 */
  }

  /* "asynctnt/iproto/rbuffer.pyx":35
 *             PyMem_Free(self.buf)
 *             self.buf = NULL
 *         self.initial_buffer_size = 0             # <<<<<<<<<<<<<<
 *         self.len = 0
 *         self.use = 0
 */
  __pyx_v_self->initial_buffer_size = 0;

  /* "asynctnt/iproto/rbuffer.pyx":36
 *             self.buf = NULL
 *         self.initial_buffer_size = 0
 *         self.len = 0             # <<<<<<<<<<<<<<
 *         self.use = 0
 * 
 */
  __pyx_v_self->len = 0;

  /* "asynctnt/iproto/rbuffer.pyx":37
 *         self.initial_buffer_size = 0
 *         self.len = 0
 *         self.use = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef void _reallocate(self, size_t new_size) except *:
 */
  __pyx_v_self->use = 0;

  /* "asynctnt/iproto/rbuffer.pyx":31
 *         return b
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.buf is not NULL:
 *             PyMem_Free(self.buf)
 */

  /* function exit code */
}

/* "asynctnt/iproto/rbuffer.pyx":39
 *         self.use = 0
 * 
 *     cdef void _reallocate(self, size_t new_size) except *:             # <<<<<<<<<<<<<<
 *         cdef char *new_buf
 * 
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer__reallocate(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, size_t __pyx_v_new_size) {
  char *__pyx_v_new_buf;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/rbuffer.pyx":43
 * 
 *         # print('ReadBuffer reallocate: {}'.format(new_size))
 *         new_buf = <char*>PyMem_Realloc(<void*>self.buf, <size_t>new_size)             # <<<<<<<<<<<<<<
 *         if new_buf is NULL:
 *             PyMem_Free(self.buf)
 */
  __pyx_v_new_buf = ((char *)PyMem_Realloc(((void *)__pyx_v_self->buf), ((size_t)__pyx_v_new_size)));

  /* "asynctnt/iproto/rbuffer.pyx":44
 *         # print('ReadBuffer reallocate: {}'.format(new_size))
 *         new_buf = <char*>PyMem_Realloc(<void*>self.buf, <size_t>new_size)
 *         if new_buf is NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.buf)
 *             self.buf = NULL
 */
  __pyx_t_1 = (__pyx_v_new_buf == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "asynctnt/iproto/rbuffer.pyx":45
 *         new_buf = <char*>PyMem_Realloc(<void*>self.buf, <size_t>new_size)
 *         if new_buf is NULL:
 *             PyMem_Free(self.buf)             # <<<<<<<<<<<<<<
 *             self.buf = NULL
 *             self.initial_buffer_size = 0
 */
    PyMem_Free(__pyx_v_self->buf);

    /* "asynctnt/iproto/rbuffer.pyx":46
 *         if new_buf is NULL:
 *             PyMem_Free(self.buf)
 *             self.buf = NULL             # <<<<<<<<<<<<<<
 *             self.initial_buffer_size = 0
 *             self.len = 0
 */
    __pyx_v_self->buf = NULL;

    /* "asynctnt/iproto/rbuffer.pyx":47
 *             PyMem_Free(self.buf)
 *             self.buf = NULL
 *             self.initial_buffer_size = 0             # <<<<<<<<<<<<<<
 *             self.len = 0
 *             self.use = 0
 */
    __pyx_v_self->initial_buffer_size = 0;

    /* "asynctnt/iproto/rbuffer.pyx":48
 *             self.buf = NULL
 *             self.initial_buffer_size = 0
 *             self.len = 0             # <<<<<<<<<<<<<<
 *             self.use = 0
 *             raise MemoryError
 */
    __pyx_v_self->len = 0;

    /* "asynctnt/iproto/rbuffer.pyx":49
 *             self.initial_buffer_size = 0
 *             self.len = 0
 *             self.use = 0             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         self.buf = new_buf
 */
    __pyx_v_self->use = 0;

    /* "asynctnt/iproto/rbuffer.pyx":50
 *             self.len = 0
 *             self.use = 0
 *             raise MemoryError             # <<<<<<<<<<<<<<
 *         self.buf = new_buf
 *         self.len = new_size
 */
    PyErr_NoMemory(); __PYX_ERR(12, 50, __pyx_L1_error)

    /* "asynctnt/iproto/rbuffer.pyx":44
 *         # print('ReadBuffer reallocate: {}'.format(new_size))
 *         new_buf = <char*>PyMem_Realloc(<void*>self.buf, <size_t>new_size)
 *         if new_buf is NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.buf)
 *             self.buf = NULL
 */
  }

  /* "asynctnt/iproto/rbuffer.pyx":51
 *             self.use = 0
 *             raise MemoryError
 *         self.buf = new_buf             # <<<<<<<<<<<<<<
 *         self.len = new_size
 * 
 */
  __pyx_v_self->buf = __pyx_v_new_buf;

  /* "asynctnt/iproto/rbuffer.pyx":52
 *             raise MemoryError
 *         self.buf = new_buf
 *         self.len = new_size             # <<<<<<<<<<<<<<
 * 
 *     cdef int extend(self, const char *data, size_t len) except -1:
 */
  __pyx_v_self->len = __pyx_v_new_size;

  /* "asynctnt/iproto/rbuffer.pyx":39
 *         self.use = 0
 * 
 *     cdef void _reallocate(self, size_t new_size) except *:             # <<<<<<<<<<<<<<
 *         cdef char *new_buf
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ReadBuffer._reallocate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "asynctnt/iproto/rbuffer.pyx":54
 *         self.len = new_size
 * 
 *     cdef int extend(self, const char *data, size_t len) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t new_size
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_extend(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, char const *__pyx_v_data, size_t __pyx_v_len) {
  size_t __pyx_v_new_size;
  size_t __pyx_v_dealloc_threshold;
  int __pyx_r;
  int __pyx_t_1;
  uint32_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/rbuffer.pyx":58
 *             size_t new_size
 *             size_t dealloc_threshold
 *         new_size = self.use + len             # <<<<<<<<<<<<<<
 *         dealloc_threshold = self.len // _DEALLOCATE_RATIO
 *         if new_size > self.len:
 */
  __pyx_v_new_size = (__pyx_v_self->use + __pyx_v_len);

  /* "asynctnt/iproto/rbuffer.pyx":59
 *             size_t dealloc_threshold
 *         new_size = self.use + len
 *         dealloc_threshold = self.len // _DEALLOCATE_RATIO             # <<<<<<<<<<<<<<
 *         if new_size > self.len:
 *             self._reallocate(
 */
  __pyx_v_dealloc_threshold = (__pyx_v_self->len / 4);

  /* "asynctnt/iproto/rbuffer.pyx":60
 *         new_size = self.use + len
 *         dealloc_threshold = self.len // _DEALLOCATE_RATIO
 *         if new_size > self.len:             # <<<<<<<<<<<<<<
 *             self._reallocate(
 *                 size_t_max(nearest_power_of_2(new_size), self.len << 1)
 */
  __pyx_t_1 = (__pyx_v_new_size > __pyx_v_self->len);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/rbuffer.pyx":62
 *         if new_size > self.len:
 *             self._reallocate(
 *                 size_t_max(nearest_power_of_2(new_size), self.len << 1)             # <<<<<<<<<<<<<<
 *             )
 *         elif dealloc_threshold >= self.initial_buffer_size \
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_nearest_power_of_2(__pyx_v_new_size); if (unlikely(__pyx_t_2 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(12, 62, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_size_t_max(__pyx_t_2, (__pyx_v_self->len << 1)); if (unlikely(__pyx_t_3 == ((size_t)-1) && PyErr_Occurred())) __PYX_ERR(12, 62, __pyx_L1_error)

    /* "asynctnt/iproto/rbuffer.pyx":61
 *         dealloc_threshold = self.len // _DEALLOCATE_RATIO
 *         if new_size > self.len:
 *             self._reallocate(             # <<<<<<<<<<<<<<
 *                 size_t_max(nearest_power_of_2(new_size), self.len << 1)
 *             )
 */
    __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer__reallocate(__pyx_v_self, __pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 61, __pyx_L1_error)

    /* "asynctnt/iproto/rbuffer.pyx":60
 *         new_size = self.use + len
 *         dealloc_threshold = self.len // _DEALLOCATE_RATIO
 *         if new_size > self.len:             # <<<<<<<<<<<<<<
 *             self._reallocate(
 *                 size_t_max(nearest_power_of_2(new_size), self.len << 1)
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/rbuffer.pyx":64
 *                 size_t_max(nearest_power_of_2(new_size), self.len << 1)
 *             )
 *         elif dealloc_threshold >= self.initial_buffer_size \             # <<<<<<<<<<<<<<
 *                 and new_size < dealloc_threshold:
 *             self._reallocate(dealloc_threshold)
 */
  __pyx_t_4 = (__pyx_v_dealloc_threshold >= __pyx_v_self->initial_buffer_size);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }

  /* "asynctnt/iproto/rbuffer.pyx":65
 *             )
 *         elif dealloc_threshold >= self.initial_buffer_size \
 *                 and new_size < dealloc_threshold:             # <<<<<<<<<<<<<<
 *             self._reallocate(dealloc_threshold)
 * 
 */
  __pyx_t_4 = (__pyx_v_new_size < __pyx_v_dealloc_threshold);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;

  /* "asynctnt/iproto/rbuffer.pyx":64
 *                 size_t_max(nearest_power_of_2(new_size), self.len << 1)
 *             )
 *         elif dealloc_threshold >= self.initial_buffer_size \             # <<<<<<<<<<<<<<
 *                 and new_size < dealloc_threshold:
 *             self._reallocate(dealloc_threshold)
 */
  if (__pyx_t_1) {

    /* "asynctnt/iproto/rbuffer.pyx":66
 *         elif dealloc_threshold >= self.initial_buffer_size \
 *                 and new_size < dealloc_threshold:
 *             self._reallocate(dealloc_threshold)             # <<<<<<<<<<<<<<
 * 
 *         memcpy(&self.buf[self.use], data, len)
 */
    __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer__reallocate(__pyx_v_self, __pyx_v_dealloc_threshold); if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 66, __pyx_L1_error)

    /* "asynctnt/iproto/rbuffer.pyx":64
 *                 size_t_max(nearest_power_of_2(new_size), self.len << 1)
 *             )
 *         elif dealloc_threshold >= self.initial_buffer_size \             # <<<<<<<<<<<<<<
 *                 and new_size < dealloc_threshold:
 *             self._reallocate(dealloc_threshold)
 */
  }
  __pyx_L3:;

  /* "asynctnt/iproto/rbuffer.pyx":68
 *             self._reallocate(dealloc_threshold)
 * 
 *         memcpy(&self.buf[self.use], data, len)             # <<<<<<<<<<<<<<
 *         self.use += len
 *         return 0
 */
  (void)(memcpy((&(__pyx_v_self->buf[__pyx_v_self->use])), __pyx_v_data, __pyx_v_len));

  /* "asynctnt/iproto/rbuffer.pyx":69
 * 
 *         memcpy(&self.buf[self.use], data, len)
 *         self.use += len             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->use = (__pyx_v_self->use + __pyx_v_len);

  /* "asynctnt/iproto/rbuffer.pyx":70
 *         memcpy(&self.buf[self.use], data, len)
 *         self.use += len
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef void move(self, size_t pos):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/rbuffer.pyx":54
 *         self.len = new_size
 * 
 *     cdef int extend(self, const char *data, size_t len) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t new_size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ReadBuffer.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/rbuffer.pyx":72
 *         return 0
 * 
 *     cdef void move(self, size_t pos):             # <<<<<<<<<<<<<<
 *         cdef size_t delta = self.use - pos
 *         memmove(self.buf, &self.buf[pos], delta)
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_move(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, size_t __pyx_v_pos) {
  size_t __pyx_v_delta;

  /* "asynctnt/iproto/rbuffer.pyx":73
 * 
 *     cdef void move(self, size_t pos):
 *         cdef size_t delta = self.use - pos             # <<<<<<<<<<<<<<
 *         memmove(self.buf, &self.buf[pos], delta)
 *         self.use = delta
 */
  __pyx_v_delta = (__pyx_v_self->use - __pyx_v_pos);

  /* "asynctnt/iproto/rbuffer.pyx":74
 *     cdef void move(self, size_t pos):
 *         cdef size_t delta = self.use - pos
 *         memmove(self.buf, &self.buf[pos], delta)             # <<<<<<<<<<<<<<
 *         self.use = delta
 * 
 */
  (void)(memmove(__pyx_v_self->buf, (&(__pyx_v_self->buf[__pyx_v_pos])), __pyx_v_delta));

  /* "asynctnt/iproto/rbuffer.pyx":75
 *         cdef size_t delta = self.use - pos
 *         memmove(self.buf, &self.buf[pos], delta)
 *         self.use = delta             # <<<<<<<<<<<<<<
 * 
 *     cdef void move_offset(self, ssize_t offset, size_t size) except *:
 */
  __pyx_v_self->use = __pyx_v_delta;

  /* "asynctnt/iproto/rbuffer.pyx":72
 *         return 0
 * 
 *     cdef void move(self, size_t pos):             # <<<<<<<<<<<<<<
 *         cdef size_t delta = self.use - pos
 *         memmove(self.buf, &self.buf[pos], delta)
 */

  /* function exit code */
}

/* "asynctnt/iproto/rbuffer.pyx":77
 *         self.use = delta
 * 
 *     cdef void move_offset(self, ssize_t offset, size_t size) except *:             # <<<<<<<<<<<<<<
 *         cdef size_t dealloc_threshold = self.len // _DEALLOCATE_RATIO
 *         if offset == 0:
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_move_offset(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, Py_ssize_t __pyx_v_offset, size_t __pyx_v_size) {
  size_t __pyx_v_dealloc_threshold;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("move_offset", 1);

  /* "asynctnt/iproto/rbuffer.pyx":78
 * 
 *     cdef void move_offset(self, ssize_t offset, size_t size) except *:
 *         cdef size_t dealloc_threshold = self.len // _DEALLOCATE_RATIO             # <<<<<<<<<<<<<<
 *         if offset == 0:
 *             return
 */
  __pyx_v_dealloc_threshold = (__pyx_v_self->len / 4);

  /* "asynctnt/iproto/rbuffer.pyx":79
 *     cdef void move_offset(self, ssize_t offset, size_t size) except *:
 *         cdef size_t dealloc_threshold = self.len // _DEALLOCATE_RATIO
 *         if offset == 0:             # <<<<<<<<<<<<<<
 *             return
 *         assert offset > 0, \
 */
  __pyx_t_1 = (__pyx_v_offset == 0);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/rbuffer.pyx":80
 *         cdef size_t dealloc_threshold = self.len // _DEALLOCATE_RATIO
 *         if offset == 0:
 *             return             # <<<<<<<<<<<<<<
 *         assert offset > 0, \
 *             'Offset incorrect. Got: {}. use:{}, len:{}'.format(
 */
    goto __pyx_L0;

    /* "asynctnt/iproto/rbuffer.pyx":79
 *     cdef void move_offset(self, ssize_t offset, size_t size) except *:
 *         cdef size_t dealloc_threshold = self.len // _DEALLOCATE_RATIO
 *         if offset == 0:             # <<<<<<<<<<<<<<
 *             return
 *         assert offset > 0, \
 */
  }

  /* "asynctnt/iproto/rbuffer.pyx":81
 *         if offset == 0:
 *             return
 *         assert offset > 0, \             # <<<<<<<<<<<<<<
 *             'Offset incorrect. Got: {}. use:{}, len:{}'.format(
 *                 offset, self.use, self.len
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__pyx_v_offset > 0);
    if (unlikely(!__pyx_t_1)) {

      /* "asynctnt/iproto/rbuffer.pyx":82
 *             return
 *         assert offset > 0, \
 *             'Offset incorrect. Got: {}. use:{}, len:{}'.format(             # <<<<<<<<<<<<<<
 *                 offset, self.use, self.len
 *             )
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Offset_incorrect_Got_use_len, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 82, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "asynctnt/iproto/rbuffer.pyx":83
 *         assert offset > 0, \
 *             'Offset incorrect. Got: {}. use:{}, len:{}'.format(
 *                 offset, self.use, self.len             # <<<<<<<<<<<<<<
 *             )
 *         memmove(self.buf, &self.buf[offset], size)
 */
      __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_offset); if (unlikely(!__pyx_t_4)) __PYX_ERR(12, 83, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_self->use); if (unlikely(!__pyx_t_5)) __PYX_ERR(12, 83, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_self->len); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 83, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_7, __pyx_t_4, __pyx_t_5, __pyx_t_6};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_8, 3+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "asynctnt/iproto/rbuffer.pyx":82
 *             return
 *         assert offset > 0, \
 *             'Offset incorrect. Got: {}. use:{}, len:{}'.format(             # <<<<<<<<<<<<<<
 *                 offset, self.use, self.len
 *             )
 */
      __pyx_t_3 = PyTuple_Pack(1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 82, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_3, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(12, 81, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(12, 81, __pyx_L1_error)
  #endif

  /* "asynctnt/iproto/rbuffer.pyx":85
 *                 offset, self.use, self.len
 *             )
 *         memmove(self.buf, &self.buf[offset], size)             # <<<<<<<<<<<<<<
 * 
 *         if dealloc_threshold >= self.initial_buffer_size \
 */
  (void)(memmove(__pyx_v_self->buf, (&(__pyx_v_self->buf[__pyx_v_offset])), __pyx_v_size));

  /* "asynctnt/iproto/rbuffer.pyx":87
 *         memmove(self.buf, &self.buf[offset], size)
 * 
 *         if dealloc_threshold >= self.initial_buffer_size \             # <<<<<<<<<<<<<<
 *                 and size < dealloc_threshold:
 *             self._reallocate(dealloc_threshold)
 */
  __pyx_t_9 = (__pyx_v_dealloc_threshold >= __pyx_v_self->initial_buffer_size);
  if (__pyx_t_9) {
  } else {
    __pyx_t_1 = __pyx_t_9;
    goto __pyx_L5_bool_binop_done;
  }

  /* "asynctnt/iproto/rbuffer.pyx":88
 * 
 *         if dealloc_threshold >= self.initial_buffer_size \
 *                 and size < dealloc_threshold:             # <<<<<<<<<<<<<<
 *             self._reallocate(dealloc_threshold)
 * 
 */
  __pyx_t_9 = (__pyx_v_size < __pyx_v_dealloc_threshold);
  __pyx_t_1 = __pyx_t_9;
  __pyx_L5_bool_binop_done:;

  /* "asynctnt/iproto/rbuffer.pyx":87
 *         memmove(self.buf, &self.buf[offset], size)
 * 
 *         if dealloc_threshold >= self.initial_buffer_size \             # <<<<<<<<<<<<<<
 *                 and size < dealloc_threshold:
 *             self._reallocate(dealloc_threshold)
 */
  if (__pyx_t_1) {

    /* "asynctnt/iproto/rbuffer.pyx":89
 *         if dealloc_threshold >= self.initial_buffer_size \
 *                 and size < dealloc_threshold:
 *             self._reallocate(dealloc_threshold)             # <<<<<<<<<<<<<<
 * 
 *     cdef bytes get_slice(self, size_t begin, size_t end):
 */
    __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer__reallocate(__pyx_v_self, __pyx_v_dealloc_threshold); if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 89, __pyx_L1_error)

    /* "asynctnt/iproto/rbuffer.pyx":87
 *         memmove(self.buf, &self.buf[offset], size)
 * 
 *         if dealloc_threshold >= self.initial_buffer_size \             # <<<<<<<<<<<<<<
 *                 and size < dealloc_threshold:
 *             self._reallocate(dealloc_threshold)
 */
  }

  /* "asynctnt/iproto/rbuffer.pyx":77
 *         self.use = delta
 * 
 *     cdef void move_offset(self, ssize_t offset, size_t size) except *:             # <<<<<<<<<<<<<<
 *         cdef size_t dealloc_threshold = self.len // _DEALLOCATE_RATIO
 *         if offset == 0:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ReadBuffer.move_offset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/rbuffer.pyx":91
 *             self._reallocate(dealloc_threshold)
 * 
 *     cdef bytes get_slice(self, size_t begin, size_t end):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ssize_t diff
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, size_t __pyx_v_begin, size_t __pyx_v_end) {
  Py_ssize_t __pyx_v_diff;
  char *__pyx_v_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice", 1);

  /* "asynctnt/iproto/rbuffer.pyx":95
 *             ssize_t diff
 *             char *p
 *         p = &self.buf[begin]             # <<<<<<<<<<<<<<
 *         diff = end - begin
 *         return <bytes>p[:diff]
 */
  __pyx_v_p = (&(__pyx_v_self->buf[__pyx_v_begin]));

  /* "asynctnt/iproto/rbuffer.pyx":96
 *             char *p
 *         p = &self.buf[begin]
 *         diff = end - begin             # <<<<<<<<<<<<<<
 *         return <bytes>p[:diff]
 * 
 */
  __pyx_v_diff = (__pyx_v_end - __pyx_v_begin);

  /* "asynctnt/iproto/rbuffer.pyx":97
 *         p = &self.buf[begin]
 *         diff = end - begin
 *         return <bytes>p[:diff]             # <<<<<<<<<<<<<<
 * 
 *     cdef bytes get_slice_begin(self, size_t begin):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_p + 0, __pyx_v_diff - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = ((PyObject*)__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/rbuffer.pyx":91
 *             self._reallocate(dealloc_threshold)
 * 
 *     cdef bytes get_slice(self, size_t begin, size_t end):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ssize_t diff
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ReadBuffer.get_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/rbuffer.pyx":99
 *         return <bytes>p[:diff]
 * 
 *     cdef bytes get_slice_begin(self, size_t begin):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ssize_t diff
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice_begin(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, size_t __pyx_v_begin) {
  Py_ssize_t __pyx_v_diff;
  char *__pyx_v_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice_begin", 1);

  /* "asynctnt/iproto/rbuffer.pyx":103
 *             ssize_t diff
 *             char *p
 *         p = &self.buf[begin]             # <<<<<<<<<<<<<<
 *         diff = self.use - begin
 *         return <bytes>p[:diff]
 */
  __pyx_v_p = (&(__pyx_v_self->buf[__pyx_v_begin]));

  /* "asynctnt/iproto/rbuffer.pyx":104
 *             char *p
 *         p = &self.buf[begin]
 *         diff = self.use - begin             # <<<<<<<<<<<<<<
 *         return <bytes>p[:diff]
 * 
 */
  __pyx_v_diff = (__pyx_v_self->use - __pyx_v_begin);

  /* "asynctnt/iproto/rbuffer.pyx":105
 *         p = &self.buf[begin]
 *         diff = self.use - begin
 *         return <bytes>p[:diff]             # <<<<<<<<<<<<<<
 * 
 *     cdef bytes get_slice_end(self, size_t end):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_p + 0, __pyx_v_diff - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = ((PyObject*)__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/rbuffer.pyx":99
 *         return <bytes>p[:diff]
 * 
 *     cdef bytes get_slice_begin(self, size_t begin):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ssize_t diff
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ReadBuffer.get_slice_begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/rbuffer.pyx":107
 *         return <bytes>p[:diff]
 * 
 *     cdef bytes get_slice_end(self, size_t end):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ssize_t diff
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice_end(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, size_t __pyx_v_end) {
  Py_ssize_t __pyx_v_diff;
  char *__pyx_v_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice_end", 1);

  /* "asynctnt/iproto/rbuffer.pyx":111
 *             ssize_t diff
 *             char *p
 *         p = &self.buf[0]             # <<<<<<<<<<<<<<
 *         diff = end - 0
 *         return <bytes>p[:diff]
 */
  __pyx_v_p = (&(__pyx_v_self->buf[0]));

  /* "asynctnt/iproto/rbuffer.pyx":112
 *             char *p
 *         p = &self.buf[0]
 *         diff = end - 0             # <<<<<<<<<<<<<<
 *         return <bytes>p[:diff]
 */
  __pyx_v_diff = (__pyx_v_end - 0);

  /* "asynctnt/iproto/rbuffer.pyx":113
 *         p = &self.buf[0]
 *         diff = end - 0
 *         return <bytes>p[:diff]             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_p + 0, __pyx_v_diff - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = ((PyObject*)__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/rbuffer.pyx":107
 *         return <bytes>p[:diff]
 * 
 *     cdef bytes get_slice_end(self, size_t end):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ssize_t diff
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ReadBuffer.get_slice_end", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_10ReadBuffer_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer_4__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ReadBuffer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_10ReadBuffer_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ReadBuffer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer_6__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_10ReadBuffer_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ReadBuffer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/base.pyx":15
 *     #     self.push_subscribe = False
 * 
 *     cdef inline WriteBuffer encode(self, bytes encoding):             # <<<<<<<<<<<<<<
 *         cdef WriteBuffer buffer = WriteBuffer.create(encoding)
 *         buffer.write_header(self.sync, self.op, self.schema_id, self.stream_id)
 */

static CYTHON_INLINE struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_encode(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self, PyObject *__pyx_v_encoding) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode", 1);

  /* "asynctnt/iproto/requests/base.pyx":16
 * 
 *     cdef inline WriteBuffer encode(self, bytes encoding):
 *         cdef WriteBuffer buffer = WriteBuffer.create(encoding)             # <<<<<<<<<<<<<<
 *         buffer.write_header(self.sync, self.op, self.schema_id, self.stream_id)
 *         self.encode_body(buffer)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_create(__pyx_v_encoding)); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buffer = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/requests/base.pyx":17
 *     cdef inline WriteBuffer encode(self, bytes encoding):
 *         cdef WriteBuffer buffer = WriteBuffer.create(encoding)
 *         buffer.write_header(self.sync, self.op, self.schema_id, self.stream_id)             # <<<<<<<<<<<<<<
 *         self.encode_body(buffer)
 *         buffer.write_length()
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_header(__pyx_v_buffer, __pyx_v_self->sync, __pyx_v_self->op, __pyx_v_self->schema_id, __pyx_v_self->stream_id); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(13, 17, __pyx_L1_error)

  /* "asynctnt/iproto/requests/base.pyx":18
 *         cdef WriteBuffer buffer = WriteBuffer.create(encoding)
 *         buffer.write_header(self.sync, self.op, self.schema_id, self.stream_id)
 *         self.encode_body(buffer)             # <<<<<<<<<<<<<<
 *         buffer.write_length()
 *         return buffer
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_self->__pyx_vtab)->encode_body(__pyx_v_self, __pyx_v_buffer); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(13, 18, __pyx_L1_error)

  /* "asynctnt/iproto/requests/base.pyx":19
 *         buffer.write_header(self.sync, self.op, self.schema_id, self.stream_id)
 *         self.encode_body(buffer)
 *         buffer.write_length()             # <<<<<<<<<<<<<<
 *         return buffer
 * 
 */
  __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_length(__pyx_v_buffer); if (unlikely(PyErr_Occurred())) __PYX_ERR(13, 19, __pyx_L1_error)

  /* "asynctnt/iproto/requests/base.pyx":20
 *         self.encode_body(buffer)
 *         buffer.write_length()
 *         return buffer             # <<<<<<<<<<<<<<
 * 
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_buffer);
  __pyx_r = __pyx_v_buffer;
  goto __pyx_L0;

  /* "asynctnt/iproto/requests/base.pyx":15
 *     #     self.push_subscribe = False
 * 
 *     cdef inline WriteBuffer encode(self, bytes encoding):             # <<<<<<<<<<<<<<
 *         cdef WriteBuffer buffer = WriteBuffer.create(encoding)
 *         buffer.write_header(self.sync, self.op, self.schema_id, self.stream_id)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseRequest.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_buffer);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/base.pyx":22
 *         return buffer
 * 
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_encode_body(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  int __pyx_r;

  /* "asynctnt/iproto/requests/base.pyx":23
 * 
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):  # pragma: nocover
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/requests/base.pyx":22
 *         return buffer
 * 
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/requests/base.pyx":25
 *         return 0
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return \
 *             '<Request op={} sync={} schema_id={} push_subscribe={}>'.format(
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11BaseRequest___repr__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11BaseRequest___repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "asynctnt/iproto/requests/base.pyx":26
 * 
 *     def __repr__(self):  # pragma: nocover
 *         return \             # <<<<<<<<<<<<<<
 *             '<Request op={} sync={} schema_id={} push_subscribe={}>'.format(
 *                 self.op,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "asynctnt/iproto/requests/base.pyx":27
 *     def __repr__(self):  # pragma: nocover
 *         return \
 *             '<Request op={} sync={} schema_id={} push_subscribe={}>'.format(             # <<<<<<<<<<<<<<
 *                 self.op,
 *                 self.sync,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Request_op_sync_schema_id_push, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "asynctnt/iproto/requests/base.pyx":28
 *         return \
 *             '<Request op={} sync={} schema_id={} push_subscribe={}>'.format(
 *                 self.op,             # <<<<<<<<<<<<<<
 *                 self.sync,
 *                 self.schema_id,
 */
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->op); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "asynctnt/iproto/requests/base.pyx":29
 *             '<Request op={} sync={} schema_id={} push_subscribe={}>'.format(
 *                 self.op,
 *                 self.sync,             # <<<<<<<<<<<<<<
 *                 self.schema_id,
 *                 self.push_subscribe
 */
  __pyx_t_4 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->sync); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "asynctnt/iproto/requests/base.pyx":30
 *                 self.op,
 *                 self.sync,
 *                 self.schema_id,             # <<<<<<<<<<<<<<
 *                 self.push_subscribe
 *             )
 */
  __pyx_t_5 = __Pyx_PyInt_From_int64_t(__pyx_v_self->schema_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(13, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "asynctnt/iproto/requests/base.pyx":31
 *                 self.sync,
 *                 self.schema_id,
 *                 self.push_subscribe             # <<<<<<<<<<<<<<
 *             )
 * 
 */
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->push_subscribe); if (unlikely(!__pyx_t_6)) __PYX_ERR(13, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 4+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/requests/base.pyx":25
 *         return 0
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return \
 *             '<Request op={} sync={} schema_id={} push_subscribe={}>'.format(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseRequest.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11BaseRequest_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11BaseRequest_2__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11BaseRequest_2__reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, ((PyObject *)__pyx_v_self->space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_v_self->timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->waiter);
  __Pyx_GIVEREF(__pyx_v_self->waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_v_self->waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, None), state
 */
  /*else*/ {
    __pyx_t_11 = (((PyObject *)__pyx_v_self->space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_BaseRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150999062);
    __Pyx_GIVEREF(__pyx_int_150999062);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_150999062)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, None), state
 *     else:
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BaseRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_BaseRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150999062);
    __Pyx_GIVEREF(__pyx_int_150999062);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_150999062)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BaseRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11BaseRequest_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11BaseRequest_4__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11BaseRequest_4__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BaseRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_BaseRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BaseRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/base.pyx":35
 * 
 * 
 * cdef char *encode_key_sequence(WriteBuffer buffer,             # <<<<<<<<<<<<<<
 *                                char *p, object t,
 *                                Metadata metadata,
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_encode_key_sequence(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer, char *__pyx_v_p, PyObject *__pyx_v_t, struct C_Metadata *__pyx_v_metadata, int __pyx_v_default_none) {
  PyObject *__pyx_v_msg = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  char *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_key_sequence", 1);

  /* "asynctnt/iproto/requests/base.pyx":39
 *                                Metadata metadata,
 *                                bint default_none) except NULL:
 *     if isinstance(t, list) or t is None:             # <<<<<<<<<<<<<<
 *         return buffer.mp_encode_list(p, <list> t)
 *     elif isinstance(t, tuple):
 */
  __pyx_t_2 = PyList_Check(__pyx_v_t); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_t == Py_None);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "asynctnt/iproto/requests/base.pyx":40
 *                                bint default_none) except NULL:
 *     if isinstance(t, list) or t is None:
 *         return buffer.mp_encode_list(p, <list> t)             # <<<<<<<<<<<<<<
 *     elif isinstance(t, tuple):
 *         return buffer.mp_encode_tuple(p, <tuple> t)
 */
    __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_list(__pyx_v_buffer, __pyx_v_p, ((PyObject*)__pyx_v_t)); if (unlikely(__pyx_t_3 == ((char *)NULL))) __PYX_ERR(13, 40, __pyx_L1_error)
    __pyx_r = __pyx_t_3;
    goto __pyx_L0;

    /* "asynctnt/iproto/requests/base.pyx":39
 *                                Metadata metadata,
 *                                bint default_none) except NULL:
 *     if isinstance(t, list) or t is None:             # <<<<<<<<<<<<<<
 *         return buffer.mp_encode_list(p, <list> t)
 *     elif isinstance(t, tuple):
 */
  }

  /* "asynctnt/iproto/requests/base.pyx":41
 *     if isinstance(t, list) or t is None:
 *         return buffer.mp_encode_list(p, <list> t)
 *     elif isinstance(t, tuple):             # <<<<<<<<<<<<<<
 *         return buffer.mp_encode_tuple(p, <tuple> t)
 *     elif isinstance(t, dict) and metadata is not None:
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_t); 
  if (__pyx_t_1) {

    /* "asynctnt/iproto/requests/base.pyx":42
 *         return buffer.mp_encode_list(p, <list> t)
 *     elif isinstance(t, tuple):
 *         return buffer.mp_encode_tuple(p, <tuple> t)             # <<<<<<<<<<<<<<
 *     elif isinstance(t, dict) and metadata is not None:
 *         return buffer.mp_encode_list(
 */
    __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_tuple(__pyx_v_buffer, __pyx_v_p, ((PyObject*)__pyx_v_t)); if (unlikely(__pyx_t_3 == ((char *)NULL))) __PYX_ERR(13, 42, __pyx_L1_error)
    __pyx_r = __pyx_t_3;
    goto __pyx_L0;

    /* "asynctnt/iproto/requests/base.pyx":41
 *     if isinstance(t, list) or t is None:
 *         return buffer.mp_encode_list(p, <list> t)
 *     elif isinstance(t, tuple):             # <<<<<<<<<<<<<<
 *         return buffer.mp_encode_tuple(p, <tuple> t)
 *     elif isinstance(t, dict) and metadata is not None:
 */
  }

  /* "asynctnt/iproto/requests/base.pyx":43
 *     elif isinstance(t, tuple):
 *         return buffer.mp_encode_tuple(p, <tuple> t)
 *     elif isinstance(t, dict) and metadata is not None:             # <<<<<<<<<<<<<<
 *         return buffer.mp_encode_list(
 *             p, dict_to_list_fields(<dict> t, metadata, default_none)
 */
  __pyx_t_2 = PyDict_Check(__pyx_v_t); 
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_2 = (((PyObject *)__pyx_v_metadata) != Py_None);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_1) {

    /* "asynctnt/iproto/requests/base.pyx":45
 *     elif isinstance(t, dict) and metadata is not None:
 *         return buffer.mp_encode_list(
 *             p, dict_to_list_fields(<dict> t, metadata, default_none)             # <<<<<<<<<<<<<<
 *         )
 *     else:
 */
    __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_dict_to_list_fields(((PyObject*)__pyx_v_t), __pyx_v_metadata, __pyx_v_default_none); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "asynctnt/iproto/requests/base.pyx":44
 *         return buffer.mp_encode_tuple(p, <tuple> t)
 *     elif isinstance(t, dict) and metadata is not None:
 *         return buffer.mp_encode_list(             # <<<<<<<<<<<<<<
 *             p, dict_to_list_fields(<dict> t, metadata, default_none)
 *         )
 */
    __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_list(__pyx_v_buffer, __pyx_v_p, ((PyObject*)__pyx_t_4)); if (unlikely(__pyx_t_3 == ((char *)NULL))) __PYX_ERR(13, 44, __pyx_L1_error)
    __pyx_r = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/requests/base.pyx":43
 *     elif isinstance(t, tuple):
 *         return buffer.mp_encode_tuple(p, <tuple> t)
 *     elif isinstance(t, dict) and metadata is not None:             # <<<<<<<<<<<<<<
 *         return buffer.mp_encode_list(
 *             p, dict_to_list_fields(<dict> t, metadata, default_none)
 */
  }

  /* "asynctnt/iproto/requests/base.pyx":48
 *         )
 *     else:
 *         if metadata is not None:             # <<<<<<<<<<<<<<
 *             msg = 'sequence must be either list, tuple or dict'
 *         else:
 */
  /*else*/ {
    __pyx_t_1 = (((PyObject *)__pyx_v_metadata) != Py_None);
    if (__pyx_t_1) {

      /* "asynctnt/iproto/requests/base.pyx":49
 *     else:
 *         if metadata is not None:
 *             msg = 'sequence must be either list, tuple or dict'             # <<<<<<<<<<<<<<
 *         else:
 *             msg = 'sequence must be either list or tuple'
 */
      __Pyx_INCREF(__pyx_kp_u_sequence_must_be_either_list_tup);
      __pyx_v_msg = __pyx_kp_u_sequence_must_be_either_list_tup;

      /* "asynctnt/iproto/requests/base.pyx":48
 *         )
 *     else:
 *         if metadata is not None:             # <<<<<<<<<<<<<<
 *             msg = 'sequence must be either list, tuple or dict'
 *         else:
 */
      goto __pyx_L8;
    }

    /* "asynctnt/iproto/requests/base.pyx":51
 *             msg = 'sequence must be either list, tuple or dict'
 *         else:
 *             msg = 'sequence must be either list or tuple'             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             '{}, got: {}'.format(msg, type(t))
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_kp_u_sequence_must_be_either_list_or);
      __pyx_v_msg = __pyx_kp_u_sequence_must_be_either_list_or;
    }
    __pyx_L8:;

    /* "asynctnt/iproto/requests/base.pyx":53
 *             msg = 'sequence must be either list or tuple'
 *         raise TypeError(
 *             '{}, got: {}'.format(msg, type(t))             # <<<<<<<<<<<<<<
 *         )
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_got, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(13, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_msg, ((PyObject *)Py_TYPE(__pyx_v_t))};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 53, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }

    /* "asynctnt/iproto/requests/base.pyx":52
 *         else:
 *             msg = 'sequence must be either list or tuple'
 *         raise TypeError(             # <<<<<<<<<<<<<<
 *             '{}, got: {}'.format(msg, type(t))
 *         )
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(13, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(13, 52, __pyx_L1_error)
  }

  /* "asynctnt/iproto/requests/base.pyx":35
 * 
 * 
 * cdef char *encode_key_sequence(WriteBuffer buffer,             # <<<<<<<<<<<<<<
 *                                char *p, object t,
 *                                Metadata metadata,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.encode_key_sequence", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11PingRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11PingRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11PingRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11PingRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11PingRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11PingRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, None), state
 */
  /*else*/ {
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_PingRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150999062);
    __Pyx_GIVEREF(__pyx_int_150999062);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_150999062)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, None), state
 *     else:
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PingRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_PingRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150999062);
    __Pyx_GIVEREF(__pyx_int_150999062);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_150999062)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PingRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PingRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11PingRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11PingRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11PingRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11PingRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PingRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11PingRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11PingRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PingRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PingRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PingRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PingRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/call.pyx":8
 * @cython.final
 * cdef class CallRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_11CallRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  char *__pyx_v_begin;
  char *__pyx_v_p;
  uint32_t __pyx_v_body_map_sz;
  uint32_t __pyx_v_max_body_len;
  PyObject *__pyx_v_func_name_temp = 0;
  char *__pyx_v_func_name_str;
  Py_ssize_t __pyx_v_func_name_len;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  char *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_body", 1);

  /* "asynctnt/iproto/requests/call.pyx":19
 *             ssize_t func_name_len
 * 
 *         func_name_str = NULL             # <<<<<<<<<<<<<<
 *         func_name_len = 0
 * 
 */
  __pyx_v_func_name_str = NULL;

  /* "asynctnt/iproto/requests/call.pyx":20
 * 
 *         func_name_str = NULL
 *         func_name_len = 0             # <<<<<<<<<<<<<<
 * 
 *         func_name_temp = encode_unicode_string(self.func_name, buffer._encoding)
 */
  __pyx_v_func_name_len = 0;

  /* "asynctnt/iproto/requests/call.pyx":22
 *         func_name_len = 0
 * 
 *         func_name_temp = encode_unicode_string(self.func_name, buffer._encoding)             # <<<<<<<<<<<<<<
 *         cpython.bytes.PyBytes_AsStringAndSize(func_name_temp,
 *                                               &func_name_str,
 */
  __pyx_t_1 = __pyx_v_self->func_name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_buffer->_encoding;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string(__pyx_t_1, ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(14, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_func_name_temp = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/requests/call.pyx":23
 * 
 *         func_name_temp = encode_unicode_string(self.func_name, buffer._encoding)
 *         cpython.bytes.PyBytes_AsStringAndSize(func_name_temp,             # <<<<<<<<<<<<<<
 *                                               &func_name_str,
 *                                               &func_name_len)
 */
  __pyx_t_4 = PyBytes_AsStringAndSize(__pyx_v_func_name_temp, (&__pyx_v_func_name_str), ((Py_ssize_t *)(&__pyx_v_func_name_len))); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(14, 23, __pyx_L1_error)

  /* "asynctnt/iproto/requests/call.pyx":26
 *                                               &func_name_str,
 *                                               &func_name_len)
 *         body_map_sz = 2             # <<<<<<<<<<<<<<
 *         # Size description:
 *         # mp_sizeof_map()
 */
  __pyx_v_body_map_sz = 2;

  /* "asynctnt/iproto/requests/call.pyx":35
 *                        + 1 \
 *                        + mp_sizeof_str(<uint32_t> func_name_len) \
 *                        + 1             # <<<<<<<<<<<<<<
 * 
 *         buffer.ensure_allocated(max_body_len)
 */
  __pyx_v_max_body_len = ((0x2 + mp_sizeof_str(((uint32_t)__pyx_v_func_name_len))) + 1);

  /* "asynctnt/iproto/requests/call.pyx":37
 *                        + 1
 * 
 *         buffer.ensure_allocated(max_body_len)             # <<<<<<<<<<<<<<
 * 
 *         p = begin = &buffer._buf[buffer._length]
 */
  __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_buffer, __pyx_v_max_body_len); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(14, 37, __pyx_L1_error)

  /* "asynctnt/iproto/requests/call.pyx":39
 *         buffer.ensure_allocated(max_body_len)
 * 
 *         p = begin = &buffer._buf[buffer._length]             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FUNCTION_NAME)
 */
  __pyx_t_5 = (&(__pyx_v_buffer->_buf[__pyx_v_buffer->_length]));
  __pyx_v_p = __pyx_t_5;
  __pyx_v_begin = __pyx_t_5;

  /* "asynctnt/iproto/requests/call.pyx":40
 * 
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_FUNCTION_NAME)
 *         p = mp_encode_str(p, func_name_str, <uint32_t> func_name_len)
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_body_map_sz);

  /* "asynctnt/iproto/requests/call.pyx":41
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FUNCTION_NAME)             # <<<<<<<<<<<<<<
 *         p = mp_encode_str(p, func_name_str, <uint32_t> func_name_len)
 * 
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FUNCTION_NAME);

  /* "asynctnt/iproto/requests/call.pyx":42
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FUNCTION_NAME)
 *         p = mp_encode_str(p, func_name_str, <uint32_t> func_name_len)             # <<<<<<<<<<<<<<
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 */
  __pyx_v_p = mp_encode_str(__pyx_v_p, __pyx_v_func_name_str, ((uint32_t)__pyx_v_func_name_len));

  /* "asynctnt/iproto/requests/call.pyx":44
 *         p = mp_encode_str(p, func_name_str, <uint32_t> func_name_len)
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)             # <<<<<<<<<<<<<<
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.args, None, False)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TUPLE);

  /* "asynctnt/iproto/requests/call.pyx":45
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 *         buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 *         p = encode_key_sequence(buffer, p, self.args, None, False)
 */
  __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/requests/call.pyx":46
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.args, None, False)             # <<<<<<<<<<<<<<
 */
  __pyx_t_3 = __pyx_v_self->args;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_8asynctnt_6iproto_8protocol_encode_key_sequence(__pyx_v_buffer, __pyx_v_p, __pyx_t_3, ((struct C_Metadata *)Py_None), 0); if (unlikely(__pyx_t_5 == ((char *)NULL))) __PYX_ERR(14, 46, __pyx_L1_error)
  __pyx_v_p = __pyx_t_5;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/requests/call.pyx":8
 * @cython.final
 * cdef class CallRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CallRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_func_name_temp);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11CallRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11CallRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11CallRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11CallRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11CallRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11CallRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.args, self.check_schema_change, self.func_name, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(13); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_v_self->args);
  __Pyx_GIVEREF(__pyx_v_self->args);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_self->args)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->func_name);
  __Pyx_GIVEREF(__pyx_v_self->func_name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_self->func_name)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 11, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 12, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.args, self.check_schema_change, self.func_name, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.args, self.check_schema_change, self.func_name, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.args is not None or self.func_name is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.args, self.check_schema_change, self.func_name, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.args is not None or self.func_name is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, None), state
 */
  /*else*/ {
    __pyx_t_11 = (__pyx_v_self->args != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->func_name != ((PyObject*)Py_None));
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.args is not None or self.func_name is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.args is not None or self.func_name is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_CallRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_47556637);
    __Pyx_GIVEREF(__pyx_int_47556637);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_47556637)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.args is not None or self.func_name is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, None), state
 *     else:
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CallRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_CallRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_47556637);
    __Pyx_GIVEREF(__pyx_int_47556637);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_47556637)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CallRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CallRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11CallRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11CallRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11CallRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11CallRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CallRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11CallRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11CallRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CallRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CallRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CallRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CallRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/eval.pyx":6
 * @cython.final
 * cdef class EvalRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_11EvalRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  char *__pyx_v_begin;
  char *__pyx_v_p;
  uint32_t __pyx_v_body_map_sz;
  uint32_t __pyx_v_max_body_len;
  PyObject *__pyx_v_expression_temp = 0;
  char *__pyx_v_expression_str;
  Py_ssize_t __pyx_v_expression_len;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  char *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_body", 1);

  /* "asynctnt/iproto/requests/eval.pyx":17
 *             ssize_t expression_len
 * 
 *         expression_str = NULL             # <<<<<<<<<<<<<<
 *         expression_len = 0
 * 
 */
  __pyx_v_expression_str = NULL;

  /* "asynctnt/iproto/requests/eval.pyx":18
 * 
 *         expression_str = NULL
 *         expression_len = 0             # <<<<<<<<<<<<<<
 * 
 *         expression_temp = encode_unicode_string(self.expression, buffer._encoding)
 */
  __pyx_v_expression_len = 0;

  /* "asynctnt/iproto/requests/eval.pyx":20
 *         expression_len = 0
 * 
 *         expression_temp = encode_unicode_string(self.expression, buffer._encoding)             # <<<<<<<<<<<<<<
 *         cpython.bytes.PyBytes_AsStringAndSize(expression_temp,
 *                                               &expression_str,
 */
  __pyx_t_1 = __pyx_v_self->expression;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_buffer->_encoding;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string(__pyx_t_1, ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(15, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_expression_temp = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/requests/eval.pyx":21
 * 
 *         expression_temp = encode_unicode_string(self.expression, buffer._encoding)
 *         cpython.bytes.PyBytes_AsStringAndSize(expression_temp,             # <<<<<<<<<<<<<<
 *                                               &expression_str,
 *                                               &expression_len)
 */
  __pyx_t_4 = PyBytes_AsStringAndSize(__pyx_v_expression_temp, (&__pyx_v_expression_str), ((Py_ssize_t *)(&__pyx_v_expression_len))); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(15, 21, __pyx_L1_error)

  /* "asynctnt/iproto/requests/eval.pyx":24
 *                                               &expression_str,
 *                                               &expression_len)
 *         body_map_sz = 2             # <<<<<<<<<<<<<<
 *         # Size description:
 *         # mp_sizeof_map()
 */
  __pyx_v_body_map_sz = 2;

  /* "asynctnt/iproto/requests/eval.pyx":33
 *                        + 1 \
 *                        + mp_sizeof_str(<uint32_t> expression_len) \
 *                        + 1             # <<<<<<<<<<<<<<
 * 
 *         buffer.ensure_allocated(max_body_len)
 */
  __pyx_v_max_body_len = ((0x2 + mp_sizeof_str(((uint32_t)__pyx_v_expression_len))) + 1);

  /* "asynctnt/iproto/requests/eval.pyx":35
 *                        + 1
 * 
 *         buffer.ensure_allocated(max_body_len)             # <<<<<<<<<<<<<<
 * 
 *         p = begin = &buffer._buf[buffer._length]
 */
  __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_buffer, __pyx_v_max_body_len); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(15, 35, __pyx_L1_error)

  /* "asynctnt/iproto/requests/eval.pyx":37
 *         buffer.ensure_allocated(max_body_len)
 * 
 *         p = begin = &buffer._buf[buffer._length]             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_EXPR)
 */
  __pyx_t_5 = (&(__pyx_v_buffer->_buf[__pyx_v_buffer->_length]));
  __pyx_v_p = __pyx_t_5;
  __pyx_v_begin = __pyx_t_5;

  /* "asynctnt/iproto/requests/eval.pyx":38
 * 
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_EXPR)
 *         p = mp_encode_str(p, expression_str, <uint32_t> expression_len)
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_body_map_sz);

  /* "asynctnt/iproto/requests/eval.pyx":39
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_EXPR)             # <<<<<<<<<<<<<<
 *         p = mp_encode_str(p, expression_str, <uint32_t> expression_len)
 * 
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_EXPR);

  /* "asynctnt/iproto/requests/eval.pyx":40
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_EXPR)
 *         p = mp_encode_str(p, expression_str, <uint32_t> expression_len)             # <<<<<<<<<<<<<<
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 */
  __pyx_v_p = mp_encode_str(__pyx_v_p, __pyx_v_expression_str, ((uint32_t)__pyx_v_expression_len));

  /* "asynctnt/iproto/requests/eval.pyx":42
 *         p = mp_encode_str(p, expression_str, <uint32_t> expression_len)
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)             # <<<<<<<<<<<<<<
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.args, None, False)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TUPLE);

  /* "asynctnt/iproto/requests/eval.pyx":43
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 *         buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 *         p = encode_key_sequence(buffer, p, self.args, None, False)
 */
  __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/requests/eval.pyx":44
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.args, None, False)             # <<<<<<<<<<<<<<
 */
  __pyx_t_3 = __pyx_v_self->args;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = __pyx_f_8asynctnt_6iproto_8protocol_encode_key_sequence(__pyx_v_buffer, __pyx_v_p, __pyx_t_3, ((struct C_Metadata *)Py_None), 0); if (unlikely(__pyx_t_5 == ((char *)NULL))) __PYX_ERR(15, 44, __pyx_L1_error)
  __pyx_v_p = __pyx_t_5;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/requests/eval.pyx":6
 * @cython.final
 * cdef class EvalRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.EvalRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_expression_temp);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11EvalRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11EvalRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11EvalRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11EvalRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11EvalRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11EvalRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.args, self.check_schema_change, self.expression, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(13); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_v_self->args);
  __Pyx_GIVEREF(__pyx_v_self->args);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_self->args)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->expression);
  __Pyx_GIVEREF(__pyx_v_self->expression);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_self->expression)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 11, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 12, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.args, self.check_schema_change, self.expression, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.args, self.check_schema_change, self.expression, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.args is not None or self.expression is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.args, self.check_schema_change, self.expression, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.args is not None or self.expression is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, None), state
 */
  /*else*/ {
    __pyx_t_11 = (__pyx_v_self->args != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->expression != ((PyObject*)Py_None));
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.args is not None or self.expression is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.args is not None or self.expression is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_EvalRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184220263);
    __Pyx_GIVEREF(__pyx_int_184220263);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_184220263)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.args is not None or self.expression is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, None), state
 *     else:
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_EvalRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_EvalRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_184220263);
    __Pyx_GIVEREF(__pyx_int_184220263);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_184220263)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.EvalRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EvalRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11EvalRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11EvalRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11EvalRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11EvalRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.EvalRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11EvalRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11EvalRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_EvalRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_EvalRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EvalRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.EvalRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/select.pyx":6
 * @cython.final
 * cdef class SelectRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_13SelectRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  char *__pyx_v_begin;
  char *__pyx_v_p;
  uint32_t __pyx_v_body_map_sz;
  uint32_t __pyx_v_max_body_len;
  uint32_t __pyx_v_space_id;
  uint32_t __pyx_v_index_id;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  char *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_body", 1);

  /* "asynctnt/iproto/requests/select.pyx":14
 *             uint32_t space_id, index_id
 * 
 *         space_id = self.space.sid             # <<<<<<<<<<<<<<
 *         index_id = self.index.iid
 * 
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.space->sid;
  __pyx_v_space_id = __pyx_t_1;

  /* "asynctnt/iproto/requests/select.pyx":15
 * 
 *         space_id = self.space.sid
 *         index_id = self.index.iid             # <<<<<<<<<<<<<<
 * 
 *         body_map_sz = 3 \
 */
  __pyx_t_1 = __pyx_v_self->index->iid;
  __pyx_v_index_id = __pyx_t_1;

  /* "asynctnt/iproto/requests/select.pyx":20
 *                       + <uint32_t> (index_id > 0) \
 *                       + <uint32_t> (self.offset > 0) \
 *                       + <uint32_t> (self.iterator > 0)             # <<<<<<<<<<<<<<
 *         # Size description:
 *         # mp_sizeof_map(body_map_sz)
 */
  __pyx_v_body_map_sz = (((3 + ((uint32_t)(__pyx_v_index_id > 0))) + ((uint32_t)(__pyx_v_self->offset > 0))) + ((uint32_t)(__pyx_v_self->iterator > 0)));

  /* "asynctnt/iproto/requests/select.pyx":31
 *                        + 9 \
 *                        + 1 \
 *                        + 9             # <<<<<<<<<<<<<<
 * 
 *         if index_id > 0:
 */
  __pyx_v_max_body_len = 0x15;

  /* "asynctnt/iproto/requests/select.pyx":33
 *                        + 9
 * 
 *         if index_id > 0:             # <<<<<<<<<<<<<<
 *             # mp_sizeof_uint(TP_INDEX) + mp_sizeof_uint(index_id)
 *             max_body_len += 1 + 9
 */
  __pyx_t_2 = (__pyx_v_index_id > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/requests/select.pyx":35
 *         if index_id > 0:
 *             # mp_sizeof_uint(TP_INDEX) + mp_sizeof_uint(index_id)
 *             max_body_len += 1 + 9             # <<<<<<<<<<<<<<
 *         if self.offset > 0:
 *             # mp_sizeof_uint(TP_OFFSET) + mp_sizeof_uint(offset)
 */
    __pyx_v_max_body_len = (__pyx_v_max_body_len + 0xa);

    /* "asynctnt/iproto/requests/select.pyx":33
 *                        + 9
 * 
 *         if index_id > 0:             # <<<<<<<<<<<<<<
 *             # mp_sizeof_uint(TP_INDEX) + mp_sizeof_uint(index_id)
 *             max_body_len += 1 + 9
 */
  }

  /* "asynctnt/iproto/requests/select.pyx":36
 *             # mp_sizeof_uint(TP_INDEX) + mp_sizeof_uint(index_id)
 *             max_body_len += 1 + 9
 *         if self.offset > 0:             # <<<<<<<<<<<<<<
 *             # mp_sizeof_uint(TP_OFFSET) + mp_sizeof_uint(offset)
 *             max_body_len += 1 + 9
 */
  __pyx_t_2 = (__pyx_v_self->offset > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/requests/select.pyx":38
 *         if self.offset > 0:
 *             # mp_sizeof_uint(TP_OFFSET) + mp_sizeof_uint(offset)
 *             max_body_len += 1 + 9             # <<<<<<<<<<<<<<
 *         if self.iterator > 0:
 *             # mp_sizeof_uint(TP_ITERATOR) + mp_sizeof_uint(iterator)
 */
    __pyx_v_max_body_len = (__pyx_v_max_body_len + 0xa);

    /* "asynctnt/iproto/requests/select.pyx":36
 *             # mp_sizeof_uint(TP_INDEX) + mp_sizeof_uint(index_id)
 *             max_body_len += 1 + 9
 *         if self.offset > 0:             # <<<<<<<<<<<<<<
 *             # mp_sizeof_uint(TP_OFFSET) + mp_sizeof_uint(offset)
 *             max_body_len += 1 + 9
 */
  }

  /* "asynctnt/iproto/requests/select.pyx":39
 *             # mp_sizeof_uint(TP_OFFSET) + mp_sizeof_uint(offset)
 *             max_body_len += 1 + 9
 *         if self.iterator > 0:             # <<<<<<<<<<<<<<
 *             # mp_sizeof_uint(TP_ITERATOR) + mp_sizeof_uint(iterator)
 *             max_body_len += 1 + 1
 */
  __pyx_t_2 = (__pyx_v_self->iterator > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/requests/select.pyx":41
 *         if self.iterator > 0:
 *             # mp_sizeof_uint(TP_ITERATOR) + mp_sizeof_uint(iterator)
 *             max_body_len += 1 + 1             # <<<<<<<<<<<<<<
 * 
 *         max_body_len += 1  # mp_sizeof_uint(TP_KEY);
 */
    __pyx_v_max_body_len = (__pyx_v_max_body_len + 0x2);

    /* "asynctnt/iproto/requests/select.pyx":39
 *             # mp_sizeof_uint(TP_OFFSET) + mp_sizeof_uint(offset)
 *             max_body_len += 1 + 9
 *         if self.iterator > 0:             # <<<<<<<<<<<<<<
 *             # mp_sizeof_uint(TP_ITERATOR) + mp_sizeof_uint(iterator)
 *             max_body_len += 1 + 1
 */
  }

  /* "asynctnt/iproto/requests/select.pyx":43
 *             max_body_len += 1 + 1
 * 
 *         max_body_len += 1  # mp_sizeof_uint(TP_KEY);             # <<<<<<<<<<<<<<
 * 
 *         buffer.ensure_allocated(max_body_len)
 */
  __pyx_v_max_body_len = (__pyx_v_max_body_len + 1);

  /* "asynctnt/iproto/requests/select.pyx":45
 *         max_body_len += 1  # mp_sizeof_uint(TP_KEY);
 * 
 *         buffer.ensure_allocated(max_body_len)             # <<<<<<<<<<<<<<
 * 
 *         p = begin = &buffer._buf[buffer._length]
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_buffer, __pyx_v_max_body_len); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(16, 45, __pyx_L1_error)

  /* "asynctnt/iproto/requests/select.pyx":47
 *         buffer.ensure_allocated(max_body_len)
 * 
 *         p = begin = &buffer._buf[buffer._length]             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 */
  __pyx_t_3 = (&(__pyx_v_buffer->_buf[__pyx_v_buffer->_length]));
  __pyx_v_p = __pyx_t_3;
  __pyx_v_begin = __pyx_t_3;

  /* "asynctnt/iproto/requests/select.pyx":48
 * 
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 *         p = mp_encode_uint(p, space_id)
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_body_map_sz);

  /* "asynctnt/iproto/requests/select.pyx":49
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, space_id)
 *         p = mp_encode_uint(p, tarantool.IPROTO_LIMIT)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SPACE_ID);

  /* "asynctnt/iproto/requests/select.pyx":50
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 *         p = mp_encode_uint(p, space_id)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_LIMIT)
 *         p = mp_encode_uint(p, self.limit)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_space_id);

  /* "asynctnt/iproto/requests/select.pyx":51
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 *         p = mp_encode_uint(p, space_id)
 *         p = mp_encode_uint(p, tarantool.IPROTO_LIMIT)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, self.limit)
 * 
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_LIMIT);

  /* "asynctnt/iproto/requests/select.pyx":52
 *         p = mp_encode_uint(p, space_id)
 *         p = mp_encode_uint(p, tarantool.IPROTO_LIMIT)
 *         p = mp_encode_uint(p, self.limit)             # <<<<<<<<<<<<<<
 * 
 *         if index_id > 0:
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_self->limit);

  /* "asynctnt/iproto/requests/select.pyx":54
 *         p = mp_encode_uint(p, self.limit)
 * 
 *         if index_id > 0:             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *             p = mp_encode_uint(p, index_id)
 */
  __pyx_t_2 = (__pyx_v_index_id > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/requests/select.pyx":55
 * 
 *         if index_id > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, index_id)
 *         if self.offset > 0:
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_INDEX_ID);

    /* "asynctnt/iproto/requests/select.pyx":56
 *         if index_id > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *             p = mp_encode_uint(p, index_id)             # <<<<<<<<<<<<<<
 *         if self.offset > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_OFFSET)
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_index_id);

    /* "asynctnt/iproto/requests/select.pyx":54
 *         p = mp_encode_uint(p, self.limit)
 * 
 *         if index_id > 0:             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *             p = mp_encode_uint(p, index_id)
 */
  }

  /* "asynctnt/iproto/requests/select.pyx":57
 *             p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *             p = mp_encode_uint(p, index_id)
 *         if self.offset > 0:             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_OFFSET)
 *             p = mp_encode_uint(p, self.offset)
 */
  __pyx_t_2 = (__pyx_v_self->offset > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/requests/select.pyx":58
 *             p = mp_encode_uint(p, index_id)
 *         if self.offset > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_OFFSET)             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, self.offset)
 *         if self.iterator > 0:
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OFFSET);

    /* "asynctnt/iproto/requests/select.pyx":59
 *         if self.offset > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_OFFSET)
 *             p = mp_encode_uint(p, self.offset)             # <<<<<<<<<<<<<<
 *         if self.iterator > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_ITERATOR)
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_self->offset);

    /* "asynctnt/iproto/requests/select.pyx":57
 *             p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *             p = mp_encode_uint(p, index_id)
 *         if self.offset > 0:             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_OFFSET)
 *             p = mp_encode_uint(p, self.offset)
 */
  }

  /* "asynctnt/iproto/requests/select.pyx":60
 *             p = mp_encode_uint(p, tarantool.IPROTO_OFFSET)
 *             p = mp_encode_uint(p, self.offset)
 *         if self.iterator > 0:             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_ITERATOR)
 *             p = mp_encode_uint(p, self.iterator)
 */
  __pyx_t_2 = (__pyx_v_self->iterator > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/requests/select.pyx":61
 *             p = mp_encode_uint(p, self.offset)
 *         if self.iterator > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_ITERATOR)             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, self.iterator)
 * 
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_ITERATOR);

    /* "asynctnt/iproto/requests/select.pyx":62
 *         if self.iterator > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_ITERATOR)
 *             p = mp_encode_uint(p, self.iterator)             # <<<<<<<<<<<<<<
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_KEY)
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_self->iterator);

    /* "asynctnt/iproto/requests/select.pyx":60
 *             p = mp_encode_uint(p, tarantool.IPROTO_OFFSET)
 *             p = mp_encode_uint(p, self.offset)
 *         if self.iterator > 0:             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_ITERATOR)
 *             p = mp_encode_uint(p, self.iterator)
 */
  }

  /* "asynctnt/iproto/requests/select.pyx":64
 *             p = mp_encode_uint(p, self.iterator)
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_KEY)             # <<<<<<<<<<<<<<
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.key, self.index.metadata, False)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_KEY);

  /* "asynctnt/iproto/requests/select.pyx":65
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_KEY)
 *         buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 *         p = encode_key_sequence(buffer, p, self.key, self.index.metadata, False)
 */
  __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/requests/select.pyx":66
 *         p = mp_encode_uint(p, tarantool.IPROTO_KEY)
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.key, self.index.metadata, False)             # <<<<<<<<<<<<<<
 */
  __pyx_t_4 = __pyx_v_self->key;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = ((PyObject *)__pyx_v_self->index->metadata);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_encode_key_sequence(__pyx_v_buffer, __pyx_v_p, __pyx_t_4, ((struct C_Metadata *)__pyx_t_5), 0); if (unlikely(__pyx_t_3 == ((char *)NULL))) __PYX_ERR(16, 66, __pyx_L1_error)
  __pyx_v_p = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "asynctnt/iproto/requests/select.pyx":6
 * @cython.final
 * cdef class SelectRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SelectRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13SelectRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13SelectRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13SelectRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13SelectRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13SelectRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13SelectRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.index, self.iterator, self.key, self.limit, self.offset, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->iterator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->limit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->offset); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = PyTuple_New(16); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->index);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->index);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, ((PyObject *)__pyx_v_self->index))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->key);
  __Pyx_GIVEREF(__pyx_v_self->key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 3, __pyx_v_self->key)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 4, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 5, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 6, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 7, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 8, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 9, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 10, __pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 11, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 12, __pyx_t_10)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_11);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 13, __pyx_t_11)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 14, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 15, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_12);
  __pyx_t_12 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.index, self.iterator, self.key, self.limit, self.offset, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_12 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_v__dict = __pyx_t_12;
  __pyx_t_12 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.index, self.iterator, self.key, self.limit, self.offset, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_13 = (__pyx_v__dict != Py_None);
  if (__pyx_t_13) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_12 = PyTuple_New(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_11 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_11));
    __pyx_t_11 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.index, self.iterator, self.key, self.limit, self.offset, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, None), state
 */
  /*else*/ {
    __pyx_t_14 = (((PyObject *)__pyx_v_self->index) != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->key != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_14 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_13 = __pyx_t_14;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_13;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.index is not None or self.key is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_pyx_unpickle_SelectRequest); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_149085227);
    __Pyx_GIVEREF(__pyx_int_149085227);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_int_149085227)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_11);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_11)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_12);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_12)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, None), state
 *     else:
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SelectRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_pyx_unpickle_SelectRequest); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_149085227);
    __Pyx_GIVEREF(__pyx_int_149085227);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_int_149085227)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_12);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_12)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_10 = 0;
    __pyx_t_12 = 0;
    __pyx_r = __pyx_t_11;
    __pyx_t_11 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SelectRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SelectRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13SelectRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13SelectRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13SelectRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13SelectRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SelectRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13SelectRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13SelectRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SelectRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_SelectRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SelectRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.SelectRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/insert.pyx":7
 * cdef class InsertRequest(BaseRequest):
 * 
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_13InsertRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  char *__pyx_v_begin;
  char *__pyx_v_p;
  uint32_t __pyx_v_body_map_sz;
  uint32_t __pyx_v_max_body_len;
  uint32_t __pyx_v_space_id;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  char *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_body", 1);

  /* "asynctnt/iproto/requests/insert.pyx":15
 *             uint32_t space_id
 * 
 *         space_id = self.space.sid             # <<<<<<<<<<<<<<
 * 
 *         body_map_sz = 2
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.space->sid;
  __pyx_v_space_id = __pyx_t_1;

  /* "asynctnt/iproto/requests/insert.pyx":17
 *         space_id = self.space.sid
 * 
 *         body_map_sz = 2             # <<<<<<<<<<<<<<
 *         # Size description:
 *         # mp_sizeof_map(body_map_sz)
 */
  __pyx_v_body_map_sz = 2;

  /* "asynctnt/iproto/requests/insert.pyx":26
 *                        + 1 \
 *                        + 9 \
 *                        + 1             # <<<<<<<<<<<<<<
 * 
 *         buffer.ensure_allocated(max_body_len)
 */
  __pyx_v_max_body_len = 0xc;

  /* "asynctnt/iproto/requests/insert.pyx":28
 *                        + 1
 * 
 *         buffer.ensure_allocated(max_body_len)             # <<<<<<<<<<<<<<
 * 
 *         p = begin = &buffer._buf[buffer._length]
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_buffer, __pyx_v_max_body_len); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(17, 28, __pyx_L1_error)

  /* "asynctnt/iproto/requests/insert.pyx":30
 *         buffer.ensure_allocated(max_body_len)
 * 
 *         p = begin = &buffer._buf[buffer._length]             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 */
  __pyx_t_2 = (&(__pyx_v_buffer->_buf[__pyx_v_buffer->_length]));
  __pyx_v_p = __pyx_t_2;
  __pyx_v_begin = __pyx_t_2;

  /* "asynctnt/iproto/requests/insert.pyx":31
 * 
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 *         p = mp_encode_uint(p, space_id)
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_body_map_sz);

  /* "asynctnt/iproto/requests/insert.pyx":32
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, space_id)
 * 
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SPACE_ID);

  /* "asynctnt/iproto/requests/insert.pyx":33
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 *         p = mp_encode_uint(p, space_id)             # <<<<<<<<<<<<<<
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_space_id);

  /* "asynctnt/iproto/requests/insert.pyx":35
 *         p = mp_encode_uint(p, space_id)
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)             # <<<<<<<<<<<<<<
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.t, self.space.metadata, True)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TUPLE);

  /* "asynctnt/iproto/requests/insert.pyx":36
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 *         buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 *         p = encode_key_sequence(buffer, p, self.t, self.space.metadata, True)
 */
  __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/requests/insert.pyx":37
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.t, self.space.metadata, True)             # <<<<<<<<<<<<<<
 */
  __pyx_t_3 = __pyx_v_self->t;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = ((PyObject *)__pyx_v_self->__pyx_base.space->metadata);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_encode_key_sequence(__pyx_v_buffer, __pyx_v_p, __pyx_t_3, ((struct C_Metadata *)__pyx_t_4), 1); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(17, 37, __pyx_L1_error)
  __pyx_v_p = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "asynctnt/iproto/requests/insert.pyx":7
 * cdef class InsertRequest(BaseRequest):
 * 
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.InsertRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13InsertRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13InsertRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13InsertRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13InsertRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13InsertRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13InsertRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.t, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(12); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->t);
  __Pyx_GIVEREF(__pyx_v_self->t);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_v_self->t)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 11, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.t, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.t, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.space is not None or self.t is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.t, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.space is not None or self.t is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, None), state
 */
  /*else*/ {
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->t != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.t is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.space is not None or self.t is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_InsertRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_194105842);
    __Pyx_GIVEREF(__pyx_int_194105842);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_194105842)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.t is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, None), state
 *     else:
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InsertRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_InsertRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_194105842);
    __Pyx_GIVEREF(__pyx_int_194105842);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_194105842)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.InsertRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InsertRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13InsertRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13InsertRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13InsertRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13InsertRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.InsertRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13InsertRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13InsertRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_InsertRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_InsertRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InsertRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.InsertRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/delete.pyx":7
 * cdef class DeleteRequest(BaseRequest):
 * 
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *p
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_13DeleteRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  char *__pyx_v_p;
  char *__pyx_v_begin;
  uint32_t __pyx_v_body_map_sz;
  uint32_t __pyx_v_max_body_len;
  uint32_t __pyx_v_space_id;
  uint32_t __pyx_v_index_id;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  char *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_body", 1);

  /* "asynctnt/iproto/requests/delete.pyx":15
 *             uint32_t space_id, index_id
 * 
 *         space_id = self.space.sid             # <<<<<<<<<<<<<<
 *         index_id = self.index.iid
 * 
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.space->sid;
  __pyx_v_space_id = __pyx_t_1;

  /* "asynctnt/iproto/requests/delete.pyx":16
 * 
 *         space_id = self.space.sid
 *         index_id = self.index.iid             # <<<<<<<<<<<<<<
 * 
 *         body_map_sz = 2 \
 */
  __pyx_t_1 = __pyx_v_self->index->iid;
  __pyx_v_index_id = __pyx_t_1;

  /* "asynctnt/iproto/requests/delete.pyx":19
 * 
 *         body_map_sz = 2 \
 *                       + <uint32_t> (index_id > 0)             # <<<<<<<<<<<<<<
 *         # Size description:
 *         # mp_sizeof_map(body_map_sz)
 */
  __pyx_v_body_map_sz = (2 + ((uint32_t)(__pyx_v_index_id > 0)));

  /* "asynctnt/iproto/requests/delete.pyx":26
 *         max_body_len = 1 \
 *                        + 1 \
 *                        + 9             # <<<<<<<<<<<<<<
 * 
 *         if index_id > 0:
 */
  __pyx_v_max_body_len = 0xb;

  /* "asynctnt/iproto/requests/delete.pyx":28
 *                        + 9
 * 
 *         if index_id > 0:             # <<<<<<<<<<<<<<
 *             # mp_sizeof_uint(TP_INDEX) + mp_sizeof_uint(index)
 *             max_body_len += 1 + 9
 */
  __pyx_t_2 = (__pyx_v_index_id > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/requests/delete.pyx":30
 *         if index_id > 0:
 *             # mp_sizeof_uint(TP_INDEX) + mp_sizeof_uint(index)
 *             max_body_len += 1 + 9             # <<<<<<<<<<<<<<
 * 
 *         max_body_len += 1  # mp_sizeof_uint(TP_KEY);
 */
    __pyx_v_max_body_len = (__pyx_v_max_body_len + 0xa);

    /* "asynctnt/iproto/requests/delete.pyx":28
 *                        + 9
 * 
 *         if index_id > 0:             # <<<<<<<<<<<<<<
 *             # mp_sizeof_uint(TP_INDEX) + mp_sizeof_uint(index)
 *             max_body_len += 1 + 9
 */
  }

  /* "asynctnt/iproto/requests/delete.pyx":32
 *             max_body_len += 1 + 9
 * 
 *         max_body_len += 1  # mp_sizeof_uint(TP_KEY);             # <<<<<<<<<<<<<<
 * 
 *         buffer.ensure_allocated(max_body_len)
 */
  __pyx_v_max_body_len = (__pyx_v_max_body_len + 1);

  /* "asynctnt/iproto/requests/delete.pyx":34
 *         max_body_len += 1  # mp_sizeof_uint(TP_KEY);
 * 
 *         buffer.ensure_allocated(max_body_len)             # <<<<<<<<<<<<<<
 * 
 *         p = begin = &buffer._buf[buffer._length]
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_buffer, __pyx_v_max_body_len); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(18, 34, __pyx_L1_error)

  /* "asynctnt/iproto/requests/delete.pyx":36
 *         buffer.ensure_allocated(max_body_len)
 * 
 *         p = begin = &buffer._buf[buffer._length]             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 */
  __pyx_t_3 = (&(__pyx_v_buffer->_buf[__pyx_v_buffer->_length]));
  __pyx_v_p = __pyx_t_3;
  __pyx_v_begin = __pyx_t_3;

  /* "asynctnt/iproto/requests/delete.pyx":37
 * 
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 *         p = mp_encode_uint(p, space_id)
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_body_map_sz);

  /* "asynctnt/iproto/requests/delete.pyx":38
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, space_id)
 * 
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SPACE_ID);

  /* "asynctnt/iproto/requests/delete.pyx":39
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 *         p = mp_encode_uint(p, space_id)             # <<<<<<<<<<<<<<
 * 
 *         if index_id > 0:
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_space_id);

  /* "asynctnt/iproto/requests/delete.pyx":41
 *         p = mp_encode_uint(p, space_id)
 * 
 *         if index_id > 0:             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *             p = mp_encode_uint(p, index_id)
 */
  __pyx_t_2 = (__pyx_v_index_id > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/requests/delete.pyx":42
 * 
 *         if index_id > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, index_id)
 * 
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_INDEX_ID);

    /* "asynctnt/iproto/requests/delete.pyx":43
 *         if index_id > 0:
 *             p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *             p = mp_encode_uint(p, index_id)             # <<<<<<<<<<<<<<
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_KEY)
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_index_id);

    /* "asynctnt/iproto/requests/delete.pyx":41
 *         p = mp_encode_uint(p, space_id)
 * 
 *         if index_id > 0:             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *             p = mp_encode_uint(p, index_id)
 */
  }

  /* "asynctnt/iproto/requests/delete.pyx":45
 *             p = mp_encode_uint(p, index_id)
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_KEY)             # <<<<<<<<<<<<<<
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.key, self.index.metadata, False)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_KEY);

  /* "asynctnt/iproto/requests/delete.pyx":46
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_KEY)
 *         buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 *         p = encode_key_sequence(buffer, p, self.key, self.index.metadata, False)
 */
  __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/requests/delete.pyx":47
 *         p = mp_encode_uint(p, tarantool.IPROTO_KEY)
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.key, self.index.metadata, False)             # <<<<<<<<<<<<<<
 */
  __pyx_t_4 = __pyx_v_self->key;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = ((PyObject *)__pyx_v_self->index->metadata);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_encode_key_sequence(__pyx_v_buffer, __pyx_v_p, __pyx_t_4, ((struct C_Metadata *)__pyx_t_5), 0); if (unlikely(__pyx_t_3 == ((char *)NULL))) __PYX_ERR(18, 47, __pyx_L1_error)
  __pyx_v_p = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "asynctnt/iproto/requests/delete.pyx":7
 * cdef class DeleteRequest(BaseRequest):
 * 
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *p
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.DeleteRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13DeleteRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13DeleteRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13DeleteRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13DeleteRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13DeleteRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13DeleteRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.index, self.key, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(13); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->index);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->index);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_self->index))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->key);
  __Pyx_GIVEREF(__pyx_v_self->key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_self->key)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 11, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 12, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.index, self.key, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.index, self.key, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.index, self.key, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, None), state
 */
  /*else*/ {
    __pyx_t_11 = (((PyObject *)__pyx_v_self->index) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->key != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.index is not None or self.key is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_DeleteRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_255344631);
    __Pyx_GIVEREF(__pyx_int_255344631);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_255344631)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, None), state
 *     else:
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DeleteRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_DeleteRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_255344631);
    __Pyx_GIVEREF(__pyx_int_255344631);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_255344631)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.DeleteRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DeleteRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13DeleteRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13DeleteRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13DeleteRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13DeleteRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.DeleteRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13DeleteRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13DeleteRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DeleteRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_DeleteRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DeleteRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.DeleteRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/update.pyx":6
 * @cython.final
 * cdef class UpdateRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         return encode_request_update(buffer, self.space, self.index,
 *                                      self.key, self.operations, <bint> False)
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_13UpdateRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_body", 1);

  /* "asynctnt/iproto/requests/update.pyx":7
 * cdef class UpdateRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         return encode_request_update(buffer, self.space, self.index,             # <<<<<<<<<<<<<<
 *                                      self.key, self.operations, <bint> False)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_v_self->index);
  __Pyx_INCREF(__pyx_t_2);

  /* "asynctnt/iproto/requests/update.pyx":8
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         return encode_request_update(buffer, self.space, self.index,
 *                                      self.key, self.operations, <bint> False)             # <<<<<<<<<<<<<<
 * 
 * cdef char *encode_update_ops(WriteBuffer buffer,
 */
  __pyx_t_3 = __pyx_v_self->key;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = __pyx_v_self->operations;
  __Pyx_INCREF(__pyx_t_4);

  /* "asynctnt/iproto/requests/update.pyx":7
 * cdef class UpdateRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         return encode_request_update(buffer, self.space, self.index,             # <<<<<<<<<<<<<<
 *                                      self.key, self.operations, <bint> False)
 * 
 */
  __pyx_t_5 = __pyx_f_8asynctnt_6iproto_8protocol_encode_request_update(__pyx_v_buffer, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_t_2), __pyx_t_3, ((PyObject*)__pyx_t_4), 0); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(4, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "asynctnt/iproto/requests/update.pyx":6
 * @cython.final
 * cdef class UpdateRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         return encode_request_update(buffer, self.space, self.index,
 *                                      self.key, self.operations, <bint> False)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.UpdateRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13UpdateRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13UpdateRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13UpdateRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13UpdateRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13UpdateRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13UpdateRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.index, self.key, self.op, self.operations, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(14); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->index);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->index);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_self->index))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->key);
  __Pyx_GIVEREF(__pyx_v_self->key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_self->key)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->operations);
  __Pyx_GIVEREF(__pyx_v_self->operations);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_v_self->operations)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 11, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 12, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 13, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.index, self.key, self.op, self.operations, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.index, self.key, self.op, self.operations, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.operations is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.index, self.key, self.op, self.operations, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.operations is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, None), state
 */
  /*else*/ {
    __pyx_t_11 = (((PyObject *)__pyx_v_self->index) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->key != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->operations != ((PyObject*)Py_None));
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.operations is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.index is not None or self.key is not None or self.operations is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_UpdateRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_135652490);
    __Pyx_GIVEREF(__pyx_int_135652490);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_135652490)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.index is not None or self.key is not None or self.operations is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, None), state
 *     else:
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UpdateRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_UpdateRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_135652490);
    __Pyx_GIVEREF(__pyx_int_135652490);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_135652490)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.UpdateRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UpdateRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13UpdateRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13UpdateRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13UpdateRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13UpdateRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.UpdateRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13UpdateRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13UpdateRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UpdateRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_UpdateRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UpdateRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.UpdateRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/update.pyx":10
 *                                      self.key, self.operations, <bint> False)
 * 
 * cdef char *encode_update_ops(WriteBuffer buffer,             # <<<<<<<<<<<<<<
 *                              char *p, list operations,
 *                              SchemaSpace space) except NULL:
 */

static char *__pyx_f_8asynctnt_6iproto_8protocol_encode_update_ops(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer, char *__pyx_v_p, PyObject *__pyx_v_operations, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_space) {
  char *__pyx_v_begin;
  uint32_t __pyx_v_ops_len;
  uint32_t __pyx_v_op_len;
  PyObject *__pyx_v_str_temp = 0;
  char *__pyx_v_op_str_c;
  Py_ssize_t __pyx_v_op_str_len;
  char __pyx_v_op;
  uint32_t __pyx_v_extra_length;
  int __pyx_v_field_encode_as_str;
  uint64_t __pyx_v_field_no;
  char *__pyx_v_field_str_c;
  Py_ssize_t __pyx_v_field_str_len;
  PyObject *__pyx_v_field_no_obj = 0;
  uint32_t __pyx_v_splice_position;
  uint32_t __pyx_v_splice_offset;
  PyObject *__pyx_v_operation = NULL;
  PyObject *__pyx_v_op_type_str = NULL;
  PyObject *__pyx_v_op_argument = NULL;
  PyObject *__pyx_v_splice_position_obj = NULL;
  PyObject *__pyx_v_splice_offset_obj = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  char *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  uint32_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_update_ops", 1);

  /* "asynctnt/iproto/requests/update.pyx":34
 *         uint32_t splice_position, splice_offset
 * 
 *     field_encode_as_str = 0             # <<<<<<<<<<<<<<
 *     field_str_c = NULL
 * 
 */
  __pyx_v_field_encode_as_str = 0;

  /* "asynctnt/iproto/requests/update.pyx":35
 * 
 *     field_encode_as_str = 0
 *     field_str_c = NULL             # <<<<<<<<<<<<<<
 * 
 *     begin = NULL
 */
  __pyx_v_field_str_c = NULL;

  /* "asynctnt/iproto/requests/update.pyx":37
 *     field_str_c = NULL
 * 
 *     begin = NULL             # <<<<<<<<<<<<<<
 * 
 *     if operations is not None:
 */
  __pyx_v_begin = NULL;

  /* "asynctnt/iproto/requests/update.pyx":39
 *     begin = NULL
 * 
 *     if operations is not None:             # <<<<<<<<<<<<<<
 *         ops_len = <uint32_t> cpython.list.PyList_GET_SIZE(operations)
 *     else:
 */
  __pyx_t_1 = (__pyx_v_operations != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "asynctnt/iproto/requests/update.pyx":40
 * 
 *     if operations is not None:
 *         ops_len = <uint32_t> cpython.list.PyList_GET_SIZE(operations)             # <<<<<<<<<<<<<<
 *     else:
 *         ops_len = 0
 */
    __pyx_v_ops_len = ((uint32_t)PyList_GET_SIZE(__pyx_v_operations));

    /* "asynctnt/iproto/requests/update.pyx":39
 *     begin = NULL
 * 
 *     if operations is not None:             # <<<<<<<<<<<<<<
 *         ops_len = <uint32_t> cpython.list.PyList_GET_SIZE(operations)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/requests/update.pyx":42
 *         ops_len = <uint32_t> cpython.list.PyList_GET_SIZE(operations)
 *     else:
 *         ops_len = 0             # <<<<<<<<<<<<<<
 * 
 *     p = buffer.mp_encode_array(p, ops_len)
 */
  /*else*/ {
    __pyx_v_ops_len = 0;
  }
  __pyx_L3:;

  /* "asynctnt/iproto/requests/update.pyx":44
 *         ops_len = 0
 * 
 *     p = buffer.mp_encode_array(p, ops_len)             # <<<<<<<<<<<<<<
 *     if ops_len == 0:
 *         return p
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_array(__pyx_v_buffer, __pyx_v_p, __pyx_v_ops_len); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(4, 44, __pyx_L1_error)
  __pyx_v_p = __pyx_t_2;

  /* "asynctnt/iproto/requests/update.pyx":45
 * 
 *     p = buffer.mp_encode_array(p, ops_len)
 *     if ops_len == 0:             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  __pyx_t_1 = (__pyx_v_ops_len == 0);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/requests/update.pyx":46
 *     p = buffer.mp_encode_array(p, ops_len)
 *     if ops_len == 0:
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     for operation in operations:
 */
    __pyx_r = __pyx_v_p;
    goto __pyx_L0;

    /* "asynctnt/iproto/requests/update.pyx":45
 * 
 *     p = buffer.mp_encode_array(p, ops_len)
 *     if ops_len == 0:             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
  }

  /* "asynctnt/iproto/requests/update.pyx":48
 *         return p
 * 
 *     for operation in operations:             # <<<<<<<<<<<<<<
 *         if isinstance(operation, tuple):
 *             op_len = cpython.tuple.PyTuple_GET_SIZE(operation)
 */
  if (unlikely(__pyx_v_operations == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(4, 48, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_operations; __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 48, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(4, 48, __pyx_L1_error)
    #else
    __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_operation, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "asynctnt/iproto/requests/update.pyx":49
 * 
 *     for operation in operations:
 *         if isinstance(operation, tuple):             # <<<<<<<<<<<<<<
 *             op_len = cpython.tuple.PyTuple_GET_SIZE(operation)
 *         elif isinstance(operation, list):
 */
    __pyx_t_1 = PyTuple_Check(__pyx_v_operation); 
    if (__pyx_t_1) {

      /* "asynctnt/iproto/requests/update.pyx":50
 *     for operation in operations:
 *         if isinstance(operation, tuple):
 *             op_len = cpython.tuple.PyTuple_GET_SIZE(operation)             # <<<<<<<<<<<<<<
 *         elif isinstance(operation, list):
 *             op_len = cpython.list.PyList_GET_SIZE(operation)
 */
      __pyx_v_op_len = PyTuple_GET_SIZE(__pyx_v_operation);

      /* "asynctnt/iproto/requests/update.pyx":49
 * 
 *     for operation in operations:
 *         if isinstance(operation, tuple):             # <<<<<<<<<<<<<<
 *             op_len = cpython.tuple.PyTuple_GET_SIZE(operation)
 *         elif isinstance(operation, list):
 */
      goto __pyx_L7;
    }

    /* "asynctnt/iproto/requests/update.pyx":51
 *         if isinstance(operation, tuple):
 *             op_len = cpython.tuple.PyTuple_GET_SIZE(operation)
 *         elif isinstance(operation, list):             # <<<<<<<<<<<<<<
 *             op_len = cpython.list.PyList_GET_SIZE(operation)
 *         else:
 */
    __pyx_t_1 = PyList_Check(__pyx_v_operation); 
    if (likely(__pyx_t_1)) {

      /* "asynctnt/iproto/requests/update.pyx":52
 *             op_len = cpython.tuple.PyTuple_GET_SIZE(operation)
 *         elif isinstance(operation, list):
 *             op_len = cpython.list.PyList_GET_SIZE(operation)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError(
 */
      __pyx_v_op_len = PyList_GET_SIZE(__pyx_v_operation);

      /* "asynctnt/iproto/requests/update.pyx":51
 *         if isinstance(operation, tuple):
 *             op_len = cpython.tuple.PyTuple_GET_SIZE(operation)
 *         elif isinstance(operation, list):             # <<<<<<<<<<<<<<
 *             op_len = cpython.list.PyList_GET_SIZE(operation)
 *         else:
 */
      goto __pyx_L7;
    }

    /* "asynctnt/iproto/requests/update.pyx":54
 *             op_len = cpython.list.PyList_GET_SIZE(operation)
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 'Single operation must be a tuple or list')
 *         if op_len < 3:
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 54, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(4, 54, __pyx_L1_error)
    }
    __pyx_L7:;

    /* "asynctnt/iproto/requests/update.pyx":56
 *             raise TypeError(
 *                 'Single operation must be a tuple or list')
 *         if op_len < 3:             # <<<<<<<<<<<<<<
 *             raise IndexError(
 *                 'Operation length must be at least 3')
 */
    __pyx_t_1 = (__pyx_v_op_len < 3);
    if (unlikely(__pyx_t_1)) {

      /* "asynctnt/iproto/requests/update.pyx":57
 *                 'Single operation must be a tuple or list')
 *         if op_len < 3:
 *             raise IndexError(             # <<<<<<<<<<<<<<
 *                 'Operation length must be at least 3')
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 57, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(4, 57, __pyx_L1_error)

      /* "asynctnt/iproto/requests/update.pyx":56
 *             raise TypeError(
 *                 'Single operation must be a tuple or list')
 *         if op_len < 3:             # <<<<<<<<<<<<<<
 *             raise IndexError(
 *                 'Operation length must be at least 3')
 */
    }

    /* "asynctnt/iproto/requests/update.pyx":60
 *                 'Operation length must be at least 3')
 * 
 *         op_type_str = operation[0]             # <<<<<<<<<<<<<<
 *         if isinstance(op_type_str, str):
 *             str_temp = encode_unicode_string(op_type_str, buffer._encoding)
 */
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_operation, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_op_type_str, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "asynctnt/iproto/requests/update.pyx":61
 * 
 *         op_type_str = operation[0]
 *         if isinstance(op_type_str, str):             # <<<<<<<<<<<<<<
 *             str_temp = encode_unicode_string(op_type_str, buffer._encoding)
 *         elif isinstance(op_type_str, bytes):
 */
    __pyx_t_1 = PyUnicode_Check(__pyx_v_op_type_str); 
    if (__pyx_t_1) {

      /* "asynctnt/iproto/requests/update.pyx":62
 *         op_type_str = operation[0]
 *         if isinstance(op_type_str, str):
 *             str_temp = encode_unicode_string(op_type_str, buffer._encoding)             # <<<<<<<<<<<<<<
 *         elif isinstance(op_type_str, bytes):
 *             str_temp = <bytes> op_type_str
 */
      __pyx_t_5 = __pyx_v_buffer->_encoding;
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_6 = __pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string(__pyx_v_op_type_str, ((PyObject*)__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 62, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_str_temp, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "asynctnt/iproto/requests/update.pyx":61
 * 
 *         op_type_str = operation[0]
 *         if isinstance(op_type_str, str):             # <<<<<<<<<<<<<<
 *             str_temp = encode_unicode_string(op_type_str, buffer._encoding)
 *         elif isinstance(op_type_str, bytes):
 */
      goto __pyx_L9;
    }

    /* "asynctnt/iproto/requests/update.pyx":63
 *         if isinstance(op_type_str, str):
 *             str_temp = encode_unicode_string(op_type_str, buffer._encoding)
 *         elif isinstance(op_type_str, bytes):             # <<<<<<<<<<<<<<
 *             str_temp = <bytes> op_type_str
 *         else:
 */
    __pyx_t_1 = PyBytes_Check(__pyx_v_op_type_str); 
    if (likely(__pyx_t_1)) {

      /* "asynctnt/iproto/requests/update.pyx":64
 *             str_temp = encode_unicode_string(op_type_str, buffer._encoding)
 *         elif isinstance(op_type_str, bytes):
 *             str_temp = <bytes> op_type_str             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError(
 */
      __pyx_t_6 = __pyx_v_op_type_str;
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_str_temp, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "asynctnt/iproto/requests/update.pyx":63
 *         if isinstance(op_type_str, str):
 *             str_temp = encode_unicode_string(op_type_str, buffer._encoding)
 *         elif isinstance(op_type_str, bytes):             # <<<<<<<<<<<<<<
 *             str_temp = <bytes> op_type_str
 *         else:
 */
      goto __pyx_L9;
    }

    /* "asynctnt/iproto/requests/update.pyx":66
 *             str_temp = <bytes> op_type_str
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 'Operation type must of a str or bytes type')
 * 
 */
    /*else*/ {
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 66, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(4, 66, __pyx_L1_error)
    }
    __pyx_L9:;

    /* "asynctnt/iproto/requests/update.pyx":69
 *                 'Operation type must of a str or bytes type')
 * 
 *         cpython.bytes.PyBytes_AsStringAndSize(str_temp, &op_str_c,             # <<<<<<<<<<<<<<
 *                                               &op_str_len)
 * 
 */
    __pyx_t_7 = PyBytes_AsStringAndSize(__pyx_v_str_temp, (&__pyx_v_op_str_c), ((Py_ssize_t *)(&__pyx_v_op_str_len))); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(4, 69, __pyx_L1_error)

    /* "asynctnt/iproto/requests/update.pyx":72
 *                                               &op_str_len)
 * 
 *         field_no_obj = operation[1]             # <<<<<<<<<<<<<<
 *         if isinstance(field_no_obj, int):
 *             field_no = <int> field_no_obj
 */
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_operation, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_field_no_obj, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "asynctnt/iproto/requests/update.pyx":73
 * 
 *         field_no_obj = operation[1]
 *         if isinstance(field_no_obj, int):             # <<<<<<<<<<<<<<
 *             field_no = <int> field_no_obj
 *         elif isinstance(field_no_obj, str):
 */
    __pyx_t_1 = PyInt_Check(__pyx_v_field_no_obj); 
    if (__pyx_t_1) {

      /* "asynctnt/iproto/requests/update.pyx":74
 *         field_no_obj = operation[1]
 *         if isinstance(field_no_obj, int):
 *             field_no = <int> field_no_obj             # <<<<<<<<<<<<<<
 *         elif isinstance(field_no_obj, str):
 *             if space.metadata is not None:
 */
      __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_field_no_obj); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 74, __pyx_L1_error)
      __pyx_v_field_no = ((int)__pyx_t_7);

      /* "asynctnt/iproto/requests/update.pyx":73
 * 
 *         field_no_obj = operation[1]
 *         if isinstance(field_no_obj, int):             # <<<<<<<<<<<<<<
 *             field_no = <int> field_no_obj
 *         elif isinstance(field_no_obj, str):
 */
      goto __pyx_L10;
    }

    /* "asynctnt/iproto/requests/update.pyx":75
 *         if isinstance(field_no_obj, int):
 *             field_no = <int> field_no_obj
 *         elif isinstance(field_no_obj, str):             # <<<<<<<<<<<<<<
 *             if space.metadata is not None:
 *                 field_no = <int> space.metadata.id_by_name_safe(field_no_obj)
 */
    __pyx_t_1 = PyUnicode_Check(__pyx_v_field_no_obj); 
    if (likely(__pyx_t_1)) {

      /* "asynctnt/iproto/requests/update.pyx":76
 *             field_no = <int> field_no_obj
 *         elif isinstance(field_no_obj, str):
 *             if space.metadata is not None:             # <<<<<<<<<<<<<<
 *                 field_no = <int> space.metadata.id_by_name_safe(field_no_obj)
 *                 if field_no == -1:
 */
      __pyx_t_1 = (((PyObject *)__pyx_v_space->metadata) != Py_None);
      if (__pyx_t_1) {

        /* "asynctnt/iproto/requests/update.pyx":77
 *         elif isinstance(field_no_obj, str):
 *             if space.metadata is not None:
 *                 field_no = <int> space.metadata.id_by_name_safe(field_no_obj)             # <<<<<<<<<<<<<<
 *                 if field_no == -1:
 *                     field_encode_as_str = 1
 */
        if (!(likely(PyUnicode_CheckExact(__pyx_v_field_no_obj))||((__pyx_v_field_no_obj) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_field_no_obj))) __PYX_ERR(4, 77, __pyx_L1_error)
        __pyx_t_7 = __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_id_by_name_safe(__pyx_v_space->metadata, ((PyObject*)__pyx_v_field_no_obj)); if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 77, __pyx_L1_error)
        __pyx_v_field_no = ((int)__pyx_t_7);

        /* "asynctnt/iproto/requests/update.pyx":78
 *             if space.metadata is not None:
 *                 field_no = <int> space.metadata.id_by_name_safe(field_no_obj)
 *                 if field_no == -1:             # <<<<<<<<<<<<<<
 *                     field_encode_as_str = 1
 *             else:
 */
        __pyx_t_1 = (__pyx_v_field_no == -1LL);
        if (__pyx_t_1) {

          /* "asynctnt/iproto/requests/update.pyx":79
 *                 field_no = <int> space.metadata.id_by_name_safe(field_no_obj)
 *                 if field_no == -1:
 *                     field_encode_as_str = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 field_encode_as_str = 1
 */
          __pyx_v_field_encode_as_str = 1;

          /* "asynctnt/iproto/requests/update.pyx":78
 *             if space.metadata is not None:
 *                 field_no = <int> space.metadata.id_by_name_safe(field_no_obj)
 *                 if field_no == -1:             # <<<<<<<<<<<<<<
 *                     field_encode_as_str = 1
 *             else:
 */
        }

        /* "asynctnt/iproto/requests/update.pyx":76
 *             field_no = <int> field_no_obj
 *         elif isinstance(field_no_obj, str):
 *             if space.metadata is not None:             # <<<<<<<<<<<<<<
 *                 field_no = <int> space.metadata.id_by_name_safe(field_no_obj)
 *                 if field_no == -1:
 */
        goto __pyx_L11;
      }

      /* "asynctnt/iproto/requests/update.pyx":81
 *                     field_encode_as_str = 1
 *             else:
 *                 field_encode_as_str = 1             # <<<<<<<<<<<<<<
 * 
 *             if field_encode_as_str:
 */
      /*else*/ {
        __pyx_v_field_encode_as_str = 1;
      }
      __pyx_L11:;

      /* "asynctnt/iproto/requests/update.pyx":83
 *                 field_encode_as_str = 1
 * 
 *             if field_encode_as_str:             # <<<<<<<<<<<<<<
 *                 str_temp = encode_unicode_string(field_no_obj, buffer._encoding)
 *                 cpython.bytes.PyBytes_AsStringAndSize(str_temp, &field_str_c, &field_str_len)
 */
      if (__pyx_v_field_encode_as_str) {

        /* "asynctnt/iproto/requests/update.pyx":84
 * 
 *             if field_encode_as_str:
 *                 str_temp = encode_unicode_string(field_no_obj, buffer._encoding)             # <<<<<<<<<<<<<<
 *                 cpython.bytes.PyBytes_AsStringAndSize(str_temp, &field_str_c, &field_str_len)
 *         else:
 */
        __pyx_t_6 = __pyx_v_buffer->_encoding;
        __Pyx_INCREF(__pyx_t_6);
        __pyx_t_5 = __pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string(__pyx_v_field_no_obj, ((PyObject*)__pyx_t_6)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 84, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF_SET(__pyx_v_str_temp, ((PyObject*)__pyx_t_5));
        __pyx_t_5 = 0;

        /* "asynctnt/iproto/requests/update.pyx":85
 *             if field_encode_as_str:
 *                 str_temp = encode_unicode_string(field_no_obj, buffer._encoding)
 *                 cpython.bytes.PyBytes_AsStringAndSize(str_temp, &field_str_c, &field_str_len)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError(
 */
        __pyx_t_7 = PyBytes_AsStringAndSize(__pyx_v_str_temp, (&__pyx_v_field_str_c), ((Py_ssize_t *)(&__pyx_v_field_str_len))); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(4, 85, __pyx_L1_error)

        /* "asynctnt/iproto/requests/update.pyx":83
 *                 field_encode_as_str = 1
 * 
 *             if field_encode_as_str:             # <<<<<<<<<<<<<<
 *                 str_temp = encode_unicode_string(field_no_obj, buffer._encoding)
 *                 cpython.bytes.PyBytes_AsStringAndSize(str_temp, &field_str_c, &field_str_len)
 */
      }

      /* "asynctnt/iproto/requests/update.pyx":75
 *         if isinstance(field_no_obj, int):
 *             field_no = <int> field_no_obj
 *         elif isinstance(field_no_obj, str):             # <<<<<<<<<<<<<<
 *             if space.metadata is not None:
 *                 field_no = <int> space.metadata.id_by_name_safe(field_no_obj)
 */
      goto __pyx_L10;
    }

    /* "asynctnt/iproto/requests/update.pyx":87
 *                 cpython.bytes.PyBytes_AsStringAndSize(str_temp, &field_str_c, &field_str_len)
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 'Operation field_no must be of either int or str type')
 * 
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 87, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(4, 87, __pyx_L1_error)
    }
    __pyx_L10:;

    /* "asynctnt/iproto/requests/update.pyx":90
 *                 'Operation field_no must be of either int or str type')
 * 
 *         op = <char> 0             # <<<<<<<<<<<<<<
 *         if op_str_len == 1:
 *             op = op_str_c[0]
 */
    __pyx_v_op = ((char)0);

    /* "asynctnt/iproto/requests/update.pyx":91
 * 
 *         op = <char> 0
 *         if op_str_len == 1:             # <<<<<<<<<<<<<<
 *             op = op_str_c[0]
 * 
 */
    __pyx_t_1 = (__pyx_v_op_str_len == 1);
    if (__pyx_t_1) {

      /* "asynctnt/iproto/requests/update.pyx":92
 *         op = <char> 0
 *         if op_str_len == 1:
 *             op = op_str_c[0]             # <<<<<<<<<<<<<<
 * 
 *         if op == tarantool.IPROTO_OP_ADD \
 */
      __pyx_v_op = (__pyx_v_op_str_c[0]);

      /* "asynctnt/iproto/requests/update.pyx":91
 * 
 *         op = <char> 0
 *         if op_str_len == 1:             # <<<<<<<<<<<<<<
 *             op = op_str_c[0]
 * 
 */
    }

    /* "asynctnt/iproto/requests/update.pyx":94
 *             op = op_str_c[0]
 * 
 *         if op == tarantool.IPROTO_OP_ADD \             # <<<<<<<<<<<<<<
 *                 or op == tarantool.IPROTO_OP_SUB \
 *                 or op == tarantool.IPROTO_OP_AND \
 */
    switch (__pyx_v_op) {
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_ADD:
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_SUB:

      /* "asynctnt/iproto/requests/update.pyx":95
 * 
 *         if op == tarantool.IPROTO_OP_ADD \
 *                 or op == tarantool.IPROTO_OP_SUB \             # <<<<<<<<<<<<<<
 *                 or op == tarantool.IPROTO_OP_AND \
 *                 or op == tarantool.IPROTO_OP_XOR \
 */
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_AND:

      /* "asynctnt/iproto/requests/update.pyx":96
 *         if op == tarantool.IPROTO_OP_ADD \
 *                 or op == tarantool.IPROTO_OP_SUB \
 *                 or op == tarantool.IPROTO_OP_AND \             # <<<<<<<<<<<<<<
 *                 or op == tarantool.IPROTO_OP_XOR \
 *                 or op == tarantool.IPROTO_OP_OR \
 */
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_XOR:

      /* "asynctnt/iproto/requests/update.pyx":97
 *                 or op == tarantool.IPROTO_OP_SUB \
 *                 or op == tarantool.IPROTO_OP_AND \
 *                 or op == tarantool.IPROTO_OP_XOR \             # <<<<<<<<<<<<<<
 *                 or op == tarantool.IPROTO_OP_OR \
 *                 or op == tarantool.IPROTO_OP_DELETE \
 */
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_OR:

      /* "asynctnt/iproto/requests/update.pyx":98
 *                 or op == tarantool.IPROTO_OP_AND \
 *                 or op == tarantool.IPROTO_OP_XOR \
 *                 or op == tarantool.IPROTO_OP_OR \             # <<<<<<<<<<<<<<
 *                 or op == tarantool.IPROTO_OP_DELETE \
 *                 or op == tarantool.IPROTO_OP_INSERT \
 */
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_DELETE:

      /* "asynctnt/iproto/requests/update.pyx":99
 *                 or op == tarantool.IPROTO_OP_XOR \
 *                 or op == tarantool.IPROTO_OP_OR \
 *                 or op == tarantool.IPROTO_OP_DELETE \             # <<<<<<<<<<<<<<
 *                 or op == tarantool.IPROTO_OP_INSERT \
 *                 or op == tarantool.IPROTO_OP_ASSIGN:
 */
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_INSERT:

      /* "asynctnt/iproto/requests/update.pyx":100
 *                 or op == tarantool.IPROTO_OP_OR \
 *                 or op == tarantool.IPROTO_OP_DELETE \
 *                 or op == tarantool.IPROTO_OP_INSERT \             # <<<<<<<<<<<<<<
 *                 or op == tarantool.IPROTO_OP_ASSIGN:
 *             # mp_sizeof_array(3)
 */
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_ASSIGN:

      /* "asynctnt/iproto/requests/update.pyx":105
 *             # + mp_sizeof_str(1)
 *             # + mp_sizeof_uint(field_no)
 *             extra_length = 1 + 2 + mp_sizeof_uint(field_no)             # <<<<<<<<<<<<<<
 *             p = begin = buffer._ensure_allocated(p, extra_length)
 * 
 */
      __pyx_v_extra_length = (0x3 + mp_sizeof_uint(__pyx_v_field_no));

      /* "asynctnt/iproto/requests/update.pyx":106
 *             # + mp_sizeof_uint(field_no)
 *             extra_length = 1 + 2 + mp_sizeof_uint(field_no)
 *             p = begin = buffer._ensure_allocated(p, extra_length)             # <<<<<<<<<<<<<<
 * 
 *             p = mp_encode_array(p, 3)
 */
      __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_buffer, __pyx_v_p, __pyx_v_extra_length); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(4, 106, __pyx_L1_error)
      __pyx_v_p = __pyx_t_2;
      __pyx_v_begin = __pyx_t_2;

      /* "asynctnt/iproto/requests/update.pyx":108
 *             p = begin = buffer._ensure_allocated(p, extra_length)
 * 
 *             p = mp_encode_array(p, 3)             # <<<<<<<<<<<<<<
 *             p = mp_encode_str(p, op_str_c, 1)
 *             if field_str_c == NULL:
 */
      __pyx_v_p = mp_encode_array(__pyx_v_p, 3);

      /* "asynctnt/iproto/requests/update.pyx":109
 * 
 *             p = mp_encode_array(p, 3)
 *             p = mp_encode_str(p, op_str_c, 1)             # <<<<<<<<<<<<<<
 *             if field_str_c == NULL:
 *                 p = mp_encode_uint(p, field_no)
 */
      __pyx_v_p = mp_encode_str(__pyx_v_p, __pyx_v_op_str_c, 1);

      /* "asynctnt/iproto/requests/update.pyx":110
 *             p = mp_encode_array(p, 3)
 *             p = mp_encode_str(p, op_str_c, 1)
 *             if field_str_c == NULL:             # <<<<<<<<<<<<<<
 *                 p = mp_encode_uint(p, field_no)
 *             else:
 */
      __pyx_t_1 = (__pyx_v_field_str_c == NULL);
      if (__pyx_t_1) {

        /* "asynctnt/iproto/requests/update.pyx":111
 *             p = mp_encode_str(p, op_str_c, 1)
 *             if field_str_c == NULL:
 *                 p = mp_encode_uint(p, field_no)             # <<<<<<<<<<<<<<
 *             else:
 *                 p = mp_encode_str(p, field_str_c, field_str_len)
 */
        __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_field_no);

        /* "asynctnt/iproto/requests/update.pyx":110
 *             p = mp_encode_array(p, 3)
 *             p = mp_encode_str(p, op_str_c, 1)
 *             if field_str_c == NULL:             # <<<<<<<<<<<<<<
 *                 p = mp_encode_uint(p, field_no)
 *             else:
 */
        goto __pyx_L15;
      }

      /* "asynctnt/iproto/requests/update.pyx":113
 *                 p = mp_encode_uint(p, field_no)
 *             else:
 *                 p = mp_encode_str(p, field_str_c, field_str_len)             # <<<<<<<<<<<<<<
 * 
 *             buffer._length += (p - begin)
 */
      /*else*/ {
        __pyx_v_p = mp_encode_str(__pyx_v_p, __pyx_v_field_str_c, __pyx_v_field_str_len);
      }
      __pyx_L15:;

      /* "asynctnt/iproto/requests/update.pyx":115
 *                 p = mp_encode_str(p, field_str_c, field_str_len)
 * 
 *             buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 * 
 *             op_argument = operation[2]
 */
      __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

      /* "asynctnt/iproto/requests/update.pyx":117
 *             buffer._length += (p - begin)
 * 
 *             op_argument = operation[2]             # <<<<<<<<<<<<<<
 *             p = buffer.mp_encode_obj(p, op_argument)
 * 
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_operation, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_op_argument, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "asynctnt/iproto/requests/update.pyx":118
 * 
 *             op_argument = operation[2]
 *             p = buffer.mp_encode_obj(p, op_argument)             # <<<<<<<<<<<<<<
 * 
 *         elif op == tarantool.IPROTO_OP_SPLICE:
 */
      __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_obj(__pyx_v_buffer, __pyx_v_p, __pyx_v_op_argument); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(4, 118, __pyx_L1_error)
      __pyx_v_p = __pyx_t_2;

      /* "asynctnt/iproto/requests/update.pyx":94
 *             op = op_str_c[0]
 * 
 *         if op == tarantool.IPROTO_OP_ADD \             # <<<<<<<<<<<<<<
 *                 or op == tarantool.IPROTO_OP_SUB \
 *                 or op == tarantool.IPROTO_OP_AND \
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OP_SPLICE:

      /* "asynctnt/iproto/requests/update.pyx":121
 * 
 *         elif op == tarantool.IPROTO_OP_SPLICE:
 *             if op_len < 5:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'Splice operation must have length of 5, '
 */
      __pyx_t_1 = (__pyx_v_op_len < 5);
      if (unlikely(__pyx_t_1)) {

        /* "asynctnt/iproto/requests/update.pyx":124
 *                 raise ValueError(
 *                     'Splice operation must have length of 5, '
 *                     'but got: {}'.format(op_len)             # <<<<<<<<<<<<<<
 *                 )
 * 
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Splice_operation_must_have_lengt, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 124, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = __Pyx_PyInt_From_uint32_t(__pyx_v_op_len); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 124, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_8};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 124, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }

        /* "asynctnt/iproto/requests/update.pyx":122
 *         elif op == tarantool.IPROTO_OP_SPLICE:
 *             if op_len < 5:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'Splice operation must have length of 5, '
 *                     'but got: {}'.format(op_len)
 */
        __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 122, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_Raise(__pyx_t_6, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __PYX_ERR(4, 122, __pyx_L1_error)

        /* "asynctnt/iproto/requests/update.pyx":121
 * 
 *         elif op == tarantool.IPROTO_OP_SPLICE:
 *             if op_len < 5:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'Splice operation must have length of 5, '
 */
      }

      /* "asynctnt/iproto/requests/update.pyx":127
 *                 )
 * 
 *             splice_position_obj = operation[2]             # <<<<<<<<<<<<<<
 *             splice_offset_obj = operation[3]
 *             op_argument = operation[4]
 */
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_operation, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_splice_position_obj, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "asynctnt/iproto/requests/update.pyx":128
 * 
 *             splice_position_obj = operation[2]
 *             splice_offset_obj = operation[3]             # <<<<<<<<<<<<<<
 *             op_argument = operation[4]
 *             if not isinstance(splice_position_obj, int):
 */
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_operation, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_splice_offset_obj, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "asynctnt/iproto/requests/update.pyx":129
 *             splice_position_obj = operation[2]
 *             splice_offset_obj = operation[3]
 *             op_argument = operation[4]             # <<<<<<<<<<<<<<
 *             if not isinstance(splice_position_obj, int):
 *                 raise TypeError('Splice position must be int')
 */
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_operation, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_op_argument, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "asynctnt/iproto/requests/update.pyx":130
 *             splice_offset_obj = operation[3]
 *             op_argument = operation[4]
 *             if not isinstance(splice_position_obj, int):             # <<<<<<<<<<<<<<
 *                 raise TypeError('Splice position must be int')
 *             if not isinstance(splice_offset_obj, int):
 */
      __pyx_t_1 = PyInt_Check(__pyx_v_splice_position_obj); 
      __pyx_t_10 = (!__pyx_t_1);
      if (unlikely(__pyx_t_10)) {

        /* "asynctnt/iproto/requests/update.pyx":131
 *             op_argument = operation[4]
 *             if not isinstance(splice_position_obj, int):
 *                 raise TypeError('Splice position must be int')             # <<<<<<<<<<<<<<
 *             if not isinstance(splice_offset_obj, int):
 *                 raise TypeError('Splice offset must be int')
 */
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 131, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_Raise(__pyx_t_6, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __PYX_ERR(4, 131, __pyx_L1_error)

        /* "asynctnt/iproto/requests/update.pyx":130
 *             splice_offset_obj = operation[3]
 *             op_argument = operation[4]
 *             if not isinstance(splice_position_obj, int):             # <<<<<<<<<<<<<<
 *                 raise TypeError('Splice position must be int')
 *             if not isinstance(splice_offset_obj, int):
 */
      }

      /* "asynctnt/iproto/requests/update.pyx":132
 *             if not isinstance(splice_position_obj, int):
 *                 raise TypeError('Splice position must be int')
 *             if not isinstance(splice_offset_obj, int):             # <<<<<<<<<<<<<<
 *                 raise TypeError('Splice offset must be int')
 * 
 */
      __pyx_t_10 = PyInt_Check(__pyx_v_splice_offset_obj); 
      __pyx_t_1 = (!__pyx_t_10);
      if (unlikely(__pyx_t_1)) {

        /* "asynctnt/iproto/requests/update.pyx":133
 *                 raise TypeError('Splice position must be int')
 *             if not isinstance(splice_offset_obj, int):
 *                 raise TypeError('Splice offset must be int')             # <<<<<<<<<<<<<<
 * 
 *             splice_position = <uint32_t> splice_position_obj
 */
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 133, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_Raise(__pyx_t_6, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __PYX_ERR(4, 133, __pyx_L1_error)

        /* "asynctnt/iproto/requests/update.pyx":132
 *             if not isinstance(splice_position_obj, int):
 *                 raise TypeError('Splice position must be int')
 *             if not isinstance(splice_offset_obj, int):             # <<<<<<<<<<<<<<
 *                 raise TypeError('Splice offset must be int')
 * 
 */
      }

      /* "asynctnt/iproto/requests/update.pyx":135
 *                 raise TypeError('Splice offset must be int')
 * 
 *             splice_position = <uint32_t> splice_position_obj             # <<<<<<<<<<<<<<
 *             splice_offset = <uint32_t> splice_offset_obj
 * 
 */
      __pyx_t_11 = __Pyx_PyInt_As_uint32_t(__pyx_v_splice_position_obj); if (unlikely((__pyx_t_11 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(4, 135, __pyx_L1_error)
      __pyx_v_splice_position = ((uint32_t)__pyx_t_11);

      /* "asynctnt/iproto/requests/update.pyx":136
 * 
 *             splice_position = <uint32_t> splice_position_obj
 *             splice_offset = <uint32_t> splice_offset_obj             # <<<<<<<<<<<<<<
 * 
 *             # mp_sizeof_array(5) + mp_sizeof_str(1) + ...
 */
      __pyx_t_11 = __Pyx_PyInt_As_uint32_t(__pyx_v_splice_offset_obj); if (unlikely((__pyx_t_11 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(4, 136, __pyx_L1_error)
      __pyx_v_splice_offset = ((uint32_t)__pyx_t_11);

      /* "asynctnt/iproto/requests/update.pyx":142
 *                            + mp_sizeof_uint(field_no) \
 *                            + mp_sizeof_uint(splice_position) \
 *                            + mp_sizeof_uint(splice_offset)             # <<<<<<<<<<<<<<
 *             p = begin = buffer._ensure_allocated(p, extra_length)
 * 
 */
      __pyx_v_extra_length = (((0x3 + mp_sizeof_uint(__pyx_v_field_no)) + mp_sizeof_uint(__pyx_v_splice_position)) + mp_sizeof_uint(__pyx_v_splice_offset));

      /* "asynctnt/iproto/requests/update.pyx":143
 *                            + mp_sizeof_uint(splice_position) \
 *                            + mp_sizeof_uint(splice_offset)
 *             p = begin = buffer._ensure_allocated(p, extra_length)             # <<<<<<<<<<<<<<
 * 
 *             p = mp_encode_array(p, 5)
 */
      __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated(__pyx_v_buffer, __pyx_v_p, __pyx_v_extra_length); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(4, 143, __pyx_L1_error)
      __pyx_v_p = __pyx_t_2;
      __pyx_v_begin = __pyx_t_2;

      /* "asynctnt/iproto/requests/update.pyx":145
 *             p = begin = buffer._ensure_allocated(p, extra_length)
 * 
 *             p = mp_encode_array(p, 5)             # <<<<<<<<<<<<<<
 *             p = mp_encode_str(p, op_str_c, 1)
 *             if field_str_c == NULL:
 */
      __pyx_v_p = mp_encode_array(__pyx_v_p, 5);

      /* "asynctnt/iproto/requests/update.pyx":146
 * 
 *             p = mp_encode_array(p, 5)
 *             p = mp_encode_str(p, op_str_c, 1)             # <<<<<<<<<<<<<<
 *             if field_str_c == NULL:
 *                 p = mp_encode_uint(p, field_no)
 */
      __pyx_v_p = mp_encode_str(__pyx_v_p, __pyx_v_op_str_c, 1);

      /* "asynctnt/iproto/requests/update.pyx":147
 *             p = mp_encode_array(p, 5)
 *             p = mp_encode_str(p, op_str_c, 1)
 *             if field_str_c == NULL:             # <<<<<<<<<<<<<<
 *                 p = mp_encode_uint(p, field_no)
 *             else:
 */
      __pyx_t_1 = (__pyx_v_field_str_c == NULL);
      if (__pyx_t_1) {

        /* "asynctnt/iproto/requests/update.pyx":148
 *             p = mp_encode_str(p, op_str_c, 1)
 *             if field_str_c == NULL:
 *                 p = mp_encode_uint(p, field_no)             # <<<<<<<<<<<<<<
 *             else:
 *                 p = mp_encode_str(p, field_str_c, field_str_len)
 */
        __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_field_no);

        /* "asynctnt/iproto/requests/update.pyx":147
 *             p = mp_encode_array(p, 5)
 *             p = mp_encode_str(p, op_str_c, 1)
 *             if field_str_c == NULL:             # <<<<<<<<<<<<<<
 *                 p = mp_encode_uint(p, field_no)
 *             else:
 */
        goto __pyx_L19;
      }

      /* "asynctnt/iproto/requests/update.pyx":150
 *                 p = mp_encode_uint(p, field_no)
 *             else:
 *                 p = mp_encode_str(p, field_str_c, field_str_len)             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, splice_position)
 *             p = mp_encode_uint(p, splice_offset)
 */
      /*else*/ {
        __pyx_v_p = mp_encode_str(__pyx_v_p, __pyx_v_field_str_c, __pyx_v_field_str_len);
      }
      __pyx_L19:;

      /* "asynctnt/iproto/requests/update.pyx":151
 *             else:
 *                 p = mp_encode_str(p, field_str_c, field_str_len)
 *             p = mp_encode_uint(p, splice_position)             # <<<<<<<<<<<<<<
 *             p = mp_encode_uint(p, splice_offset)
 *             buffer._length += (p - begin)
 */
      __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_splice_position);

      /* "asynctnt/iproto/requests/update.pyx":152
 *                 p = mp_encode_str(p, field_str_c, field_str_len)
 *             p = mp_encode_uint(p, splice_position)
 *             p = mp_encode_uint(p, splice_offset)             # <<<<<<<<<<<<<<
 *             buffer._length += (p - begin)
 *             p = buffer.mp_encode_obj(p, op_argument)
 */
      __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_splice_offset);

      /* "asynctnt/iproto/requests/update.pyx":153
 *             p = mp_encode_uint(p, splice_position)
 *             p = mp_encode_uint(p, splice_offset)
 *             buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 *             p = buffer.mp_encode_obj(p, op_argument)
 *         else:
 */
      __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

      /* "asynctnt/iproto/requests/update.pyx":154
 *             p = mp_encode_uint(p, splice_offset)
 *             buffer._length += (p - begin)
 *             p = buffer.mp_encode_obj(p, op_argument)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError(
 */
      __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_obj(__pyx_v_buffer, __pyx_v_p, __pyx_v_op_argument); if (unlikely(__pyx_t_2 == ((char *)NULL))) __PYX_ERR(4, 154, __pyx_L1_error)
      __pyx_v_p = __pyx_t_2;

      /* "asynctnt/iproto/requests/update.pyx":120
 *             p = buffer.mp_encode_obj(p, op_argument)
 * 
 *         elif op == tarantool.IPROTO_OP_SPLICE:             # <<<<<<<<<<<<<<
 *             if op_len < 5:
 *                 raise ValueError(
 */
      break;
      default:

      /* "asynctnt/iproto/requests/update.pyx":157
 *         else:
 *             raise TypeError(
 *                 'Unknown update operation type `{}`'.format(op_type_str))             # <<<<<<<<<<<<<<
 *     return p
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Unknown_update_operation_type, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 157, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_op_type_str};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 157, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }

      /* "asynctnt/iproto/requests/update.pyx":156
 *             p = buffer.mp_encode_obj(p, op_argument)
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 'Unknown update operation type `{}`'.format(op_type_str))
 *     return p
 */
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(4, 156, __pyx_L1_error)
      break;
    }

    /* "asynctnt/iproto/requests/update.pyx":48
 *         return p
 * 
 *     for operation in operations:             # <<<<<<<<<<<<<<
 *         if isinstance(operation, tuple):
 *             op_len = cpython.tuple.PyTuple_GET_SIZE(operation)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/requests/update.pyx":158
 *             raise TypeError(
 *                 'Unknown update operation type `{}`'.format(op_type_str))
 *     return p             # <<<<<<<<<<<<<<
 * 
 * cdef int encode_request_update(WriteBuffer buffer,
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "asynctnt/iproto/requests/update.pyx":10
 *                                      self.key, self.operations, <bint> False)
 * 
 * cdef char *encode_update_ops(WriteBuffer buffer,             # <<<<<<<<<<<<<<
 *                              char *p, list operations,
 *                              SchemaSpace space) except NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.encode_update_ops", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_str_temp);
  __Pyx_XDECREF(__pyx_v_field_no_obj);
  __Pyx_XDECREF(__pyx_v_operation);
  __Pyx_XDECREF(__pyx_v_op_type_str);
  __Pyx_XDECREF(__pyx_v_op_argument);
  __Pyx_XDECREF(__pyx_v_splice_position_obj);
  __Pyx_XDECREF(__pyx_v_splice_offset_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/update.pyx":160
 *     return p
 * 
 * cdef int encode_request_update(WriteBuffer buffer,             # <<<<<<<<<<<<<<
 *                                SchemaSpace space, SchemaIndex index,
 *                                key_tuple, list operations,
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_encode_request_update(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_space, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_index, PyObject *__pyx_v_key_tuple, PyObject *__pyx_v_operations, int __pyx_v_is_upsert) {
  char *__pyx_v_begin;
  char *__pyx_v_p;
  uint32_t __pyx_v_body_map_sz;
  uint32_t __pyx_v_max_body_len;
  uint32_t __pyx_v_space_id;
  uint32_t __pyx_v_index_id;
  uint32_t __pyx_v_key_of_tuple;
  uint32_t __pyx_v_key_of_operations;
  struct C_Metadata *__pyx_v_metadata = 0;
  int __pyx_v_default_fields_none;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_request_update", 1);

  /* "asynctnt/iproto/requests/update.pyx":174
 *         bint default_fields_none
 * 
 *     space_id = space.sid             # <<<<<<<<<<<<<<
 *     index_id = index.iid
 * 
 */
  __pyx_t_1 = __pyx_v_space->sid;
  __pyx_v_space_id = __pyx_t_1;

  /* "asynctnt/iproto/requests/update.pyx":175
 * 
 *     space_id = space.sid
 *     index_id = index.iid             # <<<<<<<<<<<<<<
 * 
 *     if not is_upsert:
 */
  __pyx_t_1 = __pyx_v_index->iid;
  __pyx_v_index_id = __pyx_t_1;

  /* "asynctnt/iproto/requests/update.pyx":177
 *     index_id = index.iid
 * 
 *     if not is_upsert:             # <<<<<<<<<<<<<<
 *         key_of_tuple = tarantool.IPROTO_KEY
 *         key_of_operations = tarantool.IPROTO_TUPLE
 */
  __pyx_t_2 = (!__pyx_v_is_upsert);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/requests/update.pyx":178
 * 
 *     if not is_upsert:
 *         key_of_tuple = tarantool.IPROTO_KEY             # <<<<<<<<<<<<<<
 *         key_of_operations = tarantool.IPROTO_TUPLE
 *         metadata = index.metadata
 */
    __pyx_v_key_of_tuple = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_KEY;

    /* "asynctnt/iproto/requests/update.pyx":179
 *     if not is_upsert:
 *         key_of_tuple = tarantool.IPROTO_KEY
 *         key_of_operations = tarantool.IPROTO_TUPLE             # <<<<<<<<<<<<<<
 *         metadata = index.metadata
 *         default_fields_none = False
 */
    __pyx_v_key_of_operations = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TUPLE;

    /* "asynctnt/iproto/requests/update.pyx":180
 *         key_of_tuple = tarantool.IPROTO_KEY
 *         key_of_operations = tarantool.IPROTO_TUPLE
 *         metadata = index.metadata             # <<<<<<<<<<<<<<
 *         default_fields_none = False
 *     else:
 */
    __pyx_t_3 = ((PyObject *)__pyx_v_index->metadata);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_metadata = ((struct C_Metadata *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "asynctnt/iproto/requests/update.pyx":181
 *         key_of_operations = tarantool.IPROTO_TUPLE
 *         metadata = index.metadata
 *         default_fields_none = False             # <<<<<<<<<<<<<<
 *     else:
 *         key_of_tuple = tarantool.IPROTO_TUPLE
 */
    __pyx_v_default_fields_none = 0;

    /* "asynctnt/iproto/requests/update.pyx":177
 *     index_id = index.iid
 * 
 *     if not is_upsert:             # <<<<<<<<<<<<<<
 *         key_of_tuple = tarantool.IPROTO_KEY
 *         key_of_operations = tarantool.IPROTO_TUPLE
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/requests/update.pyx":183
 *         default_fields_none = False
 *     else:
 *         key_of_tuple = tarantool.IPROTO_TUPLE             # <<<<<<<<<<<<<<
 *         key_of_operations = tarantool.IPROTO_OPS
 *         metadata = space.metadata
 */
  /*else*/ {
    __pyx_v_key_of_tuple = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TUPLE;

    /* "asynctnt/iproto/requests/update.pyx":184
 *     else:
 *         key_of_tuple = tarantool.IPROTO_TUPLE
 *         key_of_operations = tarantool.IPROTO_OPS             # <<<<<<<<<<<<<<
 *         metadata = space.metadata
 *         default_fields_none = True
 */
    __pyx_v_key_of_operations = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_OPS;

    /* "asynctnt/iproto/requests/update.pyx":185
 *         key_of_tuple = tarantool.IPROTO_TUPLE
 *         key_of_operations = tarantool.IPROTO_OPS
 *         metadata = space.metadata             # <<<<<<<<<<<<<<
 *         default_fields_none = True
 * 
 */
    __pyx_t_3 = ((PyObject *)__pyx_v_space->metadata);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_metadata = ((struct C_Metadata *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "asynctnt/iproto/requests/update.pyx":186
 *         key_of_operations = tarantool.IPROTO_OPS
 *         metadata = space.metadata
 *         default_fields_none = True             # <<<<<<<<<<<<<<
 * 
 *     body_map_sz = 3 + <uint32_t> (index_id > 0)
 */
    __pyx_v_default_fields_none = 1;
  }
  __pyx_L3:;

  /* "asynctnt/iproto/requests/update.pyx":188
 *         default_fields_none = True
 * 
 *     body_map_sz = 3 + <uint32_t> (index_id > 0)             # <<<<<<<<<<<<<<
 *     # Size description:
 *     # mp_sizeof_map(body_map_sz)
 */
  __pyx_v_body_map_sz = (3 + ((uint32_t)(__pyx_v_index_id > 0)));

  /* "asynctnt/iproto/requests/update.pyx":195
 *     max_body_len = 1 \
 *                    + 1 \
 *                    + 9             # <<<<<<<<<<<<<<
 * 
 *     if index_id > 0:
 */
  __pyx_v_max_body_len = 0xb;

  /* "asynctnt/iproto/requests/update.pyx":197
 *                    + 9
 * 
 *     if index_id > 0:             # <<<<<<<<<<<<<<
 *         # + mp_sizeof_uint(TP_INDEX)
 *         # + mp_sizeof_uint(index)
 */
  __pyx_t_2 = (__pyx_v_index_id > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/requests/update.pyx":200
 *         # + mp_sizeof_uint(TP_INDEX)
 *         # + mp_sizeof_uint(index)
 *         max_body_len += 1 + 9             # <<<<<<<<<<<<<<
 * 
 *     max_body_len += 1  # + mp_sizeof_uint(TP_KEY)
 */
    __pyx_v_max_body_len = (__pyx_v_max_body_len + 0xa);

    /* "asynctnt/iproto/requests/update.pyx":197
 *                    + 9
 * 
 *     if index_id > 0:             # <<<<<<<<<<<<<<
 *         # + mp_sizeof_uint(TP_INDEX)
 *         # + mp_sizeof_uint(index)
 */
  }

  /* "asynctnt/iproto/requests/update.pyx":202
 *         max_body_len += 1 + 9
 * 
 *     max_body_len += 1  # + mp_sizeof_uint(TP_KEY)             # <<<<<<<<<<<<<<
 *     max_body_len += 1  # + mp_sizeof_uint(TP_TUPLE)
 * 
 */
  __pyx_v_max_body_len = (__pyx_v_max_body_len + 1);

  /* "asynctnt/iproto/requests/update.pyx":203
 * 
 *     max_body_len += 1  # + mp_sizeof_uint(TP_KEY)
 *     max_body_len += 1  # + mp_sizeof_uint(TP_TUPLE)             # <<<<<<<<<<<<<<
 * 
 *     buffer.ensure_allocated(max_body_len)
 */
  __pyx_v_max_body_len = (__pyx_v_max_body_len + 1);

  /* "asynctnt/iproto/requests/update.pyx":205
 *     max_body_len += 1  # + mp_sizeof_uint(TP_TUPLE)
 * 
 *     buffer.ensure_allocated(max_body_len)             # <<<<<<<<<<<<<<
 * 
 *     p = begin = &buffer._buf[buffer._length]
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_buffer, __pyx_v_max_body_len); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(4, 205, __pyx_L1_error)

  /* "asynctnt/iproto/requests/update.pyx":207
 *     buffer.ensure_allocated(max_body_len)
 * 
 *     p = begin = &buffer._buf[buffer._length]             # <<<<<<<<<<<<<<
 *     p = mp_encode_map(p, body_map_sz)
 *     p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 */
  __pyx_t_4 = (&(__pyx_v_buffer->_buf[__pyx_v_buffer->_length]));
  __pyx_v_p = __pyx_t_4;
  __pyx_v_begin = __pyx_t_4;

  /* "asynctnt/iproto/requests/update.pyx":208
 * 
 *     p = begin = &buffer._buf[buffer._length]
 *     p = mp_encode_map(p, body_map_sz)             # <<<<<<<<<<<<<<
 *     p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 *     p = mp_encode_uint(p, space_id)
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_body_map_sz);

  /* "asynctnt/iproto/requests/update.pyx":209
 *     p = begin = &buffer._buf[buffer._length]
 *     p = mp_encode_map(p, body_map_sz)
 *     p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)             # <<<<<<<<<<<<<<
 *     p = mp_encode_uint(p, space_id)
 * 
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SPACE_ID);

  /* "asynctnt/iproto/requests/update.pyx":210
 *     p = mp_encode_map(p, body_map_sz)
 *     p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
 *     p = mp_encode_uint(p, space_id)             # <<<<<<<<<<<<<<
 * 
 *     if index_id > 0:
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_space_id);

  /* "asynctnt/iproto/requests/update.pyx":212
 *     p = mp_encode_uint(p, space_id)
 * 
 *     if index_id > 0:             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *         p = mp_encode_uint(p, index_id)
 */
  __pyx_t_2 = (__pyx_v_index_id > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/requests/update.pyx":213
 * 
 *     if index_id > 0:
 *         p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, index_id)
 *     buffer._length += (p - begin)
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_INDEX_ID);

    /* "asynctnt/iproto/requests/update.pyx":214
 *     if index_id > 0:
 *         p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *         p = mp_encode_uint(p, index_id)             # <<<<<<<<<<<<<<
 *     buffer._length += (p - begin)
 * 
 */
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_index_id);

    /* "asynctnt/iproto/requests/update.pyx":212
 *     p = mp_encode_uint(p, space_id)
 * 
 *     if index_id > 0:             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *         p = mp_encode_uint(p, index_id)
 */
  }

  /* "asynctnt/iproto/requests/update.pyx":215
 *         p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
 *         p = mp_encode_uint(p, index_id)
 *     buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 * 
 *     p = buffer.mp_encode_uint(p, key_of_tuple)
 */
  __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/requests/update.pyx":217
 *     buffer._length += (p - begin)
 * 
 *     p = buffer.mp_encode_uint(p, key_of_tuple)             # <<<<<<<<<<<<<<
 *     p = encode_key_sequence(buffer, p, key_tuple, metadata, default_fields_none)
 * 
 */
  __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uint(__pyx_v_buffer, __pyx_v_p, __pyx_v_key_of_tuple); if (unlikely(__pyx_t_4 == ((char *)NULL))) __PYX_ERR(4, 217, __pyx_L1_error)
  __pyx_v_p = __pyx_t_4;

  /* "asynctnt/iproto/requests/update.pyx":218
 * 
 *     p = buffer.mp_encode_uint(p, key_of_tuple)
 *     p = encode_key_sequence(buffer, p, key_tuple, metadata, default_fields_none)             # <<<<<<<<<<<<<<
 * 
 *     p = buffer.mp_encode_uint(p, key_of_operations)
 */
  __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_encode_key_sequence(__pyx_v_buffer, __pyx_v_p, __pyx_v_key_tuple, __pyx_v_metadata, __pyx_v_default_fields_none); if (unlikely(__pyx_t_4 == ((char *)NULL))) __PYX_ERR(4, 218, __pyx_L1_error)
  __pyx_v_p = __pyx_t_4;

  /* "asynctnt/iproto/requests/update.pyx":220
 *     p = encode_key_sequence(buffer, p, key_tuple, metadata, default_fields_none)
 * 
 *     p = buffer.mp_encode_uint(p, key_of_operations)             # <<<<<<<<<<<<<<
 *     p = encode_update_ops(buffer, p, operations, space)
 */
  __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uint(__pyx_v_buffer, __pyx_v_p, __pyx_v_key_of_operations); if (unlikely(__pyx_t_4 == ((char *)NULL))) __PYX_ERR(4, 220, __pyx_L1_error)
  __pyx_v_p = __pyx_t_4;

  /* "asynctnt/iproto/requests/update.pyx":221
 * 
 *     p = buffer.mp_encode_uint(p, key_of_operations)
 *     p = encode_update_ops(buffer, p, operations, space)             # <<<<<<<<<<<<<<
 */
  __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_encode_update_ops(__pyx_v_buffer, __pyx_v_p, __pyx_v_operations, __pyx_v_space); if (unlikely(__pyx_t_4 == ((char *)NULL))) __PYX_ERR(4, 221, __pyx_L1_error)
  __pyx_v_p = __pyx_t_4;

  /* "asynctnt/iproto/requests/update.pyx":160
 *     return p
 * 
 * cdef int encode_request_update(WriteBuffer buffer,             # <<<<<<<<<<<<<<
 *                                SchemaSpace space, SchemaIndex index,
 *                                key_tuple, list operations,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.encode_request_update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_metadata);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/upsert.pyx":6
 * @cython.final
 * cdef class UpsertRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         return encode_request_update(buffer, self.space, self.space.get_index(0),
 *                                      self.t, self.operations, <bint> True)
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_13UpsertRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_body", 1);

  /* "asynctnt/iproto/requests/upsert.pyx":7
 * cdef class UpsertRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         return encode_request_update(buffer, self.space, self.space.get_index(0),             # <<<<<<<<<<<<<<
 *                                      self.t, self.operations, <bint> True)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index(__pyx_v_self->__pyx_base.space, __pyx_int_0, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(19, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "asynctnt/iproto/requests/upsert.pyx":8
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         return encode_request_update(buffer, self.space, self.space.get_index(0),
 *                                      self.t, self.operations, <bint> True)             # <<<<<<<<<<<<<<
 */
  __pyx_t_3 = __pyx_v_self->t;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = __pyx_v_self->operations;
  __Pyx_INCREF(__pyx_t_4);

  /* "asynctnt/iproto/requests/upsert.pyx":7
 * cdef class UpsertRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:
 *         return encode_request_update(buffer, self.space, self.space.get_index(0),             # <<<<<<<<<<<<<<
 *                                      self.t, self.operations, <bint> True)
 */
  __pyx_t_5 = __pyx_f_8asynctnt_6iproto_8protocol_encode_request_update(__pyx_v_buffer, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_t_2), __pyx_t_3, ((PyObject*)__pyx_t_4), 1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(19, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "asynctnt/iproto/requests/upsert.pyx":6
 * @cython.final
 * cdef class UpsertRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         return encode_request_update(buffer, self.space, self.space.get_index(0),
 *                                      self.t, self.operations, <bint> True)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.UpsertRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13UpsertRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13UpsertRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13UpsertRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13UpsertRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13UpsertRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13UpsertRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.operations, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.t, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(13); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->operations);
  __Pyx_GIVEREF(__pyx_v_self->operations);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_self->operations)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->t);
  __Pyx_GIVEREF(__pyx_v_self->t);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_v_self->t)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 11, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 12, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.operations, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.t, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.operations, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.t, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.operations is not None or self.space is not None or self.t is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.operations, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.t, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.operations is not None or self.space is not None or self.t is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, None), state
 */
  /*else*/ {
    __pyx_t_11 = (__pyx_v_self->operations != ((PyObject*)Py_None));
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->t != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.operations is not None or self.space is not None or self.t is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.operations is not None or self.space is not None or self.t is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_UpsertRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_209859363);
    __Pyx_GIVEREF(__pyx_int_209859363);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_209859363)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.operations is not None or self.space is not None or self.t is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, None), state
 *     else:
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UpsertRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_UpsertRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_209859363);
    __Pyx_GIVEREF(__pyx_int_209859363);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_209859363)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.UpsertRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UpsertRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13UpsertRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13UpsertRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13UpsertRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13UpsertRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.UpsertRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13UpsertRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13UpsertRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UpsertRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_UpsertRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UpsertRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.UpsertRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/prepare.pyx":6
 * @cython.final
 * cdef class PrepareRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_14PrepareRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  char *__pyx_v_begin;
  char *__pyx_v_p;
  uint32_t __pyx_v_body_map_sz;
  uint32_t __pyx_v_max_body_len;
  PyObject *__pyx_v_query_temp = 0;
  char *__pyx_v_query_str;
  Py_ssize_t __pyx_v_query_len;
  uint32_t __pyx_v_kind;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_body", 1);

  /* "asynctnt/iproto/requests/prepare.pyx":18
 *             uint32_t kind
 * 
 *         body_map_sz = 1             # <<<<<<<<<<<<<<
 *         max_body_len = 0
 * 
 */
  __pyx_v_body_map_sz = 1;

  /* "asynctnt/iproto/requests/prepare.pyx":19
 * 
 *         body_map_sz = 1
 *         max_body_len = 0             # <<<<<<<<<<<<<<
 * 
 *         query_str = NULL
 */
  __pyx_v_max_body_len = 0;

  /* "asynctnt/iproto/requests/prepare.pyx":21
 *         max_body_len = 0
 * 
 *         query_str = NULL             # <<<<<<<<<<<<<<
 *         query_len = 0
 * 
 */
  __pyx_v_query_str = NULL;

  /* "asynctnt/iproto/requests/prepare.pyx":22
 * 
 *         query_str = NULL
 *         query_len = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.query is not None:
 */
  __pyx_v_query_len = 0;

  /* "asynctnt/iproto/requests/prepare.pyx":24
 *         query_len = 0
 * 
 *         if self.query is not None:             # <<<<<<<<<<<<<<
 *             query_temp = encode_unicode_string(self.query, buffer._encoding)
 *             cpython.bytes.PyBytes_AsStringAndSize(query_temp,
 */
  __pyx_t_1 = (__pyx_v_self->query != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "asynctnt/iproto/requests/prepare.pyx":25
 * 
 *         if self.query is not None:
 *             query_temp = encode_unicode_string(self.query, buffer._encoding)             # <<<<<<<<<<<<<<
 *             cpython.bytes.PyBytes_AsStringAndSize(query_temp,
 *                                                   &query_str,
 */
    __pyx_t_2 = __pyx_v_self->query;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __pyx_v_buffer->_encoding;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string(__pyx_t_2, ((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 25, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_query_temp = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "asynctnt/iproto/requests/prepare.pyx":26
 *         if self.query is not None:
 *             query_temp = encode_unicode_string(self.query, buffer._encoding)
 *             cpython.bytes.PyBytes_AsStringAndSize(query_temp,             # <<<<<<<<<<<<<<
 *                                                   &query_str,
 *                                                   &query_len)
 */
    __pyx_t_5 = PyBytes_AsStringAndSize(__pyx_v_query_temp, (&__pyx_v_query_str), ((Py_ssize_t *)(&__pyx_v_query_len))); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(20, 26, __pyx_L1_error)

    /* "asynctnt/iproto/requests/prepare.pyx":37
 *                            + 1 \
 *                            + mp_sizeof_str(<uint32_t> query_len) \
 *                            + 1             # <<<<<<<<<<<<<<
 *             kind = tarantool.IPROTO_SQL_TEXT
 *         else:
 */
    __pyx_v_max_body_len = ((0x2 + mp_sizeof_str(((uint32_t)__pyx_v_query_len))) + 1);

    /* "asynctnt/iproto/requests/prepare.pyx":38
 *                            + mp_sizeof_str(<uint32_t> query_len) \
 *                            + 1
 *             kind = tarantool.IPROTO_SQL_TEXT             # <<<<<<<<<<<<<<
 *         else:
 *             # Size description:
 */
    __pyx_v_kind = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SQL_TEXT;

    /* "asynctnt/iproto/requests/prepare.pyx":24
 *         query_len = 0
 * 
 *         if self.query is not None:             # <<<<<<<<<<<<<<
 *             query_temp = encode_unicode_string(self.query, buffer._encoding)
 *             cpython.bytes.PyBytes_AsStringAndSize(query_temp,
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/requests/prepare.pyx":45
 *             # + mp_sizeof_int(self.statement_id)
 *             # + mp_sizeof_uint(TP_SQL_BIND)
 *             max_body_len = 1 \             # <<<<<<<<<<<<<<
 *                            + 1 \
 *                            + 9 \
 */
  /*else*/ {

    /* "asynctnt/iproto/requests/prepare.pyx":48
 *                            + 1 \
 *                            + 9 \
 *                            + 1             # <<<<<<<<<<<<<<
 *             kind = tarantool.IPROTO_STMT_ID
 * 
 */
    __pyx_v_max_body_len = 0xc;

    /* "asynctnt/iproto/requests/prepare.pyx":49
 *                            + 9 \
 *                            + 1
 *             kind = tarantool.IPROTO_STMT_ID             # <<<<<<<<<<<<<<
 * 
 *         buffer.ensure_allocated(max_body_len)
 */
    __pyx_v_kind = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_STMT_ID;
  }
  __pyx_L3:;

  /* "asynctnt/iproto/requests/prepare.pyx":51
 *             kind = tarantool.IPROTO_STMT_ID
 * 
 *         buffer.ensure_allocated(max_body_len)             # <<<<<<<<<<<<<<
 * 
 *         p = begin = &buffer._buf[buffer._length]
 */
  __pyx_t_5 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_buffer, __pyx_v_max_body_len); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(20, 51, __pyx_L1_error)

  /* "asynctnt/iproto/requests/prepare.pyx":53
 *         buffer.ensure_allocated(max_body_len)
 * 
 *         p = begin = &buffer._buf[buffer._length]             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, kind)
 */
  __pyx_t_6 = (&(__pyx_v_buffer->_buf[__pyx_v_buffer->_length]));
  __pyx_v_p = __pyx_t_6;
  __pyx_v_begin = __pyx_t_6;

  /* "asynctnt/iproto/requests/prepare.pyx":54
 * 
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, kind)
 *         if query_str != NULL:
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_body_map_sz);

  /* "asynctnt/iproto/requests/prepare.pyx":55
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, kind)             # <<<<<<<<<<<<<<
 *         if query_str != NULL:
 *             p = mp_encode_str(p, query_str, <uint32_t> query_len)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_kind);

  /* "asynctnt/iproto/requests/prepare.pyx":56
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, kind)
 *         if query_str != NULL:             # <<<<<<<<<<<<<<
 *             p = mp_encode_str(p, query_str, <uint32_t> query_len)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_query_str != NULL);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/requests/prepare.pyx":57
 *         p = mp_encode_uint(p, kind)
 *         if query_str != NULL:
 *             p = mp_encode_str(p, query_str, <uint32_t> query_len)             # <<<<<<<<<<<<<<
 *         else:
 *             p = mp_encode_uint(p, self.statement_id)
 */
    __pyx_v_p = mp_encode_str(__pyx_v_p, __pyx_v_query_str, ((uint32_t)__pyx_v_query_len));

    /* "asynctnt/iproto/requests/prepare.pyx":56
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, kind)
 *         if query_str != NULL:             # <<<<<<<<<<<<<<
 *             p = mp_encode_str(p, query_str, <uint32_t> query_len)
 *         else:
 */
    goto __pyx_L4;
  }

  /* "asynctnt/iproto/requests/prepare.pyx":59
 *             p = mp_encode_str(p, query_str, <uint32_t> query_len)
 *         else:
 *             p = mp_encode_uint(p, self.statement_id)             # <<<<<<<<<<<<<<
 * 
 *         buffer._length += (p - begin)
 */
  /*else*/ {
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_self->statement_id);
  }
  __pyx_L4:;

  /* "asynctnt/iproto/requests/prepare.pyx":61
 *             p = mp_encode_uint(p, self.statement_id)
 * 
 *         buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 */
  __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/requests/prepare.pyx":6
 * @cython.final
 * cdef class PrepareRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PrepareRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_query_temp);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_14PrepareRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_14PrepareRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_14PrepareRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_14PrepareRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_14PrepareRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_14PrepareRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.query, self.schema_id, self.space, self.statement_id, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->statement_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyTuple_New(13); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->query);
  __Pyx_GIVEREF(__pyx_v_self->query);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 5, __pyx_v_self->query)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 6, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 7, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 8, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 9, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 10, __pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 11, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 12, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.query, self.schema_id, self.space, self.statement_id, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_10 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_v__dict = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.query, self.schema_id, self.space, self.statement_id, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_11 = (__pyx_v__dict != Py_None);
  if (__pyx_t_11) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_9 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.query is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.query, self.schema_id, self.space, self.statement_id, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.query is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, None), state
 */
  /*else*/ {
    __pyx_t_12 = (__pyx_v_self->query != ((PyObject*)Py_None));
    if (!__pyx_t_12) {
    } else {
      __pyx_t_11 = __pyx_t_12;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_12 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_12) {
    } else {
      __pyx_t_11 = __pyx_t_12;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_12 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_12) {
    } else {
      __pyx_t_11 = __pyx_t_12;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_12 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_11 = __pyx_t_12;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_11;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.query is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.query is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_pyx_unpickle_PrepareRequest); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_218707107);
    __Pyx_GIVEREF(__pyx_int_218707107);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_218707107)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_10)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_9 = 0;
    __pyx_t_10 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.query is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, None), state
 *     else:
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PrepareRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_PrepareRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_218707107);
    __Pyx_GIVEREF(__pyx_int_218707107);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_218707107)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_10)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_10 = 0;
    __pyx_r = __pyx_t_9;
    __pyx_t_9 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PrepareRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PrepareRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_14PrepareRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_14PrepareRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_14PrepareRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_14PrepareRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PrepareRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_14PrepareRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_14PrepareRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PrepareRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PrepareRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PrepareRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PrepareRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/execute.pyx":6
 * @cython.final
 * cdef class ExecuteRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_14ExecuteRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  char *__pyx_v_begin;
  char *__pyx_v_p;
  uint32_t __pyx_v_body_map_sz;
  uint32_t __pyx_v_max_body_len;
  PyObject *__pyx_v_query_temp = 0;
  char *__pyx_v_query_str;
  Py_ssize_t __pyx_v_query_len;
  uint32_t __pyx_v_kind;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_body", 1);

  /* "asynctnt/iproto/requests/execute.pyx":18
 *             uint32_t kind
 * 
 *         body_map_sz = 2             # <<<<<<<<<<<<<<
 *         max_body_len = 0
 * 
 */
  __pyx_v_body_map_sz = 2;

  /* "asynctnt/iproto/requests/execute.pyx":19
 * 
 *         body_map_sz = 2
 *         max_body_len = 0             # <<<<<<<<<<<<<<
 * 
 *         query_str = NULL
 */
  __pyx_v_max_body_len = 0;

  /* "asynctnt/iproto/requests/execute.pyx":21
 *         max_body_len = 0
 * 
 *         query_str = NULL             # <<<<<<<<<<<<<<
 *         query_len = 0
 * 
 */
  __pyx_v_query_str = NULL;

  /* "asynctnt/iproto/requests/execute.pyx":22
 * 
 *         query_str = NULL
 *         query_len = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.query is not None:
 */
  __pyx_v_query_len = 0;

  /* "asynctnt/iproto/requests/execute.pyx":24
 *         query_len = 0
 * 
 *         if self.query is not None:             # <<<<<<<<<<<<<<
 *             query_temp = encode_unicode_string(self.query, buffer._encoding)
 *             cpython.bytes.PyBytes_AsStringAndSize(query_temp,
 */
  __pyx_t_1 = (__pyx_v_self->query != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "asynctnt/iproto/requests/execute.pyx":25
 * 
 *         if self.query is not None:
 *             query_temp = encode_unicode_string(self.query, buffer._encoding)             # <<<<<<<<<<<<<<
 *             cpython.bytes.PyBytes_AsStringAndSize(query_temp,
 *                                                   &query_str,
 */
    __pyx_t_2 = __pyx_v_self->query;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __pyx_v_buffer->_encoding;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string(__pyx_t_2, ((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(21, 25, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_query_temp = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "asynctnt/iproto/requests/execute.pyx":26
 *         if self.query is not None:
 *             query_temp = encode_unicode_string(self.query, buffer._encoding)
 *             cpython.bytes.PyBytes_AsStringAndSize(query_temp,             # <<<<<<<<<<<<<<
 *                                                   &query_str,
 *                                                   &query_len)
 */
    __pyx_t_5 = PyBytes_AsStringAndSize(__pyx_v_query_temp, (&__pyx_v_query_str), ((Py_ssize_t *)(&__pyx_v_query_len))); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(21, 26, __pyx_L1_error)

    /* "asynctnt/iproto/requests/execute.pyx":37
 *                            + 1 \
 *                            + mp_sizeof_str(<uint32_t> query_len) \
 *                            + 1             # <<<<<<<<<<<<<<
 *             kind = tarantool.IPROTO_SQL_TEXT
 *         else:
 */
    __pyx_v_max_body_len = ((0x2 + mp_sizeof_str(((uint32_t)__pyx_v_query_len))) + 1);

    /* "asynctnt/iproto/requests/execute.pyx":38
 *                            + mp_sizeof_str(<uint32_t> query_len) \
 *                            + 1
 *             kind = tarantool.IPROTO_SQL_TEXT             # <<<<<<<<<<<<<<
 *         else:
 *             # Size description:
 */
    __pyx_v_kind = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SQL_TEXT;

    /* "asynctnt/iproto/requests/execute.pyx":24
 *         query_len = 0
 * 
 *         if self.query is not None:             # <<<<<<<<<<<<<<
 *             query_temp = encode_unicode_string(self.query, buffer._encoding)
 *             cpython.bytes.PyBytes_AsStringAndSize(query_temp,
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/requests/execute.pyx":45
 *             # + mp_sizeof_int(self.statement_id)
 *             # + mp_sizeof_uint(TP_SQL_BIND)
 *             max_body_len = 1 \             # <<<<<<<<<<<<<<
 *                            + 1 \
 *                            + 9 \
 */
  /*else*/ {

    /* "asynctnt/iproto/requests/execute.pyx":48
 *                            + 1 \
 *                            + 9 \
 *                            + 1             # <<<<<<<<<<<<<<
 *             kind = tarantool.IPROTO_STMT_ID
 * 
 */
    __pyx_v_max_body_len = 0xc;

    /* "asynctnt/iproto/requests/execute.pyx":49
 *                            + 9 \
 *                            + 1
 *             kind = tarantool.IPROTO_STMT_ID             # <<<<<<<<<<<<<<
 * 
 *         buffer.ensure_allocated(max_body_len)
 */
    __pyx_v_kind = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_STMT_ID;
  }
  __pyx_L3:;

  /* "asynctnt/iproto/requests/execute.pyx":51
 *             kind = tarantool.IPROTO_STMT_ID
 * 
 *         buffer.ensure_allocated(max_body_len)             # <<<<<<<<<<<<<<
 * 
 *         p = begin = &buffer._buf[buffer._length]
 */
  __pyx_t_5 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_buffer, __pyx_v_max_body_len); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(21, 51, __pyx_L1_error)

  /* "asynctnt/iproto/requests/execute.pyx":53
 *         buffer.ensure_allocated(max_body_len)
 * 
 *         p = begin = &buffer._buf[buffer._length]             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, kind)
 */
  __pyx_t_6 = (&(__pyx_v_buffer->_buf[__pyx_v_buffer->_length]));
  __pyx_v_p = __pyx_t_6;
  __pyx_v_begin = __pyx_t_6;

  /* "asynctnt/iproto/requests/execute.pyx":54
 * 
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, kind)
 *         if query_str != NULL:
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_body_map_sz);

  /* "asynctnt/iproto/requests/execute.pyx":55
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, kind)             # <<<<<<<<<<<<<<
 *         if query_str != NULL:
 *             p = mp_encode_str(p, query_str, <uint32_t> query_len)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_kind);

  /* "asynctnt/iproto/requests/execute.pyx":56
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, kind)
 *         if query_str != NULL:             # <<<<<<<<<<<<<<
 *             p = mp_encode_str(p, query_str, <uint32_t> query_len)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_query_str != NULL);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/requests/execute.pyx":57
 *         p = mp_encode_uint(p, kind)
 *         if query_str != NULL:
 *             p = mp_encode_str(p, query_str, <uint32_t> query_len)             # <<<<<<<<<<<<<<
 *         else:
 *             p = mp_encode_uint(p, self.statement_id)
 */
    __pyx_v_p = mp_encode_str(__pyx_v_p, __pyx_v_query_str, ((uint32_t)__pyx_v_query_len));

    /* "asynctnt/iproto/requests/execute.pyx":56
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, kind)
 *         if query_str != NULL:             # <<<<<<<<<<<<<<
 *             p = mp_encode_str(p, query_str, <uint32_t> query_len)
 *         else:
 */
    goto __pyx_L4;
  }

  /* "asynctnt/iproto/requests/execute.pyx":59
 *             p = mp_encode_str(p, query_str, <uint32_t> query_len)
 *         else:
 *             p = mp_encode_uint(p, self.statement_id)             # <<<<<<<<<<<<<<
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_SQL_BIND)
 */
  /*else*/ {
    __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_self->statement_id);
  }
  __pyx_L4:;

  /* "asynctnt/iproto/requests/execute.pyx":61
 *             p = mp_encode_uint(p, self.statement_id)
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_SQL_BIND)             # <<<<<<<<<<<<<<
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.args, None, False)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SQL_BIND);

  /* "asynctnt/iproto/requests/execute.pyx":62
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_SQL_BIND)
 *         buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 *         p = encode_key_sequence(buffer, p, self.args, None, False)
 */
  __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/requests/execute.pyx":63
 *         p = mp_encode_uint(p, tarantool.IPROTO_SQL_BIND)
 *         buffer._length += (p - begin)
 *         p = encode_key_sequence(buffer, p, self.args, None, False)             # <<<<<<<<<<<<<<
 */
  __pyx_t_4 = __pyx_v_self->args;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_6 = __pyx_f_8asynctnt_6iproto_8protocol_encode_key_sequence(__pyx_v_buffer, __pyx_v_p, __pyx_t_4, ((struct C_Metadata *)Py_None), 0); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(21, 63, __pyx_L1_error)
  __pyx_v_p = __pyx_t_6;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "asynctnt/iproto/requests/execute.pyx":6
 * @cython.final
 * cdef class ExecuteRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ExecuteRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_query_temp);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_14ExecuteRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_14ExecuteRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_14ExecuteRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_14ExecuteRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_14ExecuteRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_14ExecuteRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.args, self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.query, self.schema_id, self.space, self.statement_id, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->statement_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyTuple_New(14); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_v_self->args);
  __Pyx_GIVEREF(__pyx_v_self->args);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_self->args)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 5, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->query);
  __Pyx_GIVEREF(__pyx_v_self->query);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 6, __pyx_v_self->query)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 7, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 8, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 9, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 10, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 11, __pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 12, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 13, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.args, self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.query, self.schema_id, self.space, self.statement_id, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_10 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_v__dict = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "(tree fragment)":7
 *     state = (self.args, self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.query, self.schema_id, self.space, self.statement_id, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_11 = (__pyx_v__dict != Py_None);
  if (__pyx_t_11) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_9 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.args is not None or self.query is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.args, self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.query, self.schema_id, self.space, self.statement_id, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.args is not None or self.query is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, None), state
 */
  /*else*/ {
    __pyx_t_12 = (__pyx_v_self->args != Py_None);
    if (!__pyx_t_12) {
    } else {
      __pyx_t_11 = __pyx_t_12;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_12 = (__pyx_v_self->query != ((PyObject*)Py_None));
    if (!__pyx_t_12) {
    } else {
      __pyx_t_11 = __pyx_t_12;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_12 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_12) {
    } else {
      __pyx_t_11 = __pyx_t_12;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_12 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_12) {
    } else {
      __pyx_t_11 = __pyx_t_12;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_12 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_11 = __pyx_t_12;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_11;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.args is not None or self.query is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.args is not None or self.query is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_pyx_unpickle_ExecuteRequest); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_51694216);
    __Pyx_GIVEREF(__pyx_int_51694216);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_51694216)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_10)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_9 = 0;
    __pyx_t_10 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.args is not None or self.query is not None or self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, None), state
 *     else:
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ExecuteRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_ExecuteRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_51694216);
    __Pyx_GIVEREF(__pyx_int_51694216);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_51694216)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_10)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_10 = 0;
    __pyx_r = __pyx_t_9;
    __pyx_t_9 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ExecuteRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ExecuteRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_14ExecuteRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_14ExecuteRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_14ExecuteRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_14ExecuteRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ExecuteRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_14ExecuteRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_14ExecuteRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ExecuteRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_ExecuteRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ExecuteRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.ExecuteRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/id.pyx":7
 * @cython.final
 * cdef class IDRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *p
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_9IDRequest_encode_body(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  char *__pyx_v_p;
  char *__pyx_v_begin;
  uint32_t __pyx_v_body_map_sz;
  uint32_t __pyx_v_max_body_len;
  int __pyx_r;
  int __pyx_t_1;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/requests/id.pyx":14
 *             uint32_t max_body_len
 * 
 *         body_map_sz = 2             # <<<<<<<<<<<<<<
 *         # Size description:
 *         # mp_sizeof_map(body_map_sz)
 */
  __pyx_v_body_map_sz = 2;

  /* "asynctnt/iproto/requests/id.pyx":27
 *                        + 1 \
 *                        + 1 \
 *                        + 4  # Note: maximum 4 elements in array             # <<<<<<<<<<<<<<
 * 
 *         buffer.ensure_allocated(max_body_len)
 */
  __pyx_v_max_body_len = 0x9;

  /* "asynctnt/iproto/requests/id.pyx":29
 *                        + 4  # Note: maximum 4 elements in array
 * 
 *         buffer.ensure_allocated(max_body_len)             # <<<<<<<<<<<<<<
 * 
 *         p = begin = &buffer._buf[buffer._length]
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_buffer, __pyx_v_max_body_len); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(22, 29, __pyx_L1_error)

  /* "asynctnt/iproto/requests/id.pyx":31
 *         buffer.ensure_allocated(max_body_len)
 * 
 *         p = begin = &buffer._buf[buffer._length]             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_VERSION)
 */
  __pyx_t_2 = (&(__pyx_v_buffer->_buf[__pyx_v_buffer->_length]));
  __pyx_v_p = __pyx_t_2;
  __pyx_v_begin = __pyx_t_2;

  /* "asynctnt/iproto/requests/id.pyx":32
 * 
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_VERSION)
 *         p = mp_encode_uint(p, IPROTO_VERSION)
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_body_map_sz);

  /* "asynctnt/iproto/requests/id.pyx":33
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_VERSION)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, IPROTO_VERSION)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURES)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_VERSION);

  /* "asynctnt/iproto/requests/id.pyx":34
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_VERSION)
 *         p = mp_encode_uint(p, IPROTO_VERSION)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURES)
 *         p = mp_encode_array(p, 3)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, 3);

  /* "asynctnt/iproto/requests/id.pyx":35
 *         p = mp_encode_uint(p, tarantool.IPROTO_VERSION)
 *         p = mp_encode_uint(p, IPROTO_VERSION)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURES)             # <<<<<<<<<<<<<<
 *         p = mp_encode_array(p, 3)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_STREAMS)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FEATURES);

  /* "asynctnt/iproto/requests/id.pyx":36
 *         p = mp_encode_uint(p, IPROTO_VERSION)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURES)
 *         p = mp_encode_array(p, 3)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_STREAMS)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_TRANSACTIONS)
 */
  __pyx_v_p = mp_encode_array(__pyx_v_p, 3);

  /* "asynctnt/iproto/requests/id.pyx":37
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURES)
 *         p = mp_encode_array(p, 3)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_STREAMS)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_TRANSACTIONS)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_ERROR_EXTENSION)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FEATURE_STREAMS);

  /* "asynctnt/iproto/requests/id.pyx":38
 *         p = mp_encode_array(p, 3)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_STREAMS)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_TRANSACTIONS)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_ERROR_EXTENSION)
 * 
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FEATURE_TRANSACTIONS);

  /* "asynctnt/iproto/requests/id.pyx":39
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_STREAMS)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_TRANSACTIONS)
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_ERROR_EXTENSION)             # <<<<<<<<<<<<<<
 * 
 *         buffer._length += (p - begin)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FEATURE_ERROR_EXTENSION);

  /* "asynctnt/iproto/requests/id.pyx":41
 *         p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_ERROR_EXTENSION)
 * 
 *         buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 */
  __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/requests/id.pyx":7
 * @cython.final
 * cdef class IDRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *p
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IDRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_9IDRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_9IDRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_9IDRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_9IDRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_9IDRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_9IDRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, None), state
 */
  /*else*/ {
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_IDRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150999062);
    __Pyx_GIVEREF(__pyx_int_150999062);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_150999062)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, None), state
 *     else:
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IDRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_IDRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150999062);
    __Pyx_GIVEREF(__pyx_int_150999062);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_150999062)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IDRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IDRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_9IDRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_9IDRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_9IDRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_9IDRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IDRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_9IDRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_9IDRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IDRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IDRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IDRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IDRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/auth.pyx":8
 * 
 * 
 * cdef inline bytes _sha1(tuple values):             # <<<<<<<<<<<<<<
 *     cdef object sha = hashlib.sha1()
 *     for i in values:
 */

static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol__sha1(PyObject *__pyx_v_values) {
  PyObject *__pyx_v_sha = 0;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sha1", 1);

  /* "asynctnt/iproto/requests/auth.pyx":9
 * 
 * cdef inline bytes _sha1(tuple values):
 *     cdef object sha = hashlib.sha1()             # <<<<<<<<<<<<<<
 *     for i in values:
 *         if i is not None:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_hashlib); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sha1); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_sha = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/requests/auth.pyx":10
 * cdef inline bytes _sha1(tuple values):
 *     cdef object sha = hashlib.sha1()
 *     for i in values:             # <<<<<<<<<<<<<<
 *         if i is not None:
 *             sha.update(i)
 */
  if (unlikely(__pyx_v_values == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(23, 10, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_values; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_5 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(23, 10, __pyx_L1_error)
      #endif
      if (__pyx_t_5 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(23, 10, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "asynctnt/iproto/requests/auth.pyx":11
 *     cdef object sha = hashlib.sha1()
 *     for i in values:
 *         if i is not None:             # <<<<<<<<<<<<<<
 *             sha.update(i)
 *     return sha.digest()
 */
    __pyx_t_6 = (__pyx_v_i != Py_None);
    if (__pyx_t_6) {

      /* "asynctnt/iproto/requests/auth.pyx":12
 *     for i in values:
 *         if i is not None:
 *             sha.update(i)             # <<<<<<<<<<<<<<
 *     return sha.digest()
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sha, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 12, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_i};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 12, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "asynctnt/iproto/requests/auth.pyx":11
 *     cdef object sha = hashlib.sha1()
 *     for i in values:
 *         if i is not None:             # <<<<<<<<<<<<<<
 *             sha.update(i)
 *     return sha.digest()
 */
    }

    /* "asynctnt/iproto/requests/auth.pyx":10
 * cdef inline bytes _sha1(tuple values):
 *     cdef object sha = hashlib.sha1()
 *     for i in values:             # <<<<<<<<<<<<<<
 *         if i is not None:
 *             sha.update(i)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/requests/auth.pyx":13
 *         if i is not None:
 *             sha.update(i)
 *     return sha.digest()             # <<<<<<<<<<<<<<
 * 
 * cdef inline bytes _strxor(bytes hash1, bytes scramble):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sha, __pyx_n_s_digest); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(23, 13, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/requests/auth.pyx":8
 * 
 * 
 * cdef inline bytes _sha1(tuple values):             # <<<<<<<<<<<<<<
 *     cdef object sha = hashlib.sha1()
 *     for i in values:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynctnt.iproto.protocol._sha1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sha);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/auth.pyx":15
 *     return sha.digest()
 * 
 * cdef inline bytes _strxor(bytes hash1, bytes scramble):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char *hash1_str
 */

static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol__strxor(PyObject *__pyx_v_hash1, PyObject *__pyx_v_scramble) {
  char *__pyx_v_hash1_str;
  Py_ssize_t __pyx_v_hash1_len;
  char *__pyx_v_scramble_str;
  Py_ssize_t __pyx_v_scramble_len;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_strxor", 1);

  /* "asynctnt/iproto/requests/auth.pyx":22
 *         char *scramble_str
 *         ssize_t scramble_len
 *     cpython.bytes.PyBytes_AsStringAndSize(hash1,             # <<<<<<<<<<<<<<
 *                                           &hash1_str, &hash1_len)
 *     cpython.bytes.PyBytes_AsStringAndSize(scramble,
 */
  __pyx_t_1 = PyBytes_AsStringAndSize(__pyx_v_hash1, (&__pyx_v_hash1_str), ((Py_ssize_t *)(&__pyx_v_hash1_len))); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(23, 22, __pyx_L1_error)

  /* "asynctnt/iproto/requests/auth.pyx":24
 *     cpython.bytes.PyBytes_AsStringAndSize(hash1,
 *                                           &hash1_str, &hash1_len)
 *     cpython.bytes.PyBytes_AsStringAndSize(scramble,             # <<<<<<<<<<<<<<
 *                                           &scramble_str, &scramble_len)
 *     for i in range(scramble_len):
 */
  __pyx_t_1 = PyBytes_AsStringAndSize(__pyx_v_scramble, (&__pyx_v_scramble_str), ((Py_ssize_t *)(&__pyx_v_scramble_len))); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(23, 24, __pyx_L1_error)

  /* "asynctnt/iproto/requests/auth.pyx":26
 *     cpython.bytes.PyBytes_AsStringAndSize(scramble,
 *                                           &scramble_str, &scramble_len)
 *     for i in range(scramble_len):             # <<<<<<<<<<<<<<
 *         scramble_str[i] = hash1_str[i] ^ scramble_str[i]
 *     return scramble
 */
  __pyx_t_2 = __pyx_v_scramble_len;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "asynctnt/iproto/requests/auth.pyx":27
 *                                           &scramble_str, &scramble_len)
 *     for i in range(scramble_len):
 *         scramble_str[i] = hash1_str[i] ^ scramble_str[i]             # <<<<<<<<<<<<<<
 *     return scramble
 * 
 */
    (__pyx_v_scramble_str[__pyx_v_i]) = ((__pyx_v_hash1_str[__pyx_v_i]) ^ (__pyx_v_scramble_str[__pyx_v_i]));
  }

  /* "asynctnt/iproto/requests/auth.pyx":28
 *     for i in range(scramble_len):
 *         scramble_str[i] = hash1_str[i] ^ scramble_str[i]
 *     return scramble             # <<<<<<<<<<<<<<
 * 
 * @cython.final
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_scramble);
  __pyx_r = __pyx_v_scramble;
  goto __pyx_L0;

  /* "asynctnt/iproto/requests/auth.pyx":15
 *     return sha.digest()
 * 
 * cdef inline bytes _strxor(bytes hash1, bytes scramble):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char *hash1_str
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol._strxor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/auth.pyx":32
 * @cython.final
 * cdef class AuthRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_11AuthRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  char *__pyx_v_begin;
  char *__pyx_v_p;
  uint32_t __pyx_v_body_map_sz;
  uint32_t __pyx_v_max_body_len;
  char *__pyx_v_username_str;
  Py_ssize_t __pyx_v_username_len;
  char *__pyx_v_scramble_str;
  Py_ssize_t __pyx_v_scramble_len;
  PyObject *__pyx_v_username_bytes = NULL;
  PyObject *__pyx_v_password_bytes = NULL;
  PyObject *__pyx_v_hash1 = NULL;
  PyObject *__pyx_v_hash2 = NULL;
  PyObject *__pyx_v_scramble = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  char *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_body", 1);

  /* "asynctnt/iproto/requests/auth.pyx":45
 *             ssize_t scramble_len
 * 
 *         username_bytes = encode_unicode_string(self.username, buffer._encoding)             # <<<<<<<<<<<<<<
 *         password_bytes = encode_unicode_string(self.password, buffer._encoding)
 * 
 */
  __pyx_t_1 = __pyx_v_self->username;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_buffer->_encoding;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string(__pyx_t_1, ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_username_bytes = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/requests/auth.pyx":46
 * 
 *         username_bytes = encode_unicode_string(self.username, buffer._encoding)
 *         password_bytes = encode_unicode_string(self.password, buffer._encoding)             # <<<<<<<<<<<<<<
 * 
 *         hash1 = _sha1((password_bytes,))
 */
  __pyx_t_3 = __pyx_v_self->password;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_2 = __pyx_v_buffer->_encoding;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string(__pyx_t_3, ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_password_bytes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/requests/auth.pyx":48
 *         password_bytes = encode_unicode_string(self.password, buffer._encoding)
 * 
 *         hash1 = _sha1((password_bytes,))             # <<<<<<<<<<<<<<
 *         hash2 = _sha1((hash1,))
 *         scramble = _sha1((self.salt, hash2))
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_password_bytes);
  __Pyx_GIVEREF(__pyx_v_password_bytes);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_password_bytes)) __PYX_ERR(23, 48, __pyx_L1_error);
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol__sha1(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_hash1 = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/requests/auth.pyx":49
 * 
 *         hash1 = _sha1((password_bytes,))
 *         hash2 = _sha1((hash1,))             # <<<<<<<<<<<<<<
 *         scramble = _sha1((self.salt, hash2))
 *         scramble = _strxor(hash1, scramble)
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_hash1);
  __Pyx_GIVEREF(__pyx_v_hash1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_hash1)) __PYX_ERR(23, 49, __pyx_L1_error);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol__sha1(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_hash2 = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/requests/auth.pyx":50
 *         hash1 = _sha1((password_bytes,))
 *         hash2 = _sha1((hash1,))
 *         scramble = _sha1((self.salt, hash2))             # <<<<<<<<<<<<<<
 *         scramble = _strxor(hash1, scramble)
 * 
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->salt);
  __Pyx_GIVEREF(__pyx_v_self->salt);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->salt)) __PYX_ERR(23, 50, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_hash2);
  __Pyx_GIVEREF(__pyx_v_hash2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_hash2)) __PYX_ERR(23, 50, __pyx_L1_error);
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol__sha1(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_scramble = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/requests/auth.pyx":51
 *         hash2 = _sha1((hash1,))
 *         scramble = _sha1((self.salt, hash2))
 *         scramble = _strxor(hash1, scramble)             # <<<<<<<<<<<<<<
 * 
 *         cpython.bytes.PyBytes_AsStringAndSize(username_bytes,
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol__strxor(__pyx_v_hash1, __pyx_v_scramble); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_scramble, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/requests/auth.pyx":53
 *         scramble = _strxor(hash1, scramble)
 * 
 *         cpython.bytes.PyBytes_AsStringAndSize(username_bytes,             # <<<<<<<<<<<<<<
 *                                               &username_str, &username_len)
 *         cpython.bytes.PyBytes_AsStringAndSize(scramble,
 */
  __pyx_t_4 = PyBytes_AsStringAndSize(__pyx_v_username_bytes, (&__pyx_v_username_str), ((Py_ssize_t *)(&__pyx_v_username_len))); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(23, 53, __pyx_L1_error)

  /* "asynctnt/iproto/requests/auth.pyx":55
 *         cpython.bytes.PyBytes_AsStringAndSize(username_bytes,
 *                                               &username_str, &username_len)
 *         cpython.bytes.PyBytes_AsStringAndSize(scramble,             # <<<<<<<<<<<<<<
 *                                               &scramble_str, &scramble_len)
 *         body_map_sz = 2
 */
  __pyx_t_4 = PyBytes_AsStringAndSize(__pyx_v_scramble, (&__pyx_v_scramble_str), ((Py_ssize_t *)(&__pyx_v_scramble_len))); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(23, 55, __pyx_L1_error)

  /* "asynctnt/iproto/requests/auth.pyx":57
 *         cpython.bytes.PyBytes_AsStringAndSize(scramble,
 *                                               &scramble_str, &scramble_len)
 *         body_map_sz = 2             # <<<<<<<<<<<<<<
 *         # Size description:
 *         # mp_sizeof_map()
 */
  __pyx_v_body_map_sz = 2;

  /* "asynctnt/iproto/requests/auth.pyx":72
 *                        + 1 \
 *                        + 1 + 9 \
 *                        + mp_sizeof_str(<uint32_t> scramble_len)             # <<<<<<<<<<<<<<
 * 
 *         buffer.ensure_allocated(max_body_len)
 */
  __pyx_v_max_body_len = ((((((0x2 + mp_sizeof_str(((uint32_t)__pyx_v_username_len))) + 1) + 1) + 1) + 9) + mp_sizeof_str(((uint32_t)__pyx_v_scramble_len)));

  /* "asynctnt/iproto/requests/auth.pyx":74
 *                        + mp_sizeof_str(<uint32_t> scramble_len)
 * 
 *         buffer.ensure_allocated(max_body_len)             # <<<<<<<<<<<<<<
 * 
 *         p = begin = &buffer._buf[buffer._length]
 */
  __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_buffer, __pyx_v_max_body_len); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(23, 74, __pyx_L1_error)

  /* "asynctnt/iproto/requests/auth.pyx":76
 *         buffer.ensure_allocated(max_body_len)
 * 
 *         p = begin = &buffer._buf[buffer._length]             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_USER_NAME)
 */
  __pyx_t_5 = (&(__pyx_v_buffer->_buf[__pyx_v_buffer->_length]));
  __pyx_v_p = __pyx_t_5;
  __pyx_v_begin = __pyx_t_5;

  /* "asynctnt/iproto/requests/auth.pyx":77
 * 
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_USER_NAME)
 *         p = mp_encode_str(p, username_str, <uint32_t> username_len)
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_body_map_sz);

  /* "asynctnt/iproto/requests/auth.pyx":78
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_USER_NAME)             # <<<<<<<<<<<<<<
 *         p = mp_encode_str(p, username_str, <uint32_t> username_len)
 * 
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_USER_NAME);

  /* "asynctnt/iproto/requests/auth.pyx":79
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_USER_NAME)
 *         p = mp_encode_str(p, username_str, <uint32_t> username_len)             # <<<<<<<<<<<<<<
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 */
  __pyx_v_p = mp_encode_str(__pyx_v_p, __pyx_v_username_str, ((uint32_t)__pyx_v_username_len));

  /* "asynctnt/iproto/requests/auth.pyx":81
 *         p = mp_encode_str(p, username_str, <uint32_t> username_len)
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)             # <<<<<<<<<<<<<<
 *         p = mp_encode_array(p, 2)
 *         p = mp_encode_str(p, "chap-sha1", 9)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TUPLE);

  /* "asynctnt/iproto/requests/auth.pyx":82
 * 
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 *         p = mp_encode_array(p, 2)             # <<<<<<<<<<<<<<
 *         p = mp_encode_str(p, "chap-sha1", 9)
 *         p = mp_encode_str(p, scramble_str, <uint32_t> scramble_len)
 */
  __pyx_v_p = mp_encode_array(__pyx_v_p, 2);

  /* "asynctnt/iproto/requests/auth.pyx":83
 *         p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
 *         p = mp_encode_array(p, 2)
 *         p = mp_encode_str(p, "chap-sha1", 9)             # <<<<<<<<<<<<<<
 *         p = mp_encode_str(p, scramble_str, <uint32_t> scramble_len)
 *         buffer._length += (p - begin)
 */
  __pyx_v_p = mp_encode_str(__pyx_v_p, ((char const *)"chap-sha1"), 9);

  /* "asynctnt/iproto/requests/auth.pyx":84
 *         p = mp_encode_array(p, 2)
 *         p = mp_encode_str(p, "chap-sha1", 9)
 *         p = mp_encode_str(p, scramble_str, <uint32_t> scramble_len)             # <<<<<<<<<<<<<<
 *         buffer._length += (p - begin)
 */
  __pyx_v_p = mp_encode_str(__pyx_v_p, __pyx_v_scramble_str, ((uint32_t)__pyx_v_scramble_len));

  /* "asynctnt/iproto/requests/auth.pyx":85
 *         p = mp_encode_str(p, "chap-sha1", 9)
 *         p = mp_encode_str(p, scramble_str, <uint32_t> scramble_len)
 *         buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 */
  __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/requests/auth.pyx":32
 * @cython.final
 * cdef class AuthRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *begin
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.AuthRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_username_bytes);
  __Pyx_XDECREF(__pyx_v_password_bytes);
  __Pyx_XDECREF(__pyx_v_hash1);
  __Pyx_XDECREF(__pyx_v_hash2);
  __Pyx_XDECREF(__pyx_v_scramble);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11AuthRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11AuthRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11AuthRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11AuthRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11AuthRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11AuthRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.password, self.push_subscribe, self.salt, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.username, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(14); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->password);
  __Pyx_GIVEREF(__pyx_v_self->password);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_v_self->password)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->salt);
  __Pyx_GIVEREF(__pyx_v_self->salt);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, __pyx_v_self->salt)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 11, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->username);
  __Pyx_GIVEREF(__pyx_v_self->username);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 12, __pyx_v_self->username)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 13, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.password, self.push_subscribe, self.salt, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.username, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.password, self.push_subscribe, self.salt, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.username, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.password is not None or self.salt is not None or self.space is not None or self.timeout_handle is not None or self.username is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.password, self.push_subscribe, self.salt, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.username, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.password is not None or self.salt is not None or self.space is not None or self.timeout_handle is not None or self.username is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, None), state
 */
  /*else*/ {
    __pyx_t_11 = (__pyx_v_self->password != ((PyObject*)Py_None));
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->salt != ((PyObject*)Py_None));
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->username != ((PyObject*)Py_None));
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.password is not None or self.salt is not None or self.space is not None or self.timeout_handle is not None or self.username is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.password is not None or self.salt is not None or self.space is not None or self.timeout_handle is not None or self.username is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_AuthRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_53862661);
    __Pyx_GIVEREF(__pyx_int_53862661);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_53862661)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.password is not None or self.salt is not None or self.space is not None or self.timeout_handle is not None or self.username is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, None), state
 *     else:
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_AuthRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_AuthRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_53862661);
    __Pyx_GIVEREF(__pyx_int_53862661);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_53862661)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.AuthRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AuthRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11AuthRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11AuthRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11AuthRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11AuthRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.AuthRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11AuthRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11AuthRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_AuthRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_AuthRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AuthRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.AuthRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/requests/streams.pyx":6
 * @cython.final
 * cdef class BeginRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *p
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_12BeginRequest_encode_body(struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_v_buffer) {
  char *__pyx_v_p;
  char *__pyx_v_begin;
  uint32_t __pyx_v_body_map_sz;
  uint32_t __pyx_v_max_body_len;
  int __pyx_r;
  int __pyx_t_1;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/requests/streams.pyx":13
 *             uint32_t max_body_len
 * 
 *         body_map_sz = 2             # <<<<<<<<<<<<<<
 *         # Size description:
 *         max_body_len = mp_sizeof_map(body_map_sz) \
 */
  __pyx_v_body_map_sz = 2;

  /* "asynctnt/iproto/requests/streams.pyx":19
 *                        + mp_sizeof_uint(self.isolation) \
 *                        + mp_sizeof_uint(tarantool.IPROTO_TIMEOUT) \
 *                        + mp_sizeof_double(self.tx_timeout)             # <<<<<<<<<<<<<<
 * 
 *         buffer.ensure_allocated(max_body_len)
 */
  __pyx_v_max_body_len = ((((mp_sizeof_map(__pyx_v_body_map_sz) + mp_sizeof_uint(__pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TXN_ISOLATION)) + mp_sizeof_uint(__pyx_v_self->isolation)) + mp_sizeof_uint(__pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TIMEOUT)) + mp_sizeof_double(__pyx_v_self->tx_timeout));

  /* "asynctnt/iproto/requests/streams.pyx":21
 *                        + mp_sizeof_double(self.tx_timeout)
 * 
 *         buffer.ensure_allocated(max_body_len)             # <<<<<<<<<<<<<<
 * 
 *         p = begin = &buffer._buf[buffer._length]
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated(__pyx_v_buffer, __pyx_v_max_body_len); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(24, 21, __pyx_L1_error)

  /* "asynctnt/iproto/requests/streams.pyx":23
 *         buffer.ensure_allocated(max_body_len)
 * 
 *         p = begin = &buffer._buf[buffer._length]             # <<<<<<<<<<<<<<
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_TXN_ISOLATION)
 */
  __pyx_t_2 = (&(__pyx_v_buffer->_buf[__pyx_v_buffer->_length]));
  __pyx_v_p = __pyx_t_2;
  __pyx_v_begin = __pyx_t_2;

  /* "asynctnt/iproto/requests/streams.pyx":24
 * 
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_TXN_ISOLATION)
 *         p = mp_encode_uint(p, self.isolation)
 */
  __pyx_v_p = mp_encode_map(__pyx_v_p, __pyx_v_body_map_sz);

  /* "asynctnt/iproto/requests/streams.pyx":25
 *         p = begin = &buffer._buf[buffer._length]
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_TXN_ISOLATION)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, self.isolation)
 *         p = mp_encode_uint(p, tarantool.IPROTO_TIMEOUT)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TXN_ISOLATION);

  /* "asynctnt/iproto/requests/streams.pyx":26
 *         p = mp_encode_map(p, body_map_sz)
 *         p = mp_encode_uint(p, tarantool.IPROTO_TXN_ISOLATION)
 *         p = mp_encode_uint(p, self.isolation)             # <<<<<<<<<<<<<<
 *         p = mp_encode_uint(p, tarantool.IPROTO_TIMEOUT)
 *         p = mp_encode_double(p, self.tx_timeout)
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_v_self->isolation);

  /* "asynctnt/iproto/requests/streams.pyx":27
 *         p = mp_encode_uint(p, tarantool.IPROTO_TXN_ISOLATION)
 *         p = mp_encode_uint(p, self.isolation)
 *         p = mp_encode_uint(p, tarantool.IPROTO_TIMEOUT)             # <<<<<<<<<<<<<<
 *         p = mp_encode_double(p, self.tx_timeout)
 * 
 */
  __pyx_v_p = mp_encode_uint(__pyx_v_p, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_TIMEOUT);

  /* "asynctnt/iproto/requests/streams.pyx":28
 *         p = mp_encode_uint(p, self.isolation)
 *         p = mp_encode_uint(p, tarantool.IPROTO_TIMEOUT)
 *         p = mp_encode_double(p, self.tx_timeout)             # <<<<<<<<<<<<<<
 * 
 *         buffer._length += (p - begin)
 */
  __pyx_v_p = mp_encode_double(__pyx_v_p, __pyx_v_self->tx_timeout);

  /* "asynctnt/iproto/requests/streams.pyx":30
 *         p = mp_encode_double(p, self.tx_timeout)
 * 
 *         buffer._length += (p - begin)             # <<<<<<<<<<<<<<
 * 
 * @cython.final
 */
  __pyx_v_buffer->_length = (__pyx_v_buffer->_length + (__pyx_v_p - __pyx_v_begin));

  /* "asynctnt/iproto/requests/streams.pyx":6
 * @cython.final
 * cdef class BeginRequest(BaseRequest):
 *     cdef int encode_body(self, WriteBuffer buffer) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             char *p
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BeginRequest.encode_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BeginRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BeginRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BeginRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BeginRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BeginRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BeginRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.isolation, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.tx_timeout, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_self->isolation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_self->tx_timeout); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = PyTuple_New(13); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 3, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 4, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 5, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 6, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 7, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 8, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 9, __pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 10, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 11, __pyx_t_10)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 12, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.isolation, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.tx_timeout, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_11 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_v__dict = __pyx_t_11;
  __pyx_t_11 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.isolation, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.tx_timeout, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_12 = (__pyx_v__dict != Py_None);
  if (__pyx_t_12) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_10 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_10));
    __pyx_t_10 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.isolation, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.tx_timeout, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, None), state
 */
  /*else*/ {
    __pyx_t_13 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_12 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_12 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_12 = __pyx_t_13;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_12;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_pyx_unpickle_BeginRequest); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_51353055);
    __Pyx_GIVEREF(__pyx_int_51353055);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_int_51353055)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_10)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_11);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_11)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_10 = 0;
    __pyx_t_11 = 0;
    __pyx_r = __pyx_t_9;
    __pyx_t_9 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, None), state
 *     else:
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BeginRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_pyx_unpickle_BeginRequest); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_51353055);
    __Pyx_GIVEREF(__pyx_int_51353055);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_int_51353055)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_11);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_11)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_9 = 0;
    __pyx_t_11 = 0;
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BeginRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BeginRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BeginRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BeginRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BeginRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BeginRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BeginRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BeginRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BeginRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BeginRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_BeginRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BeginRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BeginRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13CommitRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13CommitRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13CommitRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13CommitRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13CommitRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13CommitRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, None), state
 */
  /*else*/ {
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_CommitRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150999062);
    __Pyx_GIVEREF(__pyx_int_150999062);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_150999062)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, None), state
 *     else:
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CommitRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_CommitRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150999062);
    __Pyx_GIVEREF(__pyx_int_150999062);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_150999062)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CommitRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CommitRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13CommitRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13CommitRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13CommitRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13CommitRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CommitRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_13CommitRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_13CommitRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CommitRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CommitRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CommitRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CommitRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_15RollbackRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_15RollbackRequest_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_15RollbackRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_15RollbackRequest_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_15RollbackRequest___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_15RollbackRequest___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_v_self->__pyx_base.op); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_as_tuples); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.parse_metadata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.push_subscribe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base.schema_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.stream_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->__pyx_base.sync); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.space);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.space);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, ((PyObject *)__pyx_v_self->__pyx_base.space))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_t_7)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 8, __pyx_t_8)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.timeout_handle);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.timeout_handle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 9, __pyx_v_self->__pyx_base.timeout_handle)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.waiter);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 10, __pyx_v_self->__pyx_base.waiter)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_9 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v__dict = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_10 = (__pyx_v__dict != Py_None);
  if (__pyx_t_10) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_schema_change, self.op, self.parse_as_tuples, self.parse_metadata, self.push_subscribe, self.schema_id, self.space, self.stream_id, self.sync, self.timeout_handle, self.waiter)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, None), state
 */
  /*else*/ {
    __pyx_t_11 = (((PyObject *)__pyx_v_self->__pyx_base.space) != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.timeout_handle != Py_None);
    if (!__pyx_t_11) {
    } else {
      __pyx_t_10 = __pyx_t_11;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_11 = (__pyx_v_self->__pyx_base.waiter != Py_None);
    __pyx_t_10 = __pyx_t_11;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_10;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_pyx_unpickle_RollbackRequest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150999062);
    __Pyx_GIVEREF(__pyx_int_150999062);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_150999062)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_9)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.space is not None or self.timeout_handle is not None or self.waiter is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, None), state
 *     else:
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RollbackRequest__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pyx_unpickle_RollbackRequest); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150999062);
    __Pyx_GIVEREF(__pyx_int_150999062);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_150999062)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.RollbackRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RollbackRequest__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_15RollbackRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_15RollbackRequest_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_15RollbackRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_15RollbackRequest_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.RollbackRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_15RollbackRequest_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_15RollbackRequest_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RollbackRequest__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_RollbackRequest__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RollbackRequest__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.RollbackRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":17
 * @cython.final
 * cdef class IProtoErrorStackFrame:
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<Frame type={} code={} message={}>".format(
 *             self.error_type,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_1__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_1__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame___repr__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame___repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "asynctnt/iproto/response.pyx":18
 * cdef class IProtoErrorStackFrame:
 *     def __repr__(self):
 *         return "<Frame type={} code={} message={}>".format(             # <<<<<<<<<<<<<<
 *             self.error_type,
 *             self.code,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Frame_type_code_message, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "asynctnt/iproto/response.pyx":20
 *         return "<Frame type={} code={} message={}>".format(
 *             self.error_type,
 *             self.code,             # <<<<<<<<<<<<<<
 *             self.message,
 *         )
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->code); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "asynctnt/iproto/response.pyx":21
 *             self.error_type,
 *             self.code,
 *             self.message,             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_v_self->error_type, __pyx_t_3, __pyx_v_self->message};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 18, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":17
 * @cython.final
 * cdef class IProtoErrorStackFrame:
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<Frame type={} code={} message={}>".format(
 *             self.error_type,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IProtoErrorStackFrame.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":13
 * cdef class IProtoErrorStackFrame:
 *     cdef:
 *         readonly str error_type             # <<<<<<<<<<<<<<
 *         readonly str file
 *         readonly int line
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_10error_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_10error_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_10error_type___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_10error_type___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->error_type);
  __pyx_r = __pyx_v_self->error_type;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":14
 *     cdef:
 *         readonly str error_type
 *         readonly str file             # <<<<<<<<<<<<<<
 *         readonly int line
 *         readonly str message
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4file_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4file_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4file___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4file___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->file);
  __pyx_r = __pyx_v_self->file;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":15
 *         readonly str error_type
 *         readonly str file
 *         readonly int line             # <<<<<<<<<<<<<<
 *         readonly str message
 *         readonly int err_no
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4line_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4line_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4line___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4line___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->line); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IProtoErrorStackFrame.line.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":16
 *         readonly str file
 *         readonly int line
 *         readonly str message             # <<<<<<<<<<<<<<
 *         readonly int err_no
 *         readonly int code
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_7message_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_7message_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_7message___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_7message___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->message);
  __pyx_r = __pyx_v_self->message;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":17
 *         readonly int line
 *         readonly str message
 *         readonly int err_no             # <<<<<<<<<<<<<<
 *         readonly int code
 *         readonly dict fields
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6err_no_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6err_no_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6err_no___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6err_no___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->err_no); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IProtoErrorStackFrame.err_no.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":18
 *         readonly str message
 *         readonly int err_no
 *         readonly int code             # <<<<<<<<<<<<<<
 *         readonly dict fields
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4code_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4code_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4code___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4code___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->code); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IProtoErrorStackFrame.code.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":19
 *         readonly int err_no
 *         readonly int code
 *         readonly dict fields             # <<<<<<<<<<<<<<
 * 
 * cdef class IProtoError:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6fields_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6fields_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6fields___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6fields___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->fields);
  __pyx_r = __pyx_v_self->fields;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_2__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_2__reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.code, self.err_no, self.error_type, self.fields, self.file, self.line, self.message)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->err_no); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->line); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->error_type);
  __Pyx_GIVEREF(__pyx_v_self->error_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_self->error_type)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->fields);
  __Pyx_GIVEREF(__pyx_v_self->fields);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_self->fields)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->file);
  __Pyx_GIVEREF(__pyx_v_self->file);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_self->file)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->message);
  __Pyx_GIVEREF(__pyx_v_self->message);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_self->message)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.code, self.err_no, self.error_type, self.fields, self.file, self.line, self.message)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.code, self.err_no, self.error_type, self.fields, self.file, self.line, self.message)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.error_type is not None or self.fields is not None or self.file is not None or self.message is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.code, self.err_no, self.error_type, self.fields, self.file, self.line, self.message)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.error_type is not None or self.fields is not None or self.file is not None or self.message is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, None), state
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_self->error_type != ((PyObject*)Py_None));
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->fields != ((PyObject*)Py_None));
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->file != ((PyObject*)Py_None));
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->message != ((PyObject*)Py_None));
    __pyx_t_5 = __pyx_t_6;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_5;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.error_type is not None or self.fields is not None or self.file is not None or self.message is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.error_type is not None or self.fields is not None or self.file is not None or self.message is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_IProtoErrorStackF); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_179467967);
    __Pyx_GIVEREF(__pyx_int_179467967);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_179467967)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.error_type is not None or self.fields is not None or self.file is not None or self.message is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, None), state
 *     else:
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IProtoErrorStackFrame__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_IProtoErrorStackF); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_179467967);
    __Pyx_GIVEREF(__pyx_int_179467967);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_179467967)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IProtoErrorStackFrame.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IProtoErrorStackFrame__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IProtoErrorStackFrame.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IProtoErrorStackFrame__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IProtoErrorStackFrame__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IProtoErrorStackFrame__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IProtoErrorStackFrame.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":23
 * cdef class IProtoError:
 *     cdef:
 *         readonly list trace             # <<<<<<<<<<<<<<
 * 
 * cdef class Response:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11IProtoError_5trace_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11IProtoError_5trace_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11IProtoError_5trace___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11IProtoError_5trace___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->trace);
  __pyx_r = __pyx_v_self->trace;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11IProtoError_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11IProtoError_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11IProtoError_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11IProtoError_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11IProtoError___reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11IProtoError___reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.trace,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->trace);
  __Pyx_GIVEREF(__pyx_v_self->trace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->trace)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.trace,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.trace,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.trace is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.trace,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.trace is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->trace != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.trace is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.trace is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_IProtoError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_246409365);
    __Pyx_GIVEREF(__pyx_int_246409365);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_246409365)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.trace is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, None), state
 *     else:
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IProtoError__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_IProtoError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_246409365);
    __Pyx_GIVEREF(__pyx_int_246409365);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_246409365)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IProtoError.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IProtoError__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11IProtoError_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11IProtoError_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11IProtoError_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11IProtoError_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IProtoError.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_11IProtoError_2__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_11IProtoError_2__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IProtoError__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IProtoError__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IProtoError__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.IProtoError.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":35
 *     """
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.request_ = None
 *         self.sync_ = 0
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_8Response_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8asynctnt_6iproto_8protocol_8Response_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response___cinit__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_8Response___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "asynctnt/iproto/response.pyx":36
 * 
 *     def __cinit__(self):
 *         self.request_ = None             # <<<<<<<<<<<<<<
 *         self.sync_ = 0
 *         self.code_ = -1
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->request_);
  __Pyx_DECREF((PyObject *)__pyx_v_self->request_);
  __pyx_v_self->request_ = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)Py_None);

  /* "asynctnt/iproto/response.pyx":37
 *     def __cinit__(self):
 *         self.request_ = None
 *         self.sync_ = 0             # <<<<<<<<<<<<<<
 *         self.code_ = -1
 *         self.return_code_ = -1
 */
  __pyx_v_self->sync_ = 0;

  /* "asynctnt/iproto/response.pyx":38
 *         self.request_ = None
 *         self.sync_ = 0
 *         self.code_ = -1             # <<<<<<<<<<<<<<
 *         self.return_code_ = -1
 *         self.schema_id_ = -1
 */
  __pyx_v_self->code_ = -1;

  /* "asynctnt/iproto/response.pyx":39
 *         self.sync_ = 0
 *         self.code_ = -1
 *         self.return_code_ = -1             # <<<<<<<<<<<<<<
 *         self.schema_id_ = -1
 *         self.errmsg = None
 */
  __pyx_v_self->return_code_ = -1;

  /* "asynctnt/iproto/response.pyx":40
 *         self.code_ = -1
 *         self.return_code_ = -1
 *         self.schema_id_ = -1             # <<<<<<<<<<<<<<
 *         self.errmsg = None
 *         self.error = None
 */
  __pyx_v_self->schema_id_ = -1L;

  /* "asynctnt/iproto/response.pyx":41
 *         self.return_code_ = -1
 *         self.schema_id_ = -1
 *         self.errmsg = None             # <<<<<<<<<<<<<<
 *         self.error = None
 *         self._rowcount = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->errmsg);
  __Pyx_DECREF(__pyx_v_self->errmsg);
  __pyx_v_self->errmsg = ((PyObject*)Py_None);

  /* "asynctnt/iproto/response.pyx":42
 *         self.schema_id_ = -1
 *         self.errmsg = None
 *         self.error = None             # <<<<<<<<<<<<<<
 *         self._rowcount = 0
 *         self.body = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->error);
  __Pyx_DECREF((PyObject *)__pyx_v_self->error);
  __pyx_v_self->error = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)Py_None);

  /* "asynctnt/iproto/response.pyx":43
 *         self.errmsg = None
 *         self.error = None
 *         self._rowcount = 0             # <<<<<<<<<<<<<<
 *         self.body = None
 *         self.encoding = None
 */
  __pyx_v_self->_rowcount = 0;

  /* "asynctnt/iproto/response.pyx":44
 *         self.error = None
 *         self._rowcount = 0
 *         self.body = None             # <<<<<<<<<<<<<<
 *         self.encoding = None
 *         self.metadata = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->body);
  __Pyx_DECREF(__pyx_v_self->body);
  __pyx_v_self->body = ((PyObject*)Py_None);

  /* "asynctnt/iproto/response.pyx":45
 *         self._rowcount = 0
 *         self.body = None
 *         self.encoding = None             # <<<<<<<<<<<<<<
 *         self.metadata = None
 *         self.params = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->encoding);
  __Pyx_DECREF(__pyx_v_self->encoding);
  __pyx_v_self->encoding = ((PyObject*)Py_None);

  /* "asynctnt/iproto/response.pyx":46
 *         self.body = None
 *         self.encoding = None
 *         self.metadata = None             # <<<<<<<<<<<<<<
 *         self.params = None
 *         self.params_count = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->metadata);
  __Pyx_DECREF((PyObject *)__pyx_v_self->metadata);
  __pyx_v_self->metadata = ((struct C_Metadata *)Py_None);

  /* "asynctnt/iproto/response.pyx":47
 *         self.encoding = None
 *         self.metadata = None
 *         self.params = None             # <<<<<<<<<<<<<<
 *         self.params_count = 0
 *         self.autoincrement_ids = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->params);
  __Pyx_DECREF((PyObject *)__pyx_v_self->params);
  __pyx_v_self->params = ((struct C_Metadata *)Py_None);

  /* "asynctnt/iproto/response.pyx":48
 *         self.metadata = None
 *         self.params = None
 *         self.params_count = 0             # <<<<<<<<<<<<<<
 *         self.autoincrement_ids = None
 *         self.stmt_id_ = 0
 */
  __pyx_v_self->params_count = 0;

  /* "asynctnt/iproto/response.pyx":49
 *         self.params = None
 *         self.params_count = 0
 *         self.autoincrement_ids = None             # <<<<<<<<<<<<<<
 *         self.stmt_id_ = 0
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->autoincrement_ids);
  __Pyx_DECREF(__pyx_v_self->autoincrement_ids);
  __pyx_v_self->autoincrement_ids = ((PyObject*)Py_None);

  /* "asynctnt/iproto/response.pyx":50
 *         self.params_count = 0
 *         self.autoincrement_ids = None
 *         self.stmt_id_ = 0             # <<<<<<<<<<<<<<
 * 
 *         self._push_subscribe = False
 */
  __pyx_v_self->stmt_id_ = 0;

  /* "asynctnt/iproto/response.pyx":52
 *         self.stmt_id_ = 0
 * 
 *         self._push_subscribe = False             # <<<<<<<<<<<<<<
 *         self._q = None
 *         self._push_event = None
 */
  __pyx_v_self->_push_subscribe = 0;

  /* "asynctnt/iproto/response.pyx":53
 * 
 *         self._push_subscribe = False
 *         self._q = None             # <<<<<<<<<<<<<<
 *         self._push_event = None
 *         self._q_append = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_q);
  __Pyx_DECREF(__pyx_v_self->_q);
  __pyx_v_self->_q = Py_None;

  /* "asynctnt/iproto/response.pyx":54
 *         self._push_subscribe = False
 *         self._q = None
 *         self._push_event = None             # <<<<<<<<<<<<<<
 *         self._q_append = None
 *         self._q_popleft = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_push_event);
  __Pyx_DECREF(__pyx_v_self->_push_event);
  __pyx_v_self->_push_event = Py_None;

  /* "asynctnt/iproto/response.pyx":55
 *         self._q = None
 *         self._push_event = None
 *         self._q_append = None             # <<<<<<<<<<<<<<
 *         self._q_popleft = None
 *         self._push_event_set = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_q_append);
  __Pyx_DECREF(__pyx_v_self->_q_append);
  __pyx_v_self->_q_append = Py_None;

  /* "asynctnt/iproto/response.pyx":56
 *         self._push_event = None
 *         self._q_append = None
 *         self._q_popleft = None             # <<<<<<<<<<<<<<
 *         self._push_event_set = None
 *         self._push_event_clear = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_q_popleft);
  __Pyx_DECREF(__pyx_v_self->_q_popleft);
  __pyx_v_self->_q_popleft = Py_None;

  /* "asynctnt/iproto/response.pyx":57
 *         self._q_append = None
 *         self._q_popleft = None
 *         self._push_event_set = None             # <<<<<<<<<<<<<<
 *         self._push_event_clear = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_push_event_set);
  __Pyx_DECREF(__pyx_v_self->_push_event_set);
  __pyx_v_self->_push_event_set = Py_None;

  /* "asynctnt/iproto/response.pyx":58
 *         self._q_popleft = None
 *         self._push_event_set = None
 *         self._push_event_clear = None             # <<<<<<<<<<<<<<
 * 
 *     cdef inline bint is_error(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_push_event_clear);
  __Pyx_DECREF(__pyx_v_self->_push_event_clear);
  __pyx_v_self->_push_event_clear = Py_None;

  /* "asynctnt/iproto/response.pyx":35
 *     """
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.request_ = None
 *         self.sync_ = 0
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":60
 *         self._push_event_clear = None
 * 
 *     cdef inline bint is_error(self):             # <<<<<<<<<<<<<<
 *         return self.code_ >= 0x8000
 * 
 */

static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Response_is_error(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  int __pyx_r;

  /* "asynctnt/iproto/response.pyx":61
 * 
 *     cdef inline bint is_error(self):
 *         return self.code_ >= 0x8000             # <<<<<<<<<<<<<<
 * 
 *     # noinspection PyAttributeOutsideInit
 */
  __pyx_r = (__pyx_v_self->code_ >= 0x8000);
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":60
 *         self._push_event_clear = None
 * 
 *     cdef inline bint is_error(self):             # <<<<<<<<<<<<<<
 *         return self.code_ >= 0x8000
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":64
 * 
 *     # noinspection PyAttributeOutsideInit
 *     cdef inline void init_push(self):             # <<<<<<<<<<<<<<
 *         self._push_subscribe = True
 *         self._q = collections.deque()
 */

static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_init_push(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init_push", 1);

  /* "asynctnt/iproto/response.pyx":65
 *     # noinspection PyAttributeOutsideInit
 *     cdef inline void init_push(self):
 *         self._push_subscribe = True             # <<<<<<<<<<<<<<
 *         self._q = collections.deque()
 *         self._push_event = asyncio.Event()
 */
  __pyx_v_self->_push_subscribe = 1;

  /* "asynctnt/iproto/response.pyx":66
 *     cdef inline void init_push(self):
 *         self._push_subscribe = True
 *         self._q = collections.deque()             # <<<<<<<<<<<<<<
 *         self._push_event = asyncio.Event()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_collections); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_deque); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_q);
  __Pyx_DECREF(__pyx_v_self->_q);
  __pyx_v_self->_q = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/response.pyx":67
 *         self._push_subscribe = True
 *         self._q = collections.deque()
 *         self._push_event = asyncio.Event()             # <<<<<<<<<<<<<<
 * 
 *         self._q_append = self._q.append
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Event); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_push_event);
  __Pyx_DECREF(__pyx_v_self->_push_event);
  __pyx_v_self->_push_event = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/response.pyx":69
 *         self._push_event = asyncio.Event()
 * 
 *         self._q_append = self._q.append             # <<<<<<<<<<<<<<
 *         self._q_popleft = self._q.popleft
 *         self._push_event_set = self._push_event.set
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_q, __pyx_n_s_append); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_q_append);
  __Pyx_DECREF(__pyx_v_self->_q_append);
  __pyx_v_self->_q_append = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/response.pyx":70
 * 
 *         self._q_append = self._q.append
 *         self._q_popleft = self._q.popleft             # <<<<<<<<<<<<<<
 *         self._push_event_set = self._push_event.set
 *         self._push_event_clear = self._push_event.clear
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_q, __pyx_n_s_popleft); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_q_popleft);
  __Pyx_DECREF(__pyx_v_self->_q_popleft);
  __pyx_v_self->_q_popleft = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/response.pyx":71
 *         self._q_append = self._q.append
 *         self._q_popleft = self._q.popleft
 *         self._push_event_set = self._push_event.set             # <<<<<<<<<<<<<<
 *         self._push_event_clear = self._push_event.clear
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_push_event, __pyx_n_s_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_push_event_set);
  __Pyx_DECREF(__pyx_v_self->_push_event_set);
  __pyx_v_self->_push_event_set = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/response.pyx":72
 *         self._q_popleft = self._q.popleft
 *         self._push_event_set = self._push_event.set
 *         self._push_event_clear = self._push_event.clear             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void add_push(self, push):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_push_event, __pyx_n_s_clear); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_push_event_clear);
  __Pyx_DECREF(__pyx_v_self->_push_event_clear);
  __pyx_v_self->_push_event_clear = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/response.pyx":64
 * 
 *     # noinspection PyAttributeOutsideInit
 *     cdef inline void init_push(self):             # <<<<<<<<<<<<<<
 *         self._push_subscribe = True
 *         self._q = collections.deque()
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.init_push", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/response.pyx":74
 *         self._push_event_clear = self._push_event.clear
 * 
 *     cdef inline void add_push(self, push):             # <<<<<<<<<<<<<<
 *         if not self._push_subscribe:
 *             return
 */

static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_add_push(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self, PyObject *__pyx_v_push) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_push", 1);

  /* "asynctnt/iproto/response.pyx":75
 * 
 *     cdef inline void add_push(self, push):
 *         if not self._push_subscribe:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (!__pyx_v_self->_push_subscribe);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/response.pyx":76
 *     cdef inline void add_push(self, push):
 *         if not self._push_subscribe:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._q_append(push)
 */
    goto __pyx_L0;

    /* "asynctnt/iproto/response.pyx":75
 * 
 *     cdef inline void add_push(self, push):
 *         if not self._push_subscribe:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "asynctnt/iproto/response.pyx":78
 *             return
 * 
 *         self._q_append(push)             # <<<<<<<<<<<<<<
 *         self.notify()
 * 
 */
  __Pyx_INCREF(__pyx_v_self->_q_append);
  __pyx_t_3 = __pyx_v_self->_q_append; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_push};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/response.pyx":79
 * 
 *         self._q_append(push)
 *         self.notify()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline int push_len(self):
 */
  __pyx_f_8asynctnt_6iproto_8protocol_8Response_notify(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 79, __pyx_L1_error)

  /* "asynctnt/iproto/response.pyx":74
 *         self._push_event_clear = self._push_event.clear
 * 
 *     cdef inline void add_push(self, push):             # <<<<<<<<<<<<<<
 *         if not self._push_subscribe:
 *             return
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.add_push", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/response.pyx":81
 *         self.notify()
 * 
 *     cdef inline int push_len(self):             # <<<<<<<<<<<<<<
 *         return len(self._q)
 * 
 */

static CYTHON_INLINE int __pyx_f_8asynctnt_6iproto_8protocol_8Response_push_len(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("push_len", 1);

  /* "asynctnt/iproto/response.pyx":82
 * 
 *     cdef inline int push_len(self):
 *         return len(self._q)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline object pop_push(self):
 */
  __pyx_t_1 = __pyx_v_self->_q;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(5, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":81
 *         self.notify()
 * 
 *     cdef inline int push_len(self):             # <<<<<<<<<<<<<<
 *         return len(self._q)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.push_len", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":84
 *         return len(self._q)
 * 
 *     cdef inline object pop_push(self):             # <<<<<<<<<<<<<<
 *         if not self._push_subscribe:
 *             raise RuntimeError('Cannot pop push from a non-async response')
 */

static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_8Response_pop_push(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_v_push = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_push", 1);

  /* "asynctnt/iproto/response.pyx":85
 * 
 *     cdef inline object pop_push(self):
 *         if not self._push_subscribe:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Cannot pop push from a non-async response')
 * 
 */
  __pyx_t_1 = (!__pyx_v_self->_push_subscribe);
  if (unlikely(__pyx_t_1)) {

    /* "asynctnt/iproto/response.pyx":86
 *     cdef inline object pop_push(self):
 *         if not self._push_subscribe:
 *             raise RuntimeError('Cannot pop push from a non-async response')             # <<<<<<<<<<<<<<
 * 
 *         push = self._q_popleft()
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(5, 86, __pyx_L1_error)

    /* "asynctnt/iproto/response.pyx":85
 * 
 *     cdef inline object pop_push(self):
 *         if not self._push_subscribe:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Cannot pop push from a non-async response')
 * 
 */
  }

  /* "asynctnt/iproto/response.pyx":88
 *             raise RuntimeError('Cannot pop push from a non-async response')
 * 
 *         push = self._q_popleft()             # <<<<<<<<<<<<<<
 *         if len(self._q) == 0:
 *             self._push_event_clear()
 */
  __Pyx_INCREF(__pyx_v_self->_q_popleft);
  __pyx_t_3 = __pyx_v_self->_q_popleft; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_push = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/response.pyx":89
 * 
 *         push = self._q_popleft()
 *         if len(self._q) == 0:             # <<<<<<<<<<<<<<
 *             self._push_event_clear()
 *         return push
 */
  __pyx_t_2 = __pyx_v_self->_q;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_6 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(5, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = (__pyx_t_6 == 0);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/response.pyx":90
 *         push = self._q_popleft()
 *         if len(self._q) == 0:
 *             self._push_event_clear()             # <<<<<<<<<<<<<<
 *         return push
 * 
 */
    __Pyx_INCREF(__pyx_v_self->_push_event_clear);
    __pyx_t_3 = __pyx_v_self->_push_event_clear; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 90, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynctnt/iproto/response.pyx":89
 * 
 *         push = self._q_popleft()
 *         if len(self._q) == 0:             # <<<<<<<<<<<<<<
 *             self._push_event_clear()
 *         return push
 */
  }

  /* "asynctnt/iproto/response.pyx":91
 *         if len(self._q) == 0:
 *             self._push_event_clear()
 *         return push             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void set_data(self, list data):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_push);
  __pyx_r = __pyx_v_push;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":84
 *         return len(self._q)
 * 
 *     cdef inline object pop_push(self):             # <<<<<<<<<<<<<<
 *         if not self._push_subscribe:
 *             raise RuntimeError('Cannot pop push from a non-async response')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.pop_push", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_push);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":93
 *         return push
 * 
 *     cdef inline void set_data(self, list data):             # <<<<<<<<<<<<<<
 *         self.body = data
 *         self.notify()
 */

static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_set_data(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self, PyObject *__pyx_v_data) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_data", 1);

  /* "asynctnt/iproto/response.pyx":94
 * 
 *     cdef inline void set_data(self, list data):
 *         self.body = data             # <<<<<<<<<<<<<<
 *         self.notify()
 * 
 */
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  __Pyx_GOTREF(__pyx_v_self->body);
  __Pyx_DECREF(__pyx_v_self->body);
  __pyx_v_self->body = __pyx_v_data;

  /* "asynctnt/iproto/response.pyx":95
 *     cdef inline void set_data(self, list data):
 *         self.body = data
 *         self.notify()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void set_exception(self, exc):
 */
  __pyx_f_8asynctnt_6iproto_8protocol_8Response_notify(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 95, __pyx_L1_error)

  /* "asynctnt/iproto/response.pyx":93
 *         return push
 * 
 *     cdef inline void set_data(self, list data):             # <<<<<<<<<<<<<<
 *         self.body = data
 *         self.notify()
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.set_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/response.pyx":97
 *         self.notify()
 * 
 *     cdef inline void set_exception(self, exc):             # <<<<<<<<<<<<<<
 *         self._exception = exc
 *         self.notify()
 */

static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_set_exception(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self, PyObject *__pyx_v_exc) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_exception", 1);

  /* "asynctnt/iproto/response.pyx":98
 * 
 *     cdef inline void set_exception(self, exc):
 *         self._exception = exc             # <<<<<<<<<<<<<<
 *         self.notify()
 * 
 */
  __Pyx_INCREF(__pyx_v_exc);
  __Pyx_GIVEREF(__pyx_v_exc);
  __Pyx_GOTREF(__pyx_v_self->_exception);
  __Pyx_DECREF(__pyx_v_self->_exception);
  __pyx_v_self->_exception = __pyx_v_exc;

  /* "asynctnt/iproto/response.pyx":99
 *     cdef inline void set_exception(self, exc):
 *         self._exception = exc
 *         self.notify()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline object get_exception(self):
 */
  __pyx_f_8asynctnt_6iproto_8protocol_8Response_notify(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 99, __pyx_L1_error)

  /* "asynctnt/iproto/response.pyx":97
 *         self.notify()
 * 
 *     cdef inline void set_exception(self, exc):             # <<<<<<<<<<<<<<
 *         self._exception = exc
 *         self.notify()
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.set_exception", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/response.pyx":101
 *         self.notify()
 * 
 *     cdef inline object get_exception(self):             # <<<<<<<<<<<<<<
 *         return self._exception
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_8asynctnt_6iproto_8protocol_8Response_get_exception(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_exception", 1);

  /* "asynctnt/iproto/response.pyx":102
 * 
 *     cdef inline object get_exception(self):
 *         return self._exception             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void notify(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_exception);
  __pyx_r = __pyx_v_self->_exception;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":101
 *         self.notify()
 * 
 *     cdef inline object get_exception(self):             # <<<<<<<<<<<<<<
 *         return self._exception
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":104
 *         return self._exception
 * 
 *     cdef inline void notify(self):             # <<<<<<<<<<<<<<
 *         if self._push_subscribe:
 *             self._push_event_set()  # Notify that there is no more data
 */

static CYTHON_INLINE void __pyx_f_8asynctnt_6iproto_8protocol_8Response_notify(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("notify", 1);

  /* "asynctnt/iproto/response.pyx":105
 * 
 *     cdef inline void notify(self):
 *         if self._push_subscribe:             # <<<<<<<<<<<<<<
 *             self._push_event_set()  # Notify that there is no more data
 * 
 */
  if (__pyx_v_self->_push_subscribe) {

    /* "asynctnt/iproto/response.pyx":106
 *     cdef inline void notify(self):
 *         if self._push_subscribe:
 *             self._push_event_set()  # Notify that there is no more data             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):  # pragma: nocover
 */
    __Pyx_INCREF(__pyx_v_self->_push_event_set);
    __pyx_t_2 = __pyx_v_self->_push_event_set; __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 106, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/response.pyx":105
 * 
 *     cdef inline void notify(self):
 *         if self._push_subscribe:             # <<<<<<<<<<<<<<
 *             self._push_event_set()  # Notify that there is no more data
 * 
 */
  }

  /* "asynctnt/iproto/response.pyx":104
 *         return self._exception
 * 
 *     cdef inline void notify(self):             # <<<<<<<<<<<<<<
 *         if self._push_subscribe:
 *             self._push_event_set()  # Notify that there is no more data
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.notify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/response.pyx":108
 *             self._push_event_set()  # Notify that there is no more data
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         data = self.body
 *         if data is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_2__repr__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":112
 *         if data is not None:
 *             if len(data) > 10:
 *                 parts = map(lambda x: ', '.join(map(repr, x)), [             # <<<<<<<<<<<<<<
 *                     data[:5],
 *                     data[-2:]
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_8__repr___lambda1(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_8Response_8__repr___lambda1 = {"lambda1", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_8__repr___lambda1, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_8__repr___lambda1(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_x = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda1 (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_x)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 112, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "lambda1") < 0)) __PYX_ERR(5, 112, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_x = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda1", 1, 1, 1, __pyx_nargs); __PYX_ERR(5, 112, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.__repr__.lambda1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda1(__pyx_self, __pyx_v_x);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda1", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetBuiltinName(__pyx_n_s_repr); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(5, 112, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_x);
  __Pyx_GIVEREF(__pyx_v_x);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_x)) __PYX_ERR(5, 112, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyUnicode_Join(__pyx_kp_u__14, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.__repr__.lambda1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":108
 *             self._push_event_set()  # Notify that there is no more data
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         data = self.body
 *         if data is not None:
 */

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_2__repr__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_parts = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "asynctnt/iproto/response.pyx":109
 * 
 *     def __repr__(self):  # pragma: nocover
 *         data = self.body             # <<<<<<<<<<<<<<
 *         if data is not None:
 *             if len(data) > 10:
 */
  __pyx_t_1 = __pyx_v_self->body;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/response.pyx":110
 *     def __repr__(self):  # pragma: nocover
 *         data = self.body
 *         if data is not None:             # <<<<<<<<<<<<<<
 *             if len(data) > 10:
 *                 parts = map(lambda x: ', '.join(map(repr, x)), [
 */
  __pyx_t_2 = (__pyx_v_data != Py_None);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/response.pyx":111
 *         data = self.body
 *         if data is not None:
 *             if len(data) > 10:             # <<<<<<<<<<<<<<
 *                 parts = map(lambda x: ', '.join(map(repr, x)), [
 *                     data[:5],
 */
    __pyx_t_3 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(5, 111, __pyx_L1_error)
    __pyx_t_2 = (__pyx_t_3 > 10);
    if (__pyx_t_2) {

      /* "asynctnt/iproto/response.pyx":112
 *         if data is not None:
 *             if len(data) > 10:
 *                 parts = map(lambda x: ', '.join(map(repr, x)), [             # <<<<<<<<<<<<<<
 *                     data[:5],
 *                     data[-2:]
 */
      __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_8Response_8__repr___lambda1, 0, __pyx_n_s_repr___locals_lambda, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "asynctnt/iproto/response.pyx":113
 *             if len(data) > 10:
 *                 parts = map(lambda x: ', '.join(map(repr, x)), [
 *                     data[:5],             # <<<<<<<<<<<<<<
 *                     data[-2:]
 *                 ])
 */
      __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_data, 0, 5, NULL, NULL, &__pyx_slice__15, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "asynctnt/iproto/response.pyx":114
 *                 parts = map(lambda x: ', '.join(map(repr, x)), [
 *                     data[:5],
 *                     data[-2:]             # <<<<<<<<<<<<<<
 *                 ])
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_data, -2L, 0, NULL, NULL, &__pyx_slice__16, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 114, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "asynctnt/iproto/response.pyx":112
 *         if data is not None:
 *             if len(data) > 10:
 *                 parts = map(lambda x: ', '.join(map(repr, x)), [             # <<<<<<<<<<<<<<
 *                     data[:5],
 *                     data[-2:]
 */
      __pyx_t_6 = PyList_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_4);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_4)) __PYX_ERR(5, 112, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_5);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_5)) __PYX_ERR(5, 112, __pyx_L1_error);
      __pyx_t_4 = 0;
      __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(5, 112, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_6);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6)) __PYX_ERR(5, 112, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_parts = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "asynctnt/iproto/response.pyx":117
 *                 ])
 * 
 *                 data = ' ... '.join(parts)             # <<<<<<<<<<<<<<
 *                 data = '[' + data + ']'
 * 
 */
      __pyx_t_6 = PyUnicode_Join(__pyx_kp_u__17, __pyx_v_parts); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "asynctnt/iproto/response.pyx":118
 * 
 *                 data = ' ... '.join(parts)
 *                 data = '[' + data + ']'             # <<<<<<<<<<<<<<
 * 
 *         return '<{} sync={} rowcount={} data={}>'.format(
 */
      __pyx_t_6 = __Pyx_PyUnicode_ConcatSafe(__pyx_kp_u__18, __pyx_v_data); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_6, __pyx_kp_u__19); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "asynctnt/iproto/response.pyx":111
 *         data = self.body
 *         if data is not None:
 *             if len(data) > 10:             # <<<<<<<<<<<<<<
 *                 parts = map(lambda x: ', '.join(map(repr, x)), [
 *                     data[:5],
 */
    }

    /* "asynctnt/iproto/response.pyx":110
 *     def __repr__(self):  # pragma: nocover
 *         data = self.body
 *         if data is not None:             # <<<<<<<<<<<<<<
 *             if len(data) > 10:
 *                 parts = map(lambda x: ', '.join(map(repr, x)), [
 */
  }

  /* "asynctnt/iproto/response.pyx":120
 *                 data = '[' + data + ']'
 * 
 *         return '<{} sync={} rowcount={} data={}>'.format(             # <<<<<<<<<<<<<<
 *             self.__class__.__name__, self.sync, self.rowcount, data)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_sync_rowcount_data, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "asynctnt/iproto/response.pyx":121
 * 
 *         return '<{} sync={} rowcount={} data={}>'.format(
 *             self.__class__.__name__, self.sync, self.rowcount, data)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sync); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rowcount); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_8, __pyx_t_4, __pyx_t_1, __pyx_t_7, __pyx_v_data};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_9, 4+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":108
 *             self._push_event_set()  # Notify that there is no more data
 * 
 *     def __repr__(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         data = self.body
 *         if data is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_parts);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":123
 *             self.__class__.__name__, self.sync, self.rowcount, data)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sync(self) -> int:
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_4sync_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_4sync_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_4sync___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_4sync___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "asynctnt/iproto/response.pyx":128
 *             Response's sync (incremental id) for the corresponding request
 *         """
 *         return <int> self.sync_             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(((int)__pyx_v_self->sync_)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":123
 *             self.__class__.__name__, self.sync, self.rowcount, data)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sync(self) -> int:
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.sync.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":130
 *         return <int> self.sync_
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def code(self) -> int:
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_4code_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_4code_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_4code___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_4code___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "asynctnt/iproto/response.pyx":135
 *             Response code (0 - success)
 *         """
 *         return <int> self.code_             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(((int)__pyx_v_self->code_)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":130
 *         return <int> self.sync_
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def code(self) -> int:
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.code.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":137
 *         return <int> self.code_
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def return_code(self) -> int:
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_11return_code_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_11return_code_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_11return_code___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_11return_code___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "asynctnt/iproto/response.pyx":142
 *             Response return code (It's essentially a code & 0x7FFF)
 *         """
 *         return <int> self.return_code_             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(((int)__pyx_v_self->return_code_)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":137
 *         return <int> self.code_
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def return_code(self) -> int:
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.return_code.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":144
 *         return <int> self.return_code_
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def schema_id(self) -> int:
 *         return <int> self.schema_id_
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_9schema_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_9schema_id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_9schema_id___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_9schema_id___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "asynctnt/iproto/response.pyx":146
 *     @property
 *     def schema_id(self) -> int:
 *         return <int> self.schema_id_             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(((int)__pyx_v_self->schema_id_)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":144
 *         return <int> self.return_code_
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def schema_id(self) -> int:
 *         return <int> self.schema_id_
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.schema_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":148
 *         return <int> self.schema_id_
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rowcount(self) -> int:
 *         if self.body is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_8rowcount_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_8rowcount_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_8rowcount___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_8rowcount___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  uint32_t __pyx_v_self_len;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  uint32_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "asynctnt/iproto/response.pyx":150
 *     @property
 *     def rowcount(self) -> int:
 *         if self.body is not None:             # <<<<<<<<<<<<<<
 *             self_len = self._len()
 *             if self_len > 0:
 */
  __pyx_t_1 = (__pyx_v_self->body != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "asynctnt/iproto/response.pyx":151
 *     def rowcount(self) -> int:
 *         if self.body is not None:
 *             self_len = self._len()             # <<<<<<<<<<<<<<
 *             if self_len > 0:
 *                 return self_len
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_8Response__len(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 151, __pyx_L1_error)
    __pyx_v_self_len = __pyx_t_2;

    /* "asynctnt/iproto/response.pyx":152
 *         if self.body is not None:
 *             self_len = self._len()
 *             if self_len > 0:             # <<<<<<<<<<<<<<
 *                 return self_len
 *         return self._rowcount
 */
    __pyx_t_1 = (__pyx_v_self_len > 0);
    if (__pyx_t_1) {

      /* "asynctnt/iproto/response.pyx":153
 *             self_len = self._len()
 *             if self_len > 0:
 *                 return self_len             # <<<<<<<<<<<<<<
 *         return self._rowcount
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_uint32_t(__pyx_v_self_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "asynctnt/iproto/response.pyx":152
 *         if self.body is not None:
 *             self_len = self._len()
 *             if self_len > 0:             # <<<<<<<<<<<<<<
 *                 return self_len
 *         return self._rowcount
 */
    }

    /* "asynctnt/iproto/response.pyx":150
 *     @property
 *     def rowcount(self) -> int:
 *         if self.body is not None:             # <<<<<<<<<<<<<<
 *             self_len = self._len()
 *             if self_len > 0:
 */
  }

  /* "asynctnt/iproto/response.pyx":154
 *             if self_len > 0:
 *                 return self_len
 *         return self._rowcount             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->_rowcount); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":148
 *         return <int> self.schema_id_
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rowcount(self) -> int:
 *         if self.body is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.rowcount.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":156
 *         return self._rowcount
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def stmt_id(self) -> Optional[int]:
 *         return self.stmt_id_
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_7stmt_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_7stmt_id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_7stmt_id___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_7stmt_id___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "asynctnt/iproto/response.pyx":158
 *     @property
 *     def stmt_id(self) -> Optional[int]:
 *         return self.stmt_id_             # <<<<<<<<<<<<<<
 * 
 *     def done(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->stmt_id_); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":156
 *         return self._rowcount
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def stmt_id(self) -> Optional[int]:
 *         return self.stmt_id_
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.stmt_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":160
 *         return self.stmt_id_
 * 
 *     def done(self):             # <<<<<<<<<<<<<<
 *         return self.code_ >= 0
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_5done(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_8Response_5done = {"done", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_5done, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_5done(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("done (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("done", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "done", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_4done(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_4done(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("done", 1);

  /* "asynctnt/iproto/response.pyx":161
 * 
 *     def done(self):
 *         return self.code_ >= 0             # <<<<<<<<<<<<<<
 * 
 *     cdef inline uint32_t _len(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->code_ >= 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":160
 *         return self.stmt_id_
 * 
 *     def done(self):             # <<<<<<<<<<<<<<
 *         return self.code_ >= 0
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.done", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":163
 *         return self.code_ >= 0
 * 
 *     cdef inline uint32_t _len(self):             # <<<<<<<<<<<<<<
 *         return <uint32_t> cpython.list.PyList_GET_SIZE(self.body)
 * 
 */

static CYTHON_INLINE uint32_t __pyx_f_8asynctnt_6iproto_8protocol_8Response__len(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  uint32_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_len", 1);

  /* "asynctnt/iproto/response.pyx":164
 * 
 *     cdef inline uint32_t _len(self):
 *         return <uint32_t> cpython.list.PyList_GET_SIZE(self.body)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self) -> int:
 */
  __pyx_t_1 = __pyx_v_self->body;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_r = ((uint32_t)PyList_GET_SIZE(__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":163
 *         return self.code_ >= 0
 * 
 *     cdef inline uint32_t _len(self):             # <<<<<<<<<<<<<<
 *         return <uint32_t> cpython.list.PyList_GET_SIZE(self.body)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":166
 *         return <uint32_t> cpython.list.PyList_GET_SIZE(self.body)
 * 
 *     def __len__(self) -> int:             # <<<<<<<<<<<<<<
 *         if self.body is not None:
 *             return <int> self._len()
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_8asynctnt_6iproto_8protocol_8Response_7__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_8asynctnt_6iproto_8protocol_8Response_7__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_6__len__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_8asynctnt_6iproto_8protocol_8Response_6__len__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  uint32_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/response.pyx":167
 * 
 *     def __len__(self) -> int:
 *         if self.body is not None:             # <<<<<<<<<<<<<<
 *             return <int> self._len()
 *         return 0
 */
  __pyx_t_1 = (__pyx_v_self->body != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "asynctnt/iproto/response.pyx":168
 *     def __len__(self) -> int:
 *         if self.body is not None:
 *             return <int> self._len()             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_8Response__len(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 168, __pyx_L1_error)
    __pyx_r = ((int)__pyx_t_2);
    goto __pyx_L0;

    /* "asynctnt/iproto/response.pyx":167
 * 
 *     def __len__(self) -> int:
 *         if self.body is not None:             # <<<<<<<<<<<<<<
 *             return <int> self._len()
 *         return 0
 */
  }

  /* "asynctnt/iproto/response.pyx":169
 *         if self.body is not None:
 *             return <int> self._len()
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, i):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":166
 *         return <uint32_t> cpython.list.PyList_GET_SIZE(self.body)
 * 
 *     def __len__(self) -> int:             # <<<<<<<<<<<<<<
 *         if self.body is not None:
 *             return <int> self._len()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":171
 *         return 0
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.body[i]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_8__getitem__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self), ((PyObject *)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_8__getitem__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 1);

  /* "asynctnt/iproto/response.pyx":172
 * 
 *     def __getitem__(self, i):
 *         return self.body[i]             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->body == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(5, 172, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->body, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":171
 *         return 0
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.body[i]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":174
 *         return self.body[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.body)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_11__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_10__iter__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_10__iter__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 1);

  /* "asynctnt/iproto/response.pyx":175
 * 
 *     def __iter__(self):
 *         return iter(self.body)             # <<<<<<<<<<<<<<
 * 
 * cdef object _decode_obj(const char ** p, bytes encoding):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->body;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":174
 *         return self.body[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.body)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":31
 *         uint64_t sync_
 *         int64_t schema_id_
 *         readonly str errmsg             # <<<<<<<<<<<<<<
 *         readonly IProtoError error
 *         int _rowcount
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_6errmsg_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_6errmsg_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_6errmsg___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_6errmsg___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->errmsg);
  __pyx_r = __pyx_v_self->errmsg;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":32
 *         int64_t schema_id_
 *         readonly str errmsg
 *         readonly IProtoError error             # <<<<<<<<<<<<<<
 *         int _rowcount
 *         readonly list body
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_5error_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_5error_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_5error___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_5error___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->error);
  __pyx_r = ((PyObject *)__pyx_v_self->error);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":34
 *         readonly IProtoError error
 *         int _rowcount
 *         readonly list body             # <<<<<<<<<<<<<<
 *         readonly bytes encoding
 *         readonly Metadata metadata
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_4body_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_4body_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_4body___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_4body___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->body);
  __pyx_r = __pyx_v_self->body;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":35
 *         int _rowcount
 *         readonly list body
 *         readonly bytes encoding             # <<<<<<<<<<<<<<
 *         readonly Metadata metadata
 *         readonly Metadata params
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_8encoding_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_8encoding_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_8encoding___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_8encoding___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->encoding);
  __pyx_r = __pyx_v_self->encoding;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":36
 *         readonly list body
 *         readonly bytes encoding
 *         readonly Metadata metadata             # <<<<<<<<<<<<<<
 *         readonly Metadata params
 *         readonly int params_count
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_8metadata_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_8metadata_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_8metadata___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_8metadata___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->metadata);
  __pyx_r = ((PyObject *)__pyx_v_self->metadata);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":37
 *         readonly bytes encoding
 *         readonly Metadata metadata
 *         readonly Metadata params             # <<<<<<<<<<<<<<
 *         readonly int params_count
 *         readonly list autoincrement_ids
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_6params_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_6params_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_6params___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_6params___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->params);
  __pyx_r = ((PyObject *)__pyx_v_self->params);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":38
 *         readonly Metadata metadata
 *         readonly Metadata params
 *         readonly int params_count             # <<<<<<<<<<<<<<
 *         readonly list autoincrement_ids
 *         uint64_t stmt_id_
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_12params_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_12params_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_12params_count___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_12params_count___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->params_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.params_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":39
 *         readonly Metadata params
 *         readonly int params_count
 *         readonly list autoincrement_ids             # <<<<<<<<<<<<<<
 *         uint64_t stmt_id_
 *         bint _push_subscribe
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_17autoincrement_ids_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_17autoincrement_ids_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_17autoincrement_ids___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_17autoincrement_ids___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->autoincrement_ids);
  __pyx_r = __pyx_v_self->autoincrement_ids;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":45
 *         object _exception
 * 
 *         readonly object _q             # <<<<<<<<<<<<<<
 *         readonly object _push_event
 *         object _q_append
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_2_q_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_2_q_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_2_q___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_2_q___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_q);
  __pyx_r = __pyx_v_self->_q;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pxd":46
 * 
 *         readonly object _q
 *         readonly object _push_event             # <<<<<<<<<<<<<<
 *         object _q_append
 *         object _q_popleft
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_11_push_event_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_11_push_event_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_11_push_event___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_11_push_event___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_push_event);
  __pyx_r = __pyx_v_self->_push_event;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_8Response_13__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_12__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_8Response_15__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_8Response_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8Response_14__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8Response_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Response.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":177
 *         return iter(self.body)
 * 
 * cdef object _decode_obj(const char ** p, bytes encoding):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t i
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol__decode_obj(char const **__pyx_v_p, PyObject *__pyx_v_encoding) {
  uint32_t __pyx_v_i;
  enum mp_type __pyx_v_obj_type;
  char const *__pyx_v_s;
  uint32_t __pyx_v_s_len;
  uint32_t __pyx_v_arr_size;
  uint32_t __pyx_v_map_size;
  PyObject *__pyx_v_map = 0;
  enum mp_type __pyx_v_map_key_type;
  char const *__pyx_v_map_key_str;
  uint32_t __pyx_v_map_key_len;
  PyObject *__pyx_v_map_key = 0;
  int8_t __pyx_v_ext_type;
  struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime __pyx_v_dt;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_el = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  uint32_t __pyx_t_9;
  uint32_t __pyx_t_10;
  uint32_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_decode_obj", 1);

  /* "asynctnt/iproto/response.pyx":198
 *         IProtoDateTime dt
 * 
 *     obj_type = mp_typeof(p[0][0])             # <<<<<<<<<<<<<<
 *     if obj_type == MP_UINT:
 *         return mp_decode_uint(p)
 */
  __pyx_v_obj_type = mp_typeof(((__pyx_v_p[0])[0]));

  /* "asynctnt/iproto/response.pyx":199
 * 
 *     obj_type = mp_typeof(p[0][0])
 *     if obj_type == MP_UINT:             # <<<<<<<<<<<<<<
 *         return mp_decode_uint(p)
 *     elif obj_type == MP_INT:
 */
  switch (__pyx_v_obj_type) {
    case MP_UINT:

    /* "asynctnt/iproto/response.pyx":200
 *     obj_type = mp_typeof(p[0][0])
 *     if obj_type == MP_UINT:
 *         return mp_decode_uint(p)             # <<<<<<<<<<<<<<
 *     elif obj_type == MP_INT:
 *         return mp_decode_int(p)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyInt_From_uint64_t(mp_decode_uint(__pyx_v_p)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/response.pyx":199
 * 
 *     obj_type = mp_typeof(p[0][0])
 *     if obj_type == MP_UINT:             # <<<<<<<<<<<<<<
 *         return mp_decode_uint(p)
 *     elif obj_type == MP_INT:
 */
    break;
    case MP_INT:

    /* "asynctnt/iproto/response.pyx":202
 *         return mp_decode_uint(p)
 *     elif obj_type == MP_INT:
 *         return mp_decode_int(p)             # <<<<<<<<<<<<<<
 *     elif obj_type == MP_STR:
 *         s = NULL
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyInt_From_int64_t(mp_decode_int(__pyx_v_p)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/response.pyx":201
 *     if obj_type == MP_UINT:
 *         return mp_decode_uint(p)
 *     elif obj_type == MP_INT:             # <<<<<<<<<<<<<<
 *         return mp_decode_int(p)
 *     elif obj_type == MP_STR:
 */
    break;
    case MP_STR:

    /* "asynctnt/iproto/response.pyx":204
 *         return mp_decode_int(p)
 *     elif obj_type == MP_STR:
 *         s = NULL             # <<<<<<<<<<<<<<
 *         s_len = 0
 *         s = mp_decode_str(p, &s_len)
 */
    __pyx_v_s = NULL;

    /* "asynctnt/iproto/response.pyx":205
 *     elif obj_type == MP_STR:
 *         s = NULL
 *         s_len = 0             # <<<<<<<<<<<<<<
 *         s = mp_decode_str(p, &s_len)
 *         try:
 */
    __pyx_v_s_len = 0;

    /* "asynctnt/iproto/response.pyx":206
 *         s = NULL
 *         s_len = 0
 *         s = mp_decode_str(p, &s_len)             # <<<<<<<<<<<<<<
 *         try:
 *             return decode_string(s[:s_len], encoding)
 */
    __pyx_v_s = mp_decode_str(__pyx_v_p, (&__pyx_v_s_len));

    /* "asynctnt/iproto/response.pyx":207
 *         s_len = 0
 *         s = mp_decode_str(p, &s_len)
 *         try:             # <<<<<<<<<<<<<<
 *             return decode_string(s[:s_len], encoding)
 *         except UnicodeDecodeError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "asynctnt/iproto/response.pyx":208
 *         s = mp_decode_str(p, &s_len)
 *         try:
 *             return decode_string(s[:s_len], encoding)             # <<<<<<<<<<<<<<
 *         except UnicodeDecodeError:
 *             return <bytes> s[:s_len]
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s + 0, __pyx_v_s_len - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 208, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = __pyx_f_8asynctnt_6iproto_8protocol_decode_string(((PyObject*)__pyx_t_1), __pyx_v_encoding); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 208, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_r = __pyx_t_5;
        __pyx_t_5 = 0;
        goto __pyx_L7_try_return;

        /* "asynctnt/iproto/response.pyx":207
 *         s_len = 0
 *         s = mp_decode_str(p, &s_len)
 *         try:             # <<<<<<<<<<<<<<
 *             return decode_string(s[:s_len], encoding)
 *         except UnicodeDecodeError:
 */
      }
      __pyx_L3_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "asynctnt/iproto/response.pyx":209
 *         try:
 *             return decode_string(s[:s_len], encoding)
 *         except UnicodeDecodeError:             # <<<<<<<<<<<<<<
 *             return <bytes> s[:s_len]
 *     elif obj_type == MP_BIN:
 */
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeDecodeError);
      if (__pyx_t_6) {
        __Pyx_AddTraceback("asynctnt.iproto.protocol._decode_obj", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_7) < 0) __PYX_ERR(5, 209, __pyx_L5_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_7);

        /* "asynctnt/iproto/response.pyx":210
 *             return decode_string(s[:s_len], encoding)
 *         except UnicodeDecodeError:
 *             return <bytes> s[:s_len]             # <<<<<<<<<<<<<<
 *     elif obj_type == MP_BIN:
 *         s = NULL
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_8 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s + 0, __pyx_v_s_len - 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 210, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(((PyObject*)__pyx_t_8));
        __pyx_r = __pyx_t_8;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L6_except_return;
      }
      goto __pyx_L5_except_error;

      /* "asynctnt/iproto/response.pyx":207
 *         s_len = 0
 *         s = mp_decode_str(p, &s_len)
 *         try:             # <<<<<<<<<<<<<<
 *             return decode_string(s[:s_len], encoding)
 *         except UnicodeDecodeError:
 */
      __pyx_L5_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L7_try_return:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L0;
      __pyx_L6_except_return:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L0;
    }

    /* "asynctnt/iproto/response.pyx":203
 *     elif obj_type == MP_INT:
 *         return mp_decode_int(p)
 *     elif obj_type == MP_STR:             # <<<<<<<<<<<<<<
 *         s = NULL
 *         s_len = 0
 */
    break;
    case MP_BIN:

    /* "asynctnt/iproto/response.pyx":212
 *             return <bytes> s[:s_len]
 *     elif obj_type == MP_BIN:
 *         s = NULL             # <<<<<<<<<<<<<<
 *         s_len = 0
 *         s = mp_decode_bin(p, &s_len)
 */
    __pyx_v_s = NULL;

    /* "asynctnt/iproto/response.pyx":213
 *     elif obj_type == MP_BIN:
 *         s = NULL
 *         s_len = 0             # <<<<<<<<<<<<<<
 *         s = mp_decode_bin(p, &s_len)
 *         return <bytes> s[:s_len]
 */
    __pyx_v_s_len = 0;

    /* "asynctnt/iproto/response.pyx":214
 *         s = NULL
 *         s_len = 0
 *         s = mp_decode_bin(p, &s_len)             # <<<<<<<<<<<<<<
 *         return <bytes> s[:s_len]
 *     elif obj_type == MP_BOOL:
 */
    __pyx_v_s = mp_decode_bin(__pyx_v_p, (&__pyx_v_s_len));

    /* "asynctnt/iproto/response.pyx":215
 *         s_len = 0
 *         s = mp_decode_bin(p, &s_len)
 *         return <bytes> s[:s_len]             # <<<<<<<<<<<<<<
 *     elif obj_type == MP_BOOL:
 *         return mp_decode_bool(p)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_7 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s + 0, __pyx_v_s_len - 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject*)__pyx_t_7));
    __pyx_r = __pyx_t_7;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/response.pyx":211
 *         except UnicodeDecodeError:
 *             return <bytes> s[:s_len]
 *     elif obj_type == MP_BIN:             # <<<<<<<<<<<<<<
 *         s = NULL
 *         s_len = 0
 */
    break;
    case MP_BOOL:

    /* "asynctnt/iproto/response.pyx":217
 *         return <bytes> s[:s_len]
 *     elif obj_type == MP_BOOL:
 *         return mp_decode_bool(p)             # <<<<<<<<<<<<<<
 *     elif obj_type == MP_FLOAT:
 *         return mp_decode_float(p)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_7 = __Pyx_PyBool_FromLong(mp_decode_bool(__pyx_v_p)); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/response.pyx":216
 *         s = mp_decode_bin(p, &s_len)
 *         return <bytes> s[:s_len]
 *     elif obj_type == MP_BOOL:             # <<<<<<<<<<<<<<
 *         return mp_decode_bool(p)
 *     elif obj_type == MP_FLOAT:
 */
    break;
    case MP_FLOAT:

    /* "asynctnt/iproto/response.pyx":219
 *         return mp_decode_bool(p)
 *     elif obj_type == MP_FLOAT:
 *         return mp_decode_float(p)             # <<<<<<<<<<<<<<
 *     elif obj_type == MP_DOUBLE:
 *         return mp_decode_double(p)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_7 = PyFloat_FromDouble(mp_decode_float(__pyx_v_p)); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/response.pyx":218
 *     elif obj_type == MP_BOOL:
 *         return mp_decode_bool(p)
 *     elif obj_type == MP_FLOAT:             # <<<<<<<<<<<<<<
 *         return mp_decode_float(p)
 *     elif obj_type == MP_DOUBLE:
 */
    break;
    case MP_DOUBLE:

    /* "asynctnt/iproto/response.pyx":221
 *         return mp_decode_float(p)
 *     elif obj_type == MP_DOUBLE:
 *         return mp_decode_double(p)             # <<<<<<<<<<<<<<
 *     elif obj_type == MP_ARRAY:
 *         arr_size = mp_decode_array(p)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_7 = PyFloat_FromDouble(mp_decode_double(__pyx_v_p)); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/response.pyx":220
 *     elif obj_type == MP_FLOAT:
 *         return mp_decode_float(p)
 *     elif obj_type == MP_DOUBLE:             # <<<<<<<<<<<<<<
 *         return mp_decode_double(p)
 *     elif obj_type == MP_ARRAY:
 */
    break;
    case MP_ARRAY:

    /* "asynctnt/iproto/response.pyx":223
 *         return mp_decode_double(p)
 *     elif obj_type == MP_ARRAY:
 *         arr_size = mp_decode_array(p)             # <<<<<<<<<<<<<<
 *         value = cpython.list.PyList_New(arr_size)
 *         for i in range(arr_size):
 */
    __pyx_v_arr_size = mp_decode_array(__pyx_v_p);

    /* "asynctnt/iproto/response.pyx":224
 *     elif obj_type == MP_ARRAY:
 *         arr_size = mp_decode_array(p)
 *         value = cpython.list.PyList_New(arr_size)             # <<<<<<<<<<<<<<
 *         for i in range(arr_size):
 *             el = _decode_obj(p, encoding)
 */
    __pyx_t_7 = PyList_New(__pyx_v_arr_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_v_value = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "asynctnt/iproto/response.pyx":225
 *         arr_size = mp_decode_array(p)
 *         value = cpython.list.PyList_New(arr_size)
 *         for i in range(arr_size):             # <<<<<<<<<<<<<<
 *             el = _decode_obj(p, encoding)
 *             cpython.Py_INCREF(el)
 */
    __pyx_t_9 = __pyx_v_arr_size;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "asynctnt/iproto/response.pyx":226
 *         value = cpython.list.PyList_New(arr_size)
 *         for i in range(arr_size):
 *             el = _decode_obj(p, encoding)             # <<<<<<<<<<<<<<
 *             cpython.Py_INCREF(el)
 *             cpython.list.PyList_SET_ITEM(value, i, el)
 */
      __pyx_t_7 = __pyx_f_8asynctnt_6iproto_8protocol__decode_obj(__pyx_v_p, __pyx_v_encoding); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_v_el, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "asynctnt/iproto/response.pyx":227
 *         for i in range(arr_size):
 *             el = _decode_obj(p, encoding)
 *             cpython.Py_INCREF(el)             # <<<<<<<<<<<<<<
 *             cpython.list.PyList_SET_ITEM(value, i, el)
 *         return value
 */
      Py_INCREF(__pyx_v_el);

      /* "asynctnt/iproto/response.pyx":228
 *             el = _decode_obj(p, encoding)
 *             cpython.Py_INCREF(el)
 *             cpython.list.PyList_SET_ITEM(value, i, el)             # <<<<<<<<<<<<<<
 *         return value
 *     elif obj_type == MP_MAP:
 */
      PyList_SET_ITEM(__pyx_v_value, __pyx_v_i, __pyx_v_el);
    }

    /* "asynctnt/iproto/response.pyx":229
 *             cpython.Py_INCREF(el)
 *             cpython.list.PyList_SET_ITEM(value, i, el)
 *         return value             # <<<<<<<<<<<<<<
 *     elif obj_type == MP_MAP:
 *         map = {}
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "asynctnt/iproto/response.pyx":222
 *     elif obj_type == MP_DOUBLE:
 *         return mp_decode_double(p)
 *     elif obj_type == MP_ARRAY:             # <<<<<<<<<<<<<<
 *         arr_size = mp_decode_array(p)
 *         value = cpython.list.PyList_New(arr_size)
 */
    break;
    case MP_MAP:

    /* "asynctnt/iproto/response.pyx":231
 *         return value
 *     elif obj_type == MP_MAP:
 *         map = {}             # <<<<<<<<<<<<<<
 *         map_size = mp_decode_map(p)
 *         for i in range(map_size):
 */
    __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_v_map = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "asynctnt/iproto/response.pyx":232
 *     elif obj_type == MP_MAP:
 *         map = {}
 *         map_size = mp_decode_map(p)             # <<<<<<<<<<<<<<
 *         for i in range(map_size):
 *             map_key_type = mp_typeof(p[0][0])
 */
    __pyx_v_map_size = mp_decode_map(__pyx_v_p);

    /* "asynctnt/iproto/response.pyx":233
 *         map = {}
 *         map_size = mp_decode_map(p)
 *         for i in range(map_size):             # <<<<<<<<<<<<<<
 *             map_key_type = mp_typeof(p[0][0])
 *             if map_key_type == MP_STR:
 */
    __pyx_t_9 = __pyx_v_map_size;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "asynctnt/iproto/response.pyx":234
 *         map_size = mp_decode_map(p)
 *         for i in range(map_size):
 *             map_key_type = mp_typeof(p[0][0])             # <<<<<<<<<<<<<<
 *             if map_key_type == MP_STR:
 *                 map_key_len = 0
 */
      __pyx_v_map_key_type = mp_typeof(((__pyx_v_p[0])[0]));

      /* "asynctnt/iproto/response.pyx":235
 *         for i in range(map_size):
 *             map_key_type = mp_typeof(p[0][0])
 *             if map_key_type == MP_STR:             # <<<<<<<<<<<<<<
 *                 map_key_len = 0
 *                 map_key_str = mp_decode_str(p, &map_key_len)
 */
      switch (__pyx_v_map_key_type) {
        case MP_STR:

        /* "asynctnt/iproto/response.pyx":236
 *             map_key_type = mp_typeof(p[0][0])
 *             if map_key_type == MP_STR:
 *                 map_key_len = 0             # <<<<<<<<<<<<<<
 *                 map_key_str = mp_decode_str(p, &map_key_len)
 *                 map_key = <object> (
 */
        __pyx_v_map_key_len = 0;

        /* "asynctnt/iproto/response.pyx":237
 *             if map_key_type == MP_STR:
 *                 map_key_len = 0
 *                 map_key_str = mp_decode_str(p, &map_key_len)             # <<<<<<<<<<<<<<
 *                 map_key = <object> (
 *                     decode_string(map_key_str[:map_key_len], encoding)
 */
        __pyx_v_map_key_str = mp_decode_str(__pyx_v_p, (&__pyx_v_map_key_len));

        /* "asynctnt/iproto/response.pyx":239
 *                 map_key_str = mp_decode_str(p, &map_key_len)
 *                 map_key = <object> (
 *                     decode_string(map_key_str[:map_key_len], encoding)             # <<<<<<<<<<<<<<
 *                 )
 *             elif map_key_type == MP_UINT:
 */
        __pyx_t_7 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_map_key_str + 0, __pyx_v_map_key_len - 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 239, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_decode_string(((PyObject*)__pyx_t_7), __pyx_v_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 239, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "asynctnt/iproto/response.pyx":238
 *                 map_key_len = 0
 *                 map_key_str = mp_decode_str(p, &map_key_len)
 *                 map_key = <object> (             # <<<<<<<<<<<<<<
 *                     decode_string(map_key_str[:map_key_len], encoding)
 *                 )
 */
        __pyx_t_7 = __pyx_t_1;
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_map_key, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "asynctnt/iproto/response.pyx":235
 *         for i in range(map_size):
 *             map_key_type = mp_typeof(p[0][0])
 *             if map_key_type == MP_STR:             # <<<<<<<<<<<<<<
 *                 map_key_len = 0
 *                 map_key_str = mp_decode_str(p, &map_key_len)
 */
        break;
        case MP_UINT:

        /* "asynctnt/iproto/response.pyx":242
 *                 )
 *             elif map_key_type == MP_UINT:
 *                 map_key = <object> mp_decode_uint(p)             # <<<<<<<<<<<<<<
 *             elif map_key_type == MP_INT:
 *                 map_key = <object> mp_decode_int(p)
 */
        __pyx_t_7 = __Pyx_PyInt_From_uint64_t(mp_decode_uint(__pyx_v_p)); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 242, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = __pyx_t_7;
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF_SET(__pyx_v_map_key, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "asynctnt/iproto/response.pyx":241
 *                     decode_string(map_key_str[:map_key_len], encoding)
 *                 )
 *             elif map_key_type == MP_UINT:             # <<<<<<<<<<<<<<
 *                 map_key = <object> mp_decode_uint(p)
 *             elif map_key_type == MP_INT:
 */
        break;
        case MP_INT:

        /* "asynctnt/iproto/response.pyx":244
 *                 map_key = <object> mp_decode_uint(p)
 *             elif map_key_type == MP_INT:
 *                 map_key = <object> mp_decode_int(p)             # <<<<<<<<<<<<<<
 *             elif map_key_type == MP_FLOAT:
 *                 map_key = <object> mp_decode_float(p)
 */
        __pyx_t_1 = __Pyx_PyInt_From_int64_t(mp_decode_int(__pyx_v_p)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 244, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1;
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_map_key, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "asynctnt/iproto/response.pyx":243
 *             elif map_key_type == MP_UINT:
 *                 map_key = <object> mp_decode_uint(p)
 *             elif map_key_type == MP_INT:             # <<<<<<<<<<<<<<
 *                 map_key = <object> mp_decode_int(p)
 *             elif map_key_type == MP_FLOAT:
 */
        break;
        case MP_FLOAT:

        /* "asynctnt/iproto/response.pyx":246
 *                 map_key = <object> mp_decode_int(p)
 *             elif map_key_type == MP_FLOAT:
 *                 map_key = <object> mp_decode_float(p)             # <<<<<<<<<<<<<<
 *             elif map_key_type == MP_DOUBLE:
 *                 map_key = <object> mp_decode_double(p)
 */
        __pyx_t_7 = PyFloat_FromDouble(mp_decode_float(__pyx_v_p)); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 246, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = __pyx_t_7;
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF_SET(__pyx_v_map_key, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "asynctnt/iproto/response.pyx":245
 *             elif map_key_type == MP_INT:
 *                 map_key = <object> mp_decode_int(p)
 *             elif map_key_type == MP_FLOAT:             # <<<<<<<<<<<<<<
 *                 map_key = <object> mp_decode_float(p)
 *             elif map_key_type == MP_DOUBLE:
 */
        break;
        case MP_DOUBLE:

        /* "asynctnt/iproto/response.pyx":248
 *                 map_key = <object> mp_decode_float(p)
 *             elif map_key_type == MP_DOUBLE:
 *                 map_key = <object> mp_decode_double(p)             # <<<<<<<<<<<<<<
 *             else:  # pragma: nocover
 *                 mp_next(p)  # skip current key
 */
        __pyx_t_1 = PyFloat_FromDouble(mp_decode_double(__pyx_v_p)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1;
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_map_key, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "asynctnt/iproto/response.pyx":247
 *             elif map_key_type == MP_FLOAT:
 *                 map_key = <object> mp_decode_float(p)
 *             elif map_key_type == MP_DOUBLE:             # <<<<<<<<<<<<<<
 *                 map_key = <object> mp_decode_double(p)
 *             else:  # pragma: nocover
 */
        break;
        default:

        /* "asynctnt/iproto/response.pyx":250
 *                 map_key = <object> mp_decode_double(p)
 *             else:  # pragma: nocover
 *                 mp_next(p)  # skip current key             # <<<<<<<<<<<<<<
 *                 mp_next(p)  # skip value
 *                 logger.warning('Unexpected key type in map: %s',
 */
        mp_next(__pyx_v_p);

        /* "asynctnt/iproto/response.pyx":251
 *             else:  # pragma: nocover
 *                 mp_next(p)  # skip current key
 *                 mp_next(p)  # skip value             # <<<<<<<<<<<<<<
 *                 logger.warning('Unexpected key type in map: %s',
 *                                map_key_type)
 */
        mp_next(__pyx_v_p);

        /* "asynctnt/iproto/response.pyx":252
 *                 mp_next(p)  # skip current key
 *                 mp_next(p)  # skip value
 *                 logger.warning('Unexpected key type in map: %s',             # <<<<<<<<<<<<<<
 *                                map_key_type)
 *                 continue
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 252, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warning); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 252, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynctnt/iproto/response.pyx":253
 *                 mp_next(p)  # skip value
 *                 logger.warning('Unexpected key type in map: %s',
 *                                map_key_type)             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
        __pyx_t_1 = __Pyx_PyInt_From_enum__mp_type(__pyx_v_map_key_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 253, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_kp_u_Unexpected_key_type_in_map_s, __pyx_t_1};
          __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 252, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "asynctnt/iproto/response.pyx":254
 *                 logger.warning('Unexpected key type in map: %s',
 *                                map_key_type)
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             map[map_key] = _decode_obj(p, encoding)
 */
        goto __pyx_L13_continue;
        break;
      }

      /* "asynctnt/iproto/response.pyx":256
 *                 continue
 * 
 *             map[map_key] = _decode_obj(p, encoding)             # <<<<<<<<<<<<<<
 * 
 *         return map
 */
      __pyx_t_7 = __pyx_f_8asynctnt_6iproto_8protocol__decode_obj(__pyx_v_p, __pyx_v_encoding); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely((PyDict_SetItem(__pyx_v_map, __pyx_v_map_key, __pyx_t_7) < 0))) __PYX_ERR(5, 256, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_L13_continue:;
    }

    /* "asynctnt/iproto/response.pyx":258
 *             map[map_key] = _decode_obj(p, encoding)
 * 
 *         return map             # <<<<<<<<<<<<<<
 *     elif obj_type == MP_NIL:
 *         mp_next(p)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_map);
    __pyx_r = __pyx_v_map;
    goto __pyx_L0;

    /* "asynctnt/iproto/response.pyx":230
 *             cpython.list.PyList_SET_ITEM(value, i, el)
 *         return value
 *     elif obj_type == MP_MAP:             # <<<<<<<<<<<<<<
 *         map = {}
 *         map_size = mp_decode_map(p)
 */
    break;
    case MP_NIL:

    /* "asynctnt/iproto/response.pyx":260
 *         return map
 *     elif obj_type == MP_NIL:
 *         mp_next(p)             # <<<<<<<<<<<<<<
 *         return None
 *     elif obj_type == MP_EXT:
 */
    mp_next(__pyx_v_p);

    /* "asynctnt/iproto/response.pyx":261
 *     elif obj_type == MP_NIL:
 *         mp_next(p)
 *         return None             # <<<<<<<<<<<<<<
 *     elif obj_type == MP_EXT:
 *         ext_type = 0
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynctnt/iproto/response.pyx":259
 * 
 *         return map
 *     elif obj_type == MP_NIL:             # <<<<<<<<<<<<<<
 *         mp_next(p)
 *         return None
 */
    break;
    case MP_EXT:

    /* "asynctnt/iproto/response.pyx":263
 *         return None
 *     elif obj_type == MP_EXT:
 *         ext_type = 0             # <<<<<<<<<<<<<<
 *         s_len = mp_decode_extl(p, &ext_type)
 *         if ext_type == tarantool.MP_DECIMAL:
 */
    __pyx_v_ext_type = 0;

    /* "asynctnt/iproto/response.pyx":264
 *     elif obj_type == MP_EXT:
 *         ext_type = 0
 *         s_len = mp_decode_extl(p, &ext_type)             # <<<<<<<<<<<<<<
 *         if ext_type == tarantool.MP_DECIMAL:
 *             return decimal_decode(p, s_len)
 */
    __pyx_v_s_len = mp_decode_extl(__pyx_v_p, (&__pyx_v_ext_type));

    /* "asynctnt/iproto/response.pyx":265
 *         ext_type = 0
 *         s_len = mp_decode_extl(p, &ext_type)
 *         if ext_type == tarantool.MP_DECIMAL:             # <<<<<<<<<<<<<<
 *             return decimal_decode(p, s_len)
 *         elif ext_type == tarantool.MP_UUID:
 */
    switch (__pyx_v_ext_type) {
      case __pyx_e_8asynctnt_6iproto_9tarantool_MP_DECIMAL:

      /* "asynctnt/iproto/response.pyx":266
 *         s_len = mp_decode_extl(p, &ext_type)
 *         if ext_type == tarantool.MP_DECIMAL:
 *             return decimal_decode(p, s_len)             # <<<<<<<<<<<<<<
 *         elif ext_type == tarantool.MP_UUID:
 *             return uuid_decode(p, s_len)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = __pyx_f_8asynctnt_6iproto_8protocol_decimal_decode(__pyx_v_p, __pyx_v_s_len); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 266, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      goto __pyx_L0;

      /* "asynctnt/iproto/response.pyx":265
 *         ext_type = 0
 *         s_len = mp_decode_extl(p, &ext_type)
 *         if ext_type == tarantool.MP_DECIMAL:             # <<<<<<<<<<<<<<
 *             return decimal_decode(p, s_len)
 *         elif ext_type == tarantool.MP_UUID:
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_MP_UUID:

      /* "asynctnt/iproto/response.pyx":268
 *             return decimal_decode(p, s_len)
 *         elif ext_type == tarantool.MP_UUID:
 *             return uuid_decode(p, s_len)             # <<<<<<<<<<<<<<
 *         elif ext_type == tarantool.MP_ERROR:
 *             return parse_iproto_error(p, encoding)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = __pyx_f_8asynctnt_6iproto_8protocol_uuid_decode(__pyx_v_p, __pyx_v_s_len); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      goto __pyx_L0;

      /* "asynctnt/iproto/response.pyx":267
 *         if ext_type == tarantool.MP_DECIMAL:
 *             return decimal_decode(p, s_len)
 *         elif ext_type == tarantool.MP_UUID:             # <<<<<<<<<<<<<<
 *             return uuid_decode(p, s_len)
 *         elif ext_type == tarantool.MP_ERROR:
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR:

      /* "asynctnt/iproto/response.pyx":270
 *             return uuid_decode(p, s_len)
 *         elif ext_type == tarantool.MP_ERROR:
 *             return parse_iproto_error(p, encoding)             # <<<<<<<<<<<<<<
 *         elif ext_type == tarantool.MP_DATETIME:
 *             datetime_zero(&dt)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_parse_iproto_error(__pyx_v_p, __pyx_v_encoding)); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      goto __pyx_L0;

      /* "asynctnt/iproto/response.pyx":269
 *         elif ext_type == tarantool.MP_UUID:
 *             return uuid_decode(p, s_len)
 *         elif ext_type == tarantool.MP_ERROR:             # <<<<<<<<<<<<<<
 *             return parse_iproto_error(p, encoding)
 *         elif ext_type == tarantool.MP_DATETIME:
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_MP_DATETIME:

      /* "asynctnt/iproto/response.pyx":272
 *             return parse_iproto_error(p, encoding)
 *         elif ext_type == tarantool.MP_DATETIME:
 *             datetime_zero(&dt)             # <<<<<<<<<<<<<<
 *             datetime_decode(p, s_len, &dt)
 *             return datetime_to_py(&dt)
 */
      __pyx_f_8asynctnt_6iproto_8protocol_datetime_zero((&__pyx_v_dt)); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 272, __pyx_L1_error)

      /* "asynctnt/iproto/response.pyx":273
 *         elif ext_type == tarantool.MP_DATETIME:
 *             datetime_zero(&dt)
 *             datetime_decode(p, s_len, &dt)             # <<<<<<<<<<<<<<
 *             return datetime_to_py(&dt)
 *         else:  # pragma: nocover
 */
      __pyx_t_6 = __pyx_f_8asynctnt_6iproto_8protocol_datetime_decode(__pyx_v_p, __pyx_v_s_len, (&__pyx_v_dt)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(5, 273, __pyx_L1_error)

      /* "asynctnt/iproto/response.pyx":274
 *             datetime_zero(&dt)
 *             datetime_decode(p, s_len, &dt)
 *             return datetime_to_py(&dt)             # <<<<<<<<<<<<<<
 *         else:  # pragma: nocover
 *             logger.warning('Unexpected ext type: %d', ext_type)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = __pyx_f_8asynctnt_6iproto_8protocol_datetime_to_py((&__pyx_v_dt)); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      goto __pyx_L0;

      /* "asynctnt/iproto/response.pyx":271
 *         elif ext_type == tarantool.MP_ERROR:
 *             return parse_iproto_error(p, encoding)
 *         elif ext_type == tarantool.MP_DATETIME:             # <<<<<<<<<<<<<<
 *             datetime_zero(&dt)
 *             datetime_decode(p, s_len, &dt)
 */
      break;
      default:

      /* "asynctnt/iproto/response.pyx":276
 *             return datetime_to_py(&dt)
 *         else:  # pragma: nocover
 *             logger.warning('Unexpected ext type: %d', ext_type)             # <<<<<<<<<<<<<<
 *             p += s_len  # skip unknown ext
 *             return None
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_logger); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_warning); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyInt_From_int8_t(__pyx_v_ext_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_kp_u_Unexpected_ext_type_d, __pyx_t_5};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 276, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "asynctnt/iproto/response.pyx":277
 *         else:  # pragma: nocover
 *             logger.warning('Unexpected ext type: %d', ext_type)
 *             p += s_len  # skip unknown ext             # <<<<<<<<<<<<<<
 *             return None
 *     else:  # pragma: nocover
 */
      __pyx_v_p = (__pyx_v_p + __pyx_v_s_len);

      /* "asynctnt/iproto/response.pyx":278
 *             logger.warning('Unexpected ext type: %d', ext_type)
 *             p += s_len  # skip unknown ext
 *             return None             # <<<<<<<<<<<<<<
 *     else:  # pragma: nocover
 *         mp_next(p)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;
      break;
    }

    /* "asynctnt/iproto/response.pyx":262
 *         mp_next(p)
 *         return None
 *     elif obj_type == MP_EXT:             # <<<<<<<<<<<<<<
 *         ext_type = 0
 *         s_len = mp_decode_extl(p, &ext_type)
 */
    break;
    default:

    /* "asynctnt/iproto/response.pyx":280
 *             return None
 *     else:  # pragma: nocover
 *         mp_next(p)             # <<<<<<<<<<<<<<
 *         logger.warning('Unexpected obj type: %s', obj_type)
 *         return None
 */
    mp_next(__pyx_v_p);

    /* "asynctnt/iproto/response.pyx":281
 *     else:  # pragma: nocover
 *         mp_next(p)
 *         logger.warning('Unexpected obj type: %s', obj_type)             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warning); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_From_enum__mp_type(__pyx_v_obj_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_kp_u_Unexpected_obj_type_s, __pyx_t_1};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "asynctnt/iproto/response.pyx":282
 *         mp_next(p)
 *         logger.warning('Unexpected obj type: %s', obj_type)
 *         return None             # <<<<<<<<<<<<<<
 * 
 * cdef list _response_parse_body_data(const char ** b,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
    break;
  }

  /* "asynctnt/iproto/response.pyx":177
 *         return iter(self.body)
 * 
 * cdef object _decode_obj(const char ** p, bytes encoding):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynctnt.iproto.protocol._decode_obj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_map);
  __Pyx_XDECREF(__pyx_v_map_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_el);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":284
 *         return None
 * 
 * cdef list _response_parse_body_data(const char ** b,             # <<<<<<<<<<<<<<
 *                                     Response resp, BaseRequest req):
 *     cdef:
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol__response_parse_body_data(char const **__pyx_v_b, struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_resp, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req) {
  uint32_t __pyx_v_size;
  uint32_t __pyx_v_tuple_size;
  PyObject *__pyx_v_tuples = 0;
  uint32_t __pyx_v_i;
  struct C_Metadata *__pyx_v_metadata = 0;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  uint32_t __pyx_t_3;
  uint32_t __pyx_t_4;
  uint32_t __pyx_t_5;
  uint32_t __pyx_t_6;
  uint32_t __pyx_t_7;
  uint32_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_response_parse_body_data", 1);

  /* "asynctnt/iproto/response.pyx":294
 *         Metadata metadata
 * 
 *     size = mp_decode_array(b)             # <<<<<<<<<<<<<<
 *     tuples = []
 * 
 */
  __pyx_v_size = mp_decode_array(__pyx_v_b);

  /* "asynctnt/iproto/response.pyx":295
 * 
 *     size = mp_decode_array(b)
 *     tuples = []             # <<<<<<<<<<<<<<
 * 
 *     if req.parse_as_tuples:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tuples = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/response.pyx":297
 *     tuples = []
 * 
 *     if req.parse_as_tuples:             # <<<<<<<<<<<<<<
 *         # decode as TarantoolTuples
 * 
 */
  if (__pyx_v_req->parse_as_tuples) {

    /* "asynctnt/iproto/response.pyx":300
 *         # decode as TarantoolTuples
 * 
 *         metadata = resp.metadata             # <<<<<<<<<<<<<<
 *         if metadata is None:
 *             metadata = req.metadata()
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_resp->metadata);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_metadata = ((struct C_Metadata *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/response.pyx":301
 * 
 *         metadata = resp.metadata
 *         if metadata is None:             # <<<<<<<<<<<<<<
 *             metadata = req.metadata()
 * 
 */
    __pyx_t_2 = (((PyObject *)__pyx_v_metadata) == Py_None);
    if (__pyx_t_2) {

      /* "asynctnt/iproto/response.pyx":302
 *         metadata = resp.metadata
 *         if metadata is None:
 *             metadata = req.metadata()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(size):
 */
      __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_metadata(__pyx_v_req)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_metadata, ((struct C_Metadata *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "asynctnt/iproto/response.pyx":301
 * 
 *         metadata = resp.metadata
 *         if metadata is None:             # <<<<<<<<<<<<<<
 *             metadata = req.metadata()
 * 
 */
    }

    /* "asynctnt/iproto/response.pyx":304
 *             metadata = req.metadata()
 * 
 *         for i in range(size):             # <<<<<<<<<<<<<<
 *             if mp_typeof(b[0][0]) != MP_ARRAY:  # pragma: nocover
 *                 raise TypeError(
 */
    __pyx_t_3 = __pyx_v_size;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "asynctnt/iproto/response.pyx":305
 * 
 *         for i in range(size):
 *             if mp_typeof(b[0][0]) != MP_ARRAY:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError(
 *                     'Tuple must be an array when decoding as TarantoolTuple'
 */
      __pyx_t_2 = (mp_typeof(((__pyx_v_b[0])[0])) != MP_ARRAY);
      if (unlikely(__pyx_t_2)) {

        /* "asynctnt/iproto/response.pyx":306
 *         for i in range(size):
 *             if mp_typeof(b[0][0]) != MP_ARRAY:  # pragma: nocover
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     'Tuple must be an array when decoding as TarantoolTuple'
 *                 )
 */
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(5, 306, __pyx_L1_error)

        /* "asynctnt/iproto/response.pyx":305
 * 
 *         for i in range(size):
 *             if mp_typeof(b[0][0]) != MP_ARRAY:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError(
 *                     'Tuple must be an array when decoding as TarantoolTuple'
 */
      }

      /* "asynctnt/iproto/response.pyx":310
 *                 )
 * 
 *             tuple_size = mp_decode_array(b)             # <<<<<<<<<<<<<<
 *             t = tupleobj.AtntTuple_New(metadata, <int> tuple_size)
 *             for i in range(tuple_size):
 */
      __pyx_v_tuple_size = mp_decode_array(__pyx_v_b);

      /* "asynctnt/iproto/response.pyx":311
 * 
 *             tuple_size = mp_decode_array(b)
 *             t = tupleobj.AtntTuple_New(metadata, <int> tuple_size)             # <<<<<<<<<<<<<<
 *             for i in range(tuple_size):
 *                 value = _decode_obj(b, resp.encoding)
 */
      __pyx_t_1 = AtntTuple_New(((PyObject *)__pyx_v_metadata), ((int)__pyx_v_tuple_size)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "asynctnt/iproto/response.pyx":312
 *             tuple_size = mp_decode_array(b)
 *             t = tupleobj.AtntTuple_New(metadata, <int> tuple_size)
 *             for i in range(tuple_size):             # <<<<<<<<<<<<<<
 *                 value = _decode_obj(b, resp.encoding)
 *                 cpython.Py_INCREF(value)
 */
      __pyx_t_6 = __pyx_v_tuple_size;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "asynctnt/iproto/response.pyx":313
 *             t = tupleobj.AtntTuple_New(metadata, <int> tuple_size)
 *             for i in range(tuple_size):
 *                 value = _decode_obj(b, resp.encoding)             # <<<<<<<<<<<<<<
 *                 cpython.Py_INCREF(value)
 *                 tupleobj.AtntTuple_SET_ITEM(t, i, value)
 */
        __pyx_t_1 = __pyx_v_resp->encoding;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_9 = __pyx_f_8asynctnt_6iproto_8protocol__decode_obj(__pyx_v_b, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 313, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_9);
        __pyx_t_9 = 0;

        /* "asynctnt/iproto/response.pyx":314
 *             for i in range(tuple_size):
 *                 value = _decode_obj(b, resp.encoding)
 *                 cpython.Py_INCREF(value)             # <<<<<<<<<<<<<<
 *                 tupleobj.AtntTuple_SET_ITEM(t, i, value)
 * 
 */
        Py_INCREF(__pyx_v_value);

        /* "asynctnt/iproto/response.pyx":315
 *                 value = _decode_obj(b, resp.encoding)
 *                 cpython.Py_INCREF(value)
 *                 tupleobj.AtntTuple_SET_ITEM(t, i, value)             # <<<<<<<<<<<<<<
 * 
 *             tuples.append(t)
 */
        AtntTuple_SET_ITEM(__pyx_v_t, __pyx_v_i, __pyx_v_value);
      }

      /* "asynctnt/iproto/response.pyx":317
 *                 tupleobj.AtntTuple_SET_ITEM(t, i, value)
 * 
 *             tuples.append(t)             # <<<<<<<<<<<<<<
 *     else:
 *         # decode as raw objects
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_tuples, __pyx_v_t); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(5, 317, __pyx_L1_error)
    }

    /* "asynctnt/iproto/response.pyx":297
 *     tuples = []
 * 
 *     if req.parse_as_tuples:             # <<<<<<<<<<<<<<
 *         # decode as TarantoolTuples
 * 
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/response.pyx":320
 *     else:
 *         # decode as raw objects
 *         for i in range(size):             # <<<<<<<<<<<<<<
 *             tuples.append(_decode_obj(b, resp.encoding))
 * 
 */
  /*else*/ {
    __pyx_t_3 = __pyx_v_size;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "asynctnt/iproto/response.pyx":321
 *         # decode as raw objects
 *         for i in range(size):
 *             tuples.append(_decode_obj(b, resp.encoding))             # <<<<<<<<<<<<<<
 * 
 *     return tuples
 */
      __pyx_t_9 = __pyx_v_resp->encoding;
      __Pyx_INCREF(__pyx_t_9);
      __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol__decode_obj(__pyx_v_b, ((PyObject*)__pyx_t_9)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 321, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_tuples, __pyx_t_1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(5, 321, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
  }
  __pyx_L3:;

  /* "asynctnt/iproto/response.pyx":323
 *             tuples.append(_decode_obj(b, resp.encoding))
 * 
 *     return tuples             # <<<<<<<<<<<<<<
 * 
 * cdef ssize_t response_parse_header(const char *buf, uint32_t buf_len,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tuples);
  __pyx_r = __pyx_v_tuples;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":284
 *         return None
 * 
 * cdef list _response_parse_body_data(const char ** b,             # <<<<<<<<<<<<<<
 *                                     Response resp, BaseRequest req):
 *     cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol._response_parse_body_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tuples);
  __Pyx_XDECREF((PyObject *)__pyx_v_metadata);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":325
 *     return tuples
 * 
 * cdef ssize_t response_parse_header(const char *buf, uint32_t buf_len,             # <<<<<<<<<<<<<<
 *                                    Header *hdr) except -1:
 *     cdef:
 */

static Py_ssize_t __pyx_f_8asynctnt_6iproto_8protocol_response_parse_header(char const *__pyx_v_buf, CYTHON_UNUSED uint32_t __pyx_v_buf_len, struct __pyx_t_8asynctnt_6iproto_8protocol_Header *__pyx_v_hdr) {
  char const *__pyx_v_b;
  uint32_t __pyx_v_size;
  uint32_t __pyx_v_key;
  CYTHON_UNUSED uint32_t __pyx_v__;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  uint32_t __pyx_t_3;
  uint32_t __pyx_t_4;
  uint32_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("response_parse_header", 1);

  /* "asynctnt/iproto/response.pyx":332
 *         uint32_t key
 * 
 *     b = <const char *> buf             # <<<<<<<<<<<<<<
 *     # mp_fprint(stdio.stdout, b)
 *     # stdio.fprintf(stdio.stdout, "\n")
 */
  __pyx_v_b = ((char const *)__pyx_v_buf);

  /* "asynctnt/iproto/response.pyx":336
 *     # stdio.fprintf(stdio.stdout, "\n")
 * 
 *     if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover             # <<<<<<<<<<<<<<
 *         raise TypeError('Response header must be a MP_MAP')
 * 
 */
  __pyx_t_1 = (mp_typeof((__pyx_v_b[0])) != MP_MAP);
  if (unlikely(__pyx_t_1)) {

    /* "asynctnt/iproto/response.pyx":337
 * 
 *     if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover
 *         raise TypeError('Response header must be a MP_MAP')             # <<<<<<<<<<<<<<
 * 
 *     # parsing header
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(5, 337, __pyx_L1_error)

    /* "asynctnt/iproto/response.pyx":336
 *     # stdio.fprintf(stdio.stdout, "\n")
 * 
 *     if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover             # <<<<<<<<<<<<<<
 *         raise TypeError('Response header must be a MP_MAP')
 * 
 */
  }

  /* "asynctnt/iproto/response.pyx":340
 * 
 *     # parsing header
 *     size = mp_decode_map(&b)             # <<<<<<<<<<<<<<
 *     for _ in range(size):
 *         if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 */
  __pyx_v_size = mp_decode_map((&__pyx_v_b));

  /* "asynctnt/iproto/response.pyx":341
 *     # parsing header
 *     size = mp_decode_map(&b)
 *     for _ in range(size):             # <<<<<<<<<<<<<<
 *         if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *             raise TypeError('Header key must be a MP_UINT')
 */
  __pyx_t_3 = __pyx_v_size;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v__ = __pyx_t_5;

    /* "asynctnt/iproto/response.pyx":342
 *     size = mp_decode_map(&b)
 *     for _ in range(size):
 *         if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *             raise TypeError('Header key must be a MP_UINT')
 * 
 */
    __pyx_t_1 = (mp_typeof((__pyx_v_b[0])) != MP_UINT);
    if (unlikely(__pyx_t_1)) {

      /* "asynctnt/iproto/response.pyx":343
 *     for _ in range(size):
 *         if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *             raise TypeError('Header key must be a MP_UINT')             # <<<<<<<<<<<<<<
 * 
 *         key = mp_decode_uint(&b)
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 343, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(5, 343, __pyx_L1_error)

      /* "asynctnt/iproto/response.pyx":342
 *     size = mp_decode_map(&b)
 *     for _ in range(size):
 *         if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *             raise TypeError('Header key must be a MP_UINT')
 * 
 */
    }

    /* "asynctnt/iproto/response.pyx":345
 *             raise TypeError('Header key must be a MP_UINT')
 * 
 *         key = mp_decode_uint(&b)             # <<<<<<<<<<<<<<
 *         if key == tarantool.IPROTO_REQUEST_TYPE:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 */
    __pyx_v_key = mp_decode_uint((&__pyx_v_b));

    /* "asynctnt/iproto/response.pyx":346
 * 
 *         key = mp_decode_uint(&b)
 *         if key == tarantool.IPROTO_REQUEST_TYPE:             # <<<<<<<<<<<<<<
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError('code type must be a MP_UINT')
 */
    switch (__pyx_v_key) {
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_REQUEST_TYPE:

      /* "asynctnt/iproto/response.pyx":347
 *         key = mp_decode_uint(&b)
 *         if key == tarantool.IPROTO_REQUEST_TYPE:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('code type must be a MP_UINT')
 * 
 */
      __pyx_t_1 = (mp_typeof((__pyx_v_b[0])) != MP_UINT);
      if (unlikely(__pyx_t_1)) {

        /* "asynctnt/iproto/response.pyx":348
 *         if key == tarantool.IPROTO_REQUEST_TYPE:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError('code type must be a MP_UINT')             # <<<<<<<<<<<<<<
 * 
 *             hdr.code = <uint32_t> mp_decode_uint(&b)
 */
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(5, 348, __pyx_L1_error)

        /* "asynctnt/iproto/response.pyx":347
 *         key = mp_decode_uint(&b)
 *         if key == tarantool.IPROTO_REQUEST_TYPE:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('code type must be a MP_UINT')
 * 
 */
      }

      /* "asynctnt/iproto/response.pyx":350
 *                 raise TypeError('code type must be a MP_UINT')
 * 
 *             hdr.code = <uint32_t> mp_decode_uint(&b)             # <<<<<<<<<<<<<<
 *             hdr.return_code = hdr.code & 0x7FFF
 *         elif key == tarantool.IPROTO_SYNC:
 */
      __pyx_v_hdr->code = ((uint32_t)mp_decode_uint((&__pyx_v_b)));

      /* "asynctnt/iproto/response.pyx":351
 * 
 *             hdr.code = <uint32_t> mp_decode_uint(&b)
 *             hdr.return_code = hdr.code & 0x7FFF             # <<<<<<<<<<<<<<
 *         elif key == tarantool.IPROTO_SYNC:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 */
      __pyx_v_hdr->return_code = (__pyx_v_hdr->code & 0x7FFF);

      /* "asynctnt/iproto/response.pyx":346
 * 
 *         key = mp_decode_uint(&b)
 *         if key == tarantool.IPROTO_REQUEST_TYPE:             # <<<<<<<<<<<<<<
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError('code type must be a MP_UINT')
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SYNC:

      /* "asynctnt/iproto/response.pyx":353
 *             hdr.return_code = hdr.code & 0x7FFF
 *         elif key == tarantool.IPROTO_SYNC:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('sync type must be a MP_UINT')
 * 
 */
      __pyx_t_1 = (mp_typeof((__pyx_v_b[0])) != MP_UINT);
      if (unlikely(__pyx_t_1)) {

        /* "asynctnt/iproto/response.pyx":354
 *         elif key == tarantool.IPROTO_SYNC:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError('sync type must be a MP_UINT')             # <<<<<<<<<<<<<<
 * 
 *             hdr.sync = mp_decode_uint(&b)
 */
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(5, 354, __pyx_L1_error)

        /* "asynctnt/iproto/response.pyx":353
 *             hdr.return_code = hdr.code & 0x7FFF
 *         elif key == tarantool.IPROTO_SYNC:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('sync type must be a MP_UINT')
 * 
 */
      }

      /* "asynctnt/iproto/response.pyx":356
 *                 raise TypeError('sync type must be a MP_UINT')
 * 
 *             hdr.sync = mp_decode_uint(&b)             # <<<<<<<<<<<<<<
 *         elif key == tarantool.IPROTO_SCHEMA_VERSION:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 */
      __pyx_v_hdr->sync = mp_decode_uint((&__pyx_v_b));

      /* "asynctnt/iproto/response.pyx":352
 *             hdr.code = <uint32_t> mp_decode_uint(&b)
 *             hdr.return_code = hdr.code & 0x7FFF
 *         elif key == tarantool.IPROTO_SYNC:             # <<<<<<<<<<<<<<
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError('sync type must be a MP_UINT')
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SCHEMA_VERSION:

      /* "asynctnt/iproto/response.pyx":358
 *             hdr.sync = mp_decode_uint(&b)
 *         elif key == tarantool.IPROTO_SCHEMA_VERSION:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('schema_id type must be a MP_UINT')
 * 
 */
      __pyx_t_1 = (mp_typeof((__pyx_v_b[0])) != MP_UINT);
      if (unlikely(__pyx_t_1)) {

        /* "asynctnt/iproto/response.pyx":359
 *         elif key == tarantool.IPROTO_SCHEMA_VERSION:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError('schema_id type must be a MP_UINT')             # <<<<<<<<<<<<<<
 * 
 *             hdr.schema_id = mp_decode_uint(&b)
 */
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 359, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(5, 359, __pyx_L1_error)

        /* "asynctnt/iproto/response.pyx":358
 *             hdr.sync = mp_decode_uint(&b)
 *         elif key == tarantool.IPROTO_SCHEMA_VERSION:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('schema_id type must be a MP_UINT')
 * 
 */
      }

      /* "asynctnt/iproto/response.pyx":361
 *                 raise TypeError('schema_id type must be a MP_UINT')
 * 
 *             hdr.schema_id = mp_decode_uint(&b)             # <<<<<<<<<<<<<<
 *         else:  # pragma: nocover
 *             logger.warning(
 */
      __pyx_v_hdr->schema_id = mp_decode_uint((&__pyx_v_b));

      /* "asynctnt/iproto/response.pyx":357
 * 
 *             hdr.sync = mp_decode_uint(&b)
 *         elif key == tarantool.IPROTO_SCHEMA_VERSION:             # <<<<<<<<<<<<<<
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError('schema_id type must be a MP_UINT')
 */
      break;
      default:

      /* "asynctnt/iproto/response.pyx":363
 *             hdr.schema_id = mp_decode_uint(&b)
 *         else:  # pragma: nocover
 *             logger.warning(             # <<<<<<<<<<<<<<
 *                 'Unknown key with code \'%d\' in header. Skipping.', key)
 *             mp_next(&b)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_logger); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_warning); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "asynctnt/iproto/response.pyx":364
 *         else:  # pragma: nocover
 *             logger.warning(
 *                 'Unknown key with code \'%d\' in header. Skipping.', key)             # <<<<<<<<<<<<<<
 *             mp_next(&b)
 * 
 */
      __pyx_t_6 = __Pyx_PyInt_From_uint32_t(__pyx_v_key); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 364, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_kp_u_Unknown_key_with_code_d_in_heade, __pyx_t_6};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 2+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "asynctnt/iproto/response.pyx":365
 *             logger.warning(
 *                 'Unknown key with code \'%d\' in header. Skipping.', key)
 *             mp_next(&b)             # <<<<<<<<<<<<<<
 * 
 *     return <ssize_t> (b - buf)
 */
      mp_next((&__pyx_v_b));
      break;
    }
  }

  /* "asynctnt/iproto/response.pyx":367
 *             mp_next(&b)
 * 
 *     return <ssize_t> (b - buf)             # <<<<<<<<<<<<<<
 * 
 * cdef Metadata response_parse_metadata(const char ** b, bytes encoding):
 */
  __pyx_r = ((Py_ssize_t)(__pyx_v_b - __pyx_v_buf));
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":325
 *     return tuples
 * 
 * cdef ssize_t response_parse_header(const char *buf, uint32_t buf_len,             # <<<<<<<<<<<<<<
 *                                    Header *hdr) except -1:
 *     cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.response_parse_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":369
 *     return <ssize_t> (b - buf)
 * 
 * cdef Metadata response_parse_metadata(const char ** b, bytes encoding):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t arr_size
 */

static struct C_Metadata *__pyx_f_8asynctnt_6iproto_8protocol_response_parse_metadata(char const **__pyx_v_b, PyObject *__pyx_v_encoding) {
  uint32_t __pyx_v_arr_size;
  uint32_t __pyx_v_field_map_size;
  uint32_t __pyx_v_key;
  uint32_t __pyx_v_s_len;
  uint32_t __pyx_v_i;
  char const *__pyx_v_s;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *__pyx_v_field = 0;
  struct C_Metadata *__pyx_v_metadata = 0;
  uint32_t __pyx_v_field_id;
  CYTHON_UNUSED uint32_t __pyx_v__;
  struct C_Metadata *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  uint32_t __pyx_t_3;
  uint32_t __pyx_t_4;
  uint32_t __pyx_t_5;
  int __pyx_t_6;
  uint32_t __pyx_t_7;
  uint32_t __pyx_t_8;
  uint32_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("response_parse_metadata", 1);

  /* "asynctnt/iproto/response.pyx":380
 *         Metadata metadata
 * 
 *     metadata = <Metadata> Metadata.__new__(Metadata)             # <<<<<<<<<<<<<<
 *     arr_size = mp_decode_array(b)
 *     for i in range(arr_size):
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Metadata(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 380, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_metadata = ((struct C_Metadata *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/response.pyx":381
 * 
 *     metadata = <Metadata> Metadata.__new__(Metadata)
 *     arr_size = mp_decode_array(b)             # <<<<<<<<<<<<<<
 *     for i in range(arr_size):
 *         field_map_size = mp_decode_map(b)
 */
  __pyx_v_arr_size = mp_decode_array(__pyx_v_b);

  /* "asynctnt/iproto/response.pyx":382
 *     metadata = <Metadata> Metadata.__new__(Metadata)
 *     arr_size = mp_decode_array(b)
 *     for i in range(arr_size):             # <<<<<<<<<<<<<<
 *         field_map_size = mp_decode_map(b)
 *         if field_map_size == 0:  # pragma: nocover
 */
  __pyx_t_3 = __pyx_v_arr_size;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "asynctnt/iproto/response.pyx":383
 *     arr_size = mp_decode_array(b)
 *     for i in range(arr_size):
 *         field_map_size = mp_decode_map(b)             # <<<<<<<<<<<<<<
 *         if field_map_size == 0:  # pragma: nocover
 *             raise RuntimeError('Field map must contain at least '
 */
    __pyx_v_field_map_size = mp_decode_map(__pyx_v_b);

    /* "asynctnt/iproto/response.pyx":384
 *     for i in range(arr_size):
 *         field_map_size = mp_decode_map(b)
 *         if field_map_size == 0:  # pragma: nocover             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Field map must contain at least '
 *                                '1 element - field_name')
 */
    __pyx_t_6 = (__pyx_v_field_map_size == 0);
    if (unlikely(__pyx_t_6)) {

      /* "asynctnt/iproto/response.pyx":385
 *         field_map_size = mp_decode_map(b)
 *         if field_map_size == 0:  # pragma: nocover
 *             raise RuntimeError('Field map must contain at least '             # <<<<<<<<<<<<<<
 *                                '1 element - field_name')
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 385, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(5, 385, __pyx_L1_error)

      /* "asynctnt/iproto/response.pyx":384
 *     for i in range(arr_size):
 *         field_map_size = mp_decode_map(b)
 *         if field_map_size == 0:  # pragma: nocover             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Field map must contain at least '
 *                                '1 element - field_name')
 */
    }

    /* "asynctnt/iproto/response.pyx":388
 *                                '1 element - field_name')
 * 
 *         field = <Field> Field.__new__(Field)             # <<<<<<<<<<<<<<
 *         field_id = i
 *         for _ in range(field_map_size):
 */
    __pyx_t_2 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Field(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Field), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 388, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
    __pyx_t_1 = ((PyObject *)__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF((PyObject *)__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/response.pyx":389
 * 
 *         field = <Field> Field.__new__(Field)
 *         field_id = i             # <<<<<<<<<<<<<<
 *         for _ in range(field_map_size):
 *             key = mp_decode_uint(b)
 */
    __pyx_v_field_id = __pyx_v_i;

    /* "asynctnt/iproto/response.pyx":390
 *         field = <Field> Field.__new__(Field)
 *         field_id = i
 *         for _ in range(field_map_size):             # <<<<<<<<<<<<<<
 *             key = mp_decode_uint(b)
 *             if key == tarantool.IPROTO_FIELD_NAME:
 */
    __pyx_t_7 = __pyx_v_field_map_size;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v__ = __pyx_t_9;

      /* "asynctnt/iproto/response.pyx":391
 *         field_id = i
 *         for _ in range(field_map_size):
 *             key = mp_decode_uint(b)             # <<<<<<<<<<<<<<
 *             if key == tarantool.IPROTO_FIELD_NAME:
 *                 s = NULL
 */
      __pyx_v_key = mp_decode_uint(__pyx_v_b);

      /* "asynctnt/iproto/response.pyx":392
 *         for _ in range(field_map_size):
 *             key = mp_decode_uint(b)
 *             if key == tarantool.IPROTO_FIELD_NAME:             # <<<<<<<<<<<<<<
 *                 s = NULL
 *                 s_len = 0
 */
      switch (__pyx_v_key) {
        case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_NAME:

        /* "asynctnt/iproto/response.pyx":393
 *             key = mp_decode_uint(b)
 *             if key == tarantool.IPROTO_FIELD_NAME:
 *                 s = NULL             # <<<<<<<<<<<<<<
 *                 s_len = 0
 *                 s = mp_decode_str(b, &s_len)
 */
        __pyx_v_s = NULL;

        /* "asynctnt/iproto/response.pyx":394
 *             if key == tarantool.IPROTO_FIELD_NAME:
 *                 s = NULL
 *                 s_len = 0             # <<<<<<<<<<<<<<
 *                 s = mp_decode_str(b, &s_len)
 *                 field.name = \
 */
        __pyx_v_s_len = 0;

        /* "asynctnt/iproto/response.pyx":395
 *                 s = NULL
 *                 s_len = 0
 *                 s = mp_decode_str(b, &s_len)             # <<<<<<<<<<<<<<
 *                 field.name = \
 *                     decode_string(s[:s_len], encoding)
 */
        __pyx_v_s = mp_decode_str(__pyx_v_b, (&__pyx_v_s_len));

        /* "asynctnt/iproto/response.pyx":397
 *                 s = mp_decode_str(b, &s_len)
 *                 field.name = \
 *                     decode_string(s[:s_len], encoding)             # <<<<<<<<<<<<<<
 * 
 *             elif key == tarantool.IPROTO_FIELD_TYPE:
 */
        __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s + 0, __pyx_v_s_len - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_decode_string(((PyObject*)__pyx_t_1), __pyx_v_encoding); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynctnt/iproto/response.pyx":396
 *                 s_len = 0
 *                 s = mp_decode_str(b, &s_len)
 *                 field.name = \             # <<<<<<<<<<<<<<
 *                     decode_string(s[:s_len], encoding)
 * 
 */
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_v_field->name);
        __Pyx_DECREF(__pyx_v_field->name);
        __pyx_v_field->name = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;

        /* "asynctnt/iproto/response.pyx":392
 *         for _ in range(field_map_size):
 *             key = mp_decode_uint(b)
 *             if key == tarantool.IPROTO_FIELD_NAME:             # <<<<<<<<<<<<<<
 *                 s = NULL
 *                 s_len = 0
 */
        break;
        case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_TYPE:

        /* "asynctnt/iproto/response.pyx":400
 * 
 *             elif key == tarantool.IPROTO_FIELD_TYPE:
 *                 s = NULL             # <<<<<<<<<<<<<<
 *                 s_len = 0
 *                 s = mp_decode_str(b, &s_len)
 */
        __pyx_v_s = NULL;

        /* "asynctnt/iproto/response.pyx":401
 *             elif key == tarantool.IPROTO_FIELD_TYPE:
 *                 s = NULL
 *                 s_len = 0             # <<<<<<<<<<<<<<
 *                 s = mp_decode_str(b, &s_len)
 *                 field.type = \
 */
        __pyx_v_s_len = 0;

        /* "asynctnt/iproto/response.pyx":402
 *                 s = NULL
 *                 s_len = 0
 *                 s = mp_decode_str(b, &s_len)             # <<<<<<<<<<<<<<
 *                 field.type = \
 *                     decode_string(s[:s_len], encoding)
 */
        __pyx_v_s = mp_decode_str(__pyx_v_b, (&__pyx_v_s_len));

        /* "asynctnt/iproto/response.pyx":404
 *                 s = mp_decode_str(b, &s_len)
 *                 field.type = \
 *                     decode_string(s[:s_len], encoding)             # <<<<<<<<<<<<<<
 * 
 *             elif key == tarantool.IPROTO_FIELD_COLL:
 */
        __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s + 0, __pyx_v_s_len - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 404, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_decode_string(((PyObject*)__pyx_t_2), __pyx_v_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 404, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "asynctnt/iproto/response.pyx":403
 *                 s_len = 0
 *                 s = mp_decode_str(b, &s_len)
 *                 field.type = \             # <<<<<<<<<<<<<<
 *                     decode_string(s[:s_len], encoding)
 * 
 */
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_v_field->type);
        __Pyx_DECREF(__pyx_v_field->type);
        __pyx_v_field->type = ((PyObject*)__pyx_t_1);
        __pyx_t_1 = 0;

        /* "asynctnt/iproto/response.pyx":399
 *                     decode_string(s[:s_len], encoding)
 * 
 *             elif key == tarantool.IPROTO_FIELD_TYPE:             # <<<<<<<<<<<<<<
 *                 s = NULL
 *                 s_len = 0
 */
        break;
        case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_COLL:

        /* "asynctnt/iproto/response.pyx":407
 * 
 *             elif key == tarantool.IPROTO_FIELD_COLL:
 *                 s = NULL             # <<<<<<<<<<<<<<
 *                 s_len = 0
 *                 s = mp_decode_str(b, &s_len)
 */
        __pyx_v_s = NULL;

        /* "asynctnt/iproto/response.pyx":408
 *             elif key == tarantool.IPROTO_FIELD_COLL:
 *                 s = NULL
 *                 s_len = 0             # <<<<<<<<<<<<<<
 *                 s = mp_decode_str(b, &s_len)
 *                 field.collation = \
 */
        __pyx_v_s_len = 0;

        /* "asynctnt/iproto/response.pyx":409
 *                 s = NULL
 *                 s_len = 0
 *                 s = mp_decode_str(b, &s_len)             # <<<<<<<<<<<<<<
 *                 field.collation = \
 *                     decode_string(s[:s_len], encoding)
 */
        __pyx_v_s = mp_decode_str(__pyx_v_b, (&__pyx_v_s_len));

        /* "asynctnt/iproto/response.pyx":411
 *                 s = mp_decode_str(b, &s_len)
 *                 field.collation = \
 *                     decode_string(s[:s_len], encoding)             # <<<<<<<<<<<<<<
 * 
 *             elif key == tarantool.IPROTO_FIELD_IS_NULLABLE:
 */
        __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s + 0, __pyx_v_s_len - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_decode_string(((PyObject*)__pyx_t_1), __pyx_v_encoding); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynctnt/iproto/response.pyx":410
 *                 s_len = 0
 *                 s = mp_decode_str(b, &s_len)
 *                 field.collation = \             # <<<<<<<<<<<<<<
 *                     decode_string(s[:s_len], encoding)
 * 
 */
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_v_field->collation);
        __Pyx_DECREF(__pyx_v_field->collation);
        __pyx_v_field->collation = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;

        /* "asynctnt/iproto/response.pyx":406
 *                     decode_string(s[:s_len], encoding)
 * 
 *             elif key == tarantool.IPROTO_FIELD_COLL:             # <<<<<<<<<<<<<<
 *                 s = NULL
 *                 s_len = 0
 */
        break;
        case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_IS_NULLABLE:

        /* "asynctnt/iproto/response.pyx":414
 * 
 *             elif key == tarantool.IPROTO_FIELD_IS_NULLABLE:
 *                 field.is_nullable = mp_decode_bool(b)             # <<<<<<<<<<<<<<
 * 
 *             elif key == tarantool.IPROTO_FIELD_IS_AUTOINCREMENT:
 */
        __pyx_t_2 = __Pyx_PyBool_FromLong(mp_decode_bool(__pyx_v_b)); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 414, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_v_field->is_nullable);
        __Pyx_DECREF(__pyx_v_field->is_nullable);
        __pyx_v_field->is_nullable = __pyx_t_2;
        __pyx_t_2 = 0;

        /* "asynctnt/iproto/response.pyx":413
 *                     decode_string(s[:s_len], encoding)
 * 
 *             elif key == tarantool.IPROTO_FIELD_IS_NULLABLE:             # <<<<<<<<<<<<<<
 *                 field.is_nullable = mp_decode_bool(b)
 * 
 */
        break;
        case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_IS_AUTOINCREMENT:

        /* "asynctnt/iproto/response.pyx":417
 * 
 *             elif key == tarantool.IPROTO_FIELD_IS_AUTOINCREMENT:
 *                 field.is_autoincrement = mp_decode_bool(b)             # <<<<<<<<<<<<<<
 * 
 *             elif key == tarantool.IPROTO_FIELD_SPAN:
 */
        __pyx_t_2 = __Pyx_PyBool_FromLong(mp_decode_bool(__pyx_v_b)); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 417, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_v_field->is_autoincrement);
        __Pyx_DECREF(__pyx_v_field->is_autoincrement);
        __pyx_v_field->is_autoincrement = __pyx_t_2;
        __pyx_t_2 = 0;

        /* "asynctnt/iproto/response.pyx":416
 *                 field.is_nullable = mp_decode_bool(b)
 * 
 *             elif key == tarantool.IPROTO_FIELD_IS_AUTOINCREMENT:             # <<<<<<<<<<<<<<
 *                 field.is_autoincrement = mp_decode_bool(b)
 * 
 */
        break;
        case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FIELD_SPAN:

        /* "asynctnt/iproto/response.pyx":420
 * 
 *             elif key == tarantool.IPROTO_FIELD_SPAN:
 *                 if mp_typeof(b[0][0]) == MP_NIL:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                     mp_next(b)
 *                     field.span = None
 */
        __pyx_t_6 = (mp_typeof(((__pyx_v_b[0])[0])) == MP_NIL);
        if (__pyx_t_6) {

          /* "asynctnt/iproto/response.pyx":421
 *             elif key == tarantool.IPROTO_FIELD_SPAN:
 *                 if mp_typeof(b[0][0]) == MP_NIL:  # pragma: nocover
 *                     mp_next(b)             # <<<<<<<<<<<<<<
 *                     field.span = None
 * 
 */
          mp_next(__pyx_v_b);

          /* "asynctnt/iproto/response.pyx":422
 *                 if mp_typeof(b[0][0]) == MP_NIL:  # pragma: nocover
 *                     mp_next(b)
 *                     field.span = None             # <<<<<<<<<<<<<<
 * 
 *                 elif mp_typeof(b[0][0]) == MP_STR:
 */
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          __Pyx_GOTREF(__pyx_v_field->span);
          __Pyx_DECREF(__pyx_v_field->span);
          __pyx_v_field->span = ((PyObject*)Py_None);

          /* "asynctnt/iproto/response.pyx":420
 * 
 *             elif key == tarantool.IPROTO_FIELD_SPAN:
 *                 if mp_typeof(b[0][0]) == MP_NIL:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                     mp_next(b)
 *                     field.span = None
 */
          goto __pyx_L8;
        }

        /* "asynctnt/iproto/response.pyx":424
 *                     field.span = None
 * 
 *                 elif mp_typeof(b[0][0]) == MP_STR:             # <<<<<<<<<<<<<<
 *                     s = NULL
 *                     s_len = 0
 */
        __pyx_t_6 = (mp_typeof(((__pyx_v_b[0])[0])) == MP_STR);
        if (likely(__pyx_t_6)) {

          /* "asynctnt/iproto/response.pyx":425
 * 
 *                 elif mp_typeof(b[0][0]) == MP_STR:
 *                     s = NULL             # <<<<<<<<<<<<<<
 *                     s_len = 0
 *                     s = mp_decode_str(b, &s_len)
 */
          __pyx_v_s = NULL;

          /* "asynctnt/iproto/response.pyx":426
 *                 elif mp_typeof(b[0][0]) == MP_STR:
 *                     s = NULL
 *                     s_len = 0             # <<<<<<<<<<<<<<
 *                     s = mp_decode_str(b, &s_len)
 *                     field.span = \
 */
          __pyx_v_s_len = 0;

          /* "asynctnt/iproto/response.pyx":427
 *                     s = NULL
 *                     s_len = 0
 *                     s = mp_decode_str(b, &s_len)             # <<<<<<<<<<<<<<
 *                     field.span = \
 *                         decode_string(s[:s_len], encoding)
 */
          __pyx_v_s = mp_decode_str(__pyx_v_b, (&__pyx_v_s_len));

          /* "asynctnt/iproto/response.pyx":429
 *                     s = mp_decode_str(b, &s_len)
 *                     field.span = \
 *                         decode_string(s[:s_len], encoding)             # <<<<<<<<<<<<<<
 * 
 *                 else:  # pragma: nocover
 */
          __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s + 0, __pyx_v_s_len - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 429, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_decode_string(((PyObject*)__pyx_t_2), __pyx_v_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 429, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "asynctnt/iproto/response.pyx":428
 *                     s_len = 0
 *                     s = mp_decode_str(b, &s_len)
 *                     field.span = \             # <<<<<<<<<<<<<<
 *                         decode_string(s[:s_len], encoding)
 * 
 */
          __Pyx_GIVEREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_v_field->span);
          __Pyx_DECREF(__pyx_v_field->span);
          __pyx_v_field->span = ((PyObject*)__pyx_t_1);
          __pyx_t_1 = 0;

          /* "asynctnt/iproto/response.pyx":424
 *                     field.span = None
 * 
 *                 elif mp_typeof(b[0][0]) == MP_STR:             # <<<<<<<<<<<<<<
 *                     s = NULL
 *                     s_len = 0
 */
          goto __pyx_L8;
        }

        /* "asynctnt/iproto/response.pyx":432
 * 
 *                 else:  # pragma: nocover
 *                     raise TypeError(             # <<<<<<<<<<<<<<
 *                         "IPROTO_FIELD_SPAN must be either STR or NIL"
 *                     )
 */
        /*else*/ {
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 432, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(5, 432, __pyx_L1_error)
        }
        __pyx_L8:;

        /* "asynctnt/iproto/response.pyx":419
 *                 field.is_autoincrement = mp_decode_bool(b)
 * 
 *             elif key == tarantool.IPROTO_FIELD_SPAN:             # <<<<<<<<<<<<<<
 *                 if mp_typeof(b[0][0]) == MP_NIL:  # pragma: nocover
 *                     mp_next(b)
 */
        break;
        default:

        /* "asynctnt/iproto/response.pyx":436
 *                     )
 *             else:  # pragma: nocover
 *                 logger.debug(             # <<<<<<<<<<<<<<
 *                     'unknown key in metadata decoding: %d', key)
 *                 mp_next(b)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 436, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 436, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "asynctnt/iproto/response.pyx":437
 *             else:  # pragma: nocover
 *                 logger.debug(
 *                     'unknown key in metadata decoding: %d', key)             # <<<<<<<<<<<<<<
 *                 mp_next(b)
 * 
 */
        __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 437, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_11 = NULL;
        __pyx_t_12 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_12 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_11, __pyx_kp_u_unknown_key_in_metadata_decoding, __pyx_t_2};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_12, 2+__pyx_t_12);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 436, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynctnt/iproto/response.pyx":438
 *                 logger.debug(
 *                     'unknown key in metadata decoding: %d', key)
 *                 mp_next(b)             # <<<<<<<<<<<<<<
 * 
 *         if field.name is None:  # pragma: nocover
 */
        mp_next(__pyx_v_b);
        break;
      }
    }

    /* "asynctnt/iproto/response.pyx":440
 *                 mp_next(b)
 * 
 *         if field.name is None:  # pragma: nocover             # <<<<<<<<<<<<<<
 *             raise RuntimeError('field.name must not be None')
 * 
 */
    __pyx_t_6 = (__pyx_v_field->name == ((PyObject*)Py_None));
    if (unlikely(__pyx_t_6)) {

      /* "asynctnt/iproto/response.pyx":441
 * 
 *         if field.name is None:  # pragma: nocover
 *             raise RuntimeError('field.name must not be None')             # <<<<<<<<<<<<<<
 * 
 *         metadata.add(<int> field_id, field)
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 441, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(5, 441, __pyx_L1_error)

      /* "asynctnt/iproto/response.pyx":440
 *                 mp_next(b)
 * 
 *         if field.name is None:  # pragma: nocover             # <<<<<<<<<<<<<<
 *             raise RuntimeError('field.name must not be None')
 * 
 */
    }

    /* "asynctnt/iproto/response.pyx":443
 *             raise RuntimeError('field.name must not be None')
 * 
 *         metadata.add(<int> field_id, field)             # <<<<<<<<<<<<<<
 *     return metadata
 * 
 */
    __pyx_f_8asynctnt_6iproto_8protocol_8Metadata_add(__pyx_v_metadata, ((int)__pyx_v_field_id), __pyx_v_field); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 443, __pyx_L1_error)
  }

  /* "asynctnt/iproto/response.pyx":444
 * 
 *         metadata.add(<int> field_id, field)
 *     return metadata             # <<<<<<<<<<<<<<
 * 
 * cdef inline IProtoErrorStackFrame parse_iproto_error_stack_frame(const char ** b, bytes encoding):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_metadata);
  __pyx_r = __pyx_v_metadata;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":369
 *     return <ssize_t> (b - buf)
 * 
 * cdef Metadata response_parse_metadata(const char ** b, bytes encoding):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t arr_size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.response_parse_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XDECREF((PyObject *)__pyx_v_metadata);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":446
 *     return metadata
 * 
 * cdef inline IProtoErrorStackFrame parse_iproto_error_stack_frame(const char ** b, bytes encoding):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t size
 */

static CYTHON_INLINE struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_f_8asynctnt_6iproto_8protocol_parse_iproto_error_stack_frame(char const **__pyx_v_b, PyObject *__pyx_v_encoding) {
  uint32_t __pyx_v_size;
  uint32_t __pyx_v_key;
  char const *__pyx_v_s;
  uint32_t __pyx_v_s_len;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_frame = 0;
  CYTHON_UNUSED uint32_t __pyx_v__;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  uint32_t __pyx_t_3;
  uint32_t __pyx_t_4;
  uint32_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse_iproto_error_stack_frame", 1);

  /* "asynctnt/iproto/response.pyx":455
 *         uint32_t unum
 * 
 *     size = 0             # <<<<<<<<<<<<<<
 *     key = 0
 * 
 */
  __pyx_v_size = 0;

  /* "asynctnt/iproto/response.pyx":456
 * 
 *     size = 0
 *     key = 0             # <<<<<<<<<<<<<<
 * 
 *     frame = <IProtoErrorStackFrame> IProtoErrorStackFrame.__new__(IProtoErrorStackFrame)
 */
  __pyx_v_key = 0;

  /* "asynctnt/iproto/response.pyx":458
 *     key = 0
 * 
 *     frame = <IProtoErrorStackFrame> IProtoErrorStackFrame.__new__(IProtoErrorStackFrame)             # <<<<<<<<<<<<<<
 * 
 *     size = mp_decode_map(b)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 458, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_frame = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/response.pyx":460
 *     frame = <IProtoErrorStackFrame> IProtoErrorStackFrame.__new__(IProtoErrorStackFrame)
 * 
 *     size = mp_decode_map(b)             # <<<<<<<<<<<<<<
 *     for _ in range(size):
 *         key = mp_decode_uint(b)
 */
  __pyx_v_size = mp_decode_map(__pyx_v_b);

  /* "asynctnt/iproto/response.pyx":461
 * 
 *     size = mp_decode_map(b)
 *     for _ in range(size):             # <<<<<<<<<<<<<<
 *         key = mp_decode_uint(b)
 * 
 */
  __pyx_t_3 = __pyx_v_size;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v__ = __pyx_t_5;

    /* "asynctnt/iproto/response.pyx":462
 *     size = mp_decode_map(b)
 *     for _ in range(size):
 *         key = mp_decode_uint(b)             # <<<<<<<<<<<<<<
 * 
 *         if key == tarantool.MP_ERROR_TYPE:
 */
    __pyx_v_key = mp_decode_uint(__pyx_v_b);

    /* "asynctnt/iproto/response.pyx":464
 *         key = mp_decode_uint(b)
 * 
 *         if key == tarantool.MP_ERROR_TYPE:             # <<<<<<<<<<<<<<
 *             s = NULL
 *             s_len = 0
 */
    switch (__pyx_v_key) {
      case __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_TYPE:

      /* "asynctnt/iproto/response.pyx":465
 * 
 *         if key == tarantool.MP_ERROR_TYPE:
 *             s = NULL             # <<<<<<<<<<<<<<
 *             s_len = 0
 *             s = mp_decode_str(b, &s_len)
 */
      __pyx_v_s = NULL;

      /* "asynctnt/iproto/response.pyx":466
 *         if key == tarantool.MP_ERROR_TYPE:
 *             s = NULL
 *             s_len = 0             # <<<<<<<<<<<<<<
 *             s = mp_decode_str(b, &s_len)
 *             frame.error_type = decode_string(s[:s_len], encoding)
 */
      __pyx_v_s_len = 0;

      /* "asynctnt/iproto/response.pyx":467
 *             s = NULL
 *             s_len = 0
 *             s = mp_decode_str(b, &s_len)             # <<<<<<<<<<<<<<
 *             frame.error_type = decode_string(s[:s_len], encoding)
 * 
 */
      __pyx_v_s = mp_decode_str(__pyx_v_b, (&__pyx_v_s_len));

      /* "asynctnt/iproto/response.pyx":468
 *             s_len = 0
 *             s = mp_decode_str(b, &s_len)
 *             frame.error_type = decode_string(s[:s_len], encoding)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.MP_ERROR_FILE:
 */
      __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s + 0, __pyx_v_s_len - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_decode_string(((PyObject*)__pyx_t_2), __pyx_v_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_v_frame->error_type);
      __Pyx_DECREF(__pyx_v_frame->error_type);
      __pyx_v_frame->error_type = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "asynctnt/iproto/response.pyx":464
 *         key = mp_decode_uint(b)
 * 
 *         if key == tarantool.MP_ERROR_TYPE:             # <<<<<<<<<<<<<<
 *             s = NULL
 *             s_len = 0
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_FILE:

      /* "asynctnt/iproto/response.pyx":471
 * 
 *         elif key == tarantool.MP_ERROR_FILE:
 *             s = NULL             # <<<<<<<<<<<<<<
 *             s_len = 0
 *             s = mp_decode_str(b, &s_len)
 */
      __pyx_v_s = NULL;

      /* "asynctnt/iproto/response.pyx":472
 *         elif key == tarantool.MP_ERROR_FILE:
 *             s = NULL
 *             s_len = 0             # <<<<<<<<<<<<<<
 *             s = mp_decode_str(b, &s_len)
 *             frame.file = decode_string(s[:s_len], encoding)
 */
      __pyx_v_s_len = 0;

      /* "asynctnt/iproto/response.pyx":473
 *             s = NULL
 *             s_len = 0
 *             s = mp_decode_str(b, &s_len)             # <<<<<<<<<<<<<<
 *             frame.file = decode_string(s[:s_len], encoding)
 * 
 */
      __pyx_v_s = mp_decode_str(__pyx_v_b, (&__pyx_v_s_len));

      /* "asynctnt/iproto/response.pyx":474
 *             s_len = 0
 *             s = mp_decode_str(b, &s_len)
 *             frame.file = decode_string(s[:s_len], encoding)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.MP_ERROR_LINE:
 */
      __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s + 0, __pyx_v_s_len - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_decode_string(((PyObject*)__pyx_t_1), __pyx_v_encoding); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_v_frame->file);
      __Pyx_DECREF(__pyx_v_frame->file);
      __pyx_v_frame->file = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "asynctnt/iproto/response.pyx":470
 *             frame.error_type = decode_string(s[:s_len], encoding)
 * 
 *         elif key == tarantool.MP_ERROR_FILE:             # <<<<<<<<<<<<<<
 *             s = NULL
 *             s_len = 0
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_LINE:

      /* "asynctnt/iproto/response.pyx":477
 * 
 *         elif key == tarantool.MP_ERROR_LINE:
 *             frame.line = <int> mp_decode_uint(b)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.MP_ERROR_MESSAGE:
 */
      __pyx_v_frame->line = ((int)mp_decode_uint(__pyx_v_b));

      /* "asynctnt/iproto/response.pyx":476
 *             frame.file = decode_string(s[:s_len], encoding)
 * 
 *         elif key == tarantool.MP_ERROR_LINE:             # <<<<<<<<<<<<<<
 *             frame.line = <int> mp_decode_uint(b)
 * 
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_MESSAGE:

      /* "asynctnt/iproto/response.pyx":480
 * 
 *         elif key == tarantool.MP_ERROR_MESSAGE:
 *             s = NULL             # <<<<<<<<<<<<<<
 *             s_len = 0
 *             s = mp_decode_str(b, &s_len)
 */
      __pyx_v_s = NULL;

      /* "asynctnt/iproto/response.pyx":481
 *         elif key == tarantool.MP_ERROR_MESSAGE:
 *             s = NULL
 *             s_len = 0             # <<<<<<<<<<<<<<
 *             s = mp_decode_str(b, &s_len)
 *             frame.message = decode_string(s[:s_len], encoding)
 */
      __pyx_v_s_len = 0;

      /* "asynctnt/iproto/response.pyx":482
 *             s = NULL
 *             s_len = 0
 *             s = mp_decode_str(b, &s_len)             # <<<<<<<<<<<<<<
 *             frame.message = decode_string(s[:s_len], encoding)
 * 
 */
      __pyx_v_s = mp_decode_str(__pyx_v_b, (&__pyx_v_s_len));

      /* "asynctnt/iproto/response.pyx":483
 *             s_len = 0
 *             s = mp_decode_str(b, &s_len)
 *             frame.message = decode_string(s[:s_len], encoding)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.MP_ERROR_ERRNO:
 */
      __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s + 0, __pyx_v_s_len - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_decode_string(((PyObject*)__pyx_t_2), __pyx_v_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_v_frame->message);
      __Pyx_DECREF(__pyx_v_frame->message);
      __pyx_v_frame->message = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "asynctnt/iproto/response.pyx":479
 *             frame.line = <int> mp_decode_uint(b)
 * 
 *         elif key == tarantool.MP_ERROR_MESSAGE:             # <<<<<<<<<<<<<<
 *             s = NULL
 *             s_len = 0
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_ERRNO:

      /* "asynctnt/iproto/response.pyx":486
 * 
 *         elif key == tarantool.MP_ERROR_ERRNO:
 *             frame.err_no = <int> mp_decode_uint(b)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.MP_ERROR_ERRCODE:
 */
      __pyx_v_frame->err_no = ((int)mp_decode_uint(__pyx_v_b));

      /* "asynctnt/iproto/response.pyx":485
 *             frame.message = decode_string(s[:s_len], encoding)
 * 
 *         elif key == tarantool.MP_ERROR_ERRNO:             # <<<<<<<<<<<<<<
 *             frame.err_no = <int> mp_decode_uint(b)
 * 
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_ERRCODE:

      /* "asynctnt/iproto/response.pyx":489
 * 
 *         elif key == tarantool.MP_ERROR_ERRCODE:
 *             frame.code = <int> mp_decode_uint(b)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.MP_ERROR_FIELDS:
 */
      __pyx_v_frame->code = ((int)mp_decode_uint(__pyx_v_b));

      /* "asynctnt/iproto/response.pyx":488
 *             frame.err_no = <int> mp_decode_uint(b)
 * 
 *         elif key == tarantool.MP_ERROR_ERRCODE:             # <<<<<<<<<<<<<<
 *             frame.code = <int> mp_decode_uint(b)
 * 
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_FIELDS:

      /* "asynctnt/iproto/response.pyx":492
 * 
 *         elif key == tarantool.MP_ERROR_FIELDS:
 *             if mp_typeof(b[0][0]) != MP_MAP:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError(f'iproto_error stack frame fields must be a '
 *                                 f'map, but got {mp_typeof(b[0][0])}')
 */
      __pyx_t_6 = (mp_typeof(((__pyx_v_b[0])[0])) != MP_MAP);
      if (unlikely(__pyx_t_6)) {

        /* "asynctnt/iproto/response.pyx":494
 *             if mp_typeof(b[0][0]) != MP_MAP:  # pragma: nocover
 *                 raise TypeError(f'iproto_error stack frame fields must be a '
 *                                 f'map, but got {mp_typeof(b[0][0])}')             # <<<<<<<<<<<<<<
 * 
 *             frame.fields = _decode_obj(b, encoding)
 */
        __pyx_t_1 = __Pyx_PyUnicode_From_enum__mp_type(mp_typeof(((__pyx_v_b[0])[0])), 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 494, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "asynctnt/iproto/response.pyx":493
 *         elif key == tarantool.MP_ERROR_FIELDS:
 *             if mp_typeof(b[0][0]) != MP_MAP:  # pragma: nocover
 *                 raise TypeError(f'iproto_error stack frame fields must be a '             # <<<<<<<<<<<<<<
 *                                 f'map, but got {mp_typeof(b[0][0])}')
 * 
 */
        __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_iproto_error_stack_frame_fields, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 493, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 493, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(5, 493, __pyx_L1_error)

        /* "asynctnt/iproto/response.pyx":492
 * 
 *         elif key == tarantool.MP_ERROR_FIELDS:
 *             if mp_typeof(b[0][0]) != MP_MAP:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError(f'iproto_error stack frame fields must be a '
 *                                 f'map, but got {mp_typeof(b[0][0])}')
 */
      }

      /* "asynctnt/iproto/response.pyx":496
 *                                 f'map, but got {mp_typeof(b[0][0])}')
 * 
 *             frame.fields = _decode_obj(b, encoding)             # <<<<<<<<<<<<<<
 * 
 *         else:  # pragma: nocover
 */
      __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol__decode_obj(__pyx_v_b, __pyx_v_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(5, 496, __pyx_L1_error)
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_v_frame->fields);
      __Pyx_DECREF(__pyx_v_frame->fields);
      __pyx_v_frame->fields = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "asynctnt/iproto/response.pyx":491
 *             frame.code = <int> mp_decode_uint(b)
 * 
 *         elif key == tarantool.MP_ERROR_FIELDS:             # <<<<<<<<<<<<<<
 *             if mp_typeof(b[0][0]) != MP_MAP:  # pragma: nocover
 *                 raise TypeError(f'iproto_error stack frame fields must be a '
 */
      break;
      default:

      /* "asynctnt/iproto/response.pyx":499
 * 
 *         else:  # pragma: nocover
 *             logger.debug(f"unknown iproto_error stack element with key {key}")             # <<<<<<<<<<<<<<
 *             mp_next(b)
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 499, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 499, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 499, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 499, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_unknown_iproto_error_stack_eleme, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 499, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_2};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 499, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynctnt/iproto/response.pyx":500
 *         else:  # pragma: nocover
 *             logger.debug(f"unknown iproto_error stack element with key {key}")
 *             mp_next(b)             # <<<<<<<<<<<<<<
 * 
 *     return frame
 */
      mp_next(__pyx_v_b);
      break;
    }
  }

  /* "asynctnt/iproto/response.pyx":502
 *             mp_next(b)
 * 
 *     return frame             # <<<<<<<<<<<<<<
 * 
 * cdef inline IProtoError parse_iproto_error(const char ** b, bytes encoding):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_frame);
  __pyx_r = __pyx_v_frame;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":446
 *     return metadata
 * 
 * cdef inline IProtoErrorStackFrame parse_iproto_error_stack_frame(const char ** b, bytes encoding):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.parse_iproto_error_stack_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_frame);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":504
 *     return frame
 * 
 * cdef inline IProtoError parse_iproto_error(const char ** b, bytes encoding):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t size
 */

static CYTHON_INLINE struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *__pyx_f_8asynctnt_6iproto_8protocol_parse_iproto_error(char const **__pyx_v_b, PyObject *__pyx_v_encoding) {
  uint32_t __pyx_v_size;
  uint32_t __pyx_v_arr_size;
  uint32_t __pyx_v_key;
  uint32_t __pyx_v_i;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *__pyx_v_error = 0;
  CYTHON_UNUSED uint32_t __pyx_v__;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v_el = NULL;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  uint32_t __pyx_t_3;
  uint32_t __pyx_t_4;
  uint32_t __pyx_t_5;
  int __pyx_t_6;
  uint32_t __pyx_t_7;
  uint32_t __pyx_t_8;
  uint32_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse_iproto_error", 1);

  /* "asynctnt/iproto/response.pyx":512
 *         IProtoError error
 * 
 *     size = 0             # <<<<<<<<<<<<<<
 *     arr_size = 0
 *     key = 0
 */
  __pyx_v_size = 0;

  /* "asynctnt/iproto/response.pyx":513
 * 
 *     size = 0
 *     arr_size = 0             # <<<<<<<<<<<<<<
 *     key = 0
 * 
 */
  __pyx_v_arr_size = 0;

  /* "asynctnt/iproto/response.pyx":514
 *     size = 0
 *     arr_size = 0
 *     key = 0             # <<<<<<<<<<<<<<
 * 
 *     error = <IProtoError> IProtoError.__new__(IProtoError)
 */
  __pyx_v_key = 0;

  /* "asynctnt/iproto/response.pyx":516
 *     key = 0
 * 
 *     error = <IProtoError> IProtoError.__new__(IProtoError)             # <<<<<<<<<<<<<<
 * 
 *     size = mp_decode_map(b)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_IProtoError(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 516, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_error = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/response.pyx":518
 *     error = <IProtoError> IProtoError.__new__(IProtoError)
 * 
 *     size = mp_decode_map(b)             # <<<<<<<<<<<<<<
 *     for _ in range(size):
 *         key = mp_decode_uint(b)
 */
  __pyx_v_size = mp_decode_map(__pyx_v_b);

  /* "asynctnt/iproto/response.pyx":519
 * 
 *     size = mp_decode_map(b)
 *     for _ in range(size):             # <<<<<<<<<<<<<<
 *         key = mp_decode_uint(b)
 * 
 */
  __pyx_t_3 = __pyx_v_size;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v__ = __pyx_t_5;

    /* "asynctnt/iproto/response.pyx":520
 *     size = mp_decode_map(b)
 *     for _ in range(size):
 *         key = mp_decode_uint(b)             # <<<<<<<<<<<<<<
 * 
 *         if key == tarantool.MP_ERROR_STACK:
 */
    __pyx_v_key = mp_decode_uint(__pyx_v_b);

    /* "asynctnt/iproto/response.pyx":522
 *         key = mp_decode_uint(b)
 * 
 *         if key == tarantool.MP_ERROR_STACK:             # <<<<<<<<<<<<<<
 *             arr_size = mp_decode_array(b)
 *             error.trace = cpython.list.PyList_New(arr_size)
 */
    __pyx_t_6 = (__pyx_v_key == __pyx_e_8asynctnt_6iproto_9tarantool_MP_ERROR_STACK);
    if (__pyx_t_6) {

      /* "asynctnt/iproto/response.pyx":523
 * 
 *         if key == tarantool.MP_ERROR_STACK:
 *             arr_size = mp_decode_array(b)             # <<<<<<<<<<<<<<
 *             error.trace = cpython.list.PyList_New(arr_size)
 *             for i in range(arr_size):
 */
      __pyx_v_arr_size = mp_decode_array(__pyx_v_b);

      /* "asynctnt/iproto/response.pyx":524
 *         if key == tarantool.MP_ERROR_STACK:
 *             arr_size = mp_decode_array(b)
 *             error.trace = cpython.list.PyList_New(arr_size)             # <<<<<<<<<<<<<<
 *             for i in range(arr_size):
 *                 el = parse_iproto_error_stack_frame(b, encoding)
 */
      __pyx_t_2 = PyList_New(__pyx_v_arr_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_v_error->trace);
      __Pyx_DECREF(__pyx_v_error->trace);
      __pyx_v_error->trace = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "asynctnt/iproto/response.pyx":525
 *             arr_size = mp_decode_array(b)
 *             error.trace = cpython.list.PyList_New(arr_size)
 *             for i in range(arr_size):             # <<<<<<<<<<<<<<
 *                 el = parse_iproto_error_stack_frame(b, encoding)
 *                 cpython.Py_INCREF(el)
 */
      __pyx_t_7 = __pyx_v_arr_size;
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_i = __pyx_t_9;

        /* "asynctnt/iproto/response.pyx":526
 *             error.trace = cpython.list.PyList_New(arr_size)
 *             for i in range(arr_size):
 *                 el = parse_iproto_error_stack_frame(b, encoding)             # <<<<<<<<<<<<<<
 *                 cpython.Py_INCREF(el)
 *                 cpython.list.PyList_SET_ITEM(error.trace, i, el)
 */
        __pyx_t_2 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_parse_iproto_error_stack_frame(__pyx_v_b, __pyx_v_encoding)); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 526, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_el, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "asynctnt/iproto/response.pyx":527
 *             for i in range(arr_size):
 *                 el = parse_iproto_error_stack_frame(b, encoding)
 *                 cpython.Py_INCREF(el)             # <<<<<<<<<<<<<<
 *                 cpython.list.PyList_SET_ITEM(error.trace, i, el)
 *         else:  # pragma: nocover
 */
        Py_INCREF(((PyObject *)__pyx_v_el));

        /* "asynctnt/iproto/response.pyx":528
 *                 el = parse_iproto_error_stack_frame(b, encoding)
 *                 cpython.Py_INCREF(el)
 *                 cpython.list.PyList_SET_ITEM(error.trace, i, el)             # <<<<<<<<<<<<<<
 *         else:  # pragma: nocover
 *             logger.debug(f"unknown iproto_error map field with key {key}")
 */
        __pyx_t_2 = __pyx_v_error->trace;
        __Pyx_INCREF(__pyx_t_2);
        PyList_SET_ITEM(__pyx_t_2, __pyx_v_i, ((PyObject *)__pyx_v_el));
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }

      /* "asynctnt/iproto/response.pyx":522
 *         key = mp_decode_uint(b)
 * 
 *         if key == tarantool.MP_ERROR_STACK:             # <<<<<<<<<<<<<<
 *             arr_size = mp_decode_array(b)
 *             error.trace = cpython.list.PyList_New(arr_size)
 */
      goto __pyx_L5;
    }

    /* "asynctnt/iproto/response.pyx":530
 *                 cpython.list.PyList_SET_ITEM(error.trace, i, el)
 *         else:  # pragma: nocover
 *             logger.debug(f"unknown iproto_error map field with key {key}")             # <<<<<<<<<<<<<<
 *             mp_next(b)
 * 
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 530, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_debug); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 530, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 530, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(5, 530, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_kp_u_unknown_iproto_error_map_field_w, __pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 530, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = NULL;
      __pyx_t_12 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_12 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_1};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 530, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "asynctnt/iproto/response.pyx":531
 *         else:  # pragma: nocover
 *             logger.debug(f"unknown iproto_error map field with key {key}")
 *             mp_next(b)             # <<<<<<<<<<<<<<
 * 
 *     return error
 */
      mp_next(__pyx_v_b);
    }
    __pyx_L5:;
  }

  /* "asynctnt/iproto/response.pyx":533
 *             mp_next(b)
 * 
 *     return error             # <<<<<<<<<<<<<<
 * 
 * cdef ssize_t response_parse_body(const char *buf, uint32_t buf_len,
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_error);
  __pyx_r = __pyx_v_error;
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":504
 *     return frame
 * 
 * cdef inline IProtoError parse_iproto_error(const char ** b, bytes encoding):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.parse_iproto_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_error);
  __Pyx_XDECREF((PyObject *)__pyx_v_el);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/response.pyx":535
 *     return error
 * 
 * cdef ssize_t response_parse_body(const char *buf, uint32_t buf_len,             # <<<<<<<<<<<<<<
 *                                  Response resp, BaseRequest req,
 *                                  bint is_chunk) except -1:
 */

static Py_ssize_t __pyx_f_8asynctnt_6iproto_8protocol_response_parse_body(char const *__pyx_v_buf, CYTHON_UNUSED uint32_t __pyx_v_buf_len, struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_resp, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req, int __pyx_v_is_chunk) {
  char const *__pyx_v_b;
  uint32_t __pyx_v_size;
  uint32_t __pyx_v_arr_size;
  uint32_t __pyx_v_field_map_size;
  uint32_t __pyx_v_key;
  uint32_t __pyx_v_s_len;
  uint32_t __pyx_v_i;
  char const *__pyx_v_s;
  PyObject *__pyx_v_data = 0;
  CYTHON_UNUSED uint32_t __pyx_v__;
  PyObject *__pyx_v_ids = NULL;
  PyObject *__pyx_v_el = NULL;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  uint32_t __pyx_t_3;
  uint32_t __pyx_t_4;
  uint32_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  uint32_t __pyx_t_8;
  uint32_t __pyx_t_9;
  uint32_t __pyx_t_10;
  uint32_t __pyx_t_11;
  uint32_t __pyx_t_12;
  uint32_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("response_parse_body", 1);

  /* "asynctnt/iproto/response.pyx":550
 *         Field field
 * 
 *     b = <const char *> buf             # <<<<<<<<<<<<<<
 *     # mp_fprint(stdio.stdout, b)
 *     # stdio.fprintf(stdio.stdout, "\n")
 */
  __pyx_v_b = ((char const *)__pyx_v_buf);

  /* "asynctnt/iproto/response.pyx":556
 *     # parsing body
 * 
 *     if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover             # <<<<<<<<<<<<<<
 *         raise TypeError('Response body must be a MP_MAP')
 * 
 */
  __pyx_t_1 = (mp_typeof((__pyx_v_b[0])) != MP_MAP);
  if (unlikely(__pyx_t_1)) {

    /* "asynctnt/iproto/response.pyx":557
 * 
 *     if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover
 *         raise TypeError('Response body must be a MP_MAP')             # <<<<<<<<<<<<<<
 * 
 *     size = mp_decode_map(&b)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 557, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(5, 557, __pyx_L1_error)

    /* "asynctnt/iproto/response.pyx":556
 *     # parsing body
 * 
 *     if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover             # <<<<<<<<<<<<<<
 *         raise TypeError('Response body must be a MP_MAP')
 * 
 */
  }

  /* "asynctnt/iproto/response.pyx":559
 *         raise TypeError('Response body must be a MP_MAP')
 * 
 *     size = mp_decode_map(&b)             # <<<<<<<<<<<<<<
 *     for _ in range(size):
 *         if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 */
  __pyx_v_size = mp_decode_map((&__pyx_v_b));

  /* "asynctnt/iproto/response.pyx":560
 * 
 *     size = mp_decode_map(&b)
 *     for _ in range(size):             # <<<<<<<<<<<<<<
 *         if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *             raise TypeError('Header key must be a MP_UINT')
 */
  __pyx_t_3 = __pyx_v_size;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v__ = __pyx_t_5;

    /* "asynctnt/iproto/response.pyx":561
 *     size = mp_decode_map(&b)
 *     for _ in range(size):
 *         if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *             raise TypeError('Header key must be a MP_UINT')
 * 
 */
    __pyx_t_1 = (mp_typeof((__pyx_v_b[0])) != MP_UINT);
    if (unlikely(__pyx_t_1)) {

      /* "asynctnt/iproto/response.pyx":562
 *     for _ in range(size):
 *         if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *             raise TypeError('Header key must be a MP_UINT')             # <<<<<<<<<<<<<<
 * 
 *         key = mp_decode_uint(&b)
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(5, 562, __pyx_L1_error)

      /* "asynctnt/iproto/response.pyx":561
 *     size = mp_decode_map(&b)
 *     for _ in range(size):
 *         if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *             raise TypeError('Header key must be a MP_UINT')
 * 
 */
    }

    /* "asynctnt/iproto/response.pyx":564
 *             raise TypeError('Header key must be a MP_UINT')
 * 
 *         key = mp_decode_uint(&b)             # <<<<<<<<<<<<<<
 *         if key == tarantool.IPROTO_ERROR_24:
 *             if mp_typeof(b[0]) != MP_STR:  # pragma: nocover
 */
    __pyx_v_key = mp_decode_uint((&__pyx_v_b));

    /* "asynctnt/iproto/response.pyx":565
 * 
 *         key = mp_decode_uint(&b)
 *         if key == tarantool.IPROTO_ERROR_24:             # <<<<<<<<<<<<<<
 *             if mp_typeof(b[0]) != MP_STR:  # pragma: nocover
 *                 raise TypeError('errstr type must be a MP_STR')
 */
    switch (__pyx_v_key) {
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_ERROR_24:

      /* "asynctnt/iproto/response.pyx":566
 *         key = mp_decode_uint(&b)
 *         if key == tarantool.IPROTO_ERROR_24:
 *             if mp_typeof(b[0]) != MP_STR:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('errstr type must be a MP_STR')
 * 
 */
      __pyx_t_1 = (mp_typeof((__pyx_v_b[0])) != MP_STR);
      if (unlikely(__pyx_t_1)) {

        /* "asynctnt/iproto/response.pyx":567
 *         if key == tarantool.IPROTO_ERROR_24:
 *             if mp_typeof(b[0]) != MP_STR:  # pragma: nocover
 *                 raise TypeError('errstr type must be a MP_STR')             # <<<<<<<<<<<<<<
 * 
 *             s = NULL
 */
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 567, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(5, 567, __pyx_L1_error)

        /* "asynctnt/iproto/response.pyx":566
 *         key = mp_decode_uint(&b)
 *         if key == tarantool.IPROTO_ERROR_24:
 *             if mp_typeof(b[0]) != MP_STR:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('errstr type must be a MP_STR')
 * 
 */
      }

      /* "asynctnt/iproto/response.pyx":569
 *                 raise TypeError('errstr type must be a MP_STR')
 * 
 *             s = NULL             # <<<<<<<<<<<<<<
 *             s_len = 0
 *             s = mp_decode_str(&b, &s_len)
 */
      __pyx_v_s = NULL;

      /* "asynctnt/iproto/response.pyx":570
 * 
 *             s = NULL
 *             s_len = 0             # <<<<<<<<<<<<<<
 *             s = mp_decode_str(&b, &s_len)
 *             resp.errmsg = decode_string(s[:s_len], resp.encoding)
 */
      __pyx_v_s_len = 0;

      /* "asynctnt/iproto/response.pyx":571
 *             s = NULL
 *             s_len = 0
 *             s = mp_decode_str(&b, &s_len)             # <<<<<<<<<<<<<<
 *             resp.errmsg = decode_string(s[:s_len], resp.encoding)
 * 
 */
      __pyx_v_s = mp_decode_str((&__pyx_v_b), (&__pyx_v_s_len));

      /* "asynctnt/iproto/response.pyx":572
 *             s_len = 0
 *             s = mp_decode_str(&b, &s_len)
 *             resp.errmsg = decode_string(s[:s_len], resp.encoding)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.IPROTO_ERROR:
 */
      __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s + 0, __pyx_v_s_len - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __pyx_v_resp->encoding;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_7 = __pyx_f_8asynctnt_6iproto_8protocol_decode_string(((PyObject*)__pyx_t_2), ((PyObject*)__pyx_t_6)); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_v_resp->errmsg);
      __Pyx_DECREF(__pyx_v_resp->errmsg);
      __pyx_v_resp->errmsg = ((PyObject*)__pyx_t_7);
      __pyx_t_7 = 0;

      /* "asynctnt/iproto/response.pyx":565
 * 
 *         key = mp_decode_uint(&b)
 *         if key == tarantool.IPROTO_ERROR_24:             # <<<<<<<<<<<<<<
 *             if mp_typeof(b[0]) != MP_STR:  # pragma: nocover
 *                 raise TypeError('errstr type must be a MP_STR')
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_ERROR:

      /* "asynctnt/iproto/response.pyx":575
 * 
 *         elif key == tarantool.IPROTO_ERROR:
 *             if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('IPROTO_ERROR type must be a MP_MAP')
 * 
 */
      __pyx_t_1 = (mp_typeof((__pyx_v_b[0])) != MP_MAP);
      if (unlikely(__pyx_t_1)) {

        /* "asynctnt/iproto/response.pyx":576
 *         elif key == tarantool.IPROTO_ERROR:
 *             if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover
 *                 raise TypeError('IPROTO_ERROR type must be a MP_MAP')             # <<<<<<<<<<<<<<
 * 
 *             resp.error = parse_iproto_error(&b, resp.encoding)
 */
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 576, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(5, 576, __pyx_L1_error)

        /* "asynctnt/iproto/response.pyx":575
 * 
 *         elif key == tarantool.IPROTO_ERROR:
 *             if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('IPROTO_ERROR type must be a MP_MAP')
 * 
 */
      }

      /* "asynctnt/iproto/response.pyx":578
 *                 raise TypeError('IPROTO_ERROR type must be a MP_MAP')
 * 
 *             resp.error = parse_iproto_error(&b, resp.encoding)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.IPROTO_STMT_ID:
 */
      __pyx_t_7 = __pyx_v_resp->encoding;
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_6 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_parse_iproto_error((&__pyx_v_b), ((PyObject*)__pyx_t_7))); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 578, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_GOTREF((PyObject *)__pyx_v_resp->error);
      __Pyx_DECREF((PyObject *)__pyx_v_resp->error);
      __pyx_v_resp->error = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)__pyx_t_6);
      __pyx_t_6 = 0;

      /* "asynctnt/iproto/response.pyx":574
 *             resp.errmsg = decode_string(s[:s_len], resp.encoding)
 * 
 *         elif key == tarantool.IPROTO_ERROR:             # <<<<<<<<<<<<<<
 *             if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover
 *                 raise TypeError('IPROTO_ERROR type must be a MP_MAP')
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_STMT_ID:

      /* "asynctnt/iproto/response.pyx":581
 * 
 *         elif key == tarantool.IPROTO_STMT_ID:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError(f'IPROTO_STMT_ID type must be a MP_UINT, but got {mp_typeof(b[0])}')
 *             resp.stmt_id_ = mp_decode_uint(&b)
 */
      __pyx_t_1 = (mp_typeof((__pyx_v_b[0])) != MP_UINT);
      if (unlikely(__pyx_t_1)) {

        /* "asynctnt/iproto/response.pyx":582
 *         elif key == tarantool.IPROTO_STMT_ID:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError(f'IPROTO_STMT_ID type must be a MP_UINT, but got {mp_typeof(b[0])}')             # <<<<<<<<<<<<<<
 *             resp.stmt_id_ = mp_decode_uint(&b)
 * 
 */
        __pyx_t_6 = __Pyx_PyUnicode_From_enum__mp_type(mp_typeof((__pyx_v_b[0])), 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 582, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyUnicode_Concat(__pyx_kp_u_IPROTO_STMT_ID_type_must_be_a_MP, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 582, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 582, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_6, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __PYX_ERR(5, 582, __pyx_L1_error)

        /* "asynctnt/iproto/response.pyx":581
 * 
 *         elif key == tarantool.IPROTO_STMT_ID:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError(f'IPROTO_STMT_ID type must be a MP_UINT, but got {mp_typeof(b[0])}')
 *             resp.stmt_id_ = mp_decode_uint(&b)
 */
      }

      /* "asynctnt/iproto/response.pyx":583
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError(f'IPROTO_STMT_ID type must be a MP_UINT, but got {mp_typeof(b[0])}')
 *             resp.stmt_id_ = mp_decode_uint(&b)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.IPROTO_METADATA:
 */
      __pyx_v_resp->stmt_id_ = mp_decode_uint((&__pyx_v_b));

      /* "asynctnt/iproto/response.pyx":580
 *             resp.error = parse_iproto_error(&b, resp.encoding)
 * 
 *         elif key == tarantool.IPROTO_STMT_ID:             # <<<<<<<<<<<<<<
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError(f'IPROTO_STMT_ID type must be a MP_UINT, but got {mp_typeof(b[0])}')
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_METADATA:

      /* "asynctnt/iproto/response.pyx":586
 * 
 *         elif key == tarantool.IPROTO_METADATA:
 *             if not req.parse_metadata:             # <<<<<<<<<<<<<<
 *                 mp_next(&b)
 *                 continue
 */
      __pyx_t_1 = (!__pyx_v_req->parse_metadata);
      if (__pyx_t_1) {

        /* "asynctnt/iproto/response.pyx":587
 *         elif key == tarantool.IPROTO_METADATA:
 *             if not req.parse_metadata:
 *                 mp_next(&b)             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
        mp_next((&__pyx_v_b));

        /* "asynctnt/iproto/response.pyx":588
 *             if not req.parse_metadata:
 *                 mp_next(&b)
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             resp.metadata = response_parse_metadata(&b, resp.encoding)
 */
        goto __pyx_L4_continue;

        /* "asynctnt/iproto/response.pyx":586
 * 
 *         elif key == tarantool.IPROTO_METADATA:
 *             if not req.parse_metadata:             # <<<<<<<<<<<<<<
 *                 mp_next(&b)
 *                 continue
 */
      }

      /* "asynctnt/iproto/response.pyx":590
 *                 continue
 * 
 *             resp.metadata = response_parse_metadata(&b, resp.encoding)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.IPROTO_BIND_METADATA:
 */
      __pyx_t_6 = __pyx_v_resp->encoding;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_7 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_response_parse_metadata((&__pyx_v_b), ((PyObject*)__pyx_t_6))); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_GOTREF((PyObject *)__pyx_v_resp->metadata);
      __Pyx_DECREF((PyObject *)__pyx_v_resp->metadata);
      __pyx_v_resp->metadata = ((struct C_Metadata *)__pyx_t_7);
      __pyx_t_7 = 0;

      /* "asynctnt/iproto/response.pyx":585
 *             resp.stmt_id_ = mp_decode_uint(&b)
 * 
 *         elif key == tarantool.IPROTO_METADATA:             # <<<<<<<<<<<<<<
 *             if not req.parse_metadata:
 *                 mp_next(&b)
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_BIND_METADATA:

      /* "asynctnt/iproto/response.pyx":593
 * 
 *         elif key == tarantool.IPROTO_BIND_METADATA:
 *             if not req.parse_metadata:             # <<<<<<<<<<<<<<
 *                 mp_next(&b)
 *                 continue
 */
      __pyx_t_1 = (!__pyx_v_req->parse_metadata);
      if (__pyx_t_1) {

        /* "asynctnt/iproto/response.pyx":594
 *         elif key == tarantool.IPROTO_BIND_METADATA:
 *             if not req.parse_metadata:
 *                 mp_next(&b)             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
        mp_next((&__pyx_v_b));

        /* "asynctnt/iproto/response.pyx":595
 *             if not req.parse_metadata:
 *                 mp_next(&b)
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             resp.params = response_parse_metadata(&b, resp.encoding)
 */
        goto __pyx_L4_continue;

        /* "asynctnt/iproto/response.pyx":593
 * 
 *         elif key == tarantool.IPROTO_BIND_METADATA:
 *             if not req.parse_metadata:             # <<<<<<<<<<<<<<
 *                 mp_next(&b)
 *                 continue
 */
      }

      /* "asynctnt/iproto/response.pyx":597
 *                 continue
 * 
 *             resp.params = response_parse_metadata(&b, resp.encoding)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.IPROTO_BIND_COUNT:
 */
      __pyx_t_7 = __pyx_v_resp->encoding;
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_6 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_response_parse_metadata((&__pyx_v_b), ((PyObject*)__pyx_t_7))); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_GOTREF((PyObject *)__pyx_v_resp->params);
      __Pyx_DECREF((PyObject *)__pyx_v_resp->params);
      __pyx_v_resp->params = ((struct C_Metadata *)__pyx_t_6);
      __pyx_t_6 = 0;

      /* "asynctnt/iproto/response.pyx":592
 *             resp.metadata = response_parse_metadata(&b, resp.encoding)
 * 
 *         elif key == tarantool.IPROTO_BIND_METADATA:             # <<<<<<<<<<<<<<
 *             if not req.parse_metadata:
 *                 mp_next(&b)
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_BIND_COUNT:

      /* "asynctnt/iproto/response.pyx":601
 *         elif key == tarantool.IPROTO_BIND_COUNT:
 * 
 *             resp.params_count = <int> mp_decode_uint(&b)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.IPROTO_SQL_INFO:
 */
      __pyx_v_resp->params_count = ((int)mp_decode_uint((&__pyx_v_b)));

      /* "asynctnt/iproto/response.pyx":599
 *             resp.params = response_parse_metadata(&b, resp.encoding)
 * 
 *         elif key == tarantool.IPROTO_BIND_COUNT:             # <<<<<<<<<<<<<<
 * 
 *             resp.params_count = <int> mp_decode_uint(&b)
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SQL_INFO:

      /* "asynctnt/iproto/response.pyx":604
 * 
 *         elif key == tarantool.IPROTO_SQL_INFO:
 *             field_map_size = mp_decode_map(&b)             # <<<<<<<<<<<<<<
 *             if field_map_size == 0:
 *                 raise RuntimeError('Field map must contain at least '
 */
      __pyx_v_field_map_size = mp_decode_map((&__pyx_v_b));

      /* "asynctnt/iproto/response.pyx":605
 *         elif key == tarantool.IPROTO_SQL_INFO:
 *             field_map_size = mp_decode_map(&b)
 *             if field_map_size == 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError('Field map must contain at least '
 *                                    '1 element - rowcount')
 */
      __pyx_t_1 = (__pyx_v_field_map_size == 0);
      if (unlikely(__pyx_t_1)) {

        /* "asynctnt/iproto/response.pyx":606
 *             field_map_size = mp_decode_map(&b)
 *             if field_map_size == 0:
 *                 raise RuntimeError('Field map must contain at least '             # <<<<<<<<<<<<<<
 *                                    '1 element - rowcount')
 * 
 */
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 606, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_Raise(__pyx_t_6, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __PYX_ERR(5, 606, __pyx_L1_error)

        /* "asynctnt/iproto/response.pyx":605
 *         elif key == tarantool.IPROTO_SQL_INFO:
 *             field_map_size = mp_decode_map(&b)
 *             if field_map_size == 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError('Field map must contain at least '
 *                                    '1 element - rowcount')
 */
      }

      /* "asynctnt/iproto/response.pyx":609
 *                                    '1 element - rowcount')
 * 
 *             for _ in range(field_map_size):             # <<<<<<<<<<<<<<
 *                 key = mp_decode_uint(&b)
 *                 if key == tarantool.SQL_INFO_ROW_COUNT:
 */
      __pyx_t_8 = __pyx_v_field_map_size;
      __pyx_t_9 = __pyx_t_8;
      for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
        __pyx_v__ = __pyx_t_10;

        /* "asynctnt/iproto/response.pyx":610
 * 
 *             for _ in range(field_map_size):
 *                 key = mp_decode_uint(&b)             # <<<<<<<<<<<<<<
 *                 if key == tarantool.SQL_INFO_ROW_COUNT:
 *                     resp._rowcount = mp_decode_uint(&b)
 */
        __pyx_v_key = mp_decode_uint((&__pyx_v_b));

        /* "asynctnt/iproto/response.pyx":611
 *             for _ in range(field_map_size):
 *                 key = mp_decode_uint(&b)
 *                 if key == tarantool.SQL_INFO_ROW_COUNT:             # <<<<<<<<<<<<<<
 *                     resp._rowcount = mp_decode_uint(&b)
 *                 elif key == tarantool.SQL_INFO_AUTOINCREMENT_IDS:
 */
        switch (__pyx_v_key) {
          case __pyx_e_8asynctnt_6iproto_9tarantool_SQL_INFO_ROW_COUNT:

          /* "asynctnt/iproto/response.pyx":612
 *                 key = mp_decode_uint(&b)
 *                 if key == tarantool.SQL_INFO_ROW_COUNT:
 *                     resp._rowcount = mp_decode_uint(&b)             # <<<<<<<<<<<<<<
 *                 elif key == tarantool.SQL_INFO_AUTOINCREMENT_IDS:
 *                     arr_size = mp_decode_array(&b)
 */
          __pyx_v_resp->_rowcount = mp_decode_uint((&__pyx_v_b));

          /* "asynctnt/iproto/response.pyx":611
 *             for _ in range(field_map_size):
 *                 key = mp_decode_uint(&b)
 *                 if key == tarantool.SQL_INFO_ROW_COUNT:             # <<<<<<<<<<<<<<
 *                     resp._rowcount = mp_decode_uint(&b)
 *                 elif key == tarantool.SQL_INFO_AUTOINCREMENT_IDS:
 */
          break;
          case __pyx_e_8asynctnt_6iproto_9tarantool_SQL_INFO_AUTOINCREMENT_IDS:

          /* "asynctnt/iproto/response.pyx":614
 *                     resp._rowcount = mp_decode_uint(&b)
 *                 elif key == tarantool.SQL_INFO_AUTOINCREMENT_IDS:
 *                     arr_size = mp_decode_array(&b)             # <<<<<<<<<<<<<<
 *                     ids = cpython.list.PyList_New(arr_size)
 *                     for i in range(arr_size):
 */
          __pyx_v_arr_size = mp_decode_array((&__pyx_v_b));

          /* "asynctnt/iproto/response.pyx":615
 *                 elif key == tarantool.SQL_INFO_AUTOINCREMENT_IDS:
 *                     arr_size = mp_decode_array(&b)
 *                     ids = cpython.list.PyList_New(arr_size)             # <<<<<<<<<<<<<<
 *                     for i in range(arr_size):
 *                         el = <object> mp_decode_uint(&b)
 */
          __pyx_t_6 = PyList_New(__pyx_v_arr_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 615, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_XDECREF_SET(__pyx_v_ids, ((PyObject*)__pyx_t_6));
          __pyx_t_6 = 0;

          /* "asynctnt/iproto/response.pyx":616
 *                     arr_size = mp_decode_array(&b)
 *                     ids = cpython.list.PyList_New(arr_size)
 *                     for i in range(arr_size):             # <<<<<<<<<<<<<<
 *                         el = <object> mp_decode_uint(&b)
 *                         cpython.Py_INCREF(el)
 */
          __pyx_t_11 = __pyx_v_arr_size;
          __pyx_t_12 = __pyx_t_11;
          for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
            __pyx_v_i = __pyx_t_13;

            /* "asynctnt/iproto/response.pyx":617
 *                     ids = cpython.list.PyList_New(arr_size)
 *                     for i in range(arr_size):
 *                         el = <object> mp_decode_uint(&b)             # <<<<<<<<<<<<<<
 *                         cpython.Py_INCREF(el)
 *                         cpython.list.PyList_SET_ITEM(ids, i, el)
 */
            __pyx_t_6 = __Pyx_PyInt_From_uint64_t(mp_decode_uint((&__pyx_v_b))); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 617, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_7 = __pyx_t_6;
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF_SET(__pyx_v_el, __pyx_t_7);
            __pyx_t_7 = 0;

            /* "asynctnt/iproto/response.pyx":618
 *                     for i in range(arr_size):
 *                         el = <object> mp_decode_uint(&b)
 *                         cpython.Py_INCREF(el)             # <<<<<<<<<<<<<<
 *                         cpython.list.PyList_SET_ITEM(ids, i, el)
 *                     resp.autoincrement_ids = ids
 */
            Py_INCREF(__pyx_v_el);

            /* "asynctnt/iproto/response.pyx":619
 *                         el = <object> mp_decode_uint(&b)
 *                         cpython.Py_INCREF(el)
 *                         cpython.list.PyList_SET_ITEM(ids, i, el)             # <<<<<<<<<<<<<<
 *                     resp.autoincrement_ids = ids
 *                 else:
 */
            PyList_SET_ITEM(__pyx_v_ids, __pyx_v_i, __pyx_v_el);
          }

          /* "asynctnt/iproto/response.pyx":620
 *                         cpython.Py_INCREF(el)
 *                         cpython.list.PyList_SET_ITEM(ids, i, el)
 *                     resp.autoincrement_ids = ids             # <<<<<<<<<<<<<<
 *                 else:
 *                     logger.debug('unknown key in sql info decoding: %d', key)
 */
          __Pyx_INCREF(__pyx_v_ids);
          __Pyx_GIVEREF(__pyx_v_ids);
          __Pyx_GOTREF(__pyx_v_resp->autoincrement_ids);
          __Pyx_DECREF(__pyx_v_resp->autoincrement_ids);
          __pyx_v_resp->autoincrement_ids = __pyx_v_ids;

          /* "asynctnt/iproto/response.pyx":613
 *                 if key == tarantool.SQL_INFO_ROW_COUNT:
 *                     resp._rowcount = mp_decode_uint(&b)
 *                 elif key == tarantool.SQL_INFO_AUTOINCREMENT_IDS:             # <<<<<<<<<<<<<<
 *                     arr_size = mp_decode_array(&b)
 *                     ids = cpython.list.PyList_New(arr_size)
 */
          break;
          default:

          /* "asynctnt/iproto/response.pyx":622
 *                     resp.autoincrement_ids = ids
 *                 else:
 *                     logger.debug('unknown key in sql info decoding: %d', key)             # <<<<<<<<<<<<<<
 *                     mp_next(&b)
 * 
 */
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_logger); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 622, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 622, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = __Pyx_PyInt_From_uint32_t(__pyx_v_key); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 622, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_14 = NULL;
          __pyx_t_15 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_15 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_14, __pyx_kp_u_unknown_key_in_sql_info_decoding, __pyx_t_6};
            __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_15, 2+__pyx_t_15);
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 622, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "asynctnt/iproto/response.pyx":623
 *                 else:
 *                     logger.debug('unknown key in sql info decoding: %d', key)
 *                     mp_next(&b)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.IPROTO_DATA:
 */
          mp_next((&__pyx_v_b));
          break;
        }
      }

      /* "asynctnt/iproto/response.pyx":603
 *             resp.params_count = <int> mp_decode_uint(&b)
 * 
 *         elif key == tarantool.IPROTO_SQL_INFO:             # <<<<<<<<<<<<<<
 *             field_map_size = mp_decode_map(&b)
 *             if field_map_size == 0:
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_DATA:

      /* "asynctnt/iproto/response.pyx":626
 * 
 *         elif key == tarantool.IPROTO_DATA:
 *             if mp_typeof(b[0]) != MP_ARRAY:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('body data type must be a MP_ARRAY')
 *             data = _response_parse_body_data(&b, resp, req)
 */
      __pyx_t_1 = (mp_typeof((__pyx_v_b[0])) != MP_ARRAY);
      if (unlikely(__pyx_t_1)) {

        /* "asynctnt/iproto/response.pyx":627
 *         elif key == tarantool.IPROTO_DATA:
 *             if mp_typeof(b[0]) != MP_ARRAY:  # pragma: nocover
 *                 raise TypeError('body data type must be a MP_ARRAY')             # <<<<<<<<<<<<<<
 *             data = _response_parse_body_data(&b, resp, req)
 *             if is_chunk:
 */
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 627, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(5, 627, __pyx_L1_error)

        /* "asynctnt/iproto/response.pyx":626
 * 
 *         elif key == tarantool.IPROTO_DATA:
 *             if mp_typeof(b[0]) != MP_ARRAY:  # pragma: nocover             # <<<<<<<<<<<<<<
 *                 raise TypeError('body data type must be a MP_ARRAY')
 *             data = _response_parse_body_data(&b, resp, req)
 */
      }

      /* "asynctnt/iproto/response.pyx":628
 *             if mp_typeof(b[0]) != MP_ARRAY:  # pragma: nocover
 *                 raise TypeError('body data type must be a MP_ARRAY')
 *             data = _response_parse_body_data(&b, resp, req)             # <<<<<<<<<<<<<<
 *             if is_chunk:
 *                 resp.add_push(data)
 */
      __pyx_t_7 = __pyx_f_8asynctnt_6iproto_8protocol__response_parse_body_data((&__pyx_v_b), __pyx_v_resp, __pyx_v_req); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 628, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_v_data, ((PyObject*)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "asynctnt/iproto/response.pyx":629
 *                 raise TypeError('body data type must be a MP_ARRAY')
 *             data = _response_parse_body_data(&b, resp, req)
 *             if is_chunk:             # <<<<<<<<<<<<<<
 *                 resp.add_push(data)
 *             else:
 */
      if (__pyx_v_is_chunk) {

        /* "asynctnt/iproto/response.pyx":630
 *             data = _response_parse_body_data(&b, resp, req)
 *             if is_chunk:
 *                 resp.add_push(data)             # <<<<<<<<<<<<<<
 *             else:
 *                 resp.set_data(data)
 */
        __pyx_f_8asynctnt_6iproto_8protocol_8Response_add_push(__pyx_v_resp, __pyx_v_data); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 630, __pyx_L1_error)

        /* "asynctnt/iproto/response.pyx":629
 *                 raise TypeError('body data type must be a MP_ARRAY')
 *             data = _response_parse_body_data(&b, resp, req)
 *             if is_chunk:             # <<<<<<<<<<<<<<
 *                 resp.add_push(data)
 *             else:
 */
        goto __pyx_L18;
      }

      /* "asynctnt/iproto/response.pyx":632
 *                 resp.add_push(data)
 *             else:
 *                 resp.set_data(data)             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.IPROTO_VERSION:
 */
      /*else*/ {
        __pyx_f_8asynctnt_6iproto_8protocol_8Response_set_data(__pyx_v_resp, __pyx_v_data); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 632, __pyx_L1_error)
      }
      __pyx_L18:;

      /* "asynctnt/iproto/response.pyx":625
 *                     mp_next(&b)
 * 
 *         elif key == tarantool.IPROTO_DATA:             # <<<<<<<<<<<<<<
 *             if mp_typeof(b[0]) != MP_ARRAY:  # pragma: nocover
 *                 raise TypeError('body data type must be a MP_ARRAY')
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_VERSION:

      /* "asynctnt/iproto/response.pyx":635
 * 
 *         elif key == tarantool.IPROTO_VERSION:
 *             logger.debug("IProto version: %s", _decode_obj(&b, resp.encoding))             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.IPROTO_FEATURES:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 635, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 635, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __pyx_v_resp->encoding;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_14 = __pyx_f_8asynctnt_6iproto_8protocol__decode_obj((&__pyx_v_b), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_14)) __PYX_ERR(5, 635, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      __pyx_t_15 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_15 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_kp_u_IProto_version_s, __pyx_t_14};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_15, 2+__pyx_t_15);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 635, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "asynctnt/iproto/response.pyx":634
 *                 resp.set_data(data)
 * 
 *         elif key == tarantool.IPROTO_VERSION:             # <<<<<<<<<<<<<<
 *             logger.debug("IProto version: %s", _decode_obj(&b, resp.encoding))
 * 
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_FEATURES:

      /* "asynctnt/iproto/response.pyx":638
 * 
 *         elif key == tarantool.IPROTO_FEATURES:
 *             logger.debug("IProto features available: %s", _decode_obj(&b, resp.encoding))             # <<<<<<<<<<<<<<
 * 
 *         elif key == tarantool.IPROTO_AUTH_TYPE:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_logger); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_debug); if (unlikely(!__pyx_t_14)) __PYX_ERR(5, 638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __pyx_v_resp->encoding;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol__decode_obj((&__pyx_v_b), ((PyObject*)__pyx_t_6)); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      __pyx_t_15 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
          __pyx_t_15 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_kp_u_IProto_features_available_s, __pyx_t_2};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_15, 2+__pyx_t_15);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 638, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "asynctnt/iproto/response.pyx":637
 *             logger.debug("IProto version: %s", _decode_obj(&b, resp.encoding))
 * 
 *         elif key == tarantool.IPROTO_FEATURES:             # <<<<<<<<<<<<<<
 *             logger.debug("IProto features available: %s", _decode_obj(&b, resp.encoding))
 * 
 */
      break;
      case __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_AUTH_TYPE:

      /* "asynctnt/iproto/response.pyx":641
 * 
 *         elif key == tarantool.IPROTO_AUTH_TYPE:
 *             logger.debug("IProto auth type: %s", _decode_obj(&b, resp.encoding))             # <<<<<<<<<<<<<<
 * 
 *         else:  # pragma: nocover
 */
      __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_logger); if (unlikely(!__pyx_t_14)) __PYX_ERR(5, 641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = __pyx_v_resp->encoding;
      __Pyx_INCREF(__pyx_t_14);
      __pyx_t_6 = __pyx_f_8asynctnt_6iproto_8protocol__decode_obj((&__pyx_v_b), ((PyObject*)__pyx_t_14)); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = NULL;
      __pyx_t_15 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_15 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_14, __pyx_kp_u_IProto_auth_type_s, __pyx_t_6};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_15, 2+__pyx_t_15);
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 641, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "asynctnt/iproto/response.pyx":640
 *             logger.debug("IProto features available: %s", _decode_obj(&b, resp.encoding))
 * 
 *         elif key == tarantool.IPROTO_AUTH_TYPE:             # <<<<<<<<<<<<<<
 *             logger.debug("IProto auth type: %s", _decode_obj(&b, resp.encoding))
 * 
 */
      break;
      default:

      /* "asynctnt/iproto/response.pyx":644
 * 
 *         else:  # pragma: nocover
 *             logger.debug('unknown key in body map: %s', hex(int(key)))             # <<<<<<<<<<<<<<
 *             mp_next(&b)
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_14 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyInt_Type)), __pyx_t_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(5, 644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_hex, __pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = NULL;
      __pyx_t_15 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_15 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_14, __pyx_kp_u_unknown_key_in_body_map_s, __pyx_t_2};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_15, 2+__pyx_t_15);
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 644, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "asynctnt/iproto/response.pyx":645
 *         else:  # pragma: nocover
 *             logger.debug('unknown key in body map: %s', hex(int(key)))
 *             mp_next(&b)             # <<<<<<<<<<<<<<
 * 
 *     return <ssize_t> (b - buf)
 */
      mp_next((&__pyx_v_b));
      break;
    }
    __pyx_L4_continue:;
  }

  /* "asynctnt/iproto/response.pyx":647
 *             mp_next(&b)
 * 
 *     return <ssize_t> (b - buf)             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((Py_ssize_t)(__pyx_v_b - __pyx_v_buf));
  goto __pyx_L0;

  /* "asynctnt/iproto/response.pyx":535
 *     return error
 * 
 * cdef ssize_t response_parse_body(const char *buf, uint32_t buf_len,             # <<<<<<<<<<<<<<
 *                                  Response resp, BaseRequest req,
 *                                  bint is_chunk) except -1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.response_parse_body", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_ids);
  __Pyx_XDECREF(__pyx_v_el);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":6
 * @cython.final
 * cdef class Db:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._stream_id = 0
 *         self._protocol = None
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_2Db_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8asynctnt_6iproto_8protocol_2Db_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db___cinit__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_2Db___cinit__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "asynctnt/iproto/db.pyx":7
 * cdef class Db:
 *     def __cinit__(self):
 *         self._stream_id = 0             # <<<<<<<<<<<<<<
 *         self._protocol = None
 *         self._encoding = None
 */
  __pyx_v_self->_stream_id = 0;

  /* "asynctnt/iproto/db.pyx":8
 *     def __cinit__(self):
 *         self._stream_id = 0
 *         self._protocol = None             # <<<<<<<<<<<<<<
 *         self._encoding = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_protocol);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_protocol);
  __pyx_v_self->_protocol = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)Py_None);

  /* "asynctnt/iproto/db.pyx":9
 *         self._stream_id = 0
 *         self._protocol = None
 *         self._encoding = None             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_encoding);
  __Pyx_DECREF(__pyx_v_self->_encoding);
  __pyx_v_self->_encoding = ((PyObject*)Py_None);

  /* "asynctnt/iproto/db.pyx":6
 * @cython.final
 * cdef class Db:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._stream_id = 0
 *         self._protocol = None
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":12
 * 
 *     @staticmethod
 *     cdef inline Db create(BaseProtocol protocol, uint64_t stream_id):             # <<<<<<<<<<<<<<
 *         cdef Db db = Db.__new__(Db)
 *         db._stream_id = stream_id
 */

static CYTHON_INLINE struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_f_8asynctnt_6iproto_8protocol_2Db_create(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_protocol, uint64_t __pyx_v_stream_id) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_db = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create", 1);

  /* "asynctnt/iproto/db.pyx":13
 *     @staticmethod
 *     cdef inline Db create(BaseProtocol protocol, uint64_t stream_id):
 *         cdef Db db = Db.__new__(Db)             # <<<<<<<<<<<<<<
 *         db._stream_id = stream_id
 *         db._protocol = protocol
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Db(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 13, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_db = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":14
 *     cdef inline Db create(BaseProtocol protocol, uint64_t stream_id):
 *         cdef Db db = Db.__new__(Db)
 *         db._stream_id = stream_id             # <<<<<<<<<<<<<<
 *         db._protocol = protocol
 *         db._encoding = protocol.encoding
 */
  __pyx_v_db->_stream_id = __pyx_v_stream_id;

  /* "asynctnt/iproto/db.pyx":15
 *         cdef Db db = Db.__new__(Db)
 *         db._stream_id = stream_id
 *         db._protocol = protocol             # <<<<<<<<<<<<<<
 *         db._encoding = protocol.encoding
 *         return db
 */
  __Pyx_INCREF((PyObject *)__pyx_v_protocol);
  __Pyx_GIVEREF((PyObject *)__pyx_v_protocol);
  __Pyx_GOTREF((PyObject *)__pyx_v_db->_protocol);
  __Pyx_DECREF((PyObject *)__pyx_v_db->_protocol);
  __pyx_v_db->_protocol = __pyx_v_protocol;

  /* "asynctnt/iproto/db.pyx":16
 *         db._stream_id = stream_id
 *         db._protocol = protocol
 *         db._encoding = protocol.encoding             # <<<<<<<<<<<<<<
 *         return db
 * 
 */
  __pyx_t_1 = __pyx_v_protocol->__pyx_base.encoding;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_db->_encoding);
  __Pyx_DECREF(__pyx_v_db->_encoding);
  __pyx_v_db->_encoding = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":17
 *         db._protocol = protocol
 *         db._encoding = protocol.encoding
 *         return db             # <<<<<<<<<<<<<<
 * 
 *     cdef inline uint64_t next_sync(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_db);
  __pyx_r = __pyx_v_db;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":12
 * 
 *     @staticmethod
 *     cdef inline Db create(BaseProtocol protocol, uint64_t stream_id):             # <<<<<<<<<<<<<<
 *         cdef Db db = Db.__new__(Db)
 *         db._stream_id = stream_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_db);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":19
 *         return db
 * 
 *     cdef inline uint64_t next_sync(self):             # <<<<<<<<<<<<<<
 *         return self._protocol.next_sync()
 * 
 */

static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self) {
  uint64_t __pyx_r;
  uint64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/db.pyx":20
 * 
 *     cdef inline uint64_t next_sync(self):
 *         return self._protocol.next_sync()             # <<<<<<<<<<<<<<
 * 
 *     cdef object _ping(self, float timeout):
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_next_sync(__pyx_v_self->_protocol); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 20, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":19
 *         return db
 * 
 *     cdef inline uint64_t next_sync(self):             # <<<<<<<<<<<<<<
 *         return self._protocol.next_sync()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.next_sync", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":22
 *         return self._protocol.next_sync()
 * 
 *     cdef object _ping(self, float timeout):             # <<<<<<<<<<<<<<
 *         cdef PingRequest req = PingRequest.__new__(PingRequest)
 *         req.op = tarantool.IPROTO_PING
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__ping(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ping", 1);

  /* "asynctnt/iproto/db.pyx":23
 * 
 *     cdef object _ping(self, float timeout):
 *         cdef PingRequest req = PingRequest.__new__(PingRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_PING
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_PingRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 23, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":24
 *     cdef object _ping(self, float timeout):
 *         cdef PingRequest req = PingRequest.__new__(PingRequest)
 *         req.op = tarantool.IPROTO_PING             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_PING;

  /* "asynctnt/iproto/db.pyx":25
 *         cdef PingRequest req = PingRequest.__new__(PingRequest)
 *         req.op = tarantool.IPROTO_PING
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 25, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":26
 *         req.op = tarantool.IPROTO_PING
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":27
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":28
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _id(self, float timeout):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":22
 *         return self._protocol.next_sync()
 * 
 *     cdef object _ping(self, float timeout):             # <<<<<<<<<<<<<<
 *         cdef PingRequest req = PingRequest.__new__(PingRequest)
 *         req.op = tarantool.IPROTO_PING
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._ping", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":30
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _id(self, float timeout):             # <<<<<<<<<<<<<<
 *         cdef IDRequest req = IDRequest.__new__(IDRequest)
 *         req.op = tarantool.IPROTO_ID
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__id(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_id", 1);

  /* "asynctnt/iproto/db.pyx":31
 * 
 *     cdef object _id(self, float timeout):
 *         cdef IDRequest req = IDRequest.__new__(IDRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_ID
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_IDRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 31, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":32
 *     cdef object _id(self, float timeout):
 *         cdef IDRequest req = IDRequest.__new__(IDRequest)
 *         req.op = tarantool.IPROTO_ID             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_ID;

  /* "asynctnt/iproto/db.pyx":33
 *         cdef IDRequest req = IDRequest.__new__(IDRequest)
 *         req.op = tarantool.IPROTO_ID
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = False
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 33, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":34
 *         req.op = tarantool.IPROTO_ID
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.check_schema_change = False
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":35
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = False             # <<<<<<<<<<<<<<
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 */
  __pyx_v_req->__pyx_base.check_schema_change = 0;

  /* "asynctnt/iproto/db.pyx":36
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = False
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _auth(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":30
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _id(self, float timeout):             # <<<<<<<<<<<<<<
 *         cdef IDRequest req = IDRequest.__new__(IDRequest)
 *         req.op = tarantool.IPROTO_ID
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":38
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _auth(self,             # <<<<<<<<<<<<<<
 *                       bytes salt,
 *                       str username,
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__auth(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_salt, PyObject *__pyx_v_username, PyObject *__pyx_v_password, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_auth", 1);

  /* "asynctnt/iproto/db.pyx":46
 *             AuthRequest req
 * 
 *         req = AuthRequest.__new__(AuthRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_AUTH
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_AuthRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 46, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":47
 * 
 *         req = AuthRequest.__new__(AuthRequest)
 *         req.op = tarantool.IPROTO_AUTH             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_AUTH;

  /* "asynctnt/iproto/db.pyx":48
 *         req = AuthRequest.__new__(AuthRequest)
 *         req.op = tarantool.IPROTO_AUTH
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.salt = salt
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 48, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":49
 *         req.op = tarantool.IPROTO_AUTH
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.salt = salt
 *         req.username = username
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":50
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.salt = salt             # <<<<<<<<<<<<<<
 *         req.username = username
 *         req.password = password
 */
  __Pyx_INCREF(__pyx_v_salt);
  __Pyx_GIVEREF(__pyx_v_salt);
  __Pyx_GOTREF(__pyx_v_req->salt);
  __Pyx_DECREF(__pyx_v_req->salt);
  __pyx_v_req->salt = __pyx_v_salt;

  /* "asynctnt/iproto/db.pyx":51
 *         req.stream_id = self._stream_id
 *         req.salt = salt
 *         req.username = username             # <<<<<<<<<<<<<<
 *         req.password = password
 *         req.push_subscribe = False
 */
  __Pyx_INCREF(__pyx_v_username);
  __Pyx_GIVEREF(__pyx_v_username);
  __Pyx_GOTREF(__pyx_v_req->username);
  __Pyx_DECREF(__pyx_v_req->username);
  __pyx_v_req->username = __pyx_v_username;

  /* "asynctnt/iproto/db.pyx":52
 *         req.salt = salt
 *         req.username = username
 *         req.password = password             # <<<<<<<<<<<<<<
 *         req.push_subscribe = False
 *         req.parse_as_tuples = False
 */
  __Pyx_INCREF(__pyx_v_password);
  __Pyx_GIVEREF(__pyx_v_password);
  __Pyx_GOTREF(__pyx_v_req->password);
  __Pyx_DECREF(__pyx_v_req->password);
  __pyx_v_req->password = __pyx_v_password;

  /* "asynctnt/iproto/db.pyx":53
 *         req.username = username
 *         req.password = password
 *         req.push_subscribe = False             # <<<<<<<<<<<<<<
 *         req.parse_as_tuples = False
 *         req.parse_metadata = False
 */
  __pyx_v_req->__pyx_base.push_subscribe = 0;

  /* "asynctnt/iproto/db.pyx":54
 *         req.password = password
 *         req.push_subscribe = False
 *         req.parse_as_tuples = False             # <<<<<<<<<<<<<<
 *         req.parse_metadata = False
 *         req.check_schema_change = False
 */
  __pyx_v_req->__pyx_base.parse_as_tuples = 0;

  /* "asynctnt/iproto/db.pyx":55
 *         req.push_subscribe = False
 *         req.parse_as_tuples = False
 *         req.parse_metadata = False             # <<<<<<<<<<<<<<
 *         req.check_schema_change = False
 * 
 */
  __pyx_v_req->__pyx_base.parse_metadata = 0;

  /* "asynctnt/iproto/db.pyx":56
 *         req.parse_as_tuples = False
 *         req.parse_metadata = False
 *         req.check_schema_change = False             # <<<<<<<<<<<<<<
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->__pyx_base.check_schema_change = 0;

  /* "asynctnt/iproto/db.pyx":58
 *         req.check_schema_change = False
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _call(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":38
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _auth(self,             # <<<<<<<<<<<<<<
 *                       bytes salt,
 *                       str username,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._auth", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":60
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _call(self,             # <<<<<<<<<<<<<<
 *                       tarantool.iproto_type op,
 *                       str func_name,
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__call(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_v_op, PyObject *__pyx_v_func_name, PyObject *__pyx_v_args, float __pyx_v_timeout, int __pyx_v_push_subscribe) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_call", 1);

  /* "asynctnt/iproto/db.pyx":66
 *                       float timeout,
 *                       bint push_subscribe):
 *         cdef CallRequest req = CallRequest.__new__(CallRequest)             # <<<<<<<<<<<<<<
 *         req.op = op
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_CallRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 66, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":67
 *                       bint push_subscribe):
 *         cdef CallRequest req = CallRequest.__new__(CallRequest)
 *         req.op = op             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_v_op;

  /* "asynctnt/iproto/db.pyx":68
 *         cdef CallRequest req = CallRequest.__new__(CallRequest)
 *         req.op = op
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.func_name = func_name
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 68, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":69
 *         req.op = op
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.func_name = func_name
 *         req.args = args
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":70
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.func_name = func_name             # <<<<<<<<<<<<<<
 *         req.args = args
 *         req.push_subscribe = push_subscribe
 */
  __Pyx_INCREF(__pyx_v_func_name);
  __Pyx_GIVEREF(__pyx_v_func_name);
  __Pyx_GOTREF(__pyx_v_req->func_name);
  __Pyx_DECREF(__pyx_v_req->func_name);
  __pyx_v_req->func_name = __pyx_v_func_name;

  /* "asynctnt/iproto/db.pyx":71
 *         req.stream_id = self._stream_id
 *         req.func_name = func_name
 *         req.args = args             # <<<<<<<<<<<<<<
 *         req.push_subscribe = push_subscribe
 *         req.check_schema_change = True
 */
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  __Pyx_GOTREF(__pyx_v_req->args);
  __Pyx_DECREF(__pyx_v_req->args);
  __pyx_v_req->args = __pyx_v_args;

  /* "asynctnt/iproto/db.pyx":72
 *         req.func_name = func_name
 *         req.args = args
 *         req.push_subscribe = push_subscribe             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->__pyx_base.push_subscribe = __pyx_v_push_subscribe;

  /* "asynctnt/iproto/db.pyx":73
 *         req.args = args
 *         req.push_subscribe = push_subscribe
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":74
 *         req.push_subscribe = push_subscribe
 *         req.check_schema_change = True
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _eval(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":60
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _call(self,             # <<<<<<<<<<<<<<
 *                       tarantool.iproto_type op,
 *                       str func_name,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":76
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _eval(self,             # <<<<<<<<<<<<<<
 *                       str expression,
 *                       object args,
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__eval(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_expression, PyObject *__pyx_v_args, float __pyx_v_timeout, int __pyx_v_push_subscribe) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_eval", 1);

  /* "asynctnt/iproto/db.pyx":81
 *                       float timeout,
 *                       bint push_subscribe):
 *         cdef EvalRequest req = EvalRequest.__new__(EvalRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_EVAL
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_EvalRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 81, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":82
 *                       bint push_subscribe):
 *         cdef EvalRequest req = EvalRequest.__new__(EvalRequest)
 *         req.op = tarantool.IPROTO_EVAL             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_EVAL;

  /* "asynctnt/iproto/db.pyx":83
 *         cdef EvalRequest req = EvalRequest.__new__(EvalRequest)
 *         req.op = tarantool.IPROTO_EVAL
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.expression = expression
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 83, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":84
 *         req.op = tarantool.IPROTO_EVAL
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.expression = expression
 *         req.args = args
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":85
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.expression = expression             # <<<<<<<<<<<<<<
 *         req.args = args
 *         req.push_subscribe = push_subscribe
 */
  __Pyx_INCREF(__pyx_v_expression);
  __Pyx_GIVEREF(__pyx_v_expression);
  __Pyx_GOTREF(__pyx_v_req->expression);
  __Pyx_DECREF(__pyx_v_req->expression);
  __pyx_v_req->expression = __pyx_v_expression;

  /* "asynctnt/iproto/db.pyx":86
 *         req.stream_id = self._stream_id
 *         req.expression = expression
 *         req.args = args             # <<<<<<<<<<<<<<
 *         req.push_subscribe = push_subscribe
 *         req.check_schema_change = True
 */
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  __Pyx_GOTREF(__pyx_v_req->args);
  __Pyx_DECREF(__pyx_v_req->args);
  __pyx_v_req->args = __pyx_v_args;

  /* "asynctnt/iproto/db.pyx":87
 *         req.expression = expression
 *         req.args = args
 *         req.push_subscribe = push_subscribe             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->__pyx_base.push_subscribe = __pyx_v_push_subscribe;

  /* "asynctnt/iproto/db.pyx":88
 *         req.args = args
 *         req.push_subscribe = push_subscribe
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":89
 *         req.push_subscribe = push_subscribe
 *         req.check_schema_change = True
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _select(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":76
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _eval(self,             # <<<<<<<<<<<<<<
 *                       str expression,
 *                       object args,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._eval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":91
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _select(self,             # <<<<<<<<<<<<<<
 *                         object space,
 *                         object index,
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__select(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_index, PyObject *__pyx_v_key, uint64_t __pyx_v_offset, uint64_t __pyx_v_limit, PyObject *__pyx_v_iterator, float __pyx_v_timeout, int __pyx_v_check_schema_change) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_sp = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_idx = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *__pyx_v_req = 0;
  uint32_t __pyx_v_iterator_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint32_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  uint64_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_select", 1);

  /* "asynctnt/iproto/db.pyx":106
 *             uint32_t iterator_value
 * 
 *         sp = self._protocol._schema.get_or_create_space(space)             # <<<<<<<<<<<<<<
 *         idx = sp.get_index(index)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_or_create_space(__pyx_v_self->_protocol->_schema, __pyx_v_space)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sp = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":107
 * 
 *         sp = self._protocol._schema.get_or_create_space(space)
 *         idx = sp.get_index(index)             # <<<<<<<<<<<<<<
 * 
 *         iterator_value = self._protocol.transform_iterator(iterator)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index(__pyx_v_sp, __pyx_v_index, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_idx = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":109
 *         idx = sp.get_index(index)
 * 
 *         iterator_value = self._protocol.transform_iterator(iterator)             # <<<<<<<<<<<<<<
 *         if key is None and isinstance(iterator, int):
 *             iterator_value = <uint32_t> iterator
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->_protocol->__pyx_base.__pyx_vtab)->transform_iterator(__pyx_v_self->_protocol, __pyx_v_iterator); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 109, __pyx_L1_error)
  __pyx_v_iterator_value = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":110
 * 
 *         iterator_value = self._protocol.transform_iterator(iterator)
 *         if key is None and isinstance(iterator, int):             # <<<<<<<<<<<<<<
 *             iterator_value = <uint32_t> iterator
 *             if iterator_value == 0:
 */
  __pyx_t_4 = (__pyx_v_key == Py_None);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyInt_Check(__pyx_v_iterator); 
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "asynctnt/iproto/db.pyx":111
 *         iterator_value = self._protocol.transform_iterator(iterator)
 *         if key is None and isinstance(iterator, int):
 *             iterator_value = <uint32_t> iterator             # <<<<<<<<<<<<<<
 *             if iterator_value == 0:
 *                 iterator_value = 2  # ALL
 */
    __pyx_t_2 = __Pyx_PyInt_As_uint32_t(__pyx_v_iterator); if (unlikely((__pyx_t_2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(26, 111, __pyx_L1_error)
    __pyx_v_iterator_value = ((uint32_t)__pyx_t_2);

    /* "asynctnt/iproto/db.pyx":112
 *         if key is None and isinstance(iterator, int):
 *             iterator_value = <uint32_t> iterator
 *             if iterator_value == 0:             # <<<<<<<<<<<<<<
 *                 iterator_value = 2  # ALL
 * 
 */
    __pyx_t_3 = (__pyx_v_iterator_value == 0);
    if (__pyx_t_3) {

      /* "asynctnt/iproto/db.pyx":113
 *             iterator_value = <uint32_t> iterator
 *             if iterator_value == 0:
 *                 iterator_value = 2  # ALL             # <<<<<<<<<<<<<<
 * 
 *         req = SelectRequest.__new__(SelectRequest)
 */
      __pyx_v_iterator_value = 2;

      /* "asynctnt/iproto/db.pyx":112
 *         if key is None and isinstance(iterator, int):
 *             iterator_value = <uint32_t> iterator
 *             if iterator_value == 0:             # <<<<<<<<<<<<<<
 *                 iterator_value = 2  # ALL
 * 
 */
    }

    /* "asynctnt/iproto/db.pyx":110
 * 
 *         iterator_value = self._protocol.transform_iterator(iterator)
 *         if key is None and isinstance(iterator, int):             # <<<<<<<<<<<<<<
 *             iterator_value = <uint32_t> iterator
 *             if iterator_value == 0:
 */
  }

  /* "asynctnt/iproto/db.pyx":115
 *                 iterator_value = 2  # ALL
 * 
 *         req = SelectRequest.__new__(SelectRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_SELECT
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_SelectRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 115, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":116
 * 
 *         req = SelectRequest.__new__(SelectRequest)
 *         req.op = tarantool.IPROTO_SELECT             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_SELECT;

  /* "asynctnt/iproto/db.pyx":117
 *         req = SelectRequest.__new__(SelectRequest)
 *         req.op = tarantool.IPROTO_SELECT
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.space = sp
 */
  __pyx_t_5 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 117, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_5;

  /* "asynctnt/iproto/db.pyx":118
 *         req.op = tarantool.IPROTO_SELECT
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.space = sp
 *         req.index = idx
 */
  __pyx_t_5 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_5;

  /* "asynctnt/iproto/db.pyx":119
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.space = sp             # <<<<<<<<<<<<<<
 *         req.index = idx
 *         req.key = key
 */
  __Pyx_INCREF((PyObject *)__pyx_v_sp);
  __Pyx_GIVEREF((PyObject *)__pyx_v_sp);
  __Pyx_GOTREF((PyObject *)__pyx_v_req->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v_req->__pyx_base.space);
  __pyx_v_req->__pyx_base.space = __pyx_v_sp;

  /* "asynctnt/iproto/db.pyx":120
 *         req.stream_id = self._stream_id
 *         req.space = sp
 *         req.index = idx             # <<<<<<<<<<<<<<
 *         req.key = key
 *         req.offset = offset
 */
  __Pyx_INCREF((PyObject *)__pyx_v_idx);
  __Pyx_GIVEREF((PyObject *)__pyx_v_idx);
  __Pyx_GOTREF((PyObject *)__pyx_v_req->index);
  __Pyx_DECREF((PyObject *)__pyx_v_req->index);
  __pyx_v_req->index = __pyx_v_idx;

  /* "asynctnt/iproto/db.pyx":121
 *         req.space = sp
 *         req.index = idx
 *         req.key = key             # <<<<<<<<<<<<<<
 *         req.offset = offset
 *         req.limit = limit
 */
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  __Pyx_GOTREF(__pyx_v_req->key);
  __Pyx_DECREF(__pyx_v_req->key);
  __pyx_v_req->key = __pyx_v_key;

  /* "asynctnt/iproto/db.pyx":122
 *         req.index = idx
 *         req.key = key
 *         req.offset = offset             # <<<<<<<<<<<<<<
 *         req.limit = limit
 *         req.iterator = iterator_value
 */
  __pyx_v_req->offset = __pyx_v_offset;

  /* "asynctnt/iproto/db.pyx":123
 *         req.key = key
 *         req.offset = offset
 *         req.limit = limit             # <<<<<<<<<<<<<<
 *         req.iterator = iterator_value
 *         req.push_subscribe = False
 */
  __pyx_v_req->limit = __pyx_v_limit;

  /* "asynctnt/iproto/db.pyx":124
 *         req.offset = offset
 *         req.limit = limit
 *         req.iterator = iterator_value             # <<<<<<<<<<<<<<
 *         req.push_subscribe = False
 *         req.check_schema_change = check_schema_change
 */
  __pyx_v_req->iterator = __pyx_v_iterator_value;

  /* "asynctnt/iproto/db.pyx":125
 *         req.limit = limit
 *         req.iterator = iterator_value
 *         req.push_subscribe = False             # <<<<<<<<<<<<<<
 *         req.check_schema_change = check_schema_change
 *         req.parse_as_tuples = True
 */
  __pyx_v_req->__pyx_base.push_subscribe = 0;

  /* "asynctnt/iproto/db.pyx":126
 *         req.iterator = iterator_value
 *         req.push_subscribe = False
 *         req.check_schema_change = check_schema_change             # <<<<<<<<<<<<<<
 *         req.parse_as_tuples = True
 * 
 */
  __pyx_v_req->__pyx_base.check_schema_change = __pyx_v_check_schema_change;

  /* "asynctnt/iproto/db.pyx":127
 *         req.push_subscribe = False
 *         req.check_schema_change = check_schema_change
 *         req.parse_as_tuples = True             # <<<<<<<<<<<<<<
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->__pyx_base.parse_as_tuples = 1;

  /* "asynctnt/iproto/db.pyx":129
 *         req.parse_as_tuples = True
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _insert(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_6)) __PYX_ERR(26, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":91
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _select(self,             # <<<<<<<<<<<<<<
 *                         object space,
 *                         object index,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._select", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sp);
  __Pyx_XDECREF((PyObject *)__pyx_v_idx);
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":131
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _insert(self,             # <<<<<<<<<<<<<<
 *                         object space,
 *                         object t,
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__insert(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_t, int __pyx_v_replace, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_sp = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_2;
  uint64_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_insert", 1);

  /* "asynctnt/iproto/db.pyx":140
 *             InsertRequest req
 * 
 *         sp = self._protocol._schema.get_or_create_space(space)             # <<<<<<<<<<<<<<
 * 
 *         req = InsertRequest.__new__(InsertRequest)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_or_create_space(__pyx_v_self->_protocol->_schema, __pyx_v_space)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sp = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":142
 *         sp = self._protocol._schema.get_or_create_space(space)
 * 
 *         req = InsertRequest.__new__(InsertRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_REPLACE if replace else tarantool.IPROTO_INSERT
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_InsertRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 142, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":143
 * 
 *         req = InsertRequest.__new__(InsertRequest)
 *         req.op = tarantool.IPROTO_REPLACE if replace else tarantool.IPROTO_INSERT             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  if (__pyx_v_replace) {
    __pyx_t_2 = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_REPLACE;
  } else {
    __pyx_t_2 = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_INSERT;
  }
  __pyx_v_req->__pyx_base.op = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":144
 *         req = InsertRequest.__new__(InsertRequest)
 *         req.op = tarantool.IPROTO_REPLACE if replace else tarantool.IPROTO_INSERT
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.space = sp
 */
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 144, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_3;

  /* "asynctnt/iproto/db.pyx":145
 *         req.op = tarantool.IPROTO_REPLACE if replace else tarantool.IPROTO_INSERT
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.space = sp
 *         req.t = t
 */
  __pyx_t_3 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_3;

  /* "asynctnt/iproto/db.pyx":146
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.space = sp             # <<<<<<<<<<<<<<
 *         req.t = t
 *         req.push_subscribe = False
 */
  __Pyx_INCREF((PyObject *)__pyx_v_sp);
  __Pyx_GIVEREF((PyObject *)__pyx_v_sp);
  __Pyx_GOTREF((PyObject *)__pyx_v_req->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v_req->__pyx_base.space);
  __pyx_v_req->__pyx_base.space = __pyx_v_sp;

  /* "asynctnt/iproto/db.pyx":147
 *         req.stream_id = self._stream_id
 *         req.space = sp
 *         req.t = t             # <<<<<<<<<<<<<<
 *         req.push_subscribe = False
 *         req.check_schema_change = True
 */
  __Pyx_INCREF(__pyx_v_t);
  __Pyx_GIVEREF(__pyx_v_t);
  __Pyx_GOTREF(__pyx_v_req->t);
  __Pyx_DECREF(__pyx_v_req->t);
  __pyx_v_req->t = __pyx_v_t;

  /* "asynctnt/iproto/db.pyx":148
 *         req.space = sp
 *         req.t = t
 *         req.push_subscribe = False             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True
 */
  __pyx_v_req->__pyx_base.push_subscribe = 0;

  /* "asynctnt/iproto/db.pyx":149
 *         req.t = t
 *         req.push_subscribe = False
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         req.parse_as_tuples = True
 * 
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":150
 *         req.push_subscribe = False
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True             # <<<<<<<<<<<<<<
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->__pyx_base.parse_as_tuples = 1;

  /* "asynctnt/iproto/db.pyx":152
 *         req.parse_as_tuples = True
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _delete(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_4 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":131
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _insert(self,             # <<<<<<<<<<<<<<
 *                         object space,
 *                         object t,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sp);
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":154
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _delete(self,             # <<<<<<<<<<<<<<
 *                         object space,
 *                         object index,
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__delete(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_index, PyObject *__pyx_v_key, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_sp = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_idx = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_delete", 1);

  /* "asynctnt/iproto/db.pyx":164
 *             DeleteRequest req
 * 
 *         sp = self._protocol._schema.get_or_create_space(space)             # <<<<<<<<<<<<<<
 *         idx = sp.get_index(index)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_or_create_space(__pyx_v_self->_protocol->_schema, __pyx_v_space)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sp = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":165
 * 
 *         sp = self._protocol._schema.get_or_create_space(space)
 *         idx = sp.get_index(index)             # <<<<<<<<<<<<<<
 * 
 *         req = DeleteRequest.__new__(DeleteRequest)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index(__pyx_v_sp, __pyx_v_index, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_idx = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":167
 *         idx = sp.get_index(index)
 * 
 *         req = DeleteRequest.__new__(DeleteRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_DELETE
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_DeleteRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 167, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":168
 * 
 *         req = DeleteRequest.__new__(DeleteRequest)
 *         req.op = tarantool.IPROTO_DELETE             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_DELETE;

  /* "asynctnt/iproto/db.pyx":169
 *         req = DeleteRequest.__new__(DeleteRequest)
 *         req.op = tarantool.IPROTO_DELETE
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.space = sp
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 169, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":170
 *         req.op = tarantool.IPROTO_DELETE
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.space = sp
 *         req.index = idx
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":171
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.space = sp             # <<<<<<<<<<<<<<
 *         req.index = idx
 *         req.key = key
 */
  __Pyx_INCREF((PyObject *)__pyx_v_sp);
  __Pyx_GIVEREF((PyObject *)__pyx_v_sp);
  __Pyx_GOTREF((PyObject *)__pyx_v_req->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v_req->__pyx_base.space);
  __pyx_v_req->__pyx_base.space = __pyx_v_sp;

  /* "asynctnt/iproto/db.pyx":172
 *         req.stream_id = self._stream_id
 *         req.space = sp
 *         req.index = idx             # <<<<<<<<<<<<<<
 *         req.key = key
 *         req.push_subscribe = False
 */
  __Pyx_INCREF((PyObject *)__pyx_v_idx);
  __Pyx_GIVEREF((PyObject *)__pyx_v_idx);
  __Pyx_GOTREF((PyObject *)__pyx_v_req->index);
  __Pyx_DECREF((PyObject *)__pyx_v_req->index);
  __pyx_v_req->index = __pyx_v_idx;

  /* "asynctnt/iproto/db.pyx":173
 *         req.space = sp
 *         req.index = idx
 *         req.key = key             # <<<<<<<<<<<<<<
 *         req.push_subscribe = False
 *         req.check_schema_change = True
 */
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  __Pyx_GOTREF(__pyx_v_req->key);
  __Pyx_DECREF(__pyx_v_req->key);
  __pyx_v_req->key = __pyx_v_key;

  /* "asynctnt/iproto/db.pyx":174
 *         req.index = idx
 *         req.key = key
 *         req.push_subscribe = False             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True
 */
  __pyx_v_req->__pyx_base.push_subscribe = 0;

  /* "asynctnt/iproto/db.pyx":175
 *         req.key = key
 *         req.push_subscribe = False
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         req.parse_as_tuples = True
 * 
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":176
 *         req.push_subscribe = False
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True             # <<<<<<<<<<<<<<
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->__pyx_base.parse_as_tuples = 1;

  /* "asynctnt/iproto/db.pyx":178
 *         req.parse_as_tuples = True
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _update(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":154
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _delete(self,             # <<<<<<<<<<<<<<
 *                         object space,
 *                         object index,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._delete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sp);
  __Pyx_XDECREF((PyObject *)__pyx_v_idx);
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":180
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _update(self,             # <<<<<<<<<<<<<<
 *                         object space,
 *                         object index,
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__update(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_index, PyObject *__pyx_v_key, PyObject *__pyx_v_operations, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_sp = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *__pyx_v_idx = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_update", 1);

  /* "asynctnt/iproto/db.pyx":191
 *             UpdateRequest req
 * 
 *         sp = self._protocol._schema.get_or_create_space(space)             # <<<<<<<<<<<<<<
 *         idx = sp.get_index(index)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_or_create_space(__pyx_v_self->_protocol->_schema, __pyx_v_space)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sp = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":192
 * 
 *         sp = self._protocol._schema.get_or_create_space(space)
 *         idx = sp.get_index(index)             # <<<<<<<<<<<<<<
 * 
 *         req = UpdateRequest.__new__(UpdateRequest)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index(__pyx_v_sp, __pyx_v_index, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_idx = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":194
 *         idx = sp.get_index(index)
 * 
 *         req = UpdateRequest.__new__(UpdateRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_UPDATE
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_UpdateRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 194, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":195
 * 
 *         req = UpdateRequest.__new__(UpdateRequest)
 *         req.op = tarantool.IPROTO_UPDATE             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_UPDATE;

  /* "asynctnt/iproto/db.pyx":196
 *         req = UpdateRequest.__new__(UpdateRequest)
 *         req.op = tarantool.IPROTO_UPDATE
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.space = sp
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 196, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":197
 *         req.op = tarantool.IPROTO_UPDATE
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.space = sp
 *         req.index = idx
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":198
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.space = sp             # <<<<<<<<<<<<<<
 *         req.index = idx
 *         req.key = key
 */
  __Pyx_INCREF((PyObject *)__pyx_v_sp);
  __Pyx_GIVEREF((PyObject *)__pyx_v_sp);
  __Pyx_GOTREF((PyObject *)__pyx_v_req->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v_req->__pyx_base.space);
  __pyx_v_req->__pyx_base.space = __pyx_v_sp;

  /* "asynctnt/iproto/db.pyx":199
 *         req.stream_id = self._stream_id
 *         req.space = sp
 *         req.index = idx             # <<<<<<<<<<<<<<
 *         req.key = key
 *         req.operations = operations
 */
  __Pyx_INCREF((PyObject *)__pyx_v_idx);
  __Pyx_GIVEREF((PyObject *)__pyx_v_idx);
  __Pyx_GOTREF((PyObject *)__pyx_v_req->index);
  __Pyx_DECREF((PyObject *)__pyx_v_req->index);
  __pyx_v_req->index = __pyx_v_idx;

  /* "asynctnt/iproto/db.pyx":200
 *         req.space = sp
 *         req.index = idx
 *         req.key = key             # <<<<<<<<<<<<<<
 *         req.operations = operations
 *         req.push_subscribe = False
 */
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  __Pyx_GOTREF(__pyx_v_req->key);
  __Pyx_DECREF(__pyx_v_req->key);
  __pyx_v_req->key = __pyx_v_key;

  /* "asynctnt/iproto/db.pyx":201
 *         req.index = idx
 *         req.key = key
 *         req.operations = operations             # <<<<<<<<<<<<<<
 *         req.push_subscribe = False
 *         req.check_schema_change = True
 */
  __Pyx_INCREF(__pyx_v_operations);
  __Pyx_GIVEREF(__pyx_v_operations);
  __Pyx_GOTREF(__pyx_v_req->operations);
  __Pyx_DECREF(__pyx_v_req->operations);
  __pyx_v_req->operations = __pyx_v_operations;

  /* "asynctnt/iproto/db.pyx":202
 *         req.key = key
 *         req.operations = operations
 *         req.push_subscribe = False             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True
 */
  __pyx_v_req->__pyx_base.push_subscribe = 0;

  /* "asynctnt/iproto/db.pyx":203
 *         req.operations = operations
 *         req.push_subscribe = False
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         req.parse_as_tuples = True
 * 
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":204
 *         req.push_subscribe = False
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True             # <<<<<<<<<<<<<<
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->__pyx_base.parse_as_tuples = 1;

  /* "asynctnt/iproto/db.pyx":206
 *         req.parse_as_tuples = True
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _upsert(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":180
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _update(self,             # <<<<<<<<<<<<<<
 *                         object space,
 *                         object index,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sp);
  __Pyx_XDECREF((PyObject *)__pyx_v_idx);
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":208
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _upsert(self,             # <<<<<<<<<<<<<<
 *                         object space,
 *                         object t,
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__upsert(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_t, PyObject *__pyx_v_operations, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *__pyx_v_sp = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_upsert", 1);

  /* "asynctnt/iproto/db.pyx":217
 *             UpsertRequest req
 * 
 *         sp = self._protocol._schema.get_or_create_space(space)             # <<<<<<<<<<<<<<
 * 
 *         req = UpsertRequest.__new__(UpsertRequest)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_or_create_space(__pyx_v_self->_protocol->_schema, __pyx_v_space)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sp = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":219
 *         sp = self._protocol._schema.get_or_create_space(space)
 * 
 *         req = UpsertRequest.__new__(UpsertRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_UPSERT
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_UpsertRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 219, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":220
 * 
 *         req = UpsertRequest.__new__(UpsertRequest)
 *         req.op = tarantool.IPROTO_UPSERT             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_UPSERT;

  /* "asynctnt/iproto/db.pyx":221
 *         req = UpsertRequest.__new__(UpsertRequest)
 *         req.op = tarantool.IPROTO_UPSERT
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.space = sp
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 221, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":222
 *         req.op = tarantool.IPROTO_UPSERT
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.space = sp
 *         req.t = t
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":223
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.space = sp             # <<<<<<<<<<<<<<
 *         req.t = t
 *         req.operations = operations
 */
  __Pyx_INCREF((PyObject *)__pyx_v_sp);
  __Pyx_GIVEREF((PyObject *)__pyx_v_sp);
  __Pyx_GOTREF((PyObject *)__pyx_v_req->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v_req->__pyx_base.space);
  __pyx_v_req->__pyx_base.space = __pyx_v_sp;

  /* "asynctnt/iproto/db.pyx":224
 *         req.stream_id = self._stream_id
 *         req.space = sp
 *         req.t = t             # <<<<<<<<<<<<<<
 *         req.operations = operations
 *         req.push_subscribe = False
 */
  __Pyx_INCREF(__pyx_v_t);
  __Pyx_GIVEREF(__pyx_v_t);
  __Pyx_GOTREF(__pyx_v_req->t);
  __Pyx_DECREF(__pyx_v_req->t);
  __pyx_v_req->t = __pyx_v_t;

  /* "asynctnt/iproto/db.pyx":225
 *         req.space = sp
 *         req.t = t
 *         req.operations = operations             # <<<<<<<<<<<<<<
 *         req.push_subscribe = False
 *         req.check_schema_change = True
 */
  __Pyx_INCREF(__pyx_v_operations);
  __Pyx_GIVEREF(__pyx_v_operations);
  __Pyx_GOTREF(__pyx_v_req->operations);
  __Pyx_DECREF(__pyx_v_req->operations);
  __pyx_v_req->operations = __pyx_v_operations;

  /* "asynctnt/iproto/db.pyx":226
 *         req.t = t
 *         req.operations = operations
 *         req.push_subscribe = False             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True
 */
  __pyx_v_req->__pyx_base.push_subscribe = 0;

  /* "asynctnt/iproto/db.pyx":227
 *         req.operations = operations
 *         req.push_subscribe = False
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         req.parse_as_tuples = True
 * 
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":228
 *         req.push_subscribe = False
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True             # <<<<<<<<<<<<<<
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->__pyx_base.parse_as_tuples = 1;

  /* "asynctnt/iproto/db.pyx":230
 *         req.parse_as_tuples = True
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _execute(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":208
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _upsert(self,             # <<<<<<<<<<<<<<
 *                         object space,
 *                         object t,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._upsert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sp);
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":232
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _execute(self,             # <<<<<<<<<<<<<<
 *                          object query,
 *                          object args,
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__execute(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_query, PyObject *__pyx_v_args, int __pyx_v_parse_metadata, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_execute", 1);

  /* "asynctnt/iproto/db.pyx":240
 *             ExecuteRequest req
 * 
 *         req = ExecuteRequest.__new__(ExecuteRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_EXECUTE
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_ExecuteRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 240, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":241
 * 
 *         req = ExecuteRequest.__new__(ExecuteRequest)
 *         req.op = tarantool.IPROTO_EXECUTE             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_EXECUTE;

  /* "asynctnt/iproto/db.pyx":242
 *         req = ExecuteRequest.__new__(ExecuteRequest)
 *         req.op = tarantool.IPROTO_EXECUTE
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 * 
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 242, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":243
 *         req.op = tarantool.IPROTO_EXECUTE
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(query, str):
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":245
 *         req.stream_id = self._stream_id
 * 
 *         if isinstance(query, str):             # <<<<<<<<<<<<<<
 *             req.query = query
 *             req.statement_id = 0
 */
  __pyx_t_3 = PyUnicode_Check(__pyx_v_query); 
  if (__pyx_t_3) {

    /* "asynctnt/iproto/db.pyx":246
 * 
 *         if isinstance(query, str):
 *             req.query = query             # <<<<<<<<<<<<<<
 *             req.statement_id = 0
 *         elif isinstance(query, int):
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_v_query))||((__pyx_v_query) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_query))) __PYX_ERR(26, 246, __pyx_L1_error)
    __pyx_t_1 = __pyx_v_query;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_req->query);
    __Pyx_DECREF(__pyx_v_req->query);
    __pyx_v_req->query = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/db.pyx":247
 *         if isinstance(query, str):
 *             req.query = query
 *             req.statement_id = 0             # <<<<<<<<<<<<<<
 *         elif isinstance(query, int):
 *             req.query = None
 */
    __pyx_v_req->statement_id = 0;

    /* "asynctnt/iproto/db.pyx":245
 *         req.stream_id = self._stream_id
 * 
 *         if isinstance(query, str):             # <<<<<<<<<<<<<<
 *             req.query = query
 *             req.statement_id = 0
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/db.pyx":248
 *             req.query = query
 *             req.statement_id = 0
 *         elif isinstance(query, int):             # <<<<<<<<<<<<<<
 *             req.query = None
 *             req.statement_id = <uint64_t> query
 */
  __pyx_t_3 = PyInt_Check(__pyx_v_query); 
  if (likely(__pyx_t_3)) {

    /* "asynctnt/iproto/db.pyx":249
 *             req.statement_id = 0
 *         elif isinstance(query, int):
 *             req.query = None             # <<<<<<<<<<<<<<
 *             req.statement_id = <uint64_t> query
 *         else:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_req->query);
    __Pyx_DECREF(__pyx_v_req->query);
    __pyx_v_req->query = ((PyObject*)Py_None);

    /* "asynctnt/iproto/db.pyx":250
 *         elif isinstance(query, int):
 *             req.query = None
 *             req.statement_id = <uint64_t> query             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('query must be either str or int')
 */
    __pyx_t_2 = __Pyx_PyInt_As_uint64_t(__pyx_v_query); if (unlikely((__pyx_t_2 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(26, 250, __pyx_L1_error)
    __pyx_v_req->statement_id = ((uint64_t)__pyx_t_2);

    /* "asynctnt/iproto/db.pyx":248
 *             req.query = query
 *             req.statement_id = 0
 *         elif isinstance(query, int):             # <<<<<<<<<<<<<<
 *             req.query = None
 *             req.statement_id = <uint64_t> query
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/db.pyx":252
 *             req.statement_id = <uint64_t> query
 *         else:
 *             raise TypeError('query must be either str or int')             # <<<<<<<<<<<<<<
 * 
 *         req.args = args
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(26, 252, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "asynctnt/iproto/db.pyx":254
 *             raise TypeError('query must be either str or int')
 * 
 *         req.args = args             # <<<<<<<<<<<<<<
 *         req.parse_metadata = parse_metadata
 *         req.push_subscribe = False
 */
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  __Pyx_GOTREF(__pyx_v_req->args);
  __Pyx_DECREF(__pyx_v_req->args);
  __pyx_v_req->args = __pyx_v_args;

  /* "asynctnt/iproto/db.pyx":255
 * 
 *         req.args = args
 *         req.parse_metadata = parse_metadata             # <<<<<<<<<<<<<<
 *         req.push_subscribe = False
 *         req.check_schema_change = True
 */
  __pyx_v_req->__pyx_base.parse_metadata = __pyx_v_parse_metadata;

  /* "asynctnt/iproto/db.pyx":256
 *         req.args = args
 *         req.parse_metadata = parse_metadata
 *         req.push_subscribe = False             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True
 */
  __pyx_v_req->__pyx_base.push_subscribe = 0;

  /* "asynctnt/iproto/db.pyx":257
 *         req.parse_metadata = parse_metadata
 *         req.push_subscribe = False
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         req.parse_as_tuples = True
 * 
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":258
 *         req.push_subscribe = False
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True             # <<<<<<<<<<<<<<
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->__pyx_base.parse_as_tuples = 1;

  /* "asynctnt/iproto/db.pyx":260
 *         req.parse_as_tuples = True
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _prepare(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_4 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":232
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _execute(self,             # <<<<<<<<<<<<<<
 *                          object query,
 *                          object args,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":262
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _prepare(self,             # <<<<<<<<<<<<<<
 *                          object query,
 *                          bint parse_metadata,
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__prepare(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_query, int __pyx_v_parse_metadata, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_prepare", 1);

  /* "asynctnt/iproto/db.pyx":269
 *             PrepareRequest req
 * 
 *         req = PrepareRequest.__new__(PrepareRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_PREPARE
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_PrepareRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 269, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":270
 * 
 *         req = PrepareRequest.__new__(PrepareRequest)
 *         req.op = tarantool.IPROTO_PREPARE             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_PREPARE;

  /* "asynctnt/iproto/db.pyx":271
 *         req = PrepareRequest.__new__(PrepareRequest)
 *         req.op = tarantool.IPROTO_PREPARE
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 * 
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 271, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":272
 *         req.op = tarantool.IPROTO_PREPARE
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(query, str):
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":274
 *         req.stream_id = self._stream_id
 * 
 *         if isinstance(query, str):             # <<<<<<<<<<<<<<
 *             req.query = query
 *             req.statement_id = 0
 */
  __pyx_t_3 = PyUnicode_Check(__pyx_v_query); 
  if (__pyx_t_3) {

    /* "asynctnt/iproto/db.pyx":275
 * 
 *         if isinstance(query, str):
 *             req.query = query             # <<<<<<<<<<<<<<
 *             req.statement_id = 0
 *         elif isinstance(query, int):
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_v_query))||((__pyx_v_query) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_query))) __PYX_ERR(26, 275, __pyx_L1_error)
    __pyx_t_1 = __pyx_v_query;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_req->query);
    __Pyx_DECREF(__pyx_v_req->query);
    __pyx_v_req->query = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/db.pyx":276
 *         if isinstance(query, str):
 *             req.query = query
 *             req.statement_id = 0             # <<<<<<<<<<<<<<
 *         elif isinstance(query, int):
 *             req.query = None
 */
    __pyx_v_req->statement_id = 0;

    /* "asynctnt/iproto/db.pyx":274
 *         req.stream_id = self._stream_id
 * 
 *         if isinstance(query, str):             # <<<<<<<<<<<<<<
 *             req.query = query
 *             req.statement_id = 0
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/db.pyx":277
 *             req.query = query
 *             req.statement_id = 0
 *         elif isinstance(query, int):             # <<<<<<<<<<<<<<
 *             req.query = None
 *             req.statement_id = <uint64_t> query
 */
  __pyx_t_3 = PyInt_Check(__pyx_v_query); 
  if (likely(__pyx_t_3)) {

    /* "asynctnt/iproto/db.pyx":278
 *             req.statement_id = 0
 *         elif isinstance(query, int):
 *             req.query = None             # <<<<<<<<<<<<<<
 *             req.statement_id = <uint64_t> query
 *         else:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_req->query);
    __Pyx_DECREF(__pyx_v_req->query);
    __pyx_v_req->query = ((PyObject*)Py_None);

    /* "asynctnt/iproto/db.pyx":279
 *         elif isinstance(query, int):
 *             req.query = None
 *             req.statement_id = <uint64_t> query             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('query must be either str or int')
 */
    __pyx_t_2 = __Pyx_PyInt_As_uint64_t(__pyx_v_query); if (unlikely((__pyx_t_2 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(26, 279, __pyx_L1_error)
    __pyx_v_req->statement_id = ((uint64_t)__pyx_t_2);

    /* "asynctnt/iproto/db.pyx":277
 *             req.query = query
 *             req.statement_id = 0
 *         elif isinstance(query, int):             # <<<<<<<<<<<<<<
 *             req.query = None
 *             req.statement_id = <uint64_t> query
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/db.pyx":281
 *             req.statement_id = <uint64_t> query
 *         else:
 *             raise TypeError('query must be either str or int')             # <<<<<<<<<<<<<<
 * 
 *         req.push_subscribe = False
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(26, 281, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "asynctnt/iproto/db.pyx":283
 *             raise TypeError('query must be either str or int')
 * 
 *         req.push_subscribe = False             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True
 */
  __pyx_v_req->__pyx_base.push_subscribe = 0;

  /* "asynctnt/iproto/db.pyx":284
 * 
 *         req.push_subscribe = False
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         req.parse_as_tuples = True
 *         req.parse_metadata = parse_metadata
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":285
 *         req.push_subscribe = False
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True             # <<<<<<<<<<<<<<
 *         req.parse_metadata = parse_metadata
 * 
 */
  __pyx_v_req->__pyx_base.parse_as_tuples = 1;

  /* "asynctnt/iproto/db.pyx":286
 *         req.check_schema_change = True
 *         req.parse_as_tuples = True
 *         req.parse_metadata = parse_metadata             # <<<<<<<<<<<<<<
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->__pyx_base.parse_metadata = __pyx_v_parse_metadata;

  /* "asynctnt/iproto/db.pyx":288
 *         req.parse_metadata = parse_metadata
 * 
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _begin(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_4 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":262
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _prepare(self,             # <<<<<<<<<<<<<<
 *                          object query,
 *                          bint parse_metadata,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._prepare", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":290
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _begin(self,             # <<<<<<<<<<<<<<
 *                        uint32_t isolation,
 *                        double tx_timeout,
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__begin(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, uint32_t __pyx_v_isolation, double __pyx_v_tx_timeout, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_begin", 1);

  /* "asynctnt/iproto/db.pyx":294
 *                        double tx_timeout,
 *                        float timeout):
 *         cdef BeginRequest req = BeginRequest.__new__(BeginRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_BEGIN
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_BeginRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 294, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":295
 *                        float timeout):
 *         cdef BeginRequest req = BeginRequest.__new__(BeginRequest)
 *         req.op = tarantool.IPROTO_BEGIN             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_BEGIN;

  /* "asynctnt/iproto/db.pyx":296
 *         cdef BeginRequest req = BeginRequest.__new__(BeginRequest)
 *         req.op = tarantool.IPROTO_BEGIN
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 296, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":297
 *         req.op = tarantool.IPROTO_BEGIN
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         req.push_subscribe = False
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":298
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         req.push_subscribe = False
 *         req.isolation = isolation
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":299
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True
 *         req.push_subscribe = False             # <<<<<<<<<<<<<<
 *         req.isolation = isolation
 *         req.tx_timeout = tx_timeout
 */
  __pyx_v_req->__pyx_base.push_subscribe = 0;

  /* "asynctnt/iproto/db.pyx":300
 *         req.check_schema_change = True
 *         req.push_subscribe = False
 *         req.isolation = isolation             # <<<<<<<<<<<<<<
 *         req.tx_timeout = tx_timeout
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->isolation = __pyx_v_isolation;

  /* "asynctnt/iproto/db.pyx":301
 *         req.push_subscribe = False
 *         req.isolation = isolation
 *         req.tx_timeout = tx_timeout             # <<<<<<<<<<<<<<
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 */
  __pyx_v_req->tx_timeout = __pyx_v_tx_timeout;

  /* "asynctnt/iproto/db.pyx":302
 *         req.isolation = isolation
 *         req.tx_timeout = tx_timeout
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _commit(self, float timeout):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":290
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _begin(self,             # <<<<<<<<<<<<<<
 *                        uint32_t isolation,
 *                        double tx_timeout,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":304
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _commit(self, float timeout):             # <<<<<<<<<<<<<<
 *         cdef CommitRequest req = CommitRequest.__new__(CommitRequest)
 *         req.op = tarantool.IPROTO_COMMIT
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__commit(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_commit", 1);

  /* "asynctnt/iproto/db.pyx":305
 * 
 *     cdef object _commit(self, float timeout):
 *         cdef CommitRequest req = CommitRequest.__new__(CommitRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_COMMIT
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_CommitRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 305, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":306
 *     cdef object _commit(self, float timeout):
 *         cdef CommitRequest req = CommitRequest.__new__(CommitRequest)
 *         req.op = tarantool.IPROTO_COMMIT             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_COMMIT;

  /* "asynctnt/iproto/db.pyx":307
 *         cdef CommitRequest req = CommitRequest.__new__(CommitRequest)
 *         req.op = tarantool.IPROTO_COMMIT
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 307, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":308
 *         req.op = tarantool.IPROTO_COMMIT
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         req.push_subscribe = False
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":309
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         req.push_subscribe = False
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":310
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True
 *         req.push_subscribe = False             # <<<<<<<<<<<<<<
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 */
  __pyx_v_req->__pyx_base.push_subscribe = 0;

  /* "asynctnt/iproto/db.pyx":311
 *         req.check_schema_change = True
 *         req.push_subscribe = False
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _rollback(self, float timeout):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":304
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _commit(self, float timeout):             # <<<<<<<<<<<<<<
 *         cdef CommitRequest req = CommitRequest.__new__(CommitRequest)
 *         req.op = tarantool.IPROTO_COMMIT
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._commit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":313
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _rollback(self, float timeout):             # <<<<<<<<<<<<<<
 *         cdef RollbackRequest req = RollbackRequest.__new__(RollbackRequest)
 *         req.op = tarantool.IPROTO_ROLLBACK
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_2Db__rollback(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint64_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_rollback", 1);

  /* "asynctnt/iproto/db.pyx":314
 * 
 *     cdef object _rollback(self, float timeout):
 *         cdef RollbackRequest req = RollbackRequest.__new__(RollbackRequest)             # <<<<<<<<<<<<<<
 *         req.op = tarantool.IPROTO_ROLLBACK
 *         req.sync = self.next_sync()
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_RollbackRequest(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 314, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/db.pyx":315
 *     cdef object _rollback(self, float timeout):
 *         cdef RollbackRequest req = RollbackRequest.__new__(RollbackRequest)
 *         req.op = tarantool.IPROTO_ROLLBACK             # <<<<<<<<<<<<<<
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 */
  __pyx_v_req->__pyx_base.op = __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_ROLLBACK;

  /* "asynctnt/iproto/db.pyx":316
 *         cdef RollbackRequest req = RollbackRequest.__new__(RollbackRequest)
 *         req.op = tarantool.IPROTO_ROLLBACK
 *         req.sync = self.next_sync()             # <<<<<<<<<<<<<<
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 316, __pyx_L1_error)
  __pyx_v_req->__pyx_base.sync = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":317
 *         req.op = tarantool.IPROTO_ROLLBACK
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id             # <<<<<<<<<<<<<<
 *         req.check_schema_change = True
 *         return self._protocol.execute(self._protocol, req, timeout)
 */
  __pyx_t_2 = __pyx_v_self->_stream_id;
  __pyx_v_req->__pyx_base.stream_id = __pyx_t_2;

  /* "asynctnt/iproto/db.pyx":318
 *         req.sync = self.next_sync()
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True             # <<<<<<<<<<<<<<
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 */
  __pyx_v_req->__pyx_base.check_schema_change = 1;

  /* "asynctnt/iproto/db.pyx":319
 *         req.stream_id = self._stream_id
 *         req.check_schema_change = True
 *         return self._protocol.execute(self._protocol, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     # public methods
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_protocol);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_self->_protocol->execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_t_1), ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v_req), __pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":313
 *         return self._protocol.execute(self._protocol, req, timeout)
 * 
 *     cdef object _rollback(self, float timeout):             # <<<<<<<<<<<<<<
 *         cdef RollbackRequest req = RollbackRequest.__new__(RollbackRequest)
 *         req.op = tarantool.IPROTO_ROLLBACK
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db._rollback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":323
 *     # public methods
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def stream_id(self):
 *         return <int> self._stream_id
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_9stream_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_9stream_id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_9stream_id___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_9stream_id___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "asynctnt/iproto/db.pyx":325
 *     @property
 *     def stream_id(self):
 *         return <int> self._stream_id             # <<<<<<<<<<<<<<
 * 
 *     def set_stream_id(self, int stream_id):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(((int)__pyx_v_self->_stream_id)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":323
 *     # public methods
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def stream_id(self):
 *         return <int> self._stream_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.stream_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":327
 *         return <int> self._stream_id
 * 
 *     def set_stream_id(self, int stream_id):             # <<<<<<<<<<<<<<
 *         self._stream_id = <uint64_t> stream_id
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_3set_stream_id(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_3set_stream_id = {"set_stream_id", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_3set_stream_id, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_3set_stream_id(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_stream_id;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_stream_id (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_stream_id,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_stream_id)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 327, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_stream_id") < 0)) __PYX_ERR(26, 327, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_stream_id = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_stream_id == (int)-1) && PyErr_Occurred())) __PYX_ERR(26, 327, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_stream_id", 1, 1, 1, __pyx_nargs); __PYX_ERR(26, 327, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.set_stream_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_2set_stream_id(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_stream_id);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_2set_stream_id(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, int __pyx_v_stream_id) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_stream_id", 1);

  /* "asynctnt/iproto/db.pyx":328
 * 
 *     def set_stream_id(self, int stream_id):
 *         self._stream_id = <uint64_t> stream_id             # <<<<<<<<<<<<<<
 * 
 *     def ping(self, float timeout=-1):
 */
  __pyx_v_self->_stream_id = ((uint64_t)__pyx_v_stream_id);

  /* "asynctnt/iproto/db.pyx":327
 *         return <int> self._stream_id
 * 
 *     def set_stream_id(self, int stream_id):             # <<<<<<<<<<<<<<
 *         self._stream_id = <uint64_t> stream_id
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":330
 *         self._stream_id = <uint64_t> stream_id
 * 
 *     def ping(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._ping(timeout)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_5ping(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_5ping = {"ping", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_5ping, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_5ping(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  float __pyx_v_timeout;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ping (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_timeout,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 330, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ping") < 0)) __PYX_ERR(26, 330, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[0]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 330, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ping", 0, 0, 1, __pyx_nargs); __PYX_ERR(26, 330, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.ping", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_4ping(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_timeout);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_4ping(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ping", 1);

  /* "asynctnt/iproto/db.pyx":331
 * 
 *     def ping(self, float timeout=-1):
 *         return self._ping(timeout)             # <<<<<<<<<<<<<<
 * 
 *     def call16(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__ping(__pyx_v_self, __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":330
 *         self._stream_id = <uint64_t> stream_id
 * 
 *     def ping(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._ping(timeout)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.ping", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":333
 *         return self._ping(timeout)
 * 
 *     def call16(self,             # <<<<<<<<<<<<<<
 *                str func_name,
 *                object args=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_7call16(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_7call16 = {"call16", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_7call16, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_7call16(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func_name = 0;
  PyObject *__pyx_v_args = 0;
  float __pyx_v_timeout;
  int __pyx_v_push_subscribe;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call16 (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func_name,&__pyx_n_s_args,&__pyx_n_s_timeout,&__pyx_n_s_push_subscribe,0};

    /* "asynctnt/iproto/db.pyx":335
 *     def call16(self,
 *                str func_name,
 *                object args=None,             # <<<<<<<<<<<<<<
 *                float timeout=-1,
 *                bint push_subscribe=False):
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 333, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_args);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 333, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 333, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_push_subscribe);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 333, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "call16") < 0)) __PYX_ERR(26, 333, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func_name = ((PyObject*)values[0]);
    __pyx_v_args = values[1];
    if (values[2]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 336, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
    if (values[3]) {
      __pyx_v_push_subscribe = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_push_subscribe == (int)-1) && PyErr_Occurred())) __PYX_ERR(26, 337, __pyx_L3_error)
    } else {

      /* "asynctnt/iproto/db.pyx":337
 *                object args=None,
 *                float timeout=-1,
 *                bint push_subscribe=False):             # <<<<<<<<<<<<<<
 *         return self._call(tarantool.IPROTO_CALL_16,
 *                           func_name,
 */
      __pyx_v_push_subscribe = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call16", 0, 1, 4, __pyx_nargs); __PYX_ERR(26, 333, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.call16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_func_name), (&PyUnicode_Type), 1, "func_name", 1))) __PYX_ERR(26, 334, __pyx_L1_error)
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_6call16(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_func_name, __pyx_v_args, __pyx_v_timeout, __pyx_v_push_subscribe);

  /* "asynctnt/iproto/db.pyx":333
 *         return self._ping(timeout)
 * 
 *     def call16(self,             # <<<<<<<<<<<<<<
 *                str func_name,
 *                object args=None,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_6call16(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_func_name, PyObject *__pyx_v_args, float __pyx_v_timeout, int __pyx_v_push_subscribe) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call16", 1);

  /* "asynctnt/iproto/db.pyx":338
 *                float timeout=-1,
 *                bint push_subscribe=False):
 *         return self._call(tarantool.IPROTO_CALL_16,             # <<<<<<<<<<<<<<
 *                           func_name,
 *                           args,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "asynctnt/iproto/db.pyx":342
 *                           args,
 *                           timeout,
 *                           <bint> push_subscribe)             # <<<<<<<<<<<<<<
 * 
 *     def call(self,
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__call(__pyx_v_self, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_CALL_16, __pyx_v_func_name, __pyx_v_args, __pyx_v_timeout, __pyx_v_push_subscribe); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":333
 *         return self._ping(timeout)
 * 
 *     def call16(self,             # <<<<<<<<<<<<<<
 *                str func_name,
 *                object args=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.call16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":344
 *                           <bint> push_subscribe)
 * 
 *     def call(self,             # <<<<<<<<<<<<<<
 *              str func_name,
 *              object args=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_9call(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_9call = {"call", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_9call, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_9call(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func_name = 0;
  PyObject *__pyx_v_args = 0;
  float __pyx_v_timeout;
  int __pyx_v_push_subscribe;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func_name,&__pyx_n_s_args,&__pyx_n_s_timeout,&__pyx_n_s_push_subscribe,0};

    /* "asynctnt/iproto/db.pyx":346
 *     def call(self,
 *              str func_name,
 *              object args=None,             # <<<<<<<<<<<<<<
 *              float timeout=-1,
 *              bint push_subscribe=False):
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 344, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_args);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 344, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 344, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_push_subscribe);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 344, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "call") < 0)) __PYX_ERR(26, 344, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func_name = ((PyObject*)values[0]);
    __pyx_v_args = values[1];
    if (values[2]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 347, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
    if (values[3]) {
      __pyx_v_push_subscribe = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_push_subscribe == (int)-1) && PyErr_Occurred())) __PYX_ERR(26, 348, __pyx_L3_error)
    } else {

      /* "asynctnt/iproto/db.pyx":348
 *              object args=None,
 *              float timeout=-1,
 *              bint push_subscribe=False):             # <<<<<<<<<<<<<<
 *         return self._call(tarantool.IPROTO_CALL,
 *                           func_name,
 */
      __pyx_v_push_subscribe = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call", 0, 1, 4, __pyx_nargs); __PYX_ERR(26, 344, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_func_name), (&PyUnicode_Type), 1, "func_name", 1))) __PYX_ERR(26, 345, __pyx_L1_error)
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_8call(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_func_name, __pyx_v_args, __pyx_v_timeout, __pyx_v_push_subscribe);

  /* "asynctnt/iproto/db.pyx":344
 *                           <bint> push_subscribe)
 * 
 *     def call(self,             # <<<<<<<<<<<<<<
 *              str func_name,
 *              object args=None,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_8call(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_func_name, PyObject *__pyx_v_args, float __pyx_v_timeout, int __pyx_v_push_subscribe) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call", 1);

  /* "asynctnt/iproto/db.pyx":349
 *              float timeout=-1,
 *              bint push_subscribe=False):
 *         return self._call(tarantool.IPROTO_CALL,             # <<<<<<<<<<<<<<
 *                           func_name,
 *                           args,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "asynctnt/iproto/db.pyx":353
 *                           args,
 *                           timeout,
 *                           <bint> push_subscribe)             # <<<<<<<<<<<<<<
 * 
 *     def eval(self,
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__call(__pyx_v_self, __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_CALL, __pyx_v_func_name, __pyx_v_args, __pyx_v_timeout, __pyx_v_push_subscribe); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":344
 *                           <bint> push_subscribe)
 * 
 *     def call(self,             # <<<<<<<<<<<<<<
 *              str func_name,
 *              object args=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":355
 *                           <bint> push_subscribe)
 * 
 *     def eval(self,             # <<<<<<<<<<<<<<
 *              str expression,
 *              object args=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_11eval(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_11eval = {"eval", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_11eval, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_11eval(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_expression = 0;
  PyObject *__pyx_v_args = 0;
  float __pyx_v_timeout;
  int __pyx_v_push_subscribe;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("eval (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_expression,&__pyx_n_s_args,&__pyx_n_s_timeout,&__pyx_n_s_push_subscribe,0};

    /* "asynctnt/iproto/db.pyx":357
 *     def eval(self,
 *              str expression,
 *              object args=None,             # <<<<<<<<<<<<<<
 *              float timeout=-1,
 *              bint push_subscribe=False):
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_expression)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 355, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_args);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_push_subscribe);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 355, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "eval") < 0)) __PYX_ERR(26, 355, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_expression = ((PyObject*)values[0]);
    __pyx_v_args = values[1];
    if (values[2]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 358, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
    if (values[3]) {
      __pyx_v_push_subscribe = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_push_subscribe == (int)-1) && PyErr_Occurred())) __PYX_ERR(26, 359, __pyx_L3_error)
    } else {

      /* "asynctnt/iproto/db.pyx":359
 *              object args=None,
 *              float timeout=-1,
 *              bint push_subscribe=False):             # <<<<<<<<<<<<<<
 *         return self._eval(expression,
 *                           args,
 */
      __pyx_v_push_subscribe = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("eval", 0, 1, 4, __pyx_nargs); __PYX_ERR(26, 355, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.eval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expression), (&PyUnicode_Type), 1, "expression", 1))) __PYX_ERR(26, 356, __pyx_L1_error)
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_10eval(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_expression, __pyx_v_args, __pyx_v_timeout, __pyx_v_push_subscribe);

  /* "asynctnt/iproto/db.pyx":355
 *                           <bint> push_subscribe)
 * 
 *     def eval(self,             # <<<<<<<<<<<<<<
 *              str expression,
 *              object args=None,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_10eval(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_expression, PyObject *__pyx_v_args, float __pyx_v_timeout, int __pyx_v_push_subscribe) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eval", 1);

  /* "asynctnt/iproto/db.pyx":360
 *              float timeout=-1,
 *              bint push_subscribe=False):
 *         return self._eval(expression,             # <<<<<<<<<<<<<<
 *                           args,
 *                           timeout,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "asynctnt/iproto/db.pyx":363
 *                           args,
 *                           timeout,
 *                           <bint> push_subscribe)             # <<<<<<<<<<<<<<
 * 
 *     def select(self,
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__eval(__pyx_v_self, __pyx_v_expression, __pyx_v_args, __pyx_v_timeout, __pyx_v_push_subscribe); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":355
 *                           <bint> push_subscribe)
 * 
 *     def eval(self,             # <<<<<<<<<<<<<<
 *              str expression,
 *              object args=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.eval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":365
 *                           <bint> push_subscribe)
 * 
 *     def select(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_13select(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_13select = {"select", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_13select, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_13select(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_space = 0;
  PyObject *__pyx_v_key = 0;
  int __pyx_v_offset;
  int __pyx_v_limit;
  PyObject *__pyx_v_index = 0;
  PyObject *__pyx_v_iterator = 0;
  float __pyx_v_timeout;
  int __pyx_v_check_schema_change;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[8] = {0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("select (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_space,&__pyx_n_s_key,&__pyx_n_s_offset,&__pyx_n_s_limit,&__pyx_n_s_index,&__pyx_n_s_iterator,&__pyx_n_s_timeout,&__pyx_n_s_check_schema_change,0};

    /* "asynctnt/iproto/db.pyx":367
 *     def select(self,
 *                object space,
 *                object key=None,             # <<<<<<<<<<<<<<
 *                int offset=0,
 *                int limit=0xffffffff,
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[4] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));
    values[5] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  8: values[7] = __Pyx_Arg_FASTCALL(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_space)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 365, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_offset);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_limit);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_index);
          if (value) { values[4] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_iterator);
          if (value) { values[5] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[6] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_check_schema_change);
          if (value) { values[7] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 365, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "select") < 0)) __PYX_ERR(26, 365, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  8: values[7] = __Pyx_Arg_FASTCALL(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_space = values[0];
    __pyx_v_key = values[1];
    if (values[2]) {
      __pyx_v_offset = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_offset == (int)-1) && PyErr_Occurred())) __PYX_ERR(26, 368, __pyx_L3_error)
    } else {
      __pyx_v_offset = ((int)0);
    }
    if (values[3]) {
      __pyx_v_limit = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_limit == (int)-1) && PyErr_Occurred())) __PYX_ERR(26, 369, __pyx_L3_error)
    } else {
      __pyx_v_limit = ((int)0xffffffff);
    }
    __pyx_v_index = values[4];
    __pyx_v_iterator = values[5];
    if (values[6]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[6]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 372, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
    if (values[7]) {
      __pyx_v_check_schema_change = __Pyx_PyObject_IsTrue(values[7]); if (unlikely((__pyx_v_check_schema_change == (int)-1) && PyErr_Occurred())) __PYX_ERR(26, 373, __pyx_L3_error)
    } else {

      /* "asynctnt/iproto/db.pyx":373
 *                object iterator=0,
 *                float timeout=-1,
 *                bint check_schema_change=True):             # <<<<<<<<<<<<<<
 *         return self._select(space, index, key, offset, limit, iterator,
 *                             timeout, check_schema_change)
 */
      __pyx_v_check_schema_change = ((int)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("select", 0, 1, 8, __pyx_nargs); __PYX_ERR(26, 365, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.select", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_12select(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_space, __pyx_v_key, __pyx_v_offset, __pyx_v_limit, __pyx_v_index, __pyx_v_iterator, __pyx_v_timeout, __pyx_v_check_schema_change);

  /* "asynctnt/iproto/db.pyx":365
 *                           <bint> push_subscribe)
 * 
 *     def select(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key=None,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_12select(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_key, int __pyx_v_offset, int __pyx_v_limit, PyObject *__pyx_v_index, PyObject *__pyx_v_iterator, float __pyx_v_timeout, int __pyx_v_check_schema_change) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("select", 1);

  /* "asynctnt/iproto/db.pyx":374
 *                float timeout=-1,
 *                bint check_schema_change=True):
 *         return self._select(space, index, key, offset, limit, iterator,             # <<<<<<<<<<<<<<
 *                             timeout, check_schema_change)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "asynctnt/iproto/db.pyx":375
 *                bint check_schema_change=True):
 *         return self._select(space, index, key, offset, limit, iterator,
 *                             timeout, check_schema_change)             # <<<<<<<<<<<<<<
 * 
 *     def insert(self,
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__select(__pyx_v_self, __pyx_v_space, __pyx_v_index, __pyx_v_key, __pyx_v_offset, __pyx_v_limit, __pyx_v_iterator, __pyx_v_timeout, __pyx_v_check_schema_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":365
 *                           <bint> push_subscribe)
 * 
 *     def select(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.select", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":377
 *                             timeout, check_schema_change)
 * 
 *     def insert(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object t,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_15insert(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_15insert = {"insert", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_15insert, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_15insert(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_space = 0;
  PyObject *__pyx_v_t = 0;
  int __pyx_v_replace;
  float __pyx_v_timeout;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_space,&__pyx_n_s_t,&__pyx_n_s_replace,&__pyx_n_s_timeout,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_space)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 377, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_t)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 377, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 0, 2, 4, 1); __PYX_ERR(26, 377, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_replace);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 377, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 377, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "insert") < 0)) __PYX_ERR(26, 377, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_space = values[0];
    __pyx_v_t = values[1];
    if (values[2]) {
      __pyx_v_replace = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_replace == (int)-1) && PyErr_Occurred())) __PYX_ERR(26, 380, __pyx_L3_error)
    } else {

      /* "asynctnt/iproto/db.pyx":380
 *                object space,
 *                object t,
 *                bint replace=False,             # <<<<<<<<<<<<<<
 *                float timeout=-1):
 *         return self._insert(space, t, <bint> replace, timeout)
 */
      __pyx_v_replace = ((int)0);
    }
    if (values[3]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[3]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 381, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 0, 2, 4, __pyx_nargs); __PYX_ERR(26, 377, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_14insert(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_space, __pyx_v_t, __pyx_v_replace, __pyx_v_timeout);

  /* "asynctnt/iproto/db.pyx":377
 *                             timeout, check_schema_change)
 * 
 *     def insert(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object t,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_14insert(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_t, int __pyx_v_replace, float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("insert", 1);

  /* "asynctnt/iproto/db.pyx":382
 *                bint replace=False,
 *                float timeout=-1):
 *         return self._insert(space, t, <bint> replace, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def replace(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__insert(__pyx_v_self, __pyx_v_space, __pyx_v_t, __pyx_v_replace, __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":377
 *                             timeout, check_schema_change)
 * 
 *     def insert(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object t,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":384
 *         return self._insert(space, t, <bint> replace, timeout)
 * 
 *     def replace(self,             # <<<<<<<<<<<<<<
 *                 object space,
 *                 object t,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_17replace(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_17replace = {"replace", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_17replace, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_17replace(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_space = 0;
  PyObject *__pyx_v_t = 0;
  float __pyx_v_timeout;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("replace (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_space,&__pyx_n_s_t,&__pyx_n_s_timeout,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_space)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 384, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_t)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 384, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("replace", 0, 2, 3, 1); __PYX_ERR(26, 384, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 384, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "replace") < 0)) __PYX_ERR(26, 384, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_space = values[0];
    __pyx_v_t = values[1];
    if (values[2]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 387, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("replace", 0, 2, 3, __pyx_nargs); __PYX_ERR(26, 384, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.replace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_16replace(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_space, __pyx_v_t, __pyx_v_timeout);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_16replace(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_t, float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("replace", 1);

  /* "asynctnt/iproto/db.pyx":388
 *                 object t,
 *                 float timeout=-1):
 *         return self._insert(space, t, <bint> True, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def delete(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__insert(__pyx_v_self, __pyx_v_space, __pyx_v_t, 1, __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":384
 *         return self._insert(space, t, <bint> replace, timeout)
 * 
 *     def replace(self,             # <<<<<<<<<<<<<<
 *                 object space,
 *                 object t,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.replace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":390
 *         return self._insert(space, t, <bint> True, timeout)
 * 
 *     def delete(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_19delete(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_19delete = {"delete", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_19delete, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_19delete(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_space = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_index = 0;
  float __pyx_v_timeout;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("delete (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_space,&__pyx_n_s_key,&__pyx_n_s_index,&__pyx_n_s_timeout,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_space)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 390, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 390, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("delete", 0, 2, 4, 1); __PYX_ERR(26, 390, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_index);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 390, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 390, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "delete") < 0)) __PYX_ERR(26, 390, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_space = values[0];
    __pyx_v_key = values[1];
    __pyx_v_index = values[2];
    if (values[3]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[3]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 394, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("delete", 0, 2, 4, __pyx_nargs); __PYX_ERR(26, 390, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.delete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_18delete(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_space, __pyx_v_key, __pyx_v_index, __pyx_v_timeout);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_18delete(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_key, PyObject *__pyx_v_index, float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("delete", 1);

  /* "asynctnt/iproto/db.pyx":395
 *                object index=0,
 *                float timeout=-1):
 *         return self._delete(space, index, key, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def update(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__delete(__pyx_v_self, __pyx_v_space, __pyx_v_index, __pyx_v_key, __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":390
 *         return self._insert(space, t, <bint> True, timeout)
 * 
 *     def delete(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.delete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":397
 *         return self._delete(space, index, key, timeout)
 * 
 *     def update(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_21update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_21update = {"update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_21update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_21update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_space = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_operations = 0;
  PyObject *__pyx_v_index = 0;
  float __pyx_v_timeout;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_space,&__pyx_n_s_key,&__pyx_n_s_operations,&__pyx_n_s_index,&__pyx_n_s_timeout,0};
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_space)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 397, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 397, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("update", 0, 3, 5, 1); __PYX_ERR(26, 397, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_operations)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 397, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("update", 0, 3, 5, 2); __PYX_ERR(26, 397, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_index);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 397, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[4] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 397, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "update") < 0)) __PYX_ERR(26, 397, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_space = values[0];
    __pyx_v_key = values[1];
    __pyx_v_operations = ((PyObject*)values[2]);
    __pyx_v_index = values[3];
    if (values[4]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[4]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 402, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update", 0, 3, 5, __pyx_nargs); __PYX_ERR(26, 397, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_operations), (&PyList_Type), 1, "operations", 1))) __PYX_ERR(26, 400, __pyx_L1_error)
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_20update(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_space, __pyx_v_key, __pyx_v_operations, __pyx_v_index, __pyx_v_timeout);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_20update(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_key, PyObject *__pyx_v_operations, PyObject *__pyx_v_index, float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 1);

  /* "asynctnt/iproto/db.pyx":403
 *                object index=0,
 *                float timeout=-1):
 *         return self._update(space, index, key, operations, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def upsert(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__update(__pyx_v_self, __pyx_v_space, __pyx_v_index, __pyx_v_key, __pyx_v_operations, __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":397
 *         return self._delete(space, index, key, timeout)
 * 
 *     def update(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":405
 *         return self._update(space, index, key, operations, timeout)
 * 
 *     def upsert(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object t,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_23upsert(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_23upsert = {"upsert", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_23upsert, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_23upsert(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_space = 0;
  PyObject *__pyx_v_t = 0;
  PyObject *__pyx_v_operations = 0;
  float __pyx_v_timeout;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("upsert (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_space,&__pyx_n_s_t,&__pyx_n_s_operations,&__pyx_n_s_timeout,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_space)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 405, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_t)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 405, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("upsert", 0, 3, 4, 1); __PYX_ERR(26, 405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_operations)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 405, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("upsert", 0, 3, 4, 2); __PYX_ERR(26, 405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 405, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "upsert") < 0)) __PYX_ERR(26, 405, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_space = values[0];
    __pyx_v_t = values[1];
    __pyx_v_operations = ((PyObject*)values[2]);
    if (values[3]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[3]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 409, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("upsert", 0, 3, 4, __pyx_nargs); __PYX_ERR(26, 405, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.upsert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_operations), (&PyList_Type), 1, "operations", 1))) __PYX_ERR(26, 408, __pyx_L1_error)
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_22upsert(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_space, __pyx_v_t, __pyx_v_operations, __pyx_v_timeout);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_22upsert(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_space, PyObject *__pyx_v_t, PyObject *__pyx_v_operations, float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("upsert", 1);

  /* "asynctnt/iproto/db.pyx":410
 *                list operations,
 *                float timeout=-1):
 *         return self._upsert(space, t, operations, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def execute(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__upsert(__pyx_v_self, __pyx_v_space, __pyx_v_t, __pyx_v_operations, __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":405
 *         return self._update(space, index, key, operations, timeout)
 * 
 *     def upsert(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object t,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.upsert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":412
 *         return self._upsert(space, t, operations, timeout)
 * 
 *     def execute(self,             # <<<<<<<<<<<<<<
 *                 object query,
 *                 object args,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_25execute(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_25execute = {"execute", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_25execute, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_25execute(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_query = 0;
  PyObject *__pyx_v_args = 0;
  int __pyx_v_parse_metadata;
  float __pyx_v_timeout;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_query,&__pyx_n_s_args,&__pyx_n_s_parse_metadata,&__pyx_n_s_timeout,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_query)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 412, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_args)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 412, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("execute", 0, 2, 4, 1); __PYX_ERR(26, 412, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_parse_metadata);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 412, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 412, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "execute") < 0)) __PYX_ERR(26, 412, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_query = values[0];
    __pyx_v_args = values[1];
    if (values[2]) {
      __pyx_v_parse_metadata = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_parse_metadata == (int)-1) && PyErr_Occurred())) __PYX_ERR(26, 415, __pyx_L3_error)
    } else {

      /* "asynctnt/iproto/db.pyx":415
 *                 object query,
 *                 object args,
 *                 bint parse_metadata=True,             # <<<<<<<<<<<<<<
 *                 float timeout=-1):
 *         return self._execute(query, args, <bint> parse_metadata, timeout)
 */
      __pyx_v_parse_metadata = ((int)1);
    }
    if (values[3]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[3]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 416, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("execute", 0, 2, 4, __pyx_nargs); __PYX_ERR(26, 412, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_24execute(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_query, __pyx_v_args, __pyx_v_parse_metadata, __pyx_v_timeout);

  /* "asynctnt/iproto/db.pyx":412
 *         return self._upsert(space, t, operations, timeout)
 * 
 *     def execute(self,             # <<<<<<<<<<<<<<
 *                 object query,
 *                 object args,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_24execute(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_query, PyObject *__pyx_v_args, int __pyx_v_parse_metadata, float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute", 1);

  /* "asynctnt/iproto/db.pyx":417
 *                 bint parse_metadata=True,
 *                 float timeout=-1):
 *         return self._execute(query, args, <bint> parse_metadata, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def prepare(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__execute(__pyx_v_self, __pyx_v_query, __pyx_v_args, __pyx_v_parse_metadata, __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":412
 *         return self._upsert(space, t, operations, timeout)
 * 
 *     def execute(self,             # <<<<<<<<<<<<<<
 *                 object query,
 *                 object args,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":419
 *         return self._execute(query, args, <bint> parse_metadata, timeout)
 * 
 *     def prepare(self,             # <<<<<<<<<<<<<<
 *                 object query,
 *                 bint parse_metadata=True,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_27prepare(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_27prepare = {"prepare", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_27prepare, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_27prepare(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_query = 0;
  int __pyx_v_parse_metadata;
  float __pyx_v_timeout;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("prepare (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_query,&__pyx_n_s_parse_metadata,&__pyx_n_s_timeout,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_query)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 419, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_parse_metadata);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 419, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 419, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "prepare") < 0)) __PYX_ERR(26, 419, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_query = values[0];
    if (values[1]) {
      __pyx_v_parse_metadata = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_parse_metadata == (int)-1) && PyErr_Occurred())) __PYX_ERR(26, 421, __pyx_L3_error)
    } else {

      /* "asynctnt/iproto/db.pyx":421
 *     def prepare(self,
 *                 object query,
 *                 bint parse_metadata=True,             # <<<<<<<<<<<<<<
 *                 float timeout=-1):
 *         return self._prepare(query, <bint> parse_metadata, timeout)
 */
      __pyx_v_parse_metadata = ((int)1);
    }
    if (values[2]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 422, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("prepare", 0, 1, 3, __pyx_nargs); __PYX_ERR(26, 419, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.prepare", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_26prepare(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_query, __pyx_v_parse_metadata, __pyx_v_timeout);

  /* "asynctnt/iproto/db.pyx":419
 *         return self._execute(query, args, <bint> parse_metadata, timeout)
 * 
 *     def prepare(self,             # <<<<<<<<<<<<<<
 *                 object query,
 *                 bint parse_metadata=True,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_26prepare(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, PyObject *__pyx_v_query, int __pyx_v_parse_metadata, float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prepare", 1);

  /* "asynctnt/iproto/db.pyx":423
 *                 bint parse_metadata=True,
 *                 float timeout=-1):
 *         return self._prepare(query, <bint> parse_metadata, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def begin(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__prepare(__pyx_v_self, __pyx_v_query, __pyx_v_parse_metadata, __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":419
 *         return self._execute(query, args, <bint> parse_metadata, timeout)
 * 
 *     def prepare(self,             # <<<<<<<<<<<<<<
 *                 object query,
 *                 bint parse_metadata=True,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.prepare", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":425
 *         return self._prepare(query, <bint> parse_metadata, timeout)
 * 
 *     def begin(self,             # <<<<<<<<<<<<<<
 *               uint32_t isolation,
 *               float tx_timeout=0,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_29begin(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_29begin = {"begin", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_29begin, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_29begin(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  uint32_t __pyx_v_isolation;
  float __pyx_v_tx_timeout;
  float __pyx_v_timeout;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("begin (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_isolation,&__pyx_n_s_tx_timeout,&__pyx_n_s_timeout,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_isolation)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 425, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tx_timeout);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 425, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 425, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "begin") < 0)) __PYX_ERR(26, 425, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_isolation = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_isolation == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(26, 426, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_tx_timeout = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_tx_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 427, __pyx_L3_error)
    } else {
      __pyx_v_tx_timeout = ((float)0.0);
    }
    if (values[2]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 428, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("begin", 0, 1, 3, __pyx_nargs); __PYX_ERR(26, 425, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_28begin(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_isolation, __pyx_v_tx_timeout, __pyx_v_timeout);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_28begin(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, uint32_t __pyx_v_isolation, float __pyx_v_tx_timeout, float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("begin", 1);

  /* "asynctnt/iproto/db.pyx":429
 *               float tx_timeout=0,
 *               float timeout=-1):
 *         return self._begin(isolation, <double> tx_timeout, timeout)             # <<<<<<<<<<<<<<
 * 
 *     def commit(self, float timeout=-1):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__begin(__pyx_v_self, __pyx_v_isolation, ((double)__pyx_v_tx_timeout), __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":425
 *         return self._prepare(query, <bint> parse_metadata, timeout)
 * 
 *     def begin(self,             # <<<<<<<<<<<<<<
 *               uint32_t isolation,
 *               float tx_timeout=0,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":431
 *         return self._begin(isolation, <double> tx_timeout, timeout)
 * 
 *     def commit(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._commit(timeout)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_31commit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_31commit = {"commit", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_31commit, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_31commit(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  float __pyx_v_timeout;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("commit (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_timeout,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 431, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "commit") < 0)) __PYX_ERR(26, 431, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[0]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 431, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("commit", 0, 0, 1, __pyx_nargs); __PYX_ERR(26, 431, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.commit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_30commit(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_timeout);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_30commit(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("commit", 1);

  /* "asynctnt/iproto/db.pyx":432
 * 
 *     def commit(self, float timeout=-1):
 *         return self._commit(timeout)             # <<<<<<<<<<<<<<
 * 
 *     def rollback(self, float timeout=-1):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__commit(__pyx_v_self, __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":431
 *         return self._begin(isolation, <double> tx_timeout, timeout)
 * 
 *     def commit(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._commit(timeout)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.commit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/db.pyx":434
 *         return self._commit(timeout)
 * 
 *     def rollback(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._rollback(timeout)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_33rollback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_33rollback = {"rollback", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_33rollback, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_33rollback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  float __pyx_v_timeout;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rollback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_timeout,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 434, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "rollback") < 0)) __PYX_ERR(26, 434, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_timeout = __pyx_PyFloat_AsFloat(values[0]); if (unlikely((__pyx_v_timeout == (float)-1) && PyErr_Occurred())) __PYX_ERR(26, 434, __pyx_L3_error)
    } else {
      __pyx_v_timeout = ((float)-1.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("rollback", 0, 0, 1, __pyx_nargs); __PYX_ERR(26, 434, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.rollback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_32rollback(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v_timeout);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_32rollback(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("rollback", 1);

  /* "asynctnt/iproto/db.pyx":435
 * 
 *     def rollback(self, float timeout=-1):
 *         return self._rollback(timeout)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__rollback(__pyx_v_self, __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/db.pyx":434
 *         return self._commit(timeout)
 * 
 *     def rollback(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._rollback(timeout)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.rollback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_35__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_35__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_35__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_35__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_34__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_34__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_37__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_2Db_37__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_37__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_2Db_37__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2Db_36__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2Db_36__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.Db.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/push.pyx":9
 * 
 * cdef class PushIterator:
 *     def __init__(self, fut: asyncio.Future):             # <<<<<<<<<<<<<<
 *         """
 *             Creates PushIterator object. In order to receive push notifications
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_8asynctnt_6iproto_8protocol_12PushIterator___init__, "\n            Creates PushIterator object. In order to receive push notifications\n            this iterator must be created.\n\n            Example:\n\n            .. code-block:: pycon\n\n                # tarantool function:\n                # function f()\n                #     box.session.push('hello')\n                #     return 'finished'\n                # end\n\n                >>> fut = conn.call('async_function', push_subscribe=True)\n                >>> async for item in PushIterator(fut):\n                ...     print(item)\n\n\n            :param fut: Future object returned from call_async, eval_sync\n                        functions\n            :type fut: asyncio.Future\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_8asynctnt_6iproto_8protocol_12PushIterator___init__;
#endif
static int __pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fut = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fut,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fut)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(27, 9, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(27, 9, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_fut = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(27, 9, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PushIterator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator___init__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)__pyx_v_self), __pyx_v_fut);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator___init__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self, PyObject *__pyx_v_fut) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_response = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_request = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "asynctnt/iproto/push.pyx":36
 *             Response response
 *             BaseRequest request
 *         if not hasattr(fut, '_response'):             # <<<<<<<<<<<<<<
 *             raise ValueError('Future is invalid. Make sure to call with '
 *                              'a future returned from a method with '
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_fut, __pyx_n_u_response); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(27, 36, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (unlikely(__pyx_t_2)) {

    /* "asynctnt/iproto/push.pyx":37
 *             BaseRequest request
 *         if not hasattr(fut, '_response'):
 *             raise ValueError('Future is invalid. Make sure to call with '             # <<<<<<<<<<<<<<
 *                              'a future returned from a method with '
 *                              'push_subscribe=True flag')
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(27, 37, __pyx_L1_error)

    /* "asynctnt/iproto/push.pyx":36
 *             Response response
 *             BaseRequest request
 *         if not hasattr(fut, '_response'):             # <<<<<<<<<<<<<<
 *             raise ValueError('Future is invalid. Make sure to call with '
 *                              'a future returned from a method with '
 */
  }

  /* "asynctnt/iproto/push.pyx":41
 *                              'push_subscribe=True flag')
 * 
 *         response = <Response> fut._response             # <<<<<<<<<<<<<<
 *         request = response.request_
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_response); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_response = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "asynctnt/iproto/push.pyx":42
 * 
 *         response = <Response> fut._response
 *         request = response.request_             # <<<<<<<<<<<<<<
 * 
 *         if not request.push_subscribe:
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_response->request_);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_request = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "asynctnt/iproto/push.pyx":44
 *         request = response.request_
 * 
 *         if not request.push_subscribe:             # <<<<<<<<<<<<<<
 *             raise ValueError('Future is invalid. Make sure to call with '
 *                              'a future returned from a method with '
 */
  __pyx_t_2 = (!__pyx_v_request->push_subscribe);
  if (unlikely(__pyx_t_2)) {

    /* "asynctnt/iproto/push.pyx":45
 * 
 *         if not request.push_subscribe:
 *             raise ValueError('Future is invalid. Make sure to call with '             # <<<<<<<<<<<<<<
 *                              'a future returned from a method with '
 *                              'push_subscribe=True flag')
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(27, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(27, 45, __pyx_L1_error)

    /* "asynctnt/iproto/push.pyx":44
 *         request = response.request_
 * 
 *         if not request.push_subscribe:             # <<<<<<<<<<<<<<
 *             raise ValueError('Future is invalid. Make sure to call with '
 *                              'a future returned from a method with '
 */
  }

  /* "asynctnt/iproto/push.pyx":49
 *                              'push_subscribe=True flag')
 * 
 *         self._fut = fut             # <<<<<<<<<<<<<<
 *         self._request = request
 *         self._response = response
 */
  __Pyx_INCREF(__pyx_v_fut);
  __Pyx_GIVEREF(__pyx_v_fut);
  __Pyx_GOTREF(__pyx_v_self->_fut);
  __Pyx_DECREF(__pyx_v_self->_fut);
  __pyx_v_self->_fut = __pyx_v_fut;

  /* "asynctnt/iproto/push.pyx":50
 * 
 *         self._fut = fut
 *         self._request = request             # <<<<<<<<<<<<<<
 *         self._response = response
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_request);
  __Pyx_GIVEREF((PyObject *)__pyx_v_request);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_request);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_request);
  __pyx_v_self->_request = __pyx_v_request;

  /* "asynctnt/iproto/push.pyx":51
 *         self._fut = fut
 *         self._request = request
 *         self._response = response             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_response);
  __Pyx_GIVEREF((PyObject *)__pyx_v_response);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_response);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_response);
  __pyx_v_self->_response = __pyx_v_response;

  /* "asynctnt/iproto/push.pyx":9
 * 
 * cdef class PushIterator:
 *     def __init__(self, fut: asyncio.Future):             # <<<<<<<<<<<<<<
 *         """
 *             Creates PushIterator object. In order to receive push notifications
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PushIterator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_response);
  __Pyx_XDECREF((PyObject *)__pyx_v_request);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/push.pyx":53
 *         self._response = response
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Cannot use iter with PushIterator - use aiter')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_3__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_2__iter__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_2__iter__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 1);

  /* "asynctnt/iproto/push.pyx":54
 * 
 *     def __iter__(self):
 *         raise RuntimeError('Cannot use iter with PushIterator - use aiter')             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(27, 54, __pyx_L1_error)

  /* "asynctnt/iproto/push.pyx":53
 *         self._response = response
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Cannot use iter with PushIterator - use aiter')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PushIterator.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/push.pyx":56
 *         raise RuntimeError('Cannot use iter with PushIterator - use aiter')
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Cannot use next with PushIterator - use anext')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_5__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_5__next__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_4__next__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_4__next__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__next__", 1);

  /* "asynctnt/iproto/push.pyx":57
 * 
 *     def __next__(self):
 *         raise RuntimeError('Cannot use next with PushIterator - use anext')             # <<<<<<<<<<<<<<
 * 
 *     def __aiter__(self):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(27, 57, __pyx_L1_error)

  /* "asynctnt/iproto/push.pyx":56
 *         raise RuntimeError('Cannot use iter with PushIterator - use aiter')
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Cannot use next with PushIterator - use anext')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PushIterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/push.pyx":59
 *         raise RuntimeError('Cannot use next with PushIterator - use anext')
 * 
 *     def __aiter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_7__aiter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_7__aiter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__aiter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_6__aiter__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_6__aiter__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__aiter__", 1);

  /* "asynctnt/iproto/push.pyx":60
 * 
 *     def __aiter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "asynctnt/iproto/push.pyx":59
 *         raise RuntimeError('Cannot use next with PushIterator - use anext')
 * 
 *     def __aiter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8asynctnt_6iproto_8protocol_12PushIterator_10generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "asynctnt/iproto/push.pyx":62
 *         return self
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __anext__(self):
 *         cdef Response response
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_9__anext__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_9__anext__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__anext__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_8__anext__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_8__anext__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__anext__", 0);
  __pyx_cur_scope = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(27, 62, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_8asynctnt_6iproto_8protocol_12PushIterator_10generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_anext, __pyx_n_s_PushIterator___anext, __pyx_n_s_asynctnt_iproto_protocol); if (unlikely(!gen)) __PYX_ERR(27, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PushIterator.__anext__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8asynctnt_6iproto_8protocol_12PushIterator_10generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ *__pyx_cur_scope = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__anext__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L8_resume_from_await;
    case 2: goto __pyx_L10_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 62, __pyx_L1_error)

  /* "asynctnt/iproto/push.pyx":65
 *     async def __anext__(self):
 *         cdef Response response
 *         response = self._response             # <<<<<<<<<<<<<<
 * 
 *         if response.push_len() == 0 and response.code_ >= 0:
 */
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_self->_response);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_response = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/push.pyx":67
 *         response = self._response
 * 
 *         if response.push_len() == 0 and response.code_ >= 0:             # <<<<<<<<<<<<<<
 *             # no more data left
 *             raise StopAsyncIteration
 */
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_8Response_push_len(__pyx_cur_scope->__pyx_v_response); if (unlikely(PyErr_Occurred())) __PYX_ERR(27, 67, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 == 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = (__pyx_cur_scope->__pyx_v_response->code_ >= 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "asynctnt/iproto/push.pyx":69
 *         if response.push_len() == 0 and response.code_ >= 0:
 *             # no more data left
 *             raise StopAsyncIteration             # <<<<<<<<<<<<<<
 * 
 *         if response.push_len() > 0:
 */
    __Pyx_Raise(((PyObject *)(&((PyTypeObject*)__Pyx_PyExc_StopAsyncIteration)[0])), 0, 0, 0);
    __PYX_ERR(27, 69, __pyx_L1_error)

    /* "asynctnt/iproto/push.pyx":67
 *         response = self._response
 * 
 *         if response.push_len() == 0 and response.code_ >= 0:             # <<<<<<<<<<<<<<
 *             # no more data left
 *             raise StopAsyncIteration
 */
  }

  /* "asynctnt/iproto/push.pyx":71
 *             raise StopAsyncIteration
 * 
 *         if response.push_len() > 0:             # <<<<<<<<<<<<<<
 *             return response.pop_push()
 * 
 */
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_8Response_push_len(__pyx_cur_scope->__pyx_v_response); if (unlikely(PyErr_Occurred())) __PYX_ERR(27, 71, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_3 > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/push.pyx":72
 * 
 *         if response.push_len() > 0:
 *             return response.pop_push()             # <<<<<<<<<<<<<<
 * 
 *         ev = response._push_event
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_8Response_pop_push(__pyx_cur_scope->__pyx_v_response); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/push.pyx":71
 *             raise StopAsyncIteration
 * 
 *         if response.push_len() > 0:             # <<<<<<<<<<<<<<
 *             return response.pop_push()
 * 
 */
  }

  /* "asynctnt/iproto/push.pyx":74
 *             return response.pop_push()
 * 
 *         ev = response._push_event             # <<<<<<<<<<<<<<
 *         await ev.wait()
 * 
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_response->_push_event;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_ev = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/push.pyx":75
 * 
 *         ev = response._push_event
 *         await ev.wait()             # <<<<<<<<<<<<<<
 * 
 *         exc = response.get_exception()
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ev, __pyx_n_s_wait); if (unlikely(!__pyx_t_5)) __PYX_ERR(27, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_3 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_3 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_3, 0+__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L8_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 75, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(27, 75, __pyx_L1_error)
    }
  }

  /* "asynctnt/iproto/push.pyx":77
 *         await ev.wait()
 * 
 *         exc = response.get_exception()             # <<<<<<<<<<<<<<
 *         if exc is not None:
 *             # someone needs to await the underlying future
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_8Response_get_exception(__pyx_cur_scope->__pyx_v_response); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_exc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/push.pyx":78
 * 
 *         exc = response.get_exception()
 *         if exc is not None:             # <<<<<<<<<<<<<<
 *             # someone needs to await the underlying future
 *             # so we do it here, and most probably (like 99%) the exception
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_exc != Py_None);
  if (unlikely(__pyx_t_2)) {

    /* "asynctnt/iproto/push.pyx":85
 *             # But if it doesnt't throw we still await the future and throw
 *             # the exception ourselves.
 *             await self._fut             # <<<<<<<<<<<<<<
 *             raise exc
 * 
 */
    __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_self->_fut);
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, awaiting value */
      __pyx_generator->resume_label = 2;
      return __pyx_r;
      __pyx_L10_resume_from_await:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 85, __pyx_L1_error)
    } else {
      PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
      if (exc_type) {
        if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
        else __PYX_ERR(27, 85, __pyx_L1_error)
      }
    }

    /* "asynctnt/iproto/push.pyx":86
 *             # the exception ourselves.
 *             await self._fut
 *             raise exc             # <<<<<<<<<<<<<<
 * 
 *         if response.push_len() > 0:
 */
    __Pyx_Raise(__pyx_cur_scope->__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(27, 86, __pyx_L1_error)

    /* "asynctnt/iproto/push.pyx":78
 * 
 *         exc = response.get_exception()
 *         if exc is not None:             # <<<<<<<<<<<<<<
 *             # someone needs to await the underlying future
 *             # so we do it here, and most probably (like 99%) the exception
 */
  }

  /* "asynctnt/iproto/push.pyx":88
 *             raise exc
 * 
 *         if response.push_len() > 0:             # <<<<<<<<<<<<<<
 *             return response.pop_push()
 * 
 */
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_8Response_push_len(__pyx_cur_scope->__pyx_v_response); if (unlikely(PyErr_Occurred())) __PYX_ERR(27, 88, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_3 > 0);
  if (__pyx_t_2) {

    /* "asynctnt/iproto/push.pyx":89
 * 
 *         if response.push_len() > 0:
 *             return response.pop_push()             # <<<<<<<<<<<<<<
 * 
 *         if response.code_ >= 0:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_8Response_pop_push(__pyx_cur_scope->__pyx_v_response); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/push.pyx":88
 *             raise exc
 * 
 *         if response.push_len() > 0:             # <<<<<<<<<<<<<<
 *             return response.pop_push()
 * 
 */
  }

  /* "asynctnt/iproto/push.pyx":91
 *             return response.pop_push()
 * 
 *         if response.code_ >= 0:             # <<<<<<<<<<<<<<
 *             ev.clear()
 *             raise StopAsyncIteration
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_response->code_ >= 0);
  if (unlikely(__pyx_t_2)) {

    /* "asynctnt/iproto/push.pyx":92
 * 
 *         if response.code_ >= 0:
 *             ev.clear()             # <<<<<<<<<<<<<<
 *             raise StopAsyncIteration
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ev, __pyx_n_s_clear); if (unlikely(!__pyx_t_5)) __PYX_ERR(27, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_3 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_3 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_3, 0+__pyx_t_3);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 92, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/push.pyx":93
 *         if response.code_ >= 0:
 *             ev.clear()
 *             raise StopAsyncIteration             # <<<<<<<<<<<<<<
 * 
 *         assert False, 'Impossible condition happened. ' \
 */
    __Pyx_Raise(((PyObject *)(&((PyTypeObject*)__Pyx_PyExc_StopAsyncIteration)[0])), 0, 0, 0);
    __PYX_ERR(27, 93, __pyx_L1_error)

    /* "asynctnt/iproto/push.pyx":91
 *             return response.pop_push()
 * 
 *         if response.code_ >= 0:             # <<<<<<<<<<<<<<
 *             ev.clear()
 *             raise StopAsyncIteration
 */
  }

  /* "asynctnt/iproto/push.pyx":95
 *             raise StopAsyncIteration
 * 
 *         assert False, 'Impossible condition happened. ' \             # <<<<<<<<<<<<<<
 *                       'Please file a bug to ' \
 *                       'https://github.com/igorcoding/asynctnt'
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    if (unlikely(!0)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Impossible_condition_happened_Pl, 0, 0);
      __PYX_ERR(27, 95, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(27, 95, __pyx_L1_error)
  #endif
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "asynctnt/iproto/push.pyx":62
 *         return self
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __anext__(self):
 *         cdef Response response
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("__anext__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/push.pyx":99
 *                       'https://github.com/igorcoding/asynctnt'
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def response(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_8response_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_8response_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_8response___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_8response___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "asynctnt/iproto/push.pyx":107
 *             :rtype: asynctnt.Response
 *         """
 *         return self._response             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_response);
  __pyx_r = ((PyObject *)__pyx_v_self->_response);
  goto __pyx_L0;

  /* "asynctnt/iproto/push.pyx":99
 *                       'https://github.com/igorcoding/asynctnt'
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def response(self):
 *         """
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_12__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12PushIterator_12__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_12__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_12__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_11__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_11__reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._fut, self._request, self._response)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_fut);
  __Pyx_GIVEREF(__pyx_v_self->_fut);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->_fut)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_request);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_request);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->_request))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_response);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_response);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->_response))) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._fut, self._request, self._response)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._fut, self._request, self._response)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._fut is not None or self._request is not None or self._response is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._fut, self._request, self._response)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._fut is not None or self._request is not None or self._response is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->_fut != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->_request) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->_response) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._fut is not None or self._request is not None or self._response is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._fut is not None or self._request is not None or self._response is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_PushIterator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_159463343);
    __Pyx_GIVEREF(__pyx_int_159463343);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_159463343)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._fut is not None or self._request is not None or self._response is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, None), state
 *     else:
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PushIterator__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_PushIterator); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_159463343);
    __Pyx_GIVEREF(__pyx_int_159463343);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_159463343)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PushIterator.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PushIterator__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_14__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12PushIterator_14__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_14__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_14__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PushIterator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_13__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12PushIterator_13__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PushIterator__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PushIterator__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PushIterator__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.PushIterator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/coreproto.pyx":19
 * 
 * cdef class CoreProtocol:
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  host, port,
 *                  encoding=None,
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_host = 0;
  PyObject *__pyx_v_port = 0;
  PyObject *__pyx_v_encoding = 0;
  PyObject *__pyx_v_initial_read_buffer_size = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_host,&__pyx_n_s_port,&__pyx_n_s_encoding,&__pyx_n_s_initial_read_buffer_size,0};

    /* "asynctnt/iproto/coreproto.pyx":21
 *     def __init__(self,
 *                  host, port,
 *                  encoding=None,             # <<<<<<<<<<<<<<
 *                  initial_read_buffer_size=None):
 *         self.host = host
 */
    values[2] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));

    /* "asynctnt/iproto/coreproto.pyx":22
 *                  host, port,
 *                  encoding=None,
 *                  initial_read_buffer_size=None):             # <<<<<<<<<<<<<<
 *         self.host = host
 *         self.port = port
 */
    values[3] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_host)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 19, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_port)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 19, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(28, 19, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_encoding);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 19, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_initial_read_buffer_size);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 19, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(28, 19, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_host = values[0];
    __pyx_v_port = values[1];
    __pyx_v_encoding = values[2];
    __pyx_v_initial_read_buffer_size = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, __pyx_nargs); __PYX_ERR(28, 19, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol___init__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self), __pyx_v_host, __pyx_v_port, __pyx_v_encoding, __pyx_v_initial_read_buffer_size);

  /* "asynctnt/iproto/coreproto.pyx":19
 * 
 * cdef class CoreProtocol:
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  host, port,
 *                  encoding=None,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol___init__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_host, PyObject *__pyx_v_port, PyObject *__pyx_v_encoding, PyObject *__pyx_v_initial_read_buffer_size) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  size_t __pyx_t_6;
  struct __pyx_opt_args_8asynctnt_6iproto_8protocol_10ReadBuffer_create __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_encoding);
  __Pyx_INCREF(__pyx_v_initial_read_buffer_size);

  /* "asynctnt/iproto/coreproto.pyx":23
 *                  encoding=None,
 *                  initial_read_buffer_size=None):
 *         self.host = host             # <<<<<<<<<<<<<<
 *         self.port = port
 * 
 */
  __Pyx_INCREF(__pyx_v_host);
  __Pyx_GIVEREF(__pyx_v_host);
  __Pyx_GOTREF(__pyx_v_self->host);
  __Pyx_DECREF(__pyx_v_self->host);
  __pyx_v_self->host = __pyx_v_host;

  /* "asynctnt/iproto/coreproto.pyx":24
 *                  initial_read_buffer_size=None):
 *         self.host = host
 *         self.port = port             # <<<<<<<<<<<<<<
 * 
 *         encoding = encoding or b'utf-8'
 */
  __Pyx_INCREF(__pyx_v_port);
  __Pyx_GIVEREF(__pyx_v_port);
  __Pyx_GOTREF(__pyx_v_self->port);
  __Pyx_DECREF(__pyx_v_self->port);
  __pyx_v_self->port = __pyx_v_port;

  /* "asynctnt/iproto/coreproto.pyx":26
 *         self.port = port
 * 
 *         encoding = encoding or b'utf-8'             # <<<<<<<<<<<<<<
 *         if isinstance(encoding, str):
 *             self.encoding = encoding.encode()
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_encoding); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(28, 26, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_encoding);
    __pyx_t_1 = __pyx_v_encoding;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_kp_b_utf_8);
  __pyx_t_1 = __pyx_kp_b_utf_8;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF_SET(__pyx_v_encoding, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/coreproto.pyx":27
 * 
 *         encoding = encoding or b'utf-8'
 *         if isinstance(encoding, str):             # <<<<<<<<<<<<<<
 *             self.encoding = encoding.encode()
 *         elif isinstance(self.encoding, bytes):
 */
  __pyx_t_2 = PyUnicode_Check(__pyx_v_encoding); 
  if (__pyx_t_2) {

    /* "asynctnt/iproto/coreproto.pyx":28
 *         encoding = encoding or b'utf-8'
 *         if isinstance(encoding, str):
 *             self.encoding = encoding.encode()             # <<<<<<<<<<<<<<
 *         elif isinstance(self.encoding, bytes):
 *             self.encoding = encoding
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_encoding, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 28, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 28, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(28, 28, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->encoding);
    __Pyx_DECREF(__pyx_v_self->encoding);
    __pyx_v_self->encoding = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/coreproto.pyx":27
 * 
 *         encoding = encoding or b'utf-8'
 *         if isinstance(encoding, str):             # <<<<<<<<<<<<<<
 *             self.encoding = encoding.encode()
 *         elif isinstance(self.encoding, bytes):
 */
    goto __pyx_L5;
  }

  /* "asynctnt/iproto/coreproto.pyx":29
 *         if isinstance(encoding, str):
 *             self.encoding = encoding.encode()
 *         elif isinstance(self.encoding, bytes):             # <<<<<<<<<<<<<<
 *             self.encoding = encoding
 *         else:
 */
  __pyx_t_1 = __pyx_v_self->encoding;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyBytes_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_2)) {

    /* "asynctnt/iproto/coreproto.pyx":30
 *             self.encoding = encoding.encode()
 *         elif isinstance(self.encoding, bytes):
 *             self.encoding = encoding             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('encoding must be either str or bytes')
 */
    if (!(likely(PyBytes_CheckExact(__pyx_v_encoding))||((__pyx_v_encoding) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_v_encoding))) __PYX_ERR(28, 30, __pyx_L1_error)
    __pyx_t_1 = __pyx_v_encoding;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->encoding);
    __Pyx_DECREF(__pyx_v_self->encoding);
    __pyx_v_self->encoding = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/coreproto.pyx":29
 *         if isinstance(encoding, str):
 *             self.encoding = encoding.encode()
 *         elif isinstance(self.encoding, bytes):             # <<<<<<<<<<<<<<
 *             self.encoding = encoding
 *         else:
 */
    goto __pyx_L5;
  }

  /* "asynctnt/iproto/coreproto.pyx":32
 *             self.encoding = encoding
 *         else:
 *             raise TypeError('encoding must be either str or bytes')             # <<<<<<<<<<<<<<
 * 
 *         initial_read_buffer_size = initial_read_buffer_size or 0x20000
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(28, 32, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "asynctnt/iproto/coreproto.pyx":34
 *             raise TypeError('encoding must be either str or bytes')
 * 
 *         initial_read_buffer_size = initial_read_buffer_size or 0x20000             # <<<<<<<<<<<<<<
 *         self.transport = None
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_initial_read_buffer_size); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(28, 34, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_initial_read_buffer_size);
    __pyx_t_1 = __pyx_v_initial_read_buffer_size;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyInt_From_long(0x20000); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF_SET(__pyx_v_initial_read_buffer_size, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/coreproto.pyx":35
 * 
 *         initial_read_buffer_size = initial_read_buffer_size or 0x20000
 *         self.transport = None             # <<<<<<<<<<<<<<
 * 
 *         self.rbuf = ReadBuffer.create(encoding, initial_read_buffer_size)
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_transport, Py_None) < 0) __PYX_ERR(28, 35, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":37
 *         self.transport = None
 * 
 *         self.rbuf = ReadBuffer.create(encoding, initial_read_buffer_size)             # <<<<<<<<<<<<<<
 *         self.state = PROTOCOL_IDLE
 *         self.con_state = CONNECTION_BAD
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_encoding))||((__pyx_v_encoding) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_encoding))) __PYX_ERR(28, 37, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_size_t(__pyx_v_initial_read_buffer_size); if (unlikely((__pyx_t_6 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(28, 37, __pyx_L1_error)
  __pyx_t_7.__pyx_n = 1;
  __pyx_t_7.initial_buffer_size = __pyx_t_6;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_8asynctnt_6iproto_8protocol_ReadBuffer->create(((PyObject*)__pyx_v_encoding), &__pyx_t_7)); if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->rbuf);
  __Pyx_DECREF((PyObject *)__pyx_v_self->rbuf);
  __pyx_v_self->rbuf = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/coreproto.pyx":38
 * 
 *         self.rbuf = ReadBuffer.create(encoding, initial_read_buffer_size)
 *         self.state = PROTOCOL_IDLE             # <<<<<<<<<<<<<<
 *         self.con_state = CONNECTION_BAD
 * 
 */
  __pyx_v_self->state = __pyx_e_8asynctnt_6iproto_8protocol_PROTOCOL_IDLE;

  /* "asynctnt/iproto/coreproto.pyx":39
 *         self.rbuf = ReadBuffer.create(encoding, initial_read_buffer_size)
 *         self.state = PROTOCOL_IDLE
 *         self.con_state = CONNECTION_BAD             # <<<<<<<<<<<<<<
 * 
 *         self.version = None
 */
  __pyx_v_self->con_state = __pyx_e_8asynctnt_6iproto_8protocol_CONNECTION_BAD;

  /* "asynctnt/iproto/coreproto.pyx":41
 *         self.con_state = CONNECTION_BAD
 * 
 *         self.version = None             # <<<<<<<<<<<<<<
 *         self.salt = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->version);
  __Pyx_DECREF(__pyx_v_self->version);
  __pyx_v_self->version = ((PyObject*)Py_None);

  /* "asynctnt/iproto/coreproto.pyx":42
 * 
 *         self.version = None
 *         self.salt = None             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _is_connected(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->salt);
  __Pyx_DECREF(__pyx_v_self->salt);
  __pyx_v_self->salt = ((PyObject*)Py_None);

  /* "asynctnt/iproto/coreproto.pyx":19
 * 
 * cdef class CoreProtocol:
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  host, port,
 *                  encoding=None,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoding);
  __Pyx_XDECREF(__pyx_v_initial_read_buffer_size);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/coreproto.pyx":44
 *         self.salt = None
 * 
 *     cdef bint _is_connected(self):             # <<<<<<<<<<<<<<
 *         return self.con_state != CONNECTION_BAD
 * 
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__is_connected(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self) {
  int __pyx_r;

  /* "asynctnt/iproto/coreproto.pyx":45
 * 
 *     cdef bint _is_connected(self):
 *         return self.con_state != CONNECTION_BAD             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _is_fully_connected(self):
 */
  __pyx_r = (__pyx_v_self->con_state != __pyx_e_8asynctnt_6iproto_8protocol_CONNECTION_BAD);
  goto __pyx_L0;

  /* "asynctnt/iproto/coreproto.pyx":44
 *         self.salt = None
 * 
 *     cdef bint _is_connected(self):             # <<<<<<<<<<<<<<
 *         return self.con_state != CONNECTION_BAD
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/coreproto.pyx":47
 *         return self.con_state != CONNECTION_BAD
 * 
 *     cdef bint _is_fully_connected(self):             # <<<<<<<<<<<<<<
 *         return self.con_state == CONNECTION_FULL
 * 
 */

static int __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__is_fully_connected(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self) {
  int __pyx_r;

  /* "asynctnt/iproto/coreproto.pyx":48
 * 
 *     cdef bint _is_fully_connected(self):
 *         return self.con_state == CONNECTION_FULL             # <<<<<<<<<<<<<<
 * 
 *     def is_connected(self):
 */
  __pyx_r = (__pyx_v_self->con_state == __pyx_e_8asynctnt_6iproto_8protocol_CONNECTION_FULL);
  goto __pyx_L0;

  /* "asynctnt/iproto/coreproto.pyx":47
 *         return self.con_state != CONNECTION_BAD
 * 
 *     cdef bint _is_fully_connected(self):             # <<<<<<<<<<<<<<
 *         return self.con_state == CONNECTION_FULL
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/coreproto.pyx":50
 *         return self.con_state == CONNECTION_FULL
 * 
 *     def is_connected(self):             # <<<<<<<<<<<<<<
 *         return self._is_connected()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_3is_connected(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_3is_connected = {"is_connected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_3is_connected, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_3is_connected(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_connected (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_connected", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_connected", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_2is_connected(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_2is_connected(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_connected", 1);

  /* "asynctnt/iproto/coreproto.pyx":51
 * 
 *     def is_connected(self):
 *         return self._is_connected()             # <<<<<<<<<<<<<<
 * 
 *     def is_fully_connected(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self->__pyx_vtab)->_is_connected(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 51, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/coreproto.pyx":50
 *         return self.con_state == CONNECTION_FULL
 * 
 *     def is_connected(self):             # <<<<<<<<<<<<<<
 *         return self._is_connected()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.is_connected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/coreproto.pyx":53
 *         return self._is_connected()
 * 
 *     def is_fully_connected(self):             # <<<<<<<<<<<<<<
 *         return self._is_fully_connected()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_5is_fully_connected(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_5is_fully_connected = {"is_fully_connected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_5is_fully_connected, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_5is_fully_connected(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_fully_connected (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_fully_connected", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_fully_connected", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_4is_fully_connected(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_4is_fully_connected(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_fully_connected", 1);

  /* "asynctnt/iproto/coreproto.pyx":54
 * 
 *     def is_fully_connected(self):
 *         return self._is_fully_connected()             # <<<<<<<<<<<<<<
 * 
 *     def get_version(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self->__pyx_vtab)->_is_fully_connected(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 54, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/coreproto.pyx":53
 *         return self._is_connected()
 * 
 *     def is_fully_connected(self):             # <<<<<<<<<<<<<<
 *         return self._is_fully_connected()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.is_fully_connected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/coreproto.pyx":56
 *         return self._is_fully_connected()
 * 
 *     def get_version(self):             # <<<<<<<<<<<<<<
 *         return self.version
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_7get_version(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_7get_version = {"get_version", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_7get_version, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_7get_version(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_version (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_version", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_version", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_6get_version(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_6get_version(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_version", 1);

  /* "asynctnt/iproto/coreproto.pyx":57
 * 
 *     def get_version(self):
 *         return self.version             # <<<<<<<<<<<<<<
 * 
 *     cdef void _write(self, buf) except *:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->version);
  __pyx_r = __pyx_v_self->version;
  goto __pyx_L0;

  /* "asynctnt/iproto/coreproto.pyx":56
 *         return self._is_fully_connected()
 * 
 *     def get_version(self):             # <<<<<<<<<<<<<<
 *         return self.version
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/coreproto.pyx":59
 *         return self.version
 * 
 *     cdef void _write(self, buf) except *:             # <<<<<<<<<<<<<<
 *         self.transport.write(memoryview(buf))
 * 
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__write(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_buf) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_write", 1);

  /* "asynctnt/iproto/coreproto.pyx":60
 * 
 *     cdef void _write(self, buf) except *:
 *         self.transport.write(memoryview(buf))             # <<<<<<<<<<<<<<
 * 
 *     cdef void _on_data_received(self, data):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_transport); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyMemoryView_FromObject(__pyx_v_buf); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/coreproto.pyx":59
 *         return self.version
 * 
 *     cdef void _write(self, buf) except *:             # <<<<<<<<<<<<<<
 *         self.transport.write(memoryview(buf))
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol._write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/coreproto.pyx":62
 *         self.transport.write(memoryview(buf))
 * 
 *     cdef void _on_data_received(self, data):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t ruse, curr
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_data_received(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_data) {
  char const *__pyx_v_p;
  char const *__pyx_v_q;
  uint32_t __pyx_v_packet_len;
  char *__pyx_v_data_str;
  Py_ssize_t __pyx_v_data_len;
  Py_ssize_t __pyx_v_buf_len;
  char *__pyx_v_end;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_data_received", 1);

  /* "asynctnt/iproto/coreproto.pyx":73
 *             ssize_t buf_len
 * 
 *         data_str = NULL             # <<<<<<<<<<<<<<
 *         data_len = 0
 * 
 */
  __pyx_v_data_str = NULL;

  /* "asynctnt/iproto/coreproto.pyx":74
 * 
 *         data_str = NULL
 *         data_len = 0             # <<<<<<<<<<<<<<
 * 
 *         if cpython.PyBytes_CheckExact(data):
 */
  __pyx_v_data_len = 0;

  /* "asynctnt/iproto/coreproto.pyx":76
 *         data_len = 0
 * 
 *         if cpython.PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *             cpython.bytes.PyBytes_AsStringAndSize(<bytes> data,
 *                                                   &data_str,
 */
  __pyx_t_1 = PyBytes_CheckExact(__pyx_v_data);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/coreproto.pyx":77
 * 
 *         if cpython.PyBytes_CheckExact(data):
 *             cpython.bytes.PyBytes_AsStringAndSize(<bytes> data,             # <<<<<<<<<<<<<<
 *                                                   &data_str,
 *                                                   &data_len)
 */
    __pyx_t_2 = PyBytes_AsStringAndSize(__pyx_v_data, (&__pyx_v_data_str), ((Py_ssize_t *)(&__pyx_v_data_len))); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(28, 77, __pyx_L1_error)

    /* "asynctnt/iproto/coreproto.pyx":80
 *                                                   &data_str,
 *                                                   &data_len)
 *             self.rbuf.extend(data_str, data_len)             # <<<<<<<<<<<<<<
 *         elif cpython.bytearray.PyByteArray_CheckExact(data):
 *             data_len = cpython.bytearray.PyByteArray_Size(<bytearray> data)
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_extend(__pyx_v_self->rbuf, __pyx_v_data_str, __pyx_v_data_len); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(28, 80, __pyx_L1_error)

    /* "asynctnt/iproto/coreproto.pyx":76
 *         data_len = 0
 * 
 *         if cpython.PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *             cpython.bytes.PyBytes_AsStringAndSize(<bytes> data,
 *                                                   &data_str,
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/coreproto.pyx":81
 *                                                   &data_len)
 *             self.rbuf.extend(data_str, data_len)
 *         elif cpython.bytearray.PyByteArray_CheckExact(data):             # <<<<<<<<<<<<<<
 *             data_len = cpython.bytearray.PyByteArray_Size(<bytearray> data)
 *             self.rbuf.extend(cpython.bytearray.PyByteArray_AsString(<bytearray> data), data_len)
 */
  __pyx_t_1 = PyByteArray_CheckExact(__pyx_v_data);
  if (likely(__pyx_t_1)) {

    /* "asynctnt/iproto/coreproto.pyx":82
 *             self.rbuf.extend(data_str, data_len)
 *         elif cpython.bytearray.PyByteArray_CheckExact(data):
 *             data_len = cpython.bytearray.PyByteArray_Size(<bytearray> data)             # <<<<<<<<<<<<<<
 *             self.rbuf.extend(cpython.bytearray.PyByteArray_AsString(<bytearray> data), data_len)
 *         else:
 */
    __pyx_v_data_len = PyByteArray_Size(__pyx_v_data);

    /* "asynctnt/iproto/coreproto.pyx":83
 *         elif cpython.bytearray.PyByteArray_CheckExact(data):
 *             data_len = cpython.bytearray.PyByteArray_Size(<bytearray> data)
 *             self.rbuf.extend(cpython.bytearray.PyByteArray_AsString(<bytearray> data), data_len)             # <<<<<<<<<<<<<<
 *         else:
 *             raise BufferError(
 */
    __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_extend(__pyx_v_self->rbuf, PyByteArray_AsString(__pyx_v_data), __pyx_v_data_len); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(28, 83, __pyx_L1_error)

    /* "asynctnt/iproto/coreproto.pyx":81
 *                                                   &data_len)
 *             self.rbuf.extend(data_str, data_len)
 *         elif cpython.bytearray.PyByteArray_CheckExact(data):             # <<<<<<<<<<<<<<
 *             data_len = cpython.bytearray.PyByteArray_Size(<bytearray> data)
 *             self.rbuf.extend(cpython.bytearray.PyByteArray_AsString(<bytearray> data), data_len)
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/coreproto.pyx":85
 *             self.rbuf.extend(cpython.bytearray.PyByteArray_AsString(<bytearray> data), data_len)
 *         else:
 *             raise BufferError(             # <<<<<<<<<<<<<<
 *                 '_on_data_received: expected bytes or bytearray object, got {}'.format(
 *                     type(data),
 */
  /*else*/ {

    /* "asynctnt/iproto/coreproto.pyx":86
 *         else:
 *             raise BufferError(
 *                 '_on_data_received: expected bytes or bytearray object, got {}'.format(             # <<<<<<<<<<<<<<
 *                     type(data),
 *                 ))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_on_data_received_expected_bytes, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(28, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "asynctnt/iproto/coreproto.pyx":87
 *             raise BufferError(
 *                 '_on_data_received: expected bytes or bytearray object, got {}'.format(
 *                     type(data),             # <<<<<<<<<<<<<<
 *                 ))
 * 
 */
    __pyx_t_5 = NULL;
    __pyx_t_2 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_2 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)Py_TYPE(__pyx_v_data))};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_2, 1+__pyx_t_2);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 86, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }

    /* "asynctnt/iproto/coreproto.pyx":85
 *             self.rbuf.extend(cpython.bytearray.PyByteArray_AsString(<bytearray> data), data_len)
 *         else:
 *             raise BufferError(             # <<<<<<<<<<<<<<
 *                 '_on_data_received: expected bytes or bytearray object, got {}'.format(
 *                     type(data),
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_BufferError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(28, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(28, 85, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "asynctnt/iproto/coreproto.pyx":90
 *                 ))
 * 
 *         if data_len == 0:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_data_len == 0);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/coreproto.pyx":91
 * 
 *         if data_len == 0:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if self.state == PROTOCOL_GREETING:
 */
    goto __pyx_L0;

    /* "asynctnt/iproto/coreproto.pyx":90
 *                 ))
 * 
 *         if data_len == 0:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "asynctnt/iproto/coreproto.pyx":93
 *             return
 * 
 *         if self.state == PROTOCOL_GREETING:             # <<<<<<<<<<<<<<
 *             if self.rbuf.use < IPROTO_GREETING_SIZE:
 *                 # not enough for greeting
 */
  switch (__pyx_v_self->state) {
    case __pyx_e_8asynctnt_6iproto_8protocol_PROTOCOL_GREETING:

    /* "asynctnt/iproto/coreproto.pyx":94
 * 
 *         if self.state == PROTOCOL_GREETING:
 *             if self.rbuf.use < IPROTO_GREETING_SIZE:             # <<<<<<<<<<<<<<
 *                 # not enough for greeting
 *                 return
 */
    __pyx_t_1 = (__pyx_v_self->rbuf->use < 0x80);
    if (__pyx_t_1) {

      /* "asynctnt/iproto/coreproto.pyx":96
 *             if self.rbuf.use < IPROTO_GREETING_SIZE:
 *                 # not enough for greeting
 *                 return             # <<<<<<<<<<<<<<
 *             self._process__greeting()
 *             self.rbuf.move(IPROTO_GREETING_SIZE)
 */
      goto __pyx_L0;

      /* "asynctnt/iproto/coreproto.pyx":94
 * 
 *         if self.state == PROTOCOL_GREETING:
 *             if self.rbuf.use < IPROTO_GREETING_SIZE:             # <<<<<<<<<<<<<<
 *                 # not enough for greeting
 *                 return
 */
    }

    /* "asynctnt/iproto/coreproto.pyx":97
 *                 # not enough for greeting
 *                 return
 *             self._process__greeting()             # <<<<<<<<<<<<<<
 *             self.rbuf.move(IPROTO_GREETING_SIZE)
 *         elif self.state == PROTOCOL_NORMAL:
 */
    ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self->__pyx_vtab)->_process__greeting(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 97, __pyx_L1_error)

    /* "asynctnt/iproto/coreproto.pyx":98
 *                 return
 *             self._process__greeting()
 *             self.rbuf.move(IPROTO_GREETING_SIZE)             # <<<<<<<<<<<<<<
 *         elif self.state == PROTOCOL_NORMAL:
 *             p = self.rbuf.buf
 */
    __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_move(__pyx_v_self->rbuf, 0x80); if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 98, __pyx_L1_error)

    /* "asynctnt/iproto/coreproto.pyx":93
 *             return
 * 
 *         if self.state == PROTOCOL_GREETING:             # <<<<<<<<<<<<<<
 *             if self.rbuf.use < IPROTO_GREETING_SIZE:
 *                 # not enough for greeting
 */
    break;
    case __pyx_e_8asynctnt_6iproto_8protocol_PROTOCOL_NORMAL:

    /* "asynctnt/iproto/coreproto.pyx":100
 *             self.rbuf.move(IPROTO_GREETING_SIZE)
 *         elif self.state == PROTOCOL_NORMAL:
 *             p = self.rbuf.buf             # <<<<<<<<<<<<<<
 *             end = &self.rbuf.buf[self.rbuf.use]
 * 
 */
    __pyx_t_6 = __pyx_v_self->rbuf->buf;
    __pyx_v_p = __pyx_t_6;

    /* "asynctnt/iproto/coreproto.pyx":101
 *         elif self.state == PROTOCOL_NORMAL:
 *             p = self.rbuf.buf
 *             end = &self.rbuf.buf[self.rbuf.use]             # <<<<<<<<<<<<<<
 * 
 *             while p < end:
 */
    __pyx_v_end = (&(__pyx_v_self->rbuf->buf[__pyx_v_self->rbuf->use]));

    /* "asynctnt/iproto/coreproto.pyx":103
 *             end = &self.rbuf.buf[self.rbuf.use]
 * 
 *             while p < end:             # <<<<<<<<<<<<<<
 *                 q = p  # q is temporary to parse packet length
 *                 buf_len = end - p
 */
    while (1) {
      __pyx_t_1 = (__pyx_v_p < __pyx_v_end);
      if (!__pyx_t_1) break;

      /* "asynctnt/iproto/coreproto.pyx":104
 * 
 *             while p < end:
 *                 q = p  # q is temporary to parse packet length             # <<<<<<<<<<<<<<
 *                 buf_len = end - p
 *                 if buf_len < 5:
 */
      __pyx_v_q = __pyx_v_p;

      /* "asynctnt/iproto/coreproto.pyx":105
 *             while p < end:
 *                 q = p  # q is temporary to parse packet length
 *                 buf_len = end - p             # <<<<<<<<<<<<<<
 *                 if buf_len < 5:
 *                     # not enough
 */
      __pyx_v_buf_len = (__pyx_v_end - __pyx_v_p);

      /* "asynctnt/iproto/coreproto.pyx":106
 *                 q = p  # q is temporary to parse packet length
 *                 buf_len = end - p
 *                 if buf_len < 5:             # <<<<<<<<<<<<<<
 *                     # not enough
 *                     break
 */
      __pyx_t_1 = (__pyx_v_buf_len < 5);
      if (__pyx_t_1) {

        /* "asynctnt/iproto/coreproto.pyx":108
 *                 if buf_len < 5:
 *                     # not enough
 *                     break             # <<<<<<<<<<<<<<
 * 
 *                 q = &q[1]  # skip to 2nd byte of packet length
 */
        goto __pyx_L7_break;

        /* "asynctnt/iproto/coreproto.pyx":106
 *                 q = p  # q is temporary to parse packet length
 *                 buf_len = end - p
 *                 if buf_len < 5:             # <<<<<<<<<<<<<<
 *                     # not enough
 *                     break
 */
      }

      /* "asynctnt/iproto/coreproto.pyx":110
 *                     break
 * 
 *                 q = &q[1]  # skip to 2nd byte of packet length             # <<<<<<<<<<<<<<
 *                 packet_len = mp_load_u32(&q)
 * 
 */
      __pyx_v_q = (&(__pyx_v_q[1]));

      /* "asynctnt/iproto/coreproto.pyx":111
 * 
 *                 q = &q[1]  # skip to 2nd byte of packet length
 *                 packet_len = mp_load_u32(&q)             # <<<<<<<<<<<<<<
 * 
 *                 if buf_len < 5 + packet_len:
 */
      __pyx_v_packet_len = mp_load_u32((&__pyx_v_q));

      /* "asynctnt/iproto/coreproto.pyx":113
 *                 packet_len = mp_load_u32(&q)
 * 
 *                 if buf_len < 5 + packet_len:             # <<<<<<<<<<<<<<
 *                     # not enough to read an entire packet
 *                     break
 */
      __pyx_t_1 = (__pyx_v_buf_len < (5 + __pyx_v_packet_len));
      if (__pyx_t_1) {

        /* "asynctnt/iproto/coreproto.pyx":115
 *                 if buf_len < 5 + packet_len:
 *                     # not enough to read an entire packet
 *                     break             # <<<<<<<<<<<<<<
 * 
 *                 p = &p[5]  # skip length header
 */
        goto __pyx_L7_break;

        /* "asynctnt/iproto/coreproto.pyx":113
 *                 packet_len = mp_load_u32(&q)
 * 
 *                 if buf_len < 5 + packet_len:             # <<<<<<<<<<<<<<
 *                     # not enough to read an entire packet
 *                     break
 */
      }

      /* "asynctnt/iproto/coreproto.pyx":117
 *                     break
 * 
 *                 p = &p[5]  # skip length header             # <<<<<<<<<<<<<<
 *                 self._on_response_received(p, packet_len)
 *                 p = &p[packet_len]
 */
      __pyx_v_p = (&(__pyx_v_p[5]));

      /* "asynctnt/iproto/coreproto.pyx":118
 * 
 *                 p = &p[5]  # skip length header
 *                 self._on_response_received(p, packet_len)             # <<<<<<<<<<<<<<
 *                 p = &p[packet_len]
 * 
 */
      ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self->__pyx_vtab)->_on_response_received(__pyx_v_self, __pyx_v_p, __pyx_v_packet_len); if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 118, __pyx_L1_error)

      /* "asynctnt/iproto/coreproto.pyx":119
 *                 p = &p[5]  # skip length header
 *                 self._on_response_received(p, packet_len)
 *                 p = &p[packet_len]             # <<<<<<<<<<<<<<
 * 
 *                 if p == end:
 */
      __pyx_v_p = (&(__pyx_v_p[__pyx_v_packet_len]));

      /* "asynctnt/iproto/coreproto.pyx":121
 *                 p = &p[packet_len]
 * 
 *                 if p == end:             # <<<<<<<<<<<<<<
 *                     self.rbuf.use = 0
 *                     break
 */
      __pyx_t_1 = (__pyx_v_p == __pyx_v_end);
      if (__pyx_t_1) {

        /* "asynctnt/iproto/coreproto.pyx":122
 * 
 *                 if p == end:
 *                     self.rbuf.use = 0             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        __pyx_v_self->rbuf->use = 0;

        /* "asynctnt/iproto/coreproto.pyx":123
 *                 if p == end:
 *                     self.rbuf.use = 0
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             self.rbuf.use = end - p
 */
        goto __pyx_L7_break;

        /* "asynctnt/iproto/coreproto.pyx":121
 *                 p = &p[packet_len]
 * 
 *                 if p == end:             # <<<<<<<<<<<<<<
 *                     self.rbuf.use = 0
 *                     break
 */
      }
    }
    __pyx_L7_break:;

    /* "asynctnt/iproto/coreproto.pyx":125
 *                     break
 * 
 *             self.rbuf.use = end - p             # <<<<<<<<<<<<<<
 *             if self.rbuf.use > 0:
 *                 self.rbuf.move_offset(p - self.rbuf.buf, self.rbuf.use)
 */
    __pyx_v_self->rbuf->use = (__pyx_v_end - __pyx_v_p);

    /* "asynctnt/iproto/coreproto.pyx":126
 * 
 *             self.rbuf.use = end - p
 *             if self.rbuf.use > 0:             # <<<<<<<<<<<<<<
 *                 self.rbuf.move_offset(p - self.rbuf.buf, self.rbuf.use)
 *         else:
 */
    __pyx_t_1 = (__pyx_v_self->rbuf->use > 0);
    if (__pyx_t_1) {

      /* "asynctnt/iproto/coreproto.pyx":127
 *             self.rbuf.use = end - p
 *             if self.rbuf.use > 0:
 *                 self.rbuf.move_offset(p - self.rbuf.buf, self.rbuf.use)             # <<<<<<<<<<<<<<
 *         else:
 *             # TODO: raise exception
 */
      __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_move_offset(__pyx_v_self->rbuf, (__pyx_v_p - __pyx_v_self->rbuf->buf), __pyx_v_self->rbuf->use); if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 127, __pyx_L1_error)

      /* "asynctnt/iproto/coreproto.pyx":126
 * 
 *             self.rbuf.use = end - p
 *             if self.rbuf.use > 0:             # <<<<<<<<<<<<<<
 *                 self.rbuf.move_offset(p - self.rbuf.buf, self.rbuf.use)
 *         else:
 */
    }

    /* "asynctnt/iproto/coreproto.pyx":99
 *             self._process__greeting()
 *             self.rbuf.move(IPROTO_GREETING_SIZE)
 *         elif self.state == PROTOCOL_NORMAL:             # <<<<<<<<<<<<<<
 *             p = self.rbuf.buf
 *             end = &self.rbuf.buf[self.rbuf.use]
 */
    break;
    default:
    break;
  }

  /* "asynctnt/iproto/coreproto.pyx":62
 *         self.transport.write(memoryview(buf))
 * 
 *     cdef void _on_data_received(self, data):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t ruse, curr
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol._on_data_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/coreproto.pyx":132
 *             pass
 * 
 *     cdef void _process__greeting(self):             # <<<<<<<<<<<<<<
 *         cdef size_t ver_length = TARANTOOL_VERSION_LENGTH
 *         rbuf = self.rbuf
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__process__greeting(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self) {
  size_t __pyx_v_ver_length;
  CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *__pyx_v_rbuf = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_process__greeting", 1);

  /* "asynctnt/iproto/coreproto.pyx":133
 * 
 *     cdef void _process__greeting(self):
 *         cdef size_t ver_length = TARANTOOL_VERSION_LENGTH             # <<<<<<<<<<<<<<
 *         rbuf = self.rbuf
 *         self.version = self._parse_version(self.rbuf.get_slice_end(ver_length))
 */
  __pyx_v_ver_length = 64;

  /* "asynctnt/iproto/coreproto.pyx":134
 *     cdef void _process__greeting(self):
 *         cdef size_t ver_length = TARANTOOL_VERSION_LENGTH
 *         rbuf = self.rbuf             # <<<<<<<<<<<<<<
 *         self.version = self._parse_version(self.rbuf.get_slice_end(ver_length))
 *         self.salt = base64.b64decode(
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->rbuf);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_rbuf = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/coreproto.pyx":135
 *         cdef size_t ver_length = TARANTOOL_VERSION_LENGTH
 *         rbuf = self.rbuf
 *         self.version = self._parse_version(self.rbuf.get_slice_end(ver_length))             # <<<<<<<<<<<<<<
 *         self.salt = base64.b64decode(
 *             self.rbuf.get_slice(ver_length,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_parse_version); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice_end(__pyx_v_self->rbuf, __pyx_v_ver_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(28, 135, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->version);
  __Pyx_DECREF(__pyx_v_self->version);
  __pyx_v_self->version = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/coreproto.pyx":136
 *         rbuf = self.rbuf
 *         self.version = self._parse_version(self.rbuf.get_slice_end(ver_length))
 *         self.salt = base64.b64decode(             # <<<<<<<<<<<<<<
 *             self.rbuf.get_slice(ver_length,
 *                                 ver_length + SALT_LENGTH)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_base64); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_b64decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/coreproto.pyx":137
 *         self.version = self._parse_version(self.rbuf.get_slice_end(ver_length))
 *         self.salt = base64.b64decode(
 *             self.rbuf.get_slice(ver_length,             # <<<<<<<<<<<<<<
 *                                 ver_length + SALT_LENGTH)
 *         )[:SCRAMBLE_SIZE]
 */
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice(__pyx_v_self->rbuf, __pyx_v_ver_length, (__pyx_v_ver_length + 44)); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "asynctnt/iproto/coreproto.pyx":139
 *             self.rbuf.get_slice(ver_length,
 *                                 ver_length + SALT_LENGTH)
 *         )[:SCRAMBLE_SIZE]             # <<<<<<<<<<<<<<
 *         self.state = PROTOCOL_NORMAL
 *         self._on_greeting_received()
 */
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 20, NULL, NULL, &__pyx_slice__39, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(28, 139, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":136
 *         rbuf = self.rbuf
 *         self.version = self._parse_version(self.rbuf.get_slice_end(ver_length))
 *         self.salt = base64.b64decode(             # <<<<<<<<<<<<<<
 *             self.rbuf.get_slice(ver_length,
 *                                 ver_length + SALT_LENGTH)
 */
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->salt);
  __Pyx_DECREF(__pyx_v_self->salt);
  __pyx_v_self->salt = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/coreproto.pyx":140
 *                                 ver_length + SALT_LENGTH)
 *         )[:SCRAMBLE_SIZE]
 *         self.state = PROTOCOL_NORMAL             # <<<<<<<<<<<<<<
 *         self._on_greeting_received()
 * 
 */
  __pyx_v_self->state = __pyx_e_8asynctnt_6iproto_8protocol_PROTOCOL_NORMAL;

  /* "asynctnt/iproto/coreproto.pyx":141
 *         )[:SCRAMBLE_SIZE]
 *         self.state = PROTOCOL_NORMAL
 *         self._on_greeting_received()             # <<<<<<<<<<<<<<
 * 
 *     def _parse_version(self, version):
 */
  ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self->__pyx_vtab)->_on_greeting_received(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 141, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":132
 *             pass
 * 
 *     cdef void _process__greeting(self):             # <<<<<<<<<<<<<<
 *         cdef size_t ver_length = TARANTOOL_VERSION_LENGTH
 *         rbuf = self.rbuf
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol._process__greeting", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_rbuf);
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/coreproto.pyx":143
 *         self._on_greeting_received()
 * 
 *     def _parse_version(self, version):             # <<<<<<<<<<<<<<
 *         m = VERSION_STRING_REGEX.match(version.decode('ascii'))
 *         if m is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_9_parse_version(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_9_parse_version = {"_parse_version", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_9_parse_version, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_9_parse_version(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_version = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_version (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_version,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_version)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 143, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_parse_version") < 0)) __PYX_ERR(28, 143, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_version = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_parse_version", 1, 1, 1, __pyx_nargs); __PYX_ERR(28, 143, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol._parse_version", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_8_parse_version(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self), __pyx_v_version);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_8_parse_version(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_version) {
  PyObject *__pyx_v_m = NULL;
  PyObject *__pyx_v_ver = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_version", 1);

  /* "asynctnt/iproto/coreproto.pyx":144
 * 
 *     def _parse_version(self, version):
 *         m = VERSION_STRING_REGEX.match(version.decode('ascii'))             # <<<<<<<<<<<<<<
 *         if m is not None:
 *             ver = m.group(1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_VERSION_STRING_REGEX); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_version, __pyx_n_s_decode); if (unlikely(!__pyx_t_4)) __PYX_ERR(28, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_n_u_ascii};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_m = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/coreproto.pyx":145
 *     def _parse_version(self, version):
 *         m = VERSION_STRING_REGEX.match(version.decode('ascii'))
 *         if m is not None:             # <<<<<<<<<<<<<<
 *             ver = m.group(1)
 *             return tuple(map(int, ver.split('.')))
 */
  __pyx_t_7 = (__pyx_v_m != Py_None);
  if (__pyx_t_7) {

    /* "asynctnt/iproto/coreproto.pyx":146
 *         m = VERSION_STRING_REGEX.match(version.decode('ascii'))
 *         if m is not None:
 *             ver = m.group(1)             # <<<<<<<<<<<<<<
 *             return tuple(map(int, ver.split('.')))
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_int_1};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_v_ver = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/coreproto.pyx":147
 *         if m is not None:
 *             ver = m.group(1)
 *             return tuple(map(int, ver.split('.')))             # <<<<<<<<<<<<<<
 * 
 *     cdef void _on_greeting_received(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ver, __pyx_n_s_split); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_kp_u__2};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 147, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF((PyObject *)(&PyInt_Type));
    __Pyx_GIVEREF((PyObject *)(&PyInt_Type));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyInt_Type)))) __PYX_ERR(28, 147, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(28, 147, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "asynctnt/iproto/coreproto.pyx":145
 *     def _parse_version(self, version):
 *         m = VERSION_STRING_REGEX.match(version.decode('ascii'))
 *         if m is not None:             # <<<<<<<<<<<<<<
 *             ver = m.group(1)
 *             return tuple(map(int, ver.split('.')))
 */
  }

  /* "asynctnt/iproto/coreproto.pyx":143
 *         self._on_greeting_received()
 * 
 *     def _parse_version(self, version):             # <<<<<<<<<<<<<<
 *         m = VERSION_STRING_REGEX.match(version.decode('ascii'))
 *         if m is not None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol._parse_version", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_XDECREF(__pyx_v_ver);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/coreproto.pyx":149
 *             return tuple(map(int, ver.split('.')))
 * 
 *     cdef void _on_greeting_received(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_greeting_received(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self) {

  /* function exit code */
}

/* "asynctnt/iproto/coreproto.pyx":152
 *         pass
 * 
 *     cdef void _on_response_received(self, const char *buf, uint32_t buf_len):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_response_received(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, CYTHON_UNUSED char const *__pyx_v_buf, CYTHON_UNUSED uint32_t __pyx_v_buf_len) {

  /* function exit code */
}

/* "asynctnt/iproto/coreproto.pyx":155
 *         pass
 * 
 *     cdef void _on_connection_made(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_connection_made(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self) {

  /* function exit code */
}

/* "asynctnt/iproto/coreproto.pyx":158
 *         pass
 * 
 *     cdef void _on_connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_connection_lost(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_exc) {

  /* function exit code */
}

/* "asynctnt/iproto/coreproto.pyx":163
 *     # asyncio callbacks
 * 
 *     def data_received(self, data):             # <<<<<<<<<<<<<<
 *         self._on_data_received(data)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_11data_received(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_11data_received = {"data_received", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_11data_received, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_11data_received(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("data_received (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 163, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "data_received") < 0)) __PYX_ERR(28, 163, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("data_received", 1, 1, 1, __pyx_nargs); __PYX_ERR(28, 163, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.data_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_10data_received(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self), __pyx_v_data);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_10data_received(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("data_received", 1);

  /* "asynctnt/iproto/coreproto.pyx":164
 * 
 *     def data_received(self, data):
 *         self._on_data_received(data)             # <<<<<<<<<<<<<<
 * 
 *     def connection_made(self, transport):
 */
  ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self->__pyx_vtab)->_on_data_received(__pyx_v_self, __pyx_v_data); if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 164, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":163
 *     # asyncio callbacks
 * 
 *     def data_received(self, data):             # <<<<<<<<<<<<<<
 *         self._on_data_received(data)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.data_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/coreproto.pyx":166
 *         self._on_data_received(data)
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         self.transport = transport
 *         self.con_state = CONNECTION_CONNECTED
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_13connection_made(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_13connection_made = {"connection_made", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_13connection_made, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_13connection_made(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_transport = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connection_made (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_transport,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_transport)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 166, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "connection_made") < 0)) __PYX_ERR(28, 166, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_transport = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("connection_made", 1, 1, 1, __pyx_nargs); __PYX_ERR(28, 166, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_12connection_made(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self), __pyx_v_transport);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_12connection_made(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_transport) {
  PyObject *__pyx_v_sock = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connection_made", 1);

  /* "asynctnt/iproto/coreproto.pyx":167
 * 
 *     def connection_made(self, transport):
 *         self.transport = transport             # <<<<<<<<<<<<<<
 *         self.con_state = CONNECTION_CONNECTED
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_transport, __pyx_v_transport) < 0) __PYX_ERR(28, 167, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":168
 *     def connection_made(self, transport):
 *         self.transport = transport
 *         self.con_state = CONNECTION_CONNECTED             # <<<<<<<<<<<<<<
 * 
 *         sock = transport.get_extra_info('socket')
 */
  __pyx_v_self->con_state = __pyx_e_8asynctnt_6iproto_8protocol_CONNECTION_CONNECTED;

  /* "asynctnt/iproto/coreproto.pyx":170
 *         self.con_state = CONNECTION_CONNECTED
 * 
 *         sock = transport.get_extra_info('socket')             # <<<<<<<<<<<<<<
 *         if sock is not None \
 *                 and (not hasattr(socket, 'AF_UNIX')
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_transport, __pyx_n_s_get_extra_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_u_socket};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_sock = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/coreproto.pyx":171
 * 
 *         sock = transport.get_extra_info('socket')
 *         if sock is not None \             # <<<<<<<<<<<<<<
 *                 and (not hasattr(socket, 'AF_UNIX')
 *                      or sock.family != socket.AF_UNIX):
 */
  __pyx_t_6 = (__pyx_v_sock != Py_None);
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }

  /* "asynctnt/iproto/coreproto.pyx":172
 *         sock = transport.get_extra_info('socket')
 *         if sock is not None \
 *                 and (not hasattr(socket, 'AF_UNIX')             # <<<<<<<<<<<<<<
 *                      or sock.family != socket.AF_UNIX):
 *             sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_socket); if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_u_AF_UNIX); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(28, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (!__pyx_t_6);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }

  /* "asynctnt/iproto/coreproto.pyx":173
 *         if sock is not None \
 *                 and (not hasattr(socket, 'AF_UNIX')
 *                      or sock.family != socket.AF_UNIX):             # <<<<<<<<<<<<<<
 *             sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_family); if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_AF_UNIX); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(28, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;

  /* "asynctnt/iproto/coreproto.pyx":171
 * 
 *         sock = transport.get_extra_info('socket')
 *         if sock is not None \             # <<<<<<<<<<<<<<
 *                 and (not hasattr(socket, 'AF_UNIX')
 *                      or sock.family != socket.AF_UNIX):
 */
  if (__pyx_t_5) {

    /* "asynctnt/iproto/coreproto.pyx":174
 *                 and (not hasattr(socket, 'AF_UNIX')
 *                      or sock.family != socket.AF_UNIX):
 *             sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)             # <<<<<<<<<<<<<<
 * 
 *         self.state = PROTOCOL_GREETING
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_setsockopt); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_socket); if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_IPPROTO_TCP); if (unlikely(!__pyx_t_8)) __PYX_ERR(28, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_socket); if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_TCP_NODELAY); if (unlikely(!__pyx_t_9)) __PYX_ERR(28, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_1, __pyx_t_8, __pyx_t_9, __pyx_int_1};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynctnt/iproto/coreproto.pyx":171
 * 
 *         sock = transport.get_extra_info('socket')
 *         if sock is not None \             # <<<<<<<<<<<<<<
 *                 and (not hasattr(socket, 'AF_UNIX')
 *                      or sock.family != socket.AF_UNIX):
 */
  }

  /* "asynctnt/iproto/coreproto.pyx":176
 *             sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
 * 
 *         self.state = PROTOCOL_GREETING             # <<<<<<<<<<<<<<
 *         self._on_connection_made()
 * 
 */
  __pyx_v_self->state = __pyx_e_8asynctnt_6iproto_8protocol_PROTOCOL_GREETING;

  /* "asynctnt/iproto/coreproto.pyx":177
 * 
 *         self.state = PROTOCOL_GREETING
 *         self._on_connection_made()             # <<<<<<<<<<<<<<
 * 
 *     def connection_lost(self, exc):
 */
  ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self->__pyx_vtab)->_on_connection_made(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 177, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":166
 *         self._on_data_received(data)
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         self.transport = transport
 *         self.con_state = CONNECTION_CONNECTED
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sock);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/coreproto.pyx":179
 *         self._on_connection_made()
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         self.con_state = CONNECTION_BAD
 *         self.version = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_15connection_lost(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_15connection_lost = {"connection_lost", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_15connection_lost, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_15connection_lost(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_exc = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connection_lost (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_exc,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 179, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "connection_lost") < 0)) __PYX_ERR(28, 179, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_exc = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("connection_lost", 1, 1, 1, __pyx_nargs); __PYX_ERR(28, 179, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.connection_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_14connection_lost(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self), __pyx_v_exc);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_14connection_lost(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connection_lost", 1);

  /* "asynctnt/iproto/coreproto.pyx":180
 * 
 *     def connection_lost(self, exc):
 *         self.con_state = CONNECTION_BAD             # <<<<<<<<<<<<<<
 *         self.version = None
 *         self.salt = None
 */
  __pyx_v_self->con_state = __pyx_e_8asynctnt_6iproto_8protocol_CONNECTION_BAD;

  /* "asynctnt/iproto/coreproto.pyx":181
 *     def connection_lost(self, exc):
 *         self.con_state = CONNECTION_BAD
 *         self.version = None             # <<<<<<<<<<<<<<
 *         self.salt = None
 *         self.rbuf = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->version);
  __Pyx_DECREF(__pyx_v_self->version);
  __pyx_v_self->version = ((PyObject*)Py_None);

  /* "asynctnt/iproto/coreproto.pyx":182
 *         self.con_state = CONNECTION_BAD
 *         self.version = None
 *         self.salt = None             # <<<<<<<<<<<<<<
 *         self.rbuf = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->salt);
  __Pyx_DECREF(__pyx_v_self->salt);
  __pyx_v_self->salt = ((PyObject*)Py_None);

  /* "asynctnt/iproto/coreproto.pyx":183
 *         self.version = None
 *         self.salt = None
 *         self.rbuf = None             # <<<<<<<<<<<<<<
 * 
 *         self._on_connection_lost(exc)
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->rbuf);
  __Pyx_DECREF((PyObject *)__pyx_v_self->rbuf);
  __pyx_v_self->rbuf = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)Py_None);

  /* "asynctnt/iproto/coreproto.pyx":185
 *         self.rbuf = None
 * 
 *         self._on_connection_lost(exc)             # <<<<<<<<<<<<<<
 */
  ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self->__pyx_vtab)->_on_connection_lost(__pyx_v_self, __pyx_v_exc); if (unlikely(PyErr_Occurred())) __PYX_ERR(28, 185, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":179
 *         self._on_connection_made()
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         self.con_state = CONNECTION_BAD
 *         self.version = None
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.connection_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_17__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_17__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_17__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_17__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_16__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_16__reduce_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.con_state, self.encoding, self.host, self.port, self.rbuf, self.salt, self.state, self.version)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_8protocol_ConnectionState(__pyx_v_self->con_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_8protocol_ProtocolState(__pyx_v_self->state); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->encoding);
  __Pyx_GIVEREF(__pyx_v_self->encoding);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->encoding)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->host);
  __Pyx_GIVEREF(__pyx_v_self->host);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->host)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->port);
  __Pyx_GIVEREF(__pyx_v_self->port);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->port)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->rbuf);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->rbuf);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 4, ((PyObject *)__pyx_v_self->rbuf))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->salt);
  __Pyx_GIVEREF(__pyx_v_self->salt);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_v_self->salt)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->version);
  __Pyx_GIVEREF(__pyx_v_self->version);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_v_self->version)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.con_state, self.encoding, self.host, self.port, self.rbuf, self.salt, self.state, self.version)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.con_state, self.encoding, self.host, self.port, self.rbuf, self.salt, self.state, self.version)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.encoding is not None or self.host is not None or self.port is not None or self.rbuf is not None or self.salt is not None or self.version is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.con_state, self.encoding, self.host, self.port, self.rbuf, self.salt, self.state, self.version)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.encoding is not None or self.host is not None or self.port is not None or self.rbuf is not None or self.salt is not None or self.version is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->encoding != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->host != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->port != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->rbuf) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->salt != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->version != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.encoding is not None or self.host is not None or self.port is not None or self.rbuf is not None or self.salt is not None or self.version is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.encoding is not None or self.host is not None or self.port is not None or self.rbuf is not None or self.salt is not None or self.version is not None
 *     if use_setstate:
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_CoreProtocol); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_144520718);
    __Pyx_GIVEREF(__pyx_int_144520718);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_144520718)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.encoding is not None or self.host is not None or self.port is not None or self.rbuf is not None or self.salt is not None or self.version is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, None), state
 *     else:
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CoreProtocol__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_CoreProtocol); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_144520718);
    __Pyx_GIVEREF(__pyx_int_144520718);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_144520718)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CoreProtocol__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_19__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_19__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_19__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_19__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_18__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12CoreProtocol_18__setstate_cython__(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CoreProtocol__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CoreProtocol__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CoreProtocol__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.CoreProtocol.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":63
 * 
 * cdef class BaseProtocol(CoreProtocol):
 *     def __init__(self, host, port,             # <<<<<<<<<<<<<<
 *                  username, password,
 *                  fetch_schema,
 */

/* Python wrapper */
static int __pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_host = 0;
  PyObject *__pyx_v_port = 0;
  PyObject *__pyx_v_username = 0;
  PyObject *__pyx_v_password = 0;
  PyObject *__pyx_v_fetch_schema = 0;
  PyObject *__pyx_v_auto_refetch_schema = 0;
  PyObject *__pyx_v_connected_fut = 0;
  PyObject *__pyx_v_on_connection_made = 0;
  PyObject *__pyx_v_on_connection_lost = 0;
  PyObject *__pyx_v_loop = 0;
  PyObject *__pyx_v_request_timeout = 0;
  PyObject *__pyx_v_encoding = 0;
  PyObject *__pyx_v_initial_read_buffer_size = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_host,&__pyx_n_s_port,&__pyx_n_s_username,&__pyx_n_s_password,&__pyx_n_s_fetch_schema,&__pyx_n_s_auto_refetch_schema,&__pyx_n_s_connected_fut,&__pyx_n_s_on_connection_made,&__pyx_n_s_on_connection_lost,&__pyx_n_s_loop,&__pyx_n_s_request_timeout,&__pyx_n_s_encoding,&__pyx_n_s_initial_read_buffer_size,0};

    /* "asynctnt/iproto/protocol.pyx":70
 *                  on_connection_made, on_connection_lost,
 *                  loop,
 *                  request_timeout=None,             # <<<<<<<<<<<<<<
 *                  encoding=None,
 *                  initial_read_buffer_size=None):
 */
    values[10] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));

    /* "asynctnt/iproto/protocol.pyx":71
 *                  loop,
 *                  request_timeout=None,
 *                  encoding=None,             # <<<<<<<<<<<<<<
 *                  initial_read_buffer_size=None):
 *         CoreProtocol.__init__(self, host, port, encoding,
 */
    values[11] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));

    /* "asynctnt/iproto/protocol.pyx":72
 *                  request_timeout=None,
 *                  encoding=None,
 *                  initial_read_buffer_size=None):             # <<<<<<<<<<<<<<
 *         CoreProtocol.__init__(self, host, port, encoding,
 *                               initial_read_buffer_size)
 */
    values[12] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case 13: values[12] = __Pyx_Arg_VARARGS(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = __Pyx_Arg_VARARGS(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = __Pyx_Arg_VARARGS(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = __Pyx_Arg_VARARGS(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = __Pyx_Arg_VARARGS(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = __Pyx_Arg_VARARGS(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_VARARGS(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_host)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_port)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 10, 13, 1); __PYX_ERR(6, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_username)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 10, 13, 2); __PYX_ERR(6, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_password)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 10, 13, 3); __PYX_ERR(6, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fetch_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 10, 13, 4); __PYX_ERR(6, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_auto_refetch_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[5]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 10, 13, 5); __PYX_ERR(6, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_connected_fut)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[6]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 10, 13, 6); __PYX_ERR(6, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_on_connection_made)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[7]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 10, 13, 7); __PYX_ERR(6, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_on_connection_lost)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[8]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 10, 13, 8); __PYX_ERR(6, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[9]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 10, 13, 9); __PYX_ERR(6, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_request_timeout);
          if (value) { values[10] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_encoding);
          if (value) { values[11] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_initial_read_buffer_size);
          if (value) { values[12] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 63, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(6, 63, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case 13: values[12] = __Pyx_Arg_VARARGS(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = __Pyx_Arg_VARARGS(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = __Pyx_Arg_VARARGS(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = __Pyx_Arg_VARARGS(__pyx_args, 9);
        values[8] = __Pyx_Arg_VARARGS(__pyx_args, 8);
        values[7] = __Pyx_Arg_VARARGS(__pyx_args, 7);
        values[6] = __Pyx_Arg_VARARGS(__pyx_args, 6);
        values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
        values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_host = values[0];
    __pyx_v_port = values[1];
    __pyx_v_username = values[2];
    __pyx_v_password = values[3];
    __pyx_v_fetch_schema = values[4];
    __pyx_v_auto_refetch_schema = values[5];
    __pyx_v_connected_fut = values[6];
    __pyx_v_on_connection_made = values[7];
    __pyx_v_on_connection_lost = values[8];
    __pyx_v_loop = values[9];
    __pyx_v_request_timeout = values[10];
    __pyx_v_encoding = values[11];
    __pyx_v_initial_read_buffer_size = values[12];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 10, 13, __pyx_nargs); __PYX_ERR(6, 63, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol___init__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self), __pyx_v_host, __pyx_v_port, __pyx_v_username, __pyx_v_password, __pyx_v_fetch_schema, __pyx_v_auto_refetch_schema, __pyx_v_connected_fut, __pyx_v_on_connection_made, __pyx_v_on_connection_lost, __pyx_v_loop, __pyx_v_request_timeout, __pyx_v_encoding, __pyx_v_initial_read_buffer_size);

  /* "asynctnt/iproto/protocol.pyx":63
 * 
 * cdef class BaseProtocol(CoreProtocol):
 *     def __init__(self, host, port,             # <<<<<<<<<<<<<<
 *                  username, password,
 *                  fetch_schema,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol___init__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_host, PyObject *__pyx_v_port, PyObject *__pyx_v_username, PyObject *__pyx_v_password, PyObject *__pyx_v_fetch_schema, PyObject *__pyx_v_auto_refetch_schema, PyObject *__pyx_v_connected_fut, PyObject *__pyx_v_on_connection_made, PyObject *__pyx_v_on_connection_lost, PyObject *__pyx_v_loop, PyObject *__pyx_v_request_timeout, PyObject *__pyx_v_encoding, PyObject *__pyx_v_initial_read_buffer_size) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  float __pyx_t_6;
  float __pyx_t_7;
  PyObject *(*__pyx_t_8)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, float);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "asynctnt/iproto/protocol.pyx":73
 *                  encoding=None,
 *                  initial_read_buffer_size=None):
 *         CoreProtocol.__init__(self, host, port, encoding,             # <<<<<<<<<<<<<<
 *                               initial_read_buffer_size)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "asynctnt/iproto/protocol.pyx":74
 *                  initial_read_buffer_size=None):
 *         CoreProtocol.__init__(self, host, port, encoding,
 *                               initial_read_buffer_size)             # <<<<<<<<<<<<<<
 * 
 *         self.loop = loop
 */
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[6] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_host, __pyx_v_port, __pyx_v_encoding, __pyx_v_initial_read_buffer_size};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 5+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":76
 *                               initial_read_buffer_size)
 * 
 *         self.loop = loop             # <<<<<<<<<<<<<<
 * 
 *         self.username = username
 */
  __Pyx_INCREF(__pyx_v_loop);
  __Pyx_GIVEREF(__pyx_v_loop);
  __Pyx_GOTREF(__pyx_v_self->loop);
  __Pyx_DECREF(__pyx_v_self->loop);
  __pyx_v_self->loop = __pyx_v_loop;

  /* "asynctnt/iproto/protocol.pyx":78
 *         self.loop = loop
 * 
 *         self.username = username             # <<<<<<<<<<<<<<
 *         self.password = password
 *         self.fetch_schema = fetch_schema
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_username))||((__pyx_v_username) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_username))) __PYX_ERR(6, 78, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_username;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->username);
  __Pyx_DECREF(__pyx_v_self->username);
  __pyx_v_self->username = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":79
 * 
 *         self.username = username
 *         self.password = password             # <<<<<<<<<<<<<<
 *         self.fetch_schema = fetch_schema
 *         self.auto_refetch_schema = auto_refetch_schema
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_password))||((__pyx_v_password) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_password))) __PYX_ERR(6, 79, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_password;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->password);
  __Pyx_DECREF(__pyx_v_self->password);
  __pyx_v_self->password = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":80
 *         self.username = username
 *         self.password = password
 *         self.fetch_schema = fetch_schema             # <<<<<<<<<<<<<<
 *         self.auto_refetch_schema = auto_refetch_schema
 *         self.request_timeout = request_timeout or 0
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_fetch_schema); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 80, __pyx_L1_error)
  __pyx_v_self->fetch_schema = __pyx_t_5;

  /* "asynctnt/iproto/protocol.pyx":81
 *         self.password = password
 *         self.fetch_schema = fetch_schema
 *         self.auto_refetch_schema = auto_refetch_schema             # <<<<<<<<<<<<<<
 *         self.request_timeout = request_timeout or 0
 *         self.post_con_state = POST_CONNECTION_NONE
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_auto_refetch_schema); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 81, __pyx_L1_error)
  __pyx_v_self->auto_refetch_schema = __pyx_t_5;

  /* "asynctnt/iproto/protocol.pyx":82
 *         self.fetch_schema = fetch_schema
 *         self.auto_refetch_schema = auto_refetch_schema
 *         self.request_timeout = request_timeout or 0             # <<<<<<<<<<<<<<
 *         self.post_con_state = POST_CONNECTION_NONE
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_request_timeout); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 82, __pyx_L1_error)
  if (!__pyx_t_5) {
  } else {
    __pyx_t_7 = __pyx_PyFloat_AsFloat(__pyx_v_request_timeout); if (unlikely((__pyx_t_7 == (float)-1) && PyErr_Occurred())) __PYX_ERR(6, 82, __pyx_L1_error)
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_6 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_v_self->request_timeout = __pyx_t_6;

  /* "asynctnt/iproto/protocol.pyx":83
 *         self.auto_refetch_schema = auto_refetch_schema
 *         self.request_timeout = request_timeout or 0
 *         self.post_con_state = POST_CONNECTION_NONE             # <<<<<<<<<<<<<<
 * 
 *         self.connected_fut = connected_fut
 */
  __pyx_v_self->post_con_state = __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_NONE;

  /* "asynctnt/iproto/protocol.pyx":85
 *         self.post_con_state = POST_CONNECTION_NONE
 * 
 *         self.connected_fut = connected_fut             # <<<<<<<<<<<<<<
 *         self.on_connection_made_cb = on_connection_made
 *         self.on_connection_lost_cb = on_connection_lost
 */
  __Pyx_INCREF(__pyx_v_connected_fut);
  __Pyx_GIVEREF(__pyx_v_connected_fut);
  __Pyx_GOTREF(__pyx_v_self->connected_fut);
  __Pyx_DECREF(__pyx_v_self->connected_fut);
  __pyx_v_self->connected_fut = __pyx_v_connected_fut;

  /* "asynctnt/iproto/protocol.pyx":86
 * 
 *         self.connected_fut = connected_fut
 *         self.on_connection_made_cb = on_connection_made             # <<<<<<<<<<<<<<
 *         self.on_connection_lost_cb = on_connection_lost
 *         self._closing = False
 */
  __Pyx_INCREF(__pyx_v_on_connection_made);
  __Pyx_GIVEREF(__pyx_v_on_connection_made);
  __Pyx_GOTREF(__pyx_v_self->on_connection_made_cb);
  __Pyx_DECREF(__pyx_v_self->on_connection_made_cb);
  __pyx_v_self->on_connection_made_cb = __pyx_v_on_connection_made;

  /* "asynctnt/iproto/protocol.pyx":87
 *         self.connected_fut = connected_fut
 *         self.on_connection_made_cb = on_connection_made
 *         self.on_connection_lost_cb = on_connection_lost             # <<<<<<<<<<<<<<
 *         self._closing = False
 * 
 */
  __Pyx_INCREF(__pyx_v_on_connection_lost);
  __Pyx_GIVEREF(__pyx_v_on_connection_lost);
  __Pyx_GOTREF(__pyx_v_self->on_connection_lost_cb);
  __Pyx_DECREF(__pyx_v_self->on_connection_lost_cb);
  __pyx_v_self->on_connection_lost_cb = __pyx_v_on_connection_lost;

  /* "asynctnt/iproto/protocol.pyx":88
 *         self.on_connection_made_cb = on_connection_made
 *         self.on_connection_lost_cb = on_connection_lost
 *         self._closing = False             # <<<<<<<<<<<<<<
 * 
 *         self._on_request_completed_cb = self._on_request_completed
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closing, Py_False) < 0) __PYX_ERR(6, 88, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":90
 *         self._closing = False
 * 
 *         self._on_request_completed_cb = self._on_request_completed             # <<<<<<<<<<<<<<
 *         self._on_request_timeout_cb = self._on_request_timeout
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_on_request_completed); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_on_request_completed_cb);
  __Pyx_DECREF(__pyx_v_self->_on_request_completed_cb);
  __pyx_v_self->_on_request_completed_cb = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":91
 * 
 *         self._on_request_completed_cb = self._on_request_completed
 *         self._on_request_timeout_cb = self._on_request_timeout             # <<<<<<<<<<<<<<
 * 
 *         self._reqs = {}
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_on_request_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_on_request_timeout_cb);
  __Pyx_DECREF(__pyx_v_self->_on_request_timeout_cb);
  __pyx_v_self->_on_request_timeout_cb = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":93
 *         self._on_request_timeout_cb = self._on_request_timeout
 * 
 *         self._reqs = {}             # <<<<<<<<<<<<<<
 *         self._sync = 0
 *         self._last_stream_id = 0
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_reqs);
  __Pyx_DECREF(__pyx_v_self->_reqs);
  __pyx_v_self->_reqs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":94
 * 
 *         self._reqs = {}
 *         self._sync = 0             # <<<<<<<<<<<<<<
 *         self._last_stream_id = 0
 *         self._schema_id = -1
 */
  __pyx_v_self->_sync = 0;

  /* "asynctnt/iproto/protocol.pyx":95
 *         self._reqs = {}
 *         self._sync = 0
 *         self._last_stream_id = 0             # <<<<<<<<<<<<<<
 *         self._schema_id = -1
 *         self._schema = Schema.__new__(Schema, self._schema_id)
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_last_stream_id, __pyx_int_0) < 0) __PYX_ERR(6, 95, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":96
 *         self._sync = 0
 *         self._last_stream_id = 0
 *         self._schema_id = -1             # <<<<<<<<<<<<<<
 *         self._schema = Schema.__new__(Schema, self._schema_id)
 *         self._schema_fetch_in_progress = False
 */
  __pyx_v_self->_schema_id = -1L;

  /* "asynctnt/iproto/protocol.pyx":97
 *         self._last_stream_id = 0
 *         self._schema_id = -1
 *         self._schema = Schema.__new__(Schema, self._schema_id)             # <<<<<<<<<<<<<<
 *         self._schema_fetch_in_progress = False
 *         self._refetch_schema_future = None
 */
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->_schema_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(6, 97, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Schema(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Schema), __pyx_t_2, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 97, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF((PyObject *)__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_schema);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_schema);
  __pyx_v_self->_schema = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":98
 *         self._schema_id = -1
 *         self._schema = Schema.__new__(Schema, self._schema_id)
 *         self._schema_fetch_in_progress = False             # <<<<<<<<<<<<<<
 *         self._refetch_schema_future = None
 *         self._db = self._create_db(<bint> False)
 */
  __pyx_v_self->_schema_fetch_in_progress = 0;

  /* "asynctnt/iproto/protocol.pyx":99
 *         self._schema = Schema.__new__(Schema, self._schema_id)
 *         self._schema_fetch_in_progress = False
 *         self._refetch_schema_future = None             # <<<<<<<<<<<<<<
 *         self._db = self._create_db(<bint> False)
 *         self.execute = self._execute_bad
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_refetch_schema_future);
  __Pyx_DECREF(__pyx_v_self->_refetch_schema_future);
  __pyx_v_self->_refetch_schema_future = Py_None;

  /* "asynctnt/iproto/protocol.pyx":100
 *         self._schema_fetch_in_progress = False
 *         self._refetch_schema_future = None
 *         self._db = self._create_db(<bint> False)             # <<<<<<<<<<<<<<
 *         self.execute = self._execute_bad
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_create_db(__pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_db);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_db);
  __pyx_v_self->_db = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":101
 *         self._refetch_schema_future = None
 *         self._db = self._create_db(<bint> False)
 *         self.execute = self._execute_bad             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_execute_bad;
  __pyx_v_self->execute = __pyx_t_8;

  /* "asynctnt/iproto/protocol.pyx":103
 *         self.execute = self._execute_bad
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self.create_future = self.loop.create_future
 *         except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "asynctnt/iproto/protocol.pyx":104
 * 
 *         try:
 *             self.create_future = self.loop.create_future             # <<<<<<<<<<<<<<
 *         except AttributeError:
 *             self.create_future = self._create_future_fallback
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->loop, __pyx_n_s_create_future); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 104, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_v_self->create_future);
      __Pyx_DECREF(__pyx_v_self->create_future);
      __pyx_v_self->create_future = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "asynctnt/iproto/protocol.pyx":103
 *         self.execute = self._execute_bad
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self.create_future = self.loop.create_future
 *         except AttributeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L10_try_end;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "asynctnt/iproto/protocol.pyx":105
 *         try:
 *             self.create_future = self.loop.create_future
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             self.create_future = self._create_future_fallback
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_4) {
      __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(6, 105, __pyx_L7_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);

      /* "asynctnt/iproto/protocol.pyx":106
 *             self.create_future = self.loop.create_future
 *         except AttributeError:
 *             self.create_future = self._create_future_fallback             # <<<<<<<<<<<<<<
 * 
 *     def _create_future_fallback(self):  # pragma: nocover
 */
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_create_future_fallback); if (unlikely(!__pyx_t_12)) __PYX_ERR(6, 106, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_GIVEREF(__pyx_t_12);
      __Pyx_GOTREF(__pyx_v_self->create_future);
      __Pyx_DECREF(__pyx_v_self->create_future);
      __pyx_v_self->create_future = __pyx_t_12;
      __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_exception_handled;
    }
    goto __pyx_L7_except_error;

    /* "asynctnt/iproto/protocol.pyx":103
 *         self.execute = self._execute_bad
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self.create_future = self.loop.create_future
 *         except AttributeError:
 */
    __pyx_L7_except_error:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L6_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    __pyx_L10_try_end:;
  }

  /* "asynctnt/iproto/protocol.pyx":63
 * 
 * cdef class BaseProtocol(CoreProtocol):
 *     def __init__(self, host, port,             # <<<<<<<<<<<<<<
 *                  username, password,
 *                  fetch_schema,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":108
 *             self.create_future = self._create_future_fallback
 * 
 *     def _create_future_fallback(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return asyncio.Future(loop=self.loop)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_3_create_future_fallback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_3_create_future_fallback = {"_create_future_fallback", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_3_create_future_fallback, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_3_create_future_fallback(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_create_future_fallback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("_create_future_fallback", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "_create_future_fallback", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_2_create_future_fallback(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_2_create_future_fallback(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create_future_fallback", 1);

  /* "asynctnt/iproto/protocol.pyx":109
 * 
 *     def _create_future_fallback(self):  # pragma: nocover
 *         return asyncio.Future(loop=self.loop)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Future); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_loop, __pyx_v_self->loop) < 0) __PYX_ERR(6, 109, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":108
 *             self.create_future = self._create_future_fallback
 * 
 *     def _create_future_fallback(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return asyncio.Future(loop=self.loop)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._create_future_fallback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":111
 *         return asyncio.Future(loop=self.loop)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def schema(self):
 *         return self._schema
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_6schema_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_6schema_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_6schema___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_6schema___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "asynctnt/iproto/protocol.pyx":113
 *     @property
 *     def schema(self):
 *         return self._schema             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_schema);
  __pyx_r = ((PyObject *)__pyx_v_self->_schema);
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":111
 *         return asyncio.Future(loop=self.loop)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def schema(self):
 *         return self._schema
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":115
 *         return self._schema
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def schema_id(self):
 *         return self._schema_id
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_9schema_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_9schema_id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_9schema_id___get__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_9schema_id___get__(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "asynctnt/iproto/protocol.pyx":117
 *     @property
 *     def schema_id(self):
 *         return self._schema_id             # <<<<<<<<<<<<<<
 * 
 *     cdef void _set_connection_ready(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->_schema_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":115
 *         return self._schema
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def schema_id(self):
 *         return self._schema_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.schema_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":119
 *         return self._schema_id
 * 
 *     cdef void _set_connection_ready(self):             # <<<<<<<<<<<<<<
 *         if not self.connected_fut.done():
 *             self.connected_fut.set_result(True)
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__set_connection_ready(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_connection_ready", 1);

  /* "asynctnt/iproto/protocol.pyx":120
 * 
 *     cdef void _set_connection_ready(self):
 *         if not self.connected_fut.done():             # <<<<<<<<<<<<<<
 *             self.connected_fut.set_result(True)
 *             self.con_state = CONNECTION_FULL
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->connected_fut, __pyx_n_s_done); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "asynctnt/iproto/protocol.pyx":121
 *     cdef void _set_connection_ready(self):
 *         if not self.connected_fut.done():
 *             self.connected_fut.set_result(True)             # <<<<<<<<<<<<<<
 *             self.con_state = CONNECTION_FULL
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->connected_fut, __pyx_n_s_set_result); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, Py_True};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":122
 *         if not self.connected_fut.done():
 *             self.connected_fut.set_result(True)
 *             self.con_state = CONNECTION_FULL             # <<<<<<<<<<<<<<
 * 
 *     cdef void _set_connection_error(self, e):
 */
    __pyx_v_self->__pyx_base.con_state = __pyx_e_8asynctnt_6iproto_8protocol_CONNECTION_FULL;

    /* "asynctnt/iproto/protocol.pyx":120
 * 
 *     cdef void _set_connection_ready(self):
 *         if not self.connected_fut.done():             # <<<<<<<<<<<<<<
 *             self.connected_fut.set_result(True)
 *             self.con_state = CONNECTION_FULL
 */
  }

  /* "asynctnt/iproto/protocol.pyx":119
 *         return self._schema_id
 * 
 *     cdef void _set_connection_ready(self):             # <<<<<<<<<<<<<<
 *         if not self.connected_fut.done():
 *             self.connected_fut.set_result(True)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._set_connection_ready", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/protocol.pyx":124
 *             self.con_state = CONNECTION_FULL
 * 
 *     cdef void _set_connection_error(self, e):             # <<<<<<<<<<<<<<
 *         if not self.connected_fut.done():
 *             self.connected_fut.set_exception(e)
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__set_connection_error(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_e) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *(*__pyx_t_7)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, float);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_connection_error", 1);

  /* "asynctnt/iproto/protocol.pyx":125
 * 
 *     cdef void _set_connection_error(self, e):
 *         if not self.connected_fut.done():             # <<<<<<<<<<<<<<
 *             self.connected_fut.set_exception(e)
 *             self.con_state = CONNECTION_BAD
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->connected_fut, __pyx_n_s_done); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "asynctnt/iproto/protocol.pyx":126
 *     cdef void _set_connection_error(self, e):
 *         if not self.connected_fut.done():
 *             self.connected_fut.set_exception(e)             # <<<<<<<<<<<<<<
 *             self.con_state = CONNECTION_BAD
 *             self.post_con_state = POST_CONNECTION_NONE
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->connected_fut, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_e};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 126, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":127
 *         if not self.connected_fut.done():
 *             self.connected_fut.set_exception(e)
 *             self.con_state = CONNECTION_BAD             # <<<<<<<<<<<<<<
 *             self.post_con_state = POST_CONNECTION_NONE
 *             self.execute = self._execute_bad
 */
    __pyx_v_self->__pyx_base.con_state = __pyx_e_8asynctnt_6iproto_8protocol_CONNECTION_BAD;

    /* "asynctnt/iproto/protocol.pyx":128
 *             self.connected_fut.set_exception(e)
 *             self.con_state = CONNECTION_BAD
 *             self.post_con_state = POST_CONNECTION_NONE             # <<<<<<<<<<<<<<
 *             self.execute = self._execute_bad
 * 
 */
    __pyx_v_self->post_con_state = __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_NONE;

    /* "asynctnt/iproto/protocol.pyx":129
 *             self.con_state = CONNECTION_BAD
 *             self.post_con_state = POST_CONNECTION_NONE
 *             self.execute = self._execute_bad             # <<<<<<<<<<<<<<
 * 
 *     cdef void _on_greeting_received(self):
 */
    __pyx_t_7 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_execute_bad;
    __pyx_v_self->execute = __pyx_t_7;

    /* "asynctnt/iproto/protocol.pyx":125
 * 
 *     cdef void _set_connection_error(self, e):
 *         if not self.connected_fut.done():             # <<<<<<<<<<<<<<
 *             self.connected_fut.set_exception(e)
 *             self.con_state = CONNECTION_BAD
 */
  }

  /* "asynctnt/iproto/protocol.pyx":124
 *             self.con_state = CONNECTION_FULL
 * 
 *     cdef void _set_connection_error(self, e):             # <<<<<<<<<<<<<<
 *         if not self.connected_fut.done():
 *             self.connected_fut.set_exception(e)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._set_connection_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/protocol.pyx":131
 *             self.execute = self._execute_bad
 * 
 *     cdef void _on_greeting_received(self):             # <<<<<<<<<<<<<<
 *         self.post_con_state = POST_CONNECTION_ID
 *         self.execute = self._execute_normal
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_greeting_received(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  PyObject *(*__pyx_t_1)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, float);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "asynctnt/iproto/protocol.pyx":132
 * 
 *     cdef void _on_greeting_received(self):
 *         self.post_con_state = POST_CONNECTION_ID             # <<<<<<<<<<<<<<
 *         self.execute = self._execute_normal
 *         self._post_con_state_machine()
 */
  __pyx_v_self->post_con_state = __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_ID;

  /* "asynctnt/iproto/protocol.pyx":133
 *     cdef void _on_greeting_received(self):
 *         self.post_con_state = POST_CONNECTION_ID
 *         self.execute = self._execute_normal             # <<<<<<<<<<<<<<
 *         self._post_con_state_machine()
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_execute_normal;
  __pyx_v_self->execute = __pyx_t_1;

  /* "asynctnt/iproto/protocol.pyx":134
 *         self.post_con_state = POST_CONNECTION_ID
 *         self.execute = self._execute_normal
 *         self._post_con_state_machine()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _post_con_state_machine(self):
 */
  ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_post_con_state_machine(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 134, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":131
 *             self.execute = self._execute_bad
 * 
 *     cdef void _on_greeting_received(self):             # <<<<<<<<<<<<<<
 *         self.post_con_state = POST_CONNECTION_ID
 *         self.execute = self._execute_normal
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._on_greeting_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "asynctnt/iproto/protocol.pyx":136
 *         self._post_con_state_machine()
 * 
 *     cdef void _post_con_state_machine(self):             # <<<<<<<<<<<<<<
 *         if self.post_con_state == POST_CONNECTION_ID:
 *             assert self.version is not None
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__post_con_state_machine(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_post_con_state_machine", 1);

  /* "asynctnt/iproto/protocol.pyx":137
 * 
 *     cdef void _post_con_state_machine(self):
 *         if self.post_con_state == POST_CONNECTION_ID:             # <<<<<<<<<<<<<<
 *             assert self.version is not None
 *             if self.version >= (2, 10, 0):
 */
  __pyx_t_1 = (__pyx_v_self->post_con_state == __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_ID);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/protocol.pyx":138
 *     cdef void _post_con_state_machine(self):
 *         if self.post_con_state == POST_CONNECTION_ID:
 *             assert self.version is not None             # <<<<<<<<<<<<<<
 *             if self.version >= (2, 10, 0):
 *                 # send <id> request
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_1 = (__pyx_v_self->__pyx_base.version != ((PyObject*)Py_None));
      if (unlikely(!__pyx_t_1)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
        __PYX_ERR(6, 138, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(6, 138, __pyx_L1_error)
    #endif

    /* "asynctnt/iproto/protocol.pyx":139
 *         if self.post_con_state == POST_CONNECTION_ID:
 *             assert self.version is not None
 *             if self.version >= (2, 10, 0):             # <<<<<<<<<<<<<<
 *                 # send <id> request
 *                 self._do_id()
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_self->__pyx_base.version, __pyx_tuple__40, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 139, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(6, 139, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_1) {

      /* "asynctnt/iproto/protocol.pyx":141
 *             if self.version >= (2, 10, 0):
 *                 # send <id> request
 *                 self._do_id()             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
      ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_do_id(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 141, __pyx_L1_error)

      /* "asynctnt/iproto/protocol.pyx":142
 *                 # send <id> request
 *                 self._do_id()
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 # Tarantool does not support id call - call auth request
 */
      goto __pyx_L0;

      /* "asynctnt/iproto/protocol.pyx":139
 *         if self.post_con_state == POST_CONNECTION_ID:
 *             assert self.version is not None
 *             if self.version >= (2, 10, 0):             # <<<<<<<<<<<<<<
 *                 # send <id> request
 *                 self._do_id()
 */
    }

    /* "asynctnt/iproto/protocol.pyx":145
 *             else:
 *                 # Tarantool does not support id call - call auth request
 *                 self.post_con_state = POST_CONNECTION_AUTH             # <<<<<<<<<<<<<<
 * 
 *         if self.post_con_state == POST_CONNECTION_AUTH:
 */
    /*else*/ {
      __pyx_v_self->post_con_state = __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_AUTH;
    }

    /* "asynctnt/iproto/protocol.pyx":137
 * 
 *     cdef void _post_con_state_machine(self):
 *         if self.post_con_state == POST_CONNECTION_ID:             # <<<<<<<<<<<<<<
 *             assert self.version is not None
 *             if self.version >= (2, 10, 0):
 */
  }

  /* "asynctnt/iproto/protocol.pyx":147
 *                 self.post_con_state = POST_CONNECTION_AUTH
 * 
 *         if self.post_con_state == POST_CONNECTION_AUTH:             # <<<<<<<<<<<<<<
 *             if self.username and self.password:
 *                 # send <auth> request
 */
  __pyx_t_1 = (__pyx_v_self->post_con_state == __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_AUTH);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/protocol.pyx":148
 * 
 *         if self.post_con_state == POST_CONNECTION_AUTH:
 *             if self.username and self.password:             # <<<<<<<<<<<<<<
 *                 # send <auth> request
 *                 self._do_auth(self.username, self.password)
 */
    __pyx_t_3 = (__pyx_v_self->username != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_v_self->username) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_3 = (__pyx_v_self->password != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_v_self->password) != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_1) {

      /* "asynctnt/iproto/protocol.pyx":150
 *             if self.username and self.password:
 *                 # send <auth> request
 *                 self._do_auth(self.username, self.password)             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
      __pyx_t_2 = __pyx_v_self->username;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_4 = __pyx_v_self->password;
      __Pyx_INCREF(__pyx_t_4);
      ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_do_auth(__pyx_v_self, ((PyObject*)__pyx_t_2), ((PyObject*)__pyx_t_4)); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 150, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "asynctnt/iproto/protocol.pyx":151
 *                 # send <auth> request
 *                 self._do_auth(self.username, self.password)
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 # no need to auth - fetch schema
 */
      goto __pyx_L0;

      /* "asynctnt/iproto/protocol.pyx":148
 * 
 *         if self.post_con_state == POST_CONNECTION_AUTH:
 *             if self.username and self.password:             # <<<<<<<<<<<<<<
 *                 # send <auth> request
 *                 self._do_auth(self.username, self.password)
 */
    }

    /* "asynctnt/iproto/protocol.pyx":154
 *             else:
 *                 # no need to auth - fetch schema
 *                 self.post_con_state = POST_CONNECTION_SCHEMA             # <<<<<<<<<<<<<<
 * 
 *         if self.post_con_state == POST_CONNECTION_SCHEMA:
 */
    /*else*/ {
      __pyx_v_self->post_con_state = __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_SCHEMA;
    }

    /* "asynctnt/iproto/protocol.pyx":147
 *                 self.post_con_state = POST_CONNECTION_AUTH
 * 
 *         if self.post_con_state == POST_CONNECTION_AUTH:             # <<<<<<<<<<<<<<
 *             if self.username and self.password:
 *                 # send <auth> request
 */
  }

  /* "asynctnt/iproto/protocol.pyx":156
 *                 self.post_con_state = POST_CONNECTION_SCHEMA
 * 
 *         if self.post_con_state == POST_CONNECTION_SCHEMA:             # <<<<<<<<<<<<<<
 *             if self.fetch_schema:
 *                 self._refetch_schema()
 */
  __pyx_t_1 = (__pyx_v_self->post_con_state == __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_SCHEMA);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/protocol.pyx":157
 * 
 *         if self.post_con_state == POST_CONNECTION_SCHEMA:
 *             if self.fetch_schema:             # <<<<<<<<<<<<<<
 *                 self._refetch_schema()
 *                 return
 */
    if (__pyx_v_self->fetch_schema) {

      /* "asynctnt/iproto/protocol.pyx":158
 *         if self.post_con_state == POST_CONNECTION_SCHEMA:
 *             if self.fetch_schema:
 *                 self._refetch_schema()             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_refetch_schema(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "asynctnt/iproto/protocol.pyx":159
 *             if self.fetch_schema:
 *                 self._refetch_schema()
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 # nothing else to do - connection is done
 */
      goto __pyx_L0;

      /* "asynctnt/iproto/protocol.pyx":157
 * 
 *         if self.post_con_state == POST_CONNECTION_SCHEMA:
 *             if self.fetch_schema:             # <<<<<<<<<<<<<<
 *                 self._refetch_schema()
 *                 return
 */
    }

    /* "asynctnt/iproto/protocol.pyx":162
 *             else:
 *                 # nothing else to do - connection is done
 *                 self.post_con_state = POST_CONNECTION_DONE             # <<<<<<<<<<<<<<
 * 
 *         if self.post_con_state == POST_CONNECTION_DONE:
 */
    /*else*/ {
      __pyx_v_self->post_con_state = __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_DONE;
    }

    /* "asynctnt/iproto/protocol.pyx":156
 *                 self.post_con_state = POST_CONNECTION_SCHEMA
 * 
 *         if self.post_con_state == POST_CONNECTION_SCHEMA:             # <<<<<<<<<<<<<<
 *             if self.fetch_schema:
 *                 self._refetch_schema()
 */
  }

  /* "asynctnt/iproto/protocol.pyx":164
 *                 self.post_con_state = POST_CONNECTION_DONE
 * 
 *         if self.post_con_state == POST_CONNECTION_DONE:             # <<<<<<<<<<<<<<
 *             self._set_connection_ready()
 *             return
 */
  __pyx_t_1 = (__pyx_v_self->post_con_state == __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_DONE);
  if (__pyx_t_1) {

    /* "asynctnt/iproto/protocol.pyx":165
 * 
 *         if self.post_con_state == POST_CONNECTION_DONE:
 *             self._set_connection_ready()             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_set_connection_ready(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 165, __pyx_L1_error)

    /* "asynctnt/iproto/protocol.pyx":166
 *         if self.post_con_state == POST_CONNECTION_DONE:
 *             self._set_connection_ready()
 *             return             # <<<<<<<<<<<<<<
 * 
 *     cdef void _on_response_received(self, const char *buf, uint32_t buf_len):
 */
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":164
 *                 self.post_con_state = POST_CONNECTION_DONE
 * 
 *         if self.post_con_state == POST_CONNECTION_DONE:             # <<<<<<<<<<<<<<
 *             self._set_connection_ready()
 *             return
 */
  }

  /* "asynctnt/iproto/protocol.pyx":136
 *         self._post_con_state_machine()
 * 
 *     cdef void _post_con_state_machine(self):             # <<<<<<<<<<<<<<
 *         if self.post_con_state == POST_CONNECTION_ID:
 *             assert self.version is not None
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._post_con_state_machine", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/protocol.pyx":168
 *             return
 * 
 *     cdef void _on_response_received(self, const char *buf, uint32_t buf_len):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *req_p
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_response_received(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, char const *__pyx_v_buf, uint32_t __pyx_v_buf_len) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_response = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req = 0;
  struct __pyx_t_8asynctnt_6iproto_8protocol_Header __pyx_v_hdr;
  int __pyx_v_is_chunk;
  PyObject *__pyx_v_waiter = 0;
  PyObject *__pyx_v_sync_obj = 0;
  PyObject *__pyx_v_err = 0;
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_v_response_p;
  PyObject *__pyx_v_e = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  uint64_t __pyx_t_8;
  int64_t __pyx_t_9;
  int32_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_response_received", 1);

  /* "asynctnt/iproto/protocol.pyx":181
 *             ssize_t length
 * 
 *         length = response_parse_header(buf, buf_len, &hdr)             # <<<<<<<<<<<<<<
 *         buf_len -= length
 *         buf = &buf[length]  # skip header
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_response_parse_header(__pyx_v_buf, __pyx_v_buf_len, (&__pyx_v_hdr)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1L))) __PYX_ERR(6, 181, __pyx_L1_error)
  __pyx_v_length = __pyx_t_1;

  /* "asynctnt/iproto/protocol.pyx":182
 * 
 *         length = response_parse_header(buf, buf_len, &hdr)
 *         buf_len -= length             # <<<<<<<<<<<<<<
 *         buf = &buf[length]  # skip header
 * 
 */
  __pyx_v_buf_len = (__pyx_v_buf_len - __pyx_v_length);

  /* "asynctnt/iproto/protocol.pyx":183
 *         length = response_parse_header(buf, buf_len, &hdr)
 *         buf_len -= length
 *         buf = &buf[length]  # skip header             # <<<<<<<<<<<<<<
 * 
 *         sync_obj = <object> hdr.sync
 */
  __pyx_v_buf = (&(__pyx_v_buf[__pyx_v_length]));

  /* "asynctnt/iproto/protocol.pyx":185
 *         buf = &buf[length]  # skip header
 * 
 *         sync_obj = <object> hdr.sync             # <<<<<<<<<<<<<<
 * 
 *         response_p = cpython.dict.PyDict_GetItem(self._reqs, sync_obj)
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint64_t(__pyx_v_hdr.sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sync_obj = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/protocol.pyx":187
 *         sync_obj = <object> hdr.sync
 * 
 *         response_p = cpython.dict.PyDict_GetItem(self._reqs, sync_obj)             # <<<<<<<<<<<<<<
 *         if response_p is NULL:
 *             logger.warning('sync %d not found', hdr.sync)
 */
  __pyx_t_3 = __pyx_v_self->_reqs;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_response_p = PyDict_GetItem(__pyx_t_3, __pyx_v_sync_obj);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/protocol.pyx":188
 * 
 *         response_p = cpython.dict.PyDict_GetItem(self._reqs, sync_obj)
 *         if response_p is NULL:             # <<<<<<<<<<<<<<
 *             logger.warning('sync %d not found', hdr.sync)
 *             return
 */
  __pyx_t_4 = (__pyx_v_response_p == NULL);
  if (__pyx_t_4) {

    /* "asynctnt/iproto/protocol.pyx":189
 *         response_p = cpython.dict.PyDict_GetItem(self._reqs, sync_obj)
 *         if response_p is NULL:
 *             logger.warning('sync %d not found', hdr.sync)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warning); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_From_uint64_t(__pyx_v_hdr.sync); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_kp_u_sync_d_not_found, __pyx_t_2};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "asynctnt/iproto/protocol.pyx":190
 *         if response_p is NULL:
 *             logger.warning('sync %d not found', hdr.sync)
 *             return             # <<<<<<<<<<<<<<
 * 
 *         is_chunk = (hdr.code == tarantool.IPROTO_CHUNK)
 */
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":188
 * 
 *         response_p = cpython.dict.PyDict_GetItem(self._reqs, sync_obj)
 *         if response_p is NULL:             # <<<<<<<<<<<<<<
 *             logger.warning('sync %d not found', hdr.sync)
 *             return
 */
  }

  /* "asynctnt/iproto/protocol.pyx":192
 *             return
 * 
 *         is_chunk = (hdr.code == tarantool.IPROTO_CHUNK)             # <<<<<<<<<<<<<<
 * 
 *         response = <Response> response_p
 */
  __pyx_v_is_chunk = (__pyx_v_hdr.code == __pyx_e_8asynctnt_6iproto_9tarantool_IPROTO_CHUNK);

  /* "asynctnt/iproto/protocol.pyx":194
 *         is_chunk = (hdr.code == tarantool.IPROTO_CHUNK)
 * 
 *         response = <Response> response_p             # <<<<<<<<<<<<<<
 *         req = response.request_
 *         response.sync_ = hdr.sync
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_response_p);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_response = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/protocol.pyx":195
 * 
 *         response = <Response> response_p
 *         req = response.request_             # <<<<<<<<<<<<<<
 *         response.sync_ = hdr.sync
 *         response.schema_id_ = hdr.schema_id
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_response->request_);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/protocol.pyx":196
 *         response = <Response> response_p
 *         req = response.request_
 *         response.sync_ = hdr.sync             # <<<<<<<<<<<<<<
 *         response.schema_id_ = hdr.schema_id
 *         if not is_chunk:
 */
  __pyx_t_8 = __pyx_v_hdr.sync;
  __pyx_v_response->sync_ = __pyx_t_8;

  /* "asynctnt/iproto/protocol.pyx":197
 *         req = response.request_
 *         response.sync_ = hdr.sync
 *         response.schema_id_ = hdr.schema_id             # <<<<<<<<<<<<<<
 *         if not is_chunk:
 *             response.code_ = hdr.code
 */
  __pyx_t_9 = __pyx_v_hdr.schema_id;
  __pyx_v_response->schema_id_ = __pyx_t_9;

  /* "asynctnt/iproto/protocol.pyx":198
 *         response.sync_ = hdr.sync
 *         response.schema_id_ = hdr.schema_id
 *         if not is_chunk:             # <<<<<<<<<<<<<<
 *             response.code_ = hdr.code
 *             response.return_code_ = hdr.return_code
 */
  __pyx_t_4 = (!__pyx_v_is_chunk);
  if (__pyx_t_4) {

    /* "asynctnt/iproto/protocol.pyx":199
 *         response.schema_id_ = hdr.schema_id
 *         if not is_chunk:
 *             response.code_ = hdr.code             # <<<<<<<<<<<<<<
 *             response.return_code_ = hdr.return_code
 *             cpython.dict.PyDict_DelItem(self._reqs, sync_obj)
 */
    __pyx_t_10 = __pyx_v_hdr.code;
    __pyx_v_response->code_ = __pyx_t_10;

    /* "asynctnt/iproto/protocol.pyx":200
 *         if not is_chunk:
 *             response.code_ = hdr.code
 *             response.return_code_ = hdr.return_code             # <<<<<<<<<<<<<<
 *             cpython.dict.PyDict_DelItem(self._reqs, sync_obj)
 *         else:
 */
    __pyx_t_10 = __pyx_v_hdr.return_code;
    __pyx_v_response->return_code_ = __pyx_t_10;

    /* "asynctnt/iproto/protocol.pyx":201
 *             response.code_ = hdr.code
 *             response.return_code_ = hdr.return_code
 *             cpython.dict.PyDict_DelItem(self._reqs, sync_obj)             # <<<<<<<<<<<<<<
 *         else:
 *             if not req.push_subscribe:
 */
    __pyx_t_3 = __pyx_v_self->_reqs;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_7 = PyDict_DelItem(__pyx_t_3, __pyx_v_sync_obj); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(6, 201, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "asynctnt/iproto/protocol.pyx":198
 *         response.sync_ = hdr.sync
 *         response.schema_id_ = hdr.schema_id
 *         if not is_chunk:             # <<<<<<<<<<<<<<
 *             response.code_ = hdr.code
 *             response.return_code_ = hdr.return_code
 */
    goto __pyx_L4;
  }

  /* "asynctnt/iproto/protocol.pyx":203
 *             cpython.dict.PyDict_DelItem(self._reqs, sync_obj)
 *         else:
 *             if not req.push_subscribe:             # <<<<<<<<<<<<<<
 *                 # skip request data as no one will be waiting for it
 *                 mp_next(&buf)
 */
  /*else*/ {
    __pyx_t_4 = (!__pyx_v_req->push_subscribe);
    if (__pyx_t_4) {

      /* "asynctnt/iproto/protocol.pyx":205
 *             if not req.push_subscribe:
 *                 # skip request data as no one will be waiting for it
 *                 mp_next(&buf)             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
      mp_next((&__pyx_v_buf));

      /* "asynctnt/iproto/protocol.pyx":206
 *                 # skip request data as no one will be waiting for it
 *                 mp_next(&buf)
 *                 return             # <<<<<<<<<<<<<<
 * 
 *         err = None
 */
      goto __pyx_L0;

      /* "asynctnt/iproto/protocol.pyx":203
 *             cpython.dict.PyDict_DelItem(self._reqs, sync_obj)
 *         else:
 *             if not req.push_subscribe:             # <<<<<<<<<<<<<<
 *                 # skip request data as no one will be waiting for it
 *                 mp_next(&buf)
 */
    }
  }
  __pyx_L4:;

  /* "asynctnt/iproto/protocol.pyx":208
 *                 return
 * 
 *         err = None             # <<<<<<<<<<<<<<
 *         if buf != &buf[buf_len]:
 *             # has body
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_err = Py_None;

  /* "asynctnt/iproto/protocol.pyx":209
 * 
 *         err = None
 *         if buf != &buf[buf_len]:             # <<<<<<<<<<<<<<
 *             # has body
 *             try:
 */
  __pyx_t_4 = (__pyx_v_buf != (&(__pyx_v_buf[__pyx_v_buf_len])));
  if (__pyx_t_4) {

    /* "asynctnt/iproto/protocol.pyx":211
 *         if buf != &buf[buf_len]:
 *             # has body
 *             try:             # <<<<<<<<<<<<<<
 *                 response_parse_body(buf, buf_len, response, req, is_chunk)
 *             except Exception as e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      /*try:*/ {

        /* "asynctnt/iproto/protocol.pyx":212
 *             # has body
 *             try:
 *                 response_parse_body(buf, buf_len, response, req, is_chunk)             # <<<<<<<<<<<<<<
 *             except Exception as e:
 *                 err = e
 */
        __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_response_parse_body(__pyx_v_buf, __pyx_v_buf_len, __pyx_v_response, __pyx_v_req, __pyx_v_is_chunk); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1L))) __PYX_ERR(6, 212, __pyx_L7_error)

        /* "asynctnt/iproto/protocol.pyx":211
 *         if buf != &buf[buf_len]:
 *             # has body
 *             try:             # <<<<<<<<<<<<<<
 *                 response_parse_body(buf, buf_len, response, req, is_chunk)
 *             except Exception as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "asynctnt/iproto/protocol.pyx":213
 *             try:
 *                 response_parse_body(buf, buf_len, response, req, is_chunk)
 *             except Exception as e:             # <<<<<<<<<<<<<<
 *                 err = e
 * 
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_7) {
        __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._on_response_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_5, &__pyx_t_2) < 0) __PYX_ERR(6, 213, __pyx_L9_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __pyx_v_e = __pyx_t_5;
        /*try:*/ {

          /* "asynctnt/iproto/protocol.pyx":214
 *                 response_parse_body(buf, buf_len, response, req, is_chunk)
 *             except Exception as e:
 *                 err = e             # <<<<<<<<<<<<<<
 * 
 *         # refetch schema if it is changed
 */
          __Pyx_INCREF(__pyx_v_e);
          __Pyx_DECREF_SET(__pyx_v_err, __pyx_v_e);
        }

        /* "asynctnt/iproto/protocol.pyx":213
 *             try:
 *                 response_parse_body(buf, buf_len, response, req, is_chunk)
 *             except Exception as e:             # <<<<<<<<<<<<<<
 *                 err = e
 * 
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            goto __pyx_L19;
          }
          __pyx_L19:;
        }
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L8_exception_handled;
      }
      goto __pyx_L9_except_error;

      /* "asynctnt/iproto/protocol.pyx":211
 *         if buf != &buf[buf_len]:
 *             # has body
 *             try:             # <<<<<<<<<<<<<<
 *                 response_parse_body(buf, buf_len, response, req, is_chunk)
 *             except Exception as e:
 */
      __pyx_L9_except_error:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      goto __pyx_L1_error;
      __pyx_L8_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      __pyx_L12_try_end:;
    }

    /* "asynctnt/iproto/protocol.pyx":209
 * 
 *         err = None
 *         if buf != &buf[buf_len]:             # <<<<<<<<<<<<<<
 *             # has body
 *             try:
 */
  }

  /* "asynctnt/iproto/protocol.pyx":217
 * 
 *         # refetch schema if it is changed
 *         if self.con_state == CONNECTION_FULL \             # <<<<<<<<<<<<<<
 *                 and req.check_schema_change \
 *                 and self.auto_refetch_schema \
 */
  __pyx_t_14 = (__pyx_v_self->__pyx_base.con_state == __pyx_e_8asynctnt_6iproto_8protocol_CONNECTION_FULL);
  if (__pyx_t_14) {
  } else {
    __pyx_t_4 = __pyx_t_14;
    goto __pyx_L21_bool_binop_done;
  }

  /* "asynctnt/iproto/protocol.pyx":218
 *         # refetch schema if it is changed
 *         if self.con_state == CONNECTION_FULL \
 *                 and req.check_schema_change \             # <<<<<<<<<<<<<<
 *                 and self.auto_refetch_schema \
 *                 and response.schema_id_ > 0 \
 */
  if (__pyx_v_req->check_schema_change) {
  } else {
    __pyx_t_4 = __pyx_v_req->check_schema_change;
    goto __pyx_L21_bool_binop_done;
  }

  /* "asynctnt/iproto/protocol.pyx":219
 *         if self.con_state == CONNECTION_FULL \
 *                 and req.check_schema_change \
 *                 and self.auto_refetch_schema \             # <<<<<<<<<<<<<<
 *                 and response.schema_id_ > 0 \
 *                 and response.schema_id_ != self._schema_id:
 */
  if (__pyx_v_self->auto_refetch_schema) {
  } else {
    __pyx_t_4 = __pyx_v_self->auto_refetch_schema;
    goto __pyx_L21_bool_binop_done;
  }

  /* "asynctnt/iproto/protocol.pyx":220
 *                 and req.check_schema_change \
 *                 and self.auto_refetch_schema \
 *                 and response.schema_id_ > 0 \             # <<<<<<<<<<<<<<
 *                 and response.schema_id_ != self._schema_id:
 *             self._refetch_schema()
 */
  __pyx_t_14 = (__pyx_v_response->schema_id_ > 0);
  if (__pyx_t_14) {
  } else {
    __pyx_t_4 = __pyx_t_14;
    goto __pyx_L21_bool_binop_done;
  }

  /* "asynctnt/iproto/protocol.pyx":221
 *                 and self.auto_refetch_schema \
 *                 and response.schema_id_ > 0 \
 *                 and response.schema_id_ != self._schema_id:             # <<<<<<<<<<<<<<
 *             self._refetch_schema()
 * 
 */
  __pyx_t_14 = (__pyx_v_response->schema_id_ != __pyx_v_self->_schema_id);
  __pyx_t_4 = __pyx_t_14;
  __pyx_L21_bool_binop_done:;

  /* "asynctnt/iproto/protocol.pyx":217
 * 
 *         # refetch schema if it is changed
 *         if self.con_state == CONNECTION_FULL \             # <<<<<<<<<<<<<<
 *                 and req.check_schema_change \
 *                 and self.auto_refetch_schema \
 */
  if (__pyx_t_4) {

    /* "asynctnt/iproto/protocol.pyx":222
 *                 and response.schema_id_ > 0 \
 *                 and response.schema_id_ != self._schema_id:
 *             self._refetch_schema()             # <<<<<<<<<<<<<<
 * 
 *         # returning result
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_refetch_schema(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynctnt/iproto/protocol.pyx":217
 * 
 *         # refetch schema if it is changed
 *         if self.con_state == CONNECTION_FULL \             # <<<<<<<<<<<<<<
 *                 and req.check_schema_change \
 *                 and self.auto_refetch_schema \
 */
  }

  /* "asynctnt/iproto/protocol.pyx":225
 * 
 *         # returning result
 *         if is_chunk:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  if (__pyx_v_is_chunk) {

    /* "asynctnt/iproto/protocol.pyx":226
 *         # returning result
 *         if is_chunk:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         waiter = req.waiter
 */
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":225
 * 
 *         # returning result
 *         if is_chunk:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "asynctnt/iproto/protocol.pyx":228
 *             return
 * 
 *         waiter = req.waiter             # <<<<<<<<<<<<<<
 *         if waiter is None or waiter.done():
 *             return
 */
  __pyx_t_2 = __pyx_v_req->waiter;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_waiter = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":229
 * 
 *         waiter = req.waiter
 *         if waiter is None or waiter.done():             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_14 = (__pyx_v_waiter == Py_None);
  if (!__pyx_t_14) {
  } else {
    __pyx_t_4 = __pyx_t_14;
    goto __pyx_L28_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_done); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_14 < 0))) __PYX_ERR(6, 229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_t_14;
  __pyx_L28_bool_binop_done:;
  if (__pyx_t_4) {

    /* "asynctnt/iproto/protocol.pyx":230
 *         waiter = req.waiter
 *         if waiter is None or waiter.done():
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if err is not None:
 */
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":229
 * 
 *         waiter = req.waiter
 *         if waiter is None or waiter.done():             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "asynctnt/iproto/protocol.pyx":232
 *             return
 * 
 *         if err is not None:             # <<<<<<<<<<<<<<
 *             response.set_exception(err)
 *             waiter.set_exception(err)
 */
  __pyx_t_4 = (__pyx_v_err != Py_None);
  if (__pyx_t_4) {

    /* "asynctnt/iproto/protocol.pyx":233
 * 
 *         if err is not None:
 *             response.set_exception(err)             # <<<<<<<<<<<<<<
 *             waiter.set_exception(err)
 *             return
 */
    __pyx_f_8asynctnt_6iproto_8protocol_8Response_set_exception(__pyx_v_response, __pyx_v_err); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 233, __pyx_L1_error)

    /* "asynctnt/iproto/protocol.pyx":234
 *         if err is not None:
 *             response.set_exception(err)
 *             waiter.set_exception(err)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_err};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 234, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynctnt/iproto/protocol.pyx":235
 *             response.set_exception(err)
 *             waiter.set_exception(err)
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if response.is_error():
 */
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":232
 *             return
 * 
 *         if err is not None:             # <<<<<<<<<<<<<<
 *             response.set_exception(err)
 *             waiter.set_exception(err)
 */
  }

  /* "asynctnt/iproto/protocol.pyx":237
 *             return
 * 
 *         if response.is_error():             # <<<<<<<<<<<<<<
 *             err = TarantoolDatabaseError(response.return_code_,
 *                                          response.errmsg,
 */
  __pyx_t_4 = __pyx_f_8asynctnt_6iproto_8protocol_8Response_is_error(__pyx_v_response); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 237, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "asynctnt/iproto/protocol.pyx":238
 * 
 *         if response.is_error():
 *             err = TarantoolDatabaseError(response.return_code_,             # <<<<<<<<<<<<<<
 *                                          response.errmsg,
 *                                          response.error)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_TarantoolDatabaseError); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyInt_From_int32_t(__pyx_v_response->return_code_); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "asynctnt/iproto/protocol.pyx":240
 *             err = TarantoolDatabaseError(response.return_code_,
 *                                          response.errmsg,
 *                                          response.error)             # <<<<<<<<<<<<<<
 *             response.set_exception(err)
 *             waiter.set_exception(err)
 */
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_6, __pyx_t_3, __pyx_v_response->errmsg, ((PyObject *)__pyx_v_response->error)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 3+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 238, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_err, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynctnt/iproto/protocol.pyx":241
 *                                          response.errmsg,
 *                                          response.error)
 *             response.set_exception(err)             # <<<<<<<<<<<<<<
 *             waiter.set_exception(err)
 *             return
 */
    __pyx_f_8asynctnt_6iproto_8protocol_8Response_set_exception(__pyx_v_response, __pyx_v_err); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 241, __pyx_L1_error)

    /* "asynctnt/iproto/protocol.pyx":242
 *                                          response.error)
 *             response.set_exception(err)
 *             waiter.set_exception(err)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_err};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynctnt/iproto/protocol.pyx":243
 *             response.set_exception(err)
 *             waiter.set_exception(err)
 *             return             # <<<<<<<<<<<<<<
 * 
 *         waiter.set_result(response)
 */
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":237
 *             return
 * 
 *         if response.is_error():             # <<<<<<<<<<<<<<
 *             err = TarantoolDatabaseError(response.return_code_,
 *                                          response.errmsg,
 */
  }

  /* "asynctnt/iproto/protocol.pyx":245
 *             return
 * 
 *         waiter.set_result(response)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _do_id(self):
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_set_result); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_response)};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":168
 *             return
 * 
 *     cdef void _on_response_received(self, const char *buf, uint32_t buf_len):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *req_p
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._on_response_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_response);
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XDECREF(__pyx_v_waiter);
  __Pyx_XDECREF(__pyx_v_sync_obj);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/protocol.pyx":250
 *         fut = self._db._id(0.0)
 * 
 *         def on_id(f):             # <<<<<<<<<<<<<<
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_6_do_id_1on_id(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_6_do_id_1on_id = {"on_id", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_6_do_id_1on_id, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_6_do_id_1on_id(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_f = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("on_id (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_f,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_f)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 250, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "on_id") < 0)) __PYX_ERR(6, 250, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_f = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("on_id", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 250, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._do_id.on_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_6_do_id_on_id(__pyx_self, __pyx_v_f);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_6_do_id_on_id(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *__pyx_cur_scope;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *__pyx_outer_scope;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("on_id", 1);
  __pyx_outer_scope = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "asynctnt/iproto/protocol.pyx":251
 * 
 *         def on_id(f):
 *             if f.cancelled():             # <<<<<<<<<<<<<<
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "asynctnt/iproto/protocol.pyx":252
 *         def on_id(f):
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())             # <<<<<<<<<<<<<<
 *                 return
 *             e = f.exception()
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 252, __pyx_L1_error) }
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_futures); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CancelledError); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 252, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->_set_connection_error(__pyx_cur_scope->__pyx_v_self, __pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 252, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":253
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return             # <<<<<<<<<<<<<<
 *             e = f.exception()
 *             if not e:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":251
 * 
 *         def on_id(f):
 *             if f.cancelled():             # <<<<<<<<<<<<<<
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return
 */
  }

  /* "asynctnt/iproto/protocol.pyx":254
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return
 *             e = f.exception()             # <<<<<<<<<<<<<<
 *             if not e:
 *                 logger.debug('Tarantool[%s:%s] identified successfully',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_e = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":255
 *                 return
 *             e = f.exception()
 *             if not e:             # <<<<<<<<<<<<<<
 *                 logger.debug('Tarantool[%s:%s] identified successfully',
 *                              self.host, self.port)
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_e); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 255, __pyx_L1_error)
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "asynctnt/iproto/protocol.pyx":256
 *             e = f.exception()
 *             if not e:
 *                 logger.debug('Tarantool[%s:%s] identified successfully',             # <<<<<<<<<<<<<<
 *                              self.host, self.port)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynctnt/iproto/protocol.pyx":257
 *             if not e:
 *                 logger.debug('Tarantool[%s:%s] identified successfully',
 *                              self.host, self.port)             # <<<<<<<<<<<<<<
 * 
 *                 self.post_con_state = POST_CONNECTION_AUTH
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 257, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 257, __pyx_L1_error) }
    __pyx_t_2 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_2, __pyx_kp_u_Tarantool_s_s_identified_success, __pyx_cur_scope->__pyx_v_self->__pyx_base.host, __pyx_cur_scope->__pyx_v_self->__pyx_base.port};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":259
 *                              self.host, self.port)
 * 
 *                 self.post_con_state = POST_CONNECTION_AUTH             # <<<<<<<<<<<<<<
 *                 self._post_con_state_machine()
 *             else:
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 259, __pyx_L1_error) }
    __pyx_cur_scope->__pyx_v_self->post_con_state = __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_AUTH;

    /* "asynctnt/iproto/protocol.pyx":260
 * 
 *                 self.post_con_state = POST_CONNECTION_AUTH
 *                 self._post_con_state_machine()             # <<<<<<<<<<<<<<
 *             else:
 *                 logger.error('Tarantool[%s:%s] identification failed: %s',
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 260, __pyx_L1_error) }
    ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->_post_con_state_machine(__pyx_cur_scope->__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 260, __pyx_L1_error)

    /* "asynctnt/iproto/protocol.pyx":255
 *                 return
 *             e = f.exception()
 *             if not e:             # <<<<<<<<<<<<<<
 *                 logger.debug('Tarantool[%s:%s] identified successfully',
 *                              self.host, self.port)
 */
    goto __pyx_L4;
  }

  /* "asynctnt/iproto/protocol.pyx":262
 *                 self._post_con_state_machine()
 *             else:
 *                 logger.error('Tarantool[%s:%s] identification failed: %s',             # <<<<<<<<<<<<<<
 *                              self.host, self.port, str(e))
 *                 self._set_connection_error(e)
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_error); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "asynctnt/iproto/protocol.pyx":263
 *             else:
 *                 logger.error('Tarantool[%s:%s] identification failed: %s',
 *                              self.host, self.port, str(e))             # <<<<<<<<<<<<<<
 *                 self._set_connection_error(e)
 * 
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 263, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 263, __pyx_L1_error) }
    __pyx_t_3 = __Pyx_PyObject_Str(__pyx_v_e); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_kp_u_Tarantool_s_s_identification_fai, __pyx_cur_scope->__pyx_v_self->__pyx_base.host, __pyx_cur_scope->__pyx_v_self->__pyx_base.port, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 4+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 262, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":264
 *                 logger.error('Tarantool[%s:%s] identification failed: %s',
 *                              self.host, self.port, str(e))
 *                 self._set_connection_error(e)             # <<<<<<<<<<<<<<
 * 
 *         fut.add_done_callback(on_id)
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 264, __pyx_L1_error) }
    ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->_set_connection_error(__pyx_cur_scope->__pyx_v_self, __pyx_v_e); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 264, __pyx_L1_error)
  }
  __pyx_L4:;

  /* "asynctnt/iproto/protocol.pyx":250
 *         fut = self._db._id(0.0)
 * 
 *         def on_id(f):             # <<<<<<<<<<<<<<
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._do_id.on_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":247
 *         waiter.set_result(response)
 * 
 *     cdef void _do_id(self):             # <<<<<<<<<<<<<<
 *         fut = self._db._id(0.0)
 * 
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__do_id(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *__pyx_cur_scope;
  PyObject *__pyx_v_fut = NULL;
  PyObject *__pyx_v_on_id = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_id", 0);
  __pyx_cur_scope = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(6, 247, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "asynctnt/iproto/protocol.pyx":248
 * 
 *     cdef void _do_id(self):
 *         fut = self._db._id(0.0)             # <<<<<<<<<<<<<<
 * 
 *         def on_id(f):
 */
  __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__id(__pyx_cur_scope->__pyx_v_self->_db, 0.0); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fut = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":250
 *         fut = self._db._id(0.0)
 * 
 *         def on_id(f):             # <<<<<<<<<<<<<<
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_6_do_id_1on_id, 0, __pyx_n_s_BaseProtocol__do_id_locals_on_id, ((PyObject*)__pyx_cur_scope), __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_on_id = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":266
 *                 self._set_connection_error(e)
 * 
 *         fut.add_done_callback(on_id)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _do_auth(self, str username, str password):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_add_done_callback); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_on_id};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":247
 *         waiter.set_result(response)
 * 
 *     cdef void _do_id(self):             # <<<<<<<<<<<<<<
 *         fut = self._db._id(0.0)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._do_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XDECREF(__pyx_v_on_id);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/protocol.pyx":272
 *         fut = self._db._auth(self.salt, username, password, 0)
 * 
 *         def on_authorized(f):             # <<<<<<<<<<<<<<
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_8_do_auth_1on_authorized(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_8_do_auth_1on_authorized = {"on_authorized", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_8_do_auth_1on_authorized, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_8_do_auth_1on_authorized(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_f = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("on_authorized (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_f,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_f)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 272, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "on_authorized") < 0)) __PYX_ERR(6, 272, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_f = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("on_authorized", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 272, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._do_auth.on_authorized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_8_do_auth_on_authorized(__pyx_self, __pyx_v_f);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_8_do_auth_on_authorized(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *__pyx_cur_scope;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *__pyx_outer_scope;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("on_authorized", 1);
  __pyx_outer_scope = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "asynctnt/iproto/protocol.pyx":273
 * 
 *         def on_authorized(f):
 *             if f.cancelled():             # <<<<<<<<<<<<<<
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "asynctnt/iproto/protocol.pyx":274
 *         def on_authorized(f):
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())             # <<<<<<<<<<<<<<
 *                 return
 *             e = f.exception()
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 274, __pyx_L1_error) }
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_futures); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CancelledError); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->_set_connection_error(__pyx_cur_scope->__pyx_v_self, __pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 274, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":275
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return             # <<<<<<<<<<<<<<
 *             e = f.exception()
 *             if not e:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":273
 * 
 *         def on_authorized(f):
 *             if f.cancelled():             # <<<<<<<<<<<<<<
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return
 */
  }

  /* "asynctnt/iproto/protocol.pyx":276
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return
 *             e = f.exception()             # <<<<<<<<<<<<<<
 *             if not e:
 *                 logger.debug('Tarantool[%s:%s] Authorized successfully',
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_e = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":277
 *                 return
 *             e = f.exception()
 *             if not e:             # <<<<<<<<<<<<<<
 *                 logger.debug('Tarantool[%s:%s] Authorized successfully',
 *                              self.host, self.port)
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_e); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 277, __pyx_L1_error)
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "asynctnt/iproto/protocol.pyx":278
 *             e = f.exception()
 *             if not e:
 *                 logger.debug('Tarantool[%s:%s] Authorized successfully',             # <<<<<<<<<<<<<<
 *                              self.host, self.port)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynctnt/iproto/protocol.pyx":279
 *             if not e:
 *                 logger.debug('Tarantool[%s:%s] Authorized successfully',
 *                              self.host, self.port)             # <<<<<<<<<<<<<<
 * 
 *                 self.post_con_state = POST_CONNECTION_SCHEMA
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 279, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 279, __pyx_L1_error) }
    __pyx_t_2 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_2, __pyx_kp_u_Tarantool_s_s_Authorized_success, __pyx_cur_scope->__pyx_v_self->__pyx_base.host, __pyx_cur_scope->__pyx_v_self->__pyx_base.port};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":281
 *                              self.host, self.port)
 * 
 *                 self.post_con_state = POST_CONNECTION_SCHEMA             # <<<<<<<<<<<<<<
 *                 self._post_con_state_machine()
 *             else:
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 281, __pyx_L1_error) }
    __pyx_cur_scope->__pyx_v_self->post_con_state = __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_SCHEMA;

    /* "asynctnt/iproto/protocol.pyx":282
 * 
 *                 self.post_con_state = POST_CONNECTION_SCHEMA
 *                 self._post_con_state_machine()             # <<<<<<<<<<<<<<
 *             else:
 *                 logger.error('Tarantool[%s:%s] Authorization failed: %s',
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 282, __pyx_L1_error) }
    ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->_post_con_state_machine(__pyx_cur_scope->__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 282, __pyx_L1_error)

    /* "asynctnt/iproto/protocol.pyx":277
 *                 return
 *             e = f.exception()
 *             if not e:             # <<<<<<<<<<<<<<
 *                 logger.debug('Tarantool[%s:%s] Authorized successfully',
 *                              self.host, self.port)
 */
    goto __pyx_L4;
  }

  /* "asynctnt/iproto/protocol.pyx":284
 *                 self._post_con_state_machine()
 *             else:
 *                 logger.error('Tarantool[%s:%s] Authorization failed: %s',             # <<<<<<<<<<<<<<
 *                              self.host, self.port, str(e))
 *                 self._set_connection_error(e)
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_error); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "asynctnt/iproto/protocol.pyx":285
 *             else:
 *                 logger.error('Tarantool[%s:%s] Authorization failed: %s',
 *                              self.host, self.port, str(e))             # <<<<<<<<<<<<<<
 *                 self._set_connection_error(e)
 * 
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 285, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 285, __pyx_L1_error) }
    __pyx_t_3 = __Pyx_PyObject_Str(__pyx_v_e); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_kp_u_Tarantool_s_s_Authorization_fail, __pyx_cur_scope->__pyx_v_self->__pyx_base.host, __pyx_cur_scope->__pyx_v_self->__pyx_base.port, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 4+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":286
 *                 logger.error('Tarantool[%s:%s] Authorization failed: %s',
 *                              self.host, self.port, str(e))
 *                 self._set_connection_error(e)             # <<<<<<<<<<<<<<
 * 
 *         fut.add_done_callback(on_authorized)
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 286, __pyx_L1_error) }
    ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->_set_connection_error(__pyx_cur_scope->__pyx_v_self, __pyx_v_e); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 286, __pyx_L1_error)
  }
  __pyx_L4:;

  /* "asynctnt/iproto/protocol.pyx":272
 *         fut = self._db._auth(self.salt, username, password, 0)
 * 
 *         def on_authorized(f):             # <<<<<<<<<<<<<<
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._do_auth.on_authorized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":268
 *         fut.add_done_callback(on_id)
 * 
 *     cdef void _do_auth(self, str username, str password):             # <<<<<<<<<<<<<<
 *         # No extra error handling from Db.execute
 *         fut = self._db._auth(self.salt, username, password, 0)
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__do_auth(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_username, PyObject *__pyx_v_password) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *__pyx_cur_scope;
  PyObject *__pyx_v_fut = NULL;
  PyObject *__pyx_v_on_authorized = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_auth", 0);
  __pyx_cur_scope = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(6, 268, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "asynctnt/iproto/protocol.pyx":270
 *     cdef void _do_auth(self, str username, str password):
 *         # No extra error handling from Db.execute
 *         fut = self._db._auth(self.salt, username, password, 0)             # <<<<<<<<<<<<<<
 * 
 *         def on_authorized(f):
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->__pyx_base.salt;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_8asynctnt_6iproto_8protocol_2Db__auth(__pyx_cur_scope->__pyx_v_self->_db, ((PyObject*)__pyx_t_1), __pyx_v_username, __pyx_v_password, 0.0); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_fut = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":272
 *         fut = self._db._auth(self.salt, username, password, 0)
 * 
 *         def on_authorized(f):             # <<<<<<<<<<<<<<
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_8_do_auth_1on_authorized, 0, __pyx_n_s_BaseProtocol__do_auth_locals_on, ((PyObject*)__pyx_cur_scope), __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_on_authorized = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":288
 *                 self._set_connection_error(e)
 * 
 *         fut.add_done_callback(on_authorized)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _do_fetch_schema(self, object fut):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_add_done_callback); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_on_authorized};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":268
 *         fut.add_done_callback(on_id)
 * 
 *     cdef void _do_auth(self, str username, str password):             # <<<<<<<<<<<<<<
 *         # No extra error handling from Db.execute
 *         fut = self._db._auth(self.salt, username, password, 0)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._do_auth", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XDECREF(__pyx_v_on_authorized);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/protocol.pyx":296
 *         self._schema_fetch_in_progress = True
 * 
 *         def on_fetch(f):             # <<<<<<<<<<<<<<
 *             self._schema_fetch_in_progress = False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_16_do_fetch_schema_1on_fetch(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_16_do_fetch_schema_1on_fetch = {"on_fetch", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_16_do_fetch_schema_1on_fetch, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_16_do_fetch_schema_1on_fetch(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_f = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("on_fetch (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_f,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_f)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 296, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "on_fetch") < 0)) __PYX_ERR(6, 296, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_f = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("on_fetch", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 296, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._do_fetch_schema.on_fetch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_16_do_fetch_schema_on_fetch(__pyx_self, __pyx_v_f);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_16_do_fetch_schema_on_fetch(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *__pyx_cur_scope;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *__pyx_outer_scope;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_spaces = NULL;
  PyObject *__pyx_v_indexes = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int64_t __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  char const *__pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("on_fetch", 1);
  __pyx_outer_scope = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "asynctnt/iproto/protocol.pyx":297
 * 
 *         def on_fetch(f):
 *             self._schema_fetch_in_progress = False             # <<<<<<<<<<<<<<
 * 
 *             if f.cancelled():
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 297, __pyx_L1_error) }
  __pyx_cur_scope->__pyx_v_self->_schema_fetch_in_progress = 0;

  /* "asynctnt/iproto/protocol.pyx":299
 *             self._schema_fetch_in_progress = False
 * 
 *             if f.cancelled():             # <<<<<<<<<<<<<<
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 299, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "asynctnt/iproto/protocol.pyx":300
 * 
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())             # <<<<<<<<<<<<<<
 *                 return
 *             e = f.exception()
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 300, __pyx_L1_error) }
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_futures); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CancelledError); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->_set_connection_error(__pyx_cur_scope->__pyx_v_self, __pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 300, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":301
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return             # <<<<<<<<<<<<<<
 *             e = f.exception()
 *             if not e:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":299
 *             self._schema_fetch_in_progress = False
 * 
 *             if f.cancelled():             # <<<<<<<<<<<<<<
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return
 */
  }

  /* "asynctnt/iproto/protocol.pyx":302
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 *                 return
 *             e = f.exception()             # <<<<<<<<<<<<<<
 *             if not e:
 *                 if fut is not None and fut.cancelled():
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_e = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":303
 *                 return
 *             e = f.exception()
 *             if not e:             # <<<<<<<<<<<<<<
 *                 if fut is not None and fut.cancelled():
 *                     # if the caller has cancelled waiting
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_e); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 303, __pyx_L1_error)
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "asynctnt/iproto/protocol.pyx":304
 *             e = f.exception()
 *             if not e:
 *                 if fut is not None and fut.cancelled():             # <<<<<<<<<<<<<<
 *                     # if the caller has cancelled waiting
 *                     return
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(6, 304, __pyx_L1_error) }
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_fut != Py_None);
    if (__pyx_t_5) {
    } else {
      __pyx_t_6 = __pyx_t_5;
      goto __pyx_L6_bool_binop_done;
    }
    if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(6, 304, __pyx_L1_error) }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 304, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __pyx_t_5;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_6) {

      /* "asynctnt/iproto/protocol.pyx":306
 *                 if fut is not None and fut.cancelled():
 *                     # if the caller has cancelled waiting
 *                     return             # <<<<<<<<<<<<<<
 * 
 *                 spaces, indexes = f.result()
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "asynctnt/iproto/protocol.pyx":304
 *             e = f.exception()
 *             if not e:
 *                 if fut is not None and fut.cancelled():             # <<<<<<<<<<<<<<
 *                     # if the caller has cancelled waiting
 *                     return
 */
    }

    /* "asynctnt/iproto/protocol.pyx":308
 *                     return
 * 
 *                 spaces, indexes = f.result()             # <<<<<<<<<<<<<<
 *                 logger.debug('Tarantool[%s:%s] Schema fetch succeeded. '
 *                              'Version: %d, Spaces: %d, Indexes: %d.',
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_result); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(6, 308, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
      index = 0; __pyx_t_2 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_2)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_3 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_3)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(6, 308, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L9_unpacking_done;
      __pyx_L8_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(6, 308, __pyx_L1_error)
      __pyx_L9_unpacking_done:;
    }
    __pyx_v_spaces = __pyx_t_2;
    __pyx_t_2 = 0;
    __pyx_v_indexes = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "asynctnt/iproto/protocol.pyx":309
 * 
 *                 spaces, indexes = f.result()
 *                 logger.debug('Tarantool[%s:%s] Schema fetch succeeded. '             # <<<<<<<<<<<<<<
 *                              'Version: %d, Spaces: %d, Indexes: %d.',
 *                              self.host, self.port,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "asynctnt/iproto/protocol.pyx":311
 *                 logger.debug('Tarantool[%s:%s] Schema fetch succeeded. '
 *                              'Version: %d, Spaces: %d, Indexes: %d.',
 *                              self.host, self.port,             # <<<<<<<<<<<<<<
 *                              spaces.schema_id, len(spaces), len(indexes))
 *                 try:
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 311, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 311, __pyx_L1_error) }

    /* "asynctnt/iproto/protocol.pyx":312
 *                              'Version: %d, Spaces: %d, Indexes: %d.',
 *                              self.host, self.port,
 *                              spaces.schema_id, len(spaces), len(indexes))             # <<<<<<<<<<<<<<
 *                 try:
 *                     self._schema = Schema.parse(spaces.schema_id,
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_spaces, __pyx_n_s_schema_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = PyObject_Length(__pyx_v_spaces); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(6, 312, __pyx_L1_error)
    __pyx_t_7 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyObject_Length(__pyx_v_indexes); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(6, 312, __pyx_L1_error)
    __pyx_t_10 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[7] = {__pyx_t_11, __pyx_kp_u_Tarantool_s_s_Schema_fetch_succe, __pyx_cur_scope->__pyx_v_self->__pyx_base.host, __pyx_cur_scope->__pyx_v_self->__pyx_base.port, __pyx_t_3, __pyx_t_7, __pyx_t_10};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 6+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":313
 *                              self.host, self.port,
 *                              spaces.schema_id, len(spaces), len(indexes))
 *                 try:             # <<<<<<<<<<<<<<
 *                     self._schema = Schema.parse(spaces.schema_id,
 *                                                 spaces, indexes)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      /*try:*/ {

        /* "asynctnt/iproto/protocol.pyx":314
 *                              spaces.schema_id, len(spaces), len(indexes))
 *                 try:
 *                     self._schema = Schema.parse(spaces.schema_id,             # <<<<<<<<<<<<<<
 *                                                 spaces, indexes)
 *                 except Exception as e:
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_spaces, __pyx_n_s_schema_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 314, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_15 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_15 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(6, 314, __pyx_L10_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynctnt/iproto/protocol.pyx":315
 *                 try:
 *                     self._schema = Schema.parse(spaces.schema_id,
 *                                                 spaces, indexes)             # <<<<<<<<<<<<<<
 *                 except Exception as e:
 *                     logger.exception(e)
 */
        __pyx_t_1 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse(__pyx_t_15, __pyx_v_spaces, __pyx_v_indexes)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 314, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "asynctnt/iproto/protocol.pyx":314
 *                              spaces.schema_id, len(spaces), len(indexes))
 *                 try:
 *                     self._schema = Schema.parse(spaces.schema_id,             # <<<<<<<<<<<<<<
 *                                                 spaces, indexes)
 *                 except Exception as e:
 */
        if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 314, __pyx_L10_error) }
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_self->_schema);
        __Pyx_DECREF((PyObject *)__pyx_cur_scope->__pyx_v_self->_schema);
        __pyx_cur_scope->__pyx_v_self->_schema = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)__pyx_t_1);
        __pyx_t_1 = 0;

        /* "asynctnt/iproto/protocol.pyx":313
 *                              self.host, self.port,
 *                              spaces.schema_id, len(spaces), len(indexes))
 *                 try:             # <<<<<<<<<<<<<<
 *                     self._schema = Schema.parse(spaces.schema_id,
 *                                                 spaces, indexes)
 */
      }
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      goto __pyx_L15_try_end;
      __pyx_L10_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "asynctnt/iproto/protocol.pyx":316
 *                     self._schema = Schema.parse(spaces.schema_id,
 *                                                 spaces, indexes)
 *                 except Exception as e:             # <<<<<<<<<<<<<<
 *                     logger.exception(e)
 *                     logger.error('Error happened while parsing schema. '
 */
      __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_4) {
        __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._do_fetch_schema.on_fetch", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_10) < 0) __PYX_ERR(6, 316, __pyx_L12_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_e, __pyx_t_2);
        /*try:*/ {

          /* "asynctnt/iproto/protocol.pyx":317
 *                                                 spaces, indexes)
 *                 except Exception as e:
 *                     logger.exception(e)             # <<<<<<<<<<<<<<
 *                     logger.error('Error happened while parsing schema. '
 *                                  'Space, fields and index names currently '
 */
          __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 317, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_exception); if (unlikely(!__pyx_t_11)) __PYX_ERR(6, 317, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_e};
            __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 317, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "asynctnt/iproto/protocol.pyx":318
 *                 except Exception as e:
 *                     logger.exception(e)
 *                     logger.error('Error happened while parsing schema. '             # <<<<<<<<<<<<<<
 *                                  'Space, fields and index names currently '
 *                                  'not working. Please file an issue at '
 */
          __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_logger); if (unlikely(!__pyx_t_11)) __PYX_ERR(6, 318, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 318, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_Error_happened_while_parsing_sch};
            __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 318, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "asynctnt/iproto/protocol.pyx":322
 *                                  'not working. Please file an issue at '
 *                                  'https://github.com/igorcoding/asynctnt')
 *                     self.auto_refetch_schema = False             # <<<<<<<<<<<<<<
 *                     self.fetch_schema = False
 *                     self._schema_id = -1
 */
          if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 322, __pyx_L21_error) }
          __pyx_cur_scope->__pyx_v_self->auto_refetch_schema = 0;

          /* "asynctnt/iproto/protocol.pyx":323
 *                                  'https://github.com/igorcoding/asynctnt')
 *                     self.auto_refetch_schema = False
 *                     self.fetch_schema = False             # <<<<<<<<<<<<<<
 *                     self._schema_id = -1
 *                     self._set_connection_ready()
 */
          if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 323, __pyx_L21_error) }
          __pyx_cur_scope->__pyx_v_self->fetch_schema = 0;

          /* "asynctnt/iproto/protocol.pyx":324
 *                     self.auto_refetch_schema = False
 *                     self.fetch_schema = False
 *                     self._schema_id = -1             # <<<<<<<<<<<<<<
 *                     self._set_connection_ready()
 *                     if fut is not None and not fut.done():
 */
          if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 324, __pyx_L21_error) }
          __pyx_cur_scope->__pyx_v_self->_schema_id = -1L;

          /* "asynctnt/iproto/protocol.pyx":325
 *                     self.fetch_schema = False
 *                     self._schema_id = -1
 *                     self._set_connection_ready()             # <<<<<<<<<<<<<<
 *                     if fut is not None and not fut.done():
 *                         fut.set_result(None)
 */
          if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 325, __pyx_L21_error) }
          ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->_set_connection_ready(__pyx_cur_scope->__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 325, __pyx_L21_error)

          /* "asynctnt/iproto/protocol.pyx":326
 *                     self._schema_id = -1
 *                     self._set_connection_ready()
 *                     if fut is not None and not fut.done():             # <<<<<<<<<<<<<<
 *                         fut.set_result(None)
 *                     return
 */
          if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(6, 326, __pyx_L21_error) }
          __pyx_t_5 = (__pyx_cur_scope->__pyx_v_fut != Py_None);
          if (__pyx_t_5) {
          } else {
            __pyx_t_6 = __pyx_t_5;
            goto __pyx_L24_bool_binop_done;
          }
          if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(6, 326, __pyx_L21_error) }
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_done); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 326, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_11, NULL};
            __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 326, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 326, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_16 = (!__pyx_t_5);
          __pyx_t_6 = __pyx_t_16;
          __pyx_L24_bool_binop_done:;
          if (__pyx_t_6) {

            /* "asynctnt/iproto/protocol.pyx":327
 *                     self._set_connection_ready()
 *                     if fut is not None and not fut.done():
 *                         fut.set_result(None)             # <<<<<<<<<<<<<<
 *                     return
 * 
 */
            if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(6, 327, __pyx_L21_error) }
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_set_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 327, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_11 = NULL;
            __pyx_t_4 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_11)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_11);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
                __pyx_t_4 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_11, Py_None};
              __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 327, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

            /* "asynctnt/iproto/protocol.pyx":326
 *                     self._schema_id = -1
 *                     self._set_connection_ready()
 *                     if fut is not None and not fut.done():             # <<<<<<<<<<<<<<
 *                         fut.set_result(None)
 *                     return
 */
          }

          /* "asynctnt/iproto/protocol.pyx":328
 *                     if fut is not None and not fut.done():
 *                         fut.set_result(None)
 *                     return             # <<<<<<<<<<<<<<
 * 
 *                 self._schema_id = self._schema.id
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_r = Py_None; __Pyx_INCREF(Py_None);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L20_return;
        }

        /* "asynctnt/iproto/protocol.pyx":316
 *                     self._schema = Schema.parse(spaces.schema_id,
 *                                                 spaces, indexes)
 *                 except Exception as e:             # <<<<<<<<<<<<<<
 *                     logger.exception(e)
 *                     logger.error('Error happened while parsing schema. '
 */
        /*finally:*/ {
          __pyx_L21_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21) < 0)) __Pyx_ErrFetch(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_23);
            __Pyx_XGOTREF(__pyx_t_24);
            __pyx_t_4 = __pyx_lineno; __pyx_t_17 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_XGIVEREF(__pyx_t_23);
              __Pyx_XGIVEREF(__pyx_t_24);
              __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_23, __pyx_t_24);
            }
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_ErrRestore(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
            __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_17; __pyx_filename = __pyx_t_18;
            goto __pyx_L12_except_error;
          }
          __pyx_L20_return: {
            __pyx_t_24 = __pyx_r;
            __pyx_r = 0;
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            __pyx_r = __pyx_t_24;
            __pyx_t_24 = 0;
            goto __pyx_L13_except_return;
          }
        }
      }
      goto __pyx_L12_except_error;

      /* "asynctnt/iproto/protocol.pyx":313
 *                              self.host, self.port,
 *                              spaces.schema_id, len(spaces), len(indexes))
 *                 try:             # <<<<<<<<<<<<<<
 *                     self._schema = Schema.parse(spaces.schema_id,
 *                                                 spaces, indexes)
 */
      __pyx_L12_except_error:;
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      goto __pyx_L1_error;
      __pyx_L13_except_return:;
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      goto __pyx_L0;
      __pyx_L15_try_end:;
    }

    /* "asynctnt/iproto/protocol.pyx":330
 *                     return
 * 
 *                 self._schema_id = self._schema.id             # <<<<<<<<<<<<<<
 *                 self.post_con_state = POST_CONNECTION_DONE
 *                 self._post_con_state_machine()
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 330, __pyx_L1_error) }
    __pyx_t_17 = __pyx_cur_scope->__pyx_v_self->_schema->id;
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 330, __pyx_L1_error) }
    __pyx_cur_scope->__pyx_v_self->_schema_id = __pyx_t_17;

    /* "asynctnt/iproto/protocol.pyx":331
 * 
 *                 self._schema_id = self._schema.id
 *                 self.post_con_state = POST_CONNECTION_DONE             # <<<<<<<<<<<<<<
 *                 self._post_con_state_machine()
 *                 if fut is not None and not fut.done():
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 331, __pyx_L1_error) }
    __pyx_cur_scope->__pyx_v_self->post_con_state = __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_DONE;

    /* "asynctnt/iproto/protocol.pyx":332
 *                 self._schema_id = self._schema.id
 *                 self.post_con_state = POST_CONNECTION_DONE
 *                 self._post_con_state_machine()             # <<<<<<<<<<<<<<
 *                 if fut is not None and not fut.done():
 *                     fut.set_result(self._schema)
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 332, __pyx_L1_error) }
    ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->_post_con_state_machine(__pyx_cur_scope->__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 332, __pyx_L1_error)

    /* "asynctnt/iproto/protocol.pyx":333
 *                 self.post_con_state = POST_CONNECTION_DONE
 *                 self._post_con_state_machine()
 *                 if fut is not None and not fut.done():             # <<<<<<<<<<<<<<
 *                     fut.set_result(self._schema)
 *             else:
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(6, 333, __pyx_L1_error) }
    __pyx_t_16 = (__pyx_cur_scope->__pyx_v_fut != Py_None);
    if (__pyx_t_16) {
    } else {
      __pyx_t_6 = __pyx_t_16;
      goto __pyx_L31_bool_binop_done;
    }
    if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(6, 333, __pyx_L1_error) }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_done); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    __pyx_t_17 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_17 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
      __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_17, 0+__pyx_t_17);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 333, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(6, 333, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_5 = (!__pyx_t_16);
    __pyx_t_6 = __pyx_t_5;
    __pyx_L31_bool_binop_done:;
    if (__pyx_t_6) {

      /* "asynctnt/iproto/protocol.pyx":334
 *                 self._post_con_state_machine()
 *                 if fut is not None and not fut.done():
 *                     fut.set_result(self._schema)             # <<<<<<<<<<<<<<
 *             else:
 *                 if self._closing:
 */
      if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(6, 334, __pyx_L1_error) }
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_set_result); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 334, __pyx_L1_error) }
      __pyx_t_1 = NULL;
      __pyx_t_17 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_17 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, ((PyObject *)__pyx_cur_scope->__pyx_v_self->_schema)};
        __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_17, 1+__pyx_t_17);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "asynctnt/iproto/protocol.pyx":333
 *                 self.post_con_state = POST_CONNECTION_DONE
 *                 self._post_con_state_machine()
 *                 if fut is not None and not fut.done():             # <<<<<<<<<<<<<<
 *                     fut.set_result(self._schema)
 *             else:
 */
    }

    /* "asynctnt/iproto/protocol.pyx":303
 *                 return
 *             e = f.exception()
 *             if not e:             # <<<<<<<<<<<<<<
 *                 if fut is not None and fut.cancelled():
 *                     # if the caller has cancelled waiting
 */
    goto __pyx_L4;
  }

  /* "asynctnt/iproto/protocol.pyx":336
 *                     fut.set_result(self._schema)
 *             else:
 *                 if self._closing:             # <<<<<<<<<<<<<<
 *                     # show a diag message rather than Lost connection to Tarantool when disconnected (#19)
 *                     logger.debug('Schema fetch stopped: connection is closed')
 */
  /*else*/ {
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 336, __pyx_L1_error) }
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_closing); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(6, 336, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (__pyx_t_6) {

      /* "asynctnt/iproto/protocol.pyx":338
 *                 if self._closing:
 *                     # show a diag message rather than Lost connection to Tarantool when disconnected (#19)
 *                     logger.debug('Schema fetch stopped: connection is closed')             # <<<<<<<<<<<<<<
 *                     return
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      __pyx_t_17 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_17 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_kp_u_Schema_fetch_stopped_connection};
        __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_17, 1+__pyx_t_17);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 338, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "asynctnt/iproto/protocol.pyx":339
 *                     # show a diag message rather than Lost connection to Tarantool when disconnected (#19)
 *                     logger.debug('Schema fetch stopped: connection is closed')
 *                     return             # <<<<<<<<<<<<<<
 * 
 *                 logger.error('Tarantool[%s:%s] Schema fetch failed: %s',
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "asynctnt/iproto/protocol.pyx":336
 *                     fut.set_result(self._schema)
 *             else:
 *                 if self._closing:             # <<<<<<<<<<<<<<
 *                     # show a diag message rather than Lost connection to Tarantool when disconnected (#19)
 *                     logger.debug('Schema fetch stopped: connection is closed')
 */
    }

    /* "asynctnt/iproto/protocol.pyx":341
 *                     return
 * 
 *                 logger.error('Tarantool[%s:%s] Schema fetch failed: %s',             # <<<<<<<<<<<<<<
 *                              self.host, self.port, str(e))
 *                 if isinstance(e, asyncio.TimeoutError):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_error); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":342
 * 
 *                 logger.error('Tarantool[%s:%s] Schema fetch failed: %s',
 *                              self.host, self.port, str(e))             # <<<<<<<<<<<<<<
 *                 if isinstance(e, asyncio.TimeoutError):
 *                     e = asyncio.TimeoutError('Schema fetch timeout')
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 342, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 342, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_Str(__pyx_v_e); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    __pyx_t_17 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_17 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_kp_u_Tarantool_s_s_Schema_fetch_faile, __pyx_cur_scope->__pyx_v_self->__pyx_base.host, __pyx_cur_scope->__pyx_v_self->__pyx_base.port, __pyx_t_1};
      __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_17, 4+__pyx_t_17);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 341, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "asynctnt/iproto/protocol.pyx":343
 *                 logger.error('Tarantool[%s:%s] Schema fetch failed: %s',
 *                              self.host, self.port, str(e))
 *                 if isinstance(e, asyncio.TimeoutError):             # <<<<<<<<<<<<<<
 *                     e = asyncio.TimeoutError('Schema fetch timeout')
 *                 self._set_connection_error(e)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_TimeoutError); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_6 = PyObject_IsInstance(__pyx_v_e, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(6, 343, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_6) {

      /* "asynctnt/iproto/protocol.pyx":344
 *                              self.host, self.port, str(e))
 *                 if isinstance(e, asyncio.TimeoutError):
 *                     e = asyncio.TimeoutError('Schema fetch timeout')             # <<<<<<<<<<<<<<
 *                 self._set_connection_error(e)
 *                 if fut is not None and not fut.done():
 */
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 344, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_TimeoutError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 344, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      __pyx_t_17 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_17 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_kp_u_Schema_fetch_timeout};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_17, 1+__pyx_t_17);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 344, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF_SET(__pyx_v_e, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "asynctnt/iproto/protocol.pyx":343
 *                 logger.error('Tarantool[%s:%s] Schema fetch failed: %s',
 *                              self.host, self.port, str(e))
 *                 if isinstance(e, asyncio.TimeoutError):             # <<<<<<<<<<<<<<
 *                     e = asyncio.TimeoutError('Schema fetch timeout')
 *                 self._set_connection_error(e)
 */
    }

    /* "asynctnt/iproto/protocol.pyx":345
 *                 if isinstance(e, asyncio.TimeoutError):
 *                     e = asyncio.TimeoutError('Schema fetch timeout')
 *                 self._set_connection_error(e)             # <<<<<<<<<<<<<<
 *                 if fut is not None and not fut.done():
 *                     fut.set_exception(e)
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(6, 345, __pyx_L1_error) }
    ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->_set_connection_error(__pyx_cur_scope->__pyx_v_self, __pyx_v_e); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 345, __pyx_L1_error)

    /* "asynctnt/iproto/protocol.pyx":346
 *                     e = asyncio.TimeoutError('Schema fetch timeout')
 *                 self._set_connection_error(e)
 *                 if fut is not None and not fut.done():             # <<<<<<<<<<<<<<
 *                     fut.set_exception(e)
 * 
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(6, 346, __pyx_L1_error) }
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_fut != Py_None);
    if (__pyx_t_5) {
    } else {
      __pyx_t_6 = __pyx_t_5;
      goto __pyx_L36_bool_binop_done;
    }
    if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(6, 346, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_done); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = NULL;
    __pyx_t_17 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_17 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_17, 0+__pyx_t_17);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 346, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_16 = (!__pyx_t_5);
    __pyx_t_6 = __pyx_t_16;
    __pyx_L36_bool_binop_done:;
    if (__pyx_t_6) {

      /* "asynctnt/iproto/protocol.pyx":347
 *                 self._set_connection_error(e)
 *                 if fut is not None and not fut.done():
 *                     fut.set_exception(e)             # <<<<<<<<<<<<<<
 * 
 *         fut_vspace = self._db.select(SPACE_VSPACE, timeout=0,
 */
      if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(6, 347, __pyx_L1_error) }
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = NULL;
      __pyx_t_17 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_17 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_v_e};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_17, 1+__pyx_t_17);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 347, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "asynctnt/iproto/protocol.pyx":346
 *                     e = asyncio.TimeoutError('Schema fetch timeout')
 *                 self._set_connection_error(e)
 *                 if fut is not None and not fut.done():             # <<<<<<<<<<<<<<
 *                     fut.set_exception(e)
 * 
 */
    }
  }
  __pyx_L4:;

  /* "asynctnt/iproto/protocol.pyx":296
 *         self._schema_fetch_in_progress = True
 * 
 *         def on_fetch(f):             # <<<<<<<<<<<<<<
 *             self._schema_fetch_in_progress = False
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._do_fetch_schema.on_fetch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_spaces);
  __Pyx_XDECREF(__pyx_v_indexes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":290
 *         fut.add_done_callback(on_authorized)
 * 
 *     cdef void _do_fetch_schema(self, object fut):             # <<<<<<<<<<<<<<
 *         if self._schema_fetch_in_progress:
 *             return
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__do_fetch_schema(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_fut) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *__pyx_cur_scope;
  PyObject *__pyx_v_on_fetch = 0;
  PyObject *__pyx_v_fut_vspace = NULL;
  PyObject *__pyx_v_fut_vindex = NULL;
  PyObject *__pyx_v_gather_fut = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_fetch_schema", 0);
  __pyx_cur_scope = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(6, 290, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_fut = __pyx_v_fut;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_fut);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_fut);

  /* "asynctnt/iproto/protocol.pyx":291
 * 
 *     cdef void _do_fetch_schema(self, object fut):
 *         if self._schema_fetch_in_progress:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  if (__pyx_cur_scope->__pyx_v_self->_schema_fetch_in_progress) {

    /* "asynctnt/iproto/protocol.pyx":292
 *     cdef void _do_fetch_schema(self, object fut):
 *         if self._schema_fetch_in_progress:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._schema_fetch_in_progress = True
 */
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":291
 * 
 *     cdef void _do_fetch_schema(self, object fut):
 *         if self._schema_fetch_in_progress:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "asynctnt/iproto/protocol.pyx":294
 *             return
 * 
 *         self._schema_fetch_in_progress = True             # <<<<<<<<<<<<<<
 * 
 *         def on_fetch(f):
 */
  __pyx_cur_scope->__pyx_v_self->_schema_fetch_in_progress = 1;

  /* "asynctnt/iproto/protocol.pyx":296
 *         self._schema_fetch_in_progress = True
 * 
 *         def on_fetch(f):             # <<<<<<<<<<<<<<
 *             self._schema_fetch_in_progress = False
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_16_do_fetch_schema_1on_fetch, 0, __pyx_n_s_BaseProtocol__do_fetch_schema_lo, ((PyObject*)__pyx_cur_scope), __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_on_fetch = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":349
 *                     fut.set_exception(e)
 * 
 *         fut_vspace = self._db.select(SPACE_VSPACE, timeout=0,             # <<<<<<<<<<<<<<
 *                                      check_schema_change=False)
 *         fut_vindex = self._db.select(SPACE_VINDEX, timeout=0,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->_db), __pyx_n_s_select); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_timeout, __pyx_int_0) < 0) __PYX_ERR(6, 349, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":350
 * 
 *         fut_vspace = self._db.select(SPACE_VSPACE, timeout=0,
 *                                      check_schema_change=False)             # <<<<<<<<<<<<<<
 *         fut_vindex = self._db.select(SPACE_VINDEX, timeout=0,
 *                                      check_schema_change=False)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_check_schema_change, Py_False) < 0) __PYX_ERR(6, 349, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":349
 *                     fut.set_exception(e)
 * 
 *         fut_vspace = self._db.select(SPACE_VSPACE, timeout=0,             # <<<<<<<<<<<<<<
 *                                      check_schema_change=False)
 *         fut_vindex = self._db.select(SPACE_VINDEX, timeout=0,
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__46, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_fut_vspace = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/protocol.pyx":351
 *         fut_vspace = self._db.select(SPACE_VSPACE, timeout=0,
 *                                      check_schema_change=False)
 *         fut_vindex = self._db.select(SPACE_VINDEX, timeout=0,             # <<<<<<<<<<<<<<
 *                                      check_schema_change=False)
 *         gather_fut = asyncio.gather(fut_vspace, fut_vindex,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->_db), __pyx_n_s_select); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_timeout, __pyx_int_0) < 0) __PYX_ERR(6, 351, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":352
 *                                      check_schema_change=False)
 *         fut_vindex = self._db.select(SPACE_VINDEX, timeout=0,
 *                                      check_schema_change=False)             # <<<<<<<<<<<<<<
 *         gather_fut = asyncio.gather(fut_vspace, fut_vindex,
 *                                     return_exceptions=False)
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_check_schema_change, Py_False) < 0) __PYX_ERR(6, 351, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":351
 *         fut_vspace = self._db.select(SPACE_VSPACE, timeout=0,
 *                                      check_schema_change=False)
 *         fut_vindex = self._db.select(SPACE_VINDEX, timeout=0,             # <<<<<<<<<<<<<<
 *                                      check_schema_change=False)
 *         gather_fut = asyncio.gather(fut_vspace, fut_vindex,
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__47, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_fut_vindex = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":353
 *         fut_vindex = self._db.select(SPACE_VINDEX, timeout=0,
 *                                      check_schema_change=False)
 *         gather_fut = asyncio.gather(fut_vspace, fut_vindex,             # <<<<<<<<<<<<<<
 *                                     return_exceptions=False)
 *         gather_fut.add_done_callback(on_fetch)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_gather); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_fut_vspace);
  __Pyx_GIVEREF(__pyx_v_fut_vspace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_fut_vspace)) __PYX_ERR(6, 353, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_fut_vindex);
  __Pyx_GIVEREF(__pyx_v_fut_vindex);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_fut_vindex)) __PYX_ERR(6, 353, __pyx_L1_error);

  /* "asynctnt/iproto/protocol.pyx":354
 *                                      check_schema_change=False)
 *         gather_fut = asyncio.gather(fut_vspace, fut_vindex,
 *                                     return_exceptions=False)             # <<<<<<<<<<<<<<
 *         gather_fut.add_done_callback(on_fetch)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return_exceptions, Py_False) < 0) __PYX_ERR(6, 354, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":353
 *         fut_vindex = self._db.select(SPACE_VINDEX, timeout=0,
 *                                      check_schema_change=False)
 *         gather_fut = asyncio.gather(fut_vspace, fut_vindex,             # <<<<<<<<<<<<<<
 *                                     return_exceptions=False)
 *         gather_fut.add_done_callback(on_fetch)
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_gather_fut = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "asynctnt/iproto/protocol.pyx":355
 *         gather_fut = asyncio.gather(fut_vspace, fut_vindex,
 *                                     return_exceptions=False)
 *         gather_fut.add_done_callback(on_fetch)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _on_connection_made(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_gather_fut, __pyx_n_s_add_done_callback); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_on_fetch};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "asynctnt/iproto/protocol.pyx":290
 *         fut.add_done_callback(on_authorized)
 * 
 *     cdef void _do_fetch_schema(self, object fut):             # <<<<<<<<<<<<<<
 *         if self._schema_fetch_in_progress:
 *             return
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._do_fetch_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_on_fetch);
  __Pyx_XDECREF(__pyx_v_fut_vspace);
  __Pyx_XDECREF(__pyx_v_fut_vindex);
  __Pyx_XDECREF(__pyx_v_gather_fut);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/protocol.pyx":357
 *         gather_fut.add_done_callback(on_fetch)
 * 
 *     cdef void _on_connection_made(self):             # <<<<<<<<<<<<<<
 *         CoreProtocol._on_connection_made(self)
 * 
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_connection_made(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_connection_made", 1);

  /* "asynctnt/iproto/protocol.pyx":358
 * 
 *     cdef void _on_connection_made(self):
 *         CoreProtocol._on_connection_made(self)             # <<<<<<<<<<<<<<
 * 
 *         if self.on_connection_made_cb:
 */
  __pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_connection_made(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 358, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":360
 *         CoreProtocol._on_connection_made(self)
 * 
 *         if self.on_connection_made_cb:             # <<<<<<<<<<<<<<
 *             self.on_connection_made_cb()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->on_connection_made_cb); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(6, 360, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "asynctnt/iproto/protocol.pyx":361
 * 
 *         if self.on_connection_made_cb:
 *             self.on_connection_made_cb()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _on_connection_lost(self, exc):
 */
    __Pyx_INCREF(__pyx_v_self->on_connection_made_cb);
    __pyx_t_3 = __pyx_v_self->on_connection_made_cb; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 361, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynctnt/iproto/protocol.pyx":360
 *         CoreProtocol._on_connection_made(self)
 * 
 *         if self.on_connection_made_cb:             # <<<<<<<<<<<<<<
 *             self.on_connection_made_cb()
 * 
 */
  }

  /* "asynctnt/iproto/protocol.pyx":357
 *         gather_fut.add_done_callback(on_fetch)
 * 
 *     cdef void _on_connection_made(self):             # <<<<<<<<<<<<<<
 *         CoreProtocol._on_connection_made(self)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._on_connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/protocol.pyx":363
 *             self.on_connection_made_cb()
 * 
 *     cdef void _on_connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         cdef:
 *             BaseRequest req
 */

static void __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_connection_lost(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_exc) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_response = 0;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pvalue;
  Py_ssize_t __pyx_v_pos;
  CYTHON_UNUSED uint64_t __pyx_v_sync;
  PyObject *__pyx_v_waiter = NULL;
  PyObject *__pyx_v_err = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *(*__pyx_t_3)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, float);
  uint64_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_connection_lost", 1);

  /* "asynctnt/iproto/protocol.pyx":372
 *             Py_ssize_t pos
 * 
 *         if self._closing:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closing); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "asynctnt/iproto/protocol.pyx":373
 * 
 *         if self._closing:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._closing = True
 */
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":372
 *             Py_ssize_t pos
 * 
 *         if self._closing:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "asynctnt/iproto/protocol.pyx":375
 *             return
 * 
 *         self._closing = True             # <<<<<<<<<<<<<<
 *         self.post_con_state = POST_CONNECTION_NONE
 *         self.execute = self._execute_bad
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closing, Py_True) < 0) __PYX_ERR(6, 375, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":376
 * 
 *         self._closing = True
 *         self.post_con_state = POST_CONNECTION_NONE             # <<<<<<<<<<<<<<
 *         self.execute = self._execute_bad
 * 
 */
  __pyx_v_self->post_con_state = __pyx_e_8asynctnt_6iproto_8protocol_POST_CONNECTION_NONE;

  /* "asynctnt/iproto/protocol.pyx":377
 *         self._closing = True
 *         self.post_con_state = POST_CONNECTION_NONE
 *         self.execute = self._execute_bad             # <<<<<<<<<<<<<<
 * 
 *         pos = 0
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_execute_bad;
  __pyx_v_self->execute = __pyx_t_3;

  /* "asynctnt/iproto/protocol.pyx":379
 *         self.execute = self._execute_bad
 * 
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while cpython.dict.PyDict_Next(self._reqs, &pos, &pkey, &pvalue):
 *             sync = <uint64_t> <object> pkey
 */
  __pyx_v_pos = 0;

  /* "asynctnt/iproto/protocol.pyx":380
 * 
 *         pos = 0
 *         while cpython.dict.PyDict_Next(self._reqs, &pos, &pkey, &pvalue):             # <<<<<<<<<<<<<<
 *             sync = <uint64_t> <object> pkey
 *             response = <Response> pvalue
 */
  while (1) {
    __pyx_t_1 = __pyx_v_self->_reqs;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = (PyDict_Next(__pyx_t_1, (&__pyx_v_pos), (&__pyx_v_pkey), (&__pyx_v_pvalue)) != 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_2) break;

    /* "asynctnt/iproto/protocol.pyx":381
 *         pos = 0
 *         while cpython.dict.PyDict_Next(self._reqs, &pos, &pkey, &pvalue):
 *             sync = <uint64_t> <object> pkey             # <<<<<<<<<<<<<<
 *             response = <Response> pvalue
 *             req = response.request_
 */
    __pyx_t_4 = __Pyx_PyInt_As_uint64_t(((PyObject *)__pyx_v_pkey)); if (unlikely((__pyx_t_4 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(6, 381, __pyx_L1_error)
    __pyx_v_sync = ((uint64_t)__pyx_t_4);

    /* "asynctnt/iproto/protocol.pyx":382
 *         while cpython.dict.PyDict_Next(self._reqs, &pos, &pkey, &pvalue):
 *             sync = <uint64_t> <object> pkey
 *             response = <Response> pvalue             # <<<<<<<<<<<<<<
 *             req = response.request_
 * 
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_pvalue);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_response, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":383
 *             sync = <uint64_t> <object> pkey
 *             response = <Response> pvalue
 *             req = response.request_             # <<<<<<<<<<<<<<
 * 
 *             waiter = req.waiter
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_response->request_);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_req, ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":385
 *             req = response.request_
 * 
 *             waiter = req.waiter             # <<<<<<<<<<<<<<
 *             if waiter and not waiter.done():
 *                 err = None
 */
    __pyx_t_1 = __pyx_v_req->waiter;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_waiter, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":386
 * 
 *             waiter = req.waiter
 *             if waiter and not waiter.done():             # <<<<<<<<<<<<<<
 *                 err = None
 *                 if exc is None:
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_waiter); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 386, __pyx_L1_error)
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_done); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 386, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 386, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = (!__pyx_t_5);
    __pyx_t_2 = __pyx_t_9;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_2) {

      /* "asynctnt/iproto/protocol.pyx":387
 *             waiter = req.waiter
 *             if waiter and not waiter.done():
 *                 err = None             # <<<<<<<<<<<<<<
 *                 if exc is None:
 *                     err = TarantoolNotConnectedError(
 */
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_err, Py_None);

      /* "asynctnt/iproto/protocol.pyx":388
 *             if waiter and not waiter.done():
 *                 err = None
 *                 if exc is None:             # <<<<<<<<<<<<<<
 *                     err = TarantoolNotConnectedError(
 *                         'Lost connection to Tarantool'
 */
      __pyx_t_2 = (__pyx_v_exc == Py_None);
      if (__pyx_t_2) {

        /* "asynctnt/iproto/protocol.pyx":389
 *                 err = None
 *                 if exc is None:
 *                     err = TarantoolNotConnectedError(             # <<<<<<<<<<<<<<
 *                         'Lost connection to Tarantool'
 *                     )
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_TarantoolNotConnectedError); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 389, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_kp_u_Lost_connection_to_Tarantool};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 389, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF_SET(__pyx_v_err, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "asynctnt/iproto/protocol.pyx":388
 *             if waiter and not waiter.done():
 *                 err = None
 *                 if exc is None:             # <<<<<<<<<<<<<<
 *                     err = TarantoolNotConnectedError(
 *                         'Lost connection to Tarantool'
 */
        goto __pyx_L9;
      }

      /* "asynctnt/iproto/protocol.pyx":392
 *                         'Lost connection to Tarantool'
 *                     )
 *                 elif isinstance(exc, (ConnectionRefusedError,             # <<<<<<<<<<<<<<
 *                                       ConnectionResetError)):
 *                     err = TarantoolNotConnectedError(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ConnectionRefusedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "asynctnt/iproto/protocol.pyx":393
 *                     )
 *                 elif isinstance(exc, (ConnectionRefusedError,
 *                                       ConnectionResetError)):             # <<<<<<<<<<<<<<
 *                     err = TarantoolNotConnectedError(
 *                         'Lost connection to Tarantool: {}: {}'.format(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ConnectionResetError); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "asynctnt/iproto/protocol.pyx":392
 *                         'Lost connection to Tarantool'
 *                     )
 *                 elif isinstance(exc, (ConnectionRefusedError,             # <<<<<<<<<<<<<<
 *                                       ConnectionResetError)):
 *                     err = TarantoolNotConnectedError(
 */
      __pyx_t_9 = PyObject_IsInstance(__pyx_v_exc, __pyx_t_1); 
      if (!__pyx_t_9) {
      } else {
        __pyx_t_2 = __pyx_t_9;
        goto __pyx_L10_bool_binop_done;
      }

      /* "asynctnt/iproto/protocol.pyx":393
 *                     )
 *                 elif isinstance(exc, (ConnectionRefusedError,
 *                                       ConnectionResetError)):             # <<<<<<<<<<<<<<
 *                     err = TarantoolNotConnectedError(
 *                         'Lost connection to Tarantool: {}: {}'.format(
 */
      __pyx_t_9 = PyObject_IsInstance(__pyx_v_exc, __pyx_t_6); 
      __pyx_t_2 = __pyx_t_9;
      __pyx_L10_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynctnt/iproto/protocol.pyx":392
 *                         'Lost connection to Tarantool'
 *                     )
 *                 elif isinstance(exc, (ConnectionRefusedError,             # <<<<<<<<<<<<<<
 *                                       ConnectionResetError)):
 *                     err = TarantoolNotConnectedError(
 */
      if (__pyx_t_2) {

        /* "asynctnt/iproto/protocol.pyx":394
 *                 elif isinstance(exc, (ConnectionRefusedError,
 *                                       ConnectionResetError)):
 *                     err = TarantoolNotConnectedError(             # <<<<<<<<<<<<<<
 *                         'Lost connection to Tarantool: {}: {}'.format(
 *                             exc.__class__.__name__, str(exc))
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_TarantoolNotConnectedError); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "asynctnt/iproto/protocol.pyx":395
 *                                       ConnectionResetError)):
 *                     err = TarantoolNotConnectedError(
 *                         'Lost connection to Tarantool: {}: {}'.format(             # <<<<<<<<<<<<<<
 *                             exc.__class__.__name__, str(exc))
 *                     )
 */
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Lost_connection_to_Tarantool_2, __pyx_n_s_format); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);

        /* "asynctnt/iproto/protocol.pyx":396
 *                     err = TarantoolNotConnectedError(
 *                         'Lost connection to Tarantool: {}: {}'.format(
 *                             exc.__class__.__name__, str(exc))             # <<<<<<<<<<<<<<
 *                     )
 *                 elif exc is ConnectionRefusedError \
 */
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc, __pyx_n_s_class); if (unlikely(!__pyx_t_11)) __PYX_ERR(6, 396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_name); if (unlikely(!__pyx_t_12)) __PYX_ERR(6, 396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = __Pyx_PyObject_Str(__pyx_v_exc); if (unlikely(!__pyx_t_11)) __PYX_ERR(6, 396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_13 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_13, __pyx_t_12, __pyx_t_11};
          __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 395, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __pyx_t_10 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_7};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF_SET(__pyx_v_err, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "asynctnt/iproto/protocol.pyx":392
 *                         'Lost connection to Tarantool'
 *                     )
 *                 elif isinstance(exc, (ConnectionRefusedError,             # <<<<<<<<<<<<<<
 *                                       ConnectionResetError)):
 *                     err = TarantoolNotConnectedError(
 */
        goto __pyx_L9;
      }

      /* "asynctnt/iproto/protocol.pyx":398
 *                             exc.__class__.__name__, str(exc))
 *                     )
 *                 elif exc is ConnectionRefusedError \             # <<<<<<<<<<<<<<
 *                         or exc is ConnectionResetError:
 *                     err = TarantoolNotConnectedError(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ConnectionRefusedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = (__pyx_v_exc == __pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!__pyx_t_9) {
      } else {
        __pyx_t_2 = __pyx_t_9;
        goto __pyx_L12_bool_binop_done;
      }

      /* "asynctnt/iproto/protocol.pyx":399
 *                     )
 *                 elif exc is ConnectionRefusedError \
 *                         or exc is ConnectionResetError:             # <<<<<<<<<<<<<<
 *                     err = TarantoolNotConnectedError(
 *                         'Lost connection to Tarantool: {}'.format(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ConnectionResetError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 399, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = (__pyx_v_exc == __pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = __pyx_t_9;
      __pyx_L12_bool_binop_done:;

      /* "asynctnt/iproto/protocol.pyx":398
 *                             exc.__class__.__name__, str(exc))
 *                     )
 *                 elif exc is ConnectionRefusedError \             # <<<<<<<<<<<<<<
 *                         or exc is ConnectionResetError:
 *                     err = TarantoolNotConnectedError(
 */
      if (__pyx_t_2) {

        /* "asynctnt/iproto/protocol.pyx":400
 *                 elif exc is ConnectionRefusedError \
 *                         or exc is ConnectionResetError:
 *                     err = TarantoolNotConnectedError(             # <<<<<<<<<<<<<<
 *                         'Lost connection to Tarantool: {}'.format(
 *                             exc.__name__)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_TarantoolNotConnectedError); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 400, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "asynctnt/iproto/protocol.pyx":401
 *                         or exc is ConnectionResetError:
 *                     err = TarantoolNotConnectedError(
 *                         'Lost connection to Tarantool: {}'.format(             # <<<<<<<<<<<<<<
 *                             exc.__name__)
 *                     )
 */
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Lost_connection_to_Tarantool_3, __pyx_n_s_format); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);

        /* "asynctnt/iproto/protocol.pyx":402
 *                     err = TarantoolNotConnectedError(
 *                         'Lost connection to Tarantool: {}'.format(
 *                             exc.__name__)             # <<<<<<<<<<<<<<
 *                     )
 *                 else:
 */
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc, __pyx_n_s_name); if (unlikely(!__pyx_t_11)) __PYX_ERR(6, 402, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_11};
          __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 401, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __pyx_t_10 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_7};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 400, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF_SET(__pyx_v_err, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "asynctnt/iproto/protocol.pyx":398
 *                             exc.__class__.__name__, str(exc))
 *                     )
 *                 elif exc is ConnectionRefusedError \             # <<<<<<<<<<<<<<
 *                         or exc is ConnectionResetError:
 *                     err = TarantoolNotConnectedError(
 */
        goto __pyx_L9;
      }

      /* "asynctnt/iproto/protocol.pyx":405
 *                     )
 *                 else:
 *                     err = exc             # <<<<<<<<<<<<<<
 * 
 *                 if err is not None:
 */
      /*else*/ {
        __Pyx_INCREF(__pyx_v_exc);
        __Pyx_DECREF_SET(__pyx_v_err, __pyx_v_exc);
      }
      __pyx_L9:;

      /* "asynctnt/iproto/protocol.pyx":407
 *                     err = exc
 * 
 *                 if err is not None:             # <<<<<<<<<<<<<<
 *                     waiter.set_exception(err)
 *                     if response is not None:
 */
      __pyx_t_2 = (__pyx_v_err != Py_None);
      if (__pyx_t_2) {

        /* "asynctnt/iproto/protocol.pyx":408
 * 
 *                 if err is not None:
 *                     waiter.set_exception(err)             # <<<<<<<<<<<<<<
 *                     if response is not None:
 *                         response.set_exception(err)
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 408, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_err};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 408, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynctnt/iproto/protocol.pyx":409
 *                 if err is not None:
 *                     waiter.set_exception(err)
 *                     if response is not None:             # <<<<<<<<<<<<<<
 *                         response.set_exception(err)
 * 
 */
        __pyx_t_2 = (((PyObject *)__pyx_v_response) != Py_None);
        if (__pyx_t_2) {

          /* "asynctnt/iproto/protocol.pyx":410
 *                     waiter.set_exception(err)
 *                     if response is not None:
 *                         response.set_exception(err)             # <<<<<<<<<<<<<<
 * 
 *         if self.on_connection_lost_cb:
 */
          __pyx_f_8asynctnt_6iproto_8protocol_8Response_set_exception(__pyx_v_response, __pyx_v_err); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 410, __pyx_L1_error)

          /* "asynctnt/iproto/protocol.pyx":409
 *                 if err is not None:
 *                     waiter.set_exception(err)
 *                     if response is not None:             # <<<<<<<<<<<<<<
 *                         response.set_exception(err)
 * 
 */
        }

        /* "asynctnt/iproto/protocol.pyx":407
 *                     err = exc
 * 
 *                 if err is not None:             # <<<<<<<<<<<<<<
 *                     waiter.set_exception(err)
 *                     if response is not None:
 */
      }

      /* "asynctnt/iproto/protocol.pyx":386
 * 
 *             waiter = req.waiter
 *             if waiter and not waiter.done():             # <<<<<<<<<<<<<<
 *                 err = None
 *                 if exc is None:
 */
    }
  }

  /* "asynctnt/iproto/protocol.pyx":412
 *                         response.set_exception(err)
 * 
 *         if self.on_connection_lost_cb:             # <<<<<<<<<<<<<<
 *             self.on_connection_lost_cb(exc)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->on_connection_lost_cb); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 412, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "asynctnt/iproto/protocol.pyx":413
 * 
 *         if self.on_connection_lost_cb:
 *             self.on_connection_lost_cb(exc)             # <<<<<<<<<<<<<<
 * 
 *         self._reqs = {}  # reset requests map
 */
    __Pyx_INCREF(__pyx_v_self->on_connection_lost_cb);
    __pyx_t_6 = __pyx_v_self->on_connection_lost_cb; __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_exc};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":412
 *                         response.set_exception(err)
 * 
 *         if self.on_connection_lost_cb:             # <<<<<<<<<<<<<<
 *             self.on_connection_lost_cb(exc)
 * 
 */
  }

  /* "asynctnt/iproto/protocol.pyx":415
 *             self.on_connection_lost_cb(exc)
 * 
 *         self._reqs = {}  # reset requests map             # <<<<<<<<<<<<<<
 * 
 *     cdef inline uint64_t next_sync(self):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_reqs);
  __Pyx_DECREF(__pyx_v_self->_reqs);
  __pyx_v_self->_reqs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":363
 *             self.on_connection_made_cb()
 * 
 *     cdef void _on_connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         cdef:
 *             BaseRequest req
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._on_connection_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XDECREF((PyObject *)__pyx_v_response);
  __Pyx_XDECREF(__pyx_v_waiter);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_RefNannyFinishContext();
}

/* "asynctnt/iproto/protocol.pyx":417
 *         self._reqs = {}  # reset requests map
 * 
 *     cdef inline uint64_t next_sync(self):             # <<<<<<<<<<<<<<
 *         self._sync += 1
 *         return self._sync
 */

static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_next_sync(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  uint64_t __pyx_r;

  /* "asynctnt/iproto/protocol.pyx":418
 * 
 *     cdef inline uint64_t next_sync(self):
 *         self._sync += 1             # <<<<<<<<<<<<<<
 *         return self._sync
 * 
 */
  __pyx_v_self->_sync = (__pyx_v_self->_sync + 1);

  /* "asynctnt/iproto/protocol.pyx":419
 *     cdef inline uint64_t next_sync(self):
 *         self._sync += 1
 *         return self._sync             # <<<<<<<<<<<<<<
 * 
 *     cdef inline uint64_t next_stream_id(self):
 */
  __pyx_r = __pyx_v_self->_sync;
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":417
 *         self._reqs = {}  # reset requests map
 * 
 *     cdef inline uint64_t next_sync(self):             # <<<<<<<<<<<<<<
 *         self._sync += 1
 *         return self._sync
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":421
 *         return self._sync
 * 
 *     cdef inline uint64_t next_stream_id(self):             # <<<<<<<<<<<<<<
 *         self._last_stream_id += 1
 *         return self._last_stream_id
 */

static CYTHON_INLINE uint64_t __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_next_stream_id(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  uint64_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("next_stream_id", 1);

  /* "asynctnt/iproto/protocol.pyx":422
 * 
 *     cdef inline uint64_t next_stream_id(self):
 *         self._last_stream_id += 1             # <<<<<<<<<<<<<<
 *         return self._last_stream_id
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_last_stream_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_last_stream_id, __pyx_t_2) < 0) __PYX_ERR(6, 422, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":423
 *     cdef inline uint64_t next_stream_id(self):
 *         self._last_stream_id += 1
 *         return self._last_stream_id             # <<<<<<<<<<<<<<
 * 
 *     def _on_request_timeout(self, waiter):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_last_stream_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_uint64_t(__pyx_t_2); if (unlikely((__pyx_t_3 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(6, 423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":421
 *         return self._sync
 * 
 *     cdef inline uint64_t next_stream_id(self):             # <<<<<<<<<<<<<<
 *         self._last_stream_id += 1
 *         return self._last_stream_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.next_stream_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":425
 *         return self._last_stream_id
 * 
 *     def _on_request_timeout(self, waiter):             # <<<<<<<<<<<<<<
 *         cdef:
 *             BaseRequest req
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_5_on_request_timeout(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_5_on_request_timeout = {"_on_request_timeout", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_5_on_request_timeout, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_5_on_request_timeout(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_waiter = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_on_request_timeout (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_waiter,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_waiter)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 425, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_on_request_timeout") < 0)) __PYX_ERR(6, 425, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_waiter = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_on_request_timeout", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 425, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._on_request_timeout", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_4_on_request_timeout(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self), __pyx_v_waiter);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_4_on_request_timeout(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_waiter) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req = 0;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_response = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_request_timeout", 1);

  /* "asynctnt/iproto/protocol.pyx":430
 *             Response response
 * 
 *         if waiter.done():             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_done); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(6, 430, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "asynctnt/iproto/protocol.pyx":431
 * 
 *         if waiter.done():
 *             return             # <<<<<<<<<<<<<<
 * 
 *         response = waiter._response
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":430
 *             Response response
 * 
 *         if waiter.done():             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "asynctnt/iproto/protocol.pyx":433
 *             return
 * 
 *         response = waiter._response             # <<<<<<<<<<<<<<
 *         req = response.request_
 *         req.timeout_handle.cancel()
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_response); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_Response))))) __PYX_ERR(6, 433, __pyx_L1_error)
  __pyx_v_response = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":434
 * 
 *         response = waiter._response
 *         req = response.request_             # <<<<<<<<<<<<<<
 *         req.timeout_handle.cancel()
 *         req.timeout_handle = None
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_response->request_);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":435
 *         response = waiter._response
 *         req = response.request_
 *         req.timeout_handle.cancel()             # <<<<<<<<<<<<<<
 *         req.timeout_handle = None
 *         waiter.set_exception(
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_req->timeout_handle, __pyx_n_s_cancel); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":436
 *         req = response.request_
 *         req.timeout_handle.cancel()
 *         req.timeout_handle = None             # <<<<<<<<<<<<<<
 *         waiter.set_exception(
 *             asyncio.TimeoutError(
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_req->timeout_handle);
  __Pyx_DECREF(__pyx_v_req->timeout_handle);
  __pyx_v_req->timeout_handle = Py_None;

  /* "asynctnt/iproto/protocol.pyx":437
 *         req.timeout_handle.cancel()
 *         req.timeout_handle = None
 *         waiter.set_exception(             # <<<<<<<<<<<<<<
 *             asyncio.TimeoutError(
 *                 '{} exceeded timeout'.format(req.__class__.__name__))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "asynctnt/iproto/protocol.pyx":438
 *         req.timeout_handle = None
 *         waiter.set_exception(
 *             asyncio.TimeoutError(             # <<<<<<<<<<<<<<
 *                 '{} exceeded timeout'.format(req.__class__.__name__))
 *         )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_TimeoutError); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "asynctnt/iproto/protocol.pyx":439
 *         waiter.set_exception(
 *             asyncio.TimeoutError(
 *                 '{} exceeded timeout'.format(req.__class__.__name__))             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_exceeded_timeout, __pyx_n_s_format); if (unlikely(!__pyx_t_8)) __PYX_ERR(6, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_req), __pyx_n_s_class); if (unlikely(!__pyx_t_9)) __PYX_ERR(6, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_name); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_10};
    __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __pyx_t_8 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_t_7 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":425
 *         return self._last_stream_id
 * 
 *     def _on_request_timeout(self, waiter):             # <<<<<<<<<<<<<<
 *         cdef:
 *             BaseRequest req
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._on_request_timeout", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XDECREF((PyObject *)__pyx_v_response);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":442
 *         )
 * 
 *     def _on_request_completed(self, fut):             # <<<<<<<<<<<<<<
 *         cdef BaseRequest req = (<Response> fut._response).request_
 *         fut._response = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_7_on_request_completed(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_7_on_request_completed = {"_on_request_completed", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_7_on_request_completed, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_7_on_request_completed(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fut = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_on_request_completed (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fut,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fut)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 442, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_on_request_completed") < 0)) __PYX_ERR(6, 442, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_fut = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_on_request_completed", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 442, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._on_request_completed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_6_on_request_completed(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self), __pyx_v_fut);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_6_on_request_completed(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_fut) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_request_completed", 1);

  /* "asynctnt/iproto/protocol.pyx":443
 * 
 *     def _on_request_completed(self, fut):
 *         cdef BaseRequest req = (<Response> fut._response).request_             # <<<<<<<<<<<<<<
 *         fut._response = None
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_response); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_t_1)->request_);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_req = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":444
 *     def _on_request_completed(self, fut):
 *         cdef BaseRequest req = (<Response> fut._response).request_
 *         fut._response = None             # <<<<<<<<<<<<<<
 * 
 *         if req.timeout_handle is not None:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_fut, __pyx_n_s_response, Py_None) < 0) __PYX_ERR(6, 444, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":446
 *         fut._response = None
 * 
 *         if req.timeout_handle is not None:             # <<<<<<<<<<<<<<
 *             req.timeout_handle.cancel()
 *             req.timeout_handle = None
 */
  __pyx_t_3 = (__pyx_v_req->timeout_handle != Py_None);
  if (__pyx_t_3) {

    /* "asynctnt/iproto/protocol.pyx":447
 * 
 *         if req.timeout_handle is not None:
 *             req.timeout_handle.cancel()             # <<<<<<<<<<<<<<
 *             req.timeout_handle = None
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_req->timeout_handle, __pyx_n_s_cancel); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynctnt/iproto/protocol.pyx":448
 *         if req.timeout_handle is not None:
 *             req.timeout_handle.cancel()
 *             req.timeout_handle = None             # <<<<<<<<<<<<<<
 * 
 *     cdef object _new_waiter_for_request(self, Response response, BaseRequest req, float timeout):
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_req->timeout_handle);
    __Pyx_DECREF(__pyx_v_req->timeout_handle);
    __pyx_v_req->timeout_handle = Py_None;

    /* "asynctnt/iproto/protocol.pyx":446
 *         fut._response = None
 * 
 *         if req.timeout_handle is not None:             # <<<<<<<<<<<<<<
 *             req.timeout_handle.cancel()
 *             req.timeout_handle = None
 */
  }

  /* "asynctnt/iproto/protocol.pyx":442
 *         )
 * 
 *     def _on_request_completed(self, fut):             # <<<<<<<<<<<<<<
 *         cdef BaseRequest req = (<Response> fut._response).request_
 *         fut._response = None
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._on_request_completed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":450
 *             req.timeout_handle = None
 * 
 *     cdef object _new_waiter_for_request(self, Response response, BaseRequest req, float timeout):             # <<<<<<<<<<<<<<
 *         fut = self.create_future()
 *         req.waiter = fut
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__new_waiter_for_request(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_response, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req, float __pyx_v_timeout) {
  PyObject *__pyx_v_fut = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  float __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_new_waiter_for_request", 1);

  /* "asynctnt/iproto/protocol.pyx":451
 * 
 *     cdef object _new_waiter_for_request(self, Response response, BaseRequest req, float timeout):
 *         fut = self.create_future()             # <<<<<<<<<<<<<<
 *         req.waiter = fut
 *         fut._response = response  # to be able to retrieve request after done()
 */
  __Pyx_INCREF(__pyx_v_self->create_future);
  __pyx_t_2 = __pyx_v_self->create_future; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_fut = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":452
 *     cdef object _new_waiter_for_request(self, Response response, BaseRequest req, float timeout):
 *         fut = self.create_future()
 *         req.waiter = fut             # <<<<<<<<<<<<<<
 *         fut._response = response  # to be able to retrieve request after done()
 * 
 */
  __Pyx_INCREF(__pyx_v_fut);
  __Pyx_GIVEREF(__pyx_v_fut);
  __Pyx_GOTREF(__pyx_v_req->waiter);
  __Pyx_DECREF(__pyx_v_req->waiter);
  __pyx_v_req->waiter = __pyx_v_fut;

  /* "asynctnt/iproto/protocol.pyx":453
 *         fut = self.create_future()
 *         req.waiter = fut
 *         fut._response = response  # to be able to retrieve request after done()             # <<<<<<<<<<<<<<
 * 
 *         if timeout < 0:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_fut, __pyx_n_s_response, ((PyObject *)__pyx_v_response)) < 0) __PYX_ERR(6, 453, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":455
 *         fut._response = response  # to be able to retrieve request after done()
 * 
 *         if timeout < 0:             # <<<<<<<<<<<<<<
 *             timeout = self.request_timeout
 *         if timeout is not None and timeout > 0:
 */
  __pyx_t_5 = (__pyx_v_timeout < 0.0);
  if (__pyx_t_5) {

    /* "asynctnt/iproto/protocol.pyx":456
 * 
 *         if timeout < 0:
 *             timeout = self.request_timeout             # <<<<<<<<<<<<<<
 *         if timeout is not None and timeout > 0:
 *             req.timeout_handle = \
 */
    __pyx_t_6 = __pyx_v_self->request_timeout;
    __pyx_v_timeout = __pyx_t_6;

    /* "asynctnt/iproto/protocol.pyx":455
 *         fut._response = response  # to be able to retrieve request after done()
 * 
 *         if timeout < 0:             # <<<<<<<<<<<<<<
 *             timeout = self.request_timeout
 *         if timeout is not None and timeout > 0:
 */
  }

  /* "asynctnt/iproto/protocol.pyx":457
 *         if timeout < 0:
 *             timeout = self.request_timeout
 *         if timeout is not None and timeout > 0:             # <<<<<<<<<<<<<<
 *             req.timeout_handle = \
 *                 self.loop.call_later(timeout,
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_7 = (__pyx_v_timeout > 0.0);
  __pyx_t_5 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_5) {

    /* "asynctnt/iproto/protocol.pyx":459
 *         if timeout is not None and timeout > 0:
 *             req.timeout_handle = \
 *                 self.loop.call_later(timeout,             # <<<<<<<<<<<<<<
 *                                      self._on_request_timeout_cb, fut)
 *             fut.add_done_callback(self._on_request_completed_cb)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->loop, __pyx_n_s_call_later); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyFloat_FromDouble(__pyx_v_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "asynctnt/iproto/protocol.pyx":460
 *             req.timeout_handle = \
 *                 self.loop.call_later(timeout,
 *                                      self._on_request_timeout_cb, fut)             # <<<<<<<<<<<<<<
 *             fut.add_done_callback(self._on_request_completed_cb)
 *         return fut
 */
    __pyx_t_8 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_8, __pyx_t_3, __pyx_v_self->_on_request_timeout_cb, __pyx_v_fut};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 459, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }

    /* "asynctnt/iproto/protocol.pyx":458
 *             timeout = self.request_timeout
 *         if timeout is not None and timeout > 0:
 *             req.timeout_handle = \             # <<<<<<<<<<<<<<
 *                 self.loop.call_later(timeout,
 *                                      self._on_request_timeout_cb, fut)
 */
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_req->timeout_handle);
    __Pyx_DECREF(__pyx_v_req->timeout_handle);
    __pyx_v_req->timeout_handle = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":461
 *                 self.loop.call_later(timeout,
 *                                      self._on_request_timeout_cb, fut)
 *             fut.add_done_callback(self._on_request_completed_cb)             # <<<<<<<<<<<<<<
 *         return fut
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_add_done_callback); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_self->_on_request_completed_cb};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 461, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynctnt/iproto/protocol.pyx":457
 *         if timeout < 0:
 *             timeout = self.request_timeout
 *         if timeout is not None and timeout > 0:             # <<<<<<<<<<<<<<
 *             req.timeout_handle = \
 *                 self.loop.call_later(timeout,
 */
  }

  /* "asynctnt/iproto/protocol.pyx":462
 *                                      self._on_request_timeout_cb, fut)
 *             fut.add_done_callback(self._on_request_completed_cb)
 *         return fut             # <<<<<<<<<<<<<<
 * 
 *     cdef Db _create_db(self, bint gen_stream_id):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fut);
  __pyx_r = __pyx_v_fut;
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":450
 *             req.timeout_handle = None
 * 
 *     cdef object _new_waiter_for_request(self, Response response, BaseRequest req, float timeout):             # <<<<<<<<<<<<<<
 *         fut = self.create_future()
 *         req.waiter = fut
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._new_waiter_for_request", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":464
 *         return fut
 * 
 *     cdef Db _create_db(self, bint gen_stream_id):             # <<<<<<<<<<<<<<
 *         cdef uint64_t stream_id
 *         if gen_stream_id:
 */

static struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__create_db(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, int __pyx_v_gen_stream_id) {
  uint64_t __pyx_v_stream_id;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create_db", 1);

  /* "asynctnt/iproto/protocol.pyx":466
 *     cdef Db _create_db(self, bint gen_stream_id):
 *         cdef uint64_t stream_id
 *         if gen_stream_id:             # <<<<<<<<<<<<<<
 *             stream_id = self.next_stream_id()
 *         else:
 */
  if (__pyx_v_gen_stream_id) {

    /* "asynctnt/iproto/protocol.pyx":467
 *         cdef uint64_t stream_id
 *         if gen_stream_id:
 *             stream_id = self.next_stream_id()             # <<<<<<<<<<<<<<
 *         else:
 *             stream_id = 0
 */
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_next_stream_id(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 467, __pyx_L1_error)
    __pyx_v_stream_id = __pyx_t_1;

    /* "asynctnt/iproto/protocol.pyx":466
 *     cdef Db _create_db(self, bint gen_stream_id):
 *         cdef uint64_t stream_id
 *         if gen_stream_id:             # <<<<<<<<<<<<<<
 *             stream_id = self.next_stream_id()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "asynctnt/iproto/protocol.pyx":469
 *             stream_id = self.next_stream_id()
 *         else:
 *             stream_id = 0             # <<<<<<<<<<<<<<
 * 
 *         return Db.create(self, stream_id)
 */
  /*else*/ {
    __pyx_v_stream_id = 0;
  }
  __pyx_L3:;

  /* "asynctnt/iproto/protocol.pyx":471
 *             stream_id = 0
 * 
 *         return Db.create(self, stream_id)             # <<<<<<<<<<<<<<
 * 
 *     def create_db(self, bint gen_stream_id = False):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_2Db_create(__pyx_v_self, __pyx_v_stream_id)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":464
 *         return fut
 * 
 *     cdef Db _create_db(self, bint gen_stream_id):             # <<<<<<<<<<<<<<
 *         cdef uint64_t stream_id
 *         if gen_stream_id:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._create_db", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":473
 *         return Db.create(self, stream_id)
 * 
 *     def create_db(self, bint gen_stream_id = False):             # <<<<<<<<<<<<<<
 *         return self._create_db(gen_stream_id)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_9create_db(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_9create_db = {"create_db", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_9create_db, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_9create_db(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_gen_stream_id;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_db (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gen_stream_id,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gen_stream_id);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 473, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "create_db") < 0)) __PYX_ERR(6, 473, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_gen_stream_id = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_gen_stream_id == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 473, __pyx_L3_error)
    } else {
      __pyx_v_gen_stream_id = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_db", 0, 0, 1, __pyx_nargs); __PYX_ERR(6, 473, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.create_db", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_8create_db(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self), __pyx_v_gen_stream_id);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_8create_db(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, int __pyx_v_gen_stream_id) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_db", 1);

  /* "asynctnt/iproto/protocol.pyx":474
 * 
 *     def create_db(self, bint gen_stream_id = False):
 *         return self._create_db(gen_stream_id)             # <<<<<<<<<<<<<<
 * 
 *     def get_common_db(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_create_db(__pyx_v_self, __pyx_v_gen_stream_id)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":473
 *         return Db.create(self, stream_id)
 * 
 *     def create_db(self, bint gen_stream_id = False):             # <<<<<<<<<<<<<<
 *         return self._create_db(gen_stream_id)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.create_db", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":476
 *         return self._create_db(gen_stream_id)
 * 
 *     def get_common_db(self):             # <<<<<<<<<<<<<<
 *         return self._db
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_11get_common_db(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_11get_common_db = {"get_common_db", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_11get_common_db, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_11get_common_db(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_common_db (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_common_db", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_common_db", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_10get_common_db(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_10get_common_db(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_common_db", 1);

  /* "asynctnt/iproto/protocol.pyx":477
 * 
 *     def get_common_db(self):
 *         return self._db             # <<<<<<<<<<<<<<
 * 
 *     cdef object _execute_bad(self, BaseRequest req, float timeout):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_db);
  __pyx_r = ((PyObject *)__pyx_v_self->_db);
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":476
 *         return self._create_db(gen_stream_id)
 * 
 *     def get_common_db(self):             # <<<<<<<<<<<<<<
 *         return self._db
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":479
 *         return self._db
 * 
 *     cdef object _execute_bad(self, BaseRequest req, float timeout):             # <<<<<<<<<<<<<<
 *         raise TarantoolNotConnectedError('Tarantool is not connected')
 * 
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__execute_bad(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req, CYTHON_UNUSED float __pyx_v_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_execute_bad", 1);

  /* "asynctnt/iproto/protocol.pyx":480
 * 
 *     cdef object _execute_bad(self, BaseRequest req, float timeout):
 *         raise TarantoolNotConnectedError('Tarantool is not connected')             # <<<<<<<<<<<<<<
 * 
 *     cdef object _execute_normal(self, BaseRequest req, float timeout):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_TarantoolNotConnectedError); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_Tarantool_is_not_connected};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(6, 480, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":479
 *         return self._db
 * 
 *     cdef object _execute_bad(self, BaseRequest req, float timeout):             # <<<<<<<<<<<<<<
 *         raise TarantoolNotConnectedError('Tarantool is not connected')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._execute_bad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":482
 *         raise TarantoolNotConnectedError('Tarantool is not connected')
 * 
 *     cdef object _execute_normal(self, BaseRequest req, float timeout):             # <<<<<<<<<<<<<<
 *         cdef Response response
 *         response = <Response> Response.__new__(Response)
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__execute_normal(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v_req, float __pyx_v_timeout) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_v_response = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_execute_normal", 1);

  /* "asynctnt/iproto/protocol.pyx":484
 *     cdef object _execute_normal(self, BaseRequest req, float timeout):
 *         cdef Response response
 *         response = <Response> Response.__new__(Response)             # <<<<<<<<<<<<<<
 *         response.request_ = req
 *         response.encoding = self.encoding
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Response(((PyTypeObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Response), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 484, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_response = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":485
 *         cdef Response response
 *         response = <Response> Response.__new__(Response)
 *         response.request_ = req             # <<<<<<<<<<<<<<
 *         response.encoding = self.encoding
 *         if req.push_subscribe:
 */
  __Pyx_INCREF((PyObject *)__pyx_v_req);
  __Pyx_GIVEREF((PyObject *)__pyx_v_req);
  __Pyx_GOTREF((PyObject *)__pyx_v_response->request_);
  __Pyx_DECREF((PyObject *)__pyx_v_response->request_);
  __pyx_v_response->request_ = __pyx_v_req;

  /* "asynctnt/iproto/protocol.pyx":486
 *         response = <Response> Response.__new__(Response)
 *         response.request_ = req
 *         response.encoding = self.encoding             # <<<<<<<<<<<<<<
 *         if req.push_subscribe:
 *             response.init_push()
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.encoding;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_response->encoding);
  __Pyx_DECREF(__pyx_v_response->encoding);
  __pyx_v_response->encoding = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":487
 *         response.request_ = req
 *         response.encoding = self.encoding
 *         if req.push_subscribe:             # <<<<<<<<<<<<<<
 *             response.init_push()
 *         cpython.dict.PyDict_SetItem(self._reqs, req.sync, response)
 */
  if (__pyx_v_req->push_subscribe) {

    /* "asynctnt/iproto/protocol.pyx":488
 *         response.encoding = self.encoding
 *         if req.push_subscribe:
 *             response.init_push()             # <<<<<<<<<<<<<<
 *         cpython.dict.PyDict_SetItem(self._reqs, req.sync, response)
 *         self._write(req.encode(self.encoding))
 */
    __pyx_f_8asynctnt_6iproto_8protocol_8Response_init_push(__pyx_v_response); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 488, __pyx_L1_error)

    /* "asynctnt/iproto/protocol.pyx":487
 *         response.request_ = req
 *         response.encoding = self.encoding
 *         if req.push_subscribe:             # <<<<<<<<<<<<<<
 *             response.init_push()
 *         cpython.dict.PyDict_SetItem(self._reqs, req.sync, response)
 */
  }

  /* "asynctnt/iproto/protocol.pyx":489
 *         if req.push_subscribe:
 *             response.init_push()
 *         cpython.dict.PyDict_SetItem(self._reqs, req.sync, response)             # <<<<<<<<<<<<<<
 *         self._write(req.encode(self.encoding))
 * 
 */
  __pyx_t_2 = __pyx_v_self->_reqs;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_req->sync); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyDict_SetItem(__pyx_t_2, __pyx_t_1, ((PyObject *)__pyx_v_response)); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(6, 489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynctnt/iproto/protocol.pyx":490
 *             response.init_push()
 *         cpython.dict.PyDict_SetItem(self._reqs, req.sync, response)
 *         self._write(req.encode(self.encoding))             # <<<<<<<<<<<<<<
 * 
 *         return self._new_waiter_for_request(response, req, timeout)
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.encoding;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_encode(__pyx_v_req, ((PyObject*)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._write(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v_self), __pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":492
 *         self._write(req.encode(self.encoding))
 * 
 *         return self._new_waiter_for_request(response, req, timeout)             # <<<<<<<<<<<<<<
 * 
 *     cdef uint32_t transform_iterator(self, iterator) except *:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_new_waiter_for_request(__pyx_v_self, __pyx_v_response, __pyx_v_req, __pyx_v_timeout); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":482
 *         raise TarantoolNotConnectedError('Tarantool is not connected')
 * 
 *     cdef object _execute_normal(self, BaseRequest req, float timeout):             # <<<<<<<<<<<<<<
 *         cdef Response response
 *         response = <Response> Response.__new__(Response)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._execute_normal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_response);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":494
 *         return self._new_waiter_for_request(response, req, timeout)
 * 
 *     cdef uint32_t transform_iterator(self, iterator) except *:             # <<<<<<<<<<<<<<
 *         if isinstance(iterator, int):
 *             return iterator
 */

static uint32_t __pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_transform_iterator(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, PyObject *__pyx_v_iterator) {
  uint32_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  uint32_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("transform_iterator", 1);

  /* "asynctnt/iproto/protocol.pyx":495
 * 
 *     cdef uint32_t transform_iterator(self, iterator) except *:
 *         if isinstance(iterator, int):             # <<<<<<<<<<<<<<
 *             return iterator
 *         if isinstance(iterator, Iterator):
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_iterator); 
  if (__pyx_t_1) {

    /* "asynctnt/iproto/protocol.pyx":496
 *     cdef uint32_t transform_iterator(self, iterator) except *:
 *         if isinstance(iterator, int):
 *             return iterator             # <<<<<<<<<<<<<<
 *         if isinstance(iterator, Iterator):
 *             return iterator.value
 */
    __pyx_t_2 = __Pyx_PyInt_As_uint32_t(__pyx_v_iterator); if (unlikely((__pyx_t_2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(6, 496, __pyx_L1_error)
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":495
 * 
 *     cdef uint32_t transform_iterator(self, iterator) except *:
 *         if isinstance(iterator, int):             # <<<<<<<<<<<<<<
 *             return iterator
 *         if isinstance(iterator, Iterator):
 */
  }

  /* "asynctnt/iproto/protocol.pyx":497
 *         if isinstance(iterator, int):
 *             return iterator
 *         if isinstance(iterator, Iterator):             # <<<<<<<<<<<<<<
 *             return iterator.value
 *         if isinstance(iterator, str):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Iterator); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_iterator, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(6, 497, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "asynctnt/iproto/protocol.pyx":498
 *             return iterator
 *         if isinstance(iterator, Iterator):
 *             return iterator.value             # <<<<<<<<<<<<<<
 *         if isinstance(iterator, str):
 *             return Iterator[iterator]
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_iterator, __pyx_n_s_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 498, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(6, 498, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":497
 *         if isinstance(iterator, int):
 *             return iterator
 *         if isinstance(iterator, Iterator):             # <<<<<<<<<<<<<<
 *             return iterator.value
 *         if isinstance(iterator, str):
 */
  }

  /* "asynctnt/iproto/protocol.pyx":499
 *         if isinstance(iterator, Iterator):
 *             return iterator.value
 *         if isinstance(iterator, str):             # <<<<<<<<<<<<<<
 *             return Iterator[iterator]
 *         else:
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_iterator); 
  if (likely(__pyx_t_1)) {

    /* "asynctnt/iproto/protocol.pyx":500
 *             return iterator.value
 *         if isinstance(iterator, str):
 *             return Iterator[iterator]             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Iterator is of unsupported type '
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Iterator); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_v_iterator); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyInt_As_uint32_t(__pyx_t_4); if (unlikely((__pyx_t_2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(6, 500, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    goto __pyx_L0;

    /* "asynctnt/iproto/protocol.pyx":499
 *         if isinstance(iterator, Iterator):
 *             return iterator.value
 *         if isinstance(iterator, str):             # <<<<<<<<<<<<<<
 *             return Iterator[iterator]
 *         else:
 */
  }

  /* "asynctnt/iproto/protocol.pyx":502
 *             return Iterator[iterator]
 *         else:
 *             raise TypeError('Iterator is of unsupported type '             # <<<<<<<<<<<<<<
 *                             '(asynctnt.Iterator, int, str)')
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(6, 502, __pyx_L1_error)
  }

  /* "asynctnt/iproto/protocol.pyx":494
 *         return self._new_waiter_for_request(response, req, timeout)
 * 
 *     cdef uint32_t transform_iterator(self, iterator) except *:             # <<<<<<<<<<<<<<
 *         if isinstance(iterator, int):
 *             return iterator
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.transform_iterator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":505
 *                             '(asynctnt.Iterator, int, str)')
 * 
 *     cdef object _refetch_schema(self):             # <<<<<<<<<<<<<<
 *         if self._refetch_schema_future is not None and not self._refetch_schema_future.done():
 *             self._schema_fetch_in_progress = False
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__refetch_schema(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_refetch_schema", 1);

  /* "asynctnt/iproto/protocol.pyx":506
 * 
 *     cdef object _refetch_schema(self):
 *         if self._refetch_schema_future is not None and not self._refetch_schema_future.done():             # <<<<<<<<<<<<<<
 *             self._schema_fetch_in_progress = False
 *             self._refetch_schema_future.cancel()
 */
  __pyx_t_2 = (__pyx_v_self->_refetch_schema_future != Py_None);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_refetch_schema_future, __pyx_n_s_done); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 506, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "asynctnt/iproto/protocol.pyx":507
 *     cdef object _refetch_schema(self):
 *         if self._refetch_schema_future is not None and not self._refetch_schema_future.done():
 *             self._schema_fetch_in_progress = False             # <<<<<<<<<<<<<<
 *             self._refetch_schema_future.cancel()
 * 
 */
    __pyx_v_self->_schema_fetch_in_progress = 0;

    /* "asynctnt/iproto/protocol.pyx":508
 *         if self._refetch_schema_future is not None and not self._refetch_schema_future.done():
 *             self._schema_fetch_in_progress = False
 *             self._refetch_schema_future.cancel()             # <<<<<<<<<<<<<<
 * 
 *         self._refetch_schema_future = self.create_future()
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_refetch_schema_future, __pyx_n_s_cancel); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 508, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "asynctnt/iproto/protocol.pyx":506
 * 
 *     cdef object _refetch_schema(self):
 *         if self._refetch_schema_future is not None and not self._refetch_schema_future.done():             # <<<<<<<<<<<<<<
 *             self._schema_fetch_in_progress = False
 *             self._refetch_schema_future.cancel()
 */
  }

  /* "asynctnt/iproto/protocol.pyx":510
 *             self._refetch_schema_future.cancel()
 * 
 *         self._refetch_schema_future = self.create_future()             # <<<<<<<<<<<<<<
 *         self._do_fetch_schema(self._refetch_schema_future)
 * 
 */
  __Pyx_INCREF(__pyx_v_self->create_future);
  __pyx_t_4 = __pyx_v_self->create_future; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_refetch_schema_future);
  __Pyx_DECREF(__pyx_v_self->_refetch_schema_future);
  __pyx_v_self->_refetch_schema_future = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "asynctnt/iproto/protocol.pyx":511
 * 
 *         self._refetch_schema_future = self.create_future()
 *         self._do_fetch_schema(self._refetch_schema_future)             # <<<<<<<<<<<<<<
 * 
 *         return self._refetch_schema_future
 */
  __pyx_t_3 = __pyx_v_self->_refetch_schema_future;
  __Pyx_INCREF(__pyx_t_3);
  ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_do_fetch_schema(__pyx_v_self, __pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/protocol.pyx":513
 *         self._do_fetch_schema(self._refetch_schema_future)
 * 
 *         return self._refetch_schema_future             # <<<<<<<<<<<<<<
 * 
 *     def refetch_schema(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_refetch_schema_future);
  __pyx_r = __pyx_v_self->_refetch_schema_future;
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":505
 *                             '(asynctnt.Iterator, int, str)')
 * 
 *     cdef object _refetch_schema(self):             # <<<<<<<<<<<<<<
 *         if self._refetch_schema_future is not None and not self._refetch_schema_future.done():
 *             self._schema_fetch_in_progress = False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol._refetch_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynctnt/iproto/protocol.pyx":515
 *         return self._refetch_schema_future
 * 
 *     def refetch_schema(self):             # <<<<<<<<<<<<<<
 *         return self._refetch_schema()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_13refetch_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_13refetch_schema = {"refetch_schema", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_13refetch_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_13refetch_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("refetch_schema (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("refetch_schema", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "refetch_schema", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_12refetch_schema(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_12refetch_schema(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("refetch_schema", 1);

  /* "asynctnt/iproto/protocol.pyx":516
 * 
 *     def refetch_schema(self):
 *         return self._refetch_schema()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self->__pyx_base.__pyx_vtab)->_refetch_schema(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynctnt/iproto/protocol.pyx":515
 *         return self._refetch_schema_future
 * 
 *     def refetch_schema(self):             # <<<<<<<<<<<<<<
 *         return self._refetch_schema()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.refetch_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_15__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_14__reduce_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_execute_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_17__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_16__setstate_cython__(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12BaseProtocol_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_execute_cannot_be_converted, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynctnt.iproto.protocol.BaseProtocol.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_BaseRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_1__pyx_unpickle_BaseRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_1__pyx_unpickle_BaseRequest = {"__pyx_unpickle_BaseRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_1__pyx_unpickle_BaseRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_1__pyx_unpickle_BaseRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_BaseRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BaseRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BaseRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_BaseRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BaseRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_BaseRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol___pyx_unpickle_BaseRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol___pyx_unpickle_BaseRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BaseRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__49, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = BaseRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = BaseRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = BaseRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BaseRequest__set_state(<BaseRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = BaseRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BaseRequest__set_state(<BaseRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = BaseRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BaseRequest__set_state(<BaseRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_BaseRequest__set_state(BaseRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_BaseRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = BaseRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BaseRequest__set_state(<BaseRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BaseRequest__set_state(<BaseRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_BaseRequest__set_state(BaseRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BaseRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_BaseRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_BaseRequest__set_state(<BaseRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BaseRequest__set_state(BaseRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_BaseRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BaseRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_BaseRequest__set_state(BaseRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->space);
  __pyx_v___pyx_result->space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->timeout_handle);
  __pyx_v___pyx_result->timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->waiter);
  __pyx_v___pyx_result->waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_BaseRequest__set_state(BaseRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 11);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_BaseRequest__set_state(BaseRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_BaseRequest__set_state(<BaseRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BaseRequest__set_state(BaseRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_BaseRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PingRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_3__pyx_unpickle_PingRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_3__pyx_unpickle_PingRequest = {"__pyx_unpickle_PingRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_3__pyx_unpickle_PingRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_3__pyx_unpickle_PingRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PingRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PingRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PingRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PingRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PingRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_PingRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_2__pyx_unpickle_PingRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_2__pyx_unpickle_PingRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PingRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__49, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = PingRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PingRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = PingRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PingRequest__set_state(<PingRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = PingRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PingRequest__set_state(<PingRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PingRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PingRequest__set_state(<PingRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PingRequest__set_state(PingRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PingRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = PingRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PingRequest__set_state(<PingRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PingRequest__set_state(<PingRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PingRequest__set_state(PingRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PingRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_PingRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PingRequest__set_state(<PingRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PingRequest__set_state(PingRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PingRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PingRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PingRequest__set_state(PingRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PingRequest__set_state(PingRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 11);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PingRequest__set_state(PingRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PingRequest__set_state(<PingRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PingRequest__set_state(PingRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_PingRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CallRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5__pyx_unpickle_CallRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_5__pyx_unpickle_CallRequest = {"__pyx_unpickle_CallRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_5__pyx_unpickle_CallRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_5__pyx_unpickle_CallRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CallRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CallRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CallRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_CallRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CallRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_CallRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_4__pyx_unpickle_CallRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_4__pyx_unpickle_CallRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CallRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x2d5a81d, 0xac90343, 0xcd9f35b):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2d5a81d, 0xac90343, 0xcd9f35b) = (args, check_schema_change, func_name, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__50, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x2d5a81d, 0xac90343, 0xcd9f35b):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2d5a81d, 0xac90343, 0xcd9f35b) = (args, check_schema_change, func_name, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = CallRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x2d5a81d, 0xac90343, 0xcd9f35b):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2d5a81d, 0xac90343, 0xcd9f35b) = (args, check_schema_change, func_name, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = CallRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x2d5a81d, 0xac90343, 0xcd9f35b):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2d5a81d, 0xac90343, 0xcd9f35b) = (args, check_schema_change, func_name, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2d5a81d, 0xac90343, 0xcd9f35b) = (args, check_schema_change, func_name, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = CallRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CallRequest__set_state(<CallRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2d5a81d, 0xac90343, 0xcd9f35b) = (args, check_schema_change, func_name, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = CallRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CallRequest__set_state(<CallRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = CallRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CallRequest__set_state(<CallRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CallRequest__set_state(CallRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CallRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x2d5a81d, 0xac90343, 0xcd9f35b) = (args, check_schema_change, func_name, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = CallRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CallRequest__set_state(<CallRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CallRequest__set_state(<CallRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CallRequest__set_state(CallRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.func_name = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CallRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_CallRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CallRequest__set_state(<CallRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CallRequest__set_state(CallRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.func_name = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CallRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CallRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CallRequest__set_state(CallRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.func_name = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->args);
  __Pyx_DECREF(__pyx_v___pyx_result->args);
  __pyx_v___pyx_result->args = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->func_name);
  __Pyx_DECREF(__pyx_v___pyx_result->func_name);
  __pyx_v___pyx_result->func_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_CallRequest__set_state(CallRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.func_name = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 13);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.func_name = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CallRequest__set_state(CallRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.func_name = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CallRequest__set_state(<CallRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CallRequest__set_state(CallRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.func_name = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_CallRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_EvalRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_7__pyx_unpickle_EvalRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_7__pyx_unpickle_EvalRequest = {"__pyx_unpickle_EvalRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_7__pyx_unpickle_EvalRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_7__pyx_unpickle_EvalRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_EvalRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EvalRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EvalRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_EvalRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EvalRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_EvalRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_6__pyx_unpickle_EvalRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_6__pyx_unpickle_EvalRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_EvalRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xafafa67, 0x3c16c97, 0x904102b):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xafafa67, 0x3c16c97, 0x904102b) = (args, check_schema_change, expression, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__51, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xafafa67, 0x3c16c97, 0x904102b):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xafafa67, 0x3c16c97, 0x904102b) = (args, check_schema_change, expression, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = EvalRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xafafa67, 0x3c16c97, 0x904102b):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xafafa67, 0x3c16c97, 0x904102b) = (args, check_schema_change, expression, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = EvalRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xafafa67, 0x3c16c97, 0x904102b):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xafafa67, 0x3c16c97, 0x904102b) = (args, check_schema_change, expression, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xafafa67, 0x3c16c97, 0x904102b) = (args, check_schema_change, expression, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = EvalRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EvalRequest__set_state(<EvalRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xafafa67, 0x3c16c97, 0x904102b) = (args, check_schema_change, expression, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = EvalRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_EvalRequest__set_state(<EvalRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = EvalRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EvalRequest__set_state(<EvalRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_EvalRequest__set_state(EvalRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_EvalRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xafafa67, 0x3c16c97, 0x904102b) = (args, check_schema_change, expression, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = EvalRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_EvalRequest__set_state(<EvalRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EvalRequest__set_state(<EvalRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_EvalRequest__set_state(EvalRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.expression = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_EvalRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_EvalRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_EvalRequest__set_state(<EvalRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EvalRequest__set_state(EvalRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.expression = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_EvalRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_EvalRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_EvalRequest__set_state(EvalRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.expression = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->args);
  __Pyx_DECREF(__pyx_v___pyx_result->args);
  __pyx_v___pyx_result->args = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->expression);
  __Pyx_DECREF(__pyx_v___pyx_result->expression);
  __pyx_v___pyx_result->expression = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_EvalRequest__set_state(EvalRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.expression = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 13);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.expression = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_EvalRequest__set_state(EvalRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.expression = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_EvalRequest__set_state(<EvalRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EvalRequest__set_state(EvalRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.expression = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_EvalRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_SelectRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_9__pyx_unpickle_SelectRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_9__pyx_unpickle_SelectRequest = {"__pyx_unpickle_SelectRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_9__pyx_unpickle_SelectRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_9__pyx_unpickle_SelectRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_SelectRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SelectRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SelectRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_SelectRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SelectRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_SelectRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_8__pyx_unpickle_SelectRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_8__pyx_unpickle_SelectRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_SelectRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8e2dc2b, 0x62f76f3, 0x57fe0b4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8e2dc2b, 0x62f76f3, 0x57fe0b4) = (check_schema_change, index, iterator, key, limit, offset, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__52, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8e2dc2b, 0x62f76f3, 0x57fe0b4):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8e2dc2b, 0x62f76f3, 0x57fe0b4) = (check_schema_change, index, iterator, key, limit, offset, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = SelectRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x8e2dc2b, 0x62f76f3, 0x57fe0b4):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8e2dc2b, 0x62f76f3, 0x57fe0b4) = (check_schema_change, index, iterator, key, limit, offset, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = SelectRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8e2dc2b, 0x62f76f3, 0x57fe0b4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8e2dc2b, 0x62f76f3, 0x57fe0b4) = (check_schema_change, index, iterator, key, limit, offset, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8e2dc2b, 0x62f76f3, 0x57fe0b4) = (check_schema_change, index, iterator, key, limit, offset, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = SelectRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SelectRequest__set_state(<SelectRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8e2dc2b, 0x62f76f3, 0x57fe0b4) = (check_schema_change, index, iterator, key, limit, offset, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = SelectRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SelectRequest__set_state(<SelectRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = SelectRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SelectRequest__set_state(<SelectRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_SelectRequest__set_state(SelectRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_SelectRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8e2dc2b, 0x62f76f3, 0x57fe0b4) = (check_schema_change, index, iterator, key, limit, offset, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = SelectRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SelectRequest__set_state(<SelectRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SelectRequest__set_state(<SelectRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_SelectRequest__set_state(SelectRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.iterator = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.limit = __pyx_state[4]; __pyx_result.offset = __pyx_state[5]; __pyx_result.op = __pyx_state[6]; __pyx_result.parse_as_tuples = __pyx_state[7]; __pyx_result.parse_metadata = __pyx_state[8]; __pyx_result.push_subscribe = __pyx_state[9]; __pyx_result.schema_id = __pyx_state[10]; __pyx_result.space = __pyx_state[11]; __pyx_result.stream_id = __pyx_state[12]; __pyx_result.sync = __pyx_state[13]; __pyx_result.timeout_handle = __pyx_state[14]; __pyx_result.waiter = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_SelectRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_SelectRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_SelectRequest__set_state(<SelectRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SelectRequest__set_state(SelectRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.iterator = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.limit = __pyx_state[4]; __pyx_result.offset = __pyx_state[5]; __pyx_result.op = __pyx_state[6]; __pyx_result.parse_as_tuples = __pyx_state[7]; __pyx_result.parse_metadata = __pyx_state[8]; __pyx_result.push_subscribe = __pyx_state[9]; __pyx_result.schema_id = __pyx_state[10]; __pyx_result.space = __pyx_state[11]; __pyx_result.stream_id = __pyx_state[12]; __pyx_result.sync = __pyx_state[13]; __pyx_result.timeout_handle = __pyx_state[14]; __pyx_result.waiter = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_SelectRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  uint32_t __pyx_t_3;
  uint64_t __pyx_t_4;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_5;
  int64_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_SelectRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_SelectRequest__set_state(SelectRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.iterator = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.limit = __pyx_state[4]; __pyx_result.offset = __pyx_state[5]; __pyx_result.op = __pyx_state[6]; __pyx_result.parse_as_tuples = __pyx_state[7]; __pyx_result.parse_metadata = __pyx_state[8]; __pyx_result.push_subscribe = __pyx_state[9]; __pyx_result.schema_id = __pyx_state[10]; __pyx_result.space = __pyx_state[11]; __pyx_result.stream_id = __pyx_state[12]; __pyx_result.sync = __pyx_state[13]; __pyx_result.timeout_handle = __pyx_state[14]; __pyx_result.waiter = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->index);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->index);
  __pyx_v___pyx_result->index = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_uint32_t(__pyx_t_1); if (unlikely((__pyx_t_3 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->iterator = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->key);
  __Pyx_DECREF(__pyx_v___pyx_result->key);
  __pyx_v___pyx_result->key = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->limit = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->offset = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_SelectRequest__set_state(SelectRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.iterator = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.limit = __pyx_state[4]; __pyx_result.offset = __pyx_state[5]; __pyx_result.op = __pyx_state[6]; __pyx_result.parse_as_tuples = __pyx_state[7]; __pyx_result.parse_metadata = __pyx_state[8]; __pyx_result.push_subscribe = __pyx_state[9]; __pyx_result.schema_id = __pyx_state[10]; __pyx_result.space = __pyx_state[11]; __pyx_result.stream_id = __pyx_state[12]; __pyx_result.sync = __pyx_state[13]; __pyx_result.timeout_handle = __pyx_state[14]; __pyx_result.waiter = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 > 16);
  if (__pyx_t_8) {
  } else {
    __pyx_t_2 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.iterator = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.limit = __pyx_state[4]; __pyx_result.offset = __pyx_state[5]; __pyx_result.op = __pyx_state[6]; __pyx_result.parse_as_tuples = __pyx_state[7]; __pyx_result.parse_metadata = __pyx_state[8]; __pyx_result.push_subscribe = __pyx_state[9]; __pyx_result.schema_id = __pyx_state[10]; __pyx_result.space = __pyx_state[11]; __pyx_result.stream_id = __pyx_state[12]; __pyx_result.sync = __pyx_state[13]; __pyx_result.timeout_handle = __pyx_state[14]; __pyx_result.waiter = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_update); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_11 = NULL;
    __pyx_t_12 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_12 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_9};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_SelectRequest__set_state(SelectRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.iterator = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.limit = __pyx_state[4]; __pyx_result.offset = __pyx_state[5]; __pyx_result.op = __pyx_state[6]; __pyx_result.parse_as_tuples = __pyx_state[7]; __pyx_result.parse_metadata = __pyx_state[8]; __pyx_result.push_subscribe = __pyx_state[9]; __pyx_result.schema_id = __pyx_state[10]; __pyx_result.space = __pyx_state[11]; __pyx_result.stream_id = __pyx_state[12]; __pyx_result.sync = __pyx_state[13]; __pyx_result.timeout_handle = __pyx_state[14]; __pyx_result.waiter = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_SelectRequest__set_state(<SelectRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SelectRequest__set_state(SelectRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.iterator = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.limit = __pyx_state[4]; __pyx_result.offset = __pyx_state[5]; __pyx_result.op = __pyx_state[6]; __pyx_result.parse_as_tuples = __pyx_state[7]; __pyx_result.parse_metadata = __pyx_state[8]; __pyx_result.push_subscribe = __pyx_state[9]; __pyx_result.schema_id = __pyx_state[10]; __pyx_result.space = __pyx_state[11]; __pyx_result.stream_id = __pyx_state[12]; __pyx_result.sync = __pyx_state[13]; __pyx_result.timeout_handle = __pyx_state[14]; __pyx_result.waiter = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_SelectRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_InsertRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11__pyx_unpickle_InsertRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_11__pyx_unpickle_InsertRequest = {"__pyx_unpickle_InsertRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11__pyx_unpickle_InsertRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_11__pyx_unpickle_InsertRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_InsertRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InsertRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InsertRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_InsertRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_InsertRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_InsertRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_10__pyx_unpickle_InsertRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_10__pyx_unpickle_InsertRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InsertRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb91d1f2, 0xff182e4, 0x15f8d7d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb91d1f2, 0xff182e4, 0x15f8d7d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__53, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb91d1f2, 0xff182e4, 0x15f8d7d):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb91d1f2, 0xff182e4, 0x15f8d7d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = InsertRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xb91d1f2, 0xff182e4, 0x15f8d7d):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb91d1f2, 0xff182e4, 0x15f8d7d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = InsertRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb91d1f2, 0xff182e4, 0x15f8d7d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb91d1f2, 0xff182e4, 0x15f8d7d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb91d1f2, 0xff182e4, 0x15f8d7d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = InsertRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InsertRequest__set_state(<InsertRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb91d1f2, 0xff182e4, 0x15f8d7d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = InsertRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InsertRequest__set_state(<InsertRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = InsertRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InsertRequest__set_state(<InsertRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_InsertRequest__set_state(InsertRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_InsertRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xb91d1f2, 0xff182e4, 0x15f8d7d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = InsertRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_InsertRequest__set_state(<InsertRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_InsertRequest__set_state(<InsertRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_InsertRequest__set_state(InsertRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.t = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.waiter = __pyx_state[11]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_InsertRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_InsertRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_InsertRequest__set_state(<InsertRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InsertRequest__set_state(InsertRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.t = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.waiter = __pyx_state[11]
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_InsertRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_InsertRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_InsertRequest__set_state(InsertRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.t = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.waiter = __pyx_state[11]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[12])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->t);
  __Pyx_DECREF(__pyx_v___pyx_result->t);
  __pyx_v___pyx_result->t = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_InsertRequest__set_state(InsertRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.t = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.waiter = __pyx_state[11]
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[12])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 12);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.t = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.waiter = __pyx_state[11]
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[12])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_InsertRequest__set_state(InsertRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.t = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.waiter = __pyx_state[11]
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[12])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_InsertRequest__set_state(<InsertRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_InsertRequest__set_state(InsertRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.t = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.waiter = __pyx_state[11]
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_InsertRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_DeleteRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13__pyx_unpickle_DeleteRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_13__pyx_unpickle_DeleteRequest = {"__pyx_unpickle_DeleteRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13__pyx_unpickle_DeleteRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_13__pyx_unpickle_DeleteRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_DeleteRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DeleteRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DeleteRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_DeleteRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DeleteRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_DeleteRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_12__pyx_unpickle_DeleteRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_12__pyx_unpickle_DeleteRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DeleteRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf383ff7, 0x8ac4f18, 0xe9836c0):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf383ff7, 0x8ac4f18, 0xe9836c0) = (check_schema_change, index, key, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__54, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf383ff7, 0x8ac4f18, 0xe9836c0):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf383ff7, 0x8ac4f18, 0xe9836c0) = (check_schema_change, index, key, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = DeleteRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xf383ff7, 0x8ac4f18, 0xe9836c0):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf383ff7, 0x8ac4f18, 0xe9836c0) = (check_schema_change, index, key, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = DeleteRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf383ff7, 0x8ac4f18, 0xe9836c0):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf383ff7, 0x8ac4f18, 0xe9836c0) = (check_schema_change, index, key, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf383ff7, 0x8ac4f18, 0xe9836c0) = (check_schema_change, index, key, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = DeleteRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DeleteRequest__set_state(<DeleteRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf383ff7, 0x8ac4f18, 0xe9836c0) = (check_schema_change, index, key, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = DeleteRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DeleteRequest__set_state(<DeleteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = DeleteRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DeleteRequest__set_state(<DeleteRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_DeleteRequest__set_state(DeleteRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_DeleteRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf383ff7, 0x8ac4f18, 0xe9836c0) = (check_schema_change, index, key, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = DeleteRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DeleteRequest__set_state(<DeleteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DeleteRequest__set_state(<DeleteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_DeleteRequest__set_state(DeleteRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_DeleteRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_DeleteRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_DeleteRequest__set_state(<DeleteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DeleteRequest__set_state(DeleteRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_DeleteRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DeleteRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_DeleteRequest__set_state(DeleteRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->index);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->index);
  __pyx_v___pyx_result->index = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->key);
  __Pyx_DECREF(__pyx_v___pyx_result->key);
  __pyx_v___pyx_result->key = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_DeleteRequest__set_state(DeleteRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 13);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_DeleteRequest__set_state(DeleteRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_DeleteRequest__set_state(<DeleteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DeleteRequest__set_state(DeleteRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_DeleteRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_UpdateRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_15__pyx_unpickle_UpdateRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_15__pyx_unpickle_UpdateRequest = {"__pyx_unpickle_UpdateRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_15__pyx_unpickle_UpdateRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_15__pyx_unpickle_UpdateRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_UpdateRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UpdateRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UpdateRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_UpdateRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UpdateRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_UpdateRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_14__pyx_unpickle_UpdateRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_14__pyx_unpickle_UpdateRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UpdateRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x815e48a, 0x8a9c73b, 0x180001a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x815e48a, 0x8a9c73b, 0x180001a) = (check_schema_change, index, key, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__55, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x815e48a, 0x8a9c73b, 0x180001a):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x815e48a, 0x8a9c73b, 0x180001a) = (check_schema_change, index, key, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = UpdateRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x815e48a, 0x8a9c73b, 0x180001a):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x815e48a, 0x8a9c73b, 0x180001a) = (check_schema_change, index, key, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = UpdateRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x815e48a, 0x8a9c73b, 0x180001a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x815e48a, 0x8a9c73b, 0x180001a) = (check_schema_change, index, key, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x815e48a, 0x8a9c73b, 0x180001a) = (check_schema_change, index, key, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = UpdateRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UpdateRequest__set_state(<UpdateRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x815e48a, 0x8a9c73b, 0x180001a) = (check_schema_change, index, key, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = UpdateRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UpdateRequest__set_state(<UpdateRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = UpdateRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UpdateRequest__set_state(<UpdateRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_UpdateRequest__set_state(UpdateRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_UpdateRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x815e48a, 0x8a9c73b, 0x180001a) = (check_schema_change, index, key, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = UpdateRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UpdateRequest__set_state(<UpdateRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UpdateRequest__set_state(<UpdateRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_UpdateRequest__set_state(UpdateRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.operations = __pyx_state[4]; __pyx_result.parse_as_tuples = __pyx_state[5]; __pyx_result.parse_metadata = __pyx_state[6]; __pyx_result.push_subscribe = __pyx_state[7]; __pyx_result.schema_id = __pyx_state[8]; __pyx_result.space = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_UpdateRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_UpdateRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_UpdateRequest__set_state(<UpdateRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UpdateRequest__set_state(UpdateRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.operations = __pyx_state[4]; __pyx_result.parse_as_tuples = __pyx_state[5]; __pyx_result.parse_metadata = __pyx_state[6]; __pyx_result.push_subscribe = __pyx_state[7]; __pyx_result.schema_id = __pyx_state[8]; __pyx_result.space = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_UpdateRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UpdateRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_UpdateRequest__set_state(UpdateRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.operations = __pyx_state[4]; __pyx_result.parse_as_tuples = __pyx_state[5]; __pyx_result.parse_metadata = __pyx_state[6]; __pyx_result.push_subscribe = __pyx_state[7]; __pyx_result.schema_id = __pyx_state[8]; __pyx_result.space = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->index);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->index);
  __pyx_v___pyx_result->index = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->key);
  __Pyx_DECREF(__pyx_v___pyx_result->key);
  __pyx_v___pyx_result->key = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->operations);
  __Pyx_DECREF(__pyx_v___pyx_result->operations);
  __pyx_v___pyx_result->operations = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_UpdateRequest__set_state(UpdateRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.operations = __pyx_state[4]; __pyx_result.parse_as_tuples = __pyx_state[5]; __pyx_result.parse_metadata = __pyx_state[6]; __pyx_result.push_subscribe = __pyx_state[7]; __pyx_result.schema_id = __pyx_state[8]; __pyx_result.space = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 14);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.operations = __pyx_state[4]; __pyx_result.parse_as_tuples = __pyx_state[5]; __pyx_result.parse_metadata = __pyx_state[6]; __pyx_result.push_subscribe = __pyx_state[7]; __pyx_result.schema_id = __pyx_state[8]; __pyx_result.space = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_UpdateRequest__set_state(UpdateRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.operations = __pyx_state[4]; __pyx_result.parse_as_tuples = __pyx_state[5]; __pyx_result.parse_metadata = __pyx_state[6]; __pyx_result.push_subscribe = __pyx_state[7]; __pyx_result.schema_id = __pyx_state[8]; __pyx_result.space = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_UpdateRequest__set_state(<UpdateRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UpdateRequest__set_state(UpdateRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.operations = __pyx_state[4]; __pyx_result.parse_as_tuples = __pyx_state[5]; __pyx_result.parse_metadata = __pyx_state[6]; __pyx_result.push_subscribe = __pyx_state[7]; __pyx_result.schema_id = __pyx_state[8]; __pyx_result.space = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_UpdateRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_UpsertRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_17__pyx_unpickle_UpsertRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_17__pyx_unpickle_UpsertRequest = {"__pyx_unpickle_UpsertRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_17__pyx_unpickle_UpsertRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_17__pyx_unpickle_UpsertRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_UpsertRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UpsertRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UpsertRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_UpsertRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UpsertRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_UpsertRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_16__pyx_unpickle_UpsertRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_16__pyx_unpickle_UpsertRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UpsertRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc823323, 0xea6e875, 0x0493034):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc823323, 0xea6e875, 0x0493034) = (check_schema_change, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__56, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc823323, 0xea6e875, 0x0493034):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc823323, 0xea6e875, 0x0493034) = (check_schema_change, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = UpsertRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc823323, 0xea6e875, 0x0493034):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc823323, 0xea6e875, 0x0493034) = (check_schema_change, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = UpsertRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc823323, 0xea6e875, 0x0493034):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc823323, 0xea6e875, 0x0493034) = (check_schema_change, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc823323, 0xea6e875, 0x0493034) = (check_schema_change, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = UpsertRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UpsertRequest__set_state(<UpsertRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc823323, 0xea6e875, 0x0493034) = (check_schema_change, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = UpsertRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UpsertRequest__set_state(<UpsertRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = UpsertRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UpsertRequest__set_state(<UpsertRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_UpsertRequest__set_state(UpsertRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_UpsertRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc823323, 0xea6e875, 0x0493034) = (check_schema_change, op, operations, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, t, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = UpsertRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UpsertRequest__set_state(<UpsertRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UpsertRequest__set_state(<UpsertRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_UpsertRequest__set_state(UpsertRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.operations = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.t = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_UpsertRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_UpsertRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_UpsertRequest__set_state(<UpsertRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UpsertRequest__set_state(UpsertRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.operations = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.t = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_UpsertRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UpsertRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_UpsertRequest__set_state(UpsertRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.operations = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.t = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->operations);
  __Pyx_DECREF(__pyx_v___pyx_result->operations);
  __pyx_v___pyx_result->operations = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->t);
  __Pyx_DECREF(__pyx_v___pyx_result->t);
  __pyx_v___pyx_result->t = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_UpsertRequest__set_state(UpsertRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.operations = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.t = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 13);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.operations = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.t = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_UpsertRequest__set_state(UpsertRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.operations = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.t = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_UpsertRequest__set_state(<UpsertRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UpsertRequest__set_state(UpsertRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.operations = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.t = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_UpsertRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PrepareRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_19__pyx_unpickle_PrepareRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_19__pyx_unpickle_PrepareRequest = {"__pyx_unpickle_PrepareRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_19__pyx_unpickle_PrepareRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_19__pyx_unpickle_PrepareRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PrepareRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PrepareRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PrepareRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PrepareRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PrepareRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_PrepareRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_18__pyx_unpickle_PrepareRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_18__pyx_unpickle_PrepareRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PrepareRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd0934a3, 0xcc899a0, 0x7b3178d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd0934a3, 0xcc899a0, 0x7b3178d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__57, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd0934a3, 0xcc899a0, 0x7b3178d):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd0934a3, 0xcc899a0, 0x7b3178d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = PrepareRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xd0934a3, 0xcc899a0, 0x7b3178d):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd0934a3, 0xcc899a0, 0x7b3178d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PrepareRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd0934a3, 0xcc899a0, 0x7b3178d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd0934a3, 0xcc899a0, 0x7b3178d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd0934a3, 0xcc899a0, 0x7b3178d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = PrepareRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PrepareRequest__set_state(<PrepareRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd0934a3, 0xcc899a0, 0x7b3178d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = PrepareRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PrepareRequest__set_state(<PrepareRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PrepareRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PrepareRequest__set_state(<PrepareRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PrepareRequest__set_state(PrepareRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PrepareRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd0934a3, 0xcc899a0, 0x7b3178d) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = PrepareRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PrepareRequest__set_state(<PrepareRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PrepareRequest__set_state(<PrepareRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PrepareRequest__set_state(PrepareRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.query = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.statement_id = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PrepareRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_PrepareRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PrepareRequest__set_state(<PrepareRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PrepareRequest__set_state(PrepareRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.query = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.statement_id = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PrepareRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PrepareRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PrepareRequest__set_state(PrepareRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.query = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.statement_id = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->query);
  __Pyx_DECREF(__pyx_v___pyx_result->query);
  __pyx_v___pyx_result->query = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->statement_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PrepareRequest__set_state(PrepareRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.query = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.statement_id = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 13);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.query = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.statement_id = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PrepareRequest__set_state(PrepareRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.query = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.statement_id = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PrepareRequest__set_state(<PrepareRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PrepareRequest__set_state(PrepareRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.query = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.statement_id = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_PrepareRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ExecuteRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21__pyx_unpickle_ExecuteRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_21__pyx_unpickle_ExecuteRequest = {"__pyx_unpickle_ExecuteRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_21__pyx_unpickle_ExecuteRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_21__pyx_unpickle_ExecuteRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ExecuteRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ExecuteRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ExecuteRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_ExecuteRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ExecuteRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_ExecuteRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_20__pyx_unpickle_ExecuteRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_20__pyx_unpickle_ExecuteRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ExecuteRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x314ca88, 0x64ec1ad, 0xd228fa0):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x314ca88, 0x64ec1ad, 0xd228fa0) = (args, check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__58, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x314ca88, 0x64ec1ad, 0xd228fa0):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x314ca88, 0x64ec1ad, 0xd228fa0) = (args, check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = ExecuteRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x314ca88, 0x64ec1ad, 0xd228fa0):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x314ca88, 0x64ec1ad, 0xd228fa0) = (args, check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = ExecuteRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x314ca88, 0x64ec1ad, 0xd228fa0):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x314ca88, 0x64ec1ad, 0xd228fa0) = (args, check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x314ca88, 0x64ec1ad, 0xd228fa0) = (args, check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = ExecuteRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ExecuteRequest__set_state(<ExecuteRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x314ca88, 0x64ec1ad, 0xd228fa0) = (args, check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = ExecuteRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ExecuteRequest__set_state(<ExecuteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = ExecuteRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ExecuteRequest__set_state(<ExecuteRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ExecuteRequest__set_state(ExecuteRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_ExecuteRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x314ca88, 0x64ec1ad, 0xd228fa0) = (args, check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, query, schema_id, space, statement_id, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = ExecuteRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ExecuteRequest__set_state(<ExecuteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ExecuteRequest__set_state(<ExecuteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ExecuteRequest__set_state(ExecuteRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.query = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.statement_id = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ExecuteRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_ExecuteRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ExecuteRequest__set_state(<ExecuteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ExecuteRequest__set_state(ExecuteRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.query = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.statement_id = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_ExecuteRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ExecuteRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ExecuteRequest__set_state(ExecuteRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.query = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.statement_id = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->args);
  __Pyx_DECREF(__pyx_v___pyx_result->args);
  __pyx_v___pyx_result->args = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->query);
  __Pyx_DECREF(__pyx_v___pyx_result->query);
  __pyx_v___pyx_result->query = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->statement_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_ExecuteRequest__set_state(ExecuteRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.query = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.statement_id = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 14);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.query = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.statement_id = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ExecuteRequest__set_state(ExecuteRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.query = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.statement_id = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ExecuteRequest__set_state(<ExecuteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ExecuteRequest__set_state(ExecuteRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.query = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.statement_id = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_ExecuteRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IDRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_23__pyx_unpickle_IDRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_23__pyx_unpickle_IDRequest = {"__pyx_unpickle_IDRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_23__pyx_unpickle_IDRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_23__pyx_unpickle_IDRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_IDRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IDRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IDRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_IDRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IDRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_IDRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_22__pyx_unpickle_IDRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_22__pyx_unpickle_IDRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IDRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__49, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = IDRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = IDRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = IDRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IDRequest__set_state(<IDRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = IDRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IDRequest__set_state(<IDRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = IDRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IDRequest__set_state(<IDRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IDRequest__set_state(IDRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IDRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = IDRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IDRequest__set_state(<IDRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IDRequest__set_state(<IDRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IDRequest__set_state(IDRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IDRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_IDRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IDRequest__set_state(<IDRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IDRequest__set_state(IDRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IDRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IDRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IDRequest__set_state(IDRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IDRequest__set_state(IDRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 11);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IDRequest__set_state(IDRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IDRequest__set_state(<IDRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IDRequest__set_state(IDRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_IDRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_AuthRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_25__pyx_unpickle_AuthRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_25__pyx_unpickle_AuthRequest = {"__pyx_unpickle_AuthRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_25__pyx_unpickle_AuthRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_25__pyx_unpickle_AuthRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_AuthRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AuthRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AuthRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_AuthRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AuthRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_AuthRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_24__pyx_unpickle_AuthRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_24__pyx_unpickle_AuthRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_AuthRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x335e105, 0xf5a1799, 0xa2d5d5d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x335e105, 0xf5a1799, 0xa2d5d5d) = (check_schema_change, op, parse_as_tuples, parse_metadata, password, push_subscribe, salt, schema_id, space, stream_id, sync, timeout_handle, username, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__59, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x335e105, 0xf5a1799, 0xa2d5d5d):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x335e105, 0xf5a1799, 0xa2d5d5d) = (check_schema_change, op, parse_as_tuples, parse_metadata, password, push_subscribe, salt, schema_id, space, stream_id, sync, timeout_handle, username, waiter))" % __pyx_checksum
 *     __pyx_result = AuthRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x335e105, 0xf5a1799, 0xa2d5d5d):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x335e105, 0xf5a1799, 0xa2d5d5d) = (check_schema_change, op, parse_as_tuples, parse_metadata, password, push_subscribe, salt, schema_id, space, stream_id, sync, timeout_handle, username, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = AuthRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x335e105, 0xf5a1799, 0xa2d5d5d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x335e105, 0xf5a1799, 0xa2d5d5d) = (check_schema_change, op, parse_as_tuples, parse_metadata, password, push_subscribe, salt, schema_id, space, stream_id, sync, timeout_handle, username, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x335e105, 0xf5a1799, 0xa2d5d5d) = (check_schema_change, op, parse_as_tuples, parse_metadata, password, push_subscribe, salt, schema_id, space, stream_id, sync, timeout_handle, username, waiter))" % __pyx_checksum
 *     __pyx_result = AuthRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AuthRequest__set_state(<AuthRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x335e105, 0xf5a1799, 0xa2d5d5d) = (check_schema_change, op, parse_as_tuples, parse_metadata, password, push_subscribe, salt, schema_id, space, stream_id, sync, timeout_handle, username, waiter))" % __pyx_checksum
 *     __pyx_result = AuthRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_AuthRequest__set_state(<AuthRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = AuthRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AuthRequest__set_state(<AuthRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_AuthRequest__set_state(AuthRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_AuthRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x335e105, 0xf5a1799, 0xa2d5d5d) = (check_schema_change, op, parse_as_tuples, parse_metadata, password, push_subscribe, salt, schema_id, space, stream_id, sync, timeout_handle, username, waiter))" % __pyx_checksum
 *     __pyx_result = AuthRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_AuthRequest__set_state(<AuthRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AuthRequest__set_state(<AuthRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_AuthRequest__set_state(AuthRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.password = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.salt = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.username = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_AuthRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_AuthRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_AuthRequest__set_state(<AuthRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_AuthRequest__set_state(AuthRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.password = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.salt = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.username = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_AuthRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_AuthRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_AuthRequest__set_state(AuthRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.password = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.salt = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.username = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->password);
  __Pyx_DECREF(__pyx_v___pyx_result->password);
  __pyx_v___pyx_result->password = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->salt);
  __Pyx_DECREF(__pyx_v___pyx_result->salt);
  __pyx_v___pyx_result->salt = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->username);
  __Pyx_DECREF(__pyx_v___pyx_result->username);
  __pyx_v___pyx_result->username = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_AuthRequest__set_state(AuthRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.password = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.salt = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.username = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 14);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.password = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.salt = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.username = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_AuthRequest__set_state(AuthRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.password = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.salt = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.username = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_AuthRequest__set_state(<AuthRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_AuthRequest__set_state(AuthRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.password = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.salt = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.username = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_AuthRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_BeginRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_27__pyx_unpickle_BeginRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_27__pyx_unpickle_BeginRequest = {"__pyx_unpickle_BeginRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_27__pyx_unpickle_BeginRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_27__pyx_unpickle_BeginRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_BeginRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BeginRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BeginRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_BeginRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BeginRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_BeginRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_26__pyx_unpickle_BeginRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_26__pyx_unpickle_BeginRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BeginRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x30f95df, 0xa4f125c, 0xfc952a6):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x30f95df, 0xa4f125c, 0xfc952a6) = (check_schema_change, isolation, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, tx_timeout, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__60, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x30f95df, 0xa4f125c, 0xfc952a6):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x30f95df, 0xa4f125c, 0xfc952a6) = (check_schema_change, isolation, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, tx_timeout, waiter))" % __pyx_checksum
 *     __pyx_result = BeginRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x30f95df, 0xa4f125c, 0xfc952a6):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x30f95df, 0xa4f125c, 0xfc952a6) = (check_schema_change, isolation, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, tx_timeout, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = BeginRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x30f95df, 0xa4f125c, 0xfc952a6):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x30f95df, 0xa4f125c, 0xfc952a6) = (check_schema_change, isolation, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, tx_timeout, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x30f95df, 0xa4f125c, 0xfc952a6) = (check_schema_change, isolation, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, tx_timeout, waiter))" % __pyx_checksum
 *     __pyx_result = BeginRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BeginRequest__set_state(<BeginRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x30f95df, 0xa4f125c, 0xfc952a6) = (check_schema_change, isolation, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, tx_timeout, waiter))" % __pyx_checksum
 *     __pyx_result = BeginRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BeginRequest__set_state(<BeginRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = BeginRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BeginRequest__set_state(<BeginRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_BeginRequest__set_state(BeginRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_BeginRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x30f95df, 0xa4f125c, 0xfc952a6) = (check_schema_change, isolation, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, tx_timeout, waiter))" % __pyx_checksum
 *     __pyx_result = BeginRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BeginRequest__set_state(<BeginRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BeginRequest__set_state(<BeginRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_BeginRequest__set_state(BeginRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.isolation = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.tx_timeout = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BeginRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_BeginRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_BeginRequest__set_state(<BeginRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BeginRequest__set_state(BeginRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.isolation = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.tx_timeout = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_BeginRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  uint32_t __pyx_t_3;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_4;
  int64_t __pyx_t_5;
  uint64_t __pyx_t_6;
  double __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BeginRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_BeginRequest__set_state(BeginRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.isolation = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.tx_timeout = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_uint32_t(__pyx_t_1); if (unlikely((__pyx_t_3 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->isolation = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->tx_timeout = __pyx_t_7;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_BeginRequest__set_state(BeginRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.isolation = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.tx_timeout = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 > 13);
  if (__pyx_t_9) {
  } else {
    __pyx_t_2 = __pyx_t_9;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_9 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.isolation = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.tx_timeout = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    __pyx_t_13 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
        __pyx_t_13 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_10};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_13, 1+__pyx_t_13);
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_BeginRequest__set_state(BeginRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.isolation = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.tx_timeout = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_BeginRequest__set_state(<BeginRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BeginRequest__set_state(BeginRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.isolation = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.timeout_handle = __pyx_state[10]; __pyx_result.tx_timeout = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_BeginRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CommitRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_29__pyx_unpickle_CommitRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_29__pyx_unpickle_CommitRequest = {"__pyx_unpickle_CommitRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_29__pyx_unpickle_CommitRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_29__pyx_unpickle_CommitRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CommitRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CommitRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CommitRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_CommitRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CommitRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_CommitRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_28__pyx_unpickle_CommitRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_28__pyx_unpickle_CommitRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CommitRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__49, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = CommitRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = CommitRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = CommitRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CommitRequest__set_state(<CommitRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = CommitRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CommitRequest__set_state(<CommitRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = CommitRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CommitRequest__set_state(<CommitRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CommitRequest__set_state(CommitRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CommitRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = CommitRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CommitRequest__set_state(<CommitRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CommitRequest__set_state(<CommitRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CommitRequest__set_state(CommitRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CommitRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_CommitRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CommitRequest__set_state(<CommitRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CommitRequest__set_state(CommitRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CommitRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CommitRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CommitRequest__set_state(CommitRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_CommitRequest__set_state(CommitRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 11);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CommitRequest__set_state(CommitRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CommitRequest__set_state(<CommitRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CommitRequest__set_state(CommitRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_CommitRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_RollbackRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_31__pyx_unpickle_RollbackRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_31__pyx_unpickle_RollbackRequest = {"__pyx_unpickle_RollbackRequest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_31__pyx_unpickle_RollbackRequest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_31__pyx_unpickle_RollbackRequest(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_RollbackRequest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RollbackRequest", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RollbackRequest", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_RollbackRequest") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RollbackRequest", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_RollbackRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_30__pyx_unpickle_RollbackRequest(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_30__pyx_unpickle_RollbackRequest(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_RollbackRequest", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__49, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = RollbackRequest.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = RollbackRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = RollbackRequest.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RollbackRequest__set_state(<RollbackRequest> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = RollbackRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RollbackRequest__set_state(<RollbackRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = RollbackRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RollbackRequest__set_state(<RollbackRequest> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_RollbackRequest__set_state(RollbackRequest __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_RollbackRequest__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 *     __pyx_result = RollbackRequest.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RollbackRequest__set_state(<RollbackRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RollbackRequest__set_state(<RollbackRequest> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_RollbackRequest__set_state(RollbackRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_RollbackRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_RollbackRequest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_RollbackRequest__set_state(<RollbackRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RollbackRequest__set_state(RollbackRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_RollbackRequest__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __pyx_t_3;
  int64_t __pyx_t_4;
  uint64_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_RollbackRequest__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_RollbackRequest__set_state(RollbackRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.check_schema_change = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.op = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_as_tuples = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parse_metadata = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.push_subscribe = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.schema_id = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.space);
  __pyx_v___pyx_result->__pyx_base.space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.stream_id = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.sync = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.timeout_handle);
  __pyx_v___pyx_result->__pyx_base.timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.waiter);
  __pyx_v___pyx_result->__pyx_base.waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_RollbackRequest__set_state(RollbackRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 11);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_RollbackRequest__set_state(RollbackRequest __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_RollbackRequest__set_state(<RollbackRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RollbackRequest__set_state(RollbackRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_RollbackRequest__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IProtoErrorStackFrame(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_33__pyx_unpickle_IProtoErrorStackFrame(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_33__pyx_unpickle_IProtoErrorStackFrame = {"__pyx_unpickle_IProtoErrorStackFrame", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_33__pyx_unpickle_IProtoErrorStackFrame, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_33__pyx_unpickle_IProtoErrorStackFrame(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_IProtoErrorStackFrame (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IProtoErrorStackFrame", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IProtoErrorStackFrame", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_IProtoErrorStackFrame") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IProtoErrorStackFrame", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_IProtoErrorStackFrame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_32__pyx_unpickle_IProtoErrorStackFrame(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_32__pyx_unpickle_IProtoErrorStackFrame(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IProtoErrorStackFrame", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xab276bf, 0xb7ed9c7, 0x6438658):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xab276bf, 0xb7ed9c7, 0x6438658) = (code, err_no, error_type, fields, file, line, message))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__61, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xab276bf, 0xb7ed9c7, 0x6438658):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xab276bf, 0xb7ed9c7, 0x6438658) = (code, err_no, error_type, fields, file, line, message))" % __pyx_checksum
 *     __pyx_result = IProtoErrorStackFrame.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xab276bf, 0xb7ed9c7, 0x6438658):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xab276bf, 0xb7ed9c7, 0x6438658) = (code, err_no, error_type, fields, file, line, message))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = IProtoErrorStackFrame.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xab276bf, 0xb7ed9c7, 0x6438658):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xab276bf, 0xb7ed9c7, 0x6438658) = (code, err_no, error_type, fields, file, line, message))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xab276bf, 0xb7ed9c7, 0x6438658) = (code, err_no, error_type, fields, file, line, message))" % __pyx_checksum
 *     __pyx_result = IProtoErrorStackFrame.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IProtoErrorStackFrame__set_state(<IProtoErrorStackFrame> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xab276bf, 0xb7ed9c7, 0x6438658) = (code, err_no, error_type, fields, file, line, message))" % __pyx_checksum
 *     __pyx_result = IProtoErrorStackFrame.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IProtoErrorStackFrame__set_state(<IProtoErrorStackFrame> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = IProtoErrorStackFrame.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IProtoErrorStackFrame__set_state(<IProtoErrorStackFrame> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IProtoErrorStackFrame__set_state(IProtoErrorStackFrame __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IProtoErrorStackFrame__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xab276bf, 0xb7ed9c7, 0x6438658) = (code, err_no, error_type, fields, file, line, message))" % __pyx_checksum
 *     __pyx_result = IProtoErrorStackFrame.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IProtoErrorStackFrame__set_state(<IProtoErrorStackFrame> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IProtoErrorStackFrame__set_state(<IProtoErrorStackFrame> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IProtoErrorStackFrame__set_state(IProtoErrorStackFrame __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.err_no = __pyx_state[1]; __pyx_result.error_type = __pyx_state[2]; __pyx_result.fields = __pyx_state[3]; __pyx_result.file = __pyx_state[4]; __pyx_result.line = __pyx_state[5]; __pyx_result.message = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IProtoErrorStackFrame(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_IProtoErrorStackFrame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IProtoErrorStackFrame__set_state(<IProtoErrorStackFrame> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IProtoErrorStackFrame__set_state(IProtoErrorStackFrame __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.err_no = __pyx_state[1]; __pyx_result.error_type = __pyx_state[2]; __pyx_result.fields = __pyx_state[3]; __pyx_result.file = __pyx_state[4]; __pyx_result.line = __pyx_state[5]; __pyx_result.message = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IProtoErrorStackFrame__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IProtoErrorStackFrame__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IProtoErrorStackFrame__set_state(IProtoErrorStackFrame __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.err_no = __pyx_state[1]; __pyx_result.error_type = __pyx_state[2]; __pyx_result.fields = __pyx_state[3]; __pyx_result.file = __pyx_state[4]; __pyx_result.line = __pyx_state[5]; __pyx_result.message = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->code = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->err_no = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->error_type);
  __Pyx_DECREF(__pyx_v___pyx_result->error_type);
  __pyx_v___pyx_result->error_type = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->fields);
  __Pyx_DECREF(__pyx_v___pyx_result->fields);
  __pyx_v___pyx_result->fields = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->file);
  __Pyx_DECREF(__pyx_v___pyx_result->file);
  __pyx_v___pyx_result->file = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->line = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->message);
  __Pyx_DECREF(__pyx_v___pyx_result->message);
  __pyx_v___pyx_result->message = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IProtoErrorStackFrame__set_state(IProtoErrorStackFrame __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.err_no = __pyx_state[1]; __pyx_result.error_type = __pyx_state[2]; __pyx_result.fields = __pyx_state[3]; __pyx_result.file = __pyx_state[4]; __pyx_result.line = __pyx_state[5]; __pyx_result.message = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 7);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.err_no = __pyx_state[1]; __pyx_result.error_type = __pyx_state[2]; __pyx_result.fields = __pyx_state[3]; __pyx_result.file = __pyx_state[4]; __pyx_result.line = __pyx_state[5]; __pyx_result.message = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_2 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_2 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_2, 1+__pyx_t_2);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IProtoErrorStackFrame__set_state(IProtoErrorStackFrame __pyx_result, tuple __pyx_state):
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.err_no = __pyx_state[1]; __pyx_result.error_type = __pyx_state[2]; __pyx_result.fields = __pyx_state[3]; __pyx_result.file = __pyx_state[4]; __pyx_result.line = __pyx_state[5]; __pyx_result.message = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IProtoErrorStackFrame__set_state(<IProtoErrorStackFrame> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IProtoErrorStackFrame__set_state(IProtoErrorStackFrame __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.err_no = __pyx_state[1]; __pyx_result.error_type = __pyx_state[2]; __pyx_result.fields = __pyx_state[3]; __pyx_result.file = __pyx_state[4]; __pyx_result.line = __pyx_state[5]; __pyx_result.message = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_IProtoErrorStackFrame__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IProtoError(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_35__pyx_unpickle_IProtoError(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_35__pyx_unpickle_IProtoError = {"__pyx_unpickle_IProtoError", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_35__pyx_unpickle_IProtoError, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_35__pyx_unpickle_IProtoError(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_IProtoError (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IProtoError", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IProtoError", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_IProtoError") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IProtoError", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_IProtoError", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_34__pyx_unpickle_IProtoError(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_34__pyx_unpickle_IProtoError(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IProtoError", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xeafe895, 0xd0fa554, 0x04a7503):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xeafe895, 0xd0fa554, 0x04a7503) = (trace))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__62, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xeafe895, 0xd0fa554, 0x04a7503):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xeafe895, 0xd0fa554, 0x04a7503) = (trace))" % __pyx_checksum
 *     __pyx_result = IProtoError.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xeafe895, 0xd0fa554, 0x04a7503):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xeafe895, 0xd0fa554, 0x04a7503) = (trace))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = IProtoError.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xeafe895, 0xd0fa554, 0x04a7503):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xeafe895, 0xd0fa554, 0x04a7503) = (trace))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xeafe895, 0xd0fa554, 0x04a7503) = (trace))" % __pyx_checksum
 *     __pyx_result = IProtoError.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IProtoError__set_state(<IProtoError> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xeafe895, 0xd0fa554, 0x04a7503) = (trace))" % __pyx_checksum
 *     __pyx_result = IProtoError.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IProtoError__set_state(<IProtoError> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = IProtoError.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IProtoError__set_state(<IProtoError> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IProtoError__set_state(IProtoError __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IProtoError__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xeafe895, 0xd0fa554, 0x04a7503) = (trace))" % __pyx_checksum
 *     __pyx_result = IProtoError.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IProtoError__set_state(<IProtoError> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IProtoError__set_state(<IProtoError> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IProtoError__set_state(IProtoError __pyx_result, tuple __pyx_state):
 *     __pyx_result.trace = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IProtoError(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_IProtoError", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IProtoError__set_state(<IProtoError> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IProtoError__set_state(IProtoError __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.trace = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_IProtoError__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IProtoError__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IProtoError__set_state(IProtoError __pyx_result, tuple __pyx_state):
 *     __pyx_result.trace = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->trace);
  __Pyx_DECREF(__pyx_v___pyx_result->trace);
  __pyx_v___pyx_result->trace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IProtoError__set_state(IProtoError __pyx_result, tuple __pyx_state):
 *     __pyx_result.trace = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.trace = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IProtoError__set_state(IProtoError __pyx_result, tuple __pyx_state):
 *     __pyx_result.trace = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IProtoError__set_state(<IProtoError> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IProtoError__set_state(IProtoError __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.trace = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_IProtoError__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PushIterator(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_37__pyx_unpickle_PushIterator(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_37__pyx_unpickle_PushIterator = {"__pyx_unpickle_PushIterator", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_37__pyx_unpickle_PushIterator, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_37__pyx_unpickle_PushIterator(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PushIterator (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PushIterator", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PushIterator", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PushIterator") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PushIterator", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_PushIterator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_36__pyx_unpickle_PushIterator(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_36__pyx_unpickle_PushIterator(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PushIterator", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x98137af, 0xbb11453, 0xd6d3964):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x98137af, 0xbb11453, 0xd6d3964) = (_fut, _request, _response))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__63, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x98137af, 0xbb11453, 0xd6d3964):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x98137af, 0xbb11453, 0xd6d3964) = (_fut, _request, _response))" % __pyx_checksum
 *     __pyx_result = PushIterator.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x98137af, 0xbb11453, 0xd6d3964):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x98137af, 0xbb11453, 0xd6d3964) = (_fut, _request, _response))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PushIterator.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x98137af, 0xbb11453, 0xd6d3964):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x98137af, 0xbb11453, 0xd6d3964) = (_fut, _request, _response))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x98137af, 0xbb11453, 0xd6d3964) = (_fut, _request, _response))" % __pyx_checksum
 *     __pyx_result = PushIterator.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PushIterator__set_state(<PushIterator> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x98137af, 0xbb11453, 0xd6d3964) = (_fut, _request, _response))" % __pyx_checksum
 *     __pyx_result = PushIterator.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PushIterator__set_state(<PushIterator> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PushIterator.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PushIterator__set_state(<PushIterator> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PushIterator__set_state(PushIterator __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PushIterator__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x98137af, 0xbb11453, 0xd6d3964) = (_fut, _request, _response))" % __pyx_checksum
 *     __pyx_result = PushIterator.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PushIterator__set_state(<PushIterator> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PushIterator__set_state(<PushIterator> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PushIterator__set_state(PushIterator __pyx_result, tuple __pyx_state):
 *     __pyx_result._fut = __pyx_state[0]; __pyx_result._request = __pyx_state[1]; __pyx_result._response = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PushIterator(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_PushIterator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PushIterator__set_state(<PushIterator> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PushIterator__set_state(PushIterator __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._fut = __pyx_state[0]; __pyx_result._request = __pyx_state[1]; __pyx_result._response = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_PushIterator__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PushIterator__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PushIterator__set_state(PushIterator __pyx_result, tuple __pyx_state):
 *     __pyx_result._fut = __pyx_state[0]; __pyx_result._request = __pyx_state[1]; __pyx_result._response = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_fut);
  __Pyx_DECREF(__pyx_v___pyx_result->_fut);
  __pyx_v___pyx_result->_fut = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->_request);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->_request);
  __pyx_v___pyx_result->_request = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_Response))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->_response);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->_response);
  __pyx_v___pyx_result->_response = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PushIterator__set_state(PushIterator __pyx_result, tuple __pyx_state):
 *     __pyx_result._fut = __pyx_state[0]; __pyx_result._request = __pyx_state[1]; __pyx_result._response = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._fut = __pyx_state[0]; __pyx_result._request = __pyx_state[1]; __pyx_result._response = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PushIterator__set_state(PushIterator __pyx_result, tuple __pyx_state):
 *     __pyx_result._fut = __pyx_state[0]; __pyx_result._request = __pyx_state[1]; __pyx_result._response = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PushIterator__set_state(<PushIterator> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PushIterator__set_state(PushIterator __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._fut = __pyx_state[0]; __pyx_result._request = __pyx_state[1]; __pyx_result._response = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_PushIterator__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CoreProtocol(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_39__pyx_unpickle_CoreProtocol(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_8asynctnt_6iproto_8protocol_39__pyx_unpickle_CoreProtocol = {"__pyx_unpickle_CoreProtocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_39__pyx_unpickle_CoreProtocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8asynctnt_6iproto_8protocol_39__pyx_unpickle_CoreProtocol(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CoreProtocol (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CoreProtocol", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CoreProtocol", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_CoreProtocol") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CoreProtocol", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_CoreProtocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8asynctnt_6iproto_8protocol_38__pyx_unpickle_CoreProtocol(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8asynctnt_6iproto_8protocol_38__pyx_unpickle_CoreProtocol(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CoreProtocol", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x89d360e, 0xe0a6eac, 0xf188771):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x89d360e, 0xe0a6eac, 0xf188771) = (con_state, encoding, host, port, rbuf, salt, state, version))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__64, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x89d360e, 0xe0a6eac, 0xf188771):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x89d360e, 0xe0a6eac, 0xf188771) = (con_state, encoding, host, port, rbuf, salt, state, version))" % __pyx_checksum
 *     __pyx_result = CoreProtocol.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x89d360e, 0xe0a6eac, 0xf188771):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x89d360e, 0xe0a6eac, 0xf188771) = (con_state, encoding, host, port, rbuf, salt, state, version))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = CoreProtocol.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x89d360e, 0xe0a6eac, 0xf188771):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x89d360e, 0xe0a6eac, 0xf188771) = (con_state, encoding, host, port, rbuf, salt, state, version))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x89d360e, 0xe0a6eac, 0xf188771) = (con_state, encoding, host, port, rbuf, salt, state, version))" % __pyx_checksum
 *     __pyx_result = CoreProtocol.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CoreProtocol__set_state(<CoreProtocol> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x89d360e, 0xe0a6eac, 0xf188771) = (con_state, encoding, host, port, rbuf, salt, state, version))" % __pyx_checksum
 *     __pyx_result = CoreProtocol.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CoreProtocol__set_state(<CoreProtocol> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = CoreProtocol.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CoreProtocol__set_state(<CoreProtocol> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CoreProtocol__set_state(CoreProtocol __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CoreProtocol__set_state(((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x89d360e, 0xe0a6eac, 0xf188771) = (con_state, encoding, host, port, rbuf, salt, state, version))" % __pyx_checksum
 *     __pyx_result = CoreProtocol.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CoreProtocol__set_state(<CoreProtocol> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CoreProtocol__set_state(<CoreProtocol> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CoreProtocol__set_state(CoreProtocol __pyx_result, tuple __pyx_state):
 *     __pyx_result.con_state = __pyx_state[0]; __pyx_result.encoding = __pyx_state[1]; __pyx_result.host = __pyx_state[2]; __pyx_result.port = __pyx_state[3]; __pyx_result.rbuf = __pyx_state[4]; __pyx_result.salt = __pyx_state[5]; __pyx_result.state = __pyx_state[6]; __pyx_result.version = __pyx_state[7]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CoreProtocol(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_CoreProtocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CoreProtocol__set_state(<CoreProtocol> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CoreProtocol__set_state(CoreProtocol __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.con_state = __pyx_state[0]; __pyx_result.encoding = __pyx_state[1]; __pyx_result.host = __pyx_state[2]; __pyx_result.port = __pyx_state[3]; __pyx_result.rbuf = __pyx_state[4]; __pyx_result.salt = __pyx_state[5]; __pyx_result.state = __pyx_state[6]; __pyx_result.version = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_8asynctnt_6iproto_8protocol___pyx_unpickle_CoreProtocol__set_state(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState __pyx_t_2;
  enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CoreProtocol__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CoreProtocol__set_state(CoreProtocol __pyx_result, tuple __pyx_state):
 *     __pyx_result.con_state = __pyx_state[0]; __pyx_result.encoding = __pyx_state[1]; __pyx_result.host = __pyx_state[2]; __pyx_result.port = __pyx_state[3]; __pyx_result.rbuf = __pyx_state[4]; __pyx_result.salt = __pyx_state[5]; __pyx_result.state = __pyx_state[6]; __pyx_result.version = __pyx_state[7]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_8protocol_ConnectionState(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->con_state = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->encoding);
  __Pyx_DECREF(__pyx_v___pyx_result->encoding);
  __pyx_v___pyx_result->encoding = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->host);
  __Pyx_DECREF(__pyx_v___pyx_result->host);
  __pyx_v___pyx_result->host = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->port);
  __Pyx_DECREF(__pyx_v___pyx_result->port);
  __pyx_v___pyx_result->port = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->rbuf);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->rbuf);
  __pyx_v___pyx_result->rbuf = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->salt);
  __Pyx_DECREF(__pyx_v___pyx_result->salt);
  __pyx_v___pyx_result->salt = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)__Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_8protocol_ProtocolState(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->state = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->version);
  __Pyx_DECREF(__pyx_v___pyx_result->version);
  __pyx_v___pyx_result->version = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_CoreProtocol__set_state(CoreProtocol __pyx_result, tuple __pyx_state):
 *     __pyx_result.con_state = __pyx_state[0]; __pyx_result.encoding = __pyx_state[1]; __pyx_result.host = __pyx_state[2]; __pyx_result.port = __pyx_state[3]; __pyx_result.rbuf = __pyx_state[4]; __pyx_result.salt = __pyx_state[5]; __pyx_result.state = __pyx_state[6]; __pyx_result.version = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 > 8);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.con_state = __pyx_state[0]; __pyx_result.encoding = __pyx_state[1]; __pyx_result.host = __pyx_state[2]; __pyx_result.port = __pyx_state[3]; __pyx_result.rbuf = __pyx_state[4]; __pyx_result.salt = __pyx_state[5]; __pyx_result.state = __pyx_state[6]; __pyx_result.version = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CoreProtocol__set_state(CoreProtocol __pyx_result, tuple __pyx_state):
 *     __pyx_result.con_state = __pyx_state[0]; __pyx_result.encoding = __pyx_state[1]; __pyx_result.host = __pyx_state[2]; __pyx_result.port = __pyx_state[3]; __pyx_result.rbuf = __pyx_state[4]; __pyx_result.salt = __pyx_state[5]; __pyx_result.state = __pyx_state[6]; __pyx_result.version = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CoreProtocol__set_state(<CoreProtocol> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CoreProtocol__set_state(CoreProtocol __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.con_state = __pyx_state[0]; __pyx_result.encoding = __pyx_state[1]; __pyx_result.host = __pyx_state[2]; __pyx_result.port = __pyx_state[3]; __pyx_result.rbuf = __pyx_state[4]; __pyx_result.salt = __pyx_state[5]; __pyx_result.state = __pyx_state[6]; __pyx_result.version = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynctnt.iproto.protocol.__pyx_unpickle_CoreProtocol__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_Field(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)o);
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->type = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->collation = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->is_nullable = Py_None; Py_INCREF(Py_None);
  p->is_autoincrement = Py_None; Py_INCREF(Py_None);
  p->span = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8asynctnt_6iproto_8protocol_5Field_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Field(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->type);
  Py_CLEAR(p->collation);
  Py_CLEAR(p->is_nullable);
  Py_CLEAR(p->is_autoincrement);
  Py_CLEAR(p->span);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_Field(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)o;
  if (p->is_nullable) {
    e = (*v)(p->is_nullable, a); if (e) return e;
  }
  if (p->is_autoincrement) {
    e = (*v)(p->is_autoincrement, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_Field(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *)o;
  tmp = ((PyObject*)p->is_nullable);
  p->is_nullable = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->is_autoincrement);
  p->is_autoincrement = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_5Field_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_5Field_4name_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_5Field_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_5Field_4type_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_5Field_collation(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_5Field_9collation_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_5Field_is_nullable(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_5Field_11is_nullable_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_5Field_is_autoincrement(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_5Field_16is_autoincrement_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_5Field_span(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_5Field_4span_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_5Field_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_5Field_3__repr__(self);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_Field[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_5Field_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_5Field_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_5Field_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8asynctnt_6iproto_8protocol_Field[] = {
  {(char *)"name", __pyx_getprop_8asynctnt_6iproto_8protocol_5Field_name, 0, (char *)0, 0},
  {(char *)"type", __pyx_getprop_8asynctnt_6iproto_8protocol_5Field_type, 0, (char *)0, 0},
  {(char *)"collation", __pyx_getprop_8asynctnt_6iproto_8protocol_5Field_collation, 0, (char *)0, 0},
  {(char *)"is_nullable", __pyx_getprop_8asynctnt_6iproto_8protocol_5Field_is_nullable, 0, (char *)0, 0},
  {(char *)"is_autoincrement", __pyx_getprop_8asynctnt_6iproto_8protocol_5Field_is_autoincrement, 0, (char *)0, 0},
  {(char *)"span", __pyx_getprop_8asynctnt_6iproto_8protocol_5Field_span, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_Field_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Field},
  {Py_tp_repr, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_5Field_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_Field},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_Field},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_Field},
  {Py_tp_getset, (void *)__pyx_getsets_8asynctnt_6iproto_8protocol_Field},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Field},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_Field_spec = {
  "asynctnt.iproto.protocol.Field",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_Field_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_Field = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""Field", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_Field), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Field, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_5Field_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_Field, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_Field, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_Field, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8asynctnt_6iproto_8protocol_Field, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_Field, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Metadata __pyx_vtable_8asynctnt_6iproto_8protocol_Metadata;

static struct C_Metadata *__pyx_freelist_8asynctnt_6iproto_8protocol_Metadata[128];
static int __pyx_freecount_8asynctnt_6iproto_8protocol_Metadata = 0;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_Metadata(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct C_Metadata *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_8asynctnt_6iproto_8protocol_Metadata > 0) & (int)(t->tp_basicsize == sizeof(struct C_Metadata)))) {
    o = (PyObject*)__pyx_freelist_8asynctnt_6iproto_8protocol_Metadata[--__pyx_freecount_8asynctnt_6iproto_8protocol_Metadata];
    memset(o, 0, sizeof(struct C_Metadata));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct C_Metadata *)o);
  p->__pyx_vtab = __pyx_vtabptr_8asynctnt_6iproto_8protocol_Metadata;
  p->fields = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->name_id_map = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Metadata(PyObject *o) {
  struct C_Metadata *p = (struct C_Metadata *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->fields);
  Py_CLEAR(p->name_id_map);
  Py_CLEAR(p->names);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_8asynctnt_6iproto_8protocol_Metadata < 128) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct C_Metadata)))) {
    __pyx_freelist_8asynctnt_6iproto_8protocol_Metadata[__pyx_freecount_8asynctnt_6iproto_8protocol_Metadata++] = ((struct C_Metadata *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_Metadata(PyObject *o, visitproc v, void *a) {
  int e;
  struct C_Metadata *p = (struct C_Metadata *)o;
  if (p->fields) {
    e = (*v)(p->fields, a); if (e) return e;
  }
  if (p->name_id_map) {
    e = (*v)(p->name_id_map, a); if (e) return e;
  }
  if (p->names) {
    e = (*v)(p->names, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_Metadata(PyObject *o) {
  PyObject* tmp;
  struct C_Metadata *p = (struct C_Metadata *)o;
  tmp = ((PyObject*)p->fields);
  p->fields = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->name_id_map);
  p->name_id_map = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->names);
  p->names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Metadata_fields(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_6fields_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Metadata_name_id_map(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_11name_id_map_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_3__repr__(self);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_Metadata[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8asynctnt_6iproto_8protocol_Metadata[] = {
  {(char *)"fields", __pyx_getprop_8asynctnt_6iproto_8protocol_8Metadata_fields, 0, (char *)0, 0},
  {(char *)"name_id_map", __pyx_getprop_8asynctnt_6iproto_8protocol_8Metadata_name_id_map, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot C_Metadata_Type_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Metadata},
  {Py_tp_repr, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_Metadata},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_Metadata},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_Metadata},
  {Py_tp_getset, (void *)__pyx_getsets_8asynctnt_6iproto_8protocol_Metadata},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Metadata},
  {0, 0},
};
static PyType_Spec C_Metadata_Type_spec = {
  "asynctnt.iproto.protocol.Metadata",
  sizeof(struct C_Metadata),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  C_Metadata_Type_slots,
};
#else

DL_EXPORT(PyTypeObject) C_Metadata_Type = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""Metadata", /*tp_name*/
  sizeof(struct C_Metadata), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Metadata, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_8Metadata_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_Metadata, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_Metadata, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_Metadata, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8asynctnt_6iproto_8protocol_Metadata, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_Metadata, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaIndex(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)o);
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->index_type = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->unique = Py_None; Py_INCREF(Py_None);
  p->metadata = ((struct C_Metadata *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_SchemaIndex(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->index_type);
  Py_CLEAR(p->unique);
  Py_CLEAR(p->metadata);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_SchemaIndex(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)o;
  if (p->unique) {
    e = (*v)(p->unique, a); if (e) return e;
  }
  if (p->metadata) {
    e = (*v)(((PyObject *)p->metadata), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_SchemaIndex(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)o;
  tmp = ((PyObject*)p->unique);
  p->unique = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->metadata);
  p->metadata = ((struct C_Metadata *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_sid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3sid_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_iid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3iid_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_4name_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_index_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_10index_type_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_unique(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_6unique_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_metadata(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_8metadata_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3__repr__(self);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_SchemaIndex[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8asynctnt_6iproto_8protocol_SchemaIndex[] = {
  {(char *)"sid", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_sid, 0, (char *)0, 0},
  {(char *)"iid", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_iid, 0, (char *)0, 0},
  {(char *)"name", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_name, 0, (char *)0, 0},
  {(char *)"index_type", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_index_type, 0, (char *)0, 0},
  {(char *)"unique", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_unique, 0, (char *)0, 0},
  {(char *)"metadata", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaIndex_metadata, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_SchemaIndex},
  {Py_tp_repr, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_SchemaIndex},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_SchemaIndex},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_SchemaIndex},
  {Py_tp_getset, (void *)__pyx_getsets_8asynctnt_6iproto_8protocol_SchemaIndex},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaIndex},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex_spec = {
  "asynctnt.iproto.protocol.SchemaIndex",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""SchemaIndex", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_SchemaIndex, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_SchemaIndex, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_SchemaIndex, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_SchemaIndex, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8asynctnt_6iproto_8protocol_SchemaIndex, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaIndex, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_SchemaIndex},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_SchemaIndex},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaIndex},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex_spec = {
  "asynctnt.iproto.protocol.SchemaDummyIndex",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaDummyIndex),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""SchemaDummyIndex", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaDummyIndex), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_SchemaIndex, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaIndex_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_SchemaIndex, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_SchemaIndex, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaIndex, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SchemaSpace __pyx_vtable_8asynctnt_6iproto_8protocol_SchemaSpace;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaSpace(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)o);
  p->__pyx_vtab = __pyx_vtabptr_8asynctnt_6iproto_8protocol_SchemaSpace;
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->engine = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->flags = Py_None; Py_INCREF(Py_None);
  p->metadata = ((struct C_Metadata *)Py_None); Py_INCREF(Py_None);
  p->indexes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_SchemaSpace(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->engine);
  Py_CLEAR(p->flags);
  Py_CLEAR(p->metadata);
  Py_CLEAR(p->indexes);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_SchemaSpace(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)o;
  if (p->flags) {
    e = (*v)(p->flags, a); if (e) return e;
  }
  if (p->metadata) {
    e = (*v)(((PyObject *)p->metadata), a); if (e) return e;
  }
  if (p->indexes) {
    e = (*v)(p->indexes, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_SchemaSpace(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)o;
  tmp = ((PyObject*)p->flags);
  p->flags = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->metadata);
  p->metadata = ((struct C_Metadata *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->indexes);
  p->indexes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_sid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_3sid_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_owner(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_5owner_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_4name_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_engine(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_6engine_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_field_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_11field_count_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_5flags_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_metadata(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_8metadata_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_indexes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_7indexes_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_3__repr__(self);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_SchemaSpace[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8asynctnt_6iproto_8protocol_SchemaSpace[] = {
  {(char *)"sid", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_sid, 0, (char *)0, 0},
  {(char *)"owner", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_owner, 0, (char *)0, 0},
  {(char *)"name", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_name, 0, (char *)0, 0},
  {(char *)"engine", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_engine, 0, (char *)0, 0},
  {(char *)"field_count", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_field_count, 0, (char *)0, 0},
  {(char *)"flags", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_flags, 0, (char *)0, 0},
  {(char *)"metadata", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_metadata, 0, (char *)0, 0},
  {(char *)"indexes", __pyx_getprop_8asynctnt_6iproto_8protocol_11SchemaSpace_indexes, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_SchemaSpace},
  {Py_tp_repr, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_SchemaSpace},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_SchemaSpace},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_SchemaSpace},
  {Py_tp_getset, (void *)__pyx_getsets_8asynctnt_6iproto_8protocol_SchemaSpace},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaSpace},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace_spec = {
  "asynctnt.iproto.protocol.SchemaSpace",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""SchemaSpace", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_SchemaSpace, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_SchemaSpace, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_SchemaSpace, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_SchemaSpace, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8asynctnt_6iproto_8protocol_SchemaSpace, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaSpace, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SchemaDummySpace __pyx_vtable_8asynctnt_6iproto_8protocol_SchemaDummySpace;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaDummySpace(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaDummySpace *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaSpace(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaDummySpace *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SchemaSpace*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_SchemaDummySpace;
  return o;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_SchemaSpace},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_SchemaSpace},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaDummySpace},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace_spec = {
  "asynctnt.iproto.protocol.SchemaDummySpace",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaDummySpace),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""SchemaDummySpace", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaDummySpace), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_SchemaSpace, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11SchemaSpace_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_SchemaSpace, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_SchemaSpace, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_SchemaDummySpace, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Schema __pyx_vtable_8asynctnt_6iproto_8protocol_Schema;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_Schema(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)o);
  p->__pyx_vtab = __pyx_vtabptr_8asynctnt_6iproto_8protocol_Schema;
  p->spaces = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Schema(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->spaces);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_Schema(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)o;
  if (p->spaces) {
    e = (*v)(p->spaces, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_Schema(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)o;
  tmp = ((PyObject*)p->spaces);
  p->spaces = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_6Schema_spaces(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_6Schema_6spaces_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_6Schema_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_6Schema_2id_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_6Schema_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_6Schema_3__repr__(self);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_Schema[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_6Schema_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8asynctnt_6iproto_8protocol_Schema[] = {
  {(char *)"spaces", __pyx_getprop_8asynctnt_6iproto_8protocol_6Schema_spaces, 0, (char *)0, 0},
  {(char *)"id", __pyx_getprop_8asynctnt_6iproto_8protocol_6Schema_id, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_Schema_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Schema},
  {Py_tp_repr, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_6Schema_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_Schema},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_Schema},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_Schema},
  {Py_tp_getset, (void *)__pyx_getsets_8asynctnt_6iproto_8protocol_Schema},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Schema},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_Schema_spec = {
  "asynctnt.iproto.protocol.Schema",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_Schema_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_Schema = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""Schema", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Schema, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_6Schema_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_Schema, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_Schema, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_Schema, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8asynctnt_6iproto_8protocol_Schema, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_Schema, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_WriteBuffer __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer;

static struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *__pyx_freelist_8asynctnt_6iproto_8protocol_WriteBuffer[256];
static int __pyx_freecount_8asynctnt_6iproto_8protocol_WriteBuffer = 0;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_WriteBuffer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_8asynctnt_6iproto_8protocol_WriteBuffer > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer)))) {
    o = (PyObject*)__pyx_freelist_8asynctnt_6iproto_8protocol_WriteBuffer[--__pyx_freecount_8asynctnt_6iproto_8protocol_WriteBuffer];
    memset(o, 0, sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer));
    (void) PyObject_INIT(o, t);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)o);
  p->__pyx_vtab = __pyx_vtabptr_8asynctnt_6iproto_8protocol_WriteBuffer;
  p->_encoding = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_WriteBuffer(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)o;
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_encoding);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_8asynctnt_6iproto_8protocol_WriteBuffer < 256) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer)))) {
    __pyx_freelist_8asynctnt_6iproto_8protocol_WriteBuffer[__pyx_freecount_8asynctnt_6iproto_8protocol_WriteBuffer++] = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_WriteBuffer[] = {
  {"hex", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_9hex, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
#if !CYTHON_COMPILING_IN_LIMITED_API

static PyBufferProcs __pyx_tp_as_buffer_WriteBuffer = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_5__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_7__releasebuffer__, /*bf_releasebuffer*/
};
#endif
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_WriteBuffer},
  #if defined(Py_bf_getbuffer)
  {Py_bf_getbuffer, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_5__getbuffer__},
  #endif
  #if defined(Py_bf_releasebuffer)
  {Py_bf_releasebuffer, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_7__releasebuffer__},
  #endif
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_WriteBuffer},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_WriteBuffer},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer_spec = {
  "asynctnt.iproto.protocol.WriteBuffer",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer_slots,
};
#else

static PyBufferProcs __pyx_tp_as_buffer_WriteBuffer = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_5__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_8asynctnt_6iproto_8protocol_11WriteBuffer_7__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""WriteBuffer", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_WriteBuffer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_WriteBuffer, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_WriteBuffer, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_WriteBuffer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_ReadBuffer __pyx_vtable_8asynctnt_6iproto_8protocol_ReadBuffer;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_ReadBuffer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)o);
  p->__pyx_vtab = __pyx_vtabptr_8asynctnt_6iproto_8protocol_ReadBuffer;
  p->encoding = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_ReadBuffer(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)o;
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->encoding);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_ReadBuffer[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_10ReadBuffer_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_ReadBuffer},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_ReadBuffer},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_ReadBuffer},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer_spec = {
  "asynctnt.iproto.protocol.ReadBuffer",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""ReadBuffer", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_ReadBuffer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_ReadBuffer, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_ReadBuffer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest __pyx_vtable_8asynctnt_6iproto_8protocol_BaseRequest;

static struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *__pyx_freelist_8asynctnt_6iproto_8protocol_BaseRequest[256];
static int __pyx_freecount_8asynctnt_6iproto_8protocol_BaseRequest = 0;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_8asynctnt_6iproto_8protocol_BaseRequest > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_8asynctnt_6iproto_8protocol_BaseRequest[--__pyx_freecount_8asynctnt_6iproto_8protocol_BaseRequest];
    memset(o, 0, sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)o);
  p->__pyx_vtab = __pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  p->space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)Py_None); Py_INCREF(Py_None);
  p->waiter = Py_None; Py_INCREF(Py_None);
  p->timeout_handle = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->space);
  Py_CLEAR(p->waiter);
  Py_CLEAR(p->timeout_handle);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_8asynctnt_6iproto_8protocol_BaseRequest < 256) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_8asynctnt_6iproto_8protocol_BaseRequest[__pyx_freecount_8asynctnt_6iproto_8protocol_BaseRequest++] = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)o;
  if (p->space) {
    e = (*v)(((PyObject *)p->space), a); if (e) return e;
  }
  if (p->waiter) {
    e = (*v)(p->waiter, a); if (e) return e;
  }
  if (p->timeout_handle) {
    e = (*v)(p->timeout_handle, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)o;
  tmp = ((PyObject*)p->space);
  p->space = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->waiter);
  p->waiter = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->timeout_handle);
  p->timeout_handle = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__(self);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_BaseRequest[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_BaseRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_repr, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_BaseRequest_spec = {
  "asynctnt.iproto.protocol.BaseRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_BaseRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_BaseRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""BaseRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_PingRequest __pyx_vtable_8asynctnt_6iproto_8protocol_PingRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_PingRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_PingRequest;
  return o;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_PingRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11PingRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11PingRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_PingRequest_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_PingRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_PingRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_PingRequest_spec = {
  "asynctnt.iproto.protocol.PingRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_PingRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_PingRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""PingRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_PingRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_PingRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_PingRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CallRequest __pyx_vtable_8asynctnt_6iproto_8protocol_CallRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_CallRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_CallRequest;
  p->func_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->args = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_CallRequest(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->func_name);
  Py_CLEAR(p->args);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest(o);
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_CallRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *)o;
  e = __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest(o, v, a); if (e) return e;
  if (p->args) {
    e = (*v)(p->args, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_CallRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest *)o;
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest(o);
  tmp = ((PyObject*)p->args);
  p->args = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_CallRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11CallRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11CallRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_CallRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_CallRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_CallRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_CallRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_CallRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_CallRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_CallRequest_spec = {
  "asynctnt.iproto.protocol.CallRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_CallRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_CallRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""CallRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_CallRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_CallRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_CallRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_CallRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_CallRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_CallRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_EvalRequest __pyx_vtable_8asynctnt_6iproto_8protocol_EvalRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_EvalRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_EvalRequest;
  p->expression = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->args = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_EvalRequest(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->expression);
  Py_CLEAR(p->args);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest(o);
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_EvalRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *)o;
  e = __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest(o, v, a); if (e) return e;
  if (p->args) {
    e = (*v)(p->args, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_EvalRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest *)o;
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest(o);
  tmp = ((PyObject*)p->args);
  p->args = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_EvalRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11EvalRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11EvalRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_EvalRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_EvalRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_EvalRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_EvalRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_EvalRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_EvalRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_EvalRequest_spec = {
  "asynctnt.iproto.protocol.EvalRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_EvalRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_EvalRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""EvalRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_EvalRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_EvalRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_EvalRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_EvalRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_EvalRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_EvalRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_SelectRequest __pyx_vtable_8asynctnt_6iproto_8protocol_SelectRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_SelectRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_SelectRequest;
  p->index = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)Py_None); Py_INCREF(Py_None);
  p->key = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_SelectRequest(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->index);
  Py_CLEAR(p->key);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest(o);
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_SelectRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *)o;
  e = __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest(o, v, a); if (e) return e;
  if (p->index) {
    e = (*v)(((PyObject *)p->index), a); if (e) return e;
  }
  if (p->key) {
    e = (*v)(p->key, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_SelectRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest *)o;
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest(o);
  tmp = ((PyObject*)p->index);
  p->index = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->key);
  p->key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_SelectRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13SelectRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13SelectRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_SelectRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_SelectRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_SelectRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_SelectRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_SelectRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_SelectRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_SelectRequest_spec = {
  "asynctnt.iproto.protocol.SelectRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_SelectRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_SelectRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""SelectRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_SelectRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_SelectRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_SelectRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_SelectRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_SelectRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_SelectRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_InsertRequest __pyx_vtable_8asynctnt_6iproto_8protocol_InsertRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_InsertRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_InsertRequest;
  p->t = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_InsertRequest(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->t);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest(o);
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_InsertRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *)o;
  e = __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest(o, v, a); if (e) return e;
  if (p->t) {
    e = (*v)(p->t, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_InsertRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest *)o;
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest(o);
  tmp = ((PyObject*)p->t);
  p->t = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_InsertRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13InsertRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13InsertRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_InsertRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_InsertRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_InsertRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_InsertRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_InsertRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_InsertRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_InsertRequest_spec = {
  "asynctnt.iproto.protocol.InsertRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_InsertRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_InsertRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""InsertRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_InsertRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_InsertRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_InsertRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_InsertRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_InsertRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_InsertRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_DeleteRequest __pyx_vtable_8asynctnt_6iproto_8protocol_DeleteRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_DeleteRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_DeleteRequest;
  p->index = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)Py_None); Py_INCREF(Py_None);
  p->key = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_DeleteRequest(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->index);
  Py_CLEAR(p->key);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest(o);
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_DeleteRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *)o;
  e = __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest(o, v, a); if (e) return e;
  if (p->index) {
    e = (*v)(((PyObject *)p->index), a); if (e) return e;
  }
  if (p->key) {
    e = (*v)(p->key, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_DeleteRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest *)o;
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest(o);
  tmp = ((PyObject*)p->index);
  p->index = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->key);
  p->key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_DeleteRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13DeleteRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13DeleteRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_DeleteRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_DeleteRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_DeleteRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_DeleteRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_DeleteRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest_spec = {
  "asynctnt.iproto.protocol.DeleteRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""DeleteRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_DeleteRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_DeleteRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_DeleteRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_DeleteRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_DeleteRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_DeleteRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_UpdateRequest __pyx_vtable_8asynctnt_6iproto_8protocol_UpdateRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_UpdateRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_UpdateRequest;
  p->index = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)Py_None); Py_INCREF(Py_None);
  p->key = Py_None; Py_INCREF(Py_None);
  p->operations = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_UpdateRequest(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->index);
  Py_CLEAR(p->key);
  Py_CLEAR(p->operations);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest(o);
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_UpdateRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *)o;
  e = __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest(o, v, a); if (e) return e;
  if (p->index) {
    e = (*v)(((PyObject *)p->index), a); if (e) return e;
  }
  if (p->key) {
    e = (*v)(p->key, a); if (e) return e;
  }
  if (p->operations) {
    e = (*v)(p->operations, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_UpdateRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest *)o;
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest(o);
  tmp = ((PyObject*)p->index);
  p->index = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->key);
  p->key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->operations);
  p->operations = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_UpdateRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13UpdateRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13UpdateRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_UpdateRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_UpdateRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_UpdateRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_UpdateRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_UpdateRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest_spec = {
  "asynctnt.iproto.protocol.UpdateRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""UpdateRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpdateRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_UpdateRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_UpdateRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_UpdateRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_UpdateRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_UpdateRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_UpsertRequest __pyx_vtable_8asynctnt_6iproto_8protocol_UpsertRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_UpsertRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_UpsertRequest;
  p->t = Py_None; Py_INCREF(Py_None);
  p->operations = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_UpsertRequest(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->t);
  Py_CLEAR(p->operations);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest(o);
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_UpsertRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *)o;
  e = __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest(o, v, a); if (e) return e;
  if (p->t) {
    e = (*v)(p->t, a); if (e) return e;
  }
  if (p->operations) {
    e = (*v)(p->operations, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_UpsertRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest *)o;
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest(o);
  tmp = ((PyObject*)p->t);
  p->t = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->operations);
  p->operations = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_UpsertRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13UpsertRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13UpsertRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_UpsertRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_UpsertRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_UpsertRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_UpsertRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_UpsertRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest_spec = {
  "asynctnt.iproto.protocol.UpsertRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""UpsertRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_UpsertRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_UpsertRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_UpsertRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_UpsertRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_UpsertRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_UpsertRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_PrepareRequest __pyx_vtable_8asynctnt_6iproto_8protocol_PrepareRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_PrepareRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_PrepareRequest;
  p->query = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_PrepareRequest(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->query);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest(o);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_PrepareRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_14PrepareRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_14PrepareRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_PrepareRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_PrepareRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_PrepareRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest_spec = {
  "asynctnt.iproto.protocol.PrepareRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""PrepareRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_PrepareRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_PrepareRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_PrepareRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_PrepareRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_ExecuteRequest __pyx_vtable_8asynctnt_6iproto_8protocol_ExecuteRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_ExecuteRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_ExecuteRequest;
  p->query = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->args = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_ExecuteRequest(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->query);
  Py_CLEAR(p->args);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest(o);
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_ExecuteRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *)o;
  e = __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest(o, v, a); if (e) return e;
  if (p->args) {
    e = (*v)(p->args, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_ExecuteRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest *)o;
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest(o);
  tmp = ((PyObject*)p->args);
  p->args = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_ExecuteRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_14ExecuteRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_14ExecuteRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_ExecuteRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_ExecuteRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_ExecuteRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_ExecuteRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_ExecuteRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest_spec = {
  "asynctnt.iproto.protocol.ExecuteRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""ExecuteRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_ExecuteRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_ExecuteRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_ExecuteRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_ExecuteRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_ExecuteRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_ExecuteRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_IDRequest __pyx_vtable_8asynctnt_6iproto_8protocol_IDRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_IDRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_IDRequest;
  return o;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_IDRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_9IDRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_9IDRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_IDRequest_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_IDRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_IDRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_IDRequest_spec = {
  "asynctnt.iproto.protocol.IDRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_IDRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_IDRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""IDRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_IDRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_IDRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_IDRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_AuthRequest __pyx_vtable_8asynctnt_6iproto_8protocol_AuthRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_AuthRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_AuthRequest;
  p->salt = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->username = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->password = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_AuthRequest(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->salt);
  Py_CLEAR(p->username);
  Py_CLEAR(p->password);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest(o);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_AuthRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11AuthRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11AuthRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_AuthRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_AuthRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_AuthRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_AuthRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_AuthRequest_spec = {
  "asynctnt.iproto.protocol.AuthRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_AuthRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_AuthRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""AuthRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_AuthRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_AuthRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_AuthRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_AuthRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BeginRequest __pyx_vtable_8asynctnt_6iproto_8protocol_BeginRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_BeginRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_BeginRequest;
  return o;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_BeginRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BeginRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BeginRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_BeginRequest_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_BeginRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_BeginRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_BeginRequest_spec = {
  "asynctnt.iproto.protocol.BeginRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_BeginRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_BeginRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""BeginRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_BeginRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_BeginRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_BeginRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CommitRequest __pyx_vtable_8asynctnt_6iproto_8protocol_CommitRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_CommitRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_CommitRequest;
  return o;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_CommitRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13CommitRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_13CommitRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_CommitRequest_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_CommitRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_CommitRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_CommitRequest_spec = {
  "asynctnt.iproto.protocol.CommitRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_CommitRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_CommitRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""CommitRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_CommitRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_CommitRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_CommitRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_RollbackRequest __pyx_vtable_8asynctnt_6iproto_8protocol_RollbackRequest;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_RollbackRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseRequest*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_RollbackRequest;
  return o;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_RollbackRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_15RollbackRequest_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_15RollbackRequest_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_RollbackRequest},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_RollbackRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest_spec = {
  "asynctnt.iproto.protocol.RollbackRequest",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""RollbackRequest", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_RollbackRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_8asynctnt_6iproto_8protocol_11BaseRequest_1__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_RollbackRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_RollbackRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)o);
  p->error_type = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->file = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->message = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->fields = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->error_type);
  Py_CLEAR(p->file);
  Py_CLEAR(p->message);
  Py_CLEAR(p->fields);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)o;
  if (p->fields) {
    e = (*v)(p->fields, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame *)o;
  tmp = ((PyObject*)p->fields);
  p->fields = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_error_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_10error_type_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_file(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4file_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_line(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4line_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_message(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_7message_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_err_no(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6err_no_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_code(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_4code_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_fields(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_6fields_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_1__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_1__repr__(self);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_1__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame[] = {
  {(char *)"error_type", __pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_error_type, 0, (char *)0, 0},
  {(char *)"file", __pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_file, 0, (char *)0, 0},
  {(char *)"line", __pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_line, 0, (char *)0, 0},
  {(char *)"message", __pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_message, 0, (char *)0, 0},
  {(char *)"err_no", __pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_err_no, 0, (char *)0, 0},
  {(char *)"code", __pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_code, 0, (char *)0, 0},
  {(char *)"fields", __pyx_getprop_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_fields, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame},
  {Py_tp_repr, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_1__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame},
  {Py_tp_getset, (void *)__pyx_getsets_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame_spec = {
  "asynctnt.iproto.protocol.IProtoErrorStackFrame",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""IProtoErrorStackFrame", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_1__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_IProtoError(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)o);
  p->trace = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_IProtoError(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->trace);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_IProtoError(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)o;
  if (p->trace) {
    e = (*v)(p->trace, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_IProtoError(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)o;
  tmp = ((PyObject*)p->trace);
  p->trace = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_11IProtoError_trace(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_11IProtoError_5trace_1__get__(o);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_IProtoError[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11IProtoError_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_11IProtoError_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8asynctnt_6iproto_8protocol_IProtoError[] = {
  {(char *)"trace", __pyx_getprop_8asynctnt_6iproto_8protocol_11IProtoError_trace, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_IProtoError_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_IProtoError},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_IProtoError},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_IProtoError},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_IProtoError},
  {Py_tp_getset, (void *)__pyx_getsets_8asynctnt_6iproto_8protocol_IProtoError},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_IProtoError},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_IProtoError_spec = {
  "asynctnt.iproto.protocol.IProtoError",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_IProtoError_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_IProtoError = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""IProtoError", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_IProtoError, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_IProtoError, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_IProtoError, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_IProtoError, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8asynctnt_6iproto_8protocol_IProtoError, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_IProtoError, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Response __pyx_vtable_8asynctnt_6iproto_8protocol_Response;

static struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *__pyx_freelist_8asynctnt_6iproto_8protocol_Response[256];
static int __pyx_freecount_8asynctnt_6iproto_8protocol_Response = 0;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_Response(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_8asynctnt_6iproto_8protocol_Response > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response)))) {
    o = (PyObject*)__pyx_freelist_8asynctnt_6iproto_8protocol_Response[--__pyx_freecount_8asynctnt_6iproto_8protocol_Response];
    memset(o, 0, sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)o);
  p->__pyx_vtab = __pyx_vtabptr_8asynctnt_6iproto_8protocol_Response;
  p->errmsg = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->error = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)Py_None); Py_INCREF(Py_None);
  p->body = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->encoding = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->metadata = ((struct C_Metadata *)Py_None); Py_INCREF(Py_None);
  p->params = ((struct C_Metadata *)Py_None); Py_INCREF(Py_None);
  p->autoincrement_ids = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->request_ = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)Py_None); Py_INCREF(Py_None);
  p->_exception = Py_None; Py_INCREF(Py_None);
  p->_q = Py_None; Py_INCREF(Py_None);
  p->_push_event = Py_None; Py_INCREF(Py_None);
  p->_q_append = Py_None; Py_INCREF(Py_None);
  p->_q_popleft = Py_None; Py_INCREF(Py_None);
  p->_push_event_set = Py_None; Py_INCREF(Py_None);
  p->_push_event_clear = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8asynctnt_6iproto_8protocol_8Response_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Response(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->errmsg);
  Py_CLEAR(p->error);
  Py_CLEAR(p->body);
  Py_CLEAR(p->encoding);
  Py_CLEAR(p->metadata);
  Py_CLEAR(p->params);
  Py_CLEAR(p->autoincrement_ids);
  Py_CLEAR(p->request_);
  Py_CLEAR(p->_exception);
  Py_CLEAR(p->_q);
  Py_CLEAR(p->_push_event);
  Py_CLEAR(p->_q_append);
  Py_CLEAR(p->_q_popleft);
  Py_CLEAR(p->_push_event_set);
  Py_CLEAR(p->_push_event_clear);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_8asynctnt_6iproto_8protocol_Response < 256) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response)))) {
    __pyx_freelist_8asynctnt_6iproto_8protocol_Response[__pyx_freecount_8asynctnt_6iproto_8protocol_Response++] = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_Response(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)o;
  if (p->error) {
    e = (*v)(((PyObject *)p->error), a); if (e) return e;
  }
  if (p->body) {
    e = (*v)(p->body, a); if (e) return e;
  }
  if (p->metadata) {
    e = (*v)(((PyObject *)p->metadata), a); if (e) return e;
  }
  if (p->params) {
    e = (*v)(((PyObject *)p->params), a); if (e) return e;
  }
  if (p->autoincrement_ids) {
    e = (*v)(p->autoincrement_ids, a); if (e) return e;
  }
  if (p->request_) {
    e = (*v)(((PyObject *)p->request_), a); if (e) return e;
  }
  if (p->_exception) {
    e = (*v)(p->_exception, a); if (e) return e;
  }
  if (p->_q) {
    e = (*v)(p->_q, a); if (e) return e;
  }
  if (p->_push_event) {
    e = (*v)(p->_push_event, a); if (e) return e;
  }
  if (p->_q_append) {
    e = (*v)(p->_q_append, a); if (e) return e;
  }
  if (p->_q_popleft) {
    e = (*v)(p->_q_popleft, a); if (e) return e;
  }
  if (p->_push_event_set) {
    e = (*v)(p->_push_event_set, a); if (e) return e;
  }
  if (p->_push_event_clear) {
    e = (*v)(p->_push_event_clear, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_Response(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)o;
  tmp = ((PyObject*)p->error);
  p->error = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->body);
  p->body = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->metadata);
  p->metadata = ((struct C_Metadata *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->params);
  p->params = ((struct C_Metadata *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->autoincrement_ids);
  p->autoincrement_ids = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->request_);
  p->request_ = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_exception);
  p->_exception = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_q);
  p->_q = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_push_event);
  p->_push_event = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_q_append);
  p->_q_append = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_q_popleft);
  p->_q_popleft = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_push_event_set);
  p->_push_event_set = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_push_event_clear);
  p->_push_event_clear = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_8asynctnt_6iproto_8protocol_Response(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_sync(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_4sync_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_code(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_4code_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_return_code(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_11return_code_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_schema_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_9schema_id_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_rowcount(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_8rowcount_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_stmt_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_7stmt_id_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_errmsg(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_6errmsg_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_error(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_5error_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_body(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_4body_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_encoding(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_8encoding_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_metadata(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_8metadata_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_params(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_6params_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_params_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_12params_count_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response_autoincrement_ids(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_17autoincrement_ids_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response__q(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_2_q_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_8Response__push_event(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_11_push_event_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_8Response_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_8Response_3__repr__(self);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_Response[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_8Response_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"done", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_5done, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8asynctnt_6iproto_8protocol_Response[] = {
  {(char *)"sync", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_sync, 0, (char *)PyDoc_STR("\n            Response's sync (incremental id) for the corresponding request\n        "), 0},
  {(char *)"code", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_code, 0, (char *)PyDoc_STR("\n            Response code (0 - success)\n        "), 0},
  {(char *)"return_code", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_return_code, 0, (char *)PyDoc_STR("\n            Response return code (It's essentially a code & 0x7FFF)\n        "), 0},
  {(char *)"schema_id", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_schema_id, 0, (char *)0, 0},
  {(char *)"rowcount", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_rowcount, 0, (char *)0, 0},
  {(char *)"stmt_id", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_stmt_id, 0, (char *)0, 0},
  {(char *)"errmsg", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_errmsg, 0, (char *)0, 0},
  {(char *)"error", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_error, 0, (char *)0, 0},
  {(char *)"body", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_body, 0, (char *)0, 0},
  {(char *)"encoding", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_encoding, 0, (char *)0, 0},
  {(char *)"metadata", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_metadata, 0, (char *)0, 0},
  {(char *)"params", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_params, 0, (char *)0, 0},
  {(char *)"params_count", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_params_count, 0, (char *)0, 0},
  {(char *)"autoincrement_ids", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response_autoincrement_ids, 0, (char *)0, 0},
  {(char *)"_q", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response__q, 0, (char *)0, 0},
  {(char *)"_push_event", __pyx_getprop_8asynctnt_6iproto_8protocol_8Response__push_event, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_Response_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Response},
  {Py_tp_repr, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_3__repr__},
  {Py_sq_length, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_7__len__},
  {Py_sq_item, (void *)__pyx_sq_item_8asynctnt_6iproto_8protocol_Response},
  {Py_mp_length, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_7__len__},
  {Py_mp_subscript, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_9__getitem__},
  {Py_tp_doc, (void *)PyDoc_STR("\n        Response object for all the requests to Tarantool\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_Response},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_Response},
  {Py_tp_iter, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_8Response_11__iter__},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_Response},
  {Py_tp_getset, (void *)__pyx_getsets_8asynctnt_6iproto_8protocol_Response},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Response},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_Response_spec = {
  "asynctnt.iproto.protocol.Response",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_Response_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Response = {
  __pyx_pw_8asynctnt_6iproto_8protocol_8Response_7__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_8asynctnt_6iproto_8protocol_Response, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Response = {
  __pyx_pw_8asynctnt_6iproto_8protocol_8Response_7__len__, /*mp_length*/
  __pyx_pw_8asynctnt_6iproto_8protocol_8Response_9__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_Response = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""Response", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Response, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_8Response_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Response, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Response, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n        Response object for all the requests to Tarantool\n    "), /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_Response, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_Response, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_8asynctnt_6iproto_8protocol_8Response_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_Response, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8asynctnt_6iproto_8protocol_Response, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_Response, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_Db __pyx_vtable_8asynctnt_6iproto_8protocol_Db;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_Db(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)o);
  p->__pyx_vtab = __pyx_vtabptr_8asynctnt_6iproto_8protocol_Db;
  p->_protocol = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)Py_None); Py_INCREF(Py_None);
  p->_encoding = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8asynctnt_6iproto_8protocol_2Db_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Db(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_protocol);
  Py_CLEAR(p->_encoding);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_Db(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)o;
  if (p->_protocol) {
    e = (*v)(((PyObject *)p->_protocol), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_Db(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)o;
  tmp = ((PyObject*)p->_protocol);
  p->_protocol = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_2Db_stream_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_2Db_9stream_id_1__get__(o);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_Db[] = {
  {"set_stream_id", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_3set_stream_id, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"ping", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_5ping, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"call16", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_7call16, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"call", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_9call, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"eval", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_11eval, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"select", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_13select, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"insert", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_15insert, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"replace", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_17replace, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"delete", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_19delete, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_21update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"upsert", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_23upsert, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"execute", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_25execute, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"prepare", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_27prepare, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"begin", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_29begin, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"commit", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_31commit, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"rollback", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_33rollback, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_35__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_2Db_37__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8asynctnt_6iproto_8protocol_Db[] = {
  {(char *)"stream_id", __pyx_getprop_8asynctnt_6iproto_8protocol_2Db_stream_id, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_Db_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Db},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_Db},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_Db},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_Db},
  {Py_tp_getset, (void *)__pyx_getsets_8asynctnt_6iproto_8protocol_Db},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_Db},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_Db_spec = {
  "asynctnt.iproto.protocol.Db",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_Db_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_Db = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""Db", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_Db, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_Db, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_Db, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_Db, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8asynctnt_6iproto_8protocol_Db, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_Db, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_PushIterator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)o);
  p->_fut = Py_None; Py_INCREF(Py_None);
  p->_request = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)Py_None); Py_INCREF(Py_None);
  p->_response = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_PushIterator(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_fut);
  Py_CLEAR(p->_request);
  Py_CLEAR(p->_response);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_PushIterator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)o;
  if (p->_fut) {
    e = (*v)(p->_fut, a); if (e) return e;
  }
  if (p->_request) {
    e = (*v)(((PyObject *)p->_request), a); if (e) return e;
  }
  if (p->_response) {
    e = (*v)(((PyObject *)p->_response), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_PushIterator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator *)o;
  tmp = ((PyObject*)p->_fut);
  p->_fut = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_request);
  p->_request = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_response);
  p->_response = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_12PushIterator_response(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_8response_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_5__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  PyObject *res = __pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_5__next__(self);
  if (!res && !PyErr_Occurred()) { PyErr_SetNone(PyExc_StopIteration); }
  return res;
}
static PyObject *__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_7__aiter__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_7__aiter__(self);
}
static PyObject *__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_9__anext__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_9__anext__(self);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_PushIterator[] = {
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_5__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__aiter__", (PyCFunction)__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_7__aiter__, METH_NOARGS|METH_COEXIST, 0},
  {"__anext__", (PyCFunction)__pyx_specialmethod___pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_9__anext__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_12__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_14__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8asynctnt_6iproto_8protocol_PushIterator[] = {
  {(char *)"response", __pyx_getprop_8asynctnt_6iproto_8protocol_12PushIterator_response, 0, (char *)PyDoc_STR("\n            Return current Response object. Might be handy to know if the\n            request is finished already while iterating over the PushIterator\n\n            :rtype: asynctnt.Response\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_PushIterator_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_PushIterator},
  {Py_am_aiter, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_7__aiter__},
  {Py_am_anext, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_9__anext__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_PushIterator},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_PushIterator},
  {Py_tp_iter, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_3__iter__},
  {Py_tp_iternext, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_5__next__},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_PushIterator},
  {Py_tp_getset, (void *)__pyx_getsets_8asynctnt_6iproto_8protocol_PushIterator},
  {Py_tp_init, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_PushIterator},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_PushIterator_spec = {
  "asynctnt.iproto.protocol.PushIterator",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_PushIterator_slots,
};
#else

#if PY_MAJOR_VERSION >= 3
static __Pyx_PyAsyncMethodsStruct __pyx_tp_as_async_PushIterator = {
  0, /*am_await*/
  __pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_7__aiter__, /*am_aiter*/
  __pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_9__anext__, /*am_anext*/
  #if PY_VERSION_HEX >= 0x030A00A3
  0, /*am_send*/
  #endif
};
#endif

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_PushIterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""PushIterator", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_PushIterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_PushIterator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  &__pyx_tp_as_async_PushIterator, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_PushIterator, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_PushIterator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_3__iter__, /*tp_iter*/
  __pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_5__next__, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_PushIterator, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8asynctnt_6iproto_8protocol_PushIterator, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_12PushIterator_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_PushIterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol __pyx_vtable_8asynctnt_6iproto_8protocol_CoreProtocol;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_CoreProtocol(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)o);
  p->__pyx_vtab = __pyx_vtabptr_8asynctnt_6iproto_8protocol_CoreProtocol;
  p->host = Py_None; Py_INCREF(Py_None);
  p->port = Py_None; Py_INCREF(Py_None);
  p->encoding = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->rbuf = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)Py_None); Py_INCREF(Py_None);
  p->version = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->salt = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_CoreProtocol(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_CoreProtocol) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->host);
  Py_CLEAR(p->port);
  Py_CLEAR(p->encoding);
  Py_CLEAR(p->rbuf);
  Py_CLEAR(p->version);
  Py_CLEAR(p->salt);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_CoreProtocol(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)o;
  if (p->host) {
    e = (*v)(p->host, a); if (e) return e;
  }
  if (p->port) {
    e = (*v)(p->port, a); if (e) return e;
  }
  if (p->rbuf) {
    e = (*v)(((PyObject *)p->rbuf), a); if (e) return e;
  }
  if (p->version) {
    e = (*v)(p->version, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_CoreProtocol(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *)o;
  tmp = ((PyObject*)p->host);
  p->host = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->port);
  p->port = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->rbuf);
  p->rbuf = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->version);
  p->version = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_CoreProtocol[] = {
  {"is_connected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_3is_connected, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"is_fully_connected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_5is_fully_connected, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_version", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_7get_version, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_parse_version", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_9_parse_version, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"data_received", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_11data_received, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"connection_made", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_13connection_made, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"connection_lost", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_15connection_lost, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_17__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_19__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_CoreProtocol},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_CoreProtocol},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_CoreProtocol},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_CoreProtocol},
  {Py_tp_init, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_CoreProtocol},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol_spec = {
  "asynctnt.iproto.protocol.CoreProtocol",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""CoreProtocol", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_CoreProtocol, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_CoreProtocol, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_CoreProtocol, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_CoreProtocol, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_12CoreProtocol_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_CoreProtocol, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_BaseProtocol __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol_BaseProtocol(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *p;
  PyObject *o = __pyx_tp_new_8asynctnt_6iproto_8protocol_CoreProtocol(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_CoreProtocol*)__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseProtocol;
  p->loop = Py_None; Py_INCREF(Py_None);
  p->username = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->password = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->connected_fut = Py_None; Py_INCREF(Py_None);
  p->on_connection_made_cb = Py_None; Py_INCREF(Py_None);
  p->on_connection_lost_cb = Py_None; Py_INCREF(Py_None);
  p->_on_request_completed_cb = Py_None; Py_INCREF(Py_None);
  p->_on_request_timeout_cb = Py_None; Py_INCREF(Py_None);
  p->_reqs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_schema = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)Py_None); Py_INCREF(Py_None);
  p->_refetch_schema_future = Py_None; Py_INCREF(Py_None);
  p->_db = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)Py_None); Py_INCREF(Py_None);
  p->create_future = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseProtocol(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseProtocol) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->loop);
  Py_CLEAR(p->username);
  Py_CLEAR(p->password);
  Py_CLEAR(p->connected_fut);
  Py_CLEAR(p->on_connection_made_cb);
  Py_CLEAR(p->on_connection_lost_cb);
  Py_CLEAR(p->_on_request_completed_cb);
  Py_CLEAR(p->_on_request_timeout_cb);
  Py_CLEAR(p->_reqs);
  Py_CLEAR(p->_schema);
  Py_CLEAR(p->_refetch_schema_future);
  Py_CLEAR(p->_db);
  Py_CLEAR(p->create_future);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_CoreProtocol(o);
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseProtocol(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)o;
  e = __pyx_tp_traverse_8asynctnt_6iproto_8protocol_CoreProtocol(o, v, a); if (e) return e;
  if (p->loop) {
    e = (*v)(p->loop, a); if (e) return e;
  }
  if (p->connected_fut) {
    e = (*v)(p->connected_fut, a); if (e) return e;
  }
  if (p->on_connection_made_cb) {
    e = (*v)(p->on_connection_made_cb, a); if (e) return e;
  }
  if (p->on_connection_lost_cb) {
    e = (*v)(p->on_connection_lost_cb, a); if (e) return e;
  }
  if (p->_on_request_completed_cb) {
    e = (*v)(p->_on_request_completed_cb, a); if (e) return e;
  }
  if (p->_on_request_timeout_cb) {
    e = (*v)(p->_on_request_timeout_cb, a); if (e) return e;
  }
  if (p->_reqs) {
    e = (*v)(p->_reqs, a); if (e) return e;
  }
  if (p->_schema) {
    e = (*v)(((PyObject *)p->_schema), a); if (e) return e;
  }
  if (p->_refetch_schema_future) {
    e = (*v)(p->_refetch_schema_future, a); if (e) return e;
  }
  if (p->_db) {
    e = (*v)(((PyObject *)p->_db), a); if (e) return e;
  }
  if (p->create_future) {
    e = (*v)(p->create_future, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseProtocol(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)o;
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_CoreProtocol(o);
  tmp = ((PyObject*)p->loop);
  p->loop = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->connected_fut);
  p->connected_fut = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->on_connection_made_cb);
  p->on_connection_made_cb = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->on_connection_lost_cb);
  p->on_connection_lost_cb = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_on_request_completed_cb);
  p->_on_request_completed_cb = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_on_request_timeout_cb);
  p->_on_request_timeout_cb = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_reqs);
  p->_reqs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_schema);
  p->_schema = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_refetch_schema_future);
  p->_refetch_schema_future = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_db);
  p->_db = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->create_future);
  p->create_future = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_12BaseProtocol_schema(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_6schema_1__get__(o);
}

static PyObject *__pyx_getprop_8asynctnt_6iproto_8protocol_12BaseProtocol_schema_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_9schema_id_1__get__(o);
}

static PyMethodDef __pyx_methods_8asynctnt_6iproto_8protocol_BaseProtocol[] = {
  {"_create_future_fallback", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_3_create_future_fallback, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_on_request_timeout", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_5_on_request_timeout, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_on_request_completed", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_7_on_request_completed, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"create_db", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_9create_db, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_common_db", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_11get_common_db, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"refetch_schema", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_13refetch_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8asynctnt_6iproto_8protocol_BaseProtocol[] = {
  {(char *)"schema", __pyx_getprop_8asynctnt_6iproto_8protocol_12BaseProtocol_schema, 0, (char *)0, 0},
  {(char *)"schema_id", __pyx_getprop_8asynctnt_6iproto_8protocol_12BaseProtocol_schema_id, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseProtocol},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseProtocol},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseProtocol},
  {Py_tp_methods, (void *)__pyx_methods_8asynctnt_6iproto_8protocol_BaseProtocol},
  {Py_tp_getset, (void *)__pyx_getsets_8asynctnt_6iproto_8protocol_BaseProtocol},
  {Py_tp_init, (void *)__pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol_BaseProtocol},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol_spec = {
  "asynctnt.iproto.protocol.BaseProtocol",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""BaseProtocol", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol_BaseProtocol, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol_BaseProtocol, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol_BaseProtocol, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8asynctnt_6iproto_8protocol_BaseProtocol, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8asynctnt_6iproto_8protocol_BaseProtocol, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_8asynctnt_6iproto_8protocol_12BaseProtocol_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol_BaseProtocol, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex[8];
static int __pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex = 0;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex)))) {
    o = (PyObject*)__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex[--__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex];
    memset(o, 0, sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex)))) {
    __pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex[__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex++] = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex_spec = {
  "asynctnt.iproto.protocol.__pyx_scope_struct__hex",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""__pyx_scope_struct__hex", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr *__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr[--__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr[__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr_spec = {
  "asynctnt.iproto.protocol.__pyx_scope_struct_1_genexpr",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ *__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__[8];
static int __pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ = 0;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__)))) {
    o = (PyObject*)__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__[--__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__];
    memset(o, 0, sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_ev);
  Py_CLEAR(p->__pyx_v_exc);
  Py_CLEAR(p->__pyx_v_response);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__)))) {
    __pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__[__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__++] = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ *)o;
  if (p->__pyx_v_ev) {
    e = (*v)(p->__pyx_v_ev, a); if (e) return e;
  }
  if (p->__pyx_v_exc) {
    e = (*v)(p->__pyx_v_exc, a); if (e) return e;
  }
  if (p->__pyx_v_response) {
    e = (*v)(((PyObject *)p->__pyx_v_response), a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext___slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext___spec = {
  "asynctnt.iproto.protocol.__pyx_scope_struct_2___anext__",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext___slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""__pyx_scope_struct_2___anext__", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id[8];
static int __pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id = 0;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id)))) {
    o = (PyObject*)__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id[--__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id];
    memset(o, 0, sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id)))) {
    __pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id[__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id++] = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id_spec = {
  "asynctnt.iproto.protocol.__pyx_scope_struct_3__do_id",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""__pyx_scope_struct_3__do_id", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth[8];
static int __pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth = 0;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth)))) {
    o = (PyObject*)__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth[--__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth];
    memset(o, 0, sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth)))) {
    __pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth[__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth++] = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth_spec = {
  "asynctnt.iproto.protocol.__pyx_scope_struct_4__do_auth",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""__pyx_scope_struct_4__do_auth", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema[8];
static int __pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema = 0;

static PyObject *__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema)))) {
    o = (PyObject*)__pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema[--__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema];
    memset(o, 0, sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema(PyObject *o) {
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_fut);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema)))) {
    __pyx_freelist_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema[__pyx_freecount_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema++] = ((struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *)o;
  if (p->__pyx_v_fut) {
    e = (*v)(p->__pyx_v_fut, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *p = (struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema *)o;
  tmp = ((PyObject*)p->__pyx_v_fut);
  p->__pyx_v_fut = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema},
  {Py_tp_clear, (void *)__pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema},
  {Py_tp_new, (void *)__pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema},
  {0, 0},
};
static PyType_Spec __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema_spec = {
  "asynctnt.iproto.protocol.__pyx_scope_struct_5__do_fetch_schema",
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema_slots,
};
#else

static PyTypeObject __pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynctnt.iproto.protocol.""__pyx_scope_struct_5__do_fetch_schema", /*tp_name*/
  sizeof(struct __pyx_obj_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema, /*tp_traverse*/
  __pyx_tp_clear_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_kp_u_02x, __pyx_k_02x, sizeof(__pyx_k_02x), 0, 1, 0, 0},
    {&__pyx_n_s_AF_UNIX, __pyx_k_AF_UNIX, sizeof(__pyx_k_AF_UNIX), 0, 0, 1, 1},
    {&__pyx_n_u_AF_UNIX, __pyx_k_AF_UNIX, sizeof(__pyx_k_AF_UNIX), 0, 1, 0, 1},
    {&__pyx_n_s_ALL, __pyx_k_ALL, sizeof(__pyx_k_ALL), 0, 0, 1, 1},
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
    {&__pyx_n_s_AuthRequest, __pyx_k_AuthRequest, sizeof(__pyx_k_AuthRequest), 0, 0, 1, 1},
    {&__pyx_n_s_AuthRequest___reduce_cython, __pyx_k_AuthRequest___reduce_cython, sizeof(__pyx_k_AuthRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_AuthRequest___setstate_cython, __pyx_k_AuthRequest___setstate_cython, sizeof(__pyx_k_AuthRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_s_Available_Iterator_types, __pyx_k_Available_Iterator_types, sizeof(__pyx_k_Available_Iterator_types), 0, 0, 1, 0},
    {&__pyx_n_s_BITS_ALL_NOT_SET, __pyx_k_BITS_ALL_NOT_SET, sizeof(__pyx_k_BITS_ALL_NOT_SET), 0, 0, 1, 1},
    {&__pyx_n_s_BITS_ALL_SET, __pyx_k_BITS_ALL_SET, sizeof(__pyx_k_BITS_ALL_SET), 0, 0, 1, 1},
    {&__pyx_n_s_BITS_ANY_SET, __pyx_k_BITS_ANY_SET, sizeof(__pyx_k_BITS_ANY_SET), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol, __pyx_k_BaseProtocol, sizeof(__pyx_k_BaseProtocol), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol___reduce_cython, __pyx_k_BaseProtocol___reduce_cython, sizeof(__pyx_k_BaseProtocol___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol___setstate_cython, __pyx_k_BaseProtocol___setstate_cython, sizeof(__pyx_k_BaseProtocol___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol__create_future_fall, __pyx_k_BaseProtocol__create_future_fall, sizeof(__pyx_k_BaseProtocol__create_future_fall), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol__do_auth_locals_on, __pyx_k_BaseProtocol__do_auth_locals_on, sizeof(__pyx_k_BaseProtocol__do_auth_locals_on), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol__do_fetch_schema_lo, __pyx_k_BaseProtocol__do_fetch_schema_lo, sizeof(__pyx_k_BaseProtocol__do_fetch_schema_lo), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol__do_id_locals_on_id, __pyx_k_BaseProtocol__do_id_locals_on_id, sizeof(__pyx_k_BaseProtocol__do_id_locals_on_id), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol__on_request_complet, __pyx_k_BaseProtocol__on_request_complet, sizeof(__pyx_k_BaseProtocol__on_request_complet), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol__on_request_timeout, __pyx_k_BaseProtocol__on_request_timeout, sizeof(__pyx_k_BaseProtocol__on_request_timeout), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol_create_db, __pyx_k_BaseProtocol_create_db, sizeof(__pyx_k_BaseProtocol_create_db), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol_get_common_db, __pyx_k_BaseProtocol_get_common_db, sizeof(__pyx_k_BaseProtocol_get_common_db), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol_refetch_schema, __pyx_k_BaseProtocol_refetch_schema, sizeof(__pyx_k_BaseProtocol_refetch_schema), 0, 0, 1, 1},
    {&__pyx_n_s_BaseRequest, __pyx_k_BaseRequest, sizeof(__pyx_k_BaseRequest), 0, 0, 1, 1},
    {&__pyx_n_s_BaseRequest___reduce_cython, __pyx_k_BaseRequest___reduce_cython, sizeof(__pyx_k_BaseRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BaseRequest___setstate_cython, __pyx_k_BaseRequest___setstate_cython, sizeof(__pyx_k_BaseRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BeginRequest, __pyx_k_BeginRequest, sizeof(__pyx_k_BeginRequest), 0, 0, 1, 1},
    {&__pyx_n_s_BeginRequest___reduce_cython, __pyx_k_BeginRequest___reduce_cython, sizeof(__pyx_k_BeginRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BeginRequest___setstate_cython, __pyx_k_BeginRequest___setstate_cython, sizeof(__pyx_k_BeginRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BufferError, __pyx_k_BufferError, sizeof(__pyx_k_BufferError), 0, 0, 1, 1},
    {&__pyx_n_s_CallRequest, __pyx_k_CallRequest, sizeof(__pyx_k_CallRequest), 0, 0, 1, 1},
    {&__pyx_n_s_CallRequest___reduce_cython, __pyx_k_CallRequest___reduce_cython, sizeof(__pyx_k_CallRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_CallRequest___setstate_cython, __pyx_k_CallRequest___setstate_cython, sizeof(__pyx_k_CallRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_CancelledError, __pyx_k_CancelledError, sizeof(__pyx_k_CancelledError), 0, 0, 1, 1},
    {&__pyx_kp_u_Cannot_pop_push_from_a_non_async, __pyx_k_Cannot_pop_push_from_a_non_async, sizeof(__pyx_k_Cannot_pop_push_from_a_non_async), 0, 1, 0, 0},
    {&__pyx_kp_u_Cannot_use_iter_with_PushIterato, __pyx_k_Cannot_use_iter_with_PushIterato, sizeof(__pyx_k_Cannot_use_iter_with_PushIterato), 0, 1, 0, 0},
    {&__pyx_kp_u_Cannot_use_next_with_PushIterato, __pyx_k_Cannot_use_next_with_PushIterato, sizeof(__pyx_k_Cannot_use_next_with_PushIterato), 0, 1, 0, 0},
    {&__pyx_n_s_CommitRequest, __pyx_k_CommitRequest, sizeof(__pyx_k_CommitRequest), 0, 0, 1, 1},
    {&__pyx_n_s_CommitRequest___reduce_cython, __pyx_k_CommitRequest___reduce_cython, sizeof(__pyx_k_CommitRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_CommitRequest___setstate_cython, __pyx_k_CommitRequest___setstate_cython, sizeof(__pyx_k_CommitRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ConnectionRefusedError, __pyx_k_ConnectionRefusedError, sizeof(__pyx_k_ConnectionRefusedError), 0, 0, 1, 1},
    {&__pyx_n_s_ConnectionResetError, __pyx_k_ConnectionResetError, sizeof(__pyx_k_ConnectionResetError), 0, 0, 1, 1},
    {&__pyx_n_s_CoreProtocol, __pyx_k_CoreProtocol, sizeof(__pyx_k_CoreProtocol), 0, 0, 1, 1},
    {&__pyx_n_s_CoreProtocol___reduce_cython, __pyx_k_CoreProtocol___reduce_cython, sizeof(__pyx_k_CoreProtocol___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_CoreProtocol___setstate_cython, __pyx_k_CoreProtocol___setstate_cython, sizeof(__pyx_k_CoreProtocol___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_CoreProtocol__parse_version, __pyx_k_CoreProtocol__parse_version, sizeof(__pyx_k_CoreProtocol__parse_version), 0, 0, 1, 1},
    {&__pyx_n_s_CoreProtocol_connection_lost, __pyx_k_CoreProtocol_connection_lost, sizeof(__pyx_k_CoreProtocol_connection_lost), 0, 0, 1, 1},
    {&__pyx_n_s_CoreProtocol_connection_made, __pyx_k_CoreProtocol_connection_made, sizeof(__pyx_k_CoreProtocol_connection_made), 0, 0, 1, 1},
    {&__pyx_n_s_CoreProtocol_data_received, __pyx_k_CoreProtocol_data_received, sizeof(__pyx_k_CoreProtocol_data_received), 0, 0, 1, 1},
    {&__pyx_n_s_CoreProtocol_get_version, __pyx_k_CoreProtocol_get_version, sizeof(__pyx_k_CoreProtocol_get_version), 0, 0, 1, 1},
    {&__pyx_n_s_CoreProtocol_is_connected, __pyx_k_CoreProtocol_is_connected, sizeof(__pyx_k_CoreProtocol_is_connected), 0, 0, 1, 1},
    {&__pyx_n_s_CoreProtocol_is_fully_connected, __pyx_k_CoreProtocol_is_fully_connected, sizeof(__pyx_k_CoreProtocol_is_fully_connected), 0, 0, 1, 1},
    {&__pyx_kp_u_Creates_PushIterator_object_In, __pyx_k_Creates_PushIterator_object_In, sizeof(__pyx_k_Creates_PushIterator_object_In), 0, 1, 0, 0},
    {&__pyx_n_s_Db, __pyx_k_Db, sizeof(__pyx_k_Db), 0, 0, 1, 1},
    {&__pyx_n_s_Db___reduce_cython, __pyx_k_Db___reduce_cython, sizeof(__pyx_k_Db___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Db___setstate_cython, __pyx_k_Db___setstate_cython, sizeof(__pyx_k_Db___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Db_begin, __pyx_k_Db_begin, sizeof(__pyx_k_Db_begin), 0, 0, 1, 1},
    {&__pyx_n_s_Db_call, __pyx_k_Db_call, sizeof(__pyx_k_Db_call), 0, 0, 1, 1},
    {&__pyx_n_s_Db_call16, __pyx_k_Db_call16, sizeof(__pyx_k_Db_call16), 0, 0, 1, 1},
    {&__pyx_n_s_Db_commit, __pyx_k_Db_commit, sizeof(__pyx_k_Db_commit), 0, 0, 1, 1},
    {&__pyx_n_s_Db_delete, __pyx_k_Db_delete, sizeof(__pyx_k_Db_delete), 0, 0, 1, 1},
    {&__pyx_n_s_Db_eval, __pyx_k_Db_eval, sizeof(__pyx_k_Db_eval), 0, 0, 1, 1},
    {&__pyx_n_s_Db_execute, __pyx_k_Db_execute, sizeof(__pyx_k_Db_execute), 0, 0, 1, 1},
    {&__pyx_n_s_Db_insert, __pyx_k_Db_insert, sizeof(__pyx_k_Db_insert), 0, 0, 1, 1},
    {&__pyx_n_s_Db_ping, __pyx_k_Db_ping, sizeof(__pyx_k_Db_ping), 0, 0, 1, 1},
    {&__pyx_n_s_Db_prepare, __pyx_k_Db_prepare, sizeof(__pyx_k_Db_prepare), 0, 0, 1, 1},
    {&__pyx_n_s_Db_replace, __pyx_k_Db_replace, sizeof(__pyx_k_Db_replace), 0, 0, 1, 1},
    {&__pyx_n_s_Db_rollback, __pyx_k_Db_rollback, sizeof(__pyx_k_Db_rollback), 0, 0, 1, 1},
    {&__pyx_n_s_Db_select, __pyx_k_Db_select, sizeof(__pyx_k_Db_select), 0, 0, 1, 1},
    {&__pyx_n_s_Db_set_stream_id, __pyx_k_Db_set_stream_id, sizeof(__pyx_k_Db_set_stream_id), 0, 0, 1, 1},
    {&__pyx_n_s_Db_update, __pyx_k_Db_update, sizeof(__pyx_k_Db_update), 0, 0, 1, 1},
    {&__pyx_n_s_Db_upsert, __pyx_k_Db_upsert, sizeof(__pyx_k_Db_upsert), 0, 0, 1, 1},
    {&__pyx_kp_u_Deallocating_buffer_with_attache, __pyx_k_Deallocating_buffer_with_attache, sizeof(__pyx_k_Deallocating_buffer_with_attache), 0, 1, 0, 0},
    {&__pyx_n_s_Decimal, __pyx_k_Decimal, sizeof(__pyx_k_Decimal), 0, 0, 1, 1},
    {&__pyx_n_s_DeleteRequest, __pyx_k_DeleteRequest, sizeof(__pyx_k_DeleteRequest), 0, 0, 1, 1},
    {&__pyx_n_s_DeleteRequest___reduce_cython, __pyx_k_DeleteRequest___reduce_cython, sizeof(__pyx_k_DeleteRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_DeleteRequest___setstate_cython, __pyx_k_DeleteRequest___setstate_cython, sizeof(__pyx_k_DeleteRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_EQ, __pyx_k_EQ, sizeof(__pyx_k_EQ), 0, 0, 1, 1},
    {&__pyx_kp_u_Error_happened_while_parsing_sch, __pyx_k_Error_happened_while_parsing_sch, sizeof(__pyx_k_Error_happened_while_parsing_sch), 0, 1, 0, 0},
    {&__pyx_n_s_EvalRequest, __pyx_k_EvalRequest, sizeof(__pyx_k_EvalRequest), 0, 0, 1, 1},
    {&__pyx_n_s_EvalRequest___reduce_cython, __pyx_k_EvalRequest___reduce_cython, sizeof(__pyx_k_EvalRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_EvalRequest___setstate_cython, __pyx_k_EvalRequest___setstate_cython, sizeof(__pyx_k_EvalRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Event, __pyx_k_Event, sizeof(__pyx_k_Event), 0, 0, 1, 1},
    {&__pyx_n_s_ExecuteRequest, __pyx_k_ExecuteRequest, sizeof(__pyx_k_ExecuteRequest), 0, 0, 1, 1},
    {&__pyx_n_s_ExecuteRequest___reduce_cython, __pyx_k_ExecuteRequest___reduce_cython, sizeof(__pyx_k_ExecuteRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ExecuteRequest___setstate_cython, __pyx_k_ExecuteRequest___setstate_cython, sizeof(__pyx_k_ExecuteRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Field, __pyx_k_Field, sizeof(__pyx_k_Field), 0, 0, 1, 1},
    {&__pyx_n_s_Field___reduce_cython, __pyx_k_Field___reduce_cython, sizeof(__pyx_k_Field___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Field___setstate_cython, __pyx_k_Field___setstate_cython, sizeof(__pyx_k_Field___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Field_d_of_space_s_is_not_in_spa, __pyx_k_Field_d_of_space_s_is_not_in_spa, sizeof(__pyx_k_Field_d_of_space_s_is_not_in_spa), 0, 1, 0, 0},
    {&__pyx_kp_u_Field_map_must_contain_at_least, __pyx_k_Field_map_must_contain_at_least, sizeof(__pyx_k_Field_map_must_contain_at_least), 0, 1, 0, 0},
    {&__pyx_kp_u_Field_map_must_contain_at_least_2, __pyx_k_Field_map_must_contain_at_least_2, sizeof(__pyx_k_Field_map_must_contain_at_least_2), 0, 1, 0, 0},
    {&__pyx_kp_u_Field_name_type_is_nullable, __pyx_k_Field_name_type_is_nullable, sizeof(__pyx_k_Field_name_type_is_nullable), 0, 1, 0, 0},
    {&__pyx_kp_u_Field_not_found, __pyx_k_Field_not_found, sizeof(__pyx_k_Field_not_found), 0, 1, 0, 0},
    {&__pyx_kp_u_Field_s_in_supplied_dict_is_unkn, __pyx_k_Field_s_in_supplied_dict_is_unkn, sizeof(__pyx_k_Field_s_in_supplied_dict_is_unkn), 0, 1, 0, 0},
    {&__pyx_kp_u_Frame_type_code_message, __pyx_k_Frame_type_code_message, sizeof(__pyx_k_Frame_type_code_message), 0, 1, 0, 0},
    {&__pyx_n_s_Future, __pyx_k_Future, sizeof(__pyx_k_Future), 0, 0, 1, 1},
    {&__pyx_kp_u_Future_is_invalid_Make_sure_to_c, __pyx_k_Future_is_invalid_Make_sure_to_c, sizeof(__pyx_k_Future_is_invalid_Make_sure_to_c), 0, 1, 0, 0},
    {&__pyx_n_s_GE, __pyx_k_GE, sizeof(__pyx_k_GE), 0, 0, 1, 1},
    {&__pyx_n_s_GT, __pyx_k_GT, sizeof(__pyx_k_GT), 0, 0, 1, 1},
    {&__pyx_kp_u_Header_key_must_be_a_MP_UINT, __pyx_k_Header_key_must_be_a_MP_UINT, sizeof(__pyx_k_Header_key_must_be_a_MP_UINT), 0, 1, 0, 0},
    {&__pyx_n_s_IDRequest, __pyx_k_IDRequest, sizeof(__pyx_k_IDRequest), 0, 0, 1, 1},
    {&__pyx_n_s_IDRequest___reduce_cython, __pyx_k_IDRequest___reduce_cython, sizeof(__pyx_k_IDRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_IDRequest___setstate_cython, __pyx_k_IDRequest___setstate_cython, sizeof(__pyx_k_IDRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_IPPROTO_TCP, __pyx_k_IPPROTO_TCP, sizeof(__pyx_k_IPPROTO_TCP), 0, 0, 1, 1},
    {&__pyx_kp_u_IPROTO_ERROR_type_must_be_a_MP_M, __pyx_k_IPROTO_ERROR_type_must_be_a_MP_M, sizeof(__pyx_k_IPROTO_ERROR_type_must_be_a_MP_M), 0, 1, 0, 0},
    {&__pyx_kp_u_IPROTO_FIELD_SPAN_must_be_either, __pyx_k_IPROTO_FIELD_SPAN_must_be_either, sizeof(__pyx_k_IPROTO_FIELD_SPAN_must_be_either), 0, 1, 0, 0},
    {&__pyx_kp_u_IPROTO_STMT_ID_type_must_be_a_MP, __pyx_k_IPROTO_STMT_ID_type_must_be_a_MP, sizeof(__pyx_k_IPROTO_STMT_ID_type_must_be_a_MP), 0, 1, 0, 0},
    {&__pyx_n_s_IProtoError, __pyx_k_IProtoError, sizeof(__pyx_k_IProtoError), 0, 0, 1, 1},
    {&__pyx_n_s_IProtoErrorStackFrame, __pyx_k_IProtoErrorStackFrame, sizeof(__pyx_k_IProtoErrorStackFrame), 0, 0, 1, 1},
    {&__pyx_n_s_IProtoErrorStackFrame___reduce_c, __pyx_k_IProtoErrorStackFrame___reduce_c, sizeof(__pyx_k_IProtoErrorStackFrame___reduce_c), 0, 0, 1, 1},
    {&__pyx_n_s_IProtoErrorStackFrame___setstate, __pyx_k_IProtoErrorStackFrame___setstate, sizeof(__pyx_k_IProtoErrorStackFrame___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_IProtoError___reduce_cython, __pyx_k_IProtoError___reduce_cython, sizeof(__pyx_k_IProtoError___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_IProtoError___setstate_cython, __pyx_k_IProtoError___setstate_cython, sizeof(__pyx_k_IProtoError___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_IProto_auth_type_s, __pyx_k_IProto_auth_type_s, sizeof(__pyx_k_IProto_auth_type_s), 0, 1, 0, 0},
    {&__pyx_kp_u_IProto_features_available_s, __pyx_k_IProto_features_available_s, sizeof(__pyx_k_IProto_features_available_s), 0, 1, 0, 0},
    {&__pyx_kp_u_IProto_version_s, __pyx_k_IProto_version_s, sizeof(__pyx_k_IProto_version_s), 0, 1, 0, 0},
    {&__pyx_kp_u_Impossible_condition_happened_Pl, __pyx_k_Impossible_condition_happened_Pl, sizeof(__pyx_k_Impossible_condition_happened_Pl), 0, 1, 0, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_k_Incompatible_checksums_0x_x_vs_0_10, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_10), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_k_Incompatible_checksums_0x_x_vs_0_11, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_11), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12, __pyx_k_Incompatible_checksums_0x_x_vs_0_12, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_12), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13, __pyx_k_Incompatible_checksums_0x_x_vs_0_13, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_13), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14, __pyx_k_Incompatible_checksums_0x_x_vs_0_14, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_14), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15, __pyx_k_Incompatible_checksums_0x_x_vs_0_15, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_15), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16, __pyx_k_Incompatible_checksums_0x_x_vs_0_16, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_16), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_k_Incompatible_checksums_0x_x_vs_0_4, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_4), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_k_Incompatible_checksums_0x_x_vs_0_5, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_5), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_k_Incompatible_checksums_0x_x_vs_0_6, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_6), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_k_Incompatible_checksums_0x_x_vs_0_7, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_7), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8, __pyx_k_Incompatible_checksums_0x_x_vs_0_8, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_8), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9, __pyx_k_Incompatible_checksums_0x_x_vs_0_9, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_9), 0, 0, 1, 0},
    {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
    {&__pyx_kp_u_Index_must_be_either_str_or_int, __pyx_k_Index_must_be_either_str_or_int, sizeof(__pyx_k_Index_must_be_either_str_or_int), 0, 1, 0, 0},
    {&__pyx_kp_u_Index_not_found_in_space, __pyx_k_Index_not_found_in_space, sizeof(__pyx_k_Index_not_found_in_space), 0, 1, 0, 0},
    {&__pyx_kp_u_Index_s_not_found_in_space_s_s_C, __pyx_k_Index_s_not_found_in_space_s_s_C, sizeof(__pyx_k_Index_s_not_found_in_space_s_s_C), 0, 1, 0, 0},
    {&__pyx_n_s_InsertRequest, __pyx_k_InsertRequest, sizeof(__pyx_k_InsertRequest), 0, 0, 1, 1},
    {&__pyx_n_s_InsertRequest___reduce_cython, __pyx_k_InsertRequest___reduce_cython, sizeof(__pyx_k_InsertRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_InsertRequest___setstate_cython, __pyx_k_InsertRequest___setstate_cython, sizeof(__pyx_k_InsertRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_IntEnum, __pyx_k_IntEnum, sizeof(__pyx_k_IntEnum), 0, 0, 1, 1},
    {&__pyx_n_s_Iterator, __pyx_k_Iterator, sizeof(__pyx_k_Iterator), 0, 0, 1, 1},
    {&__pyx_kp_u_Iterator_is_of_unsupported_type, __pyx_k_Iterator_is_of_unsupported_type, sizeof(__pyx_k_Iterator_is_of_unsupported_type), 0, 1, 0, 0},
    {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
    {&__pyx_n_s_LE, __pyx_k_LE, sizeof(__pyx_k_LE), 0, 0, 1, 1},
    {&__pyx_n_s_LT, __pyx_k_LT, sizeof(__pyx_k_LT), 0, 0, 1, 1},
    {&__pyx_kp_u_Lost_connection_to_Tarantool, __pyx_k_Lost_connection_to_Tarantool, sizeof(__pyx_k_Lost_connection_to_Tarantool), 0, 1, 0, 0},
    {&__pyx_kp_u_Lost_connection_to_Tarantool_2, __pyx_k_Lost_connection_to_Tarantool_2, sizeof(__pyx_k_Lost_connection_to_Tarantool_2), 0, 1, 0, 0},
    {&__pyx_kp_u_Lost_connection_to_Tarantool_3, __pyx_k_Lost_connection_to_Tarantool_3, sizeof(__pyx_k_Lost_connection_to_Tarantool_3), 0, 1, 0, 0},
    {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
    {&__pyx_n_s_Metadata, __pyx_k_Metadata, sizeof(__pyx_k_Metadata), 0, 0, 1, 1},
    {&__pyx_n_s_Metadata___reduce_cython, __pyx_k_Metadata___reduce_cython, sizeof(__pyx_k_Metadata___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Metadata___setstate_cython, __pyx_k_Metadata___setstate_cython, sizeof(__pyx_k_Metadata___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Metadata_fields_count, __pyx_k_Metadata_fields_count, sizeof(__pyx_k_Metadata_fields_count), 0, 1, 0, 0},
    {&__pyx_n_s_NEIGHBOR, __pyx_k_NEIGHBOR, sizeof(__pyx_k_NEIGHBOR), 0, 0, 1, 1},
    {&__pyx_n_s_OVERLAPS, __pyx_k_OVERLAPS, sizeof(__pyx_k_OVERLAPS), 0, 0, 1, 1},
    {&__pyx_kp_u_Offset_incorrect_Got_use_len, __pyx_k_Offset_incorrect_Got_use_len, sizeof(__pyx_k_Offset_incorrect_Got_use_len), 0, 1, 0, 0},
    {&__pyx_kp_u_Operation_field_no_must_be_of_ei, __pyx_k_Operation_field_no_must_be_of_ei, sizeof(__pyx_k_Operation_field_no_must_be_of_ei), 0, 1, 0, 0},
    {&__pyx_kp_u_Operation_length_must_be_at_leas, __pyx_k_Operation_length_must_be_at_leas, sizeof(__pyx_k_Operation_length_must_be_at_leas), 0, 1, 0, 0},
    {&__pyx_kp_u_Operation_type_must_of_a_str_or, __pyx_k_Operation_type_must_of_a_str_or, sizeof(__pyx_k_Operation_type_must_of_a_str_or), 0, 1, 0, 0},
    {&__pyx_n_s_Optional, __pyx_k_Optional, sizeof(__pyx_k_Optional), 0, 0, 1, 1},
    {&__pyx_kp_u_Part_len_must_be_2, __pyx_k_Part_len_must_be_2, sizeof(__pyx_k_Part_len_must_be_2), 0, 1, 0, 0},
    {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_PingRequest, __pyx_k_PingRequest, sizeof(__pyx_k_PingRequest), 0, 0, 1, 1},
    {&__pyx_n_s_PingRequest___reduce_cython, __pyx_k_PingRequest___reduce_cython, sizeof(__pyx_k_PingRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PingRequest___setstate_cython, __pyx_k_PingRequest___setstate_cython, sizeof(__pyx_k_PingRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PrepareRequest, __pyx_k_PrepareRequest, sizeof(__pyx_k_PrepareRequest), 0, 0, 1, 1},
    {&__pyx_n_s_PrepareRequest___reduce_cython, __pyx_k_PrepareRequest___reduce_cython, sizeof(__pyx_k_PrepareRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PrepareRequest___setstate_cython, __pyx_k_PrepareRequest___setstate_cython, sizeof(__pyx_k_PrepareRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Protocol, __pyx_k_Protocol, sizeof(__pyx_k_Protocol), 0, 0, 1, 1},
    {&__pyx_n_s_PushIterator, __pyx_k_PushIterator, sizeof(__pyx_k_PushIterator), 0, 0, 1, 1},
    {&__pyx_n_s_PushIterator___anext, __pyx_k_PushIterator___anext, sizeof(__pyx_k_PushIterator___anext), 0, 0, 1, 1},
    {&__pyx_kp_u_PushIterator___init___line_9, __pyx_k_PushIterator___init___line_9, sizeof(__pyx_k_PushIterator___init___line_9), 0, 1, 0, 0},
    {&__pyx_n_s_PushIterator___reduce_cython, __pyx_k_PushIterator___reduce_cython, sizeof(__pyx_k_PushIterator___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PushIterator___setstate_cython, __pyx_k_PushIterator___setstate_cython, sizeof(__pyx_k_PushIterator___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_REQ, __pyx_k_REQ, sizeof(__pyx_k_REQ), 0, 0, 1, 1},
    {&__pyx_n_s_ReadBuffer, __pyx_k_ReadBuffer, sizeof(__pyx_k_ReadBuffer), 0, 0, 1, 1},
    {&__pyx_n_s_ReadBuffer___reduce_cython, __pyx_k_ReadBuffer___reduce_cython, sizeof(__pyx_k_ReadBuffer___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ReadBuffer___setstate_cython, __pyx_k_ReadBuffer___setstate_cython, sizeof(__pyx_k_ReadBuffer___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Request_op_sync_schema_id_push, __pyx_k_Request_op_sync_schema_id_push, sizeof(__pyx_k_Request_op_sync_schema_id_push), 0, 1, 0, 0},
    {&__pyx_n_s_Response, __pyx_k_Response, sizeof(__pyx_k_Response), 0, 0, 1, 1},
    {&__pyx_n_s_Response___reduce_cython, __pyx_k_Response___reduce_cython, sizeof(__pyx_k_Response___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Response___setstate_cython, __pyx_k_Response___setstate_cython, sizeof(__pyx_k_Response___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Response_body_must_be_a_MP_MAP, __pyx_k_Response_body_must_be_a_MP_MAP, sizeof(__pyx_k_Response_body_must_be_a_MP_MAP), 0, 1, 0, 0},
    {&__pyx_n_s_Response_done, __pyx_k_Response_done, sizeof(__pyx_k_Response_done), 0, 0, 1, 1},
    {&__pyx_kp_u_Response_header_must_be_a_MP_MAP, __pyx_k_Response_header_must_be_a_MP_MAP, sizeof(__pyx_k_Response_header_must_be_a_MP_MAP), 0, 1, 0, 0},
    {&__pyx_n_s_RollbackRequest, __pyx_k_RollbackRequest, sizeof(__pyx_k_RollbackRequest), 0, 0, 1, 1},
    {&__pyx_n_s_RollbackRequest___reduce_cython, __pyx_k_RollbackRequest___reduce_cython, sizeof(__pyx_k_RollbackRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_RollbackRequest___setstate_cytho, __pyx_k_RollbackRequest___setstate_cytho, sizeof(__pyx_k_RollbackRequest___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
    {&__pyx_n_s_Schema, __pyx_k_Schema, sizeof(__pyx_k_Schema), 0, 0, 1, 1},
    {&__pyx_n_s_SchemaDummyIndex, __pyx_k_SchemaDummyIndex, sizeof(__pyx_k_SchemaDummyIndex), 0, 0, 1, 1},
    {&__pyx_n_s_SchemaDummySpace, __pyx_k_SchemaDummySpace, sizeof(__pyx_k_SchemaDummySpace), 0, 0, 1, 1},
    {&__pyx_n_s_SchemaIndex, __pyx_k_SchemaIndex, sizeof(__pyx_k_SchemaIndex), 0, 0, 1, 1},
    {&__pyx_n_s_SchemaIndex___reduce_cython, __pyx_k_SchemaIndex___reduce_cython, sizeof(__pyx_k_SchemaIndex___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SchemaIndex___setstate_cython, __pyx_k_SchemaIndex___setstate_cython, sizeof(__pyx_k_SchemaIndex___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SchemaSpace, __pyx_k_SchemaSpace, sizeof(__pyx_k_SchemaSpace), 0, 0, 1, 1},
    {&__pyx_n_s_SchemaSpace___reduce_cython, __pyx_k_SchemaSpace___reduce_cython, sizeof(__pyx_k_SchemaSpace___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SchemaSpace___setstate_cython, __pyx_k_SchemaSpace___setstate_cython, sizeof(__pyx_k_SchemaSpace___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Schema___reduce_cython, __pyx_k_Schema___reduce_cython, sizeof(__pyx_k_Schema___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Schema___setstate_cython, __pyx_k_Schema___setstate_cython, sizeof(__pyx_k_Schema___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Schema_fetch_stopped_connection, __pyx_k_Schema_fetch_stopped_connection, sizeof(__pyx_k_Schema_fetch_stopped_connection), 0, 1, 0, 0},
    {&__pyx_kp_u_Schema_fetch_timeout, __pyx_k_Schema_fetch_timeout, sizeof(__pyx_k_Schema_fetch_timeout), 0, 1, 0, 0},
    {&__pyx_kp_u_Schema_spaces, __pyx_k_Schema_spaces, sizeof(__pyx_k_Schema_spaces), 0, 1, 0, 0},
    {&__pyx_n_s_SelectRequest, __pyx_k_SelectRequest, sizeof(__pyx_k_SelectRequest), 0, 0, 1, 1},
    {&__pyx_n_s_SelectRequest___reduce_cython, __pyx_k_SelectRequest___reduce_cython, sizeof(__pyx_k_SelectRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SelectRequest___setstate_cython, __pyx_k_SelectRequest___setstate_cython, sizeof(__pyx_k_SelectRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Single_operation_must_be_a_tuple, __pyx_k_Single_operation_must_be_a_tuple, sizeof(__pyx_k_Single_operation_must_be_a_tuple), 0, 1, 0, 0},
    {&__pyx_kp_u_Space_must_be_either_str_or_int, __pyx_k_Space_must_be_either_str_or_int, sizeof(__pyx_k_Space_must_be_either_str_or_int), 0, 1, 0, 0},
    {&__pyx_kp_u_Space_not_found, __pyx_k_Space_not_found, sizeof(__pyx_k_Space_not_found), 0, 1, 0, 0},
    {&__pyx_kp_u_Space_s_not_found_Creating_dummy, __pyx_k_Space_s_not_found_Creating_dummy, sizeof(__pyx_k_Space_s_not_found_Creating_dummy), 0, 1, 0, 0},
    {&__pyx_kp_u_Space_with_id_not_found, __pyx_k_Space_with_id_not_found, sizeof(__pyx_k_Space_with_id_not_found), 0, 1, 0, 0},
    {&__pyx_kp_u_Splice_offset_must_be_int, __pyx_k_Splice_offset_must_be_int, sizeof(__pyx_k_Splice_offset_must_be_int), 0, 1, 0, 0},
    {&__pyx_kp_u_Splice_operation_must_have_lengt, __pyx_k_Splice_operation_must_have_lengt, sizeof(__pyx_k_Splice_operation_must_have_lengt), 0, 1, 0, 0},
    {&__pyx_kp_u_Splice_position_must_be_int, __pyx_k_Splice_position_must_be_int, sizeof(__pyx_k_Splice_position_must_be_int), 0, 1, 0, 0},
    {&__pyx_n_s_TCP_NODELAY, __pyx_k_TCP_NODELAY, sizeof(__pyx_k_TCP_NODELAY), 0, 0, 1, 1},
    {&__pyx_n_s_TarantoolDatabaseError, __pyx_k_TarantoolDatabaseError, sizeof(__pyx_k_TarantoolDatabaseError), 0, 0, 1, 1},
    {&__pyx_n_s_TarantoolNotConnectedError, __pyx_k_TarantoolNotConnectedError, sizeof(__pyx_k_TarantoolNotConnectedError), 0, 0, 1, 1},
    {&__pyx_n_s_TarantoolSchemaError, __pyx_k_TarantoolSchemaError, sizeof(__pyx_k_TarantoolSchemaError), 0, 0, 1, 1},
    {&__pyx_n_s_TarantoolTuple, __pyx_k_TarantoolTuple, sizeof(__pyx_k_TarantoolTuple), 0, 0, 1, 1},
    {&__pyx_kp_u_Tarantool_is_not_connected, __pyx_k_Tarantool_is_not_connected, sizeof(__pyx_k_Tarantool_is_not_connected), 0, 1, 0, 0},
    {&__pyx_kp_u_Tarantool_s_s_Authorization_fail, __pyx_k_Tarantool_s_s_Authorization_fail, sizeof(__pyx_k_Tarantool_s_s_Authorization_fail), 0, 1, 0, 0},
    {&__pyx_kp_u_Tarantool_s_s_Authorized_success, __pyx_k_Tarantool_s_s_Authorized_success, sizeof(__pyx_k_Tarantool_s_s_Authorized_success), 0, 1, 0, 0},
    {&__pyx_kp_u_Tarantool_s_s_Schema_fetch_faile, __pyx_k_Tarantool_s_s_Schema_fetch_faile, sizeof(__pyx_k_Tarantool_s_s_Schema_fetch_faile), 0, 1, 0, 0},
    {&__pyx_kp_u_Tarantool_s_s_Schema_fetch_succe, __pyx_k_Tarantool_s_s_Schema_fetch_succe, sizeof(__pyx_k_Tarantool_s_s_Schema_fetch_succe), 0, 1, 0, 0},
    {&__pyx_kp_u_Tarantool_s_s_identification_fai, __pyx_k_Tarantool_s_s_identification_fai, sizeof(__pyx_k_Tarantool_s_s_identification_fai), 0, 1, 0, 0},
    {&__pyx_kp_u_Tarantool_s_s_identified_success, __pyx_k_Tarantool_s_s_identified_success, sizeof(__pyx_k_Tarantool_s_s_identified_success), 0, 1, 0, 0},
    {&__pyx_kp_u_Time_zones_are_not_available_fro, __pyx_k_Time_zones_are_not_available_fro, sizeof(__pyx_k_Time_zones_are_not_available_fro), 0, 1, 0, 0},
    {&__pyx_n_s_TimeoutError, __pyx_k_TimeoutError, sizeof(__pyx_k_TimeoutError), 0, 0, 1, 1},
    {&__pyx_kp_u_Tuple_must_be_an_array_when_deco, __pyx_k_Tuple_must_be_an_array_when_deco, sizeof(__pyx_k_Tuple_must_be_an_array_when_deco), 0, 1, 0, 0},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_kp_u_Type_is_not_supported_for_encodi, __pyx_k_Type_is_not_supported_for_encodi, sizeof(__pyx_k_Type_is_not_supported_for_encodi), 0, 1, 0, 0},
    {&__pyx_n_s_UUID, __pyx_k_UUID, sizeof(__pyx_k_UUID), 0, 0, 1, 1},
    {&__pyx_kp_u_Unexpected_ext_type_d, __pyx_k_Unexpected_ext_type_d, sizeof(__pyx_k_Unexpected_ext_type_d), 0, 1, 0, 0},
    {&__pyx_kp_u_Unexpected_key_type_in_map_s, __pyx_k_Unexpected_key_type_in_map_s, sizeof(__pyx_k_Unexpected_key_type_in_map_s), 0, 1, 0, 0},
    {&__pyx_kp_u_Unexpected_obj_type_s, __pyx_k_Unexpected_obj_type_s, sizeof(__pyx_k_Unexpected_obj_type_s), 0, 1, 0, 0},
    {&__pyx_n_s_UnicodeDecodeError, __pyx_k_UnicodeDecodeError, sizeof(__pyx_k_UnicodeDecodeError), 0, 0, 1, 1},
    {&__pyx_kp_u_Unknown_key_with_code_d_in_heade, __pyx_k_Unknown_key_with_code_d_in_heade, sizeof(__pyx_k_Unknown_key_with_code_d_in_heade), 0, 1, 0, 0},
    {&__pyx_kp_u_Unknown_update_operation_type, __pyx_k_Unknown_update_operation_type, sizeof(__pyx_k_Unknown_update_operation_type), 0, 1, 0, 0},
    {&__pyx_n_s_UpdateRequest, __pyx_k_UpdateRequest, sizeof(__pyx_k_UpdateRequest), 0, 0, 1, 1},
    {&__pyx_n_s_UpdateRequest___reduce_cython, __pyx_k_UpdateRequest___reduce_cython, sizeof(__pyx_k_UpdateRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UpdateRequest___setstate_cython, __pyx_k_UpdateRequest___setstate_cython, sizeof(__pyx_k_UpdateRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UpsertRequest, __pyx_k_UpsertRequest, sizeof(__pyx_k_UpsertRequest), 0, 0, 1, 1},
    {&__pyx_n_s_UpsertRequest___reduce_cython, __pyx_k_UpsertRequest___reduce_cython, sizeof(__pyx_k_UpsertRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UpsertRequest___setstate_cython, __pyx_k_UpsertRequest___setstate_cython, sizeof(__pyx_k_UpsertRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_VERSION_STRING_REGEX, __pyx_k_VERSION_STRING_REGEX, sizeof(__pyx_k_VERSION_STRING_REGEX), 0, 0, 1, 1},
    {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
    {&__pyx_n_s_WriteBuffer, __pyx_k_WriteBuffer, sizeof(__pyx_k_WriteBuffer), 0, 0, 1, 1},
    {&__pyx_n_s_WriteBuffer___reduce_cython, __pyx_k_WriteBuffer___reduce_cython, sizeof(__pyx_k_WriteBuffer___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_WriteBuffer___setstate_cython, __pyx_k_WriteBuffer___setstate_cython, sizeof(__pyx_k_WriteBuffer___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_WriteBuffer_hex, __pyx_k_WriteBuffer_hex, sizeof(__pyx_k_WriteBuffer_hex), 0, 0, 1, 1},
    {&__pyx_kp_u__14, __pyx_k__14, sizeof(__pyx_k__14), 0, 1, 0, 0},
    {&__pyx_kp_u__17, __pyx_k__17, sizeof(__pyx_k__17), 0, 1, 0, 0},
    {&__pyx_kp_u__18, __pyx_k__18, sizeof(__pyx_k__18), 0, 1, 0, 0},
    {&__pyx_kp_u__19, __pyx_k__19, sizeof(__pyx_k__19), 0, 1, 0, 0},
    {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
    {&__pyx_n_s__211, __pyx_k__211, sizeof(__pyx_k__211), 0, 0, 1, 1},
    {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
    {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
    {&__pyx_n_s__65, __pyx_k__65, sizeof(__pyx_k__65), 0, 0, 1, 1},
    {&__pyx_n_s_add_done_callback, __pyx_k_add_done_callback, sizeof(__pyx_k_add_done_callback), 0, 0, 1, 1},
    {&__pyx_n_s_anext, __pyx_k_anext, sizeof(__pyx_k_anext), 0, 0, 1, 1},
    {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_s_as_tuple, __pyx_k_as_tuple, sizeof(__pyx_k_as_tuple), 0, 0, 1, 1},
    {&__pyx_n_u_ascii, __pyx_k_ascii, sizeof(__pyx_k_ascii), 0, 1, 0, 1},
    {&__pyx_n_s_asyncio, __pyx_k_asyncio, sizeof(__pyx_k_asyncio), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_tasks, __pyx_k_asyncio_tasks, sizeof(__pyx_k_asyncio_tasks), 0, 0, 1, 1},
    {&__pyx_n_s_asynctnt, __pyx_k_asynctnt, sizeof(__pyx_k_asynctnt), 0, 0, 1, 1},
    {&__pyx_n_s_asynctnt_exceptions, __pyx_k_asynctnt_exceptions, sizeof(__pyx_k_asynctnt_exceptions), 0, 0, 1, 1},
    {&__pyx_kp_s_asynctnt_iproto_buffer_pyx, __pyx_k_asynctnt_iproto_buffer_pyx, sizeof(__pyx_k_asynctnt_iproto_buffer_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_asynctnt_iproto_coreproto_pyx, __pyx_k_asynctnt_iproto_coreproto_pyx, sizeof(__pyx_k_asynctnt_iproto_coreproto_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_k_asynctnt_iproto_db_pyx, sizeof(__pyx_k_asynctnt_iproto_db_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_asynctnt_iproto_protocol, __pyx_k_asynctnt_iproto_protocol, sizeof(__pyx_k_asynctnt_iproto_protocol), 0, 0, 1, 1},
    {&__pyx_kp_s_asynctnt_iproto_protocol_pyx, __pyx_k_asynctnt_iproto_protocol_pyx, sizeof(__pyx_k_asynctnt_iproto_protocol_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_asynctnt_iproto_response_pyx, __pyx_k_asynctnt_iproto_response_pyx, sizeof(__pyx_k_asynctnt_iproto_response_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_asynctnt_log, __pyx_k_asynctnt_log, sizeof(__pyx_k_asynctnt_log), 0, 0, 1, 1},
    {&__pyx_n_s_auto_refetch_schema, __pyx_k_auto_refetch_schema, sizeof(__pyx_k_auto_refetch_schema), 0, 0, 1, 1},
    {&__pyx_n_s_await, __pyx_k_await, sizeof(__pyx_k_await), 0, 0, 1, 1},
    {&__pyx_n_s_b64decode, __pyx_k_b64decode, sizeof(__pyx_k_b64decode), 0, 0, 1, 1},
    {&__pyx_n_s_base64, __pyx_k_base64, sizeof(__pyx_k_base64), 0, 0, 1, 1},
    {&__pyx_n_s_begin, __pyx_k_begin, sizeof(__pyx_k_begin), 0, 0, 1, 1},
    {&__pyx_kp_u_body_data_type_must_be_a_MP_ARRA, __pyx_k_body_data_type_must_be_a_MP_ARRA, sizeof(__pyx_k_body_data_type_must_be_a_MP_ARRA), 0, 1, 0, 0},
    {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
    {&__pyx_n_s_call, __pyx_k_call, sizeof(__pyx_k_call), 0, 0, 1, 1},
    {&__pyx_n_s_call16, __pyx_k_call16, sizeof(__pyx_k_call16), 0, 0, 1, 1},
    {&__pyx_n_s_call_later, __pyx_k_call_later, sizeof(__pyx_k_call_later), 0, 0, 1, 1},
    {&__pyx_n_s_cancel, __pyx_k_cancel, sizeof(__pyx_k_cancel), 0, 0, 1, 1},
    {&__pyx_n_s_cancelled, __pyx_k_cancelled, sizeof(__pyx_k_cancelled), 0, 0, 1, 1},
    {&__pyx_n_s_check_schema_change, __pyx_k_check_schema_change, sizeof(__pyx_k_check_schema_change), 0, 0, 1, 1},
    {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_clear, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_s_closing, __pyx_k_closing, sizeof(__pyx_k_closing), 0, 0, 1, 1},
    {&__pyx_kp_u_code_type_must_be_a_MP_UINT, __pyx_k_code_type_must_be_a_MP_UINT, sizeof(__pyx_k_code_type_must_be_a_MP_UINT), 0, 1, 0, 0},
    {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
    {&__pyx_n_s_commit, __pyx_k_commit, sizeof(__pyx_k_commit), 0, 0, 1, 1},
    {&__pyx_n_s_compile, __pyx_k_compile, sizeof(__pyx_k_compile), 0, 0, 1, 1},
    {&__pyx_n_s_connected_fut, __pyx_k_connected_fut, sizeof(__pyx_k_connected_fut), 0, 0, 1, 1},
    {&__pyx_n_s_connection_lost, __pyx_k_connection_lost, sizeof(__pyx_k_connection_lost), 0, 0, 1, 1},
    {&__pyx_n_s_connection_made, __pyx_k_connection_made, sizeof(__pyx_k_connection_made), 0, 0, 1, 1},
    {&__pyx_n_s_create_db, __pyx_k_create_db, sizeof(__pyx_k_create_db), 0, 0, 1, 1},
    {&__pyx_n_s_create_future, __pyx_k_create_future, sizeof(__pyx_k_create_future), 0, 0, 1, 1},
    {&__pyx_n_s_create_future_fallback, __pyx_k_create_future_fallback, sizeof(__pyx_k_create_future_fallback), 0, 0, 1, 1},
    {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
    {&__pyx_n_s_data_received, __pyx_k_data_received, sizeof(__pyx_k_data_received), 0, 0, 1, 1},
    {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
    {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
    {&__pyx_n_s_decimal, __pyx_k_decimal, sizeof(__pyx_k_decimal), 0, 0, 1, 1},
    {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
    {&__pyx_n_s_delete, __pyx_k_delete, sizeof(__pyx_k_delete), 0, 0, 1, 1},
    {&__pyx_n_s_deque, __pyx_k_deque, sizeof(__pyx_k_deque), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
    {&__pyx_n_s_digest, __pyx_k_digest, sizeof(__pyx_k_digest), 0, 0, 1, 1},
    {&__pyx_n_s_digits, __pyx_k_digits, sizeof(__pyx_k_digits), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
    {&__pyx_n_s_done, __pyx_k_done, sizeof(__pyx_k_done), 0, 0, 1, 1},
    {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
    {&__pyx_n_s_encoding, __pyx_k_encoding, sizeof(__pyx_k_encoding), 0, 0, 1, 1},
    {&__pyx_kp_u_encoding_must_be_either_str_or_b, __pyx_k_encoding_must_be_either_str_or_b, sizeof(__pyx_k_encoding_must_be_either_str_or_b), 0, 1, 0, 0},
    {&__pyx_n_s_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
    {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
    {&__pyx_kp_u_errstr_type_must_be_a_MP_STR, __pyx_k_errstr_type_must_be_a_MP_STR, sizeof(__pyx_k_errstr_type_must_be_a_MP_STR), 0, 1, 0, 0},
    {&__pyx_n_s_eval, __pyx_k_eval, sizeof(__pyx_k_eval), 0, 0, 1, 1},
    {&__pyx_n_s_exc, __pyx_k_exc, sizeof(__pyx_k_exc), 0, 0, 1, 1},
    {&__pyx_kp_u_exceeded_timeout, __pyx_k_exceeded_timeout, sizeof(__pyx_k_exceeded_timeout), 0, 1, 0, 0},
    {&__pyx_n_s_exception, __pyx_k_exception, sizeof(__pyx_k_exception), 0, 0, 1, 1},
    {&__pyx_n_s_execute, __pyx_k_execute, sizeof(__pyx_k_execute), 0, 0, 1, 1},
    {&__pyx_n_s_exponent, __pyx_k_exponent, sizeof(__pyx_k_exponent), 0, 0, 1, 1},
    {&__pyx_n_s_expression, __pyx_k_expression, sizeof(__pyx_k_expression), 0, 0, 1, 1},
    {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
    {&__pyx_n_s_family, __pyx_k_family, sizeof(__pyx_k_family), 0, 0, 1, 1},
    {&__pyx_n_s_fetch_schema, __pyx_k_fetch_schema, sizeof(__pyx_k_fetch_schema), 0, 0, 1, 1},
    {&__pyx_n_u_field, __pyx_k_field, sizeof(__pyx_k_field), 0, 1, 0, 1},
    {&__pyx_kp_u_field_name_must_not_be_None, __pyx_k_field_name_must_not_be_None, sizeof(__pyx_k_field_name_must_not_be_None), 0, 1, 0, 0},
    {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
    {&__pyx_n_s_func_name, __pyx_k_func_name, sizeof(__pyx_k_func_name), 0, 0, 1, 1},
    {&__pyx_n_s_fut, __pyx_k_fut, sizeof(__pyx_k_fut), 0, 0, 1, 1},
    {&__pyx_n_s_futures, __pyx_k_futures, sizeof(__pyx_k_futures), 0, 0, 1, 1},
    {&__pyx_n_s_gather, __pyx_k_gather, sizeof(__pyx_k_gather), 0, 0, 1, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_gen_stream_id, __pyx_k_gen_stream_id, sizeof(__pyx_k_gen_stream_id), 0, 0, 1, 1},
    {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
    {&__pyx_n_s_get_common_db, __pyx_k_get_common_db, sizeof(__pyx_k_get_common_db), 0, 0, 1, 1},
    {&__pyx_n_s_get_extra_info, __pyx_k_get_extra_info, sizeof(__pyx_k_get_extra_info), 0, 0, 1, 1},
    {&__pyx_n_s_get_version, __pyx_k_get_version, sizeof(__pyx_k_get_version), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_kp_u_got, __pyx_k_got, sizeof(__pyx_k_got), 0, 1, 0, 0},
    {&__pyx_n_s_group, __pyx_k_group, sizeof(__pyx_k_group), 0, 0, 1, 1},
    {&__pyx_n_s_hashlib, __pyx_k_hashlib, sizeof(__pyx_k_hashlib), 0, 0, 1, 1},
    {&__pyx_n_s_hex, __pyx_k_hex, sizeof(__pyx_k_hex), 0, 0, 1, 1},
    {&__pyx_n_s_hex_locals_genexpr, __pyx_k_hex_locals_genexpr, sizeof(__pyx_k_hex_locals_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_host, __pyx_k_host, sizeof(__pyx_k_host), 0, 0, 1, 1},
    {&__pyx_kp_u_id_name_engine, __pyx_k_id_name_engine, sizeof(__pyx_k_id_name_engine), 0, 1, 0, 0},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
    {&__pyx_n_s_indexes, __pyx_k_indexes, sizeof(__pyx_k_indexes), 0, 0, 1, 1},
    {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
    {&__pyx_n_s_init_subclass, __pyx_k_init_subclass, sizeof(__pyx_k_init_subclass), 0, 0, 1, 1},
    {&__pyx_n_s_initial_read_buffer_size, __pyx_k_initial_read_buffer_size, sizeof(__pyx_k_initial_read_buffer_size), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_s_insert, __pyx_k_insert, sizeof(__pyx_k_insert), 0, 0, 1, 1},
    {&__pyx_n_s_inspect, __pyx_k_inspect, sizeof(__pyx_k_inspect), 0, 0, 1, 1},
    {&__pyx_kp_u_invalid_datetime_size_got_extra, __pyx_k_invalid_datetime_size_got_extra, sizeof(__pyx_k_invalid_datetime_size_got_extra), 0, 1, 0, 0},
    {&__pyx_kp_u_iproto_error_stack_frame_fields, __pyx_k_iproto_error_stack_frame_fields, sizeof(__pyx_k_iproto_error_stack_frame_fields), 0, 1, 0, 0},
    {&__pyx_n_s_is_connected, __pyx_k_is_connected, sizeof(__pyx_k_is_connected), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_n_s_is_fully_connected, __pyx_k_is_fully_connected, sizeof(__pyx_k_is_fully_connected), 0, 0, 1, 1},
    {&__pyx_n_u_is_nullable, __pyx_k_is_nullable, sizeof(__pyx_k_is_nullable), 0, 1, 0, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_isolation, __pyx_k_isolation, sizeof(__pyx_k_isolation), 0, 0, 1, 1},
    {&__pyx_n_s_iterator, __pyx_k_iterator, sizeof(__pyx_k_iterator), 0, 0, 1, 1},
    {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
    {&__pyx_n_s_last_stream_id, __pyx_k_last_stream_id, sizeof(__pyx_k_last_stream_id), 0, 0, 1, 1},
    {&__pyx_n_s_limit, __pyx_k_limit, sizeof(__pyx_k_limit), 0, 0, 1, 1},
    {&__pyx_n_s_logger, __pyx_k_logger, sizeof(__pyx_k_logger), 0, 0, 1, 1},
    {&__pyx_n_s_loop, __pyx_k_loop, sizeof(__pyx_k_loop), 0, 0, 1, 1},
    {&__pyx_n_s_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
    {&__pyx_n_s_match, __pyx_k_match, sizeof(__pyx_k_match), 0, 0, 1, 1},
    {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
    {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
    {&__pyx_n_s_mro_entries, __pyx_k_mro_entries, sizeof(__pyx_k_mro_entries), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_u_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 1, 0, 1},
    {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
    {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
    {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
    {&__pyx_n_s_on_authorized, __pyx_k_on_authorized, sizeof(__pyx_k_on_authorized), 0, 0, 1, 1},
    {&__pyx_n_s_on_connection_lost, __pyx_k_on_connection_lost, sizeof(__pyx_k_on_connection_lost), 0, 0, 1, 1},
    {&__pyx_n_s_on_connection_made, __pyx_k_on_connection_made, sizeof(__pyx_k_on_connection_made), 0, 0, 1, 1},
    {&__pyx_kp_u_on_data_received_expected_bytes, __pyx_k_on_data_received_expected_bytes, sizeof(__pyx_k_on_data_received_expected_bytes), 0, 1, 0, 0},
    {&__pyx_n_s_on_fetch, __pyx_k_on_fetch, sizeof(__pyx_k_on_fetch), 0, 0, 1, 1},
    {&__pyx_n_s_on_id, __pyx_k_on_id, sizeof(__pyx_k_on_id), 0, 0, 1, 1},
    {&__pyx_n_s_on_request_completed, __pyx_k_on_request_completed, sizeof(__pyx_k_on_request_completed), 0, 0, 1, 1},
    {&__pyx_n_s_on_request_timeout, __pyx_k_on_request_timeout, sizeof(__pyx_k_on_request_timeout), 0, 0, 1, 1},
    {&__pyx_n_s_operations, __pyx_k_operations, sizeof(__pyx_k_operations), 0, 0, 1, 1},
    {&__pyx_n_s_parse_metadata, __pyx_k_parse_metadata, sizeof(__pyx_k_parse_metadata), 0, 0, 1, 1},
    {&__pyx_n_s_parse_version, __pyx_k_parse_version, sizeof(__pyx_k_parse_version), 0, 0, 1, 1},
    {&__pyx_n_s_password, __pyx_k_password, sizeof(__pyx_k_password), 0, 0, 1, 1},
    {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
    {&__pyx_n_s_ping, __pyx_k_ping, sizeof(__pyx_k_ping), 0, 0, 1, 1},
    {&__pyx_n_s_popleft, __pyx_k_popleft, sizeof(__pyx_k_popleft), 0, 0, 1, 1},
    {&__pyx_n_s_port, __pyx_k_port, sizeof(__pyx_k_port), 0, 0, 1, 1},
    {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
    {&__pyx_n_s_prepare_2, __pyx_k_prepare_2, sizeof(__pyx_k_prepare_2), 0, 0, 1, 1},
    {&__pyx_n_s_push_subscribe, __pyx_k_push_subscribe, sizeof(__pyx_k_push_subscribe), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_AuthRequest, __pyx_k_pyx_unpickle_AuthRequest, sizeof(__pyx_k_pyx_unpickle_AuthRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_BaseRequest, __pyx_k_pyx_unpickle_BaseRequest, sizeof(__pyx_k_pyx_unpickle_BaseRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_BeginRequest, __pyx_k_pyx_unpickle_BeginRequest, sizeof(__pyx_k_pyx_unpickle_BeginRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_CallRequest, __pyx_k_pyx_unpickle_CallRequest, sizeof(__pyx_k_pyx_unpickle_CallRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_CommitRequest, __pyx_k_pyx_unpickle_CommitRequest, sizeof(__pyx_k_pyx_unpickle_CommitRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_CoreProtocol, __pyx_k_pyx_unpickle_CoreProtocol, sizeof(__pyx_k_pyx_unpickle_CoreProtocol), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_DeleteRequest, __pyx_k_pyx_unpickle_DeleteRequest, sizeof(__pyx_k_pyx_unpickle_DeleteRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_EvalRequest, __pyx_k_pyx_unpickle_EvalRequest, sizeof(__pyx_k_pyx_unpickle_EvalRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_ExecuteRequest, __pyx_k_pyx_unpickle_ExecuteRequest, sizeof(__pyx_k_pyx_unpickle_ExecuteRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_IDRequest, __pyx_k_pyx_unpickle_IDRequest, sizeof(__pyx_k_pyx_unpickle_IDRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_IProtoError, __pyx_k_pyx_unpickle_IProtoError, sizeof(__pyx_k_pyx_unpickle_IProtoError), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_IProtoErrorStackF, __pyx_k_pyx_unpickle_IProtoErrorStackF, sizeof(__pyx_k_pyx_unpickle_IProtoErrorStackF), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_InsertRequest, __pyx_k_pyx_unpickle_InsertRequest, sizeof(__pyx_k_pyx_unpickle_InsertRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PingRequest, __pyx_k_pyx_unpickle_PingRequest, sizeof(__pyx_k_pyx_unpickle_PingRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PrepareRequest, __pyx_k_pyx_unpickle_PrepareRequest, sizeof(__pyx_k_pyx_unpickle_PrepareRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PushIterator, __pyx_k_pyx_unpickle_PushIterator, sizeof(__pyx_k_pyx_unpickle_PushIterator), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_RollbackRequest, __pyx_k_pyx_unpickle_RollbackRequest, sizeof(__pyx_k_pyx_unpickle_RollbackRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_SelectRequest, __pyx_k_pyx_unpickle_SelectRequest, sizeof(__pyx_k_pyx_unpickle_SelectRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_UpdateRequest, __pyx_k_pyx_unpickle_UpdateRequest, sizeof(__pyx_k_pyx_unpickle_UpdateRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_UpsertRequest, __pyx_k_pyx_unpickle_UpsertRequest, sizeof(__pyx_k_pyx_unpickle_UpsertRequest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
    {&__pyx_n_s_query, __pyx_k_query, sizeof(__pyx_k_query), 0, 0, 1, 1},
    {&__pyx_kp_u_query_must_be_either_str_or_int, __pyx_k_query_must_be_either_str_or_int, sizeof(__pyx_k_query_must_be_either_str_or_int), 0, 1, 0, 0},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_re, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_refetch_schema, __pyx_k_refetch_schema, sizeof(__pyx_k_refetch_schema), 0, 0, 1, 1},
    {&__pyx_n_s_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 0, 1, 1},
    {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
    {&__pyx_n_s_repr___locals_lambda, __pyx_k_repr___locals_lambda, sizeof(__pyx_k_repr___locals_lambda), 0, 0, 1, 1},
    {&__pyx_n_s_req, __pyx_k_req, sizeof(__pyx_k_req), 0, 0, 1, 1},
    {&__pyx_n_s_request_timeout, __pyx_k_request_timeout, sizeof(__pyx_k_request_timeout), 0, 0, 1, 1},
    {&__pyx_n_s_response, __pyx_k_response, sizeof(__pyx_k_response), 0, 0, 1, 1},
    {&__pyx_n_u_response, __pyx_k_response, sizeof(__pyx_k_response), 0, 1, 0, 1},
    {&__pyx_n_s_response_2, __pyx_k_response_2, sizeof(__pyx_k_response_2), 0, 0, 1, 1},
    {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
    {&__pyx_n_s_return_exceptions, __pyx_k_return_exceptions, sizeof(__pyx_k_return_exceptions), 0, 0, 1, 1},
    {&__pyx_n_s_rollback, __pyx_k_rollback, sizeof(__pyx_k_rollback), 0, 0, 1, 1},
    {&__pyx_n_s_rowcount, __pyx_k_rowcount, sizeof(__pyx_k_rowcount), 0, 0, 1, 1},
    {&__pyx_kp_u_s_Tarantool_s_d_s, __pyx_k_s_Tarantool_s_d_s, sizeof(__pyx_k_s_Tarantool_s_d_s), 0, 1, 0, 0},
    {&__pyx_n_s_schema_id, __pyx_k_schema_id, sizeof(__pyx_k_schema_id), 0, 0, 1, 1},
    {&__pyx_kp_u_schema_id_type_must_be_a_MP_UINT, __pyx_k_schema_id_type_must_be_a_MP_UINT, sizeof(__pyx_k_schema_id_type_must_be_a_MP_UINT), 0, 1, 0, 0},
    {&__pyx_n_s_select, __pyx_k_select, sizeof(__pyx_k_select), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_kp_s_self_execute_cannot_be_converted, __pyx_k_self_execute_cannot_be_converted, sizeof(__pyx_k_self_execute_cannot_be_converted), 0, 0, 1, 0},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_kp_u_sequence_must_be_either_list_or, __pyx_k_sequence_must_be_either_list_or, sizeof(__pyx_k_sequence_must_be_either_list_or), 0, 1, 0, 0},
    {&__pyx_kp_u_sequence_must_be_either_list_tup, __pyx_k_sequence_must_be_either_list_tup, sizeof(__pyx_k_sequence_must_be_either_list_tup), 0, 1, 0, 0},
    {&__pyx_n_s_set, __pyx_k_set, sizeof(__pyx_k_set), 0, 0, 1, 1},
    {&__pyx_n_s_set_exception, __pyx_k_set_exception, sizeof(__pyx_k_set_exception), 0, 0, 1, 1},
    {&__pyx_n_s_set_name, __pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 0, 1, 1},
    {&__pyx_n_s_set_result, __pyx_k_set_result, sizeof(__pyx_k_set_result), 0, 0, 1, 1},
    {&__pyx_n_s_set_stream_id, __pyx_k_set_stream_id, sizeof(__pyx_k_set_stream_id), 0, 0, 1, 1},
    {&__pyx_n_s_setsockopt, __pyx_k_setsockopt, sizeof(__pyx_k_setsockopt), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_sha1, __pyx_k_sha1, sizeof(__pyx_k_sha1), 0, 0, 1, 1},
    {&__pyx_kp_u_sid_id_name_type_unique, __pyx_k_sid_id_name_type_unique, sizeof(__pyx_k_sid_id_name_type_unique), 0, 1, 0, 0},
    {&__pyx_n_s_sign, __pyx_k_sign, sizeof(__pyx_k_sign), 0, 0, 1, 1},
    {&__pyx_n_s_sock, __pyx_k_sock, sizeof(__pyx_k_sock), 0, 0, 1, 1},
    {&__pyx_n_s_socket, __pyx_k_socket, sizeof(__pyx_k_socket), 0, 0, 1, 1},
    {&__pyx_n_u_socket, __pyx_k_socket, sizeof(__pyx_k_socket), 0, 1, 0, 1},
    {&__pyx_n_s_space, __pyx_k_space, sizeof(__pyx_k_space), 0, 0, 1, 1},
    {&__pyx_n_s_spaces, __pyx_k_spaces, sizeof(__pyx_k_spaces), 0, 0, 1, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
    {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
    {&__pyx_n_s_stream_id, __pyx_k_stream_id, sizeof(__pyx_k_stream_id), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
    {&__pyx_n_s_sync, __pyx_k_sync, sizeof(__pyx_k_sync), 0, 0, 1, 1},
    {&__pyx_kp_u_sync_d_not_found, __pyx_k_sync_d_not_found, sizeof(__pyx_k_sync_d_not_found), 0, 1, 0, 0},
    {&__pyx_kp_u_sync_rowcount_data, __pyx_k_sync_rowcount_data, sizeof(__pyx_k_sync_rowcount_data), 0, 1, 0, 0},
    {&__pyx_kp_u_sync_type_must_be_a_MP_UINT, __pyx_k_sync_type_must_be_a_MP_UINT, sizeof(__pyx_k_sync_type_must_be_a_MP_UINT), 0, 1, 0, 0},
    {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_kp_u_the_buffer_is_in_read_only_mode, __pyx_k_the_buffer_is_in_read_only_mode, sizeof(__pyx_k_the_buffer_is_in_read_only_mode), 0, 1, 0, 0},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_timeout, __pyx_k_timeout, sizeof(__pyx_k_timeout), 0, 0, 1, 1},
    {&__pyx_n_s_timestamp, __pyx_k_timestamp, sizeof(__pyx_k_timestamp), 0, 0, 1, 1},
    {&__pyx_n_s_timezone, __pyx_k_timezone, sizeof(__pyx_k_timezone), 0, 0, 1, 1},
    {&__pyx_n_s_total_seconds, __pyx_k_total_seconds, sizeof(__pyx_k_total_seconds), 0, 0, 1, 1},
    {&__pyx_n_s_transport, __pyx_k_transport, sizeof(__pyx_k_transport), 0, 0, 1, 1},
    {&__pyx_n_s_tx_timeout, __pyx_k_tx_timeout, sizeof(__pyx_k_tx_timeout), 0, 0, 1, 1},
    {&__pyx_n_u_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 1, 0, 1},
    {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
    {&__pyx_kp_u_unexpected_exponent_type, __pyx_k_unexpected_exponent_type, sizeof(__pyx_k_unexpected_exponent_type), 0, 1, 0, 0},
    {&__pyx_kp_u_unexpected_type_of_part_must_be, __pyx_k_unexpected_type_of_part_must_be, sizeof(__pyx_k_unexpected_type_of_part_must_be), 0, 1, 0, 0},
    {&__pyx_kp_u_unknown_iproto_error_map_field_w, __pyx_k_unknown_iproto_error_map_field_w, sizeof(__pyx_k_unknown_iproto_error_map_field_w), 0, 1, 0, 0},
    {&__pyx_kp_u_unknown_iproto_error_stack_eleme, __pyx_k_unknown_iproto_error_stack_eleme, sizeof(__pyx_k_unknown_iproto_error_stack_eleme), 0, 1, 0, 0},
    {&__pyx_kp_u_unknown_key_in_body_map_s, __pyx_k_unknown_key_in_body_map_s, sizeof(__pyx_k_unknown_key_in_body_map_s), 0, 1, 0, 0},
    {&__pyx_kp_u_unknown_key_in_metadata_decoding, __pyx_k_unknown_key_in_metadata_decoding, sizeof(__pyx_k_unknown_key_in_metadata_decoding), 0, 1, 0, 0},
    {&__pyx_kp_u_unknown_key_in_sql_info_decoding, __pyx_k_unknown_key_in_sql_info_decoding, sizeof(__pyx_k_unknown_key_in_sql_info_decoding), 0, 1, 0, 0},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_s_upsert, __pyx_k_upsert, sizeof(__pyx_k_upsert), 0, 0, 1, 1},
    {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_username, __pyx_k_username, sizeof(__pyx_k_username), 0, 0, 1, 1},
    {&__pyx_n_s_utcoffset, __pyx_k_utcoffset, sizeof(__pyx_k_utcoffset), 0, 0, 1, 1},
    {&__pyx_kp_b_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 0, 0, 0},
    {&__pyx_n_s_uuid, __pyx_k_uuid, sizeof(__pyx_k_uuid), 0, 0, 1, 1},
    {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
    {&__pyx_n_s_ver, __pyx_k_ver, sizeof(__pyx_k_ver), 0, 0, 1, 1},
    {&__pyx_n_s_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 0, 1, 1},
    {&__pyx_n_s_wait, __pyx_k_wait, sizeof(__pyx_k_wait), 0, 0, 1, 1},
    {&__pyx_n_s_waiter, __pyx_k_waiter, sizeof(__pyx_k_waiter), 0, 0, 1, 1},
    {&__pyx_n_s_warning, __pyx_k_warning, sizeof(__pyx_k_warning), 0, 0, 1, 1},
    {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
    {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(1, 51, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(1, 196, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(1, 224, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(2, 173, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(3, 39, __pyx_L1_error)
  __pyx_builtin_BufferError = __Pyx_GetBuiltinName(__pyx_n_s_BufferError); if (!__pyx_builtin_BufferError) __PYX_ERR(3, 67, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(3, 106, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(4, 57, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(5, 112, __pyx_L1_error)
  __pyx_builtin_UnicodeDecodeError = __Pyx_GetBuiltinName(__pyx_n_s_UnicodeDecodeError); if (!__pyx_builtin_UnicodeDecodeError) __PYX_ERR(5, 209, __pyx_L1_error)
  __pyx_builtin_hex = __Pyx_GetBuiltinName(__pyx_n_s_hex); if (!__pyx_builtin_hex) __PYX_ERR(5, 644, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(6, 105, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "cpython/datetime.pxd":308
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')             # <<<<<<<<<<<<<<
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 * 
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_Time_zones_are_not_available_fro); if (unlikely(!__pyx_tuple_)) __PYX_ERR(8, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "asynctnt/iproto/buffer.pyx":39
 * 
 *         if self._view_count:  # pragma: nocover
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'Deallocating buffer with attached memoryviews')
 * 
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_u_Deallocating_buffer_with_attache); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(3, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "asynctnt/iproto/buffer.pyx":67
 *     cdef inline _check_readonly(self):  # pragma: nocover
 *         if self._view_count:
 *             raise BufferError('the buffer is in read-only mode')             # <<<<<<<<<<<<<<
 * 
 *     cdef inline len(self):
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_u_the_buffer_is_in_read_only_mode); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(3, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "asynctnt/iproto/requests/update.pyx":54
 *             op_len = cpython.list.PyList_GET_SIZE(operation)
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 'Single operation must be a tuple or list')
 *         if op_len < 3:
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_u_Single_operation_must_be_a_tuple); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(4, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "asynctnt/iproto/requests/update.pyx":57
 *                 'Single operation must be a tuple or list')
 *         if op_len < 3:
 *             raise IndexError(             # <<<<<<<<<<<<<<
 *                 'Operation length must be at least 3')
 * 
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_u_Operation_length_must_be_at_leas); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(4, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "asynctnt/iproto/requests/update.pyx":66
 *             str_temp = <bytes> op_type_str
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 'Operation type must of a str or bytes type')
 * 
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_Operation_type_must_of_a_str_or); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(4, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "asynctnt/iproto/requests/update.pyx":87
 *                 cpython.bytes.PyBytes_AsStringAndSize(str_temp, &field_str_c, &field_str_len)
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 'Operation field_no must be of either int or str type')
 * 
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_u_Operation_field_no_must_be_of_ei); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(4, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "asynctnt/iproto/requests/update.pyx":131
 *             op_argument = operation[4]
 *             if not isinstance(splice_position_obj, int):
 *                 raise TypeError('Splice position must be int')             # <<<<<<<<<<<<<<
 *             if not isinstance(splice_offset_obj, int):
 *                 raise TypeError('Splice offset must be int')
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_Splice_position_must_be_int); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(4, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "asynctnt/iproto/requests/update.pyx":133
 *                 raise TypeError('Splice position must be int')
 *             if not isinstance(splice_offset_obj, int):
 *                 raise TypeError('Splice offset must be int')             # <<<<<<<<<<<<<<
 * 
 *             splice_position = <uint32_t> splice_position_obj
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_Splice_offset_must_be_int); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(4, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "asynctnt/iproto/response.pyx":86
 *     cdef inline object pop_push(self):
 *         if not self._push_subscribe:
 *             raise RuntimeError('Cannot pop push from a non-async response')             # <<<<<<<<<<<<<<
 * 
 *         push = self._q_popleft()
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_u_Cannot_pop_push_from_a_non_async); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(5, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "asynctnt/iproto/response.pyx":113
 *             if len(data) > 10:
 *                 parts = map(lambda x: ', '.join(map(repr, x)), [
 *                     data[:5],             # <<<<<<<<<<<<<<
 *                     data[-2:]
 *                 ])
 */
  __pyx_slice__15 = PySlice_New(Py_None, __pyx_int_5, Py_None); if (unlikely(!__pyx_slice__15)) __PYX_ERR(5, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__15);
  __Pyx_GIVEREF(__pyx_slice__15);

  /* "asynctnt/iproto/response.pyx":114
 *                 parts = map(lambda x: ', '.join(map(repr, x)), [
 *                     data[:5],
 *                     data[-2:]             # <<<<<<<<<<<<<<
 *                 ])
 * 
 */
  __pyx_slice__16 = PySlice_New(__pyx_int_neg_2, Py_None, Py_None); if (unlikely(!__pyx_slice__16)) __PYX_ERR(5, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__16);
  __Pyx_GIVEREF(__pyx_slice__16);

  /* "asynctnt/iproto/response.pyx":306
 *         for i in range(size):
 *             if mp_typeof(b[0][0]) != MP_ARRAY:  # pragma: nocover
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     'Tuple must be an array when decoding as TarantoolTuple'
 *                 )
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_u_Tuple_must_be_an_array_when_deco); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(5, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "asynctnt/iproto/response.pyx":337
 * 
 *     if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover
 *         raise TypeError('Response header must be a MP_MAP')             # <<<<<<<<<<<<<<
 * 
 *     # parsing header
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_u_Response_header_must_be_a_MP_MAP); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(5, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "asynctnt/iproto/response.pyx":343
 *     for _ in range(size):
 *         if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *             raise TypeError('Header key must be a MP_UINT')             # <<<<<<<<<<<<<<
 * 
 *         key = mp_decode_uint(&b)
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_u_Header_key_must_be_a_MP_UINT); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(5, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "asynctnt/iproto/response.pyx":348
 *         if key == tarantool.IPROTO_REQUEST_TYPE:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError('code type must be a MP_UINT')             # <<<<<<<<<<<<<<
 * 
 *             hdr.code = <uint32_t> mp_decode_uint(&b)
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_u_code_type_must_be_a_MP_UINT); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(5, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "asynctnt/iproto/response.pyx":354
 *         elif key == tarantool.IPROTO_SYNC:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError('sync type must be a MP_UINT')             # <<<<<<<<<<<<<<
 * 
 *             hdr.sync = mp_decode_uint(&b)
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_u_sync_type_must_be_a_MP_UINT); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(5, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "asynctnt/iproto/response.pyx":359
 *         elif key == tarantool.IPROTO_SCHEMA_VERSION:
 *             if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
 *                 raise TypeError('schema_id type must be a MP_UINT')             # <<<<<<<<<<<<<<
 * 
 *             hdr.schema_id = mp_decode_uint(&b)
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_u_schema_id_type_must_be_a_MP_UINT); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(5, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "asynctnt/iproto/response.pyx":385
 *         field_map_size = mp_decode_map(b)
 *         if field_map_size == 0:  # pragma: nocover
 *             raise RuntimeError('Field map must contain at least '             # <<<<<<<<<<<<<<
 *                                '1 element - field_name')
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_u_Field_map_must_contain_at_least); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(5, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "asynctnt/iproto/response.pyx":432
 * 
 *                 else:  # pragma: nocover
 *                     raise TypeError(             # <<<<<<<<<<<<<<
 *                         "IPROTO_FIELD_SPAN must be either STR or NIL"
 *                     )
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_u_IPROTO_FIELD_SPAN_must_be_either); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(5, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "asynctnt/iproto/response.pyx":441
 * 
 *         if field.name is None:  # pragma: nocover
 *             raise RuntimeError('field.name must not be None')             # <<<<<<<<<<<<<<
 * 
 *         metadata.add(<int> field_id, field)
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_u_field_name_must_not_be_None); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(5, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "asynctnt/iproto/response.pyx":557
 * 
 *     if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover
 *         raise TypeError('Response body must be a MP_MAP')             # <<<<<<<<<<<<<<
 * 
 *     size = mp_decode_map(&b)
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_u_Response_body_must_be_a_MP_MAP); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(5, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "asynctnt/iproto/response.pyx":567
 *         if key == tarantool.IPROTO_ERROR_24:
 *             if mp_typeof(b[0]) != MP_STR:  # pragma: nocover
 *                 raise TypeError('errstr type must be a MP_STR')             # <<<<<<<<<<<<<<
 * 
 *             s = NULL
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_u_errstr_type_must_be_a_MP_STR); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(5, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "asynctnt/iproto/response.pyx":576
 *         elif key == tarantool.IPROTO_ERROR:
 *             if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover
 *                 raise TypeError('IPROTO_ERROR type must be a MP_MAP')             # <<<<<<<<<<<<<<
 * 
 *             resp.error = parse_iproto_error(&b, resp.encoding)
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_u_IPROTO_ERROR_type_must_be_a_MP_M); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(5, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "asynctnt/iproto/response.pyx":606
 *             field_map_size = mp_decode_map(&b)
 *             if field_map_size == 0:
 *                 raise RuntimeError('Field map must contain at least '             # <<<<<<<<<<<<<<
 *                                    '1 element - rowcount')
 * 
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_u_Field_map_must_contain_at_least_2); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(5, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "asynctnt/iproto/response.pyx":627
 *         elif key == tarantool.IPROTO_DATA:
 *             if mp_typeof(b[0]) != MP_ARRAY:  # pragma: nocover
 *                 raise TypeError('body data type must be a MP_ARRAY')             # <<<<<<<<<<<<<<
 *             data = _response_parse_body_data(&b, resp, req)
 *             if is_chunk:
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_u_body_data_type_must_be_a_MP_ARRA); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(5, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "asynctnt/iproto/db.pyx":252
 *             req.statement_id = <uint64_t> query
 *         else:
 *             raise TypeError('query must be either str or int')             # <<<<<<<<<<<<<<
 * 
 *         req.args = args
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_u_query_must_be_either_str_or_int); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(26, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "asynctnt/iproto/push.pyx":37
 *             BaseRequest request
 *         if not hasattr(fut, '_response'):
 *             raise ValueError('Future is invalid. Make sure to call with '             # <<<<<<<<<<<<<<
 *                              'a future returned from a method with '
 *                              'push_subscribe=True flag')
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_u_Future_is_invalid_Make_sure_to_c); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(27, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "asynctnt/iproto/push.pyx":54
 * 
 *     def __iter__(self):
 *         raise RuntimeError('Cannot use iter with PushIterator - use aiter')             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_u_Cannot_use_iter_with_PushIterato); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(27, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "asynctnt/iproto/push.pyx":57
 * 
 *     def __next__(self):
 *         raise RuntimeError('Cannot use next with PushIterator - use anext')             # <<<<<<<<<<<<<<
 * 
 *     def __aiter__(self):
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_kp_u_Cannot_use_next_with_PushIterato); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(27, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "asynctnt/iproto/coreproto.pyx":32
 *             self.encoding = encoding
 *         else:
 *             raise TypeError('encoding must be either str or bytes')             # <<<<<<<<<<<<<<
 * 
 *         initial_read_buffer_size = initial_read_buffer_size or 0x20000
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_u_encoding_must_be_either_str_or_b); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(28, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "asynctnt/iproto/coreproto.pyx":139
 *             self.rbuf.get_slice(ver_length,
 *                                 ver_length + SALT_LENGTH)
 *         )[:SCRAMBLE_SIZE]             # <<<<<<<<<<<<<<
 *         self.state = PROTOCOL_NORMAL
 *         self._on_greeting_received()
 */
  __pyx_slice__39 = PySlice_New(Py_None, __pyx_int_20, Py_None); if (unlikely(!__pyx_slice__39)) __PYX_ERR(28, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__39);
  __Pyx_GIVEREF(__pyx_slice__39);

  /* "asynctnt/iproto/protocol.pyx":139
 *         if self.post_con_state == POST_CONNECTION_ID:
 *             assert self.version is not None
 *             if self.version >= (2, 10, 0):             # <<<<<<<<<<<<<<
 *                 # send <id> request
 *                 self._do_id()
 */
  __pyx_tuple__40 = PyTuple_Pack(3, __pyx_int_2, __pyx_int_10, __pyx_int_0); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(6, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "asynctnt/iproto/protocol.pyx":250
 *         fut = self._db._id(0.0)
 * 
 *         def on_id(f):             # <<<<<<<<<<<<<<
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 */
  __pyx_tuple__41 = PyTuple_Pack(2, __pyx_n_s_f, __pyx_n_s_e); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(6, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_protocol_pyx, __pyx_n_s_on_id, 250, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(6, 250, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":272
 *         fut = self._db._auth(self.salt, username, password, 0)
 * 
 *         def on_authorized(f):             # <<<<<<<<<<<<<<
 *             if f.cancelled():
 *                 self._set_connection_error(asyncio.futures.CancelledError())
 */
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_protocol_pyx, __pyx_n_s_on_authorized, 272, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(6, 272, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":296
 *         self._schema_fetch_in_progress = True
 * 
 *         def on_fetch(f):             # <<<<<<<<<<<<<<
 *             self._schema_fetch_in_progress = False
 * 
 */
  __pyx_tuple__44 = PyTuple_Pack(4, __pyx_n_s_f, __pyx_n_s_e, __pyx_n_s_spaces, __pyx_n_s_indexes); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(6, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__44, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_protocol_pyx, __pyx_n_s_on_fetch, 296, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(6, 296, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":349
 *                     fut.set_exception(e)
 * 
 *         fut_vspace = self._db.select(SPACE_VSPACE, timeout=0,             # <<<<<<<<<<<<<<
 *                                      check_schema_change=False)
 *         fut_vindex = self._db.select(SPACE_VINDEX, timeout=0,
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_int_281); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(6, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "asynctnt/iproto/protocol.pyx":351
 *         fut_vspace = self._db.select(SPACE_VSPACE, timeout=0,
 *                                      check_schema_change=False)
 *         fut_vindex = self._db.select(SPACE_VINDEX, timeout=0,             # <<<<<<<<<<<<<<
 *                                      check_schema_change=False)
 *         gather_fut = asyncio.gather(fut_vspace, fut_vindex,
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_int_289); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(6, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "asynctnt/iproto/protocol.pyx":502
 *             return Iterator[iterator]
 *         else:
 *             raise TypeError('Iterator is of unsupported type '             # <<<<<<<<<<<<<<
 *                             '(asynctnt.Iterator, int, str)')
 * 
 */
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_u_Iterator_is_of_unsupported_type); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(6, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9001016, 0xc05ea8e, 0xfe5f18a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x9001016, 0xc05ea8e, 0xfe5f18a) = (check_schema_change, op, parse_as_tuples, parse_metadata, push_subscribe, schema_id, space, stream_id, sync, timeout_handle, waiter))" % __pyx_checksum
 */
  __pyx_tuple__49 = PyTuple_Pack(3, __pyx_int_150999062, __pyx_int_201714318, __pyx_int_266727818); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_tuple__50 = PyTuple_Pack(3, __pyx_int_47556637, __pyx_int_180945731, __pyx_int_215610203); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_tuple__51 = PyTuple_Pack(3, __pyx_int_184220263, __pyx_int_63007895, __pyx_int_151261227); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);
  __pyx_tuple__52 = PyTuple_Pack(3, __pyx_int_149085227, __pyx_int_103773939, __pyx_int_92266676); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_tuple__53 = PyTuple_Pack(3, __pyx_int_194105842, __pyx_int_267485924, __pyx_int_23039357); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);
  __pyx_tuple__54 = PyTuple_Pack(3, __pyx_int_255344631, __pyx_int_145510168, __pyx_int_244856512); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_tuple__55 = PyTuple_Pack(3, __pyx_int_135652490, __pyx_int_145344315, __pyx_int_25165850); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_tuple__56 = PyTuple_Pack(3, __pyx_int_209859363, __pyx_int_245819509, __pyx_int_4796468); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);
  __pyx_tuple__57 = PyTuple_Pack(3, __pyx_int_218707107, __pyx_int_214473120, __pyx_int_129177485); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_tuple__58 = PyTuple_Pack(3, __pyx_int_51694216, __pyx_int_105824685, __pyx_int_220368800); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);
  __pyx_tuple__59 = PyTuple_Pack(3, __pyx_int_53862661, __pyx_int_257562521, __pyx_int_170745181); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_tuple__60 = PyTuple_Pack(3, __pyx_int_51353055, __pyx_int_172954204, __pyx_int_264852134); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);
  __pyx_tuple__61 = PyTuple_Pack(3, __pyx_int_179467967, __pyx_int_192862663, __pyx_int_105088600); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_tuple__62 = PyTuple_Pack(3, __pyx_int_246409365, __pyx_int_219129172, __pyx_int_4879619); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);
  __pyx_tuple__63 = PyTuple_Pack(3, __pyx_int_159463343, __pyx_int_196154451, __pyx_int_225261924); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_tuple__64 = PyTuple_Pack(3, __pyx_int_144520718, __pyx_int_235564716, __pyx_int_253265777); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_tuple__66 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_tuple__68 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__73 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__73)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "asynctnt/iproto/buffer.pyx":54
 *         self._view_count -= 1
 * 
 *     def hex(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))
 *                         for i in range(self._length))
 */
  __pyx_tuple__78 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(3, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__78, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_buffer_pyx, __pyx_n_s_hex, 54, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(3, 54, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__81 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__81)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__83 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__83)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__84 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_2, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BaseRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__86 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__86)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PingRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__89 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__89)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CallRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__90 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__90)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__91 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__91)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EvalRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__93 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__93)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SelectRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__95 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__95)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InsertRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__96 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__96)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DeleteRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__98 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__98)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__99 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__99)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UpdateRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__101 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__101)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UpsertRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__103 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__103)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PrepareRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__104 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__104)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__105 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__105)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ExecuteRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__107 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__107)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IDRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__108 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__108)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AuthRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__110 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__110)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__111 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__111)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BeginRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__112 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__112)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__113 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__113)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CommitRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__114 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__114)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__115 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__115)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RollbackRequest__set_state(self, __pyx_state)
 */
  __pyx_codeobj__116 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__116)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__117 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__117)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IProtoErrorStackFrame__set_state(self, __pyx_state)
 */
  __pyx_codeobj__118 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__118)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__119 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__119)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IProtoError__set_state(self, __pyx_state)
 */
  __pyx_codeobj__120 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__120)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "asynctnt/iproto/response.pyx":160
 *         return self.stmt_id_
 * 
 *     def done(self):             # <<<<<<<<<<<<<<
 *         return self.code_ >= 0
 * 
 */
  __pyx_codeobj__121 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_response_pyx, __pyx_n_s_done, 160, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__121)) __PYX_ERR(5, 160, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__122 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__122)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__123 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__123)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "asynctnt/iproto/db.pyx":327
 *         return <int> self._stream_id
 * 
 *     def set_stream_id(self, int stream_id):             # <<<<<<<<<<<<<<
 *         self._stream_id = <uint64_t> stream_id
 * 
 */
  __pyx_tuple__124 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_stream_id); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(26, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);
  __pyx_codeobj__125 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__124, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_set_stream_id, 327, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__125)) __PYX_ERR(26, 327, __pyx_L1_error)

  /* "asynctnt/iproto/db.pyx":330
 *         self._stream_id = <uint64_t> stream_id
 * 
 *     def ping(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._ping(timeout)
 * 
 */
  __pyx_tuple__126 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_timeout); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(26, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);
  __pyx_codeobj__127 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__126, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_ping, 330, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__127)) __PYX_ERR(26, 330, __pyx_L1_error)
  __pyx_tuple__128 = PyTuple_Pack(1, __pyx_float_neg_1_0); if (unlikely(!__pyx_tuple__128)) __PYX_ERR(26, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);

  /* "asynctnt/iproto/db.pyx":333
 *         return self._ping(timeout)
 * 
 *     def call16(self,             # <<<<<<<<<<<<<<
 *                str func_name,
 *                object args=None,
 */
  __pyx_tuple__129 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_func_name, __pyx_n_s_args, __pyx_n_s_timeout, __pyx_n_s_push_subscribe); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(26, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);
  __pyx_codeobj__130 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__129, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_call16, 333, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__130)) __PYX_ERR(26, 333, __pyx_L1_error)
  __pyx_tuple__131 = PyTuple_Pack(3, Py_None, __pyx_float_neg_1_0, Py_False); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(26, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);

  /* "asynctnt/iproto/db.pyx":344
 *                           <bint> push_subscribe)
 * 
 *     def call(self,             # <<<<<<<<<<<<<<
 *              str func_name,
 *              object args=None,
 */
  __pyx_codeobj__132 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__129, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_call, 344, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__132)) __PYX_ERR(26, 344, __pyx_L1_error)

  /* "asynctnt/iproto/db.pyx":355
 *                           <bint> push_subscribe)
 * 
 *     def eval(self,             # <<<<<<<<<<<<<<
 *              str expression,
 *              object args=None,
 */
  __pyx_tuple__133 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_expression, __pyx_n_s_args, __pyx_n_s_timeout, __pyx_n_s_push_subscribe); if (unlikely(!__pyx_tuple__133)) __PYX_ERR(26, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);
  __pyx_codeobj__134 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__133, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_eval, 355, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__134)) __PYX_ERR(26, 355, __pyx_L1_error)

  /* "asynctnt/iproto/db.pyx":365
 *                           <bint> push_subscribe)
 * 
 *     def select(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key=None,
 */
  __pyx_tuple__135 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_space, __pyx_n_s_key, __pyx_n_s_offset, __pyx_n_s_limit, __pyx_n_s_index, __pyx_n_s_iterator, __pyx_n_s_timeout, __pyx_n_s_check_schema_change); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(26, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);
  __pyx_codeobj__136 = (PyObject*)__Pyx_PyCode_New(9, 0, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_select, 365, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__136)) __PYX_ERR(26, 365, __pyx_L1_error)
  __pyx_tuple__137 = PyTuple_Pack(7, Py_None, __pyx_int_0, __pyx_int_4294967295, __pyx_int_0, __pyx_int_0, __pyx_float_neg_1_0, Py_True); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(26, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);

  /* "asynctnt/iproto/db.pyx":377
 *                             timeout, check_schema_change)
 * 
 *     def insert(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object t,
 */
  __pyx_tuple__138 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_space, __pyx_n_s_t, __pyx_n_s_replace, __pyx_n_s_timeout); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(26, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);
  __pyx_codeobj__139 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__138, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_insert, 377, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__139)) __PYX_ERR(26, 377, __pyx_L1_error)
  __pyx_tuple__140 = PyTuple_Pack(2, Py_False, __pyx_float_neg_1_0); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(26, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);

  /* "asynctnt/iproto/db.pyx":384
 *         return self._insert(space, t, <bint> replace, timeout)
 * 
 *     def replace(self,             # <<<<<<<<<<<<<<
 *                 object space,
 *                 object t,
 */
  __pyx_tuple__141 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_space, __pyx_n_s_t, __pyx_n_s_timeout); if (unlikely(!__pyx_tuple__141)) __PYX_ERR(26, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);
  __pyx_codeobj__142 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__141, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_replace, 384, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__142)) __PYX_ERR(26, 384, __pyx_L1_error)

  /* "asynctnt/iproto/db.pyx":390
 *         return self._insert(space, t, <bint> True, timeout)
 * 
 *     def delete(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key,
 */
  __pyx_tuple__143 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_space, __pyx_n_s_key, __pyx_n_s_index, __pyx_n_s_timeout); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(26, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);
  __pyx_codeobj__144 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_delete, 390, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__144)) __PYX_ERR(26, 390, __pyx_L1_error)
  __pyx_tuple__145 = PyTuple_Pack(2, __pyx_int_0, __pyx_float_neg_1_0); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(26, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);

  /* "asynctnt/iproto/db.pyx":397
 *         return self._delete(space, index, key, timeout)
 * 
 *     def update(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key,
 */
  __pyx_tuple__146 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_space, __pyx_n_s_key, __pyx_n_s_operations, __pyx_n_s_index, __pyx_n_s_timeout); if (unlikely(!__pyx_tuple__146)) __PYX_ERR(26, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);
  __pyx_codeobj__147 = (PyObject*)__Pyx_PyCode_New(6, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__146, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_update, 397, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__147)) __PYX_ERR(26, 397, __pyx_L1_error)

  /* "asynctnt/iproto/db.pyx":405
 *         return self._update(space, index, key, operations, timeout)
 * 
 *     def upsert(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object t,
 */
  __pyx_tuple__148 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_space, __pyx_n_s_t, __pyx_n_s_operations, __pyx_n_s_timeout); if (unlikely(!__pyx_tuple__148)) __PYX_ERR(26, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);
  __pyx_codeobj__149 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__148, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_upsert, 405, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__149)) __PYX_ERR(26, 405, __pyx_L1_error)

  /* "asynctnt/iproto/db.pyx":412
 *         return self._upsert(space, t, operations, timeout)
 * 
 *     def execute(self,             # <<<<<<<<<<<<<<
 *                 object query,
 *                 object args,
 */
  __pyx_tuple__150 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_query, __pyx_n_s_args, __pyx_n_s_parse_metadata, __pyx_n_s_timeout); if (unlikely(!__pyx_tuple__150)) __PYX_ERR(26, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);
  __pyx_codeobj__151 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_execute, 412, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__151)) __PYX_ERR(26, 412, __pyx_L1_error)
  __pyx_tuple__152 = PyTuple_Pack(2, Py_True, __pyx_float_neg_1_0); if (unlikely(!__pyx_tuple__152)) __PYX_ERR(26, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__152);
  __Pyx_GIVEREF(__pyx_tuple__152);

  /* "asynctnt/iproto/db.pyx":419
 *         return self._execute(query, args, <bint> parse_metadata, timeout)
 * 
 *     def prepare(self,             # <<<<<<<<<<<<<<
 *                 object query,
 *                 bint parse_metadata=True,
 */
  __pyx_tuple__153 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_query, __pyx_n_s_parse_metadata, __pyx_n_s_timeout); if (unlikely(!__pyx_tuple__153)) __PYX_ERR(26, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__153);
  __Pyx_GIVEREF(__pyx_tuple__153);
  __pyx_codeobj__154 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__153, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_prepare, 419, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__154)) __PYX_ERR(26, 419, __pyx_L1_error)

  /* "asynctnt/iproto/db.pyx":425
 *         return self._prepare(query, <bint> parse_metadata, timeout)
 * 
 *     def begin(self,             # <<<<<<<<<<<<<<
 *               uint32_t isolation,
 *               float tx_timeout=0,
 */
  __pyx_tuple__155 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_isolation, __pyx_n_s_tx_timeout, __pyx_n_s_timeout); if (unlikely(!__pyx_tuple__155)) __PYX_ERR(26, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__155);
  __Pyx_GIVEREF(__pyx_tuple__155);
  __pyx_codeobj__156 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__155, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_begin, 425, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__156)) __PYX_ERR(26, 425, __pyx_L1_error)
  __pyx_tuple__157 = PyTuple_Pack(2, __pyx_float_0_0, __pyx_float_neg_1_0); if (unlikely(!__pyx_tuple__157)) __PYX_ERR(26, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);

  /* "asynctnt/iproto/db.pyx":431
 *         return self._begin(isolation, <double> tx_timeout, timeout)
 * 
 *     def commit(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._commit(timeout)
 * 
 */
  __pyx_codeobj__158 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__126, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_commit, 431, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__158)) __PYX_ERR(26, 431, __pyx_L1_error)

  /* "asynctnt/iproto/db.pyx":434
 *         return self._commit(timeout)
 * 
 *     def rollback(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._rollback(timeout)
 */
  __pyx_codeobj__159 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__126, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_db_pyx, __pyx_n_s_rollback, 434, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__159)) __PYX_ERR(26, 434, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__160 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__160)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__161 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__161)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__162 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__162)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PushIterator__set_state(self, __pyx_state)
 */
  __pyx_codeobj__163 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__163)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":15
 * from asynctnt.log import logger
 * 
 * VERSION_STRING_REGEX = re.compile(r'\s*Tarantool\s+([\d.]+)\s+.*')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__164 = PyTuple_Pack(1, __pyx_kp_u_s_Tarantool_s_d_s); if (unlikely(!__pyx_tuple__164)) __PYX_ERR(28, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__164);
  __Pyx_GIVEREF(__pyx_tuple__164);

  /* "asynctnt/iproto/coreproto.pyx":50
 *         return self.con_state == CONNECTION_FULL
 * 
 *     def is_connected(self):             # <<<<<<<<<<<<<<
 *         return self._is_connected()
 * 
 */
  __pyx_codeobj__165 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_coreproto_pyx, __pyx_n_s_is_connected, 50, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__165)) __PYX_ERR(28, 50, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":53
 *         return self._is_connected()
 * 
 *     def is_fully_connected(self):             # <<<<<<<<<<<<<<
 *         return self._is_fully_connected()
 * 
 */
  __pyx_codeobj__166 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_coreproto_pyx, __pyx_n_s_is_fully_connected, 53, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__166)) __PYX_ERR(28, 53, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":56
 *         return self._is_fully_connected()
 * 
 *     def get_version(self):             # <<<<<<<<<<<<<<
 *         return self.version
 * 
 */
  __pyx_codeobj__167 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_coreproto_pyx, __pyx_n_s_get_version, 56, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__167)) __PYX_ERR(28, 56, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":143
 *         self._on_greeting_received()
 * 
 *     def _parse_version(self, version):             # <<<<<<<<<<<<<<
 *         m = VERSION_STRING_REGEX.match(version.decode('ascii'))
 *         if m is not None:
 */
  __pyx_tuple__168 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_version, __pyx_n_s_m, __pyx_n_s_ver); if (unlikely(!__pyx_tuple__168)) __PYX_ERR(28, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__168);
  __Pyx_GIVEREF(__pyx_tuple__168);
  __pyx_codeobj__169 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__168, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_coreproto_pyx, __pyx_n_s_parse_version, 143, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__169)) __PYX_ERR(28, 143, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":163
 *     # asyncio callbacks
 * 
 *     def data_received(self, data):             # <<<<<<<<<<<<<<
 *         self._on_data_received(data)
 * 
 */
  __pyx_tuple__170 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_data); if (unlikely(!__pyx_tuple__170)) __PYX_ERR(28, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__170);
  __Pyx_GIVEREF(__pyx_tuple__170);
  __pyx_codeobj__171 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_coreproto_pyx, __pyx_n_s_data_received, 163, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__171)) __PYX_ERR(28, 163, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":166
 *         self._on_data_received(data)
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         self.transport = transport
 *         self.con_state = CONNECTION_CONNECTED
 */
  __pyx_tuple__172 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_transport, __pyx_n_s_sock); if (unlikely(!__pyx_tuple__172)) __PYX_ERR(28, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__172);
  __Pyx_GIVEREF(__pyx_tuple__172);
  __pyx_codeobj__173 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__172, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_coreproto_pyx, __pyx_n_s_connection_made, 166, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__173)) __PYX_ERR(28, 166, __pyx_L1_error)

  /* "asynctnt/iproto/coreproto.pyx":179
 *         self._on_connection_made()
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         self.con_state = CONNECTION_BAD
 *         self.version = None
 */
  __pyx_tuple__174 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__174)) __PYX_ERR(28, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__174);
  __Pyx_GIVEREF(__pyx_tuple__174);
  __pyx_codeobj__175 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__174, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_coreproto_pyx, __pyx_n_s_connection_lost, 179, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__175)) __PYX_ERR(28, 179, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__176 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__176)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CoreProtocol__set_state(self, __pyx_state)
 */
  __pyx_codeobj__177 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__177)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":108
 *             self.create_future = self._create_future_fallback
 * 
 *     def _create_future_fallback(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return asyncio.Future(loop=self.loop)
 * 
 */
  __pyx_codeobj__178 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_protocol_pyx, __pyx_n_s_create_future_fallback, 108, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__178)) __PYX_ERR(6, 108, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":425
 *         return self._last_stream_id
 * 
 *     def _on_request_timeout(self, waiter):             # <<<<<<<<<<<<<<
 *         cdef:
 *             BaseRequest req
 */
  __pyx_tuple__179 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_waiter, __pyx_n_s_req, __pyx_n_s_response_2); if (unlikely(!__pyx_tuple__179)) __PYX_ERR(6, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__179);
  __Pyx_GIVEREF(__pyx_tuple__179);
  __pyx_codeobj__180 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__179, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_protocol_pyx, __pyx_n_s_on_request_timeout, 425, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__180)) __PYX_ERR(6, 425, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":442
 *         )
 * 
 *     def _on_request_completed(self, fut):             # <<<<<<<<<<<<<<
 *         cdef BaseRequest req = (<Response> fut._response).request_
 *         fut._response = None
 */
  __pyx_tuple__181 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_fut, __pyx_n_s_req); if (unlikely(!__pyx_tuple__181)) __PYX_ERR(6, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__181);
  __Pyx_GIVEREF(__pyx_tuple__181);
  __pyx_codeobj__182 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__181, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_protocol_pyx, __pyx_n_s_on_request_completed, 442, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__182)) __PYX_ERR(6, 442, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":473
 *         return Db.create(self, stream_id)
 * 
 *     def create_db(self, bint gen_stream_id = False):             # <<<<<<<<<<<<<<
 *         return self._create_db(gen_stream_id)
 * 
 */
  __pyx_tuple__183 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_gen_stream_id); if (unlikely(!__pyx_tuple__183)) __PYX_ERR(6, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__183);
  __Pyx_GIVEREF(__pyx_tuple__183);
  __pyx_codeobj__184 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__183, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_protocol_pyx, __pyx_n_s_create_db, 473, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__184)) __PYX_ERR(6, 473, __pyx_L1_error)
  __pyx_tuple__185 = PyTuple_Pack(1, Py_False); if (unlikely(!__pyx_tuple__185)) __PYX_ERR(6, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__185);
  __Pyx_GIVEREF(__pyx_tuple__185);

  /* "asynctnt/iproto/protocol.pyx":476
 *         return self._create_db(gen_stream_id)
 * 
 *     def get_common_db(self):             # <<<<<<<<<<<<<<
 *         return self._db
 * 
 */
  __pyx_codeobj__186 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_protocol_pyx, __pyx_n_s_get_common_db, 476, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__186)) __PYX_ERR(6, 476, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":515
 *         return self._refetch_schema_future
 * 
 *     def refetch_schema(self):             # <<<<<<<<<<<<<<
 *         return self._refetch_schema()
 * 
 */
  __pyx_codeobj__187 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_asynctnt_iproto_protocol_pyx, __pyx_n_s_refetch_schema, 515, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__187)) __PYX_ERR(6, 515, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__188 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__188)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__189 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__189)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_BaseRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__190 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__190)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__190);
  __Pyx_GIVEREF(__pyx_tuple__190);
  __pyx_codeobj__191 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_BaseRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__191)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__192 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PingRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__192)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__193 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CallRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__193)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__194 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_EvalRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__194)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__195 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_SelectRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__195)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__196 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_InsertRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__196)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__197 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_DeleteRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__197)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__198 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_UpdateRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__198)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__199 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_UpsertRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__199)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__200 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PrepareRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__200)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__201 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ExecuteRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__201)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__202 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IDRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__202)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__203 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_AuthRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__203)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__204 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_BeginRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__204)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__205 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CommitRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__205)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__206 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_RollbackRequest, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__206)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__207 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IProtoErrorStackF, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__207)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__208 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IProtoError, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__208)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__209 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PushIterator, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__209)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__210 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__190, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CoreProtocol, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__210)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(6, 1, __pyx_L1_error);
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_float_neg_1_0 = PyFloat_FromDouble(-1.0); if (unlikely(!__pyx_float_neg_1_0)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_11 = PyInt_FromLong(11); if (unlikely(!__pyx_int_11)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_20 = PyInt_FromLong(20); if (unlikely(!__pyx_int_20)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_281 = PyInt_FromLong(281); if (unlikely(!__pyx_int_281)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_289 = PyInt_FromLong(289); if (unlikely(!__pyx_int_289)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_4796468 = PyInt_FromLong(4796468L); if (unlikely(!__pyx_int_4796468)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_4879619 = PyInt_FromLong(4879619L); if (unlikely(!__pyx_int_4879619)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_23039357 = PyInt_FromLong(23039357L); if (unlikely(!__pyx_int_23039357)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_25165850 = PyInt_FromLong(25165850L); if (unlikely(!__pyx_int_25165850)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_47556637 = PyInt_FromLong(47556637L); if (unlikely(!__pyx_int_47556637)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_51353055 = PyInt_FromLong(51353055L); if (unlikely(!__pyx_int_51353055)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_51694216 = PyInt_FromLong(51694216L); if (unlikely(!__pyx_int_51694216)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_53862661 = PyInt_FromLong(53862661L); if (unlikely(!__pyx_int_53862661)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_63007895 = PyInt_FromLong(63007895L); if (unlikely(!__pyx_int_63007895)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_92266676 = PyInt_FromLong(92266676L); if (unlikely(!__pyx_int_92266676)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_103773939 = PyInt_FromLong(103773939L); if (unlikely(!__pyx_int_103773939)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_105088600 = PyInt_FromLong(105088600L); if (unlikely(!__pyx_int_105088600)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_105824685 = PyInt_FromLong(105824685L); if (unlikely(!__pyx_int_105824685)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_129177485 = PyInt_FromLong(129177485L); if (unlikely(!__pyx_int_129177485)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_135652490 = PyInt_FromLong(135652490L); if (unlikely(!__pyx_int_135652490)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_144520718 = PyInt_FromLong(144520718L); if (unlikely(!__pyx_int_144520718)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_145344315 = PyInt_FromLong(145344315L); if (unlikely(!__pyx_int_145344315)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_145510168 = PyInt_FromLong(145510168L); if (unlikely(!__pyx_int_145510168)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_149085227 = PyInt_FromLong(149085227L); if (unlikely(!__pyx_int_149085227)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_150999062 = PyInt_FromLong(150999062L); if (unlikely(!__pyx_int_150999062)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_151261227 = PyInt_FromLong(151261227L); if (unlikely(!__pyx_int_151261227)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_159463343 = PyInt_FromLong(159463343L); if (unlikely(!__pyx_int_159463343)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_170745181 = PyInt_FromLong(170745181L); if (unlikely(!__pyx_int_170745181)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_172954204 = PyInt_FromLong(172954204L); if (unlikely(!__pyx_int_172954204)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_179467967 = PyInt_FromLong(179467967L); if (unlikely(!__pyx_int_179467967)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_180945731 = PyInt_FromLong(180945731L); if (unlikely(!__pyx_int_180945731)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_184220263 = PyInt_FromLong(184220263L); if (unlikely(!__pyx_int_184220263)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_192862663 = PyInt_FromLong(192862663L); if (unlikely(!__pyx_int_192862663)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_194105842 = PyInt_FromLong(194105842L); if (unlikely(!__pyx_int_194105842)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_196154451 = PyInt_FromLong(196154451L); if (unlikely(!__pyx_int_196154451)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_201714318 = PyInt_FromLong(201714318L); if (unlikely(!__pyx_int_201714318)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_209859363 = PyInt_FromLong(209859363L); if (unlikely(!__pyx_int_209859363)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_214473120 = PyInt_FromLong(214473120L); if (unlikely(!__pyx_int_214473120)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_215610203 = PyInt_FromLong(215610203L); if (unlikely(!__pyx_int_215610203)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_218707107 = PyInt_FromLong(218707107L); if (unlikely(!__pyx_int_218707107)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_219129172 = PyInt_FromLong(219129172L); if (unlikely(!__pyx_int_219129172)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_220368800 = PyInt_FromLong(220368800L); if (unlikely(!__pyx_int_220368800)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_225261924 = PyInt_FromLong(225261924L); if (unlikely(!__pyx_int_225261924)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_235564716 = PyInt_FromLong(235564716L); if (unlikely(!__pyx_int_235564716)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_244856512 = PyInt_FromLong(244856512L); if (unlikely(!__pyx_int_244856512)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_245819509 = PyInt_FromLong(245819509L); if (unlikely(!__pyx_int_245819509)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_246409365 = PyInt_FromLong(246409365L); if (unlikely(!__pyx_int_246409365)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_253265777 = PyInt_FromLong(253265777L); if (unlikely(!__pyx_int_253265777)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_255344631 = PyInt_FromLong(255344631L); if (unlikely(!__pyx_int_255344631)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_257562521 = PyInt_FromLong(257562521L); if (unlikely(!__pyx_int_257562521)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_264852134 = PyInt_FromLong(264852134L); if (unlikely(!__pyx_int_264852134)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_266727818 = PyInt_FromLong(266727818L); if (unlikely(!__pyx_int_266727818)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_267485924 = PyInt_FromLong(267485924L); if (unlikely(!__pyx_int_267485924)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_4294967295 = PyInt_FromString((char *)"4294967295", 0, 0); if (unlikely(!__pyx_int_4294967295)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_int_neg_2 = PyInt_FromLong(-2); if (unlikely(!__pyx_int_neg_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("encode_unicode_string", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_encode_unicode_string, "PyObject *(PyObject *, PyObject *)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("decode_string", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_decode_string, "PyObject *(PyObject *, PyObject *)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dict_to_list_fields", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_dict_to_list_fields, "PyObject *(PyObject *, struct C_Metadata *, int)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("decimal_len", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_decimal_len, "uint32_t (int, uint32_t)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("decimal_encode", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_decimal_encode, "char *(char *, uint32_t, uint8_t, PyObject *, int)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("decimal_decode", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_decimal_decode, "PyObject *(char const **, uint32_t)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("uuid_decode", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_uuid_decode, "PyObject *(char const **, uint32_t)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("datetime_zero", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_datetime_zero, "void (struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("datetime_len", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_datetime_len, "uint32_t (struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("datetime_encode", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_datetime_encode, "char *(char *, struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("datetime_decode", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_datetime_decode, "int (char const **, uint32_t, struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("datetime_from_py", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_datetime_from_py, "void (PyDateTime_DateTime *, struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("datetime_to_py", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_datetime_to_py, "PyObject *(struct __pyx_t_8asynctnt_6iproto_8protocol_IProtoDateTime *)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("encode_key_sequence", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_encode_key_sequence, "char *(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *, struct C_Metadata *, int)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("encode_update_ops", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_encode_update_ops, "char *(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("encode_request_update", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_encode_request_update, "int (struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *, PyObject *, PyObject *, int)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("response_parse_header", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_response_parse_header, "Py_ssize_t (char const *, uint32_t, struct __pyx_t_8asynctnt_6iproto_8protocol_Header *)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("response_parse_body", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_response_parse_body, "Py_ssize_t (char const *, uint32_t, struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, int)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("parse_iproto_error", (void (*)(void))__pyx_f_8asynctnt_6iproto_8protocol_parse_iproto_error, "struct __pyx_obj_8asynctnt_6iproto_8protocol_IProtoError *(char const **, PyObject *)") < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_Field = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_Field_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_Field)) __PYX_ERR(1, 13, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_Field_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_Field) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_Field = &__pyx_type_8asynctnt_6iproto_8protocol_Field;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_Field) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_Field->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_Field->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_Field->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_Field->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Field, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_Field) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_Field) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_Metadata = &__pyx_vtable_8asynctnt_6iproto_8protocol_Metadata;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Metadata.len = (int (*)(struct C_Metadata *))__pyx_f_8asynctnt_6iproto_8protocol_8Metadata_len;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Metadata.add = (void (*)(struct C_Metadata *, int, struct __pyx_obj_8asynctnt_6iproto_8protocol_Field *))__pyx_f_8asynctnt_6iproto_8protocol_8Metadata_add;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Metadata.name_by_id = (PyObject *(*)(struct C_Metadata *, int))__pyx_f_8asynctnt_6iproto_8protocol_8Metadata_name_by_id;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Metadata.id_by_name = (int (*)(struct C_Metadata *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_8Metadata_id_by_name;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Metadata.id_by_name_safe = (int (*)(struct C_Metadata *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_8Metadata_id_by_name_safe;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_Metadata = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &C_Metadata_Type_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata)) __PYX_ERR(1, 29, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&C_Metadata_Type_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_Metadata) < 0) __PYX_ERR(1, 29, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_Metadata = &C_Metadata_Type;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata) < 0) __PYX_ERR(1, 29, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_Metadata->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_Metadata->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_Metadata->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata, __pyx_vtabptr_8asynctnt_6iproto_8protocol_Metadata) < 0) __PYX_ERR(1, 29, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_Metadata) < 0) __PYX_ERR(1, 29, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Metadata, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_Metadata) < 0) __PYX_ERR(1, 29, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_Metadata) < 0) __PYX_ERR(1, 29, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex)) __PYX_ERR(1, 70, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex) < 0) __PYX_ERR(1, 70, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex = &__pyx_type_8asynctnt_6iproto_8protocol_SchemaIndex;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex) < 0) __PYX_ERR(1, 70, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SchemaIndex, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex) < 0) __PYX_ERR(1, 70, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex) < 0) __PYX_ERR(1, 70, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex)) __PYX_ERR(11, 37, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex) < 0) __PYX_ERR(11, 37, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex = &__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummyIndex;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaIndex;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex) < 0) __PYX_ERR(11, 37, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SchemaDummyIndex, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummyIndex) < 0) __PYX_ERR(11, 37, __pyx_L1_error)
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_SchemaSpace = &__pyx_vtable_8asynctnt_6iproto_8protocol_SchemaSpace;
  __pyx_vtable_8asynctnt_6iproto_8protocol_SchemaSpace.add_index = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *, struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *))__pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_add_index;
  __pyx_vtable_8asynctnt_6iproto_8protocol_SchemaSpace.get_index = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *, PyObject *, struct __pyx_opt_args_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index *__pyx_optional_args))__pyx_f_8asynctnt_6iproto_8protocol_11SchemaSpace_get_index;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace)) __PYX_ERR(1, 88, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace = &__pyx_type_8asynctnt_6iproto_8protocol_SchemaSpace;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace, __pyx_vtabptr_8asynctnt_6iproto_8protocol_SchemaSpace) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SchemaSpace, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_SchemaDummySpace = &__pyx_vtable_8asynctnt_6iproto_8protocol_SchemaDummySpace;
  __pyx_vtable_8asynctnt_6iproto_8protocol_SchemaDummySpace.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_SchemaSpace;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace)) __PYX_ERR(11, 57, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace) < 0) __PYX_ERR(11, 57, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace = &__pyx_type_8asynctnt_6iproto_8protocol_SchemaDummySpace;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaSpace;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace) < 0) __PYX_ERR(11, 57, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace, __pyx_vtabptr_8asynctnt_6iproto_8protocol_SchemaDummySpace) < 0) __PYX_ERR(11, 57, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace) < 0) __PYX_ERR(11, 57, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SchemaDummySpace, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_SchemaDummySpace) < 0) __PYX_ERR(11, 57, __pyx_L1_error)
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_Schema = &__pyx_vtable_8asynctnt_6iproto_8protocol_Schema;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Schema.get_space = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_space;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Schema.create_dummy_space = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, int))__pyx_f_8asynctnt_6iproto_8protocol_6Schema_create_dummy_space;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Schema.get_or_create_space = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_6Schema_get_or_create_space;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Schema.parse_space = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaSpace *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse_space;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Schema.parse_index = (struct __pyx_obj_8asynctnt_6iproto_8protocol_SchemaIndex *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse_index;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Schema.clear = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *))__pyx_f_8asynctnt_6iproto_8protocol_6Schema_clear;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Schema.parse = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Schema *(*)(int64_t, PyObject *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_6Schema_parse;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_Schema = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_Schema_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_Schema)) __PYX_ERR(1, 145, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_Schema_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_Schema) < 0) __PYX_ERR(1, 145, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_Schema = &__pyx_type_8asynctnt_6iproto_8protocol_Schema;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_Schema) < 0) __PYX_ERR(1, 145, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_Schema->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_Schema->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_Schema->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_Schema->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_Schema, __pyx_vtabptr_8asynctnt_6iproto_8protocol_Schema) < 0) __PYX_ERR(1, 145, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_Schema) < 0) __PYX_ERR(1, 145, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Schema, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_Schema) < 0) __PYX_ERR(1, 145, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_Schema) < 0) __PYX_ERR(1, 145, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_WriteBuffer = &__pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.create = (struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *(*)(PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_create;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer._check_readonly = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__check_readonly;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.len = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_len;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.ensure_allocated = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, Py_ssize_t))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_ensure_allocated;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer._ensure_allocated = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, Py_ssize_t))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__ensure_allocated;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer._reallocate = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, Py_ssize_t))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer__reallocate;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.write_buffer = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_buffer;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.write_header = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, uint64_t, enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, int64_t, uint64_t))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_header;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.write_length = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_write_length;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_nil = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_nil;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_bool = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, int))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_bool;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_double = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, double))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_double;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_uint = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, uint64_t))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uint;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_int = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, int64_t))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_int;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_str = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, char const *, uint32_t))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_str;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_bin = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, char const *, uint32_t))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_bin;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_decimal = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_decimal;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_uuid = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_uuid;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_datetime = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_datetime;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_array = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, uint32_t))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_array;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_map = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, uint32_t))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_map;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_list = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_list;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_tuple = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_tuple;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_dict = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_dict;
  __pyx_vtable_8asynctnt_6iproto_8protocol_WriteBuffer.mp_encode_obj = (char *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *, char *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_11WriteBuffer_mp_encode_obj;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer)) __PYX_ERR(3, 24, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer->tp_as_buffer = &__pyx_tp_as_buffer_WriteBuffer;
  #elif defined(Py_bf_getbuffer) && defined(Py_bf_releasebuffer)
  /* PY_VERSION_HEX >= 0x03090000 || Py_LIMITED_API >= 0x030B0000 */
  #elif defined(_MSC_VER)
  #pragma message ("The buffer protocol is not supported in the Limited C-API < 3.11.")
  #else
  #warning "The buffer protocol is not supported in the Limited C-API < 3.11."
  #endif
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer) < 0) __PYX_ERR(3, 24, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer = &__pyx_type_8asynctnt_6iproto_8protocol_WriteBuffer;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer) < 0) __PYX_ERR(3, 24, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer, __pyx_vtabptr_8asynctnt_6iproto_8protocol_WriteBuffer) < 0) __PYX_ERR(3, 24, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer) < 0) __PYX_ERR(3, 24, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_WriteBuffer, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer) < 0) __PYX_ERR(3, 24, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer) < 0) __PYX_ERR(3, 24, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_ReadBuffer = &__pyx_vtable_8asynctnt_6iproto_8protocol_ReadBuffer;
  __pyx_vtable_8asynctnt_6iproto_8protocol_ReadBuffer.create = (struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *(*)(PyObject *, struct __pyx_opt_args_8asynctnt_6iproto_8protocol_10ReadBuffer_create *__pyx_optional_args))__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_create;
  __pyx_vtable_8asynctnt_6iproto_8protocol_ReadBuffer._reallocate = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t))__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer__reallocate;
  __pyx_vtable_8asynctnt_6iproto_8protocol_ReadBuffer.extend = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, char const *, size_t))__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_extend;
  __pyx_vtable_8asynctnt_6iproto_8protocol_ReadBuffer.move = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t))__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_move;
  __pyx_vtable_8asynctnt_6iproto_8protocol_ReadBuffer.move_offset = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, Py_ssize_t, size_t))__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_move_offset;
  __pyx_vtable_8asynctnt_6iproto_8protocol_ReadBuffer.get_slice = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t, size_t))__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice;
  __pyx_vtable_8asynctnt_6iproto_8protocol_ReadBuffer.get_slice_begin = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t))__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice_begin;
  __pyx_vtable_8asynctnt_6iproto_8protocol_ReadBuffer.get_slice_end = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_ReadBuffer *, size_t))__pyx_f_8asynctnt_6iproto_8protocol_10ReadBuffer_get_slice_end;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer)) __PYX_ERR(12, 8, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer) < 0) __PYX_ERR(12, 8, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer = &__pyx_type_8asynctnt_6iproto_8protocol_ReadBuffer;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer) < 0) __PYX_ERR(12, 8, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer, __pyx_vtabptr_8asynctnt_6iproto_8protocol_ReadBuffer) < 0) __PYX_ERR(12, 8, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer) < 0) __PYX_ERR(12, 8, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ReadBuffer, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer) < 0) __PYX_ERR(12, 8, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_ReadBuffer) < 0) __PYX_ERR(12, 8, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseRequest.metadata = (struct C_Metadata *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *))__pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_metadata;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseRequest.encode = (struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_encode;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseRequest.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_11BaseRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_BaseRequest_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest)) __PYX_ERR(13, 6, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_BaseRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest) < 0) __PYX_ERR(13, 6, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest = &__pyx_type_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest) < 0) __PYX_ERR(13, 6, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest) < 0) __PYX_ERR(13, 6, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest) < 0) __PYX_ERR(13, 6, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BaseRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest) < 0) __PYX_ERR(13, 6, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest) < 0) __PYX_ERR(13, 6, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_PingRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_PingRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_PingRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(29, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_PingRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest)) __PYX_ERR(29, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_PingRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest) < 0) __PYX_ERR(29, 5, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest = &__pyx_type_8asynctnt_6iproto_8protocol_PingRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest) < 0) __PYX_ERR(29, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_PingRequest) < 0) __PYX_ERR(29, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest) < 0) __PYX_ERR(29, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PingRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest) < 0) __PYX_ERR(29, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest) < 0) __PYX_ERR(29, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_CallRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_CallRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CallRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CallRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_11CallRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_CallRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest)) __PYX_ERR(14, 7, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_CallRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest) < 0) __PYX_ERR(14, 7, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest = &__pyx_type_8asynctnt_6iproto_8protocol_CallRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest) < 0) __PYX_ERR(14, 7, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_CallRequest) < 0) __PYX_ERR(14, 7, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest) < 0) __PYX_ERR(14, 7, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CallRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest) < 0) __PYX_ERR(14, 7, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest) < 0) __PYX_ERR(14, 7, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_EvalRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_EvalRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_EvalRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_EvalRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_11EvalRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_EvalRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest)) __PYX_ERR(15, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_EvalRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest) < 0) __PYX_ERR(15, 5, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest = &__pyx_type_8asynctnt_6iproto_8protocol_EvalRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest) < 0) __PYX_ERR(15, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_EvalRequest) < 0) __PYX_ERR(15, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest) < 0) __PYX_ERR(15, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_EvalRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest) < 0) __PYX_ERR(15, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest) < 0) __PYX_ERR(15, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_SelectRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_SelectRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_SelectRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_SelectRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_13SelectRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_SelectRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest)) __PYX_ERR(16, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_SelectRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest) < 0) __PYX_ERR(16, 5, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest = &__pyx_type_8asynctnt_6iproto_8protocol_SelectRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest) < 0) __PYX_ERR(16, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_SelectRequest) < 0) __PYX_ERR(16, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest) < 0) __PYX_ERR(16, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SelectRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest) < 0) __PYX_ERR(16, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest) < 0) __PYX_ERR(16, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_InsertRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_InsertRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_InsertRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_InsertRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_13InsertRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_InsertRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest)) __PYX_ERR(17, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_InsertRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest) < 0) __PYX_ERR(17, 5, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest = &__pyx_type_8asynctnt_6iproto_8protocol_InsertRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest) < 0) __PYX_ERR(17, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_InsertRequest) < 0) __PYX_ERR(17, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest) < 0) __PYX_ERR(17, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InsertRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest) < 0) __PYX_ERR(17, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest) < 0) __PYX_ERR(17, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_DeleteRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_DeleteRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_DeleteRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_DeleteRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_13DeleteRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest)) __PYX_ERR(18, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest) < 0) __PYX_ERR(18, 5, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest = &__pyx_type_8asynctnt_6iproto_8protocol_DeleteRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest) < 0) __PYX_ERR(18, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_DeleteRequest) < 0) __PYX_ERR(18, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest) < 0) __PYX_ERR(18, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DeleteRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest) < 0) __PYX_ERR(18, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest) < 0) __PYX_ERR(18, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_UpdateRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_UpdateRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_UpdateRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_UpdateRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_13UpdateRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest)) __PYX_ERR(4, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest) < 0) __PYX_ERR(4, 5, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest = &__pyx_type_8asynctnt_6iproto_8protocol_UpdateRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest) < 0) __PYX_ERR(4, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_UpdateRequest) < 0) __PYX_ERR(4, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest) < 0) __PYX_ERR(4, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UpdateRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest) < 0) __PYX_ERR(4, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest) < 0) __PYX_ERR(4, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_UpsertRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_UpsertRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_UpsertRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_UpsertRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_13UpsertRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest)) __PYX_ERR(19, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest) < 0) __PYX_ERR(19, 5, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest = &__pyx_type_8asynctnt_6iproto_8protocol_UpsertRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest) < 0) __PYX_ERR(19, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_UpsertRequest) < 0) __PYX_ERR(19, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest) < 0) __PYX_ERR(19, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UpsertRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest) < 0) __PYX_ERR(19, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest) < 0) __PYX_ERR(19, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_PrepareRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_PrepareRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_PrepareRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_PrepareRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_14PrepareRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest)) __PYX_ERR(20, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest = &__pyx_type_8asynctnt_6iproto_8protocol_PrepareRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_PrepareRequest) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PrepareRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_ExecuteRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_ExecuteRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_ExecuteRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_ExecuteRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_14ExecuteRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest)) __PYX_ERR(21, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest) < 0) __PYX_ERR(21, 5, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest = &__pyx_type_8asynctnt_6iproto_8protocol_ExecuteRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest) < 0) __PYX_ERR(21, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_ExecuteRequest) < 0) __PYX_ERR(21, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest) < 0) __PYX_ERR(21, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ExecuteRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest) < 0) __PYX_ERR(21, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest) < 0) __PYX_ERR(21, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_IDRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_IDRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_IDRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_IDRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_9IDRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_IDRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest)) __PYX_ERR(22, 6, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_IDRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest) < 0) __PYX_ERR(22, 6, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest = &__pyx_type_8asynctnt_6iproto_8protocol_IDRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest) < 0) __PYX_ERR(22, 6, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_IDRequest) < 0) __PYX_ERR(22, 6, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest) < 0) __PYX_ERR(22, 6, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IDRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest) < 0) __PYX_ERR(22, 6, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest) < 0) __PYX_ERR(22, 6, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_AuthRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_AuthRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_AuthRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_AuthRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_11AuthRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_AuthRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest)) __PYX_ERR(23, 31, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_AuthRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest) < 0) __PYX_ERR(23, 31, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest = &__pyx_type_8asynctnt_6iproto_8protocol_AuthRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest) < 0) __PYX_ERR(23, 31, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_AuthRequest) < 0) __PYX_ERR(23, 31, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest) < 0) __PYX_ERR(23, 31, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AuthRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest) < 0) __PYX_ERR(23, 31, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest) < 0) __PYX_ERR(23, 31, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_BeginRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_BeginRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BeginRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BeginRequest.__pyx_base.encode_body = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, struct __pyx_obj_8asynctnt_6iproto_8protocol_WriteBuffer *))__pyx_f_8asynctnt_6iproto_8protocol_12BeginRequest_encode_body;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_BeginRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest)) __PYX_ERR(24, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_BeginRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest) < 0) __PYX_ERR(24, 5, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest = &__pyx_type_8asynctnt_6iproto_8protocol_BeginRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest) < 0) __PYX_ERR(24, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_BeginRequest) < 0) __PYX_ERR(24, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest) < 0) __PYX_ERR(24, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BeginRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest) < 0) __PYX_ERR(24, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest) < 0) __PYX_ERR(24, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_CommitRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_CommitRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CommitRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_CommitRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest)) __PYX_ERR(24, 33, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_CommitRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest) < 0) __PYX_ERR(24, 33, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest = &__pyx_type_8asynctnt_6iproto_8protocol_CommitRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest) < 0) __PYX_ERR(24, 33, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_CommitRequest) < 0) __PYX_ERR(24, 33, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest) < 0) __PYX_ERR(24, 33, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CommitRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest) < 0) __PYX_ERR(24, 33, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest) < 0) __PYX_ERR(24, 33, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_RollbackRequest = &__pyx_vtable_8asynctnt_6iproto_8protocol_RollbackRequest;
  __pyx_vtable_8asynctnt_6iproto_8protocol_RollbackRequest.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseRequest;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest)) __PYX_ERR(24, 37, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest) < 0) __PYX_ERR(24, 37, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest = &__pyx_type_8asynctnt_6iproto_8protocol_RollbackRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest) < 0) __PYX_ERR(24, 37, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest, __pyx_vtabptr_8asynctnt_6iproto_8protocol_RollbackRequest) < 0) __PYX_ERR(24, 37, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest) < 0) __PYX_ERR(24, 37, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_RollbackRequest, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest) < 0) __PYX_ERR(24, 37, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest) < 0) __PYX_ERR(24, 37, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame)) __PYX_ERR(5, 16, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame) < 0) __PYX_ERR(5, 16, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame = &__pyx_type_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame) < 0) __PYX_ERR(5, 16, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IProtoErrorStackFrame, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame) < 0) __PYX_ERR(5, 16, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame) < 0) __PYX_ERR(5, 16, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_IProtoError_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError)) __PYX_ERR(5, 25, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_IProtoError_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError) < 0) __PYX_ERR(5, 25, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError = &__pyx_type_8asynctnt_6iproto_8protocol_IProtoError;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError) < 0) __PYX_ERR(5, 25, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IProtoError, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError) < 0) __PYX_ERR(5, 25, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError) < 0) __PYX_ERR(5, 25, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_Response = &__pyx_vtable_8asynctnt_6iproto_8protocol_Response;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Response.is_error = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *))__pyx_f_8asynctnt_6iproto_8protocol_8Response_is_error;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Response._len = (uint32_t (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *))__pyx_f_8asynctnt_6iproto_8protocol_8Response__len;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Response.init_push = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *))__pyx_f_8asynctnt_6iproto_8protocol_8Response_init_push;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Response.add_push = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_8Response_add_push;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Response.pop_push = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *))__pyx_f_8asynctnt_6iproto_8protocol_8Response_pop_push;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Response.push_len = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *))__pyx_f_8asynctnt_6iproto_8protocol_8Response_push_len;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Response.set_data = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_8Response_set_data;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Response.set_exception = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_8Response_set_exception;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Response.get_exception = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *))__pyx_f_8asynctnt_6iproto_8protocol_8Response_get_exception;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Response.notify = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *))__pyx_f_8asynctnt_6iproto_8protocol_8Response_notify;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_Response = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_Response_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_Response)) __PYX_ERR(5, 30, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_Response_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_Response) < 0) __PYX_ERR(5, 30, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_Response = &__pyx_type_8asynctnt_6iproto_8protocol_Response;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_Response) < 0) __PYX_ERR(5, 30, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_Response->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_Response->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_Response->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_Response->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_Response, __pyx_vtabptr_8asynctnt_6iproto_8protocol_Response) < 0) __PYX_ERR(5, 30, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_Response) < 0) __PYX_ERR(5, 30, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Response, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_Response) < 0) __PYX_ERR(5, 30, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_Response) < 0) __PYX_ERR(5, 30, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_Db = &__pyx_vtable_8asynctnt_6iproto_8protocol_Db;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db.create = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, uint64_t))__pyx_f_8asynctnt_6iproto_8protocol_2Db_create;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db.next_sync = (uint64_t (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *))__pyx_f_8asynctnt_6iproto_8protocol_2Db_next_sync;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._ping = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__ping;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._id = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__id;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._auth = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__auth;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._call = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, PyObject *, PyObject *, float, int))__pyx_f_8asynctnt_6iproto_8protocol_2Db__call;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._eval = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, float, int))__pyx_f_8asynctnt_6iproto_8protocol_2Db__eval;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._select = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, uint64_t, uint64_t, PyObject *, float, int))__pyx_f_8asynctnt_6iproto_8protocol_2Db__select;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._insert = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, int, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__insert;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._delete = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__delete;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._update = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, PyObject *, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__update;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._upsert = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, PyObject *, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__upsert;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._execute = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, PyObject *, int, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__execute;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._prepare = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, PyObject *, int, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__prepare;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._begin = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, uint32_t, double, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__begin;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._commit = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__commit;
  __pyx_vtable_8asynctnt_6iproto_8protocol_Db._rollback = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *, float))__pyx_f_8asynctnt_6iproto_8protocol_2Db__rollback;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_Db = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_Db_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_Db)) __PYX_ERR(26, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_Db_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_Db) < 0) __PYX_ERR(26, 5, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_Db = &__pyx_type_8asynctnt_6iproto_8protocol_Db;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_Db) < 0) __PYX_ERR(26, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_Db->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_Db->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_Db->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_Db->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_vtabptr_8asynctnt_6iproto_8protocol_Db) < 0) __PYX_ERR(26, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_Db) < 0) __PYX_ERR(26, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Db, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_Db) < 0) __PYX_ERR(26, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_Db) < 0) __PYX_ERR(26, 5, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_PushIterator_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator)) __PYX_ERR(27, 8, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_PushIterator_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator) < 0) __PYX_ERR(27, 8, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator = &__pyx_type_8asynctnt_6iproto_8protocol_PushIterator;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator) < 0) __PYX_ERR(27, 8, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(27, 8, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_8asynctnt_6iproto_8protocol_12PushIterator___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_8asynctnt_6iproto_8protocol_12PushIterator___init__.doc = __pyx_doc_8asynctnt_6iproto_8protocol_12PushIterator___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_8asynctnt_6iproto_8protocol_12PushIterator___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PushIterator, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator) < 0) __PYX_ERR(27, 8, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator) < 0) __PYX_ERR(27, 8, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_CoreProtocol = &__pyx_vtable_8asynctnt_6iproto_8protocol_CoreProtocol;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CoreProtocol._is_connected = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__is_connected;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CoreProtocol._is_fully_connected = (int (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__is_fully_connected;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CoreProtocol._write = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__write;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CoreProtocol._on_data_received = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_data_received;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CoreProtocol._process__greeting = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__process__greeting;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CoreProtocol._on_greeting_received = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_greeting_received;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CoreProtocol._on_response_received = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *, char const *, uint32_t))__pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_response_received;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CoreProtocol._on_connection_made = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_connection_made;
  __pyx_vtable_8asynctnt_6iproto_8protocol_CoreProtocol._on_connection_lost = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_12CoreProtocol__on_connection_lost;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol)) __PYX_ERR(28, 18, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol) < 0) __PYX_ERR(28, 18, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol = &__pyx_type_8asynctnt_6iproto_8protocol_CoreProtocol;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol) < 0) __PYX_ERR(28, 18, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol, __pyx_vtabptr_8asynctnt_6iproto_8protocol_CoreProtocol) < 0) __PYX_ERR(28, 18, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol) < 0) __PYX_ERR(28, 18, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CoreProtocol, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol) < 0) __PYX_ERR(28, 18, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol) < 0) __PYX_ERR(28, 18, __pyx_L1_error)
  #endif
  __pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseProtocol = &__pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol.__pyx_base = *__pyx_vtabptr_8asynctnt_6iproto_8protocol_CoreProtocol;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol.__pyx_base._on_greeting_received = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_greeting_received;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol.__pyx_base._on_response_received = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *, char const *, uint32_t))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_response_received;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol.__pyx_base._on_connection_made = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_connection_made;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol.__pyx_base._on_connection_lost = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_CoreProtocol *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__on_connection_lost;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol._set_connection_ready = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__set_connection_ready;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol._set_connection_error = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__set_connection_error;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol._post_con_state_machine = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__post_con_state_machine;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol._do_id = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__do_id;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol._do_auth = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, PyObject *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__do_auth;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol._do_fetch_schema = (void (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__do_fetch_schema;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol._refetch_schema = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__refetch_schema;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol.next_sync = (uint64_t (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_next_sync;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol.next_stream_id = (uint64_t (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_next_stream_id;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol.transform_iterator = (uint32_t (*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, PyObject *))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol_transform_iterator;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol._new_waiter_for_request = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, struct __pyx_obj_8asynctnt_6iproto_8protocol_Response *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, float))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__new_waiter_for_request;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol._create_db = (struct __pyx_obj_8asynctnt_6iproto_8protocol_Db *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, int))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__create_db;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol._execute_bad = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, float))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__execute_bad;
  __pyx_vtable_8asynctnt_6iproto_8protocol_BaseProtocol._execute_normal = (PyObject *(*)(struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseProtocol *, struct __pyx_obj_8asynctnt_6iproto_8protocol_BaseRequest *, float))__pyx_f_8asynctnt_6iproto_8protocol_12BaseProtocol__execute_normal;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol)) __PYX_ERR(6, 62, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol_spec, __pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol) < 0) __PYX_ERR(6, 62, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol = &__pyx_type_8asynctnt_6iproto_8protocol_BaseProtocol;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol->tp_base = __pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol) < 0) __PYX_ERR(6, 62, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol, __pyx_vtabptr_8asynctnt_6iproto_8protocol_BaseProtocol) < 0) __PYX_ERR(6, 62, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol) < 0) __PYX_ERR(6, 62, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BaseProtocol, (PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol) < 0) __PYX_ERR(6, 62, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol) < 0) __PYX_ERR(6, 62, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex)) __PYX_ERR(3, 54, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex_spec, __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex) < 0) __PYX_ERR(3, 54, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex = &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex) < 0) __PYX_ERR(3, 54, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct__hex->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr)) __PYX_ERR(3, 55, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr_spec, __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(3, 55, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr = &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(3, 55, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_1_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext___spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__)) __PYX_ERR(27, 62, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext___spec, __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__) < 0) __PYX_ERR(27, 62, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__ = &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__) < 0) __PYX_ERR(27, 62, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_2___anext__->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id)) __PYX_ERR(6, 247, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id_spec, __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id) < 0) __PYX_ERR(6, 247, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id = &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id) < 0) __PYX_ERR(6, 247, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_3__do_id->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth)) __PYX_ERR(6, 268, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth_spec, __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth) < 0) __PYX_ERR(6, 268, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth = &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth) < 0) __PYX_ERR(6, 268, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_4__do_auth->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema_spec, NULL); if (unlikely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema)) __PYX_ERR(6, 290, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema_spec, __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema) < 0) __PYX_ERR(6, 290, __pyx_L1_error)
  #else
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema = &__pyx_type_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema) < 0) __PYX_ERR(6, 290, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema->tp_dictoffset && __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_8asynctnt_6iproto_8protocol___pyx_scope_struct_5__do_fetch_schema->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(30, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_7(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_7(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_7(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_7(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_7); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(30, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(31, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_0_7(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_7(PyBoolObject),__Pyx_ImportType_CheckSize_Warn_3_0_7); if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(31, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(32, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_0_7(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_7(PyComplexObject),__Pyx_ImportType_CheckSize_Warn_3_0_7); if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(32, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("datetime"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_8datetime_date = __Pyx_ImportType_3_0_7(__pyx_t_1, "datetime", "date", sizeof(PyDateTime_Date), __PYX_GET_STRUCT_ALIGNMENT_3_0_7(PyDateTime_Date),__Pyx_ImportType_CheckSize_Warn_3_0_7); if (!__pyx_ptype_7cpython_8datetime_date) __PYX_ERR(8, 70, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_time = __Pyx_ImportType_3_0_7(__pyx_t_1, "datetime", "time", sizeof(PyDateTime_Time), __PYX_GET_STRUCT_ALIGNMENT_3_0_7(PyDateTime_Time),__Pyx_ImportType_CheckSize_Warn_3_0_7); if (!__pyx_ptype_7cpython_8datetime_time) __PYX_ERR(8, 83, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_datetime = __Pyx_ImportType_3_0_7(__pyx_t_1, "datetime", "datetime", sizeof(PyDateTime_DateTime), __PYX_GET_STRUCT_ALIGNMENT_3_0_7(PyDateTime_DateTime),__Pyx_ImportType_CheckSize_Warn_3_0_7); if (!__pyx_ptype_7cpython_8datetime_datetime) __PYX_ERR(8, 109, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_timedelta = __Pyx_ImportType_3_0_7(__pyx_t_1, "datetime", "timedelta", sizeof(PyDateTime_Delta), __PYX_GET_STRUCT_ALIGNMENT_3_0_7(PyDateTime_Delta),__Pyx_ImportType_CheckSize_Warn_3_0_7); if (!__pyx_ptype_7cpython_8datetime_timedelta) __PYX_ERR(8, 147, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_tzinfo = __Pyx_ImportType_3_0_7(__pyx_t_1, "datetime", "tzinfo", sizeof(PyDateTime_TZInfo), __PYX_GET_STRUCT_ALIGNMENT_3_0_7(PyDateTime_TZInfo),__Pyx_ImportType_CheckSize_Warn_3_0_7); if (!__pyx_ptype_7cpython_8datetime_tzinfo) __PYX_ERR(8, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_protocol(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_protocol},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "protocol",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initprotocol(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initprotocol(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_protocol(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_protocol(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_protocol(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyTypeObject *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'protocol' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("protocol", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(6, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "protocol" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(6, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(6, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(6, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(6, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_protocol(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(6, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_asynctnt__iproto__protocol) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(6, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "asynctnt.iproto.protocol")) {
      if (unlikely((PyDict_SetItemString(modules, "asynctnt.iproto.protocol", __pyx_m) < 0))) __PYX_ERR(6, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely((__Pyx_modinit_function_export_code() < 0))) __PYX_ERR(6, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(6, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(6, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  #endif

  /* "asynctnt/iproto/protocol.pyx":6
 * from cpython.datetime cimport import_datetime
 * 
 * import_datetime()             # <<<<<<<<<<<<<<
 * 
 * import asyncio
 */
  __pyx_f_7cpython_8datetime_import_datetime(); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 6, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":8
 * import_datetime()
 * 
 * import asyncio             # <<<<<<<<<<<<<<
 * import enum
 * 
 */
  __pyx_t_2 = __Pyx_patch_asyncio(__Pyx_ImportDottedModule(__pyx_n_s_asyncio, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_2) < 0) __PYX_ERR(6, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":9
 * 
 * import asyncio
 * import enum             # <<<<<<<<<<<<<<
 * 
 * from asynctnt.exceptions import TarantoolNotConnectedError
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_enum, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_enum, __pyx_t_2) < 0) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":11
 * import enum
 * 
 * from asynctnt.exceptions import TarantoolNotConnectedError             # <<<<<<<<<<<<<<
 * 
 * include "const.pxi"
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_TarantoolNotConnectedError);
  __Pyx_GIVEREF(__pyx_n_s_TarantoolNotConnectedError);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_TarantoolNotConnectedError)) __PYX_ERR(6, 11, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_asynctnt_exceptions, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_TarantoolNotConnectedError); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TarantoolNotConnectedError, __pyx_t_2) < 0) __PYX_ERR(6, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/schema.pyx":3
 * # cython: profile=True
 * 
 * from asynctnt.exceptions import TarantoolSchemaError             # <<<<<<<<<<<<<<
 * from asynctnt.log import logger
 * 
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_TarantoolSchemaError);
  __Pyx_GIVEREF(__pyx_n_s_TarantoolSchemaError);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_TarantoolSchemaError)) __PYX_ERR(1, 3, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_asynctnt_exceptions, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_TarantoolSchemaError); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TarantoolSchemaError, __pyx_t_3) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/schema.pyx":4
 * 
 * from asynctnt.exceptions import TarantoolSchemaError
 * from asynctnt.log import logger             # <<<<<<<<<<<<<<
 * 
 * cimport cpython.list
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_logger);
  __Pyx_GIVEREF(__pyx_n_s_logger);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_logger)) __PYX_ERR(1, 4, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_asynctnt_log, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logger, __pyx_t_2) < 0) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_5Field_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Field___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_5Field_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Field___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__69)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_8Metadata_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Metadata___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_8Metadata_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Metadata___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__71)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11SchemaIndex_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SchemaIndex___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11SchemaIndex_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SchemaIndex___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11SchemaSpace_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SchemaSpace___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__74)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11SchemaSpace_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SchemaSpace___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__75)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_6Schema_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_6Schema_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/ext.pyx":6
 * from libc.string cimport memcpy
 * 
 * from decimal import Decimal             # <<<<<<<<<<<<<<
 * from uuid import UUID
 * 
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_Decimal);
  __Pyx_GIVEREF(__pyx_n_s_Decimal);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Decimal)) __PYX_ERR(2, 6, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_decimal, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Decimal, __pyx_t_3) < 0) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/ext.pyx":7
 * 
 * from decimal import Decimal
 * from uuid import UUID             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_UUID);
  __Pyx_GIVEREF(__pyx_n_s_UUID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_UUID)) __PYX_ERR(2, 7, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_uuid, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_UUID); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UUID, __pyx_t_2) < 0) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/buffer.pyx":15
 * from libc.string cimport memcpy
 * 
 * from decimal import Decimal  # pragma: nocover             # <<<<<<<<<<<<<<
 * from uuid import UUID  # pragma: nocover
 * 
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_Decimal);
  __Pyx_GIVEREF(__pyx_n_s_Decimal);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Decimal)) __PYX_ERR(3, 15, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_decimal, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Decimal, __pyx_t_3) < 0) __PYX_ERR(3, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/buffer.pyx":16
 * 
 * from decimal import Decimal  # pragma: nocover
 * from uuid import UUID  # pragma: nocover             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_UUID);
  __Pyx_GIVEREF(__pyx_n_s_UUID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_UUID)) __PYX_ERR(3, 16, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_uuid, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_UUID); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UUID, __pyx_t_2) < 0) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/buffer.pyx":54
 *         self._view_count -= 1
 * 
 *     def hex(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))
 *                         for i in range(self._length))
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11WriteBuffer_9hex, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_WriteBuffer_hex, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer, __pyx_n_s_hex, __pyx_t_3) < 0) __PYX_ERR(3, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_WriteBuffer);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11WriteBuffer_11__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_WriteBuffer___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__80)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11WriteBuffer_13__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_WriteBuffer___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__81)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_10ReadBuffer_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadBuffer___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_10ReadBuffer_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadBuffer___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__83)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11BaseRequest_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BaseRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BaseRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BaseRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11BaseRequest_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BaseRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__86)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11PingRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PingRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__87)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PingRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PingRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11PingRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PingRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_PingRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11CallRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CallRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__89)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CallRequest, (type(self), 0x2d5a81d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CallRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11CallRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CallRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__90)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CallRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11EvalRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EvalRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EvalRequest, (type(self), 0xafafa67, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EvalRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11EvalRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EvalRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_EvalRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13SelectRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SelectRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__93)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SelectRequest, (type(self), 0x8e2dc2b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SelectRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13SelectRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SelectRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_SelectRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13InsertRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InsertRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__95)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_InsertRequest, (type(self), 0xb91d1f2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_InsertRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13InsertRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_InsertRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__96)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_InsertRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13DeleteRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DeleteRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DeleteRequest, (type(self), 0xf383ff7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DeleteRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13DeleteRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DeleteRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__98)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_DeleteRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13UpdateRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UpdateRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__99)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UpdateRequest, (type(self), 0x815e48a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UpdateRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13UpdateRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UpdateRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_UpdateRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13UpsertRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UpsertRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__101)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UpsertRequest, (type(self), 0xc823323, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UpsertRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13UpsertRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UpsertRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__102)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_UpsertRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_14PrepareRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrepareRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PrepareRequest, (type(self), 0xd0934a3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PrepareRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_14PrepareRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrepareRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__104)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_PrepareRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_14ExecuteRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ExecuteRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__105)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ExecuteRequest, (type(self), 0x314ca88, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ExecuteRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_14ExecuteRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ExecuteRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_ExecuteRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_9IDRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IDRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__107)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IDRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IDRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_9IDRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IDRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__108)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_IDRequest);

  /* "asynctnt/iproto/requests/auth.pyx":5
 * cimport cython
 * 
 * import hashlib             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_hashlib, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hashlib, __pyx_t_3) < 0) __PYX_ERR(23, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11AuthRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AuthRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AuthRequest, (type(self), 0x335e105, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AuthRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11AuthRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AuthRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__110)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_AuthRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BeginRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BeginRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__111)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BeginRequest, (type(self), 0x30f95df, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BeginRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BeginRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BeginRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_BeginRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13CommitRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CommitRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__113)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CommitRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CommitRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13CommitRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CommitRequest___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__114)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CommitRequest);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_15RollbackRequest_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RollbackRequest___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__115)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RollbackRequest, (type(self), 0x9001016, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RollbackRequest__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_15RollbackRequest_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RollbackRequest___setstate_cytho, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__116)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_RollbackRequest);

  /* "asynctnt/iproto/ttuple.pyx":1
 * TarantoolTuple = <object> tupleobj.AtntTuple_InitTypes()             # <<<<<<<<<<<<<<
 */
  __pyx_t_4 = AtntTuple_InitTypes(); if (unlikely(__pyx_t_4 == ((PyTypeObject *)NULL))) __PYX_ERR(33, 1, __pyx_L1_error)
  __pyx_t_3 = ((PyObject *)__pyx_t_4);
  __Pyx_INCREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TarantoolTuple, __pyx_t_3) < 0) __PYX_ERR(33, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/response.pyx":1
 * import asyncio             # <<<<<<<<<<<<<<
 * import collections
 * from typing import Optional
 */
  __pyx_t_3 = __Pyx_patch_asyncio(__Pyx_ImportDottedModule(__pyx_n_s_asyncio, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_3) < 0) __PYX_ERR(5, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/response.pyx":2
 * import asyncio
 * import collections             # <<<<<<<<<<<<<<
 * from typing import Optional
 * 
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_collections, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_collections, __pyx_t_3) < 0) __PYX_ERR(5, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/response.pyx":3
 * import asyncio
 * import collections
 * from typing import Optional             # <<<<<<<<<<<<<<
 * 
 * cimport cpython
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_Optional);
  __Pyx_GIVEREF(__pyx_n_s_Optional);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Optional)) __PYX_ERR(5, 3, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Optional); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Optional, __pyx_t_3) < 0) __PYX_ERR(5, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/response.pyx":12
 * from libc.stdint cimport uint32_t
 * 
 * from asynctnt.log import logger             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_logger);
  __Pyx_GIVEREF(__pyx_n_s_logger);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_logger)) __PYX_ERR(5, 12, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_asynctnt_log, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logger, __pyx_t_2) < 0) __PYX_ERR(5, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IProtoErrorStackFrame___reduce_c, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__117)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IProtoErrorStackFrame, (type(self), 0xab276bf, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IProtoErrorStackFrame__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_21IProtoErrorStackFrame_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IProtoErrorStackFrame___setstate, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__118)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoErrorStackFrame);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11IProtoError_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IProtoError___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__119)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IProtoError, (type(self), 0xeafe895, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IProtoError__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11IProtoError_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IProtoError___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__120)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_IProtoError);

  /* "asynctnt/iproto/response.pyx":160
 *         return self.stmt_id_
 * 
 *     def done(self):             # <<<<<<<<<<<<<<
 *         return self.code_ >= 0
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_8Response_5done, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Response_done, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__121)); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Response, __pyx_n_s_done, __pyx_t_3) < 0) __PYX_ERR(5, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Response);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_8Response_13__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Response___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__122)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_8Response_15__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Response___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__123)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/db.pyx":327
 *         return <int> self._stream_id
 * 
 *     def set_stream_id(self, int stream_id):             # <<<<<<<<<<<<<<
 *         self._stream_id = <uint64_t> stream_id
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_3set_stream_id, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_set_stream_id, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__125)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_set_stream_id, __pyx_t_3) < 0) __PYX_ERR(26, 327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":330
 *         self._stream_id = <uint64_t> stream_id
 * 
 *     def ping(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._ping(timeout)
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_5ping, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_ping, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__127)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__128);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_ping, __pyx_t_3) < 0) __PYX_ERR(26, 330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":333
 *         return self._ping(timeout)
 * 
 *     def call16(self,             # <<<<<<<<<<<<<<
 *                str func_name,
 *                object args=None,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_7call16, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_call16, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__130)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__131);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_call16, __pyx_t_3) < 0) __PYX_ERR(26, 333, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":344
 *                           <bint> push_subscribe)
 * 
 *     def call(self,             # <<<<<<<<<<<<<<
 *              str func_name,
 *              object args=None,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_9call, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_call, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__132)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__131);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_call, __pyx_t_3) < 0) __PYX_ERR(26, 344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":355
 *                           <bint> push_subscribe)
 * 
 *     def eval(self,             # <<<<<<<<<<<<<<
 *              str expression,
 *              object args=None,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_11eval, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_eval, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__134)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__131);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_eval, __pyx_t_3) < 0) __PYX_ERR(26, 355, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":365
 *                           <bint> push_subscribe)
 * 
 *     def select(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key=None,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_13select, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_select, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__136)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__137);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_select, __pyx_t_3) < 0) __PYX_ERR(26, 365, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":377
 *                             timeout, check_schema_change)
 * 
 *     def insert(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object t,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_15insert, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_insert, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__139)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__140);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_insert, __pyx_t_3) < 0) __PYX_ERR(26, 377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":384
 *         return self._insert(space, t, <bint> replace, timeout)
 * 
 *     def replace(self,             # <<<<<<<<<<<<<<
 *                 object space,
 *                 object t,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_17replace, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_replace, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__142)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__128);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_replace, __pyx_t_3) < 0) __PYX_ERR(26, 384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":390
 *         return self._insert(space, t, <bint> True, timeout)
 * 
 *     def delete(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_19delete, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_delete, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__144)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__145);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_delete, __pyx_t_3) < 0) __PYX_ERR(26, 390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":397
 *         return self._delete(space, index, key, timeout)
 * 
 *     def update(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object key,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_21update, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_update, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__147)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__145);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_update, __pyx_t_3) < 0) __PYX_ERR(26, 397, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":405
 *         return self._update(space, index, key, operations, timeout)
 * 
 *     def upsert(self,             # <<<<<<<<<<<<<<
 *                object space,
 *                object t,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_23upsert, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_upsert, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__149)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__128);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_upsert, __pyx_t_3) < 0) __PYX_ERR(26, 405, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":412
 *         return self._upsert(space, t, operations, timeout)
 * 
 *     def execute(self,             # <<<<<<<<<<<<<<
 *                 object query,
 *                 object args,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_25execute, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_execute, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__151)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__152);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_execute, __pyx_t_3) < 0) __PYX_ERR(26, 412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":419
 *         return self._execute(query, args, <bint> parse_metadata, timeout)
 * 
 *     def prepare(self,             # <<<<<<<<<<<<<<
 *                 object query,
 *                 bint parse_metadata=True,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_27prepare, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_prepare, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__154)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__152);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_prepare, __pyx_t_3) < 0) __PYX_ERR(26, 419, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":425
 *         return self._prepare(query, <bint> parse_metadata, timeout)
 * 
 *     def begin(self,             # <<<<<<<<<<<<<<
 *               uint32_t isolation,
 *               float tx_timeout=0,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_29begin, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_begin, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__156)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__157);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_begin, __pyx_t_3) < 0) __PYX_ERR(26, 425, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":431
 *         return self._begin(isolation, <double> tx_timeout, timeout)
 * 
 *     def commit(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._commit(timeout)
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_31commit, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_commit, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__158)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__128);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_commit, __pyx_t_3) < 0) __PYX_ERR(26, 431, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "asynctnt/iproto/db.pyx":434
 *         return self._commit(timeout)
 * 
 *     def rollback(self, float timeout=-1):             # <<<<<<<<<<<<<<
 *         return self._rollback(timeout)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_33rollback, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db_rollback, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__159)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__128);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_Db, __pyx_n_s_rollback, __pyx_t_3) < 0) __PYX_ERR(26, 434, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_Db);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_35__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__160)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_2Db_37__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Db___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__161)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/push.pyx":1
 * import asyncio             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */
  __pyx_t_3 = __Pyx_patch_asyncio(__Pyx_ImportDottedModule(__pyx_n_s_asyncio, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_3) < 0) __PYX_ERR(27, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/push.pyx":5
 * cimport cython
 * 
 * import asynctnt             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_asynctnt, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asynctnt, __pyx_t_3) < 0) __PYX_ERR(27, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12PushIterator_12__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PushIterator___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__162)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PushIterator, (type(self), 0x98137af, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PushIterator__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12PushIterator_14__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PushIterator___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__163)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_PushIterator);

  /* "asynctnt/iproto/coreproto.pyx":8
 * from libc.stdint cimport uint32_t, uint64_t
 * 
 * import base64             # <<<<<<<<<<<<<<
 * import re
 * import socket
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_base64, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_base64, __pyx_t_3) < 0) __PYX_ERR(28, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/coreproto.pyx":9
 * 
 * import base64
 * import re             # <<<<<<<<<<<<<<
 * import socket
 * 
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_re, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_re, __pyx_t_3) < 0) __PYX_ERR(28, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/coreproto.pyx":10
 * import base64
 * import re
 * import socket             # <<<<<<<<<<<<<<
 * 
 * from asynctnt.exceptions import TarantoolDatabaseError
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_socket, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_socket, __pyx_t_3) < 0) __PYX_ERR(28, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/coreproto.pyx":12
 * import socket
 * 
 * from asynctnt.exceptions import TarantoolDatabaseError             # <<<<<<<<<<<<<<
 * from asynctnt.log import logger
 * 
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_TarantoolDatabaseError);
  __Pyx_GIVEREF(__pyx_n_s_TarantoolDatabaseError);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_TarantoolDatabaseError)) __PYX_ERR(28, 12, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_asynctnt_exceptions, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_TarantoolDatabaseError); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TarantoolDatabaseError, __pyx_t_3) < 0) __PYX_ERR(28, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/coreproto.pyx":13
 * 
 * from asynctnt.exceptions import TarantoolDatabaseError
 * from asynctnt.log import logger             # <<<<<<<<<<<<<<
 * 
 * VERSION_STRING_REGEX = re.compile(r'\s*Tarantool\s+([\d.]+)\s+.*')
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_logger);
  __Pyx_GIVEREF(__pyx_n_s_logger);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_logger)) __PYX_ERR(28, 13, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_asynctnt_log, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logger, __pyx_t_2) < 0) __PYX_ERR(28, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/coreproto.pyx":15
 * from asynctnt.log import logger
 * 
 * VERSION_STRING_REGEX = re.compile(r'\s*Tarantool\s+([\d.]+)\s+.*')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_re); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_compile); if (unlikely(!__pyx_t_2)) __PYX_ERR(28, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__164, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VERSION_STRING_REGEX, __pyx_t_3) < 0) __PYX_ERR(28, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/coreproto.pyx":50
 *         return self.con_state == CONNECTION_FULL
 * 
 *     def is_connected(self):             # <<<<<<<<<<<<<<
 *         return self._is_connected()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_3is_connected, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoreProtocol_is_connected, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__165)); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol, __pyx_n_s_is_connected, __pyx_t_3) < 0) __PYX_ERR(28, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol);

  /* "asynctnt/iproto/coreproto.pyx":53
 *         return self._is_connected()
 * 
 *     def is_fully_connected(self):             # <<<<<<<<<<<<<<
 *         return self._is_fully_connected()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_5is_fully_connected, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoreProtocol_is_fully_connected, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__166)); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol, __pyx_n_s_is_fully_connected, __pyx_t_3) < 0) __PYX_ERR(28, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol);

  /* "asynctnt/iproto/coreproto.pyx":56
 *         return self._is_fully_connected()
 * 
 *     def get_version(self):             # <<<<<<<<<<<<<<
 *         return self.version
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_7get_version, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoreProtocol_get_version, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__167)); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol, __pyx_n_s_get_version, __pyx_t_3) < 0) __PYX_ERR(28, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol);

  /* "asynctnt/iproto/coreproto.pyx":143
 *         self._on_greeting_received()
 * 
 *     def _parse_version(self, version):             # <<<<<<<<<<<<<<
 *         m = VERSION_STRING_REGEX.match(version.decode('ascii'))
 *         if m is not None:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_9_parse_version, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoreProtocol__parse_version, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__169)); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol, __pyx_n_s_parse_version, __pyx_t_3) < 0) __PYX_ERR(28, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol);

  /* "asynctnt/iproto/coreproto.pyx":163
 *     # asyncio callbacks
 * 
 *     def data_received(self, data):             # <<<<<<<<<<<<<<
 *         self._on_data_received(data)
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_11data_received, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoreProtocol_data_received, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__171)); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol, __pyx_n_s_data_received, __pyx_t_3) < 0) __PYX_ERR(28, 163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol);

  /* "asynctnt/iproto/coreproto.pyx":166
 *         self._on_data_received(data)
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         self.transport = transport
 *         self.con_state = CONNECTION_CONNECTED
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_13connection_made, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoreProtocol_connection_made, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__173)); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol, __pyx_n_s_connection_made, __pyx_t_3) < 0) __PYX_ERR(28, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol);

  /* "asynctnt/iproto/coreproto.pyx":179
 *         self._on_connection_made()
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         self.con_state = CONNECTION_BAD
 *         self.version = None
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_15connection_lost, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoreProtocol_connection_lost, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__175)); if (unlikely(!__pyx_t_3)) __PYX_ERR(28, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol, __pyx_n_s_connection_lost, __pyx_t_3) < 0) __PYX_ERR(28, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_17__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoreProtocol___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__176)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CoreProtocol, (type(self), 0x89d360e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CoreProtocol__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12CoreProtocol_19__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CoreProtocol___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__177)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_CoreProtocol);

  /* "asynctnt/iproto/protocol.pyx":44
 * 
 * 
 * class Iterator(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     """
 *         Available Iterator types
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_enum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(6, 44, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_2, __pyx_n_s_Iterator, __pyx_n_s_Iterator, (PyObject *) NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_kp_s_Available_Iterator_types); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__pyx_t_2 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_6, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(6, 44, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynctnt/iproto/protocol.pyx":48
 *         Available Iterator types
 *     """
 *     EQ = 0             # <<<<<<<<<<<<<<
 *     REQ = 1
 *     ALL = 2
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_EQ, __pyx_int_0) < 0) __PYX_ERR(6, 48, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":49
 *     """
 *     EQ = 0
 *     REQ = 1             # <<<<<<<<<<<<<<
 *     ALL = 2
 *     LT = 3
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_REQ, __pyx_int_1) < 0) __PYX_ERR(6, 49, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":50
 *     EQ = 0
 *     REQ = 1
 *     ALL = 2             # <<<<<<<<<<<<<<
 *     LT = 3
 *     LE = 4
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_ALL, __pyx_int_2) < 0) __PYX_ERR(6, 50, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":51
 *     REQ = 1
 *     ALL = 2
 *     LT = 3             # <<<<<<<<<<<<<<
 *     LE = 4
 *     GE = 5
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_LT, __pyx_int_3) < 0) __PYX_ERR(6, 51, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":52
 *     ALL = 2
 *     LT = 3
 *     LE = 4             # <<<<<<<<<<<<<<
 *     GE = 5
 *     GT = 6
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_LE, __pyx_int_4) < 0) __PYX_ERR(6, 52, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":53
 *     LT = 3
 *     LE = 4
 *     GE = 5             # <<<<<<<<<<<<<<
 *     GT = 6
 *     BITS_ALL_SET = 7
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_GE, __pyx_int_5) < 0) __PYX_ERR(6, 53, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":54
 *     LE = 4
 *     GE = 5
 *     GT = 6             # <<<<<<<<<<<<<<
 *     BITS_ALL_SET = 7
 *     BITS_ANY_SET = 8
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_GT, __pyx_int_6) < 0) __PYX_ERR(6, 54, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":55
 *     GE = 5
 *     GT = 6
 *     BITS_ALL_SET = 7             # <<<<<<<<<<<<<<
 *     BITS_ANY_SET = 8
 *     BITS_ALL_NOT_SET = 9
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_BITS_ALL_SET, __pyx_int_7) < 0) __PYX_ERR(6, 55, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":56
 *     GT = 6
 *     BITS_ALL_SET = 7
 *     BITS_ANY_SET = 8             # <<<<<<<<<<<<<<
 *     BITS_ALL_NOT_SET = 9
 *     OVERLAPS = 10
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_BITS_ANY_SET, __pyx_int_8) < 0) __PYX_ERR(6, 56, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":57
 *     BITS_ALL_SET = 7
 *     BITS_ANY_SET = 8
 *     BITS_ALL_NOT_SET = 9             # <<<<<<<<<<<<<<
 *     OVERLAPS = 10
 *     NEIGHBOR = 11
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_BITS_ALL_NOT_SET, __pyx_int_9) < 0) __PYX_ERR(6, 57, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":58
 *     BITS_ANY_SET = 8
 *     BITS_ALL_NOT_SET = 9
 *     OVERLAPS = 10             # <<<<<<<<<<<<<<
 *     NEIGHBOR = 11
 * 
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_OVERLAPS, __pyx_int_10) < 0) __PYX_ERR(6, 58, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":59
 *     BITS_ALL_NOT_SET = 9
 *     OVERLAPS = 10
 *     NEIGHBOR = 11             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_NEIGHBOR, __pyx_int_11) < 0) __PYX_ERR(6, 59, __pyx_L1_error)

  /* "asynctnt/iproto/protocol.pyx":44
 * 
 * 
 * class Iterator(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     """
 *         Available Iterator types
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_Iterator, __pyx_t_2, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Iterator, __pyx_t_3) < 0) __PYX_ERR(6, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":108
 *             self.create_future = self._create_future_fallback
 * 
 *     def _create_future_fallback(self):  # pragma: nocover             # <<<<<<<<<<<<<<
 *         return asyncio.Future(loop=self.loop)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_3_create_future_fallback, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BaseProtocol__create_future_fall, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__178)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol, __pyx_n_s_create_future_fallback, __pyx_t_2) < 0) __PYX_ERR(6, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol);

  /* "asynctnt/iproto/protocol.pyx":425
 *         return self._last_stream_id
 * 
 *     def _on_request_timeout(self, waiter):             # <<<<<<<<<<<<<<
 *         cdef:
 *             BaseRequest req
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_5_on_request_timeout, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BaseProtocol__on_request_timeout, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__180)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol, __pyx_n_s_on_request_timeout, __pyx_t_2) < 0) __PYX_ERR(6, 425, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol);

  /* "asynctnt/iproto/protocol.pyx":442
 *         )
 * 
 *     def _on_request_completed(self, fut):             # <<<<<<<<<<<<<<
 *         cdef BaseRequest req = (<Response> fut._response).request_
 *         fut._response = None
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_7_on_request_completed, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BaseProtocol__on_request_complet, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__182)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol, __pyx_n_s_on_request_completed, __pyx_t_2) < 0) __PYX_ERR(6, 442, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol);

  /* "asynctnt/iproto/protocol.pyx":473
 *         return Db.create(self, stream_id)
 * 
 *     def create_db(self, bint gen_stream_id = False):             # <<<<<<<<<<<<<<
 *         return self._create_db(gen_stream_id)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_9create_db, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BaseProtocol_create_db, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__184)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__185);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol, __pyx_n_s_create_db, __pyx_t_2) < 0) __PYX_ERR(6, 473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol);

  /* "asynctnt/iproto/protocol.pyx":476
 *         return self._create_db(gen_stream_id)
 * 
 *     def get_common_db(self):             # <<<<<<<<<<<<<<
 *         return self._db
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_11get_common_db, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BaseProtocol_get_common_db, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__186)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol, __pyx_n_s_get_common_db, __pyx_t_2) < 0) __PYX_ERR(6, 476, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol);

  /* "asynctnt/iproto/protocol.pyx":515
 *         return self._refetch_schema_future
 * 
 *     def refetch_schema(self):             # <<<<<<<<<<<<<<
 *         return self._refetch_schema()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_13refetch_schema, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BaseProtocol_refetch_schema, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__187)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol, __pyx_n_s_refetch_schema, __pyx_t_2) < 0) __PYX_ERR(6, 515, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_15__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BaseProtocol___reduce_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__188)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.execute cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_12BaseProtocol_17__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BaseProtocol___setstate_cython, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__189)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynctnt/iproto/protocol.pyx":519
 * 
 * 
 * class Protocol(BaseProtocol, asyncio.Protocol):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Protocol); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_8asynctnt_6iproto_8protocol_BaseProtocol))) __PYX_ERR(6, 519, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5)) __PYX_ERR(6, 519, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_6, __pyx_t_5, __pyx_n_s_Protocol, __pyx_n_s_Protocol, (PyObject *) NULL, __pyx_n_s_asynctnt_iproto_protocol, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_5 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_3, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(6, 519, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_6, __pyx_n_s_Protocol, __pyx_t_5, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Protocol, __pyx_t_2) < 0) __PYX_ERR(6, 519, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BaseRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_1__pyx_unpickle_BaseRequest, 0, __pyx_n_s_pyx_unpickle_BaseRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__191)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_BaseRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_BaseRequest__set_state(<BaseRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BaseRequest__set_state(BaseRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_3__pyx_unpickle_PingRequest, 0, __pyx_n_s_pyx_unpickle_PingRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__192)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PingRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CallRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_5__pyx_unpickle_CallRequest, 0, __pyx_n_s_pyx_unpickle_CallRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__193)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CallRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_CallRequest__set_state(<CallRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CallRequest__set_state(CallRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.func_name = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_7__pyx_unpickle_EvalRequest, 0, __pyx_n_s_pyx_unpickle_EvalRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__194)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_EvalRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_SelectRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_9__pyx_unpickle_SelectRequest, 0, __pyx_n_s_pyx_unpickle_SelectRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__195)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_SelectRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_SelectRequest__set_state(<SelectRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SelectRequest__set_state(SelectRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.iterator = __pyx_state[2]; __pyx_result.key = __pyx_state[3]; __pyx_result.limit = __pyx_state[4]; __pyx_result.offset = __pyx_state[5]; __pyx_result.op = __pyx_state[6]; __pyx_result.parse_as_tuples = __pyx_state[7]; __pyx_result.parse_metadata = __pyx_state[8]; __pyx_result.push_subscribe = __pyx_state[9]; __pyx_result.schema_id = __pyx_state[10]; __pyx_result.space = __pyx_state[11]; __pyx_result.stream_id = __pyx_state[12]; __pyx_result.sync = __pyx_state[13]; __pyx_result.timeout_handle = __pyx_state[14]; __pyx_result.waiter = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_11__pyx_unpickle_InsertRequest, 0, __pyx_n_s_pyx_unpickle_InsertRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__196)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_InsertRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_DeleteRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_13__pyx_unpickle_DeleteRequest, 0, __pyx_n_s_pyx_unpickle_DeleteRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__197)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_DeleteRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_DeleteRequest__set_state(<DeleteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DeleteRequest__set_state(DeleteRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.index = __pyx_state[1]; __pyx_result.key = __pyx_state[2]; __pyx_result.op = __pyx_state[3]; __pyx_result.parse_as_tuples = __pyx_state[4]; __pyx_result.parse_metadata = __pyx_state[5]; __pyx_result.push_subscribe = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_15__pyx_unpickle_UpdateRequest, 0, __pyx_n_s_pyx_unpickle_UpdateRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__198)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_UpdateRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_UpsertRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_17__pyx_unpickle_UpsertRequest, 0, __pyx_n_s_pyx_unpickle_UpsertRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__199)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_UpsertRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_UpsertRequest__set_state(<UpsertRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UpsertRequest__set_state(UpsertRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.operations = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.schema_id = __pyx_state[6]; __pyx_result.space = __pyx_state[7]; __pyx_result.stream_id = __pyx_state[8]; __pyx_result.sync = __pyx_state[9]; __pyx_result.t = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.waiter = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_19__pyx_unpickle_PrepareRequest, 0, __pyx_n_s_pyx_unpickle_PrepareRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__200)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PrepareRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ExecuteRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_21__pyx_unpickle_ExecuteRequest, 0, __pyx_n_s_pyx_unpickle_ExecuteRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__201)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ExecuteRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_ExecuteRequest__set_state(<ExecuteRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ExecuteRequest__set_state(ExecuteRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.check_schema_change = __pyx_state[1]; __pyx_result.op = __pyx_state[2]; __pyx_result.parse_as_tuples = __pyx_state[3]; __pyx_result.parse_metadata = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.query = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.statement_id = __pyx_state[9]; __pyx_result.stream_id = __pyx_state[10]; __pyx_result.sync = __pyx_state[11]; __pyx_result.timeout_handle = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_23__pyx_unpickle_IDRequest, 0, __pyx_n_s_pyx_unpickle_IDRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__202)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IDRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_AuthRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_25__pyx_unpickle_AuthRequest, 0, __pyx_n_s_pyx_unpickle_AuthRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__203)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_AuthRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_AuthRequest__set_state(<AuthRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_AuthRequest__set_state(AuthRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.password = __pyx_state[4]; __pyx_result.push_subscribe = __pyx_state[5]; __pyx_result.salt = __pyx_state[6]; __pyx_result.schema_id = __pyx_state[7]; __pyx_result.space = __pyx_state[8]; __pyx_result.stream_id = __pyx_state[9]; __pyx_result.sync = __pyx_state[10]; __pyx_result.timeout_handle = __pyx_state[11]; __pyx_result.username = __pyx_state[12]; __pyx_result.waiter = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_27__pyx_unpickle_BeginRequest, 0, __pyx_n_s_pyx_unpickle_BeginRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__204)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_BeginRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CommitRequest(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_29__pyx_unpickle_CommitRequest, 0, __pyx_n_s_pyx_unpickle_CommitRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__205)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CommitRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_CommitRequest__set_state(<CommitRequest> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CommitRequest__set_state(CommitRequest __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_schema_change = __pyx_state[0]; __pyx_result.op = __pyx_state[1]; __pyx_result.parse_as_tuples = __pyx_state[2]; __pyx_result.parse_metadata = __pyx_state[3]; __pyx_result.push_subscribe = __pyx_state[4]; __pyx_result.schema_id = __pyx_state[5]; __pyx_result.space = __pyx_state[6]; __pyx_result.stream_id = __pyx_state[7]; __pyx_result.sync = __pyx_state[8]; __pyx_result.timeout_handle = __pyx_state[9]; __pyx_result.waiter = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_31__pyx_unpickle_RollbackRequest, 0, __pyx_n_s_pyx_unpickle_RollbackRequest, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__206)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_RollbackRequest, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IProtoErrorStackFrame(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_33__pyx_unpickle_IProtoErrorStackFrame, 0, __pyx_n_s_pyx_unpickle_IProtoErrorStackF, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__207)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IProtoErrorStackF, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_IProtoErrorStackFrame__set_state(<IProtoErrorStackFrame> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IProtoErrorStackFrame__set_state(IProtoErrorStackFrame __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.code = __pyx_state[0]; __pyx_result.err_no = __pyx_state[1]; __pyx_result.error_type = __pyx_state[2]; __pyx_result.fields = __pyx_state[3]; __pyx_result.file = __pyx_state[4]; __pyx_result.line = __pyx_state[5]; __pyx_result.message = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_35__pyx_unpickle_IProtoError, 0, __pyx_n_s_pyx_unpickle_IProtoError, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__208)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IProtoError, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PushIterator(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_37__pyx_unpickle_PushIterator, 0, __pyx_n_s_pyx_unpickle_PushIterator, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__209)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PushIterator, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_PushIterator__set_state(<PushIterator> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PushIterator__set_state(PushIterator __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._fut = __pyx_state[0]; __pyx_result._request = __pyx_state[1]; __pyx_result._response = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8asynctnt_6iproto_8protocol_39__pyx_unpickle_CoreProtocol, 0, __pyx_n_s_pyx_unpickle_CoreProtocol, NULL, __pyx_n_s_asynctnt_iproto_protocol, __pyx_d, ((PyObject *)__pyx_codeobj__210)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CoreProtocol, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "asynctnt/iproto/protocol.pyx":1
 * # cython: language_level=3             # <<<<<<<<<<<<<<
 * 
 * cimport cpython.dict
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_kp_u_PushIterator___init___line_9, __pyx_kp_u_Creates_PushIterator_object_In) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_5) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init asynctnt.iproto.protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init asynctnt.iproto.protocol");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__2);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;  // error
            return kwvalues[i];
        }
    }
    return NULL;  // not found (no exception set)
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value); // transfer ownership of value to values
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;  // ownership transferred to values
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL; // ownership transferred to values
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* UnicodeAsUCS4 */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject* x) {
   Py_ssize_t length;
   #if CYTHON_PEP393_ENABLED
   length = PyUnicode_GET_LENGTH(x);
   if (likely(length == 1)) {
       return PyUnicode_READ_CHAR(x, 0);
   }
   #else
   length = PyUnicode_GET_SIZE(x);
   if (likely(length == 1)) {
       return PyUnicode_AS_UNICODE(x)[0];
   }
   #if Py_UNICODE_SIZE == 2
   else if (PyUnicode_GET_SIZE(x) == 2) {
       Py_UCS4 high_val = PyUnicode_AS_UNICODE(x)[0];
       if (high_val >= 0xD800 && high_val <= 0xDBFF) {
           Py_UCS4 low_val = PyUnicode_AS_UNICODE(x)[1];
           if (low_val >= 0xDC00 && low_val <= 0xDFFF) {
               return 0x10000 + (((high_val & ((1<<10)-1)) << 10) | (low_val & ((1<<10)-1)));
           }
       }
   }
   #endif
   #endif
   PyErr_Format(PyExc_ValueError,
                "only single character unicode strings can be converted to Py_UCS4, "
                "got length %" CYTHON_FORMAT_SSIZE_T "d", length);
   return (Py_UCS4)-1;
}

/* object_ord */
static long __Pyx__PyObject_Ord(PyObject* c) {
    Py_ssize_t size;
    if (PyBytes_Check(c)) {
        size = PyBytes_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyBytes_AS_STRING(c)[0];
        }
#if PY_MAJOR_VERSION < 3
    } else if (PyUnicode_Check(c)) {
        return (long)__Pyx_PyUnicode_AsPy_UCS4(c);
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    } else if (PyByteArray_Check(c)) {
        size = PyByteArray_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyByteArray_AS_STRING(c)[0];
        }
#endif
    } else {
        __Pyx_TypeName c_type_name = __Pyx_PyType_GetName(Py_TYPE(c));
        PyErr_Format(PyExc_TypeError,
            "ord() expected string of length 1, but " __Pyx_FMT_TYPENAME " found",
            c_type_name);
        __Pyx_DECREF_TypeName(c_type_name);
        return (long)(Py_UCS4)-1;
    }
    PyErr_Format(PyExc_TypeError,
        "ord() expected a character, but string of length %zd found", size);
    return (long)(Py_UCS4)-1;
}

/* RaiseClosureNameError */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* UnicodeConcatInPlace */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
static int
__Pyx_unicode_modifiable(PyObject *unicode)
{
    if (Py_REFCNT(unicode) != 1)
        return 0;
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        int ret;
        __Pyx_GIVEREF(*p_left);
        ret = PyUnicode_Resize(p_left, new_len);
        __Pyx_GOTREF(*p_left);
        if (unlikely(ret != 0))
            return NULL;
        #if PY_VERSION_HEX >= 0x030d0000
        if (unlikely(PyUnicode_CopyCharacters(*p_left, left_len, right, 0, right_len) < 0)) return NULL;
        #else
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        #endif
        __Pyx_INCREF(*p_left);
        __Pyx_GIVEREF(*p_left);
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSsize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_enum__mp_type(enum mp_type value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(enum mp_type)*3+2];
    char *dpos, *end = digits + sizeof(enum mp_type)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    enum mp_type remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mp_type neg_one = (enum mp_type) -1, const_zero = (enum mp_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (enum mp_type) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (enum mp_type) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (enum mp_type) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* StopAsyncIteration */
#if PY_VERSION_HEX < 0x030500B1
#if CYTHON_USE_TYPE_SPECS
#error Using async coroutines with type specs requires Python 3.5 or later.
#else
static PyTypeObject __Pyx__PyExc_StopAsyncIteration_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "StopAsyncIteration",
    sizeof(PyBaseExceptionObject),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC,
    PyDoc_STR("Signal the end from iterator.__anext__()."),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
    0,
#endif
};
#endif
#endif
static int __pyx_StopAsyncIteration_init(PyObject *module) {
    CYTHON_UNUSED_VAR(module);
#if PY_VERSION_HEX >= 0x030500B1
    __Pyx_PyExc_StopAsyncIteration = PyExc_StopAsyncIteration;
#else
    PyObject *builtins = PyEval_GetBuiltins();
    if (likely(builtins)) {
        PyObject *exc = PyMapping_GetItemString(builtins, (char*) "StopAsyncIteration");
        if (exc) {
            __Pyx_PyExc_StopAsyncIteration = exc;
            return 0;
        }
    }
    PyErr_Clear();
    __Pyx__PyExc_StopAsyncIteration_type.tp_traverse = ((PyTypeObject*)PyExc_BaseException)->tp_traverse;
    __Pyx__PyExc_StopAsyncIteration_type.tp_clear = ((PyTypeObject*)PyExc_BaseException)->tp_clear;
    __Pyx__PyExc_StopAsyncIteration_type.tp_dictoffset = ((PyTypeObject*)PyExc_BaseException)->tp_dictoffset;
    __Pyx__PyExc_StopAsyncIteration_type.tp_base = (PyTypeObject*)PyExc_Exception;
    __Pyx_PyExc_StopAsyncIteration = (PyObject*) __Pyx_FetchCommonType(&__Pyx__PyExc_StopAsyncIteration_type);
    if (unlikely(!__Pyx_PyExc_StopAsyncIteration))
        return -1;
    if (likely(builtins) && unlikely(PyMapping_SetItemString(builtins, (char*) "StopAsyncIteration", __Pyx_PyExc_StopAsyncIteration) < 0))
        return -1;
#endif
    return 0;
}

/* ReturnWithStopIteration */
static void __Pyx__ReturnWithStopIteration(PyObject* value) {
    PyObject *exc, *args;
#if CYTHON_COMPILING_IN_CPYTHON
    __Pyx_PyThreadState_declare
    if (PY_VERSION_HEX >= 0x030C00A6
            || unlikely(PyTuple_Check(value) || PyExceptionInstance_Check(value))) {
        args = PyTuple_New(1);
        if (unlikely(!args)) return;
        Py_INCREF(value);
        PyTuple_SET_ITEM(args, 0, value);
        exc = PyType_Type.tp_call(PyExc_StopIteration, args, NULL);
        Py_DECREF(args);
        if (!exc) return;
    } else {
        Py_INCREF(value);
        exc = value;
    }
    #if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    #if CYTHON_USE_EXC_INFO_STACK
    if (!__pyx_tstate->exc_info->exc_value)
    #else
    if (!__pyx_tstate->exc_type)
    #endif
    {
        Py_INCREF(PyExc_StopIteration);
        __Pyx_ErrRestore(PyExc_StopIteration, exc, NULL);
        return;
    }
    #endif
#else
    args = PyTuple_Pack(1, value);
    if (unlikely(!args)) return;
    exc = PyObject_Call(PyExc_StopIteration, args, NULL);
    Py_DECREF(args);
    if (unlikely(!exc)) return;
#endif
    PyErr_SetObject(PyExc_StopIteration, exc);
    Py_DECREF(exc);
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Coroutine */
static void __Pyx_CoroutineAwait_dealloc(PyObject *self) {
    PyObject_GC_UnTrack(self);
    Py_CLEAR(((__pyx_CoroutineAwaitObject*)self)->coroutine);
    __Pyx_PyHeapTypeObject_GC_Del(self);
}
static int __Pyx_CoroutineAwait_traverse(__pyx_CoroutineAwaitObject *self, visitproc visit, void *arg) {
    Py_VISIT(self->coroutine);
    return 0;
}
static int __Pyx_CoroutineAwait_clear(__pyx_CoroutineAwaitObject *self) {
    Py_CLEAR(self->coroutine);
    return 0;
}
static PyObject *__Pyx_CoroutineAwait_Next(__pyx_CoroutineAwaitObject *self) {
    return __Pyx_Generator_Next(self->coroutine);
}
static PyObject *__Pyx_CoroutineAwait_Send(__pyx_CoroutineAwaitObject *self, PyObject *value) {
    return __Pyx_Coroutine_Send(self->coroutine, value);
}
static PyObject *__Pyx_CoroutineAwait_Throw(__pyx_CoroutineAwaitObject *self, PyObject *args) {
    return __Pyx_Coroutine_Throw(self->coroutine, args);
}
static PyObject *__Pyx_CoroutineAwait_Close(__pyx_CoroutineAwaitObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self->coroutine);
}
static PyObject *__Pyx_CoroutineAwait_self(PyObject *self) {
    Py_INCREF(self);
    return self;
}
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_CoroutineAwait_no_new(PyTypeObject *type, PyObject *args, PyObject *kwargs) {
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(args);
    CYTHON_UNUSED_VAR(kwargs);
    PyErr_SetString(PyExc_TypeError, "cannot instantiate type, use 'await coroutine' instead");
    return NULL;
}
#endif
static PyObject *__Pyx_CoroutineAwait_reduce_ex(__pyx_CoroutineAwaitObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    PyErr_Format(PyExc_TypeError, "cannot pickle '%.200s' object",
                         Py_TYPE(self)->tp_name);
    return NULL;
}
static PyMethodDef __pyx_CoroutineAwait_methods[] = {
    {"send", (PyCFunction) __Pyx_CoroutineAwait_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into coroutine,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_CoroutineAwait_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in coroutine,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_CoroutineAwait_Close, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside coroutine.")},
    {"__reduce_ex__", (PyCFunction) __Pyx_CoroutineAwait_reduce_ex, METH_O, 0},
    {"__reduce__", (PyCFunction) __Pyx_CoroutineAwait_reduce_ex, METH_NOARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CoroutineAwaitType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CoroutineAwait_dealloc},
    {Py_tp_traverse, (void *)__Pyx_CoroutineAwait_traverse},
    {Py_tp_clear, (void *)__Pyx_CoroutineAwait_clear},
#if !CYTHON_COMPILING_IN_PYPY
    {Py_tp_new, (void *)__Pyx_CoroutineAwait_no_new},
#endif
    {Py_tp_methods, (void *)__pyx_CoroutineAwait_methods},
    {Py_tp_iter, (void *)__Pyx_CoroutineAwait_self},
    {Py_tp_iternext, (void *)__Pyx_CoroutineAwait_Next},
    {0, 0},
};
static PyType_Spec __pyx_CoroutineAwaitType_spec = {
    __PYX_TYPE_MODULE_PREFIX "coroutine_wrapper",
    sizeof(__pyx_CoroutineAwaitObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    __pyx_CoroutineAwaitType_slots
};
#else
static PyTypeObject __pyx_CoroutineAwaitType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "coroutine_wrapper",
    sizeof(__pyx_CoroutineAwaitObject),
    0,
    (destructor) __Pyx_CoroutineAwait_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    PyDoc_STR("A wrapper object implementing __await__ for coroutines."),
    (traverseproc) __Pyx_CoroutineAwait_traverse,
    (inquiry) __Pyx_CoroutineAwait_clear,
    0,
    0,
    __Pyx_CoroutineAwait_self,
    (iternextfunc) __Pyx_CoroutineAwait_Next,
    __pyx_CoroutineAwait_methods,
    0                         ,
    0                      ,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if !CYTHON_COMPILING_IN_PYPY
    __Pyx_CoroutineAwait_no_new,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
#if PY_VERSION_HEX < 0x030500B1 || defined(__Pyx_IterableCoroutine_USED) || CYTHON_USE_ASYNC_SLOTS
static CYTHON_INLINE PyObject *__Pyx__Coroutine_await(PyObject *coroutine) {
    __pyx_CoroutineAwaitObject *await = PyObject_GC_New(__pyx_CoroutineAwaitObject, __pyx_CoroutineAwaitType);
    if (unlikely(!await)) return NULL;
    Py_INCREF(coroutine);
    await->coroutine = coroutine;
    PyObject_GC_Track(await);
    return (PyObject*)await;
}
#endif
#if PY_VERSION_HEX < 0x030500B1
static PyObject *__Pyx_Coroutine_await_method(PyObject *coroutine, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx__Coroutine_await(coroutine);
}
#endif
#if defined(__Pyx_IterableCoroutine_USED) || CYTHON_USE_ASYNC_SLOTS
static PyObject *__Pyx_Coroutine_await(PyObject *coroutine) {
    if (unlikely(!coroutine || !__Pyx_Coroutine_Check(coroutine))) {
        PyErr_SetString(PyExc_TypeError, "invalid input, expected coroutine");
        return NULL;
    }
    return __Pyx__Coroutine_await(coroutine);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
static PyObject *__Pyx_Coroutine_compare(PyObject *obj, PyObject *other, int op) {
    PyObject* result;
    switch (op) {
        case Py_EQ: result = (other == obj) ? Py_True : Py_False; break;
        case Py_NE: result = (other != obj) ? Py_True : Py_False; break;
        default:
            result = Py_NotImplemented;
    }
    Py_INCREF(result);
    return result;
}
#endif
static PyMethodDef __pyx_Coroutine_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into coroutine,\nreturn next iterated value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in coroutine,\nreturn next iterated value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside coroutine.")},
#if PY_VERSION_HEX < 0x030500B1
    {"__await__", (PyCFunction) __Pyx_Coroutine_await_method, METH_NOARGS,
     (char*) PyDoc_STR("__await__() -> return an iterator to be used in await expression.")},
#endif
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Coroutine_memberlist[] = {
    {(char *) "cr_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "cr_await", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being awaited, or None")},
    {(char*) "cr_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Coroutine_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the coroutine"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the coroutine"), 0},
    {(char *) "cr_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the coroutine"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CoroutineType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_am_await, (void *)&__Pyx_Coroutine_await},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_methods, (void *)__pyx_Coroutine_methods},
    {Py_tp_members, (void *)__pyx_Coroutine_memberlist},
    {Py_tp_getset, (void *)__pyx_Coroutine_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_CoroutineType_spec = {
    __PYX_TYPE_MODULE_PREFIX "coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_CoroutineType_slots
};
#else
#if CYTHON_USE_ASYNC_SLOTS
static __Pyx_PyAsyncMethodsStruct __pyx_Coroutine_as_async = {
    __Pyx_Coroutine_await,
    0,
    0,
#if PY_VERSION_HEX >= 0x030A00A3
    0,
#endif
};
#endif
static PyTypeObject __pyx_CoroutineType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
#if CYTHON_USE_ASYNC_SLOTS
    &__pyx_Coroutine_as_async,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
#if CYTHON_USE_ASYNC_SLOTS && CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
    __Pyx_Coroutine_compare,
#else
    0,
#endif
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    0,
    __pyx_Coroutine_methods,
    __pyx_Coroutine_memberlist,
    __pyx_Coroutine_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Coroutine_init(PyObject *module) {
    CYTHON_MAYBE_UNUSED_VAR(module);
#if CYTHON_USE_TYPE_SPECS
    __pyx_CoroutineType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CoroutineType_spec, NULL);
#else
    __pyx_CoroutineType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_CoroutineType = __Pyx_FetchCommonType(&__pyx_CoroutineType_type);
#endif
    if (unlikely(!__pyx_CoroutineType))
        return -1;
#ifdef __Pyx_IterableCoroutine_USED
    if (unlikely(__pyx_IterableCoroutine_init(module) == -1))
        return -1;
#endif
#if CYTHON_USE_TYPE_SPECS
    __pyx_CoroutineAwaitType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CoroutineAwaitType_spec, NULL);
#else
    __pyx_CoroutineAwaitType = __Pyx_FetchCommonType(&__pyx_CoroutineAwaitType_type);
#endif
    if (unlikely(!__pyx_CoroutineAwaitType))
        return -1;
    return 0;
}

/* GetAwaitIter */
static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAwaitableIter(PyObject *o) {
#ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(o)) {
        return __Pyx_NewRef(o);
    }
#endif
    return __Pyx__Coroutine_GetAwaitableIter(o);
}
static void __Pyx_Coroutine_AwaitableIterError(PyObject *source) {
#if PY_VERSION_HEX >= 0x030600B3 && PY_VERSION_HEX < 0x030d0000 || defined(_PyErr_FormatFromCause)
    __Pyx_TypeName source_type_name = __Pyx_PyType_GetName(Py_TYPE(source));
    _PyErr_FormatFromCause(PyExc_TypeError,
        "'async for' received an invalid object from __anext__: " __Pyx_FMT_TYPENAME, source_type_name);
    __Pyx_DECREF_TypeName(source_type_name);
#elif PY_MAJOR_VERSION >= 3
    PyObject *exc, *val, *val2, *tb;
    __Pyx_TypeName source_type_name = __Pyx_PyType_GetName(Py_TYPE(source));
    assert(PyErr_Occurred());
    PyErr_Fetch(&exc, &val, &tb);
    PyErr_NormalizeException(&exc, &val, &tb);
    if (tb != NULL) {
        PyException_SetTraceback(val, tb);
        Py_DECREF(tb);
    }
    Py_DECREF(exc);
    assert(!PyErr_Occurred());
    PyErr_Format(PyExc_TypeError,
        "'async for' received an invalid object from __anext__: " __Pyx_FMT_TYPENAME, source_type_name);
    __Pyx_DECREF_TypeName(source_type_name);
    PyErr_Fetch(&exc, &val2, &tb);
    PyErr_NormalizeException(&exc, &val2, &tb);
    Py_INCREF(val);
    PyException_SetCause(val2, val);
    PyException_SetContext(val2, val);
    PyErr_Restore(exc, val2, tb);
#else
    CYTHON_UNUSED_VAR(source);
#endif
}
static PyObject *__Pyx__Coroutine_GetAwaitableIter(PyObject *obj) {
    PyObject *res;
#if CYTHON_USE_ASYNC_SLOTS
    __Pyx_PyAsyncMethodsStruct* am = __Pyx_PyType_AsAsync(obj);
    if (likely(am && am->am_await)) {
        res = (*am->am_await)(obj);
    } else
#endif
#if PY_VERSION_HEX >= 0x030500B2 || defined(PyCoro_CheckExact)
    if (PyCoro_CheckExact(obj)) {
        return __Pyx_NewRef(obj);
    } else
#endif
#if CYTHON_COMPILING_IN_CPYTHON && defined(CO_ITERABLE_COROUTINE)
#if PY_VERSION_HEX >= 0x030C00A6
    if (PyGen_CheckExact(obj) && (PyGen_GetCode((PyGenObject*)obj)->co_flags & CO_ITERABLE_COROUTINE)) {
#else
    if (PyGen_CheckExact(obj) && ((PyGenObject*)obj)->gi_code && ((PyCodeObject *)((PyGenObject*)obj)->gi_code)->co_flags & CO_ITERABLE_COROUTINE) {
#endif
        return __Pyx_NewRef(obj);
    } else
#endif
    {
        PyObject *method = NULL;
        int is_method = __Pyx_PyObject_GetMethod(obj, __pyx_n_s_await, &method);
        if (likely(is_method)) {
            res = __Pyx_PyObject_CallOneArg(method, obj);
        } else if (likely(method)) {
            res = __Pyx_PyObject_CallNoArg(method);
        } else
            goto slot_error;
        Py_DECREF(method);
    }
    if (unlikely(!res)) {
        __Pyx_Coroutine_AwaitableIterError(obj);
        goto bad;
    }
    if (unlikely(!PyIter_Check(res))) {
        __Pyx_TypeName res_type_name = __Pyx_PyType_GetName(Py_TYPE(res));
        PyErr_Format(PyExc_TypeError,
            "__await__() returned non-iterator of type '" __Pyx_FMT_TYPENAME "'", res_type_name);
        __Pyx_DECREF_TypeName(res_type_name);
        Py_CLEAR(res);
    } else {
        int is_coroutine = 0;
        #ifdef __Pyx_Coroutine_USED
        is_coroutine |= __Pyx_Coroutine_Check(res);
        #endif
        #if PY_VERSION_HEX >= 0x030500B2 || defined(PyCoro_CheckExact)
        is_coroutine |= PyCoro_CheckExact(res);
        #endif
        if (unlikely(is_coroutine)) {
            /* __await__ must return an *iterator*, not
               a coroutine or another awaitable (see PEP 492) */
            PyErr_SetString(PyExc_TypeError,
                            "__await__() returned a coroutine");
            Py_CLEAR(res);
        }
    }
    return res;
slot_error:
    {
        __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
        PyErr_Format(PyExc_TypeError,
            "object " __Pyx_FMT_TYPENAME " can't be used in 'await' expression", obj_type_name);
        __Pyx_DECREF_TypeName(obj_type_name);
    }
bad:
    return NULL;
}

/* CoroutineYieldFrom */
  static PyObject* __Pyx__Coroutine_Yield_From_Generic(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *retval;
    PyObject *source_gen = __Pyx__Coroutine_GetAwaitableIter(source);
    if (unlikely(!source_gen)) {
        return NULL;
    }
    if (__Pyx_Coroutine_Check(source_gen)) {
        retval = __Pyx_Generator_Next(source_gen);
    } else {
        retval = __Pyx_PyObject_GetIterNextFunc(source_gen)(source_gen);
    }
    if (retval) {
        gen->yieldfrom = source_gen;
        return retval;
    }
    Py_DECREF(source_gen);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *retval;
    if (__Pyx_Coroutine_Check(source)) {
        if (unlikely(((__pyx_CoroutineObject*)source)->yieldfrom)) {
            PyErr_SetString(
                PyExc_RuntimeError,
                "coroutine is being awaited already");
            return NULL;
        }
        retval = __Pyx_Generator_Next(source);
#ifdef __Pyx_AsyncGen_USED
    } else if (__pyx_PyAsyncGenASend_CheckExact(source)) {
        retval = __Pyx_async_gen_asend_iternext(source);
#endif
    } else {
        return __Pyx__Coroutine_Yield_From_Generic(gen, source);
    }
    if (retval) {
        Py_INCREF(source);
        gen->yieldfrom = source;
    }
    return retval;
}

/* PyObjectSetAttrStr */
  #if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a + (unsigned long)b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op2);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* ValidateBasesTuple */
  #if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
  static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetupReduce */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
  static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType_3_0_7
#define __PYX_HAVE_RT_ImportType_3_0_7
static PyTypeObject *__Pyx_ImportType_3_0_7(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_7 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_7 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_7 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ImportDottedModule */
  #if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__65;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* PatchInspect */
  static PyObject* __Pyx_patch_inspect(PyObject* module) {
#if defined(__Pyx_Generator_USED) && (!defined(CYTHON_PATCH_INSPECT) || CYTHON_PATCH_INSPECT)
    static int inspect_patched = 0;
    if (unlikely((!inspect_patched) && module)) {
        module = __Pyx_Coroutine_patch_module(
            module, ""
"old_types = getattr(_module.isgenerator, '_cython_generator_types', None)\n"
"if old_types is None or not isinstance(old_types, set):\n"
"    old_types = set()\n"
"    def cy_wrap(orig_func, type=type, cython_generator_types=old_types):\n"
"        def cy_isgenerator(obj): return type(obj) in cython_generator_types or orig_func(obj)\n"
"        cy_isgenerator._cython_generator_types = cython_generator_types\n"
"        return cy_isgenerator\n"
"    _module.isgenerator = cy_wrap(_module.isgenerator)\n"
"old_types.add(_cython_generator_type)\n"
        );
        inspect_patched = 1;
    }
#else
    if ((0)) return __Pyx_Coroutine_patch_module(module, NULL);
#endif
    return module;
}

/* PatchAsyncIO */
  static PyObject* __Pyx_patch_asyncio(PyObject* module) {
#if PY_VERSION_HEX < 0x030500B2 &&\
        (defined(__Pyx_Coroutine_USED) || defined(__Pyx_Generator_USED)) &&\
        (!defined(CYTHON_PATCH_ASYNCIO) || CYTHON_PATCH_ASYNCIO)
    PyObject *patch_module = NULL;
    static int asyncio_patched = 0;
    if (unlikely((!asyncio_patched) && module)) {
        PyObject *package;
        package = __Pyx_Import(__pyx_n_s_asyncio_coroutines, NULL, 0);
        if (package) {
            patch_module = __Pyx_Coroutine_patch_module(
                PyObject_GetAttrString(package, "coroutines"), ""
"try:\n"
"    coro_types = _module._COROUTINE_TYPES\n"
"except AttributeError: pass\n"
"else:\n"
"    if _cython_coroutine_type is not None and _cython_coroutine_type not in coro_types:\n"
"        coro_types = tuple(coro_types) + (_cython_coroutine_type,)\n"
"    if _cython_generator_type is not None and _cython_generator_type not in coro_types:\n"
"        coro_types = tuple(coro_types) + (_cython_generator_type,)\n"
"_module._COROUTINE_TYPES = coro_types\n"
            );
        } else {
            PyErr_Clear();
            package = __Pyx_Import(__pyx_n_s_asyncio_tasks, NULL, 0);
            if (unlikely(!package)) goto asyncio_done;
            patch_module = __Pyx_Coroutine_patch_module(
                PyObject_GetAttrString(package, "tasks"), ""
"if hasattr(_module, 'iscoroutine'):\n"
"    old_types = getattr(_module.iscoroutine, '_cython_coroutine_types', None)\n"
"    if old_types is None or not isinstance(old_types, set):\n"
"        old_types = set()\n"
"        def cy_wrap(orig_func, type=type, cython_coroutine_types=old_types):\n"
"            def cy_iscoroutine(obj): return type(obj) in cython_coroutine_types or orig_func(obj)\n"
"            cy_iscoroutine._cython_coroutine_types = cython_coroutine_types\n"
"            return cy_iscoroutine\n"
"        _module.iscoroutine = cy_wrap(_module.iscoroutine)\n"
"    if _cython_coroutine_type is not None:\n"
"        old_types.add(_cython_coroutine_type)\n"
"    if _cython_generator_type is not None:\n"
"        old_types.add(_cython_generator_type)\n"
            );
        }
        Py_DECREF(package);
        if (unlikely(!patch_module)) goto ignore;
asyncio_done:
        PyErr_Clear();
        asyncio_patched = 1;
#ifdef __Pyx_Generator_USED
        {
            PyObject *inspect_module;
            if (patch_module) {
                inspect_module = PyObject_GetAttr(patch_module, __pyx_n_s_inspect);
                Py_DECREF(patch_module);
            } else {
                inspect_module = __Pyx_Import(__pyx_n_s_inspect, NULL, 0);
            }
            if (unlikely(!inspect_module)) goto ignore;
            inspect_module = __Pyx_patch_inspect(inspect_module);
            if (unlikely(!inspect_module)) {
                Py_DECREF(module);
                module = NULL;
            }
            Py_XDECREF(inspect_module);
        }
#else
        if ((0)) return __Pyx_patch_inspect(module);
#endif
    }
    return module;
ignore:
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch asyncio package with custom generator type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    if ((0)) return __Pyx_patch_inspect(__Pyx_Coroutine_patch_module(module, NULL));
#endif
    return module;
}

/* Py3UpdateBases */
  static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base, *meth, *new_base, *result, *new_bases = NULL;
    size_bases = PyTuple_GET_SIZE(bases);
    for (i = 0; i < size_bases; i++) {
        base  = PyTuple_GET_ITEM(bases, i);
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_n_s_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                base = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base);
                Py_INCREF(base);
            }
        }
        j = PyList_GET_SIZE(new_bases);
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
    return result;
error:
    Py_XDECREF(new_bases);
    return NULL;
}

/* CalculateMetaclass */
  static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases;
#if CYTHON_ASSUME_SAFE_MACROS
    nbases = PyTuple_GET_SIZE(bases);
#else
    nbases = PyTuple_Size(bases);
    if (nbases < 0) return NULL;
#endif
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
#if CYTHON_ASSUME_SAFE_MACROS
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *tmp = PyTuple_GetItem(bases, i);
        if (!tmp) return NULL;
#endif
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyObjectLookupSpecial */
  #if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return with_error ? __Pyx_PyObject_GetAttrStr(obj, attr_name) : __Pyx_PyObject_GetAttrStrNoError(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* Py3ClassCreate */
  static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_n_s_prepare_2);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
#if PY_VERSION_HEX >= 0x03030000
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
#else
    CYTHON_MAYBE_UNUSED_VAR(qualname);
#endif
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
static int __Pyx_SetNamesPEP487(PyObject *type_obj) {
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *names_to_set, *key, *value, *set_name, *tmp;
    Py_ssize_t i = 0;
#if CYTHON_USE_TYPE_SLOTS
    names_to_set = PyDict_Copy(type->tp_dict);
#else
    {
        PyObject *d = PyObject_GetAttr(type_obj, __pyx_n_s_dict);
        names_to_set = NULL;
        if (likely(d)) {
            PyObject *names_to_set = PyDict_New();
            int ret = likely(names_to_set) ? PyDict_Update(names_to_set, d) : -1;
            Py_DECREF(d);
            if (unlikely(ret < 0))
                Py_CLEAR(names_to_set);
        }
    }
#endif
    if (unlikely(names_to_set == NULL))
        goto bad;
    while (PyDict_Next(names_to_set, &i, &key, &value)) {
        set_name = __Pyx_PyObject_LookupSpecialNoError(value, __pyx_n_s_set_name);
        if (unlikely(set_name != NULL)) {
            tmp = __Pyx_PyObject_Call2Args(set_name, type_obj, key);
            Py_DECREF(set_name);
            if (unlikely(tmp == NULL)) {
                __Pyx_TypeName value_type_name =
                    __Pyx_PyType_GetName(Py_TYPE(value));
                __Pyx_TypeName type_name = __Pyx_PyType_GetName(type);
                PyErr_Format(PyExc_RuntimeError,
#if PY_MAJOR_VERSION >= 3
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %R " "in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name, key, type_name);
#else
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %.100s in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name,
                    PyString_Check(key) ? PyString_AS_STRING(key) : "?",
                    type_name);
#endif
                goto bad;
            } else {
                Py_DECREF(tmp);
            }
        }
        else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
    Py_DECREF(names_to_set);
    return 0;
bad:
    Py_XDECREF(names_to_set);
    return -1;
}
static PyObject *__Pyx_InitSubclassPEP487(PyObject *type_obj, PyObject *mkw) {
#if CYTHON_USE_TYPE_SLOTS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *mro = type->tp_mro;
    Py_ssize_t i, nbases;
    if (unlikely(!mro)) goto done;
    (void) &__Pyx_GetBuiltinName;
    Py_INCREF(mro);
    nbases = PyTuple_GET_SIZE(mro);
    assert(PyTuple_GET_ITEM(mro, 0) == type_obj);
    for (i = 1; i < nbases-1; i++) {
        PyObject *base, *dict, *meth;
        base = PyTuple_GET_ITEM(mro, i);
        dict = ((PyTypeObject *)base)->tp_dict;
        meth = __Pyx_PyDict_GetItemStrWithError(dict, __pyx_n_s_init_subclass);
        if (unlikely(meth)) {
            descrgetfunc f = Py_TYPE(meth)->tp_descr_get;
            PyObject *res;
            Py_INCREF(meth);
            if (likely(f)) {
                res = f(meth, NULL, type_obj);
                Py_DECREF(meth);
                if (unlikely(!res)) goto bad;
                meth = res;
            }
            res = __Pyx_PyObject_FastCallDict(meth, NULL, 0, mkw);
            Py_DECREF(meth);
            if (unlikely(!res)) goto bad;
            Py_DECREF(res);
            goto done;
        } else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
done:
    Py_XDECREF(mro);
    return type_obj;
bad:
    Py_XDECREF(mro);
    Py_DECREF(type_obj);
    return NULL;
#else
    PyObject *super_type, *super, *func, *res;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    super_type = __Pyx_GetBuiltinName(__pyx_n_s_super);
#else
    super_type = (PyObject*) &PySuper_Type;
    (void) &__Pyx_GetBuiltinName;
#endif
    super = likely(super_type) ? __Pyx_PyObject_Call2Args(super_type, type_obj, type_obj) : NULL;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    Py_XDECREF(super_type);
#endif
    if (unlikely(!super)) {
        Py_CLEAR(type_obj);
        goto done;
    }
    func = __Pyx_PyObject_GetAttrStrNoError(super, __pyx_n_s_init_subclass);
    Py_DECREF(super);
    if (likely(!func)) {
        if (unlikely(PyErr_Occurred()))
            Py_CLEAR(type_obj);
        goto done;
    }
    res = __Pyx_PyObject_FastCallDict(func, NULL, 0, mkw);
    Py_DECREF(func);
    if (unlikely(!res))
        Py_CLEAR(type_obj);
    Py_XDECREF(res);
done:
    return type_obj;
#endif
}
#endif
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET,
#if PY_VERSION_HEX < 0x030600A4
        (metaclass == (PyObject*)&PyType_Type) ? NULL : mkw
#else
        mkw
#endif
    );
    Py_XDECREF(owned_metaclass);
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
    if (likely(result) && likely(PyType_Check(result))) {
        if (unlikely(__Pyx_SetNamesPEP487(result) < 0)) {
            Py_CLEAR(result);
        } else {
            result = __Pyx_InitSubclassPEP487(result, mkw);
        }
    }
#else
    (void) &__Pyx_GetBuiltinName;
#endif
    return result;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type neg_one = (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) -1, const_zero = (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type __Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type neg_one = (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) -1, const_zero = (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) >= 2 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) (((((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) >= 3 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) (((((((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) >= 4 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) (((((((((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) - 1 > 2 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) (((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)-1)*(((((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) - 1 > 2 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) ((((((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) - 1 > 3 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) (((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)-1)*(((((((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) - 1 > 3 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) ((((((((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) - 1 > 4 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) (((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)-1)*(((((((((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) - 1 > 4 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) ((((((((((enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() not available, cannot convert large enums");
        return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) -1;
    } else {
        enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_9tarantool_iproto_type(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type");
    return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type");
    return (enum __pyx_t_8asynctnt_6iproto_9tarantool_iproto_type) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                                return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                                return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                                return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(int64_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(int64_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            int64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (int64_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (int64_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (int64_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (int64_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (int64_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((int64_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((int64_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (int64_t) -1;
        }
    } else {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT)) {
                                return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT)) {
                                return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT)) {
                                return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(uint64_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(uint64_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint64_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            uint64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (uint64_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (uint64_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (uint64_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (uint64_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (uint64_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(uint64_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((uint64_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(uint64_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((uint64_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((uint64_t) 1) << (sizeof(uint64_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (uint64_t) -1;
        }
    } else {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyInt_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint32_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT)) {
                                return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT)) {
                                return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT)) {
                                return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(uint32_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(uint32_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint32_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            uint32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (uint32_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (uint32_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (uint32_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (uint32_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (uint32_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(uint32_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((uint32_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(uint32_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((uint32_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((uint32_t) 1) << (sizeof(uint32_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (uint32_t) -1;
        }
    } else {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyInt_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                                return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                                return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                                return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(int) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(int) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (int) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (int) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (int) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (int) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (int) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((int) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((int) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_8protocol_ConnectionState(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState neg_one = (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) -1, const_zero = (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState __Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_8protocol_ConnectionState(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState neg_one = (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) -1, const_zero = (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) >= 2 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) (((((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) >= 3 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) (((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) >= 4 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) (((((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) - 1 > 2 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) (((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)-1)*(((((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) - 1 > 2 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) ((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) - 1 > 3 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) (((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)-1)*(((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) - 1 > 3 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) ((((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) - 1 > 4 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) (((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)-1)*(((((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) - 1 > 4 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) ((((((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() not available, cannot convert large enums");
        return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) -1;
    } else {
        enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_8protocol_ConnectionState(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState");
    return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState");
    return (enum __pyx_t_8asynctnt_6iproto_8protocol_ConnectionState) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_8asynctnt_6iproto_8protocol_ProtocolState(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState neg_one = (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) -1, const_zero = (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState __Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_8protocol_ProtocolState(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState neg_one = (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) -1, const_zero = (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) >= 2 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) (((((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) >= 3 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) (((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) >= 4 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) (((((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) - 1 > 2 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) (((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)-1)*(((((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) - 1 > 2 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) ((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) - 1 > 3 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) (((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)-1)*(((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) - 1 > 3 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) ((((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) - 1 > 4 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) (((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)-1)*(((((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) - 1 > 4 * PyLong_SHIFT)) {
                                return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) ((((((((((enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() not available, cannot convert large enums");
        return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) -1;
    } else {
        enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_8asynctnt_6iproto_8protocol_ProtocolState(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState");
    return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState");
    return (enum __pyx_t_8asynctnt_6iproto_8protocol_ProtocolState) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                                return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                                return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                                return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(long) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(long) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (long) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (long) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (long) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (long) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (long) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((long) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((long) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint8_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT)) {
                                return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT)) {
                                return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT)) {
                                return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(uint8_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(uint8_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint8_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            uint8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (uint8_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (uint8_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (uint8_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (uint8_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (uint8_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(uint8_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((uint8_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(uint8_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((uint8_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((uint8_t) 1) << (sizeof(uint8_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (uint8_t) -1;
        }
    } else {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyInt_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__mp_type(enum mp_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mp_type neg_one = (enum mp_type) -1, const_zero = (enum mp_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mp_type) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum mp_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum mp_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mp_type) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum mp_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(enum mp_type),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mp_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int8_t(int8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int8_t neg_one = (int8_t) -1, const_zero = (int8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int8_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int8_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(int8_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int8_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(size_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                                return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                                return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                                return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(size_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(size_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (size_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (size_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (size_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (size_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (size_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((size_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((size_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(int32_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
  #if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__211);
    }
    return name;
}
#endif

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* Generator */
  static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_GeneratorType_slots
};
#else
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Generator_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_GeneratorType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
#endif
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* IterableCoroutine */
  #if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_IterableCoroutineType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_am_await, (void *)&__Pyx_Coroutine_await},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)__Pyx_Coroutine_await},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Coroutine_methods},
    {Py_tp_members, (void *)__pyx_Coroutine_memberlist},
    {Py_tp_getset, (void *)__pyx_Coroutine_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_IterableCoroutineType_spec = {
    __PYX_TYPE_MODULE_PREFIX "iterable_coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_IterableCoroutineType_slots
};
#else
static PyTypeObject __pyx_IterableCoroutineType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "iterable_coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
#if CYTHON_USE_ASYNC_SLOTS
    &__pyx_Coroutine_as_async,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
#if CYTHON_USE_ASYNC_SLOTS && CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
    __Pyx_Coroutine_compare,
#else
    0,
#endif
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    __Pyx_Coroutine_await,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Coroutine_methods,
    __pyx_Coroutine_memberlist,
    __pyx_Coroutine_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if PY_VERSION_HEX >= 0x030400a1 && !CYTHON_COMPILING_IN_PYPY
    __Pyx_Coroutine_del,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_IterableCoroutine_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_IterableCoroutineType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_IterableCoroutineType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_IterableCoroutineType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_IterableCoroutineType = __Pyx_FetchCommonType(&__pyx_IterableCoroutineType_type);
#endif
    if (unlikely(!__pyx_IterableCoroutineType))
        return -1;
    return 0;
}

/* CheckBinaryVersion */
  static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* FunctionExport */
  static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
    cobj = PyCapsule_New(tmp.p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
  #if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
